
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "canon_gamut.h"
#include "illum_data.h"
#include "spectra.h"
#include "synthetic_data.h"

/* -------------------------------------------------------------------------- */

#define CANON_GAMUT_FILE      "canon_gamut.rgb"

/* -------------------------------------------------------------------------- */

typedef struct Canon_gamut 
{
    Matrix* mp;
    int     version;
    int     world_version;
    int     canon_data_version;
    int     canon_illum_version;
    int     illum_spectra_version;
    int     specular_rgb_version; 

    char    specular_rgb_config_file[ MAX_FILE_NAME_SIZE ]; 
    int     combine_multiple_hulls; 
    double  specular_factor;
    int     specular_white_enable; 

    char    config_file[ MAX_FILE_NAME_SIZE ];
    World_info world_info; 
}
Canon_gamut; 

/* -------------------------------------------------------------------------- */

/* Note that we rely on the fact that statics are initialized to all zeros. */
static Canon_gamut* fs_cached_canon_gamuts[ MAX_NUM_CANON_GAMUTS ];

static int     fs_specular_white_enable  = TRUE;
static double  fs_specular_factor        = DBL_NOT_SET;
static int     fs_combine_multiple_hulls = FALSE;
static Matrix* fs_cached_specular_rgb_mp = NULL;
static char    fs_specular_rgb_config_file[ MAX_FILE_NAME_SIZE ];
static int     fs_canon_specular_rgb_version = 0;

/* -------------------------------------------------------------------------- */

static int echo_canon_gamut_info(Canon_gamut* cg_ptr);

static int make_multi_canon_gamut_variant
(
    int         index,
    const char* file_name 
);

static int make_canon_gamut_variant(int index, const char* file_name);

static int write_canon_gamut(Canon_gamut* cg_ptr, const char* file_name);

static int write_canon_gamut_info(FILE* fp, Canon_gamut* cg_ptr);

static int read_canon_gamut_from_file
(
    Canon_gamut** cg_ptr_ptr,
    const char*   file_name 
);

static int read_canon_gamut_info(FILE* fp, Canon_gamut* cg_ptr);

static int get_canon_specular_rgb(Matrix** rgb_mpp);

static int get_canon_gamut(Matrix** target_mpp, int canon_gamut_num);

static void invalidate_canon_gamut(int index);

static Canon_gamut* create_canon_gamut(void);
static void free_canon_gamut(Canon_gamut* cg_ptr);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_canon_gamut_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  i, index;
    int  first_index;
    int  second_index;
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);
    
    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "specular-rgb-file")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-rgb-file = %s\n",
                    fs_specular_rgb_config_file[ 0 ] ? fs_specular_rgb_config_file
                                                  : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_specular_rgb_mp == NULL) 
            {
                ERE(pso("No specular reflectance database has been set.\n"));
            }
            else 
            {
                ERE(pso("Specular reflectance spectra database source is %s.\n", 
                        fs_specular_rgb_config_file));
            }
        }
        else
        {
#ifdef TRACK_MEMORY_ALLOCATION
            prepare_memory_cleanup();
#endif 
            ERE(read_matrix_from_gamut_config_file(&fs_cached_specular_rgb_mp,
                                                   (const char*)NULL, 
                                                   value, "specular RGB", 
                                                   fs_specular_rgb_config_file,
                                           sizeof(fs_specular_rgb_config_file)));

            fs_canon_specular_rgb_version++; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "specular-white-enable")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-white-enable = %s\n",
                    fs_specular_white_enable ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Specular canonical hulls %s white.\n",
                    fs_specular_white_enable ? "always includes" 
                                          : "does not necessarily include"));
        }
        else
        {
            int temp_boolean_value; 

            ERE(temp_boolean_value = get_boolean_value(value));
            fs_specular_white_enable = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "specular-factor"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (IS_NOT_NEGATIVE_DBL(fs_specular_factor))
            {
                ERE(pso("specular-factor = %.1f\n", fs_specular_factor));
            }
            else 
            {
                ERE(pso("specular-factor = off\n"));
            }
        }
        else if (value[ 0 ] == '\0') 
        {
            if (IS_NOT_NEGATIVE_DBL(fs_specular_factor))
            {
                ERE(pso("Specular factor for building 3D canon hulls is %.1f.\n",
                        fs_specular_factor));
            }
            else 
            {
                ERE(pso("Specular factor for building 3D canon hulls is not used.\n"));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_specular_factor = DBL_NOT_SET; 
        }
        else
        {
            double temp_double; 

            ERE(ss1d(value, &temp_double)); 
            fs_specular_factor = temp_double;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "combine-multiple-hulls")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("combine-multiple-hulls = %s\n",
                    fs_combine_multiple_hulls ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Multiple hulls %s combined for cononical hull.\n",
                    fs_combine_multiple_hulls ? "are" : "are not"));
        }
        else
        {
            int temp_boolean_value; 

            ERE(temp_boolean_value = get_boolean_value(value));
            fs_combine_multiple_hulls = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    ERE(parse_index(lc_option, MAX_NUM_CANON_GAMUTS,
                    &first_index, &second_index, &index, (int*)NULL));

    if (    (lc_option[ 0 ] == '\0')
          || match_pattern(lc_option, "canon-gamut-file")
          || match_pattern(lc_option, "cgf")
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = NULL;
                
                if (fs_cached_canon_gamuts[ i ] != NULL)
                {
                    file_name = fs_cached_canon_gamuts[ i ]->config_file;
                }

                ERE(pso("canon-gamut-file:%d = %s\n",
                        i + 1, (file_name != NULL) ? file_name : "off"));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = NULL;
                
                if (fs_cached_canon_gamuts[ i ] != NULL)
                {
                    file_name = fs_cached_canon_gamuts[ i ]->config_file;
                }

                ERE(pso("Canon gamut %d source is %s\n",
                        i + 1, (file_name != NULL) ? file_name : "<none>"));

                if (fs_cached_canon_gamuts[ i ] != NULL)
                {
                    ERE(echo_world_info(&(fs_cached_canon_gamuts[ i ]->world_info))); 
                    ERE(echo_canon_gamut_info(fs_cached_canon_gamuts[ i ])); 
                }
            }
        }
        else if (is_no_value_word(value))
        {
            invalidate_canon_gamut(index); 
        }
        else
        {
#ifdef TRACK_MEMORY_ALLOCATION
            prepare_memory_cleanup();
#endif 
            ERE(read_canon_gamut_from_file(&(fs_cached_canon_gamuts[ index ]), 
                                             value)); 
        }
        result = NO_ERROR; 
    }
    
    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int echo_canon_gamut_info(Canon_gamut* cg_ptr)
{

    ERE(pso("     specular_rgb_config_file = %s\n",
            cg_ptr->specular_rgb_config_file[ 0 ] 
            ?   cg_ptr->specular_rgb_config_file  
            :   "<none>")); 

    ERE(pso("     combine_multiple_hulls = %s\n",
            cg_ptr->combine_multiple_hulls ? "t" : "f"));
 
    ERE(pso("     specular_white_enable = %s\n",
            cg_ptr->specular_white_enable ? "t" : "f"));
 
    if (cg_ptr->specular_factor < 0.0)
    {
        ERE(pso("     specular_factor = off\n"));
    }
    else 
    {
        ERE(pso("     specular_factor = %.1f\n", cg_ptr->specular_factor));
    }
 
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_canon_gamut(const char* make_directive, const char* file_name)
{
    char lc_make_directive[ 100 ];
    int  result = NOT_FOUND; 
    int  index; 

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    ERE(parse_index(lc_make_directive, MAX_NUM_CANON_GAMUTS, 
                    (int*)NULL, (int*)NULL, &index, (int*)NULL));

    if (    (STRCMP_EQ(lc_make_directive, "all"))
         || (STRCMP_EQ(lc_make_directive, "gamut"))
         || (STRCMP_EQ(lc_make_directive, "canon"))
         || (STRCMP_EQ(lc_make_directive, "canon-gamut"))
       )
    {
        ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, 
                              "Canon gamut"));

        if (fs_combine_multiple_hulls)
        {
            ERE(make_multi_canon_gamut_variant(index, file_name));
        }
        else 
        {
            ERE(make_canon_gamut_variant(index, file_name));
        }

        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int make_multi_canon_gamut_variant
(
    int         index,
    const char* file_name 
)
{
    int         illum_spectra_version;
    int         canon_illum_version;
    int         specular_rgb_version;
    int         result = NO_ERROR;
    Matrix*     specular_rgb_mp = NULL;
    int         num_illum;
    Vector*     canon_rgb_vp    = NULL;
    Canon_gamut* cg_ptr; 

    if (fs_cached_canon_gamuts[ index ] == NULL)
    {
#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 
        NRE(fs_cached_canon_gamuts[ index ] = create_canon_gamut());
    }

    cg_ptr = fs_cached_canon_gamuts[ index ]; 

    ERE(illum_spectra_version = get_num_illum_spectra(&num_illum));

    ERE(canon_illum_version = get_canon_illum(&canon_rgb_vp));

    if (fs_specular_factor > 0.0) 
    {
        ERE(specular_rgb_version = get_canon_specular_rgb(&specular_rgb_mp));
    }
    else
    {
        specular_rgb_version = cg_ptr->specular_rgb_version;
    }

    if (    (world_is_up_to_date(cg_ptr->world_version)) 
         && (cg_ptr->canon_illum_version == canon_illum_version)
         && (cg_ptr->illum_spectra_version ==illum_spectra_version)
         && (IS_EQUAL_DBL(cg_ptr->specular_factor, fs_specular_factor))
         && (cg_ptr->specular_rgb_version == specular_rgb_version)
         && (cg_ptr->combine_multiple_hulls == TRUE)
       )
    {
        verbose_pso(10, "Multi-canon gamut %d is up to date. Skipping make.\n",
                    index + 1); 
    }
    else 
    {
        Matrix*  data_mp;
        Matrix*  canon_data_mp                   = NULL;
        Matrix*  canon_data_with_specular_rgb_mp = NULL;
        Matrix*  canon_gamut_mp = NULL;
        Matrix** canon_gamut_matrix_vector;
        Vector*  actual_rgb_vp                   = NULL;
        Vector*  ratio_vp                        = NULL;
        Hull*    canon_gamut_hp                  = NULL;
        int      i;
        int      world_version = NOT_SET;

        verbose_pso(5, "Multi-canon gamut %d is out of date. Remaking it.\n",
                    index + 1);

        canon_gamut_matrix_vector = N_TYPE_MALLOC(Matrix*, num_illum);

        if (canon_gamut_matrix_vector == NULL) 
        {
            result = ERROR;
        }

        if (num_illum <= 0)
        {
            SET_CANT_HAPPEN_BUG();
            result = ERROR;
        }

        for (i=0; i<num_illum; i++)
        {
            canon_gamut_matrix_vector[ i ] = NULL;
        }

        for (i=0; i<num_illum; i++)
        {
            if (result == ERROR) break;

            data_mp = NULL; 
            world_version = get_synthetic_RGB(&data_mp, i); 

            if (world_version == ERROR) result = ERROR; 

            if (result != ERROR)
            {
                result = set_actual_spectra_by_number(i + 1);
            }

            if (result != ERROR)
            {
                result = make_actual_illum("", "");
            }
            
            if (result != ERROR)
            {
                result = get_actual_illum(&actual_rgb_vp);
            }

            if (result != ERROR)
            {
                result = divide_vectors(&ratio_vp, canon_rgb_vp, 
                                        actual_rgb_vp);
            }

            if (result != ERROR)
            {

                result = ow_multiply_matrix_by_row_vector_ew(data_mp,
                                                           ratio_vp);

            }

            if (result != ERROR)
            {
                canon_gamut_matrix_vector[ i ] = data_mp; 
            }
        }

        if (result != ERROR)
        {
            const Matrix** const_canon_gamut_matrix_vector = 
                                      N_TYPE_MALLOC(const Matrix*, num_illum);

            if (const_canon_gamut_matrix_vector == NULL)
            {
                result = ERROR;
            }

            if (result != ERROR) 
            {
                for (i=0; i<num_illum; i++)
                {
                    const_canon_gamut_matrix_vector[ i ] =
                                               canon_gamut_matrix_vector[ i ]; 
                }

                result = concat_matrices_vertically(&canon_data_mp, num_illum, 
                                               const_canon_gamut_matrix_vector);
            }

            kjb_free(const_canon_gamut_matrix_vector); 
        }

        if (canon_gamut_matrix_vector != NULL)
        {
            for (i=0; i<num_illum; i++)
            {
                free_matrix(canon_gamut_matrix_vector[ i ]);
            }

            kjb_free(canon_gamut_matrix_vector);
        }
                            
        if (result != ERROR)
        {
            if (fs_specular_factor > 0.0)
            {
                const Matrix* matrix_vector[ 2 ];

                result = ow_multiply_matrix_by_scalar(specular_rgb_mp, 
                                                      fs_specular_factor);

                if (result != ERROR)
                {
                    const Matrix* const_canon_data_mp = canon_data_mp;
                    const Matrix* const_specular_rgb_mp = specular_rgb_mp;

                    matrix_vector[ 0 ] = const_canon_data_mp;
                    matrix_vector[ 1 ] = const_specular_rgb_mp;

                    result = concat_matrices_vertically(
                                              &canon_data_with_specular_rgb_mp, 
                                              2, matrix_vector);
                    canon_gamut_mp = canon_data_with_specular_rgb_mp;
                }
            }
            else 
            {
                canon_gamut_mp = canon_data_mp;
            }
        }

        if (result != ERROR)
        {
            /*
            // We always find the convex hull of the canonical gamut data when
            // we use it, but we take the convex hull here anyway to save
            // resources. 
            */
            canon_gamut_hp = find_convex_hull(canon_gamut_mp, 
                                              DEFAULT_HULL_OPTIONS);
            if (canon_gamut_hp == NULL) result = ERROR; 
        }
        
        if (result != ERROR)
        {
            result = copy_matrix(&(cg_ptr->mp), canon_gamut_hp->vertex_mp);
        }

        if (result != ERROR)
        {
            (cg_ptr->version)++;
            cg_ptr->canon_data_version = NOT_SET;
            cg_ptr->specular_factor = fs_specular_factor;
            cg_ptr->combine_multiple_hulls = TRUE;
            cg_ptr->specular_rgb_version = specular_rgb_version; 
            cg_ptr->world_version = world_version; 
            cg_ptr->illum_spectra_version = illum_spectra_version;
            cg_ptr->canon_illum_version = canon_illum_version; 
            /* The role of fs_specular_white_enable is hidden in
            // get_canon_specular_rgb, but we want to record the current state.
            */
            cg_ptr->specular_white_enable = fs_specular_white_enable; 

            BUFF_CPY(cg_ptr->specular_rgb_config_file, 
                     fs_specular_rgb_config_file); 
            BUFF_CPY(cg_ptr->config_file, "<make>"); 

            result = get_current_world_info(&(cg_ptr->world_info)); 
        }
        
        if (result != ERROR)
        {
            verbose_pso(10, "Canon gamut remade.\n"); 
        }

        free_matrix(canon_data_mp); 
        free_matrix(canon_data_with_specular_rgb_mp); 
        free_hull(canon_gamut_hp); 
        free_vector(actual_rgb_vp);
        free_vector(ratio_vp); 
    }

    free_vector(canon_rgb_vp); 
    free_matrix(specular_rgb_mp); 

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_canon_gamut(cg_ptr, file_name) == ERROR) result = ERROR; 
    }

    if (result != ERROR)
    {
        return cg_ptr->version;
    }
    else
    {
        return ERROR; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int make_canon_gamut_variant(int index, const char* file_name)
{
    int          canon_data_version;
    int          specular_rgb_version;
    int          result               = NO_ERROR;
    Matrix*      specular_rgb_mp      = NULL;
    Canon_gamut* cg_ptr;
    
    if (fs_cached_canon_gamuts[ index ] == NULL)
    {
#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 
        NRE(fs_cached_canon_gamuts[ index ] = create_canon_gamut());
    }

    cg_ptr = fs_cached_canon_gamuts[ index ]; 

    ERE(canon_data_version = make_matte_canon_data( FALSE  /* dry_run */ ));

    if (fs_specular_factor > 0.0) 
    {
        ERE(specular_rgb_version = get_canon_specular_rgb(&specular_rgb_mp));
    }
    else 
    {
        specular_rgb_version = cg_ptr->specular_rgb_version;
    }

    if (    (cg_ptr->canon_data_version == canon_data_version)
         && (IS_EQUAL_DBL(cg_ptr->specular_factor, fs_specular_factor))
         && (cg_ptr->specular_rgb_version == specular_rgb_version)
         && (cg_ptr->combine_multiple_hulls == FALSE)
       )
    {
        verbose_pso(10, "Canon gamut %d is up to date. Skipping make.\n",
                    index + 1); 
    }
    else 
    {
        Spectra* canon_illum_sp         = NULL;
        Matrix*  canon_data_mp          = NULL;
        Matrix*  canon_data_with_specular_rgb_mp = NULL;
        Matrix*  canon_gamut_mp;
        Hull*    canon_gamut_hp         = NULL;

#ifdef __GNUC__
        /* Reduce noise from error checker. */
        canon_gamut_mp = NULL;
#endif 
        verbose_pso(5, "Canon gamut %d is out of date. Remaking it.\n",
                    index + 1);

        canon_data_version = get_synthetic_matte_canon_data(&canon_data_mp); 

        if (canon_data_version == ERROR) 
        {
            result = ERROR;
        }
        else if (fs_specular_factor > 0.0)
        {
            const Matrix* matrix_vector[ 2 ];

            result = ow_multiply_matrix_by_scalar(specular_rgb_mp, 
                                                  fs_specular_factor);

            if (result != ERROR)
            {
                matrix_vector[ 0 ] = canon_data_mp;
                matrix_vector[ 1 ] = specular_rgb_mp;

                result = concat_matrices_vertically(
                                              &canon_data_with_specular_rgb_mp,
                                              2, matrix_vector);
                canon_gamut_mp = canon_data_with_specular_rgb_mp;
            }
        }
        else 
        {
            canon_gamut_mp = canon_data_mp;
        }

        if (result != ERROR)
        {
            /*
            // We always find the convex hull of the canonical gamut data when
            // we use it, but we take the convex hull here anyway to save
            // resources. 
            */
            canon_gamut_hp = find_convex_hull(canon_gamut_mp, 
                                              DEFAULT_HULL_OPTIONS);
            if (canon_gamut_hp == NULL) result = ERROR; 
        }
        
        if (result != ERROR)
        {
            result = copy_matrix(&(cg_ptr->mp), canon_gamut_hp->vertex_mp);
        }

        if (result != ERROR)
        {
            (cg_ptr->version)++;
            cg_ptr->canon_data_version = canon_data_version;
            cg_ptr->specular_factor = fs_specular_factor;
            cg_ptr->combine_multiple_hulls = FALSE;
            cg_ptr->specular_rgb_version = specular_rgb_version; 
            cg_ptr->world_version = NOT_SET; 
            cg_ptr->illum_spectra_version = NOT_SET;
            cg_ptr->canon_illum_version = NOT_SET; 
            /* The role of fs_specular_white_enable is hidden in
            // get_canon_specular_rgb, but we want to record the current state.
            */
            cg_ptr->specular_white_enable = fs_specular_white_enable; 

            BUFF_CPY(cg_ptr->specular_rgb_config_file, 
                     fs_specular_rgb_config_file); 

            BUFF_CPY(cg_ptr->config_file, "<make>"); 

            result = get_current_world_info(&(cg_ptr->world_info)); 
        }
        
        if (result != ERROR)
        {
            verbose_pso(10, "Canon gamut remade.\n"); 
        }

        free_matrix(canon_data_mp); 
        free_matrix(canon_data_with_specular_rgb_mp); 
        free_spectra(canon_illum_sp); 
        free_hull(canon_gamut_hp); 
    }

    free_matrix(specular_rgb_mp); 

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_canon_gamut(cg_ptr, file_name) == ERROR) result = ERROR; 
    }

    if (result != ERROR)
    {
        return cg_ptr->version;
    }
    else
    {
        return ERROR; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_canon_specular_rgb(Matrix** rgb_mpp)
{
    static int     previous_specular_spectra_version = NOT_SET;
    static int     previous_sensor_version = NOT_SET; 
    static int     previous_specular_white_enable = NOT_SET;
    int            specular_spectra_version; 
    int            sensor_version = NOT_SET; 
    Spectra*       canon_illum_sp                    = NULL;
    Spectra*       specular_reflect_sp               = NULL;
    Vector*        canon_illum_vp                    = NULL;
    int            result                            = NO_ERROR;
    Matrix*        specular_rgb_from_spectra_mp      = NULL;
    Matrix*        white_specular_rgb_mp             = NULL;
    
    specular_spectra_version = get_specular_spectra(&specular_reflect_sp);

    if (specular_spectra_version != ERROR)
    {
        result = get_canon_spectra(&canon_illum_sp);

        if (result != ERROR)
        {
            sensor_version = generate_RGB_data(&specular_rgb_from_spectra_mp, 
                                               canon_illum_sp,
                                               0, specular_reflect_sp, 
                                               (Spectra*)NULL);
            if (sensor_version == ERROR) result = ERROR;
        }

        if (    (result != ERROR)
             &&
                (    (specular_spectra_version != 
                                             previous_specular_spectra_version)
                  || (sensor_version != previous_sensor_version)
                )
           )
        {
            ASSERT(IS_SET(sensor_version)); 

            previous_specular_spectra_version = specular_spectra_version;
            previous_sensor_version = sensor_version;

            fs_canon_specular_rgb_version++; 
        }
    }

    if ((result != ERROR) && (fs_specular_white_enable))
    {
        result = get_canon_illum(&canon_illum_vp);

        if (result != ERROR)
        {
            result = get_target_matrix(&white_specular_rgb_mp, 1, 3);
        }

        if (result != ERROR)
        {
            result = put_matrix_row(white_specular_rgb_mp, canon_illum_vp, 0);
        }
    }

    if (    (result != ERROR) 
         && (fs_specular_white_enable != previous_specular_white_enable)
       )
    {
        previous_specular_white_enable = fs_specular_white_enable;
        fs_canon_specular_rgb_version++;
    }

    if (result != ERROR)
    {
        const Matrix*  specular_sources_matrix_list[ 3 ];

        specular_sources_matrix_list[ 0 ] = fs_cached_specular_rgb_mp;
        specular_sources_matrix_list[ 1 ] = specular_rgb_from_spectra_mp; 
        specular_sources_matrix_list[ 2 ] = white_specular_rgb_mp; 

        result = concat_matrices_vertically(rgb_mpp, 3, 
                                            specular_sources_matrix_list);
    }

    if (result != ERROR)
    {
        Vector* max_vp = NULL;

        result = get_max_matrix_row_elements(&max_vp, *rgb_mpp);

        if (result != ERROR)
        {
            result = ow_divide_matrix_by_col_vector(*rgb_mpp, max_vp);
        }

        if (result != ERROR)
        {
            result = ow_multiply_matrix_by_scalar(*rgb_mpp, 255.0);
        }

        free_vector(max_vp);
    }

    free_vector(canon_illum_vp);
    free_spectra(specular_reflect_sp);
    free_spectra(canon_illum_sp);
    free_matrix(white_specular_rgb_mp); 
    free_matrix(specular_rgb_from_spectra_mp); 

    if (result == ERROR) 
    {
        return ERROR; 
    }
    else 
    {
        return fs_canon_specular_rgb_version;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int canon_specular_rgb_is_up_to_date(int proposed_version)
{
    return (fs_canon_specular_rgb_version == proposed_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_canon_gamut(Canon_gamut* cg_ptr, const char* file_name)
{
    FILE* fp; 
    int   result; 
    Error_action save_error_action;

    NRE(fp = kjb_fopen(file_name, "w"));

    result = kjb_fprintf(fp, "\n# Canonical gamut file\n\n"); 

    if (result != ERROR)
    {
        result = write_world_info(fp, &(cg_ptr->world_info));
    }

    if (result != ERROR)
    {
        result = kjb_fprintf(fp, "\n");
    }

    if (result != ERROR)
    {
        result = write_canon_gamut_info(fp, cg_ptr);
    }

    if (result != ERROR)
    {
        result = kjb_fprintf(fp, "\n\n");
    }

    if (result != ERROR)
    {
        result = fp_write_matrix(cg_ptr->mp, fp); 
    }

    save_error_action = get_error_action();
    set_error_action(FORCE_ADD_ERROR_ON_ERROR);
    if (kjb_fclose(fp) == ERROR) result = ERROR; 
    set_error_action(save_error_action); 
        
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_canon_gamut_info(FILE* fp, Canon_gamut* cg_ptr)
{

    ERE(kjb_fprintf(fp, "#! fs_specular_rgb_config_file = %s\n",
                    cg_ptr->specular_rgb_config_file[ 0 ] 
                    ?   cg_ptr->specular_rgb_config_file  
                    :   "<none>")); 

    ERE(kjb_fprintf(fp, "#! fs_combine_multiple_hulls = %s\n",
                    cg_ptr->combine_multiple_hulls ? "t" : "f"));
 
    ERE(kjb_fprintf(fp, "#! fs_specular_white_enable = %s\n",
                    cg_ptr->specular_white_enable ? "t" : "f"));
 
    if (cg_ptr->specular_factor < 0.0)
    {
        ERE(kjb_fprintf(fp, "#! fs_specular_factor = off\n"));
    }
    else 
    {
        ERE(kjb_fprintf(fp, "#! fs_specular_factor = %.1f\n", 
                        cg_ptr->specular_factor));
    }
 
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_gamut_from_file
(
    Canon_gamut** cg_ptr_ptr,
    const char*   file_name 
)
{
    Canon_gamut* cg_ptr;
    FILE*        fp;
    int          result;

    if (*cg_ptr_ptr == NULL)
    {
        NRE(*cg_ptr_ptr = create_canon_gamut());
    }

    cg_ptr = *cg_ptr_ptr; 

    NRE(fp = open_gamut_config_file((const char*)NULL, file_name, 
                                    "canon gamut"));

    result = read_world_info(fp, &(cg_ptr->world_info));

    if (result != ERROR)
    {
        result = read_canon_gamut_info(fp, cg_ptr);
    }

    if (result != ERROR)
    {
        result = fp_read_formatted_matrix(&(cg_ptr->mp), fp);
    }

    if (result != ERROR)
    {
        BUFF_GET_FD_NAME(fileno(fp), cg_ptr->config_file);
    }

    kjb_fclose(fp); /* Only reading, so I don't care about the return. */

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_gamut_info(FILE* fp, Canon_gamut* cg_ptr)
{
    char** option_list;
    char** value_list;
    char*  value;
    char*  option;
    int    num_options;
    int    i;
    long save_file_pos; 
    int result = NO_ERROR; 
    Error_action save_error_action;

    ERE(save_file_pos = kjb_ftell(fp)); 

    while ((num_options = read_header_options(fp, &option_list, &value_list))
           != EOF)
    {
        if (num_options == ERROR) 
        {
            result = ERROR;
            break; 
        }

        for (i=0; i<num_options; i++)
        {
            option = option_list[ i ];
            value  = value_list[ i ];

            if (STRCMP_EQ(option, "fs_specular_rgb_config_file"))
            {
                BUFF_CPY(cg_ptr->specular_rgb_config_file, value);
            }
            else if (STRCMP_EQ(option, "fs_combine_multiple_hulls"))
            {
                ERE(cg_ptr->combine_multiple_hulls = get_boolean_value(value));
            }
            else if (STRCMP_EQ(option, "fs_specular_factor"))
            {
                if (is_no_value_word(value))
                {
                    cg_ptr->specular_factor = DBL_NOT_SET; 
                }
                else 
                {
                    ERE(ss1d(value, &(cg_ptr->specular_factor))); 
                }
            }
            else if (STRCMP_EQ(option, "fs_specular_white_enable"))
            {
                ERE(cg_ptr->specular_white_enable = get_boolean_value(value));
            }
        }

        free_options(num_options, option_list, value_list); 
    }

    save_error_action = get_error_action();
    set_error_action(FORCE_ADD_ERROR_ON_ERROR);
    if (kjb_fseek(fp, save_file_pos, SEEK_SET) == ERROR) result = ERROR; 
    set_error_action(save_error_action); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_canon_gamut(Matrix** target_mpp, int canon_gamut_num)
{

    ERE(get_canon_gamut(target_mpp, canon_gamut_num));
    ERE(ow_sharpen_matrix_data(*target_mpp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_canon_gamut(Matrix** target_mpp, int canon_gamut_num)
{
    if (    (canon_gamut_num == 0)
         && (fs_cached_canon_gamuts[ canon_gamut_num ] == NULL) 
       )
    {
#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 
        ERE(read_canon_gamut_from_file(&(fs_cached_canon_gamuts[ 0 ]), 
                                         CANON_GAMUT_FILE));
    }

    if (fs_cached_canon_gamuts[ canon_gamut_num ] == NULL) 
    {
        free_matrix(*target_mpp);
        *target_mpp = NULL; 
        return NO_ERROR; 
    }

    ERE(copy_matrix(target_mpp, fs_cached_canon_gamuts[ canon_gamut_num ]->mp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void invalidate_canon_gamut(int index)
{
    free_canon_gamut(fs_cached_canon_gamuts[ index ]);

    fs_cached_canon_gamuts[ index ] = NULL; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static Canon_gamut* create_canon_gamut(void)
{
    Canon_gamut* cg_ptr;

    NRN(cg_ptr = TYPE_MALLOC(Canon_gamut));

    cg_ptr->mp = NULL;
    
    cg_ptr->version                = 0;

    cg_ptr->world_version          = NOT_SET;
    cg_ptr->specular_rgb_version   = NOT_SET;
    cg_ptr->canon_data_version     = NOT_SET;
    cg_ptr->canon_illum_version    = NOT_SET;
    cg_ptr->illum_spectra_version  = NOT_SET;
    
    cg_ptr->combine_multiple_hulls = NOT_SET;
    
    cg_ptr->specular_factor        = DBL_NOT_SET;
    cg_ptr->combine_multiple_hulls = NOT_SET;
    cg_ptr->specular_white_enable  = NOT_SET;

    cg_ptr->specular_rgb_config_file[ 0 ] = '\0';

    ERN(initialize_world_info(&(cg_ptr->world_info))); 

    return cg_ptr; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void free_canon_gamut(Canon_gamut* cg_ptr)
{

    if (cg_ptr != NULL)
    {
        free_matrix(cg_ptr->mp); 
        kjb_free(cg_ptr);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    int i;
    
    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        free_canon_gamut(fs_cached_canon_gamuts[ i ]);
    }

    free_matrix(fs_cached_specular_rgb_mp); 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

