
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Lindsay Martin and Kobus Barnard (authors). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef CORR_LIB_INCLUDED
#define CORR_LIB_INCLUDED

#include "l/l_incl.h"
#include "m/m_incl.h"
#include "s/s_incl.h"

/* ========================================================================= *
|                                                                             |
|  Copyright (c) 1994-1998, by Lindsay Martin, Kobus Barnard (authors).       |
|                                                                             |
|  For use outside the SFU vision lab please contact the author(s).           |
|                                                                             |
* ========================================================================== */

/* This file contains data types and routines that are common between the
// 2-D and 3-D flavours of the Colour by Correlation algorithms.
*/

/* 
// We convolve with a normalized mask, so the convolutions are almost
// normalized. However, when we are at the edges, the effective mask weight is
// not 1. This #define optionally fixes this problem.
//
// However, the probability distribution outside the defined area is is assumed
// to be ZERO. Thus the above normalization is a bad idea, and this #define is
// recomended to be left commented out. 
//
// #define NORMALIZE_CONVOLUTIONS
*/

#define R_CHROM_INDEX                       0
#define G_CHROM_INDEX                       1
#define MIN_COL_INDEX                       0
#define MAX_COL_INDEX                       1
#define MAX_RG_CHROM                        1.0
#define TWO_DIMENSIONAL                     2
#define THREE_DIMENSIONAL                   3

typedef enum Corr_colour_space
{
    CORR_INVALID_COLOUR_SPACE = ERROR, 
    CORR_RG_SPACE  = 2,   /* Backwards compatability. */
    CORR_RGB_SPACE = 3,   /* Backwards compatability. */
    CORR_RG_LUM_SPACE = 4
}
Corr_colour_space;

/* Kobus: Reduced the value of CORRELATION_MINUS_INFINITY because using exactly
//        the smallest number lead to problems.
*/
#define CORRELATION_MINUS_INFINITY          ((double)-0.3 * DBL_MAX)

#define NEW_CORRELATION_IO_VERSION
#define FILE_MINUS_INFINITY                 ((double)12345.0) 

#define CORRELATION_MIN_LOG  (log(10000.0 * (double)DBL_MIN))
#define CORRELATION_MAX_LOG  (log((double)DBL_MAX / 10000.0))

/* Define the algorithms currently implemented in the correlation group. */
typedef enum Corr_algorithm
{
    CMA_ZERO_ONE,
    CMA_PROBABILITY,
    CMA_LOG_PROBABILITY
} Corr_algorithm;

/* Define the data source that the correlation matrix was derived from.  */
typedef enum Corr_matrix_src
{
    CMS_UNDEFINED, 
    CMS_RGB, 
    CMS_SPECTRA 
} Corr_matrix_src;

/* Define the distribution of the reflectance database used in generating
// the correlation matrix.                                                 */
typedef enum Corr_matrix_dist
{
    CMD_UNIFORM, 
    CMD_GAUSSIAN,
    CMD_NONE
} Corr_matrix_dist;

/* Define the Correlation Matrix data type.                                */
typedef struct Corr_matrix
{
  Corr_algorithm    algorithm;      /* Algorithm used to generate matrix.   */
  Corr_matrix_src   data_source;    /* Data source used to build matrix.    */
  Corr_matrix_dist  distribution;   /* Assumed prior distribution type.     */
  int               num_illum;      /* Number of illuminants in database.   */
  int               num_surf;       /* Number of surfaces used in database. */  
  int               num_bins;       /* Number of histogram bins in first two axes.*/
  int               num_bins_3;     /* Number of histogram bins in third axis.  */
  Corr_colour_space colour_space;
  double            sigma;          /* Gaussian distribution sigma.         */
  double            make_max_RGB;   /* Maximum RGB value for 3D.            */
  Matrix*           cormat_mp;      /* Pointer to correl. matrix data.      */
  Matrix*           illum_chrom_mp; /* Pointer to illuminant spectra.       */
  Spectra*          illum_sp;       /* Pointer to illuminant spectra.       */
}
Corr_matrix;

/* Correlation Matrix Data Type Functions -----------------------------------*/

int get_target_correlation_matrix
(
    Corr_matrix**     output_correl_cp_ptr,
    Corr_algorithm    algorithm,
    Corr_matrix_src   src,
    Corr_matrix_dist  dist,
    int               num_illum,
    int               num_surf,
    int               num_bins,
    int               num_bins_3,
    Corr_colour_space colour_space,
    double            sigma,
    double            make_max_RGB 
);

void free_correlation_matrix(Corr_matrix* correl_cp);

int copy_correlation_matrix
(
    Corr_matrix**      target_cp_ptr,
    const Corr_matrix* source_cp 
);

int read_correlation_matrix
(
    Corr_matrix** target_cp_ptr,
    const char*   file_name,
    const char*   error_msg 
);

int write_correlation_matrix
(
    const char*        file_name,
    const Corr_matrix* source_cp 
);

#endif /* #ifndef CORR_LIB_INCLUDED */
