
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "synthetic_data.h"
#include "image_data.h"
#include "illum_data.h"
#include "file_data.h"
#include "image_input.h"
#include "data.h"

/* -------------------------------------------------------------------------- */

typedef enum Data_source
{
    USE_IMAGE_DATA,
    USE_SYNTHETIC_DATA,
    USE_FILE_DATA,
    USE_IMAGE_VAR_DATA,
    USE_SYNTHETIC_VAR_DATA,
    USE_FILE_VAR_DATA
}
Data_source; 

/* -------------------------------------------------------------------------- */

static Data_source    fs_data_source             = USE_FILE_DATA;
static Queue_element* fs_data_source_stack_head  = NULL;
static int            fs_data_cluster_resolution = NOT_SET;
static double           fs_artificial_clip_point   = DBL_NOT_SET;
static double           fs_artificial_aperture     = DBL_NOT_SET;

/* -------------------------------------------------------------------------- */
 
#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */
 
int set_data_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "data-source"))
    {
        const char* str;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if      (fs_data_source == USE_IMAGE_DATA)         str = "image";
            else if (fs_data_source == USE_SYNTHETIC_DATA)     str = "synthetic";
            else if (fs_data_source == USE_FILE_DATA)          str = "file";
            else if (fs_data_source == USE_IMAGE_VAR_DATA)     str = "image-var";
            else if (fs_data_source == USE_SYNTHETIC_VAR_DATA) str = "synthetic-var";
            else if (fs_data_source == USE_FILE_VAR_DATA)      str = "file-var";
            else 
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }

            ERE(pso("data-source = %s\n", str));  
        }
        else if (value[ 0 ] == '\0')
        {
            if      (fs_data_source == USE_IMAGE_DATA)         str = "image";
            else if (fs_data_source == USE_SYNTHETIC_DATA)     str = "synthetic";
            else if (fs_data_source == USE_FILE_DATA)          str = "file";
            else if (fs_data_source == USE_IMAGE_VAR_DATA)     str = "image var";
            else if (fs_data_source == USE_SYNTHETIC_VAR_DATA) str = "synthetic var";
            else if (fs_data_source == USE_FILE_VAR_DATA)      str = "file var";
            else 
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }

            ERE(pso("Data source is %s data.\n", str)); 
        }
        else
        {
            if (STRCMP_EQ(value, "image"))
            {
                fs_data_source = USE_IMAGE_DATA;
            }
            else if (    (STRCMP_EQ(value, "synthetic"))
                      || (STRCMP_EQ(value, "make"))
                      || (STRCMP_EQ(value, "generate"))
                      || (STRCMP_EQ(value, "gen"))
                    )
            {
                fs_data_source = USE_SYNTHETIC_DATA;
            }
            else if (STRCMP_EQ(value, "file"))
            {
                fs_data_source = USE_FILE_DATA;
            }
            else if (STRCMP_EQ(value, "image-var"))
            {
                fs_data_source = USE_IMAGE_VAR_DATA;
            }
            else if (STRCMP_EQ(value, "synthetic-var"))
            {
                fs_data_source = USE_SYNTHETIC_VAR_DATA;
            }
            else if (STRCMP_EQ(value, "file-var"))
            {
                fs_data_source = USE_FILE_VAR_DATA;
            }
            else 
            {
                set_error("%q is an invalid data source designator.", value);
                return ERROR;
            }
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "data-cluster-resolution")))
    {
        int temp_int_value;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("data-cluster-resolution = %d.\n",
                    fs_data_cluster_resolution)); 
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_data_cluster_resolution > 0) 
            {
                ERE(pso("Data cluster resolution is %d.\n",
                        fs_data_cluster_resolution)); 
            }
            else if (fs_data_cluster_resolution == 0) 
            {
                ERE(pso("Data clustering is done by convex hull.\n")); 
            }
            else 
            {
                ERE(pso("Data clustering is not being used.\n")); 
            }
        }
        else
        {
            ERE(ss1i(value, &temp_int_value));

            if (temp_int_value > 100)
            {
                set_error("Data cluster resolution must be at most 100.");
                return ERROR;
            }
            fs_data_cluster_resolution = temp_int_value;
        }
        result = NO_ERROR; 
    }

    if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "artificial-clip-point"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_artificial_clip_point < 0.0)
            {
                ERE(pso("artificial-clip-point = off\n"));
            }
            else 
            {
                ERE(pso("artificial-clip-point = %.1f.\n", 
                        fs_artificial_clip_point));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_artificial_clip_point < 0.0)
            {
                ERE(pso("Pixels are not aritificially clipped.\n"));
            }
            else 
            {
                ERE(pso("The artificial clip point is %.1f.\n",
                        fs_artificial_clip_point));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_artificial_clip_point = DBL_NOT_SET;
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 
            fs_artificial_clip_point = temp_real;
        }
        result = NO_ERROR; 
    }
    
    if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "artificial-aperture"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_artificial_aperture < 0.0)
            {
                ERE(pso("artificial-aperture = off\n"));
            }
            else 
            {
                ERE(pso("artificial-aperture = %.1f.\n", 
                        fs_artificial_aperture));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_artificial_aperture < 0.0)
            {
                ERE(pso("No artificial aperture is set.\n"));
            }
            else 
            {
                ERE(pso("The artificial aperture is such that the max RGB "));
                ERE(pso("of white is %.1f.\n", fs_artificial_aperture));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_artificial_aperture = DBL_NOT_SET;
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 
            fs_artificial_aperture = temp_real;
        }
        result = NO_ERROR; 
    }
    
    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void get_data_source(char* source_buff, size_t max_len)
{
    if (    (fs_data_source == USE_IMAGE_DATA) 
         || (fs_data_source == USE_IMAGE_VAR_DATA)
       )
    {
        get_image_data_source(source_buff, max_len);
    }
    else if (    (fs_data_source == USE_SYNTHETIC_DATA)
              || (fs_data_source == USE_SYNTHETIC_VAR_DATA)
            )
    {
        kjb_strncpy(source_buff, "synthetic data", max_len);
    }
    else if (    (fs_data_source == USE_FILE_DATA) 
              || (fs_data_source == USE_FILE_VAR_DATA)
            )
    {
        get_file_data_source(source_buff, max_len);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG(); 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_image_data(void)
{
    if (fs_data_source != USE_IMAGE_DATA) 
    {
        verbose_pso(1, "Switching to image data.\n"); 
        fs_data_source = USE_IMAGE_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_image_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_IMAGE_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_synthetic_data(void)
{
    if (fs_data_source != USE_SYNTHETIC_DATA) 
    {
        verbose_pso(1, "Switching to synthetic data.\n"); 
        fs_data_source = USE_SYNTHETIC_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_synthetic_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_SYNTHETIC_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_file_data(void)
{
    if (fs_data_source != USE_FILE_DATA) 
    {
        verbose_pso(1, "Switching to file data.\n"); 
        fs_data_source = USE_FILE_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_file_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_FILE_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_image_var_data(void)
{
    if (fs_data_source != USE_IMAGE_VAR_DATA) 
    {
        verbose_pso(1, "Switching to image var data.\n"); 
        fs_data_source = USE_IMAGE_VAR_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_image_var_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_IMAGE_VAR_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_synthetic_var_data(void)
{
    if (fs_data_source != USE_SYNTHETIC_VAR_DATA) 
    {
        verbose_pso(1, "Switching to synthetic var data.\n"); 
        fs_data_source = USE_SYNTHETIC_VAR_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_synthetic_var_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_SYNTHETIC_VAR_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int force_file_var_data(void)
{
    if (fs_data_source != USE_FILE_VAR_DATA) 
    {
        verbose_pso(1, "Switching to file var data.\n"); 
        fs_data_source = USE_FILE_VAR_DATA;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int request_file_var_data(void)
{
    int* save_data_source_ptr;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(save_data_source_ptr = INT_MALLOC(1));
    *save_data_source_ptr = fs_data_source;

    ERE(insert_into_queue(&fs_data_source_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_data_source_ptr));

    fs_data_source = USE_FILE_VAR_DATA;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int restore_data_source(void)
{
    Queue_element* cur_elem;

    if (fs_data_source_stack_head == NULL) return NO_ERROR;  

    NRE(cur_elem = remove_first_element(&fs_data_source_stack_head,
                                        (Queue_element**)NULL));

    fs_data_source = *((Data_source*)(cur_elem->contents));

    free_queue_element(cur_elem, kjb_free); 

    return fs_data_source;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_data_points(void)
{
    if (fs_data_source == USE_IMAGE_DATA)
    {
        return get_num_image_data_points();
    }
    else if (fs_data_source == USE_SYNTHETIC_DATA)
    {
        return get_num_synthetic_data_points();
    }
    else if (fs_data_source == USE_FILE_DATA)
    {
        return get_num_file_data_points();
    }
    else if (fs_data_source == USE_IMAGE_VAR_DATA)
    {
        return get_num_image_var_data_points();
    }
    else if (fs_data_source == USE_SYNTHETIC_VAR_DATA)
    {
        return get_num_synthetic_var_data_points();
    }
    else if (fs_data_source == USE_FILE_VAR_DATA)
    {
        return get_num_file_var_data_points();
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_derived_data
(
    Matrix** result_mpp,
    int      num_data_points,
    int*     data_point_nums
)
{
    int result;

    if (fs_data_source == USE_IMAGE_DATA) 
    {
        result = get_selected_image_data(result_mpp, num_data_points, 
                                         data_point_nums);
    }
    else if (fs_data_source == USE_SYNTHETIC_DATA)
    {
        result = get_selected_synthetic_data(result_mpp, num_data_points, 
                                             data_point_nums);
    }
    else if (fs_data_source == USE_FILE_DATA)
    {
        result = get_selected_file_data(result_mpp, num_data_points, 
                                        data_point_nums);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR;
    }

    if ((result != ERROR) && (fs_artificial_aperture >= 0.0)) 
    {
        Vector* actual_illum_vp = NULL;

        result = get_actual_illum(&actual_illum_vp);

        if (result != ERROR) 
        {
            double max = max_vector_element(actual_illum_vp);
            double factor = fs_artificial_aperture / max; 

            result = ow_multiply_matrix_by_scalar(*result_mpp, factor);

            verbose_pso(3, "Scaling data by %f due to artificial aperture\n", 
                        factor); 
        }

        free_vector(actual_illum_vp);
    }

    /*
    //  If there is artificial clipping, do it here, by removing clipped points.
    //  This can lead to the possiblity that there are no good data points, so
    //  we have to check for that. The reason why we clip here, rather than
    //  before selection is to force clipping to have the required negative
    //  impact. The number of selected points is supposed to represent the
    //  number of scene surfaces, and if one is clipped, then we want to supply
    //  less data to the algorithms. Also, clipping before makes handling the
    //  number of points more difficult. 
    */
    if ((result != ERROR) && (fs_artificial_clip_point >= 0.0)) 
    {
        Matrix* result_mp = *result_mpp; 
        int num_rows = result_mp->num_rows; 
        int count = 0; 
        int i; 

        for (i=0; i<num_rows; i++) 
        {
            if (    (result_mp->elements[ i ][ 0 ] < fs_artificial_clip_point) 
                 && (result_mp->elements[ i ][ 1 ] < fs_artificial_clip_point) 
                 && (result_mp->elements[ i ][ 2 ] < fs_artificial_clip_point) 
               )
            {
                result_mp->elements[count][0] = result_mp->elements[i][0];
                result_mp->elements[count][1] = result_mp->elements[i][1];
                result_mp->elements[count][2] = result_mp->elements[i][2];
                
                count++; 
            }
        }

        if (count != result_mp->num_rows)
        {
            verbose_pso(3, "%d rows skipped due to artificial clipping.\n",
                        result_mp->num_rows - count);
        }

        result_mp->num_rows = count; 

        if (count == 0) 
        {
            set_error("No good data after artificial clipping.");
            result = ERROR; 
        }
    }

    if ((result != ERROR) && (fs_data_cluster_resolution > 0))
    {
        Matrix* clustered_data_mp = NULL; 

        verbose_pso(3, "Clustering input data with resolution %d.\n", 
                    fs_data_cluster_resolution);

        result = get_fixed_clustering_of_3D_data(&clustered_data_mp, 
                                                 *result_mpp, 
                                                 fs_data_cluster_resolution);

        if (result != ERROR) 
        {
            result = copy_matrix(result_mpp, clustered_data_mp);
        }

        free_matrix(clustered_data_mp); 
    }
    else if ((result != ERROR) && (fs_data_cluster_resolution == 0))
    {
        Hull* hp; 

        verbose_pso(3, "Clustering input data by convex hull.\n");

        hp = find_convex_hull(*result_mpp, DEFAULT_HULL_OPTIONS);

        if (hp == NULL)
        {
            result = ERROR;
        }
        else 
        {
            result = copy_matrix(result_mpp, hp->vertex_mp);

            free_hull(hp);
        }
    }

    if (result != ERROR) 
    {
#ifdef TEST 
        /*
        // In the varying illumination case, we must get unsharp data, and then
        // do clever stuff to sharpen both the result data and the var data.
        // This is so that the lower level routines for getting the data can be
        // similar. 
        */
        if (    (fs_data_source == USE_FILE_VAR_DATA) 
             || (fs_data_source == USE_SYNTHETIC_VAR_DATA) 
             || (fs_data_source == USE_IMAGE_VAR_DATA)
           )
        {
        }
        else
        {
             result = ow_sharpen_matrix_data(*result_mpp);
        }
#else 
        result = ow_sharpen_matrix_data(*result_mpp);
#endif 
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_selected_registered_data
(
    Matrix** result_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(get_selected_registered_data(result_mpp, num_data_points, 
                                     data_point_nums));
    ERE(ow_sharpen_matrix_data(*result_mpp));

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_registered_data
(
    Matrix** result_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{
    int result;

    if (fs_data_source == USE_IMAGE_DATA) 
    {
        result = get_selected_registered_image_data(result_mpp,
                                                    num_data_points, 
                                                    data_point_nums);
    }
    else if (fs_data_source == USE_SYNTHETIC_DATA)
    {
        result = get_selected_registered_synthetic_data(result_mpp, 
                                                        num_data_points, 
                                                        data_point_nums);
    }
    else if (fs_data_source == USE_FILE_DATA)
    {
        result = get_selected_registered_file_data(result_mpp,
                                                   num_data_points, 
                                                   data_point_nums);
    }
    else if (fs_data_source == USE_IMAGE_VAR_DATA)
    {
        result = get_selected_registered_image_var_data(result_mpp, 
                                                        num_data_points, 
                                                        data_point_nums);
    }
    else if (fs_data_source == USE_SYNTHETIC_VAR_DATA)
    {
        result = get_selected_registered_synthetic_var_data(result_mpp, 
                                                            num_data_points, 
                                                            data_point_nums);
    }
    else if (fs_data_source == USE_FILE_VAR_DATA)
    {
        result = get_selected_registered_file_var_data(result_mpp, 
                                                       num_data_points, 
                                                       data_point_nums);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_registered_data(Matrix** result_mpp)
{

    ERE(get_registered_data(result_mpp));
    ERE(ow_sharpen_matrix_data(*result_mpp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_registered_data(Matrix** result_mpp)
{

    /*
    // Originally, I had planned to cache the result, but there is not enough
    // to be gained here. Thus we just use this routine as a big switch to get
    // the data. Note that the data source specific routines return version
    // numbers, but here we force either ERROR or NO_ERROR, because without a
    // cache which is updated when data source is changed (as well as when the
    // data changes), version numbers do not make much sense. 
    */

    /*
    // In the case of artificial clipping: To match what we do with images, we
    // would have to clip the data and the canon_data simultaneously, and return
    // the set of points where there is no clipping in either the data or the
    // canon data, and do the same for canon data. However, I am not convinced
    // that this is what one would do in the ideal case with images, and
    // furthermore, it is extra work. (In other words, the method used with
    // images is forced on us by the lack of data--noting that the canon data is
    // only used for measuring error). Thus we will live with this slight
    // discrepency between the double and simulated case. 
    //
    // (Bottem line: Don't clip) 
    */

    if (fs_data_source == USE_IMAGE_DATA) 
    {
        ERE(get_registered_image_data(result_mpp));
    }
    else if (fs_data_source == USE_SYNTHETIC_DATA)
    {
        ERE(get_registered_synthetic_data(result_mpp));
    }
    else if (fs_data_source == USE_FILE_DATA)
    {
        ERE(get_registered_file_data(result_mpp));
    }
    else if (fs_data_source == USE_IMAGE_VAR_DATA)
    {
        ERE(get_registered_image_var_data(result_mpp));
    }
    else if (fs_data_source == USE_SYNTHETIC_VAR_DATA)
    {
        ERE(get_registered_synthetic_var_data(result_mpp));
    }
    else if (fs_data_source == USE_FILE_VAR_DATA)
    {
        ERE(get_registered_file_var_data(result_mpp));
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_selected_canon_data
(
    Matrix** selected_canon_data_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(get_selected_canon_data(selected_canon_data_mpp, num_data_points, 
                                data_point_nums));
    ERE(ow_sharpen_matrix_data(*selected_canon_data_mpp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_canon_data
(
    Matrix** selected_canon_data_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{
    int result;

    if (    (fs_data_source == USE_IMAGE_DATA)
         || (fs_data_source == USE_IMAGE_VAR_DATA)
       )
    {
        result = get_selected_image_canon_data(selected_canon_data_mpp,
                                               num_data_points,
                                               data_point_nums);
    }
    else if (    (fs_data_source == USE_SYNTHETIC_DATA)
              || (fs_data_source == USE_SYNTHETIC_VAR_DATA)
            )
    {
        result = get_selected_synthetic_canon_data(selected_canon_data_mpp,
                                                   num_data_points,
                                                   data_point_nums);
    }
    else if (    (fs_data_source == USE_FILE_DATA) 
              || (fs_data_source == USE_FILE_VAR_DATA)
            )
    {
        result = get_selected_file_canon_data(selected_canon_data_mpp, 
                                        num_data_points,
                                        data_point_nums);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_canon_data(Matrix** canon_data_mpp)
{
    int result;

    if (    (fs_data_source == USE_IMAGE_DATA)
         || (fs_data_source == USE_IMAGE_VAR_DATA)
       )
    {
        result = get_image_canon_data(canon_data_mpp);
    }
    else if (    (fs_data_source == USE_SYNTHETIC_DATA)
              || (fs_data_source == USE_SYNTHETIC_VAR_DATA)
            )
    {
        result = get_synthetic_canon_data(canon_data_mpp);
    }
    else if (    (fs_data_source == USE_FILE_DATA) 
              || (fs_data_source == USE_FILE_VAR_DATA)
            )
    {
        result = get_file_canon_data(canon_data_mpp);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_canon_data(Matrix** canon_data_mpp)
{
    int result;

    if (    (fs_data_source == USE_IMAGE_DATA)
         || (fs_data_source == USE_IMAGE_VAR_DATA)
       )
    {
        result = get_image_canon_data(canon_data_mpp);
    }
    else if (    (fs_data_source == USE_SYNTHETIC_DATA)
              || (fs_data_source == USE_SYNTHETIC_VAR_DATA)
            )
    {
        result = get_synthetic_canon_data(canon_data_mpp);
    }
    else if (    (fs_data_source == USE_FILE_DATA) 
              || (fs_data_source == USE_FILE_VAR_DATA)
            )
    {
        result = get_file_canon_data(canon_data_mpp);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR; 
    }

    if (result != ERROR) 
    {
        result = ow_sharpen_matrix_data(*canon_data_mpp);
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;  

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_queue(&fs_data_source_stack_head, (Queue_element**)NULL,
               kjb_free); 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

