
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "compute_output.h"
#include "image_data.h"
#include "display.h"

/* -------------------------------------------------------------------------- */

#define MAX_NUM_COLLAGE_IMAGES         100
#define MAX_MIN_COLLAGE_DIM              6

/* -------------------------------------------------------------------------- */

typedef struct Display_key_word_options
{
    int chromaticity_only;
    int collage_format;
    int half_size;
    int illumination;
    int pre_processed_image;
}
Display_key_word_options;   

/* -------------------------------------------------------------------------- */

static int display_collage(Queue_element* image_map_data_queue_head);
static int display_normal(Queue_element* image_map_data_queue_head);

/* -------------------------------------------------------------------------- */

int process_display_command
(
    char*                    arg,
    Colour_constancy_result* result_ptr 
)
{ 
    /*
    // Although at first glance, "pp" is more naturally handled as an algorithm
    // (as are input and canon), we don't want to send it to the collage maker,
    // and thus we handle it as a special case, by adding it in with the key
    // words--a bit of a HACK, sorry.
    */
    static const char *display_key_words[] =  
                              { "pp", "collage", "chrom", "full", "illum" };
    int        num_key_words = sizeof(display_key_words)/sizeof(char*);
    int        key_words_found_list[ sizeof(display_key_words)/sizeof(char*) ];
    const char* key_word;
    int        i, count;
    int        num_found_key_words;
    Display_key_word_options key_word_options;
    Image_map_data*          image_map_data_ptr;
    Queue_element*           image_map_data_queue_head = NULL;
    char                     request[ 100 ];
    int                      reduction_flag      = FALSE;
    int                      result              = NO_ERROR;

    trim_beg(&arg);

    /* Set key_word_options defaults.   */
    key_word_options.chromaticity_only   = FALSE;
    key_word_options.collage_format      = FALSE;
    key_word_options.half_size           = TRUE;
    key_word_options.illumination        = FALSE;
    key_word_options.pre_processed_image = FALSE;

    ERE(num_found_key_words = ic_parse_key_words(arg, num_key_words, 
                                                 display_key_words, 
                                                 key_words_found_list, 
                                                 (int*)NULL));
    for (i=0; i<num_found_key_words; i++) 
    {
        key_word = display_key_words[ key_words_found_list[ i ] ];

        if (IC_STRCMP_EQ(key_word, "full"))
            key_word_options.half_size = FALSE;
        else if (IC_STRCMP_EQ(key_word, "collage"))
            key_word_options.collage_format = TRUE; 
        else if (IC_STRCMP_EQ(key_word, "chrom"))
            key_word_options.chromaticity_only =TRUE;
        else if (IC_STRCMP_EQ(key_word, "pp"))
            key_word_options.pre_processed_image = TRUE; 
        else if (IC_STRCMP_EQ(key_word, "illum"))
            key_word_options.illumination = TRUE; 
    }

    trim_beg(&arg);

    if (key_word_options.pre_processed_image)
    {
        const KJB_image* pp_ip; 

        NRE(pp_ip = get_pp_input_image());
        ERE(kjb_display_image(pp_ip, "Preprocessed image"));

        if (arg[ 0 ] == '\0') return NO_ERROR; 
    }

    if (arg[ 0 ] == '\0')
    {
        set_error("No display request.");
        return ERROR; 
    }

    count = 0;

    if ((key_word_options.collage_format) || (key_word_options.half_size))
    {
        reduction_flag = TRUE;
    }

    while (    (count < MAX_NUM_COLLAGE_IMAGES) 
            && (result != ERROR) 
            && ((BUFF_GET_TOKEN_OK(&arg, request)))
          )
    {
        image_map_data_ptr = create_image_map_data();

        if (image_map_data_ptr == NULL) 
        {
            result = ERROR;
        }
        else 
        {
            result = parse_image_request(request, result_ptr, 
                                         reduction_flag, 
                                         key_word_options.illumination, 
                                         key_word_options.chromaticity_only, 
                                         image_map_data_ptr,
                                         (Error_by_sharp*)NULL,
                                         (double**)NULL);
        }

        /*
        // Add to queue, even if result is ERROR, so that *image_map_data_ptr
        // will be de-allocated.
        */
        
        insert_at_end_of_queue(&image_map_data_queue_head, 
                               (Queue_element**)NULL, 
                               (void*)image_map_data_ptr);
    }

    if (result != ERROR)
    {
        if (key_word_options.collage_format)
        {
            result = display_collage(image_map_data_queue_head);
        }
        else
        {
            result = display_normal(image_map_data_queue_head);
        }
    }

    free_queue(&image_map_data_queue_head, (Queue_element**)NULL, 
               queue_free_image_map_data);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int display_collage(Queue_element* image_map_data_queue_head)
{ 
    Queue_element*  cur_elem;
    Image_map_data* image_map_data_ptr;
    KJB_image*    ip;
    KJB_image*    out_ip = NULL;
    /*
    // MAX_NUM_COLLAGE_IMAGES must be "roundable" UP to the nearest 
    // multiple of MAX_MIN_COLLAGE_DIM without exceeding the bounds of 
    // the ip_list array. 
    */
    KJB_image* ip_list[ MAX_NUM_COLLAGE_IMAGES + MAX_MIN_COLLAGE_DIM ];
    char         title[ 200 ];
    int          i;
    int          result = NO_ERROR;
    int          count;

    for (i=0; i < (int)(sizeof(ip_list)/sizeof(ip_list[ 0 ])); i++) 
    {
        ip_list[ i ] = NULL;
    }

    cur_elem = image_map_data_queue_head; 
    count = 0;
    title[ 0 ] = '\0'; 

    while (    (count < MAX_NUM_COLLAGE_IMAGES) 
            && (cur_elem != NULL) 
            && (result != ERROR) 
          ) 
    {
        ip = NULL;

        image_map_data_ptr = (Image_map_data*)(cur_elem->contents); 

        result = compute_result_image(&ip, image_map_data_ptr);

        if (result != ERROR) 
        {
            if (title[ 0 ] != '\0') BUFF_CAT(title, "|");
            BUFF_CAT(title, image_map_data_ptr->request);
            ip_list[count] = ip;
        }

        count++;
        cur_elem = cur_elem->next; 
    }

    if (result != ERROR) 
    {
        if (count < 1)
        {
            SET_CANT_HAPPEN_BUG();
            result = ERROR; 
        }
        else if (count == 1) 
        {
            result = make_image_collage(&out_ip, 1, 1, ip_list);
        }
        else if (count <= 2) 
        {
            result = make_image_collage(&out_ip, 1, 2, ip_list);
        }
        else if (count <= 4) 
        {
            result = make_image_collage(&out_ip, 2, 2, ip_list);
        }
        else if (count <= 6) 
        {
            result = make_image_collage(&out_ip, 3, 2, ip_list);
        }
        else if (count <= 9) 
        {
            result = make_image_collage(&out_ip, 3, 3, ip_list);
        }
        else if (count <= 12) 
        {
            result = make_image_collage(&out_ip, 4, 3, ip_list);
        }
        else if (count <= 15) 
        {
            result = make_image_collage(&out_ip, 5, 3, ip_list);
        }
        else
        {
            int num_rows = count / 4;

            result = make_image_collage(&out_ip, num_rows, 4, ip_list);
        }
    }

    if (result != ERROR) 
        result = kjb_display_image(out_ip, title);

    kjb_free_image(out_ip);

    for (i=0; i<count; i++) 
        kjb_free_image(ip_list[i]);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int display_normal(Queue_element* image_map_data_queue_head)
{ 
    Queue_element*  cur_elem;
    int             result          = NO_ERROR;
    KJB_image*    ip              = NULL;

    cur_elem = image_map_data_queue_head; 

    while ((cur_elem != NULL) && (result != ERROR)) 
    {
        Image_map_data* image_map_data_ptr;

        image_map_data_ptr = (Image_map_data*)(cur_elem->contents); 

        result = compute_result_image(&ip, image_map_data_ptr);

        if (result != ERROR)
        {
            result = kjb_display_image(ip, image_map_data_ptr->request); 
        }

        cur_elem = cur_elem->next; 
    }

    kjb_free_image(ip); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

