
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "illum_data.h"
#include "do_best_fit.h"

/* -------------------------------------------------------------------------- */

static int fs_do_best_2D_world_fits = FALSE; 
static int fs_do_best_2D_scene_fits = FALSE; 
static int fs_do_best_3D_world_fits = TRUE; 
static int fs_do_best_3D_scene_fits = FALSE; 

/* -------------------------------------------------------------------------- */

static int do_best_3D_fit_guts
(
    Matrix*  data_mp,
    Matrix*  canon_data_mp,
    Matrix** best_3D_fit_mpp,
    Vector** best_3D_diag_fit_vpp 
);

static int do_best_2D_fit_guts
(
    Matrix*  data_mp,
    Matrix*  canon_data_mp,
    Matrix** best_2D_fit_mpp,
    Vector** best_2D_diag_fit_vpp 
);

/* -------------------------------------------------------------------------- */

int set_best_fit_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_boolean_value;
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (match_pattern(lc_option, "do-2d-best-world-fits"))
         || (match_pattern(lc_option, "do-best-2d-world-fits"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-best-2d-world-fits = %s\n",
                    fs_do_best_2D_world_fits ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D world fits %s done.\n",
                    fs_do_best_2D_world_fits ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_best_2D_world_fits = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    (match_pattern(lc_option, "do-2d-best-scene-fits"))
         || (match_pattern(lc_option, "do-best-2d-scene-fits"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-best-2d-scene-fits = %s\n",
                    fs_do_best_2D_scene_fits ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D scene fits %s done.\n",
                    fs_do_best_2D_scene_fits ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_best_2D_scene_fits = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    (match_pattern(lc_option, "do-3d-best-world-fits"))
         || (match_pattern(lc_option, "do-best-3d-world-fits"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-best-3d-world-fits = %s\n",
                    fs_do_best_3D_world_fits ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D world fits %s done.\n",
                    fs_do_best_3D_world_fits ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_best_3D_world_fits = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    if (    (match_pattern(lc_option, "do-3d-best-scene-fits"))
         || (match_pattern(lc_option, "do-best-3d-scene-fits"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-best-3d-scene-fits = %s\n",
                    fs_do_best_3D_scene_fits ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D scene fits %s done.\n",
                    fs_do_best_3D_scene_fits ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_best_3D_scene_fits = temp_boolean_value;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_best_fit
(
    Best_fit_result* best_fit_ptr,
    int              num_data_points,
    int*             data_point_nums 
)
{
    Vector* actual_illum_vp = NULL;
    Vector* canon_illum_vp  = NULL;
    Matrix* data_mp         = NULL;
    Matrix* canon_data_mp   = NULL;
    int     result          = NO_ERROR;

    /*
    // NOTE : All the following routines are robust with respect to NULL's, 
    //        so we don't want to check the returns of the routines which get
    //        the data. 
    */
    (void)get_sharp_canon_illum(&canon_illum_vp);
    (void)get_sharp_actual_illum(&actual_illum_vp);

    if ((actual_illum_vp != NULL) && (canon_illum_vp != NULL)) 
    {
        result = divide_vectors(&(best_fit_ptr->actual_vp), canon_illum_vp, 
                                actual_illum_vp);
    }

    if (result != ERROR)
    {
        (void)get_sharp_registered_data(&data_mp); 
        (void)get_sharp_canon_data(&canon_data_mp);

        if (fs_do_best_3D_world_fits)
        {
            result = do_best_3D_fit_guts(data_mp, canon_data_mp, 
                                  &(best_fit_ptr->world_best_3D_fit_mp),
                                  &(best_fit_ptr->world_best_3D_diag_fit_vp));
        }

        if ((result != ERROR) && (fs_do_best_2D_world_fits))
        {
            result = do_best_2D_fit_guts(data_mp, canon_data_mp, 
                                  &(best_fit_ptr->world_best_2D_fit_mp),
                                  &(best_fit_ptr->world_best_2D_diag_fit_vp));
        }
    }

    if (    (result != ERROR) 
         && (fs_do_best_2D_scene_fits || fs_do_best_3D_scene_fits)
       )
    {
        result = get_sharp_selected_registered_data(&data_mp, 
                                                    num_data_points, 
                                                    data_point_nums); 
        if (result != ERROR)
        {
            result = get_sharp_selected_canon_data(&canon_data_mp,
                                                   num_data_points, 
                                                   data_point_nums);
        }

        if ((result != ERROR) && (fs_do_best_3D_scene_fits))
        {
            result = do_best_3D_fit_guts(data_mp, canon_data_mp, 
                                &(best_fit_ptr->scene_best_3D_fit_mp),
                                &(best_fit_ptr->scene_best_3D_diag_fit_vp));
            db_mat(best_fit_ptr->scene_best_3D_fit_mp);
        }

        if ((result != ERROR) && (fs_do_best_2D_scene_fits))
        {
            result = do_best_2D_fit_guts(data_mp, canon_data_mp, 
                                &(best_fit_ptr->scene_best_2D_fit_mp),
                                &(best_fit_ptr->scene_best_2D_diag_fit_vp));
        }
    }

    free_matrix(data_mp);
    free_matrix(canon_data_mp);
    free_vector(canon_illum_vp);
    free_vector(actual_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_best_3D_fit_guts
(
    Matrix*  data_mp,
    Matrix*  canon_data_mp,
    Matrix** best_3D_fit_mpp,
    Vector** best_3D_diag_fit_vpp 
)
{
    int res;

    if ((data_mp == NULL) || (canon_data_mp == NULL)) 
    {
        return NO_ERROR;
    }

    if (data_mp->num_rows != canon_data_mp->num_rows)
    {
        set_error("Inconsistant number of surfaces in data and canon data");
#ifdef TEST
        add_error("Data has %d surfaces; canon data has %d surfaces.\n",
                  data_mp->num_rows, canon_data_mp->num_rows);
#endif 
        return ERROR;
    }

    res = get_best_post_map(best_3D_fit_mpp, data_mp, canon_data_mp);

    if (res == ERROR) 
    {
        file_db_mat(data_mp);
        file_db_mat(canon_data_mp);
        return ERROR;
    }

    return get_best_diagonal_post_map(best_3D_diag_fit_vpp, data_mp, 
                                      canon_data_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_best_2D_fit_guts
(
    Matrix*  data_mp,
    Matrix*  canon_data_mp,
    Matrix** best_2D_fit_mpp,
    Vector** best_2D_diag_fit_vpp 
)
{
    Matrix* projected_data_mp       = NULL;
    Matrix* projected_canon_data_mp = NULL;
    Matrix* positive_data_mp        = NULL;
    Matrix* positive_canon_data_mp  = NULL;
    int     count                   = 0;
    int     i;
    int     res;

    if ((data_mp == NULL) || (canon_data_mp == NULL)) 
    {
        return NO_ERROR;
    }

    if (data_mp->num_rows != canon_data_mp->num_rows)
    {
        set_error("Inconsistant number of surfaces in data and canon data");
#ifdef TEST
        add_error("Data has %d surfaces; canon data has %d surfaces.\n",
                  data_mp->num_rows, canon_data_mp->num_rows);
#endif 
        return ERROR;
    }

    ERE(get_target_matrix(&positive_data_mp, data_mp->num_rows,
                          data_mp->num_cols)); 

    ERE(get_target_matrix(&positive_canon_data_mp, canon_data_mp->num_rows,
                          canon_data_mp->num_cols)); 

    for (i=0; i<data_mp->num_rows; i++) 
    {
        if (    (data_mp->elements[ i ][ 2 ] > MIN_X_DIV_B)
             && (canon_data_mp->elements[ i ][ 2 ] > MIN_X_DIV_B) 
           )
       {
           int j;

           for (j=0; j<3; j++) 
           {
               positive_data_mp->elements[ count ][ j ] = 
                                        data_mp->elements[ i ][ j ];
               positive_canon_data_mp->elements[ count ][ j ] = 
                                        canon_data_mp->elements[ i ][ j ];
           }
           count++; 
       }
    }

    if (count != positive_data_mp->num_rows) 
    {
        verbose_pso(2, "Rejecting %d points for best 2D fits (%d left).\n", 
                    positive_data_mp->num_rows - count, count);
    }

    positive_data_mp->num_rows = count; 
    positive_canon_data_mp->num_rows = count; 

    if (count > 1) 
    {
        ERE(get_projection_matrix(&projected_data_mp, positive_data_mp, 
                                  DIVIDE_BY_BLUE,
                                  TRUNCATE_INVALID_CHROMATICITY, 
                                  MAX_X_DIV_B,
                                  DBL_NOT_SET, DBL_NOT_SET, 
                                  (Matrix**)NULL, (Matrix**)NULL)); 

        ERE(get_projection_matrix(&projected_canon_data_mp, 
                                  positive_canon_data_mp,
                                  DIVIDE_BY_BLUE, 
                                  TRUNCATE_INVALID_CHROMATICITY,
                                  MAX_X_DIV_B, 
                                  DBL_NOT_SET, DBL_NOT_SET, 
                                  (Matrix**)NULL, (Matrix**)NULL)); 

        res = get_best_post_map(best_2D_fit_mpp, 
                                projected_data_mp, projected_canon_data_mp);
    }
    else 
    {
        res = ERROR; 

        set_error("Unable to get 2 corresponding data and ");
        cat_error("canon rows with B>0.");
    }

    /*
    // This result is not that important, even when we skipping on error, so
    // just print the error if it does not work out, and make the result
    // NO_ERROR. 
    */
    if (res == ERROR)
    {
        verbose_pso(1, "Skipping calculation of projected best map.");
        insert_error("Unable to get best 2D diagonal map.");
        add_error("Skipping calculation of projected best map.");

        kjb_print_error();

        file_db_mat(data_mp);
        file_db_mat(canon_data_mp);
        file_db_mat(projected_data_mp);
        file_db_mat(projected_canon_data_mp);
        file_db_mat(positive_data_mp);
        file_db_mat(positive_canon_data_mp);

        /*
        // Normally we would want to return ERROR if we are skipping on
        // error, but we use the best projected map results so infrequently
        // that even in those circumstances we want to continue. 
        */ 
        res = NO_ERROR; 
    }

    if ((projected_data_mp != NULL) && (projected_canon_data_mp != NULL))
    {
        int temp_res;

        temp_res = get_best_diagonal_post_map(best_2D_diag_fit_vpp,
                                              projected_data_mp,
                                              projected_canon_data_mp);

        /*
        // Again, this result is not that important, so just print the error
        // if it does not work out. 
        */
        if (temp_res == ERROR)
        {
            insert_error("Unable to get best 2D diagonal map.");
            kjb_print_error();

            file_db_mat(data_mp);
            file_db_mat(canon_data_mp);
            file_db_mat(projected_data_mp);
            file_db_mat(projected_canon_data_mp);
            file_db_mat(positive_data_mp);
            file_db_mat(positive_canon_data_mp);
        }
    }

    free_matrix(projected_data_mp);
    free_matrix(projected_canon_data_mp);
    free_matrix(positive_data_mp);
    free_matrix(positive_canon_data_mp);

    return res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

