
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "do_gamut_cc.h"

#define DO_3D_SCIA_AND_SCCA_BY_WEIGHTED_AVERAGE 

typedef enum Weight_function_id 
{
    DISABLED_WEIGHT_FUNCTION, 
    IDENTITY_WEIGHT_FUNCTION, 
    POWER_OF_MAP_NORM,
    POWER_OF_INV_NORM,
    POWER_OF_GEOMETRIC_MEAN,
    NUMBER_OF_WEIGHT_FUNCTIONS
}
Weight_function_id;

static const char* fs_weight_function_id_strings[ NUMBER_OF_WEIGHT_FUNCTIONS ] =
{
    "off", "identity", "power-of-map-norm", "power-of-inv-norm", "power-of-gm"
};

typedef struct Weight_function
{
    Weight_function_id id;
    double             parm;
}
Weight_function;

/* -------------------------------------------------------------------------- */

static int  fs_gamut_solution_resolution = 100;
static int  fs_do_weighted_averages      = FALSE;
static int  fs_do_icsa                   = FALSE;
static int  fs_do_scca                   = FALSE;
static int  fs_do_scia                   = FALSE;
static int  fs_do_intersections_with_luminance                 = FALSE;
static int  fs_do_sil_scla               = FALSE;
static int  fs_do_sil_lcsa               = FALSE;
static int  fs_do_sil_scca               = FALSE;
static int  fs_do_sil_scia               = FALSE;
static double fs_relative_luminance_error  = 0.02;
static double fs_lcsa_max_luminance;
static double fs_lcsa_min_luminance;
static double fs_constraint_min_rgb = MIN_RGB / 10000.0; 
static int  fs_compute_scia_points = TRUE; 
static int  fs_compute_scca_points = TRUE; 
static Weight_function* fs_scia_weight_functions = NULL;
static Weight_function* fs_scca_weight_functions = NULL;

/* -------------------------------------------------------------------------- */

static int get_weight_function
(
    const char*      weight_function_str,
    Weight_function* weight_function_ptr 
);

static int get_weight_function_id
(
    char*               id_buff,
    Weight_function_id* id_ptr 
);

static int get_weight_function_string
(
    Weight_function* weight_function_ptr,
    char*            weight_function_string,
    size_t           weight_function_string_size 
);

static int weight_function_has_parm(Weight_function_id id);

static int do_hi_gamut_cc
(
    int           prompt_to_increase_error_flag,
    Constraints*  constraint_ptr,
    Gamut_result* gamut_result_ptr,
    Hull_result*  hull_result_ptr,
    Gamut_error*  cpu_time_ptr 
);

static int get_max_volume_result
(
    Hull*    hp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
);

static int get_hull_average_result
(
    Hull*    hp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
);

static int get_weighted_average_result
(
    Hull*    hp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
);

static int get_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   result_hp_ptr 
);

static int get_2D_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   dummy_result_hp_ptr 
);

static int get_3D_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   result_hp_ptr 
);

static int get_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector*  result_vp_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ],
    Matrix** points_mpp,
    double   cpu_time_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ] 
);

static int get_2D_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
);

static int get_3D_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector*  result_vp_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ],
    Matrix** points_mpp,
    double   cpu_time_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ] 
);

static int get_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector*  result_vp_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ],
    Matrix** points_mpp,
    double   cpu_time_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ] 
);

static int get_2D_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
);

static int get_3D_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     base_cpu_time,
    Vector*  result_vp_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ],
    Matrix** points_mpp,
    double   cpu_time_array[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ] 
);

static int get_luminance_constrained_result
(
    Hull*   hp,
    Vector* canon_illum_vp,
    double  luminances[ MAX_NUM_LUMINANCE_METHODS ],
    long    base_cpu_time,
    Vector* ha_result_vp[ MAX_NUM_LUMINANCE_METHODS ],
    double  ha_cpu_times[ MAX_NUM_LUMINANCE_METHODS ],
    Vector* mv_result_vp[ MAX_NUM_LUMINANCE_METHODS ],
    double  mv_cpu_times[ MAX_NUM_LUMINANCE_METHODS ],
    Hull*   result_hp[ MAX_NUM_LUMINANCE_METHODS ] 
);

static int get_scla_result
(
    Hull*   expanded_hp,
    Hull*   illum_hp,
    Vector* canon_illum_vp,
    double  sum_luminances[ MAX_NUM_LUMINANCE_METHODS ],
    long    base_cpu_time,
    Vector* result_vp[ MAX_NUM_LUMINANCE_METHODS ],
    double  cpu_times[ MAX_NUM_LUMINANCE_METHODS ] 
);

static int get_lcsa_result
(
    Hull*   hp,
    Hull*   expanded_illum_hp,
    Vector* canon_illum_vp,
    double  luminances[ MAX_NUM_LUMINANCE_METHODS ],
    long    base_cpu_time,
    Vector* result_vp[ MAX_NUM_LUMINANCE_METHODS ],
    double  cpu_times[ MAX_NUM_LUMINANCE_METHODS ],
    Hull*   result_hp[ MAX_NUM_LUMINANCE_METHODS ] 
);

static int get_sil_scia_result
(
    Hull*   expanded_hp,
    Hull*   illum_hp,
    Vector* canon_illum_vp,
    double  sum_luminances[ MAX_NUM_LUMINANCE_METHODS ],
    long    base_cpu_time,
    Vector* result_vp_array[ MAX_NUM_LUMINANCE_METHODS ],
    double  cpu_time_array[ MAX_NUM_LUMINANCE_METHODS ] 
);

static int get_sil_scca_result
(
    Hull*   expanded_hp,
    Hull*   chrom_hp,
    Vector* canon_illum_vp,
    double  sum_luminances[ MAX_NUM_LUMINANCE_METHODS ],
    long    base_cpu_time,
    Vector* result_vp_array[ MAX_NUM_LUMINANCE_METHODS ],
    double  cpu_time_array[ MAX_NUM_LUMINANCE_METHODS ] 
);

static int get_3D_map_extent_for_illum
(
    Vector* illum_vp,
    Vector* canon_illum_vp,
    Hull*   map_hp,
    Vector* map_vp 
);

static double chromaticity_weight_fn(const Vector* vp);
static double rgb_weight_fn(const Vector* vp);

static int scca_map_fn(Vector** target_vpp, const Vector* vp);
static int icsa_map_fn(Vector** target_vpp, const Vector* vp);

static int do_gamut_invert_vector
(
    Vector**      target_vpp,
    const Vector* source_vp 
);

static int lcsa_map_fn(Vector** target_vpp, const Vector* source_vp);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_gamut_options(const char* option, const char* value)
{ 
    int  temp_boolean_value;
    int  temp_integer_value;
    char lc_option[ 100 ];
    int  result = NOT_FOUND;
    int  i, index;
    int  first_index;
    int  second_index; 

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "gamut-solution-resolution"))
         || (match_pattern(lc_option, "gamut-sol-res"))
         || (match_pattern(lc_option, "gsr"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("gamut-solution-resolution = %d.\n", 
                    fs_gamut_solution_resolution));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Gamut solution resolution is %d.\n", 
                    fs_gamut_solution_resolution));
        }
        else
        {
            ERE(ss1pi(value, &temp_integer_value));

            if (temp_integer_value < 1)
            {
                set_error("gamut-solution-resolution must be at least 1.\n");
                return ERROR;
            }

            fs_gamut_solution_resolution = temp_integer_value; 
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-wa"))
         || (match_pattern(lc_option, "do-weighted-average"))
         || (match_pattern(lc_option, "do-weighted-averages"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-weighted-averages = %s\n", 
                    fs_do_weighted_averages ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Weighted averages %s done.\n",
                    fs_do_weighted_averages ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_weighted_averages = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-icsa"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-icsa = %s\n", fs_do_icsa ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illum constrained surface averages %s done.\n",
                    fs_do_icsa ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_icsa = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-scca"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-scca = %s\n", fs_do_scca ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained chrom averages %s done.\n",
                    fs_do_scca ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_scca = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-scia"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-scia = %s\n", fs_do_scia ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained illum averages %s done.\n",
                    fs_do_scia ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_scia = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "compute-scia-points"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scia-points = %s\n", 
                    fs_compute_scia_points ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained illum averages points %s computed.\n",
                    fs_compute_scia_points ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scia_points = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "compute-scca-points"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("compute-scca-points = %s\n", 
                    fs_compute_scca_points ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained chrom averages points %s computed.\n",
                    fs_compute_scca_points ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_compute_scca_points = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-luminance"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-luminance = %s\n", fs_do_intersections_with_luminance ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Estimates involving luminance constrained maps %s done.\n",
                    fs_do_intersections_with_luminance ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_intersections_with_luminance = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-sil-scla"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-sil-scla = %s\n", fs_do_sil_scla ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface, illum, and lum scla %s done.\n",
                    fs_do_sil_scla ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_sil_scla = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-sil-lcsa"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-sil-lcsa = %s\n", fs_do_sil_lcsa ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface, illum, and lum lcsa %s done.\n",
                    fs_do_sil_lcsa ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_sil_lcsa = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-sil-scia"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-sil-scia = %s\n", fs_do_sil_scia ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface, illum, and lum scia %s done.\n",
                    fs_do_sil_scia ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_sil_scia = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "do-sil-scca"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("do-sil-scca = %s\n", fs_do_sil_scca ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface, illum, and lum scca %s done.\n",
                    fs_do_sil_scca ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_do_sil_scca = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "luminance-relative-error"))
        || (match_pattern(lc_option, "relative-luminance-error"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("relative-luminance-error = %.1f%%\n", 
                    100.0 * fs_relative_luminance_error));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Relative luminance error is %.1f%%.\n",
                    100.0 * fs_relative_luminance_error));
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 

            if (temp_real < 0.0)
            {
                set_error("Relative luminance error must be positive.\n");
                return ERROR;
            }
            fs_relative_luminance_error = temp_real;
        }
        result = NO_ERROR; 
    }
    
    if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "constraint-min-rgb"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("constraint-min-rgb = %f.\n", 
                    fs_constraint_min_rgb));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Constraint min rgb is %f.\n",
                    fs_constraint_min_rgb));
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 

            if (temp_real < 0.0)
            {
                set_error("Luminance relative error must be positive.\n");
                return ERROR;
            }
            fs_constraint_min_rgb = temp_real;
        }
        result = NO_ERROR; 
    }

    ERE(parse_index(lc_option, MAX_NUM_SCA_WEIGHT_FUNCTIONS,
                    &first_index, &second_index, &index, (int*)NULL));

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "scia-weight-function"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_SCA_WEIGHT_FUNCTIONS, 
                              "Surface constrained average functions"));

#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 

        if (fs_scia_weight_functions == NULL)
        {
            int k; 

            NRE(fs_scia_weight_functions = N_TYPE_MALLOC(Weight_function, 
                                                MAX_NUM_SCA_WEIGHT_FUNCTIONS));

            for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
            {
                fs_scia_weight_functions[ k ].id = DISABLED_WEIGHT_FUNCTION;
            }
        }
        
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            char weight_function_string[ 100 ];

            for (i = first_index; i <= second_index; i++)
            {
                ERE(get_weight_function_string(&(fs_scia_weight_functions[ i ]),
                                               weight_function_string, 
                                               sizeof(weight_function_string)));

                ERE(pso("scia-weight-function:%d = %s\n", i + 1,
                        weight_function_string)); 
            }
        }
        else if (value[ 0 ] == '\0') 
        {
            char weight_function_string[ 100 ];

            for (i = first_index; i <= second_index; i++)
            {
                ERE(get_weight_function_string(&(fs_scia_weight_functions[ i ]),
                                               weight_function_string, 
                                               sizeof(weight_function_string)));

                ERE(pso("Surface constrained illum ave weight function %d is %s.\n",
                         i + 1, weight_function_string));
            }
        }
        else
        {
            ERE(get_weight_function(value, &(fs_scia_weight_functions[ index ])));
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "scca-weight-function"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_SCA_WEIGHT_FUNCTIONS, 
                              "Surface constrained average functions"));

#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 

        if (fs_scca_weight_functions == NULL)
        {
            int k; 

            NRE(fs_scca_weight_functions = N_TYPE_MALLOC(Weight_function, 
                                                MAX_NUM_SCA_WEIGHT_FUNCTIONS));

            for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
            {
                fs_scca_weight_functions[ k ].id = DISABLED_WEIGHT_FUNCTION;
            }
        }
        
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            char weight_function_string[ 100 ];

            for (i = first_index; i <= second_index; i++)
            {
                ERE(get_weight_function_string(&(fs_scca_weight_functions[ i ]),
                                               weight_function_string, 
                                               sizeof(weight_function_string)));

                ERE(pso("scca-weight-function:%d = %s\n", i + 1,
                        weight_function_string)); 
            }
        }
        else if (value[ 0 ] == '\0') 
        {
            char weight_function_string[ 100 ];

            for (i = first_index; i <= second_index; i++)
            {
                ERE(get_weight_function_string(&(fs_scca_weight_functions[ i ]),
                                               weight_function_string, 
                                               sizeof(weight_function_string)));

                ERE(pso("Surface constrained chrom ave weight function %d is %s.\n",
                         i + 1, weight_function_string));
            }
        }
        else
        {
            ERE(get_weight_function(value, &(fs_scca_weight_functions[ index ])));
        }
        result = NO_ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_weight_function
(
    const char*      weight_function_str,
    Weight_function* weight_function_ptr 
)
{
    char weight_function_id_buff[ 100 ];
    char weight_function_str_cpy[ 1000 ]; 
    char* weight_function_str_pos;

    BUFF_CPY(weight_function_str_cpy, weight_function_str);
    weight_function_str_pos = weight_function_str_cpy;

    BUFF_GEN_GET_TOKEN(&weight_function_str_pos, weight_function_id_buff, ":");
    ERE(get_weight_function_id(weight_function_id_buff, 
                               &(weight_function_ptr->id))); 

    if (weight_function_has_parm(weight_function_ptr->id))
    {
        if (*weight_function_str_pos == ':')
        {
            weight_function_str_pos++;
            ERE(ss1d(weight_function_str_pos, &(weight_function_ptr->parm)));
        }
        else 
        {
            set_error("Function with id %q requires a parameter.",
                      weight_function_id_buff);
            return ERROR;
        }
    }
    else if (*weight_function_str_pos == ':')  
    {
        set_error("Function with id %q does not have a a parameter.",
                  weight_function_id_buff);
        return ERROR; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_weight_function_id
(
    char*               id_buff,
    Weight_function_id* id_ptr 
)
{
    char lc_id_buff[ 100 ];
    Weight_function_id  i;

    if (is_no_value_word(id_buff))
    {
        *id_ptr = DISABLED_WEIGHT_FUNCTION;
        return NO_ERROR; 
    }

    EXTENDED_LC_BUFF_CPY(lc_id_buff, id_buff);

    for (i=IDENTITY_WEIGHT_FUNCTION; i<NUMBER_OF_WEIGHT_FUNCTIONS; i++)
    {
        if (STRCMP_EQ(lc_id_buff, fs_weight_function_id_strings[ i ])) 
        {
            *id_ptr = i;
            return NO_ERROR; 
        }
    }

    set_error("%q is in invalid function type.", id_buff);
    return ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_weight_function_string
(
    Weight_function* weight_function_ptr,
    char*            weight_function_string,
    size_t           weight_function_string_size 
)
{

    kjb_strncpy(weight_function_string, 
                fs_weight_function_id_strings[ weight_function_ptr->id ],
                weight_function_string_size); 

    if (weight_function_has_parm(weight_function_ptr->id))
    {
        char buff[ 100 ];

        ERE(kjb_sprintf(buff, sizeof(buff), ":%.1f",
                        weight_function_ptr->parm));
        kjb_strncat(weight_function_string, buff, weight_function_string_size);
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int weight_function_has_parm(Weight_function_id id)
{

    if (    (id == DISABLED_WEIGHT_FUNCTION)
         || (id == IDENTITY_WEIGHT_FUNCTION)
       )
    {
        return FALSE;
    }
    else 
    {
        return TRUE;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_gamut_colour_constancy
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    Constraints*              constraint_ptr,
    Gamut_result*             gamut_result_ptr,
    Hull_result*              hull_result_ptr,
    Gamut_error*              cpu_time_ptr 
)
{
    int result;

    if (cc_methods_ptr->hull_intersection) 
    {
        result = do_hi_gamut_cc(prompt_to_increase_error_flag, constraint_ptr,
                                gamut_result_ptr, hull_result_ptr, 
                                cpu_time_ptr);

        if (result != NO_ERROR) 
        {
            return result;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_hi_gamut_cc
(
    int           prompt_to_increase_error_flag,
    Constraints*  constraint_ptr,
    Gamut_result* gamut_result_ptr,
    Hull_result*  hull_result_ptr,
    Gamut_error*  cpu_time_ptr 
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    Queue_element*      cur_elem;
    Queue_element*      save_cur_elem;
    int                 result                         = NO_ERROR;
    Hull*               surface_intersect_hp           = NULL;
    Hull*               surface_illum_intersect_hp     = NULL;
    int                 intersect_cpu_time; 

    if (io_atn_flag) 
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE; 
        return ERROR;
    }

    /* 
    // I don't think this can be used without being set, but if not, making it
    // negative should make it more obvious if it is. 
    */
    intersect_cpu_time = -100000; 

    /*
    // We return NO_SOLUTION if we fail due to hull intersection failure, as in
    // this case, we want to be able to increase the error and try again. Other
    // failures return ERROR. If there is no illum hull, this is not treated as
    // an error in order that the other results are calculated. In the case that
    // we wish to skip the other results, then the calling routine will makes
    // sure that we are not called if there is no illum hull. (This is a bit of
    // spagetti, sorry). 
    // 
    // I have also coded this for the case that there is an illum hull, but no
    // surface hull, but this is not currently used. (It has been tested by
    // setting constraint_ptr->surface_hull_list_head to null in this module,
    // and seems to work). 
    */
    if (constraint_ptr->surface_hull_list_head != NULL) 
    {
        if (prompt_to_increase_error_flag)
        {
            interactive_pso("Intersection of surface hulls.\n");
        }

        ERE(init_cpu_time());

        cur_elem = constraint_ptr->surface_hull_list_head;
        while(cur_elem != NULL)
        {
            cur_elem = cur_elem->next;
        }

        result = intersect_hulls(constraint_ptr->surface_hull_list_head,
                                 DEFAULT_HULL_OPTIONS,
                                 &surface_intersect_hp);

        if (result == ERROR)
        {
            add_error("Computation of intersection of surface hulls failed.");
            return ERROR;
        }

        hull_result_ptr->surface_hp = surface_intersect_hp;

        ERE(intersect_cpu_time = get_cpu_time());

        if (result == NO_SOLUTION) 
        {
            if (prompt_to_increase_error_flag)
            {
                interactive_pso("    Failure.\n");
            }
        }
        else 
        {
            interactive_pso("Finding solutions within surface constraint result.\n");
        
            ERE(get_max_volume_result(surface_intersect_hp, 
                                      intersect_cpu_time, 
                                      &(gamut_result_ptr->mv_surface_vp), 
                                      &(cpu_time_ptr->mv_surface))); 
                                      
            ERE(get_hull_average_result(surface_intersect_hp, 
                                        intersect_cpu_time, 
                                        &(gamut_result_ptr->ha_surface_vp), 
                                        &(cpu_time_ptr->ha_surface))); 

            ERE(get_weighted_average_result(surface_intersect_hp,
                                            intersect_cpu_time,
                                            &(gamut_result_ptr->wa_surface_vp),
                                            &(cpu_time_ptr->wa_surface)));
                                            
            ERE(get_luminance_constrained_result(surface_intersect_hp,
                                                 constraint_ptr->canon_illum_vp,
                                                 constraint_ptr->sum_luminance,
                                                 intersect_cpu_time,
                                   gamut_result_ptr->ha_surface_lum_vp,
                                   cpu_time_ptr->ha_surface_lum,
                                   gamut_result_ptr->mv_surface_lum_vp,
                                   cpu_time_ptr->mv_surface_lum,
                                   hull_result_ptr->luminance_surface_hp));

            if (constraint_ptr->illum_hp != NULL)
            {
                IMPORT double min_relative_diag_error;
                Hull*         expanded_illum_hp;
                Hull*         expanded_surface_intersect_hp;

                expanded_illum_hp = expand_hull(constraint_ptr->illum_hp, 
                                                min_relative_diag_error,
                                                DEFAULT_HULL_OPTIONS);

                expanded_surface_intersect_hp = 
                                    expand_hull(surface_intersect_hp, 
                                                min_relative_diag_error,
                                                DEFAULT_HULL_OPTIONS);

                if (    (expanded_illum_hp == NULL)
                     || (expanded_surface_intersect_hp == NULL)
                   )
                {
                    result = ERROR; 
                }

                if (result != ERROR) 
                {
                    result = get_icsa_result(surface_intersect_hp, 
                                             expanded_illum_hp, 
                                             constraint_ptr->canon_illum_vp, 
                                             intersect_cpu_time,
                              &(gamut_result_ptr->icsa_surface_illum_vp),
                              &(cpu_time_ptr->icsa_surface_illum),
                              &(hull_result_ptr->illum_constrained_surface_hp));
                }

                if (result != ERROR) 
                {
                    result = get_scla_result(expanded_surface_intersect_hp, 
                                             constraint_ptr->illum_hp, 
                                             constraint_ptr->canon_illum_vp, 
                                             constraint_ptr->sum_luminance, 
                                             intersect_cpu_time,
                                    gamut_result_ptr->scla_surface_illum_lum_vp,
                                    cpu_time_ptr->scla_surface_illum_lum);
                }

                if (result != ERROR) 
                {
                    result = get_lcsa_result(surface_intersect_hp, 
                                             expanded_illum_hp, 
                                             constraint_ptr->canon_illum_vp, 
                                             constraint_ptr->sum_luminance, 
                                             intersect_cpu_time,
                             gamut_result_ptr->lcsa_surface_illum_lum_vp,
                             cpu_time_ptr->lcsa_surface_illum_lum,
                             hull_result_ptr->luminance_constrained_surface_hp);

                }

                if (result != ERROR) 
                {
                    result = get_scia_result(expanded_surface_intersect_hp, 
                                             constraint_ptr->illum_hp, 
                                             constraint_ptr->canon_illum_vp, 
                                             intersect_cpu_time,
                                    gamut_result_ptr->scia_surface_illum_vp,
                                    &(hull_result_ptr->scia_points_mp), 
                                    cpu_time_ptr->scia_surface_illum);
                }

                if (result != ERROR) 
                {
                    result = get_scca_result(expanded_surface_intersect_hp, 
                                             constraint_ptr->illum_hp, 
                                             constraint_ptr->canon_illum_vp, 
                                             intersect_cpu_time,
                                    gamut_result_ptr->scca_surface_illum_vp,
                                    &(hull_result_ptr->scca_points_mp), 
                                    cpu_time_ptr->scca_surface_illum);
                }

                if (result != ERROR) 
                {
                    result = get_sil_scia_result(expanded_surface_intersect_hp, 
                                                 constraint_ptr->illum_hp, 
                                                 constraint_ptr->canon_illum_vp,
                                                 constraint_ptr->sum_luminance, 
                                                 intersect_cpu_time,
                                    gamut_result_ptr->scia_surface_illum_lum_vp,
                                    cpu_time_ptr->scia_surface_illum_lum);
                }

                if (result != ERROR) 
                {
                    result = get_sil_scca_result(expanded_surface_intersect_hp, 
                                                 constraint_ptr->illum_hp, 
                                                 constraint_ptr->canon_illum_vp,
                                                 constraint_ptr->sum_luminance, 
                                                 intersect_cpu_time,
                                    gamut_result_ptr->scca_surface_illum_lum_vp,
                                    cpu_time_ptr->scca_surface_illum_lum);
                }

                free_hull(expanded_illum_hp);
                free_hull(expanded_surface_intersect_hp);
            }
        }
    }

    if (io_atn_flag) 
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE; 
        result = ERROR;
    }

    /* We used to set this to NULL, but I think that it has to be NULL
    // already.
    */
    ASSERT(hull_result_ptr->surface_illum_hp == NULL); 

    if (    (result == NO_ERROR)
         && (constraint_ptr->illum_hull_list_head != NULL)
       )
    {
        Queue_element* surface_illum_hull_list_head = NULL;

        if (prompt_to_increase_error_flag)
        {
            interactive_pso("Intersection of surface and illumination hulls.\n");
        }

        /*
        // This covers the case that we only have an illum hull, but no surface
        // hull (not relavent yet).
        */
        if (surface_intersect_hp != NULL)
        {
            ERE(insert_into_queue(&surface_illum_hull_list_head,
                                  (Queue_element**)NULL, 
                                  (void*)surface_intersect_hp));
        }

        ERE(insert_into_queue(&surface_illum_hull_list_head,
                              (Queue_element**)NULL, 
                              constraint_ptr->illum_hull_list_head->contents));

        ERE(init_cpu_time());

        result = intersect_hulls(surface_illum_hull_list_head,
                                 DEFAULT_HULL_OPTIONS,
                                 &surface_illum_intersect_hp);

        if (result == ERROR)
        {
            add_error("Computation of intersection of surface hulls failed.");
        }

        cur_elem = surface_illum_hull_list_head;

        while (cur_elem != (Queue_element *)NULL) 
        {
            save_cur_elem = cur_elem;
            cur_elem = cur_elem->next;
            kjb_free(save_cur_elem);
        }

        if (result == NO_ERROR) 
        {
            hull_result_ptr->surface_illum_hp = surface_illum_intersect_hp;
            ERE(intersect_cpu_time = get_cpu_time());
        }

        if (    (result == NO_SOLUTION) 
             && (prompt_to_increase_error_flag)
           )
        {
            interactive_pso("    Failure.\n");
        }
    }

    if ((result == NO_ERROR) && (surface_illum_intersect_hp != NULL)) 
    {
        interactive_pso("Finding solutions within surface and illum constraints.\n");
        
        ERE(get_max_volume_result(surface_illum_intersect_hp, 
                                  intersect_cpu_time, 
                                  &(gamut_result_ptr->mv_surface_illum_vp), 
                                  &(cpu_time_ptr->mv_surface_illum)));

        ERE(get_hull_average_result(surface_illum_intersect_hp,
                                    intersect_cpu_time, 
                                    &(gamut_result_ptr->ha_surface_illum_vp), 
                                    &(cpu_time_ptr->ha_surface_illum))); 

        ERE(get_weighted_average_result(surface_illum_intersect_hp,
                                       intersect_cpu_time,
                                       &(gamut_result_ptr->wa_surface_illum_vp),
                                       &(cpu_time_ptr->wa_surface_illum)));
                                        
        ERE(get_luminance_constrained_result(surface_illum_intersect_hp,
                                             constraint_ptr->canon_illum_vp, 
                                             constraint_ptr->sum_luminance,
                                             intersect_cpu_time,
                               gamut_result_ptr->ha_surface_illum_lum_vp,
                               cpu_time_ptr->ha_surface_illum_lum,
                               gamut_result_ptr->mv_surface_illum_lum_vp,
                               cpu_time_ptr->mv_surface_illum_lum,
                               hull_result_ptr->luminance_surface_illum_hp));
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_max_volume_result
(
    Hull*    hp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
)
{
    double product;
    long total_cpu_time;
    int  result;

    interactive_pso("    Finding max volume result.\n");
    
    ERE(init_cpu_time());

    result = find_hull_max_coord_product(hp, result_vpp, &product);
    /*
    dbe(product);
    db_rv(*result_vpp); 
    */

    /*
    result = estimate_hull_max_coord_product(hp, result_vpp, &product);
    dbe(product);
    db_rv(*result_vpp); 
    */

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 

    /*
    // Actually, I don't think we want to entertain the idea of NO_SOLUTION
    // here. In fact, it should not happen. 
    */

    if (result == NO_SOLUTION)
    {
        set_error("Switching NO_SOLUTION to ERROR in get_max_volume_result.");
        result = ERROR;
    }

    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_hull_average_result
(
    Hull*    hp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
)
{
    long total_cpu_time;
    int  result;

    interactive_pso("    Finding hull average result.\n");
    
    ERE(init_cpu_time());

    result = get_hull_CM_and_volume(hp, result_vpp, (double*)NULL);

    /*
    // result can be ERROR, NO_ERROR, or NO_SOLUTION 
    //
    // Actually, I don't think we want to entertain the idea of NO_SOLUTION
    // here. In fact, it should not happen. 
    */
    if (result == NO_SOLUTION)
    {
        set_error("Switching NO_SOLUTION to ERROR in get_hull_average_result.");
        result = ERROR;
    }

    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR) 
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_weighted_average_result
(
    Hull*    hp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
)
{
    long   total_cpu_time;
    double (*weight_fn)(const Vector*);
    int    result;

    if (!fs_do_weighted_averages) return NO_ERROR; 

    interactive_pso("    Finding weighted hull average result.\n");

    ERE(init_cpu_time());

    if (hp->dimension == 2) 
    {
        weight_fn = chromaticity_weight_fn;
    }
    else 
    {
        weight_fn = rgb_weight_fn;
    }

    result = find_weighted_hull_average(hp, weight_fn, result_vpp);

    /*
    // Actually, I don't think we want to entertain the idea of NO_SOLUTION
    // here. In fact, it should not happen. 
    */

    if (result == NO_SOLUTION)
    {
        set_error("Switching NO_SOLUTION to ERROR in get_weighted_average_result.");
        result = ERROR;
    }

    /* result can be ERROR, NO_ERROR */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   result_hp_ptr 
)
{
    if (( ! fs_do_icsa) || (expanded_illum_hp == NULL) || (canon_illum_vp == NULL))
    {
        return NO_ERROR;
    }

    interactive_pso("    "); 
    interactive_pso("Finding illum constrained average of surface result.\n");

    if (hp->dimension == 2) 
    {
        return get_2D_icsa_result(hp, expanded_illum_hp, canon_illum_vp, 
                                  base_cpu_time, result_vpp, cpu_time_ptr, 
                                  result_hp_ptr);
    }
    else if (hp->dimension == 3) 
    {
        return get_3D_icsa_result(hp, expanded_illum_hp, canon_illum_vp, 
                                  base_cpu_time, result_vpp, cpu_time_ptr, 
                                  result_hp_ptr);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// We integrate 2D solutions over the surface of a sphere in map space.
*/

/*ARGSUSED*/
static int get_2D_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   dummy_result_hp_ptr 
)
{
    int       i, j;
    int       count        = 0;
    int       count_2        = 0;
    double      theta;
    double      phi;
    double      theta_step;
    double      phi_step;
    long_double theta_sum    = 0.0;
    long_double phi_sum      = 0.0;
    double      xy_r;
    Vector*   test_vp      = NULL;
    double      max_phi      = 0.01;
    double      max_theta    = 0.01;
    double      min_phi      = M_PI / 2.0;
    double      min_theta    = M_PI / 2.0;
    double      weight;
    double      weight_sum = 0.0;
    double      u, v, w, x, y, z, sum;
    double      c_r, c_g, c_b;
    long      total_cpu_time;
    Vector*   result_vp = NULL; 
    int       result;

    ERE(init_cpu_time());

    ERE(fill_hull(expanded_illum_hp));
    ERE(fill_hull(hp)); 

    c_r = canon_illum_vp->elements[ 0 ];
    c_g = canon_illum_vp->elements[ 1 ];
    c_b = canon_illum_vp->elements[ 2 ];

    theta_step = (max_theta - min_theta) / (fs_gamut_solution_resolution - 1);
    phi_step   = (max_phi - min_phi)     / (fs_gamut_solution_resolution - 1);
    theta      = min_theta;

    result = get_target_vector(&test_vp, 2); 

    for (i=0; i<fs_gamut_solution_resolution; i++) 
    {
        if (result == ERROR) break;

        phi = min_phi;

        for (j=0; j<fs_gamut_solution_resolution; j++) 
        {
            int inside_hull; 

            xy_r = sin(phi);  /* xy_r = r * sin(phi), but r is 1 */
            x = xy_r * cos(theta);
            y = xy_r * sin(theta);
            z = cos(phi);     /*  z = r * cos(phi), but r is 1 */ 

            test_vp->elements[ 0 ] = x / z;
            test_vp->elements[ 1 ] = y / z;

            inside_hull = is_point_inside_hull(hp, test_vp);

            if (inside_hull == ERROR)
            {
                result = ERROR;
                break;
            }
            else if (inside_hull == TRUE)
            {
                count_2++;

                u = c_r / x;
                v = c_g / y; 
                w = c_b / z; 
                sum = u + v + w; 

                test_vp->elements[ 0 ] = u / sum;
                test_vp->elements[ 1 ] = v / sum;

                inside_hull = is_point_inside_hull(expanded_illum_hp, test_vp);

                if (inside_hull == ERROR)
                {
                    result = ERROR;
                    break;
                }
                else if (inside_hull == TRUE)
                {
                    weight = sin(phi); 
                    count++;
                    theta_sum += theta * weight;
                    phi_sum += phi * weight;
                    weight_sum += weight; 
                }
            }
            phi += phi_step;
        }
        theta += theta_step; 

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }
    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) result = get_target_vector(&result_vp, 3); 

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        theta = theta_sum / weight_sum; 
        phi = phi_sum / weight_sum;
        xy_r = sin(phi);

        result_vp->elements[ 0 ] = xy_r * cos(theta);
        result_vp->elements[ 1 ] = xy_r * sin(theta);
        result_vp->elements[ 2 ] = cos(phi);
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    /*
    // Convert to 2D. It is a mistake to leave the result as 3D, as the
    // dimension of the vector implicitly contains the information as to wether
    // this is a 2D or a 3D algorithm. 
    */ 
    if (result == NO_ERROR) 
    {
        result = project_vector(result_vpp, result_vp, DIVIDE_BY_BLUE);
    }

    free_vector(result_vp); 
    free_vector(test_vp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
//    Average in surface map space, constrained by the illumination. This is
//    very similar to 3D_ha_surface_illum, but here the non-convexity of the
//    illumination hull is taken into account. For convenience in
//    implementation, we integrate in diagonal map space normalized by the
//    canonical illuminant, which is the inverted illumination space. However,
//    this should be no different than integrating in the map space. 
*/

static int get_3D_icsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr,
    Hull**   result_hp_ptr 
)
{
    Matrix* inverted_illum_mp        = NULL;
    Hull*   inverted_illum_hp        = NULL;
    Matrix* result_mp                = NULL;
    Hull*   inverted_illum_result_hp = NULL;
    long    total_cpu_time;
    int     result;

    ERE(init_cpu_time());

    result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                          hp->vertex_mp,
                                          canon_illum_vp);
    if (result != ERROR) 
    {
        inverted_illum_hp = find_convex_hull(inverted_illum_mp,
                                            DEFAULT_HULL_OPTIONS);
        if (inverted_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR)
    {
        result = find_constrained_hull_average(inverted_illum_hp,
                                               expanded_illum_hp, 
                                               icsa_map_fn, NO_HULL_WEIGHT_FN, 
                                               result_vpp, 
                                               (Vector**)NULL, 
                                               &inverted_illum_result_hp);
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        result = multiply_matrix_by_row_vector_ew(&result_mp,
                                          inverted_illum_result_hp->vertex_mp,
                                          canon_illum_vp);
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        free_hull(*result_hp_ptr); /* Likely not necessary. */
        *result_hp_ptr = find_convex_hull(result_mp, DEFAULT_HULL_OPTIONS);
        if ((*result_hp_ptr) == NULL) result = ERROR;
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {  
        ow_multiply_vectors(*result_vpp, canon_illum_vp);
    }
        
    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    free_matrix(inverted_illum_mp); 
    free_hull(inverted_illum_hp);
    free_hull(inverted_illum_result_hp);
    free_matrix(result_mp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
//    Average in illum space. The average is done in polar coordinates (phi and
//    theta), of all illuminants in the illuminant set which correspond to valid
//    mappings with respect to the surface constraints. 
*/
static int get_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vp_array,
    Matrix** points_mpp,
    double*  cpu_time_array 
)
{
    if (( ! fs_do_scia) || (illum_hp == NULL) || (canon_illum_vp == NULL))
    {
        return NO_ERROR;
    }

    interactive_pso("    "); 
    interactive_pso("Finding surface constrained average of illum result.\n");

    if (expanded_hp->dimension == 2)
    {
        return get_2D_scia_result(expanded_hp, illum_hp, canon_illum_vp, 
                                  base_cpu_time, &(result_vp_array[ 0 ]),
                                  &(cpu_time_array[ 0 ]));
    }
    else 
    {
        return get_3D_scia_result(expanded_hp, illum_hp, canon_illum_vp, 
                                  base_cpu_time, result_vp_array,
                                  points_mpp, cpu_time_array);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
)
{
    int         i, j;
    int         count       = 0;
    int         count_2     = 0;
    double      theta;
    double      phi;
    double      theta_step;
    double      phi_step;
    long_double theta_sum   = 0.0;
    long_double phi_sum     = 0.0;
    double      xy_r;
    Vector*     test_vp     = NULL;
    double      max_phi     = 0.0;
    double      max_theta   = 0.0;
    double      min_phi     = M_PI / 2.0;
    double      min_theta   = M_PI / 2.0;
    double      weight;
    double      weight_sum  = 0.0;
    double      x, y, z, sum, c_r, c_g, c_b;
    double      c_r_div_c_b;
    double      c_g_div_c_b;
    long        total_cpu_time;
    Vector*     result_vp      = NULL;
    int         result;

    ERE(init_cpu_time());

    result = get_target_vector(&test_vp, 2); 

    if (result != ERROR) 
    {
        result = fill_hull(illum_hp); 
    }

    if (result != ERROR) 
    {
        fill_hull(expanded_hp); 
    }

    c_r = canon_illum_vp->elements[ 0 ];
    c_g = canon_illum_vp->elements[ 1 ];
    c_b = canon_illum_vp->elements[ 2 ];
    c_r_div_c_b = c_r / c_b; 
    c_g_div_c_b = c_g / c_b; 

    for (i = 0; i<illum_hp->vertex_mp->num_rows; i++) 
    {
        double R; 

        if (result == ERROR) break;

        x = illum_hp->vertex_mp->elements[ i ][ 0 ];
        y = illum_hp->vertex_mp->elements[ i ][ 1 ];
        z = 1.0 - x - y;

        R     = sqrt(x*x + y*y + z*z);
        phi   = acos(z/R);
        xy_r  = sqrt(x*x + y*y);
        theta = acos(x/xy_r);

        if (theta < min_theta) min_theta = theta;
        if (phi   < min_phi)   min_phi   = phi;
        if (theta > max_theta) max_theta = theta;
        if (phi   > max_phi)   max_phi   = phi;
    }

    theta_step = (max_theta - min_theta) / (fs_gamut_solution_resolution - 1);
    phi_step   = (max_phi - min_phi)     / (fs_gamut_solution_resolution - 1);
    theta      = min_theta;

    for (i=0; i<fs_gamut_solution_resolution; i++) 
    {
        if (result == ERROR) break;

        phi = min_phi;

        for (j=0; j<fs_gamut_solution_resolution; j++) 
        {
            int inside_hull;

            if (result == ERROR) break;

            xy_r = sin(phi);  /* xy_r = r * sin(phi), but r is 1 */
            x = xy_r * cos(theta);
            y = xy_r * sin(theta);
            z = cos(phi);     /* z = r * cos(phi), but r is 1 */ 
            sum = x + y + z;

            test_vp->elements[ 0 ] = x / sum;
            test_vp->elements[ 1 ] = y / sum;

            inside_hull = is_point_inside_hull(illum_hp, test_vp);

            if (inside_hull == ERROR)
            {
                result = ERROR;
                break;
            }
            else if (inside_hull == TRUE)
            {
                count_2++;

                test_vp->elements[ 0 ] = c_r_div_c_b / (x / z);
                test_vp->elements[ 1 ] = c_g_div_c_b / (y / z);

                inside_hull = is_point_inside_hull(expanded_hp, test_vp);

                if (inside_hull == ERROR)
                {
                    result = ERROR;
                    break;
                }
                else if (inside_hull == TRUE)
                {
                    weight = sin(phi); 
                    count++;
                    theta_sum += theta * weight;
                    phi_sum += phi * weight;
                    weight_sum += weight; 
                }
            }
            phi += phi_step;

            if (io_atn_flag) 
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE; 
                result = ERROR;
            }

        }
        theta += theta_step; 

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }
    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) result = get_target_vector(&result_vp, 3); 

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        theta = theta_sum / weight_sum; 
        phi = phi_sum / weight_sum;
        xy_r = sin(phi);

        result_vp->elements[ 0 ] = c_r / (xy_r * cos(theta));
        result_vp->elements[ 1 ] = c_g / (xy_r * sin(theta));
        result_vp->elements[ 2 ] = c_b / (cos(phi));
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    /*
    // Convert to 2D. It is a mistake to leave the result as 3D, as the
    // dimension of the vector implicitly contains the information as to wether
    // this is a 2D or a 3D algorithm. 
    */ 
    if (result == NO_ERROR) 
    {
        result = project_vector(result_vpp, result_vp, DIVIDE_BY_BLUE);
    }

    free_vector(result_vp); 
    free_vector(test_vp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_3D_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vp_array,
    Matrix** points_mpp,
    double*  cpu_time_array 
)
{
    int         i, j, k;
    int         count      = 0;
    int         count_2    = 0;
    double        theta;
    double        phi;
    double        theta_step;
    double        phi_step;
    long_double   theta_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    long_double   phi_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    double        xy_r;
    Vector*     test_vp   = NULL;
    double        max_phi   = 0.0;
    double        max_theta = 0.0;
    double        min_phi   = M_PI / 2.0;
    double        min_theta = M_PI / 2.0;
    double        weight; 
    long_double   weight_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    Hull*       projected_inv_illum_hp     = NULL;
    Matrix*     projected_inv_illum_mp = NULL;
    Matrix*     inverted_illum_mp = NULL;
    double        x, y, z, sum, c_r, c_g, c_b;
    Vector*     result_vp      = NULL;
    int         result;
    Vector*     map_vp     = NULL;
    Vector*     inv_vp     = NULL;
    Matrix*     temp_vertex_mp = NULL;
    Vector*     illum_vp = NULL; 

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    if (fs_scia_weight_functions == NULL)
    {
        NRE(fs_scia_weight_functions = N_TYPE_MALLOC(Weight_function, 
                                                MAX_NUM_SCA_WEIGHT_FUNCTIONS));

        /* Set some defaults. */
        fs_scia_weight_functions[ 0 ].id = POWER_OF_GEOMETRIC_MEAN;
        fs_scia_weight_functions[ 0 ].parm = 3.0;

        for (k=1; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
        {
            fs_scia_weight_functions[ k ].id = DISABLED_WEIGHT_FUNCTION;
        }
    }
        
    ERE(init_cpu_time());

    for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
    {
        theta_sum[ k ]  = 0.0; 
        phi_sum[ k ]    = 0.0; 
        weight_sum[ k ] = 0.0; 
    }

    result = get_target_vector(&test_vp, 2); 

    if (result != ERROR) result = fill_hull(illum_hp); 

    if ((result != ERROR) && (fs_compute_scia_points))
    {
        int resolution_2 = fs_gamut_solution_resolution*fs_gamut_solution_resolution;

        result = get_target_matrix(points_mpp, resolution_2, 3);
    }

    if (result != ERROR)
    {
        result = remove_origin_from_data(&temp_vertex_mp, 
                                         expanded_hp->vertex_mp,
                                         MIN_RGB / (2.0 * MAX_RGB));
    }

    if (result != ERROR)
    {
        result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                              temp_vertex_mp,
                                              canon_illum_vp);
    }

    if (result != ERROR)
    {
        result = ow_min_thresh_matrix_col(inverted_illum_mp, BLUE_INDEX, 
                                          MIN_RGB / MAX_RGB);
    }

    free_matrix(temp_vertex_mp);

    if (result != ERROR)
    {
        result = get_projection_matrix(&projected_inv_illum_mp, 
                                       inverted_illum_mp,
                                       DIVIDE_BY_BLUE, 
                                       ALL_CHROMATICITIES_MUST_BE_VALID,
                                       2.0 * MAX_RGB / MIN_RGB, 
                                       DBL_NOT_SET, DBL_NOT_SET, 
                                       (Matrix **)NULL, (Matrix **)NULL);
        if (result == ERROR)
        {
            add_error("Problem occured while computing 3D-SCIA result.");
        }
    }

    if (result != ERROR) 
    {
        projected_inv_illum_hp = find_convex_hull(projected_inv_illum_mp,
                                                  DEFAULT_HULL_OPTIONS);
        if (projected_inv_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR) result = fill_hull(projected_inv_illum_hp); 
    if (result != ERROR) result = get_target_vector(&map_vp, 3);
    if (result != ERROR) result = get_target_vector(&inv_vp, 3);
    if (result != ERROR) result = get_target_vector(&illum_vp, 3);

    c_r = canon_illum_vp->elements[ 0 ];
    c_g = canon_illum_vp->elements[ 1 ];
    c_b = canon_illum_vp->elements[ 2 ];

    for (i = 0; i<illum_hp->vertex_mp->num_rows; i++) 
    {
        double R; 

        if (result == ERROR) break;

        x = illum_hp->vertex_mp->elements[ i ][ 0 ];
        y = illum_hp->vertex_mp->elements[ i ][ 1 ];
        z = 1.0 - x - y;

        R     = sqrt(x*x + y*y + z*z);
        phi   = acos(z/R);
        xy_r  = sqrt(x*x + y*y);
        theta = acos(x/xy_r);

        if (theta < min_theta) min_theta = theta;
        if (phi   < min_phi)   min_phi   = phi;
        if (theta > max_theta) max_theta = theta;
        if (phi   > max_phi)   max_phi   = phi;
    }

    theta_step = (max_theta - min_theta) / (fs_gamut_solution_resolution - 1);
    phi_step   = (max_phi - min_phi)     / (fs_gamut_solution_resolution - 1);
    theta      = min_theta;

    for (i=0; i<fs_gamut_solution_resolution; i++) 
    {
        if (result == ERROR) break;

        phi = min_phi;

        for (j=0; j<fs_gamut_solution_resolution; j++) 
        {
            int inside_hull;

            if (result == ERROR) break;

            xy_r = sin(phi);  /* xy_r = r * sin(phi), but r is 1 */
            x = xy_r * cos(theta);
            y = xy_r * sin(theta);
            z = cos(phi);     /* z = r * cos(phi), but r is 1 */ 
            sum = x + y + z;

            test_vp->elements[ 0 ] = x / sum;
            test_vp->elements[ 1 ] = y / sum;

            inside_hull = is_point_inside_hull(illum_hp, test_vp);

            if (inside_hull == ERROR)
            {
                result = ERROR;
                break;
            }
            else if (inside_hull == TRUE)
            {
                count_2++;

                test_vp->elements[ 0 ] = z / x;
                test_vp->elements[ 1 ] = z / y;

                inside_hull = is_point_inside_hull(projected_inv_illum_hp, 
                                                   test_vp);

                if (inside_hull == ERROR)
                {
                    result = ERROR;
                    break;
                }
                else if (inside_hull == TRUE)
                {
                    illum_vp->elements[ 0 ] = x; 
                    illum_vp->elements[ 1 ] = y; 
                    illum_vp->elements[ 2 ] = z; 

                    result = get_3D_map_extent_for_illum(illum_vp, 
                                                         canon_illum_vp,
                                                         expanded_hp,
                                                         map_vp); 
                    
                    if (result == ERROR) break; 

                    result = divide_vectors(&inv_vp, map_vp, canon_illum_vp);
                    if (result == ERROR) break; 

                    for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
                    {
                        if (fs_scia_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
                        {
                            weight = xy_r;   /* xy_r == sin(phi) */

                            if (fs_scia_weight_functions[ k ].id == POWER_OF_MAP_NORM)
                            {
                                double map_R = vector_magnitude(map_vp);
                                double exponant = fs_scia_weight_functions[k].parm;

                                weight *= pow(map_R, exponant); 
                            }
                            else if (fs_scia_weight_functions[ k ].id == POWER_OF_INV_NORM)
                            {
                                double map_R = vector_magnitude(inv_vp);
                                double exponant = fs_scia_weight_functions[k].parm;

                                weight *= pow(map_R, exponant); 
                            }
                            else if (fs_scia_weight_functions[ k ].id == POWER_OF_GEOMETRIC_MEAN)
                            {
                                double exponant = fs_scia_weight_functions[k].parm;
                                double map_prod = map_vp->elements[ 0 ]
                                                    * map_vp->elements[ 1 ]
                                                    * map_vp->elements[ 2 ];

                                weight *= pow(map_prod, exponant / 3.0); 
                            }

                            theta_sum[ k ] += theta * weight; /* Ignore lint */
                            phi_sum[ k ] += phi * weight;     /* Ignore lint */
                            weight_sum[ k ] += weight;        /* Ignore lint */
                        }
                    }

                    if (fs_compute_scia_points)
                    {
                        result = put_matrix_row(*points_mpp, map_vp, count); 
                        if (result == ERROR) break; 
                    }

                    count++;
                }
            }
            phi += phi_step;

            if (io_atn_flag) 
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE; 
                result = ERROR;
            }
        }
        theta += theta_step; 

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }
    }

    if ((result != ERROR) && (fs_compute_scia_points))
    {
        (*points_mpp)->num_rows = count; 
    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) result = get_target_vector(&result_vp, 3); 

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
        {
            if (fs_scia_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
            {
                theta = theta_sum[ k ] / weight_sum[ k ]; 
                phi = phi_sum[ k ] / weight_sum[ k ];
                xy_r = sin(phi);
                
                result_vp->elements[ 0 ] = c_r / (xy_r * cos(theta));
                result_vp->elements[ 1 ] = c_g / (xy_r * sin(theta));
                result_vp->elements[ 2 ] = c_b / (cos(phi));
                
                /*
                // Convert to 2D. It is a mistake to leave the result as 3D, as
                // the dimension of the vector implicitly contains the
                // information as to whether this is a 2D or a 3D algorithm. 
                */ 
                result = project_vector(&(result_vp_array[ k ]), result_vp,
                                          DIVIDE_BY_BLUE);

                if (result == ERROR) break;
            }
        }
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            double cpu_time;
            long   total_cpu_time;

            total_cpu_time = base_cpu_time + extra_cpu_time;  

            cpu_time = (double)total_cpu_time;
            cpu_time /= (double)MILLI_SECONDS_PER_SECOND;

            for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
            {
                if (fs_scia_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
                {
                    cpu_time_array[ k ] = cpu_time;
                }
            }
        }
    }

    free_vector(map_vp);  
    free_vector(inv_vp);  
    free_vector(result_vp); 
    free_matrix(inverted_illum_mp);
    free_matrix(projected_inv_illum_mp);
    free_vector(test_vp); 
    free_vector(illum_vp); 
    free_hull(projected_inv_illum_hp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vp_array,
    Matrix** points_mpp,
    double*  cpu_time_array 
)
{

    if (( ! fs_do_scca) || (chrom_hp == NULL) || (canon_illum_vp == NULL))
    {
        return NO_ERROR;
    }

    interactive_pso("    "); 
    interactive_pso("Finding surface constrained average of chrom result.\n");

    if (expanded_hp->dimension == 2)
    {
        return get_2D_scca_result(expanded_hp, chrom_hp, canon_illum_vp, 
                                  base_cpu_time, &(result_vp_array[ 0 ]), 
                                  &(cpu_time_array[ 0 ]));
    }
    else 
    {
        return get_3D_scca_result(expanded_hp, chrom_hp, canon_illum_vp, 
                                  base_cpu_time, result_vp_array, points_mpp, 
                                  cpu_time_array);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vpp,
    double*  cpu_time_ptr 
)
{
    Vector* projected_canon_illum_vp    = NULL;
    Matrix* inverted_illum_mp           = NULL;
    Matrix* projected_inv_illum_mp = NULL;
    Hull*   projected_inv_illum_hp = NULL;
    Vector* projected_result_vp         = NULL;
    Vector* result_vp                   = NULL;
    long    total_cpu_time;
    int     result = project_vector(&projected_canon_illum_vp, canon_illum_vp, 
                                    DIVIDE_BY_BLUE);

    if (result != ERROR) 
    {
        result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                              expanded_hp->vertex_mp,
                                              projected_canon_illum_vp);
    }

    if (result != ERROR) 
    {
        projected_inv_illum_hp =find_convex_hull(inverted_illum_mp,
                                                      DEFAULT_HULL_OPTIONS);
        if (projected_inv_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR) 
    {
        result = find_constrained_hull_average(chrom_hp, 
                                               projected_inv_illum_hp,
                                               scca_map_fn, 
                                               NO_HULL_WEIGHT_FN, 
                                               &projected_result_vp, 
                                               (Vector**)NULL, 
                                               (Hull**)NULL);
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
        result = back_project_vector(&result_vp, projected_result_vp, 
                                     DIVIDE_BY_SUM);

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
        result = ow_divide_vectors(result_vp, canon_illum_vp);

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) result = ow_invert_vector(result_vp);
    
    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            total_cpu_time = base_cpu_time + extra_cpu_time;  
            *cpu_time_ptr  = (double)total_cpu_time;
            *cpu_time_ptr  /= (double)MILLI_SECONDS_PER_SECOND;
        }
    }

    /*
    // Convert to 2D. It is a mistake to leave the result as 3D, as the
    // dimension of the vector implicitly contains the information as to wether
    // this is a 2D or a 3D algorithm. 
    */ 
    if (result == NO_ERROR) 
    {
        result = project_vector(result_vpp, result_vp, DIVIDE_BY_BLUE);
    }

    free_vector(projected_canon_illum_vp);
    free_matrix(inverted_illum_mp); 
    free_matrix(projected_inv_illum_mp);
    free_hull(projected_inv_illum_hp); 
    free_vector(result_vp); 
    free_vector(projected_result_vp); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_3D_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    long     int base_cpu_time,
    Vector** result_vp_array,
    Matrix** points_mpp,
    double*  cpu_time_array 
)
{
    Vector*    projected_canon_illum_vp    = NULL;
    Matrix*    inverted_illum_mp           = NULL;
    Matrix*    projected_inv_illum_mp = NULL;
    Hull*      projected_inv_illum_hp = NULL;
    Vector*    projected_result_vp         = NULL;
    Vector*    result_vp                   = NULL;
    Vector*    illum_vp                   = NULL;
    int        result                      = NO_ERROR;
    Vector*    map_vp                  = NULL;
    Vector*     inv_vp     = NULL;
    Matrix*    temp_vertex_mp              = NULL;
    long_double  r_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    long_double  g_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    long_double  weight_sum[ MAX_NUM_SCA_WEIGHT_FUNCTIONS ];
    double       weight; 
    int        count                       = 0;
    Vector*    test_vp                     = NULL;
    int k; 
    double r, g; 
    double r_min = 0.0;
    double g_min = 0.0;
    double r_max = 0.0;
    double g_max = 0.0;
    double r_step = 0.0;
    double g_step = 0.0;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    if (fs_scca_weight_functions == NULL)
    {
        NRE(fs_scca_weight_functions = N_TYPE_MALLOC(Weight_function, 
                                                MAX_NUM_SCA_WEIGHT_FUNCTIONS));

        /* Set up a default. */
        fs_scca_weight_functions[ 0 ].id = POWER_OF_GEOMETRIC_MEAN;
        fs_scca_weight_functions[ 0 ].parm = 3.0;

        for (k=1; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
        {
            fs_scca_weight_functions[ k ].id = DISABLED_WEIGHT_FUNCTION;
        }
    }
        
    ERE(init_cpu_time());

    for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
    {
        r_sum[ k ]  = 0.0; 
        g_sum[ k ]    = 0.0; 
        weight_sum[ k ] = 0.0; 
    }

    if (fs_compute_scca_points)
    {
        int resolution_2 = fs_gamut_solution_resolution*fs_gamut_solution_resolution;

        result = get_target_matrix(points_mpp, resolution_2, 3);
    }

    if (result != ERROR)
    {
        result = remove_origin_from_data(&temp_vertex_mp, 
                                         expanded_hp->vertex_mp,
                                         MIN_RGB / (2.0 * MAX_RGB));
    }

    if (result != ERROR)
    {
        result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                              temp_vertex_mp,
                                              canon_illum_vp);
    }

    if (result != ERROR)
    {
        result = ow_min_thresh_matrix_col(inverted_illum_mp, BLUE_INDEX,
                                          MIN_RGB / (2.0 * MAX_RGB)); 
    }

    if (result != ERROR) 
    {
        result = project_matrix(&projected_inv_illum_mp,
                                inverted_illum_mp, DIVIDE_BY_BLUE);

        if (result == ERROR)
        {
            db_mat(temp_vertex_mp);
            db_mat(inverted_illum_mp);
            db_mat(projected_inv_illum_mp);
            add_error("Problem occured computing 3D SCCA result."); 
        }
    }

    free_matrix(temp_vertex_mp);

    if (result != ERROR) 
    {
        projected_inv_illum_hp = find_convex_hull(projected_inv_illum_mp,
                                                  DEFAULT_HULL_OPTIONS);
        if (projected_inv_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR) result = get_target_vector(&test_vp, 2);  
    if (result != ERROR) result = get_target_vector(&illum_vp, 3);  
    if (result != ERROR) result = get_target_vector(&map_vp, 3);
    if (result != ERROR) result = get_target_vector(&inv_vp, 3);

    if (result != ERROR)
    {
        r_min = chrom_hp->min_x;
        g_min = chrom_hp->min_y;
        r_max = chrom_hp->max_x;
        g_max = chrom_hp->max_y;
        r_step = (r_max - r_min) / fs_gamut_solution_resolution;
        g_step = (g_max - g_min) / fs_gamut_solution_resolution;
    }

    for (r = r_min; r<r_max; r += r_step)
    {
        if (result == ERROR) break;

        for (g = g_min; g<g_max; g += g_step)
        {
            if (result == ERROR) break;

            test_vp->elements[ 0 ] = r; 
            test_vp->elements[ 1 ] = g; 

            if (is_point_inside_hull(chrom_hp, test_vp))
            {
                double b = 1.0 - r - g;
                double inv_r = b / r;
                double inv_g = b / g;

                test_vp->elements[ 0 ] = inv_r; 
                test_vp->elements[ 1 ] = inv_g; 

                if (is_point_inside_hull(projected_inv_illum_hp, test_vp))
                {
                    illum_vp->elements[ 0 ] = r; 
                    illum_vp->elements[ 1 ] = g; 
                    illum_vp->elements[ 2 ] = b; 

                    result = get_3D_map_extent_for_illum(illum_vp, 
                                                         canon_illum_vp,
                                                         expanded_hp,
                                                         map_vp); 
                    
                    if (result == ERROR) break; 

                    result = divide_vectors(&inv_vp, map_vp, canon_illum_vp);
                    if (result == ERROR) break; 

                    for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
                    {
                        if (fs_scca_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
                        {
                            if (fs_scca_weight_functions[ k ].id == POWER_OF_MAP_NORM)
                            {
                                double map_R = vector_magnitude(map_vp);
                                double exponant = fs_scca_weight_functions[k].parm;

                                weight = pow(map_R, exponant); 
                            }
                            else if (fs_scca_weight_functions[ k ].id == POWER_OF_INV_NORM)
                            {
                                double map_R = vector_magnitude(inv_vp);
                                double exponant = fs_scca_weight_functions[k].parm;

                                weight = pow(map_R, exponant); 
                            }
                            else if (fs_scca_weight_functions[ k ].id == POWER_OF_GEOMETRIC_MEAN)
                            {
                                double exponant = fs_scca_weight_functions[k].parm;
                                double map_prod = 
                                    map_vp->elements[ 0 ]
                                                * map_vp->elements[ 1 ]
                                                * map_vp->elements[ 2 ];

                                weight = pow(map_prod, exponant / 3.0); 
                            }
                            else 
                            {
                                weight = 1.0;
                            }

                            r_sum[ k ] += r * weight;  /* Ignore lint */
                            g_sum[ k ] += g * weight;  /* Ignore lint */
                            weight_sum[ k ] += weight; /* Ignore lint */
 
                        }
                    }

                    if (fs_compute_scca_points)
                    {
                        result = put_matrix_row(*points_mpp, map_vp, count); 
                        if (result == ERROR) break; 
                    }

                    count++;
                }
            }

            if (io_atn_flag) 
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE; 
                result = ERROR;
            }
        }

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }
    }

    if ((result != ERROR) && (fs_compute_scca_points))
    {
        (*points_mpp)->num_rows = count;
    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result == NO_ERROR)
    {
        result = get_target_vector(&projected_result_vp, 2);
    }

    if (result == NO_ERROR)
    {
        for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
        {
            if (fs_scca_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
            {
                projected_result_vp->elements[0] = r_sum[ k ] / weight_sum[ k ];
                projected_result_vp->elements[1] = g_sum[ k ] / weight_sum[ k ]; 
                result = back_project_vector(&result_vp, projected_result_vp, 
                                             DIVIDE_BY_SUM);
                if (result == ERROR) break;

                result = ow_divide_vectors(result_vp, canon_illum_vp);
                if (result == ERROR) break;

                result = ow_invert_vector(result_vp);
                if (result == ERROR) break;

                /*
                // Convert to 2D. It is a mistake to leave the result as 3D, as
                // the dimension of the vector implicitly contains the
                // information as to wether this is a 2D or a 3D algorithm. 
                */ 
                result = project_vector(&(result_vp_array[ k ]), result_vp, 
                                          DIVIDE_BY_BLUE);
                if (result == ERROR) break;
            }
        }
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            long total_cpu_time = base_cpu_time + extra_cpu_time;  
            double cpu_time = (double)total_cpu_time;

            cpu_time /= (double)MILLI_SECONDS_PER_SECOND;

            for (k=0; k<MAX_NUM_SCA_WEIGHT_FUNCTIONS; k++) 
            {
                if (fs_scca_weight_functions[ k ].id != DISABLED_WEIGHT_FUNCTION)
                {
                    cpu_time_array[ k ] = cpu_time;
                }
            }
        }
    }

    free_vector(projected_canon_illum_vp);
    free_matrix(inverted_illum_mp); 
    free_matrix(projected_inv_illum_mp);
    free_hull(projected_inv_illum_hp); 
    free_vector(result_vp); 
    free_vector(projected_result_vp); 
    free_vector(test_vp); 
    free_vector(map_vp); 
    free_vector(inv_vp); 
    free_vector(illum_vp); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_luminance_constrained_result
(
    Hull*    hp,
    Vector*  canon_illum_vp,
    double*  luminances,
    long     int base_cpu_time,
    Vector** ha_result_vp,
    double*  ha_cpu_times,
    Vector** mv_result_vp,
    double*  mv_cpu_times,
    Hull**   result_hp 
)
{
    double       relative_error;
    double       max_luminance;
    double       min_luminance;
    Matrix*    luminance_mp   = NULL;
    Hull*      luminance_hp   = NULL;
    int        try, i;
    int        result = NO_ERROR;
    int        lum_method;
    double       luminance;

    if (    ( !fs_do_intersections_with_luminance) 
         || (hp->dimension != 3)
         || (canon_illum_vp == NULL)
       )
    {
        return NO_ERROR; 
    }

    for (lum_method = 0; lum_method < MAX_NUM_LUMINANCE_METHODS; lum_method++)
    {
        if (luminances[ lum_method ] > 0.0)
        {
            interactive_pso("    ");
            interactive_pso("Finding ave of intersection with luminance %d.\n",
                            lum_method + 1);

            luminance = luminances[ lum_method ];

            ERE(init_cpu_time());

            try = 0;
            result = NO_SOLUTION; 
            relative_error = fs_relative_luminance_error;

            /*
            // FIX
            //
            // Need a more perminant solution here. 
            */
            while ((result == NO_SOLUTION) && (try < 10))
            {
                try++;

                max_luminance = (1.0 + relative_error) * luminance;
                min_luminance = (1.0 - relative_error) * luminance;

                result = get_zero_matrix(&luminance_mp, 6, 3); 

                if (result != ERROR) 
                {
                    /*
                    // The region for wich the luminance is between the min and
                    // the max is the convex hull of 6 points; 3 for where the
                    // plane R+G+B=min intersects the axes and 3 for where the
                    // plane R+G+B=max intersects the axes. Here we actually
                    // work in a space stretched by the inverse of the canon
                    // illum, as the space hp is the maps.
                    */
                    for (i=0; i<3; i++) 
                    {
                        luminance_mp->elements[ i ][ i ] = 
                                 max_luminance / canon_illum_vp->elements[ i ]; 
                        luminance_mp->elements[ 3 + i ][ i ] = 
                                 min_luminance / canon_illum_vp->elements[ i ]; 
                    }

                    result = ow_min_thresh_matrix(luminance_mp, 0.0001);

                    if (result != ERROR)
                    {
                        luminance_hp = find_convex_hull(luminance_mp, 
                                                        DEFAULT_HULL_OPTIONS);

                        if (luminance_hp == NULL) result = ERROR;
                    }
                }

                if (result != ERROR) 
                {
                    result = find_constrained_hull_average(hp, luminance_hp, 
                                                do_gamut_invert_vector,
                                                NO_HULL_WEIGHT_FN,
                                                &(ha_result_vp[ lum_method ]), 
                                                &(mv_result_vp[ lum_method ]), 
                                                &(result_hp[ lum_method ]));
                }

                free_hull(luminance_hp); 
                relative_error *= 2.0; 
            }

            if (result == NO_SOLUTION)
            {
                verbose_pso(1, "Unable to find constrained hull with ");
                verbose_pso(1, "relative error as high as %.1f%%.\n",
                           relative_error * 100.0 );
                ASSERT(ha_result_vp[ lum_method ] == NULL); 
            }
            else if ((result == NO_ERROR) && (try != 1))
            {
                verbose_pso(15,"%d tries required to find constrained hull.\n",
                            try);
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result != ERROR) 
            {
                long extra_cpu_time = get_cpu_time();

                if (extra_cpu_time == ERROR)
                {
                    result = ERROR;
                }
                else 
                {
                    long total_cpu_time = base_cpu_time + extra_cpu_time;  

                    /*
                    // CPU time is actually the time to do both ha and mv, but
                    // an upper bound is OK for now. 
                    */
                    ha_cpu_times[ lum_method ] = (double)total_cpu_time;
                    ha_cpu_times[ lum_method ] /= (double)MILLI_SECONDS_PER_SECOND;
                    mv_cpu_times[ lum_method ] = ha_cpu_times[ lum_method ]; 
                }
            }

            if (result == ERROR) break; 
        }
    }

    free_matrix(luminance_mp); 

    if (result == ERROR)
    {
        return result;
    }
    else 
    {
        /* Don't bother reporting NO_SOLUTION. */
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_lcsa_result
(
    Hull*    hp,
    Hull*    expanded_illum_hp,
    Vector*  canon_illum_vp,
    double*  luminances,
    long     int base_cpu_time,
    Vector** result_vp,
    double*  cpu_times,
    Hull**   result_hp 
)
{
    double    relative_error;
    int     try;
    int     result         = NO_ERROR;
    int     lum_method;
    double    luminance;
    Matrix* inverted_illum_mp           = NULL;
    Hull*   inverted_illum_hp; 
    Vector* inverted_illum_result_vp = NULL;
    Hull*   inverted_illum_result_hp = NULL; 
    Matrix* inverted_illum_result_mp = NULL; 

    if (( ! fs_do_sil_lcsa) || (hp->dimension != 3) || (canon_illum_vp == NULL)) 
    {
        return NO_ERROR; 
    }

    ERE(divide_matrix_by_row_vector(&inverted_illum_mp, 
                                     hp->vertex_mp,
                                     canon_illum_vp));

    inverted_illum_hp = find_convex_hull(inverted_illum_mp, 
                                         DEFAULT_HULL_OPTIONS);

    if (inverted_illum_hp == NULL)
    {
        result = ERROR;
    }

    for (lum_method = 0; lum_method < MAX_NUM_LUMINANCE_METHODS; lum_method++)
    {
        if (result == ERROR) break; 

        if (luminances[ lum_method ] > 0.0)
        {
            interactive_pso("    Finding I and L:%d constrained result.\n",
                            lum_method + 1);

            luminance = luminances[ lum_method ];

            ERE(init_cpu_time());

            try = 0;
            result = NO_SOLUTION; 
            relative_error = fs_relative_luminance_error;

            /*
            // FIX
            //
            // Need a more perminant solution here. 
            */
            while ((result == NO_SOLUTION) && (try < 10))
            {
                try++;

                fs_lcsa_max_luminance = (1.0 + relative_error) * luminance;
                fs_lcsa_min_luminance = (1.0 - relative_error) * luminance;

                result = find_constrained_hull_average(inverted_illum_hp, 
                                                       expanded_illum_hp, 
                                                       lcsa_map_fn,
                                                       NO_HULL_WEIGHT_FN,
                                                     &inverted_illum_result_vp,
                                                     (Vector**)NULL, 
                                                     &inverted_illum_result_hp);
                relative_error *= 2.0; 
            }

            if (result == NO_ERROR) 
            {
                result = multiply_matrix_by_row_vector_ew(
                                            &inverted_illum_result_mp,
                                            inverted_illum_result_hp->vertex_mp,
                                            canon_illum_vp);
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result == NO_ERROR) 
            {
                free_hull(result_hp[ lum_method ]); /* Likely not necessary. */

                result_hp[ lum_method ] = 
                    find_convex_hull(inverted_illum_result_mp, 
                                     DEFAULT_HULL_OPTIONS);

                if (result_hp[ lum_method ] == NULL) result = ERROR;
            }

            if (result == NO_SOLUTION)
            {
                verbose_pso(1, "Unable to find constrained hull with ");
                verbose_pso(1, "relative error as high as %.1f%%.\n",
                           relative_error * 100.0 );
            }
            else if ((result == NO_ERROR) && (try != 1))
            {
                verbose_pso(15,"%d tries required to find constrained hull.\n",
                            try);
            }

            if (result == NO_ERROR)
            {
                result = multiply_vectors(&(result_vp[ lum_method ]), 
                                          inverted_illum_result_vp,
                                          canon_illum_vp);
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result != ERROR) 
            {
                long extra_cpu_time = get_cpu_time();

                if (extra_cpu_time == ERROR)
                {
                    result = ERROR;
                }
                else 
                {
                    long total_cpu_time = base_cpu_time + extra_cpu_time;  

                    cpu_times[ lum_method ] = (double)total_cpu_time;
                    cpu_times[ lum_method ] /= (double)MILLI_SECONDS_PER_SECOND;
                }
            }
        }
    }

    free_matrix(inverted_illum_mp); 
    free_hull(inverted_illum_hp); 
    free_vector(inverted_illum_result_vp); 
    free_hull(inverted_illum_result_hp);
    free_matrix(inverted_illum_result_mp); 

    if (result == ERROR)
    {
        return result;
    }
    else 
    {
        /* Don't bother reporting NO_SOLUTION. */
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Average is on the surface of a sphere whose size is set by the luminance in
// illumination space. The average is restricted to illuminations which satisfy
// both the surface and illumination constraints. 
*/
static int get_scla_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    double*  sum_luminances,
    long     int base_cpu_time,
    Vector** result_vp,
    double*  cpu_times 
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    double sum_luminance; 
    double    rms_luminance;
    Matrix*   map_mp              = NULL;
    int       i;
    int       j;
    double      theta;
    double      phi;
    long_double theta_sum           = 0.0;
    long_double phi_sum             = 0.0;
    double      xy_r;
    Vector*   surface_test_vp     = NULL;
    Vector*   illum_test_vp       = NULL;
    Vector*   illum_chrom_test_vp = NULL;
#ifdef TEST
    double      max_phi             = 0.0;
    double      max_theta           = 0.0;
    double      min_phi             = M_PI / 2.0;
    double      min_theta           = M_PI / 2.0;
#endif
    double      weight;
    double      weight_sum        = 0.0;
    Vector*   sum_vp              = NULL;
    int       result              = NO_ERROR;
    int       lum_method;

    if (    ( ! fs_do_sil_scla) || (expanded_hp->dimension != 3)
         || (illum_hp == NULL) || (canon_illum_vp == NULL)
       )
    {
        return NO_ERROR;
    }

    if (result != ERROR) result = get_target_vector(&illum_test_vp, 3); 
    if (result != ERROR) result = fill_hull(illum_hp); 
    if (result != ERROR) result = fill_hull(expanded_hp); 

    for (lum_method = 0; lum_method < MAX_NUM_LUMINANCE_METHODS; lum_method++)
    {
        if (result == ERROR) break; 

        if (sum_luminances[ lum_method ] > 0.0)
        {
            int outer_count = 0;
            int count = 0;

            interactive_pso("    Finding surface and luminance %d ",
                            lum_method + 1);
            interactive_pso("constrained average of illum result.\n");

            sum_luminance = sum_luminances[ lum_method ];

            result = init_cpu_time();

            for (i=0; i<fs_gamut_solution_resolution; i++) 
            {
                if (result == ERROR) break;

                for (j=0; j<fs_gamut_solution_resolution; j++) 
                {
                    int inside_hull; 

                    if (result == ERROR) break;

                    theta = (M_PI / 2.0 ) * i / fs_gamut_solution_resolution;
                    phi = (M_PI / 2.0 ) * j / fs_gamut_solution_resolution;

                    /*
                    // Derivation of the expression for rms_luminance:
                    //
                    // x = rms_luminance * sin(phi) * cos(theta)
                    // y = rms_luminance * sin(phi) * sin(theta)
                    // z = rms_luminance * cos(phi)
                    //
                    // sum_luminances = x + y + z
                    // sum_luminances = rms_luminance * stuff
                    // rms_luminance = sum_luminances / stuff
                    //   where 
                    // stuff = sin(phi)*(cos(theta) + sin(theta)) + cos(phi)
                    */
                    rms_luminance = sum_luminance / 
                               (sin(phi)*(cos(theta) + sin(theta)) + cos(phi));

                    xy_r = rms_luminance * sin(phi);

                    illum_test_vp->elements[ 0 ] = xy_r * cos(theta); 
                    illum_test_vp->elements[ 1 ] = xy_r * sin(theta); 
                    illum_test_vp->elements[ 2 ] = rms_luminance * cos(phi); 

                    result = project_vector(&illum_chrom_test_vp, illum_test_vp,
                                            DIVIDE_BY_SUM); 

                    if (result != ERROR) 
                    {
                        inside_hull = is_point_inside_hull(illum_hp, 
                                                     illum_chrom_test_vp);

                        if (inside_hull == ERROR)
                        {
                            result = ERROR;
                        }
                        else if (inside_hull == TRUE)
                        {
#ifdef TEST
                            if (theta < min_theta) min_theta = theta;
                            if (phi   < min_theta) min_phi   = phi;
                            if (theta > max_theta) max_theta = theta;
                            if (phi   > max_theta) max_phi   = phi;
#endif 

                            outer_count++;

                            result = divide_vectors(&surface_test_vp, 
                                                    canon_illum_vp,
                                                    illum_test_vp);

                            if (result != ERROR) 
                            {
                                inside_hull = is_point_inside_hull(expanded_hp, 
                                                             surface_test_vp);

                                if (inside_hull == ERROR)
                                {
                                    result = ERROR;
                                }
                                else if (inside_hull == TRUE)
                                {
                                    weight = sin(phi); 
                                    count++;
                                    theta_sum += theta * weight;
                                    phi_sum += phi * weight;
                                    weight_sum += weight; 
                                }
                            }
                        }
                    }
                }

                if (io_atn_flag) 
                {
                    set_error("Processing interrupted.");
                    halt_all_output = FALSE; 
                    result = ERROR;
                }
            }

#ifdef TEST
            dbf(min_theta); 
            dbf(max_theta);
            dbf(min_phi);
            dbf(max_phi); 
#endif 

            if (result != ERROR)
            {
                if (count > 0) result = NO_ERROR;
                else           result = NO_SOLUTION;
            }

            if (result != NO_ERROR)
            {
                interactive_pso("        Failure.\n");
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result == NO_ERROR) 
            {
                result = get_target_vector(&(result_vp[ lum_method ]), 3); 
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result == NO_ERROR) 
            {
                theta = theta_sum / weight_sum; 
                phi = phi_sum / weight_sum;

                rms_luminance = 
                    sum_luminance / 
                              (sin(phi)*(cos(theta) + sin(theta)) + cos(phi));

                xy_r = rms_luminance * sin(phi);

                (result_vp[ lum_method ])->elements[ 0 ] = xy_r * cos(theta);
                (result_vp[ lum_method ])->elements[ 1 ] = xy_r * sin(theta);
                (result_vp[ lum_method ])->elements[ 2 ] = 
                                                      rms_luminance * cos(phi);
            }

            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result == NO_ERROR) 
            {
                result = ow_divide_vectors(result_vp[ lum_method ], 
                                           canon_illum_vp);

                if (result != ERROR) 
                {
                    result = ow_invert_vector(result_vp[ lum_method ]);
                }
            }
            
            /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
            if (result != ERROR) 
            {
                long extra_cpu_time = get_cpu_time();

                if (extra_cpu_time == ERROR)
                {
                    result = ERROR;
                }
                else 
                {
                    long total_cpu_time = base_cpu_time + extra_cpu_time;  

                    cpu_times[ lum_method ] = (double)total_cpu_time;
                    cpu_times[ lum_method ] /= (double)MILLI_SECONDS_PER_SECOND;
                }
            }
        }
    }

    free_matrix(map_mp);
    free_vector(surface_test_vp); 
    free_vector(illum_test_vp); 
    free_vector(illum_chrom_test_vp); 
    free_vector(sum_vp); 

    if (result == ERROR)
    {
        return result;
    }
    else 
    {
        /* Don't bother reporting NO_SOLUTION. */
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_sil_scia_result
(
    Hull*    expanded_hp,
    Hull*    illum_hp,
    Vector*  canon_illum_vp,
    double*  sum_luminances,
    long     int base_cpu_time,
    Vector** result_vp_array,
    double*  cpu_time_array 
)
{
    int        i, j, k;
    int        count      = 0;
    int        count_2    = 0;
    double       theta;
    double       phi;
    double       R;
    double       theta_step;
    double       phi_step;
    long_double  theta_sum[ MAX_NUM_LUMINANCE_METHODS ];
    long_double  phi_sum[ MAX_NUM_LUMINANCE_METHODS ];
    long_double  R_sum[ MAX_NUM_LUMINANCE_METHODS ];
    double       xy_r;
    Vector*    test_vp   = NULL;
    double       max_phi   = 0.0;
    double       max_theta = 0.0;
    double       min_phi   = M_PI / 2.0;
    double       min_theta = M_PI / 2.0;
    double       weight;
    long_double  weight_sum[ MAX_NUM_LUMINANCE_METHODS ];
    Hull*      projected_inv_illum_hp = NULL;
    Matrix*    projected_inv_illum_mp = NULL;
    Matrix*    inverted_illum_mp      = NULL;
    double       x, y, z, sum, c_r, c_g, c_b;
    Vector*    result_vp      = NULL;
    int        result;
    Vector*    map_vp         = NULL;
    Matrix*    temp_vertex_mp = NULL;
    double     cpu_time;
    Vector*    illum_vp       = NULL;
    int        have_one_luminance = FALSE;

    if (    ( ! fs_do_sil_scia) 
         || (illum_hp == NULL)
         || (canon_illum_vp == NULL)
         || (expanded_hp->dimension != 3)  
       )
    {
        return NO_ERROR;
    }

    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
    {
        if (sum_luminances[ k ] > 0.0)
        {
            have_one_luminance = TRUE;
            break; 
        }
    }

    if ( ! have_one_luminance) return ERROR; 

    interactive_pso("    Finding surface, illum and luminance SCIA result.\n");

    ERE(init_cpu_time());

    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
    {
        theta_sum[  k ] = 0.0;
        phi_sum[ k ]    = 0.0;
        R_sum[ k ]      = 0.0;
        weight_sum[ k ] = 0.0;
    }

    result = get_target_vector(&test_vp, 2); 

    if (result != ERROR) result = fill_hull(illum_hp); 

    if (result != ERROR)
    {
        result = remove_origin_from_data(&temp_vertex_mp, 
                                         expanded_hp->vertex_mp, 
                                         MIN_RGB / (2.0 * MAX_RGB));
    }

    if (result != ERROR)
    {
        result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                              temp_vertex_mp,
                                              canon_illum_vp);
    }

    free_matrix(temp_vertex_mp);

    if (result != ERROR)
    {
        result = get_projection_matrix(&projected_inv_illum_mp, 
                                       inverted_illum_mp,
                                       DIVIDE_BY_BLUE, 
                                       ALL_CHROMATICITIES_MUST_BE_VALID,
                                       10.0 * MAX_X_DIV_B, 
                                       DBL_NOT_SET, DBL_NOT_SET, 
                                       (Matrix **)NULL, (Matrix **)NULL);
        if (result == ERROR)
        {
            add_error("Problem occured while computing SIL-SCIA result.");
        }
    }

    if (result != ERROR) 
    {
        projected_inv_illum_hp = find_convex_hull(projected_inv_illum_mp,
                                                  DEFAULT_HULL_OPTIONS);
        if (projected_inv_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR) result = fill_hull(projected_inv_illum_hp); 
    if (result != ERROR) result = get_target_vector(&map_vp, 3);
    if (result != ERROR) result = get_target_vector(&illum_vp, 3);

    c_r = canon_illum_vp->elements[ 0 ];
    c_g = canon_illum_vp->elements[ 1 ];
    c_b = canon_illum_vp->elements[ 2 ];

    for (i = 0; i<illum_hp->vertex_mp->num_rows; i++) 
    {
        if (result == ERROR) break;

        x = illum_hp->vertex_mp->elements[ i ][ 0 ];
        y = illum_hp->vertex_mp->elements[ i ][ 1 ];
        z = 1.0 - x - y;

        R     = sqrt(x*x + y*y + z*z);
        phi   = acos(z/R);
        xy_r  = sqrt(x*x + y*y);
        theta = acos(x/xy_r);

        if (theta < min_theta) min_theta = theta;
        if (phi   < min_phi)   min_phi   = phi;
        if (theta > max_theta) max_theta = theta;
        if (phi   > max_phi)   max_phi   = phi;
    }

    theta_step = (max_theta - min_theta) / (fs_gamut_solution_resolution - 1);
    phi_step   = (max_phi - min_phi)     / (fs_gamut_solution_resolution - 1);
    theta      = min_theta;

    for (i=0; i<fs_gamut_solution_resolution; i++) 
    {
        if (result == ERROR) break;

        phi = min_phi;

        for (j=0; j<fs_gamut_solution_resolution; j++) 
        {
            int inside_hull;

            if (result == ERROR) break;

            xy_r = sin(phi);  /* xy_r = r * sin(phi), but r is 1 */
            x = xy_r * cos(theta);
            y = xy_r * sin(theta);
            z = cos(phi);     /* z = r * cos(phi), but r is 1 */ 
            sum = x + y + z;

            test_vp->elements[ 0 ] = x / sum;
            test_vp->elements[ 1 ] = y / sum;

            inside_hull = is_point_inside_hull(illum_hp, test_vp);

            if (inside_hull == ERROR)
            {
                result = ERROR;
                break;
            }
            else if (inside_hull == TRUE)
            {
                count_2++;

                test_vp->elements[ 0 ] = z / x;
                test_vp->elements[ 1 ] = z / y;

                inside_hull = is_point_inside_hull(projected_inv_illum_hp, test_vp);

                if (inside_hull == ERROR)
                {
                    result = ERROR;
                    break;
                }
                else if (inside_hull == TRUE)
                {
                    double min_lum_R;

                    illum_vp->elements[ 0 ] = x; 
                    illum_vp->elements[ 1 ] = y; 
                    illum_vp->elements[ 2 ] = z; 

                    result = get_3D_map_extent_for_illum(illum_vp, 
                                                         canon_illum_vp,
                                                         expanded_hp,
                                                         map_vp); 
                    if (result == ERROR) break; 

                    result = ow_divide_vectors(map_vp, canon_illum_vp);
                    if (result == ERROR) break; 

                    result = ow_invert_vector(map_vp);
                    if (result == ERROR) break; 

                    min_lum_R = vector_magnitude(map_vp);

                    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
                    {
                        if (sum_luminances[ k ] > 0.0)
                        {
                            double luminance = sum_luminances[ k ];
                            double lum = luminance / 
                                (sin(phi)*(cos(theta) + sin(theta)) + cos(phi));
                            double r; 
                            double max_lum_R; 

                            if (lum > min_lum_R) 
                            {
                                max_lum_R = 2.0 * lum - min_lum_R;
                            }
                            else 
                            {
                                max_lum_R = min_lum_R;
                            }

                            weight = 0.0;
                            r      = min_lum_R;

                            do 
                            {
                                /* 5 ~ 4.5 = 1/((2)*(1/3)*(1/3)) */
                                double t = 5.0 * (lum - r) / lum;
                                double u = exp(-(t*t)); 

                                u *= xy_r;   /* xy_r == sin(phi) */  
                                weight += u; 
                                R_sum[ k ] += r * u;  /* Ignore lint */
                                r *= 1.1; 
                            }
                            while (r < max_lum_R); 

                            theta_sum[ k ] += theta * weight; /* Ignore lint */
                            phi_sum[ k ] += phi * weight; /* Ignore lint */
                            weight_sum[ k ] += weight;  /* Ignore lint */
                        }
                    }

                    count++;
                }
            }
            phi += phi_step;

            if (io_atn_flag) 
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE; 
                result = ERROR;
            }
        }
        theta += theta_step; 

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }
    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) result = get_target_vector(&result_vp, 3); 

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result == NO_ERROR) 
    {
        for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
        {
            if (sum_luminances[ k ] > 0.0)
            {
                theta = theta_sum[ k ] / weight_sum[ k ];
                phi   = phi_sum[ k ] / weight_sum[ k ];
                R     = R_sum[ k ] /  weight_sum[ k ];
                xy_r  = R * sin(phi);
                
                result_vp->elements[ 0 ] = c_r / (xy_r * cos(theta));
                result_vp->elements[ 1 ] = c_g / (xy_r * sin(theta));
                result_vp->elements[ 2 ] = c_b / (R * cos(phi));
                
                result = copy_vector(&(result_vp_array[ k ]), result_vp); 

                if (result == ERROR) break;
            }
        }
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            long total_cpu_time = base_cpu_time + extra_cpu_time;  

            cpu_time = (double)total_cpu_time;
            cpu_time /= (double)MILLI_SECONDS_PER_SECOND;

            for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
            {
                if (sum_luminances[ k ] > 0.0)
                {
                    cpu_time_array[ k ] = cpu_time;
                }
            }
        }
    }

    free_vector(map_vp);  
    free_vector(result_vp); 
    free_matrix(inverted_illum_mp);
    free_matrix(projected_inv_illum_mp);
    free_vector(test_vp); 
    free_vector(illum_vp); 
    free_hull(projected_inv_illum_hp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_sil_scca_result
(
    Hull*    expanded_hp,
    Hull*    chrom_hp,
    Vector*  canon_illum_vp,
    double*  sum_luminances,
    long     int base_cpu_time,
    Vector** result_vp_array,
    double*  cpu_time_array 
)
{
    Vector*    projected_canon_illum_vp = NULL;
    Matrix*    inverted_illum_mp        = NULL;
    Matrix*    projected_inv_illum_mp   = NULL;
    Hull*      projected_inv_illum_hp   = NULL;
    Vector*    result_vp      = NULL;
    Vector*    illum_vp       = NULL;
    int        result         = NO_ERROR;
    Vector*    map_vp         = NULL;
    Matrix*    temp_vertex_mp = NULL;
    long_double  r_sum[ MAX_NUM_LUMINANCE_METHODS ];
    long_double  g_sum[ MAX_NUM_LUMINANCE_METHODS ];
    long_double  L_sum[ MAX_NUM_LUMINANCE_METHODS ];
    long_double  weight_sum[ MAX_NUM_LUMINANCE_METHODS ];
    double       weight;
    int        count    = 0;
    Vector*    test_vp  = NULL;
    double     cpu_time;
    int        k;
    int        have_one_luminance = FALSE;
    double       L, r, g;
    double       r_min = 0.0;
    double       g_min = 0.0;
    double       r_max = 0.0;
    double       g_max = 0.0;
    double       r_step = 0.0;
    double       g_step = 0.0;

    if (    ( ! fs_do_sil_scca) 
         || (chrom_hp == NULL)
         || (canon_illum_vp == NULL)
         || (expanded_hp->dimension != 3)  
       )
    {
        return NO_ERROR;
    }

    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
    {
        if (sum_luminances[ k ] > 0.0)
        {
            have_one_luminance = TRUE;
            break; 
        }
    }

    if ( ! have_one_luminance) return ERROR; 

    interactive_pso("    Finding surface, illum and luminance SCCA result.\n");

    ERE(init_cpu_time());

    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
    {
        r_sum[ k ] = 0.0; 
        g_sum[ k ] = 0.0; 
        L_sum[ k ] = 0.0; 
        weight_sum[ k ] = 0.0; 
    }

    if (result != ERROR)
    {
        result = remove_origin_from_data(&temp_vertex_mp, 
                                         expanded_hp->vertex_mp,
                                         MIN_RGB / (2.0 * MAX_RGB));
    }

    if (result != ERROR)
    {
        result = divide_matrix_by_row_vector(&inverted_illum_mp, 
                                              temp_vertex_mp,
                                              canon_illum_vp);
    }

    free_matrix(temp_vertex_mp);

    if (result != ERROR) 
    {
        result = project_matrix(&projected_inv_illum_mp,
                                inverted_illum_mp, DIVIDE_BY_BLUE);

        if (result == ERROR)
        {
            add_error("Problem occured computing SIL-SCCA result."); 
            db_mat(projected_inv_illum_mp);
        }
    }

    if (result != ERROR) 
    {
        projected_inv_illum_hp = find_convex_hull(projected_inv_illum_mp,
                                                      DEFAULT_HULL_OPTIONS);
        if (projected_inv_illum_hp == NULL) result = ERROR;
    }

    if (result != ERROR) result = get_target_vector(&test_vp, 2);  
    if (result != ERROR) result = get_target_vector(&illum_vp, 3);  
    if (result != ERROR) result = get_target_vector(&map_vp, 3);

    if (result != ERROR)
    {
        r_min = chrom_hp->min_x;
        g_min = chrom_hp->min_y;
        r_max = chrom_hp->max_x;
        g_max = chrom_hp->max_y;
        r_step = (r_max - r_min) / fs_gamut_solution_resolution;
        g_step = (g_max - g_min) / fs_gamut_solution_resolution;
    }

    for (r = r_min; r<r_max; r += r_step)
    {
        if (result == ERROR) break;

        for (g = g_min; g<g_max; g += g_step)
        {
            if (result == ERROR) break;

            test_vp->elements[ 0 ] = r; 
            test_vp->elements[ 1 ] = g; 

            if (is_point_inside_hull(chrom_hp, test_vp))
            {
                double b = 1.0 - r - g;
                double inv_r = b / r;
                double inv_g = b / g;

                test_vp->elements[ 0 ] = inv_r; 
                test_vp->elements[ 1 ] = inv_g; 

                if (is_point_inside_hull(projected_inv_illum_hp, test_vp))
                {
                    double min_sum_lum;

                    illum_vp->elements[ 0 ] = r; 
                    illum_vp->elements[ 1 ] = g; 
                    illum_vp->elements[ 2 ] = b; 

                    result = get_3D_map_extent_for_illum(illum_vp, 
                                                         canon_illum_vp,
                                                         expanded_hp,
                                                         map_vp); 
                    if (result == ERROR) break; 

                    result = ow_divide_vectors(map_vp, canon_illum_vp);
                    if (result == ERROR) break; 

                    result = ow_invert_vector(map_vp);
                    if (result == ERROR) break; 

                    min_sum_lum = sum_vector_elements(map_vp);

                    for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
                    {
                        if (sum_luminances[ k ] > 0.0)
                        {
                            double sum_lum = sum_luminances[ k ];
                            double max_sum_lum; 

                            if (sum_lum > min_sum_lum) 
                            {
                                max_sum_lum = 2.0 * sum_lum - min_sum_lum;
                            }
                            else 
                            {
                                max_sum_lum = min_sum_lum;
                            }

                            L = min_sum_lum; 
                            weight = 0.0;

                            do 
                            {
                                /* 5 ~ 4.5 = 1/((2)*(1/3)*(1/3)) */
                                double t = 5.0 * (sum_lum - L) / sum_lum;
                                double u = exp(-(t*t)); 

                                weight += u; 
                                L_sum[ k ] += L * u;  /* Ignore lint */
                                L *= 1.1; 
                            }
                            while (L < max_sum_lum); 

                            r_sum[ k ] += r * weight; /* Ignore lint */
                            g_sum[ k ] += g * weight; /* Ignore lint */
                            weight_sum[ k ] += weight;  /* Ignore lint */
                        }
                    }

                    count++;
                }
            }

            if (io_atn_flag) 
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE; 
                result = ERROR;
            }
        }

        if (io_atn_flag) 
        {
            set_error("Processing interrupted.");
            halt_all_output = FALSE; 
            result = ERROR;
        }

    }

    if (result != ERROR)
    {
        if (count > 0) result = NO_ERROR;
        else           result = NO_SOLUTION;
    }

    if (result == NO_ERROR) result = get_target_vector(&result_vp, 3); 

    if (result == NO_ERROR)
    {
        for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
        {
            if (sum_luminances[ k ] > 0.0)
            {
                double b;

                L = L_sum[ k ] / weight_sum[ k ];  
                r = r_sum[ k ] / weight_sum[ k ]; 
                g = g_sum[ k ] / weight_sum[ k ]; 
                b = 1.0 - r -g; 

                result_vp->elements[0] = L * r; 
                result_vp->elements[1] = L * g;
                result_vp->elements[2] = L * b;

                result = ow_divide_vectors(result_vp, canon_illum_vp);
                if (result == ERROR) break;

                result = ow_invert_vector(result_vp);
                if (result == ERROR) break;

                result = copy_vector(&(result_vp_array[ k ]), result_vp); 
                if (result == ERROR) break;
            }
        }
    }

    if (result != NO_ERROR)
    {
        interactive_pso("        Failure.\n");
    }

    /* result can be ERROR, NO_ERROR, or NO_SOLUTION */ 
    if (result != ERROR) 
    {
        long extra_cpu_time = get_cpu_time();

        if (extra_cpu_time == ERROR)
        {
            result = ERROR;
        }
        else 
        {
            long total_cpu_time = base_cpu_time + extra_cpu_time;  

            cpu_time = (double)total_cpu_time;
            cpu_time /= (double)MILLI_SECONDS_PER_SECOND;

            for (k=0; k<MAX_NUM_LUMINANCE_METHODS; k++) 
            {
                if (sum_luminances[ k ] > 0.0)
                {
                    cpu_time_array[ k ] = cpu_time;
                }
            }
        }
    }

    free_vector(projected_canon_illum_vp);
    free_matrix(inverted_illum_mp); 
    free_matrix(projected_inv_illum_mp);
    free_hull(projected_inv_illum_hp); 
    free_vector(result_vp); 
    free_vector(test_vp); 
    free_vector(map_vp); 
    free_vector(illum_vp); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_3D_map_extent_for_illum
(
    Vector* illum_vp,
    Vector* canon_illum_vp,
    Hull*   map_hp,
    Vector* map_vp 
)
{
    IMPORT int kjb_debug_level; 
    double   delta_map_x;
    double   delta_map_y;
    double   delta_map_z;
    double   factor;
    int    k;
    double   map_x = canon_illum_vp->elements[ 0 ] / illum_vp->elements[ 0 ];
    double   map_y = canon_illum_vp->elements[ 1 ] / illum_vp->elements[ 1 ];
    double   map_z = canon_illum_vp->elements[ 2 ] / illum_vp->elements[ 2 ];

    /* Set the map coords outside the map hull. */
    factor = map_hp->max_x / map_x;
    map_x = map_hp->max_x;
    map_y *= factor;
    map_z *= factor;

    /* Do a binary search to get the intersection. */
    map_x /= 2.0; 
    map_y /= 2.0;
    map_z /= 2.0; 

    delta_map_x = map_x;
    delta_map_y = map_y;
    delta_map_z = map_z;

    for (k=0; k<15; k++)
    {
        delta_map_x /= 2.0; 
        delta_map_y /= 2.0;
        delta_map_z /= 2.0; 

        map_vp->elements[ 0 ] = map_x;
        map_vp->elements[ 1 ] = map_y;
        map_vp->elements[ 2 ] = map_z;

        if (is_point_inside_hull(map_hp, map_vp))
        {
            map_x += delta_map_x; 
            map_y += delta_map_y; 
            map_z += delta_map_z; 
        }
        else 
        {
            map_x -= delta_map_x; 
            map_y -= delta_map_y; 
            map_z -= delta_map_z; 
        }
    }

    if (kjb_debug_level > 3)
    {
        double distance; 

        EPETE(get_distance_to_hull(map_hp, map_vp, &distance));

        if (distance > 1e-04)
        {
            dbe(distance); 
        }
    }

    if (map_x < map_hp->max_x / 102400.0)
    {
        verbose_pso(1, "Intersection of illum direction ");
        verbose_pso(1, "with hull is unexpectedly close to zero.\n");
        verbose_pso(1, "X component of intersection is %e\n", map_x); 
    }

    map_vp->elements[ 0 ] = map_x;
    map_vp->elements[ 1 ] = map_y;
    map_vp->elements[ 2 ] = map_z;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static double chromaticity_weight_fn(const Vector* vp)
{
    double temp = MIN_OF(vp->elements[ 0 ], vp->elements[ 1 ]);

    if (temp < 10.0 * MIN_RGB)
    {
        temp = 10.0 * MIN_RGB;
    }

    return 1.0 / temp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static double rgb_weight_fn(const Vector* vp)
{
    double temp = vp->elements[ 0 ] + vp->elements[ 1 ] + vp->elements[ 2 ];

    if (temp < 10.0 * MIN_RGB)
    {
        temp = 10.0 * MIN_RGB;
    }

    return 1.0 / temp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int scca_map_fn(Vector** target_vpp, const Vector* vp)
{
    double x, y, z; 

    ASSERT(vp->length == 2);

    ERE(get_target_vector(target_vpp, 2));

    x = vp->elements[ 0 ];
    y = vp->elements[ 1 ];
    z = 1.0 - x - y;

    if (    (x < fs_constraint_min_rgb)
         || (y < fs_constraint_min_rgb)
       )
    {
        return NO_SOLUTION;
    }
    else 
    {
        (*target_vpp)->elements[ 0 ] = z / x;
        (*target_vpp)->elements[ 1 ] = z / y;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int icsa_map_fn(Vector** target_vpp, const Vector* source_vp)
{
    double r   = (source_vp->elements)[ 0 ];
    double g   = (source_vp->elements)[ 1 ];
    double b   = (source_vp->elements)[ 2 ];
    double sum;

    ASSERT(source_vp->length == 3);

    ERE(get_target_vector(target_vpp, 2));

    if (    (r < fs_constraint_min_rgb) 
         || (g < fs_constraint_min_rgb)
         || (b < fs_constraint_min_rgb)
       )
    {
        return NO_SOLUTION;
    }

    r = 1.0 / r;
    g = 1.0 / g;
    b = 1.0 / b;

    sum = r + g + b; 

    if (sum < fs_constraint_min_rgb) 
    {
        return NO_SOLUTION;
    }
    else 
    {
        (*target_vpp)->elements[ 0 ] = r / sum; 
        (*target_vpp)->elements[ 1 ] = g / sum; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_gamut_invert_vector
(
    Vector**      target_vpp,
    const Vector* source_vp 
)
{
    Vector* target_vp;
    int     i;
    
    ERE(get_target_vector(target_vpp, source_vp->length));
    target_vp = *target_vpp; 

    for (i=0; i<source_vp->length; i++) 
    {
        if ((source_vp->elements)[i] < fs_constraint_min_rgb)
        {
            return NO_SOLUTION;
        }
        else 
        {
            (target_vp->elements)[i] = 1.0 / (source_vp->elements)[i];
        }
    }

   return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int lcsa_map_fn(Vector** target_vpp, const Vector* source_vp)
{
    Vector* target_vp;
    double  r         = (source_vp->elements)[ 0 ];
    double  g         = (source_vp->elements)[ 1 ];
    double  b         = (source_vp->elements)[ 2 ];
    double  sum;

    ERE(get_target_vector(target_vpp, 2));
    target_vp = *target_vpp; 

    if (    (r < fs_constraint_min_rgb) 
         || (g < fs_constraint_min_rgb)
         || (b < fs_constraint_min_rgb)
       )
    {
        return NO_SOLUTION;
    }

    r = 1.0 / r;
    g = 1.0 / g;
    b = 1.0 / b;

    sum = r + g + b; 

    if (    (sum < fs_constraint_min_rgb) 
         || (sum < fs_lcsa_min_luminance) 
         || (sum > fs_lcsa_max_luminance)
       )
    {
        return NO_SOLUTION;
    }

    (target_vp->elements)[ 0 ] = r / sum;
    (target_vp->elements)[ 1 ] = g / sum;

   return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    kjb_free(fs_scia_weight_functions); 
    kjb_free(fs_scca_weight_functions); 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

