
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "file_data.h"

/* -------------------------------------------------------------------------- */

#define MEASURED_DATA_FILE            "measured_data.rgb"
#define GENERATED_DATA_FILE           "generated_data.rgb"
#define MEASURED_CANON_DATA_FILE      "measured_canon_data.rgb"
#define GENERATED_CANON_DATA_FILE     "generated_canon_data.rgb"

#define MAX_NUM_VAR_ILLUM_MAPS                   1000

/* -------------------------------------------------------------------------- */

static Matrix* fs_cached_data_mp                               = NULL;
static int     fs_data_measured_flag                           = NOT_SET;
static char    fs_data_config_file[ MAX_FILE_NAME_SIZE ]       = { '\0' };

static Matrix* fs_cached_canon_data_mp                         = NULL;
static int     fs_canon_data_measured_flag                     = NOT_SET;
static char    fs_canon_data_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };

static int     fs_data_version                                 = 0;
static int     fs_canon_data_version                           = 0;

/* -------------------------------------------------------------------------- */

static int initialize_file_data(void);

static int read_file_data(const char* file_name, int measured_flag);

static int set_data_matrix
(
    Matrix* data_mp,
    char*   config_file,
    int     measured_flag 
);

static int read_canon_data_from_file
(
    const char* file_name,
    int         measured_flag 
);

static int set_canon_data_matrix
(
    Matrix* canon_data_mp,
    char*   config_file,
    int     measured_flag 
);

static int initialize_file_canon_data(void);

#ifdef DISABLED_FOR_NOW
    static int read_var_data_from_file
    (
        char* file_name,
        int   measured_flag 
    );
#endif 

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_data_files(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "mdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_file_data(value, TRUE /* measured_flag */));
            result = NO_ERROR; 
        }
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "gdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_file_data(value, FALSE /* measured_flag */)); 
            result = NO_ERROR; 
        }
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "mcdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_canon_data_from_file(value, TRUE /* measured_flag */)); 
            result = NO_ERROR; 
        }
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "gcdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_canon_data_from_file(value, FALSE /* measured_flag */)); 
            result = NO_ERROR; 
        }
    }
    
#ifdef DISABLED_FOR_NOW
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "gvdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_var_data_from_file(value, FALSE /* measured_flag */)); 
            result = NO_ERROR; 
        }
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "mvdf"))
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?'))
        {
            ERE(read_var_data_from_file(value, TRUE /* measured_flag */)); 
            result = NO_ERROR; 
        }
    }
#endif 

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "df"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_data_mp == NULL) 
            {
                ERE(pso("No data has been set.\n"));
            }
            else 
            {
                ERE(pso("Data source is %s.\n", fs_data_config_file)); 

                if (IS_SET(fs_data_measured_flag))
                {
                    ERE(pso("    (It was read %s linearization).\n",
                            fs_data_measured_flag ? "with" : "without")); 
                }
            }
        }
        else 
        {
            set_error("Use either gdf or mdf to change data file."); 
            return ERROR; 
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "cdf"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_canon_data_mp == NULL) 
            {
                ERE(pso("No canonical data has been set.\n"));
            }
            else 
            {
                ERE(pso("Canonical data source is %s.\n", 
                        fs_canon_data_config_file)); 

                if (IS_SET(fs_canon_data_measured_flag))
                {
                    ERE(pso("    (It was read %s linearization).\n",
                            fs_canon_data_measured_flag ? "with" : "without")); 
                }
            }
        }
        else 
        {
            set_error("Use either gcdf or mcdf to change canonical data file.");
            return ERROR; 
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_file_data(const char* file_name, int measured_flag)
{
    Matrix* data_mp = NULL;
    char config_file[ MAX_FILE_NAME_SIZE ];
    int result; 

    ERE(read_matrix_from_gamut_config_file(&data_mp, (const char*)NULL, 
                                           file_name, 
                                           measured_flag ? "measured data"
                                                         : "generated data",
                                           config_file, sizeof(config_file)));

    if ((data_mp != NULL) && (measured_flag))
    {
        ERE(remove_camera_offset_from_data(data_mp));
    }

    result = set_data_matrix(data_mp, config_file, measured_flag);

    free_matrix(data_mp);

    if (result != ERROR) 
    {
        result = force_file_data();
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_data_matrix
(
    Matrix* new_data_mp,
    char*   config_file,
    int     measured_flag 
)
{

    ERE(copy_matrix(&fs_cached_data_mp, new_data_mp));
    BUFF_CPY(fs_data_config_file, config_file);
    fs_data_measured_flag = measured_flag; 

    fs_data_version++; 

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    return fs_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void get_file_data_source(char* source_buff, size_t max_len)
{

    kjb_strncpy(source_buff, fs_data_config_file, max_len);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_file_data_points(void)
{

    ERE(initialize_file_data());

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0)); 

    return fs_cached_data_mp->num_rows; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_file_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    return get_selected_registered_file_data(target_mpp, num_data_points, 
                                       data_point_nums);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_registered_file_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(initialize_file_data());

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0)); 

    ERE(select_data(target_mpp, fs_cached_data_mp, num_data_points, 
                    data_point_nums, (int*)NULL));
    
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_registered_file_data(Matrix** target_mpp)
{

    ERE(initialize_file_data());

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0)); 

    ERE(copy_matrix(target_mpp, fs_cached_data_mp));

    return fs_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_file_data(void)
{
    static int first_time = TRUE; 

    if ((fs_cached_data_mp == NULL) && first_time)
    {
        (void)read_file_data(MEASURED_DATA_FILE, TRUE  /* measured_flag */);
    }

    if ((fs_cached_data_mp == NULL) && first_time)
    {
        (void)read_file_data(GENERATED_DATA_FILE, FALSE  /* measured_flag */ );
    }

    if (fs_cached_data_mp == NULL)
    {
        set_error("Unable to get file data."); 

        if (first_time)
        {
            add_error("None is set and default data is not available.");
        }
        return ERROR; 
    }
    else if (fs_cached_data_mp->num_rows == 0) /* Not sure if possible. */
    {
        set_error("No good file data points.");
        return ERROR; 
    }

    first_time = FALSE;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int file_data_is_up_to_date(int proposed_current_version)
{

    return (fs_data_version == proposed_current_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_data_from_file
(
    const char* file_name,
    int         measured_flag 
)
{
    Matrix* canon_data_mp = NULL;
    char config_file[ MAX_FILE_NAME_SIZE ];
    int result; 

    ERE(read_matrix_from_gamut_config_file(&canon_data_mp, (const char*)NULL, 
                                         file_name, 
                                         measured_flag ? "measured canon data"
                                                       : "generated canon data",
                                         config_file,
                                         sizeof(config_file)));

    if ((canon_data_mp != NULL) && (measured_flag))
    {
        ERE(remove_camera_offset_from_data(canon_data_mp));
    }

    result = set_canon_data_matrix(canon_data_mp, config_file, measured_flag);

    free_matrix(canon_data_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_canon_data_matrix
(
    Matrix* new_canon_data_mp,
    char*   config_file,
    int     measured_flag 
)
{

    ERE(copy_matrix(&fs_cached_canon_data_mp, new_canon_data_mp));
    BUFF_CPY(fs_canon_data_config_file, config_file);
    fs_canon_data_measured_flag = measured_flag; 

    fs_canon_data_version++; 

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    return fs_canon_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_file_canon_data(Matrix** target_mpp)
{

    if (fs_cached_canon_data_mp == NULL) 
    {
       ERE(initialize_file_canon_data()); 
    }

    ASSERT(    (fs_cached_canon_data_mp != NULL) 
            && (fs_cached_canon_data_mp->num_rows > 0)); 

    ERE(copy_matrix(target_mpp, fs_cached_canon_data_mp));

    return fs_canon_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_file_canon_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    if (fs_cached_canon_data_mp == NULL) 
    {
       ERE(initialize_file_canon_data()); 
    }

    ASSERT(    (fs_cached_canon_data_mp != NULL) 
            && (fs_cached_canon_data_mp->num_rows > 0)); 

    ERE(select_data(target_mpp, fs_cached_canon_data_mp, num_data_points, 
                    data_point_nums, (int*)NULL));

    return fs_canon_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_file_var_data_points(void)
{
    set_error("FIle var data code not yet ready.");
    return ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_registered_file_var_data(Matrix** dummy_target_mpp)
{
#ifdef NOT_YET 
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version; 
    Matrix*        dull_data_mp = NULL; 
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR; 

    ERE(initialize_var_image_data()); 

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0)); 

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else 
    {
        result = select_data(data_mpp, var_cached_input_data_mp, 
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp); 

    return cached_var_input_data_version; 
#else
    set_error("Var data code not yet ready.");
    return ERROR; 
#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_file_var_data
(
    Matrix** dummy_target_mpp,
    int      dummy_num_data_points,
    int*     dummy_data_point_nums,
    Matrix** dummy_var_data_mpp 
)
{

#ifdef NOT_YET 
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version; 
    Matrix*        dull_data_mp = NULL; 
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR; 

    ERE(initialize_var_image_data()); 

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0)); 

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else 
    {
        result = select_data(data_mpp, var_cached_input_data_mp, 
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp); 

    return cached_var_input_data_version; 
#else
    set_error("Var data code not yet ready.");
    return ERROR; 
#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_registered_file_var_data
(
    Matrix** dummy_target_mpp,
    int      dummy_num_data_points,
    int*     dummy_data_point_nums 
)
{
#ifdef NOT_YET 
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version; 
    Matrix*        dull_data_mp = NULL; 
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR; 

    ERE(initialize_var_image_data()); 

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0)); 

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else 
    {
        result = select_data(data_mpp, var_cached_input_data_mp, 
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp); 

    return cached_var_input_data_version; 
#else
    set_error("Var data code not yet ready.");
    return ERROR; 
#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_file_canon_data(void)
{
    static int first_time = TRUE; 

    if ((fs_cached_canon_data_mp == NULL) && first_time)
    {
        (void)read_canon_data_from_file(MEASURED_CANON_DATA_FILE,
                                        TRUE  /* measured_flag */);
    }

    if ((fs_cached_canon_data_mp == NULL) && first_time)
    {
        (void)read_canon_data_from_file(GENERATED_CANON_DATA_FILE,
                                        FALSE  /* measured_flag */ );
    }

    if (fs_cached_canon_data_mp == NULL)
    {
        set_error("Unable to get file canonical data."); 

        if (first_time)
        {
            add_error("None is set and default data is not available.");
        }
        return ERROR; 
    }
    else if (fs_cached_canon_data_mp->num_rows == 0) /* Not sure if can happen. */
    {
        set_error("No good file canonical data points.");
        return ERROR; 
    }

    first_time = FALSE; 

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int file_canon_data_is_up_to_date(int proposed_current_version)
{

    /*
    // If there is no canon data, then we are up to date. Making this explicit
    // through the test (sharp_version == 0) means that a previous return from a
    // sharpening routine will work properly, even if that return was ERROR, due
    // to the lack of canon data.
    */

    return (fs_canon_data_version == proposed_current_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef DISABLED_FOR_NOW 

static int read_var_data_from_file(file_name, measured_flag)
    char* file_name;
    int   measured_flag;
{
    char    dummy_config_file_name[ MAX_FILE_NAME_SIZE ];
    Matrix* data_mp                                      = NULL;
    Matrix* var_data_mp                                  = NULL;
    Matrix* sharp_var_data_mp                            = NULL;
    Matrix* sharp_data_mp                                = NULL;
    int     cur_row;
    Vector* diag_trans_vp                                = NULL;
    double  error;

    kjb_clear_error();

    initialize_file_data();

    if (fs_cached_data_mp == NULL)
    {
        add_error("Cannot set var illum data before setting data.");
        return ERROR; 
    }

    if (cached_sharp_synth_var_mp == NULL)
    {
#ifdef TRACK_MEMORY_ALLOCATION
        prepare_memory_cleanup();
#endif 

        ERE(get_target_matrix(&cached_synth_var_mp, MAX_NUM_VAR_ILLUM_MAPS, 3));
        ERE(get_target_matrix(&cached_sharp_synth_var_mp, 
                              MAX_NUM_VAR_ILLUM_MAPS, 3));
        cached_synth_var_mp->num_rows = 0;
        cached_sharp_synth_var_mp->num_rows = 0;
    }

    if (cached_synth_var_mp->num_rows >= MAX_NUM_VAR_ILLUM_MAPS)
    {
        set_error("Max number of var illum maps exceeded.");
        return ERROR; 
    }

    ERE(read_matrix_from_gamut_config_file(&var_data_mp, (const char*)NULL, 
                                           file_name, "var data", 
                                           dummy_config_file_name,
                                           sizeof(dummy_config_file_name)));

    /*
    // FIX
    */
    if (var_data_mp == NULL) 
    {
        set_error("Can't set var data to off yet."); 
        return ERROR;
    }

    if (var_data_mp->num_rows != fs_cached_data_mp->num_rows) 
    {
        set_error("Number of var data points must be same as ");
        cat_error("that in non-var data.");
        return ERROR; 
    }

    if (measured_flag) 
    {
        ERE(remove_camera_offset_from_data(var_data_mp));
    }

    cur_row = cached_synth_var_mp->num_rows;

    /*
    // We need to compute the file transforms due to varying illum. 
    // We are looking for an estimate of the transform going from the 
    // new illuminant to the base illumination. Use least squares fit.
    */
    ERE(get_best_diagonal_post_map(&diag_trans_vp, data_mp, var_data_mp));

    ERE(get_diagonal_post_map_error(diag_trans_vp, data_mp, 
                                    var_data_mp, &error));

    (cached_synth_var_mp->num_rows)++;

    ERE(put_matrix_row(cached_synth_var_mp, diag_trans_vp, cur_row));

    if (    (sharpen_matrix_data(&sharp_data_mp, data_mp) != ERROR)
         && (sharpen_matrix_data(&sharp_var_data_mp, var_data_mp) != ERROR))
    {
        cur_row = cached_sharp_synth_var_mp->num_rows;

        ERE(get_best_diagonal_post_map(&diag_trans_vp, sharp_data_mp,
                                       sharp_var_data_mp));

        ERE(get_diagonal_post_map_error(diag_trans_vp, sharp_data_mp, 
                                        sharp_var_data_mp, &error));

        (cached_sharp_synth_var_mp->num_rows)++;

        ERE(put_matrix_row(cached_sharp_synth_var_mp, diag_trans_vp, cur_row));
    }

    free_vector(diag_trans_vp);
    free_matrix(data_mp);
    free_matrix(var_data_mp);

    free_matrix(sharp_var_data_mp);
    free_matrix(sharp_data_mp);

    return NO_ERROR; 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_matrix(fs_cached_data_mp);
    free_matrix(fs_cached_canon_data_mp);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

