
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "do_gamut_cc.h"
#include "gamut_alg_input.h"
#include "illum_data.h"
#include "canon_gamut.h"
#include "lib.h"
#include "plot.h"

#include "gamut_2D.h"

/* -------------------------------------------------------------------------- */

#define MIN_USABLE_2D_CANON_GAMUT_VALUE      (0.1)
#define MIN_ALLOWABLE_2D_CANON_GAMUT_VALUE   (0.0)
#define MIN_USABLE_2D_ILLUM_VALUE            (0.1)
#define MIN_ALLOWABLE_2D_ILLUM_VALUE         (0.0)

#define MAX_NUM_2D_PLOTS                    5

#define INTERSECTION_PLOT                   0
#define HULL_PLOT                           1
#define HULL_WITH_NO_ERROR_PLOT             2
#define ILLUM_INVERSE_HULL_PLOT             3
#define ILLUM_HULL_POINTS_PLOT              4

#define IS_ALMOST_EQUAL(x, y, z)     (((x-y)<z) && ((y-x)<z))

/* -------------------------------------------------------------------------- */

static int fs_gamut_2D_inversion_resolution = 50;
static Negative_data_action fs_negative_input_data_action =SKIP_NEGATIVE_BLUE_DATA;

/* -------------------------------------------------------------------------- */

static int do_gamut_2D_variant
(
    int                       canon_gamut_num,
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr 
);

static int do_2D_maps
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       plot_id_list[ MAX_NUM_2D_PLOTS ],
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    double                    absolute_pixel_error,
    double                    relative_pixel_error,
    Gamut_result*             result_ptr,
    Hull_result*              hull_result_ptr,
    Gamut_error*              cpu_time_ptr,
    double*                   illum_hull_x_max_ptr,
    double*                   illum_hull_y_max_ptr 
);

static int do_2D_surface_maps
(
    int             plot_id_list[ MAX_NUM_2D_PLOTS ],
    Matrix*         canon_gamut_mp,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    Queue_element** hull_list_head_ptr 
);

static int do_2D_hull_surface_maps
(
    int             plot_id_list[ MAX_NUM_2D_PLOTS ],
    Hull*           canon_hp,
    Matrix*         projected_data_mp,
    Matrix*         error_box_mp,
    Queue_element** hull_list_head_ptr 
);

static int do_2D_point_surface_maps
(
    int             plot_id_list[ MAX_NUM_2D_PLOTS ],
    int*            data_point_nums,
    Hull*           canon_hp,
    Matrix*         projected_data_mp,
    Matrix*         error_box_mp,
    Queue_element** hull_list_head_ptr 
);

static int do_2D_illum_maps
(
    int*         plot_id_list,
    Constraints* constraints_ptr,
    double*      illum_hull_x_max_ptr,
    double*      illum_hull_y_max_ptr 
);

static int do_illum_2D_inv_trans_gamut
(
    int             plot_id_list[ MAX_NUM_2D_PLOTS ],
    Vector*         illum_2D_trans_r_vp,
    Vector*         illum_2D_trans_g_vp,
    Vector*         hull_x1_vp,
    Vector*         hull_x2_vp,
    Vector*         hull_y1_vp,
    Vector*         hull_y2_vp,
    Queue_element** hull_list_head_ptr 
);

static int get_2D_hull_segments
(
    Vector*  x_vp,
    Vector*  y_vp,
    Vector** hull_x1_vpp,
    Vector** hull_x2_vpp,
    Vector** hull_y1_vpp,
    Vector** hull_y2_vpp 
);

static int initialize_2D_plots
(
    int canon_gamut_num,
    int enable_plotting,
    int plot_id_list[ MAX_NUM_2D_PLOTS ] 
);

static int plot_2D_hull_list
(
    int         plot_id,
    int         num_hulls,
    Hull**      hp_list,
    const char* title 
);

/* -------------------------------------------------------------------------- */

int set_2D_gamut_options(const char* option, const char* value)
{ 
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "gamut-2d-inversion-resolution"))
         || (match_pattern(lc_option, "2d-gamut-inversion-resolution"))
         || (match_pattern(lc_option, "2d-inversion-resolution"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("2d-gamut-inversion-resolution = %d\n",
                    fs_gamut_2D_inversion_resolution));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Gamut 2D inversion resolution is %d.\n",
                    fs_gamut_2D_inversion_resolution));
        }
        else
        {
            int temp_int;

            ERE(ss1pi(value, &temp_int));

            if (temp_int <= 0) 
            {
                set_error("Gamut inversion resolution must be positive.");
                return ERROR; 
            }
            fs_gamut_2D_inversion_resolution = temp_int;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "gamut-2d-negative-input-data-action"))
         || (match_pattern(lc_option, "2d-gamut-negative-input-data-action"))
         || (match_pattern(lc_option, "2d-negative-input-data-action"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Input data points with negative "));

            if (fs_negative_input_data_action == USE_NEGATIVE_DATA)
            {
                ERE(pso("R/G/B are used "));
            }
            else if (fs_negative_input_data_action == SKIP_NEGATIVE_DATA)
            {
                ERE(pso("R/G/B are skipped ")); 
            }
            else if (fs_negative_input_data_action == SKIP_NEGATIVE_BLUE_DATA)
            {
                ERE(pso("B are skipped ")); 
            }
            else if (fs_negative_input_data_action == ZERO_NEGATIVE_DATA)
            {
                ERE(pso("R/G/B are set to zero ")); 
            }
            else if (fs_negative_input_data_action == NEGATIVE_DATA_IS_ERROR)
            {
                ERE(pso("R/G/B cause error "));
            }
            ERE(pso("with 2D gamut mapping.\n")); 
        }
        else
        {
            char lc_value[ 100 ];

            EXTENDED_LC_BUFF_CPY(lc_value, value);

            if (STRCMP_EQ(lc_value, "use"))
            {
                fs_negative_input_data_action = USE_NEGATIVE_DATA;

                pso("Warning: "); 
                pso("Data with negative blue components will cause error.\n"); 
            }
            else if (STRCMP_EQ(lc_value, "skip"))
            {
                fs_negative_input_data_action = SKIP_NEGATIVE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "skip-blue"))
            {
                fs_negative_input_data_action = SKIP_NEGATIVE_BLUE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "zero"))
            {
                fs_negative_input_data_action = ZERO_NEGATIVE_DATA;
            }
            else if (STRCMP_EQ(lc_value, "error"))
            {
                fs_negative_input_data_action = NEGATIVE_DATA_IS_ERROR;
            }
            else 
            {
                set_error("%q is and invalid negative data action.", value);
                return ERROR;
            }
        }
        result = NO_ERROR; 
    }
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_gamut_2D
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr 
)
{
    Matrix* canon_gamut_mp       = NULL;
    int     i;
    int     result               = NO_ERROR;
    Vector* min_vp               = NULL;
    double  min_canon_gamut_blue;
    char sharp_method_string[ 100 ];

    get_current_sharp_method_string(sharp_method_string, 
                                    sizeof(sharp_method_string));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        IMPORT volatile int io_atn_flag;

        if (result == ERROR) break; 

        result = get_sharp_canon_gamut(&canon_gamut_mp, i);

        if ((result != ERROR) && (canon_gamut_mp != NULL))
        {
            result = get_min_matrix_col_elements(&min_vp, canon_gamut_mp);
        }

        if ((result != ERROR) && (canon_gamut_mp != NULL))
        {
            min_canon_gamut_blue = (min_vp->elements)[ BLUE_INDEX ];

            if (min_canon_gamut_blue < MIN_ALLOWABLE_2D_CANON_GAMUT_VALUE)
            {
                set_error("Canon gamut RGB contains blue values less than %.1f.",
                          MIN_ALLOWABLE_2D_CANON_GAMUT_VALUE);

                result = ERROR;
            }
            else if (min_canon_gamut_blue < MIN_USABLE_2D_CANON_GAMUT_VALUE)
            {
                warn_pso("Canon gamut RGB contains blue values less than %.1f.\n",
                         MIN_USABLE_2D_CANON_GAMUT_VALUE); 
                warn_pso("Min blue value is %.2e (min allowable is %.1f).\n",
                         min_canon_gamut_blue, 
                         MIN_ALLOWABLE_2D_CANON_GAMUT_VALUE);
                warn_pso("Resetting those values to %.1f.\n", 
                         MIN_USABLE_2D_CANON_GAMUT_VALUE); 
                warn_pso("Problem occured doing 2D gamut mapping with %q.\n\n", 
                         sharp_method_string);

                result = ow_min_thresh_matrix_col(canon_gamut_mp, BLUE_INDEX, 
                                              MIN_USABLE_2D_CANON_GAMUT_VALUE);
            }
        }

        if ((result != ERROR) && (canon_gamut_mp != NULL))
        {
            result = do_gamut_2D_variant(i, cc_methods_ptr,
                                         prompt_to_increase_error_flag,
                                         canon_gamut_mp,
                                         data_point_nums, data_mp, 
                                         result_ptr,
                                         &(cpu_time_ptr[ i ]), 
                                         &(input_data_error_ptr[ i ]));

        }

        if (result == ERROR) 
        {
            add_error("Problem with 2D gamut mapping variant %d with %q.", 
                     i + 1, sharp_method_string);

            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                /*EMPTY*/
                ; /* Do nothing. */
            }
            else if (cc_methods_ptr->print_algorithm_error) 
            {
                kjb_print_error();
                result = NO_ERROR; 
            }
        }
    }

    free_matrix(canon_gamut_mp); 
    free_vector(min_vp); 
    
    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_gamut_2D_variant
(
    int                       canon_gamut_num,
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr 
)
{
    IMPORT double min_absolute_pixel_error;
    IMPORT double min_relative_pixel_error;
    IMPORT double min_relative_diag_error;
    int           plot_id_list[ MAX_NUM_2D_PLOTS ];
    Hull_result   hull_result;
    double        absolute_pixel_error;
    double        relative_pixel_error;
    double        relative_diag_error;
    double        illum_hull_x_max     = NOT_SET;
    double        illum_hull_y_max     = NOT_SET;
    int           map_res = NO_ERROR;
    int           try     = 0;
    Matrix*       derived_data_mp = NULL;

    initialize_hull_result(&hull_result);

    EPE(initialize_2D_plots(canon_gamut_num, cc_methods_ptr->plot_enable, 
                            plot_id_list));

    /*
    // Note: Even if we use negative data, negative blue components will cause
    // error.
    */
    ERE(process_negative_data(&derived_data_mp, data_mp, SKIP_NEGATIVE_DATA,
                              "input data"));

    absolute_pixel_error = min_absolute_pixel_error;
    relative_pixel_error = min_relative_pixel_error;
    relative_diag_error  = min_relative_diag_error;

    while (try < 20)
    {
        try++; 

        map_res = do_2D_maps(cc_methods_ptr, plot_id_list, 
                             prompt_to_increase_error_flag, 
                             canon_gamut_mp, 
                             data_point_nums, derived_data_mp,
                             absolute_pixel_error, relative_pixel_error,
                             &(result_ptr->gamut_2D[ canon_gamut_num]),
                             &hull_result, cpu_time_ptr,
                             &illum_hull_x_max, &illum_hull_y_max);

        if (    (map_res == NO_SOLUTION)
             && (ok_to_increase_pixel_error(prompt_to_increase_error_flag,
                                            &absolute_pixel_error,
                                            &relative_pixel_error))
           )
        {
#ifdef TEST
            interactive_pso("Freeing hull result .... ");
#endif
            clear_hull_result(&hull_result);
#ifdef TEST
            interactive_pso("done.\n");
#endif

            if (plot_id_list[ INTERSECTION_PLOT ] > 0)
            {
                EPE(plot_clear(plot_id_list[ INTERSECTION_PLOT ]));
            }
        }
        else
        {
            break;
        }
    }

    if (map_res == NO_SOLUTION)
    {
        set_error("Unable to find a 2D gamut solution after %d tries.",
                  try);
        
        if (absolute_pixel_error > 0.0)
        {
            add_error("Greatest absolute pixel error used was %.1f.",
                      absolute_pixel_error);
        }
        if (relative_pixel_error > 0.0)
        {
            add_error("Greatest pixel error used was  %.1f%%.", 
                      (100.0 * relative_pixel_error));
        }

        map_res = ERROR; 
    }

    if ((map_res == NO_ERROR) && (plot_id_list[INTERSECTION_PLOT ] > 0))
    {
        EPE(plot_2D_result(plot_id_list[ INTERSECTION_PLOT ],
                           (plot_id_list[ HULL_PLOT ] > 0), 
                           canon_gamut_num, result_ptr, &hull_result, 
                           illum_hull_x_max, illum_hull_y_max));
    }

#ifdef TEST
    interactive_pso("Freeing hull result .... ");
#endif
    clear_hull_result(&hull_result);
#ifdef TEST
    interactive_pso("done.\n");
#endif

    if (map_res == NO_ERROR)
    {
        input_data_error_ptr->absolute_pixel_error = absolute_pixel_error;
        input_data_error_ptr->relative_pixel_error = relative_pixel_error;
        input_data_error_ptr->diag_error = relative_diag_error;
    }
    else
    {
        input_data_error_ptr->absolute_pixel_error = DBL_HALF_MOST_NEGATIVE;
        input_data_error_ptr->relative_pixel_error = DBL_HALF_MOST_NEGATIVE;
        input_data_error_ptr->diag_error = DBL_HALF_MOST_NEGATIVE;
    }

    free_matrix(derived_data_mp);

    return map_res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_2D_maps
(
    Colour_constancy_methods* cc_methods_ptr,
    int*                      plot_id_list,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    double                    absolute_pixel_error,
    double                    relative_pixel_error,
    Gamut_result*             result_ptr,
    Hull_result*              hull_result_ptr,
    Gamut_error*              cpu_time_ptr,
    double*                   illum_hull_x_max_ptr,
    double*                   illum_hull_y_max_ptr 
)
{
    Constraints constraints;
    int         res;

    initialize_constraints(&constraints);

    res = do_2D_surface_maps(plot_id_list, canon_gamut_mp, data_point_nums,
                             data_mp, absolute_pixel_error, 
                             relative_pixel_error,
                             &(constraints.surface_hull_list_head));
    if (res != ERROR)
    {
        /*
        // In theory this could be calculated once unless one of a certain
        // small number of things changed (illum RGB, canon illum RGB, hull
        // intersection resolution, sharpening). However, the code is
        // simpler (at least without major changes) if we just re-calculate
        // this. If we wanted to optimize the code, then the illuminant hull
        // should be calculated just once (until something relavent changes).
        //
        // NOTE: If this change is made, it will be necessary to compute
        //       both sharp and dull illuminant hulls. (And 2D/3D). As
        //       discussed in the TODO file, it would probably be best to
        //       create an environment/cache structure for this hull and
        //       the canonical RGB hull.
        //
        // Note that there is always only one illum hull even though we are
        // using a queue. This is to remain consistent with the other two hull
        // sets. (I'm not sure it this is a win). To change this, code in
        // colour_constancy.c will also have to be changed.
        */

        res = do_2D_illum_maps(plot_id_list, &constraints,
                               illum_hull_x_max_ptr, illum_hull_y_max_ptr);
    }

    if (res != ERROR)
    {
#ifdef TEST
        interactive_pso("Choosing solutions for 2D gamut mapping.\n");
#endif

        res = do_gamut_colour_constancy(cc_methods_ptr,
                                        prompt_to_increase_error_flag,
                                        &constraints,
                                        result_ptr,
                                        hull_result_ptr,
                                        cpu_time_ptr);
    }

#ifdef TEST
    interactive_pso("Freeing constraints .... ");
#endif

    free_constraints(&constraints);

#ifdef TEST
    interactive_pso("done.\n");
#endif

    return res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_2D_surface_maps
(
    int*            plot_id_list,
    Matrix*         canon_gamut_mp,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    Queue_element** hull_list_head_ptr 
)
{
    IMPORT int  use_data_hull;
    IMPORT double relative_sensor_error;
    Matrix*     projected_data_mp     = NULL;
    Matrix*     error_box_mp          = NULL;
    Matrix*     projected_gamut_mp    = NULL;
    Matrix*     canon_error_box_mp;
    Hull*       canon_hp;
    int         result                = NO_ERROR;

    if (data_mp->num_rows == 0)
    {
        *hull_list_head_ptr = NULL;
        return NO_ERROR;
    }

#ifdef TEST
    if (is_interactive()) TEST_PSO(("Doing 2D surface maps.\n"));
#endif

    ERE(get_projection_matrix(&projected_gamut_mp, canon_gamut_mp,
                              DIVIDE_BY_BLUE, TRUNCATE_INVALID_CHROMATICITY,
                              MAX_X_DIV_B, DBL_NOT_SET,
                              relative_sensor_error, &canon_error_box_mp,
                              (Matrix**) NULL));

    /*
    // The following does not seem very likely, since we have already returned
    // ERROR if there were negative blue values.
    */
    if (projected_gamut_mp->num_rows < canon_gamut_mp->num_rows)
    {
        pso("Skipping %d canonical hull points due to small blue values.\n",
            canon_gamut_mp->num_rows - projected_gamut_mp->num_rows);
    }

    if (relative_sensor_error > 10 * DBL_EPSILON)
    {
        Matrix *canon_error_mp;

        NRE(canon_error_mp = expand_error_box(canon_error_box_mp));

        canon_hp = find_convex_hull(canon_error_mp, DEFAULT_HULL_OPTIONS);
        free_matrix(canon_error_mp);
    }
    else 
    {
        canon_hp = find_convex_hull(projected_gamut_mp, DEFAULT_HULL_OPTIONS);
    }

    if (canon_hp == NULL) result = ERROR; 

    /* ---------------------------------------------------------------------- */

    if (result != ERROR) 
    {
        result = get_projection_matrix(&projected_data_mp, data_mp,
                                       DIVIDE_BY_BLUE, 
                                       TRUNCATE_INVALID_CHROMATICITY, 
                                       MAX_X_DIV_B,
                                       absolute_pixel_error, 
                                       relative_pixel_error,
                                       &error_box_mp, (Matrix**)NULL);
    }

    if (result != ERROR) 
    {
        if (use_data_hull)
        {
            result = do_2D_hull_surface_maps(plot_id_list, canon_hp,
                                             projected_data_mp, error_box_mp,
                                             hull_list_head_ptr);
        }
        else
        {
            result = do_2D_point_surface_maps(plot_id_list, data_point_nums, 
                                              canon_hp, projected_data_mp,
                                              error_box_mp, hull_list_head_ptr);
        }
    }

    free_matrix(canon_error_box_mp);
    free_matrix(error_box_mp);
    free_matrix(projected_data_mp);
    free_matrix(projected_gamut_mp); 

    free_hull(canon_hp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_2D_hull_surface_maps
(
    int*            plot_id_list,
    Hull*           canon_hp,
    Matrix*         projected_data_mp,
    Matrix*         error_box_mp,
    Queue_element** hull_list_head_ptr 
)
{
    Matrix*        projected_gamut_hull_mp;
    Matrix*        temp_mp;
    int            i, j; 
    char           name[ 200 ];
    Hull*          data_hp;
    Hull*          hp;
    Matrix*        data_hull_vertex_mp;
    int            num_data_points;
    int            num_data_hull_points;
    double           temp;
    Matrix_vector** facet_matrix_vector_list = NULL;
    Matrix*        hull_error_box_mp;
    double           min_r_div_b;
    double           max_r_div_b;
    double           min_g_div_b;
    double           max_g_div_b;
    double           c_r;
    double           c_g;
    int            count;
    int            result = NO_ERROR; 

    if (    (plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] > 0) 
         && (canon_hp != NULL))
    {
        Hull**  no_error_hp_list;
        Hull*   no_error_hp;
        Matrix* no_error_temp_mp;

        projected_gamut_hull_mp = canon_hp->vertex_mp;

        num_data_points = projected_data_mp->num_rows;

        if (num_data_points > 2)
        {
            NRE(data_hp = find_convex_hull(projected_data_mp,
                                           DEFAULT_HULL_OPTIONS));

            data_hull_vertex_mp = data_hp->vertex_mp;
        }
        else
        {
            data_hp = NULL;
            data_hull_vertex_mp = projected_data_mp;
        }

        num_data_hull_points = data_hull_vertex_mp->num_rows;

        NRE(no_error_temp_mp = create_matrix(projected_gamut_hull_mp->num_rows,
                                             2));

        NRE(no_error_hp_list = N_TYPE_MALLOC(Hull *, num_data_hull_points));

        for (i=0; i<num_data_hull_points; i++)
        {
            for (j=0; j<projected_gamut_hull_mp->num_rows; j++)
            {
                temp = (data_hull_vertex_mp->elements)[ i ][ 0 ];

                if (temp < MIN_X_DIV_B)      temp = MIN_X_DIV_B;
                else if (temp > MAX_X_DIV_B) temp = MAX_X_DIV_B;

                (no_error_temp_mp->elements)[j][ 0 ] =
                        ((projected_gamut_hull_mp->elements)[ j ][ 0 ])/temp;

                temp = (data_hull_vertex_mp->elements)[ i ][ 1 ];

                if (temp < MIN_X_DIV_B)      temp = MIN_X_DIV_B;
                else if (temp > MAX_X_DIV_B) temp = MAX_X_DIV_B;

                (no_error_temp_mp->elements)[j][ 1 ] =
                        ((projected_gamut_hull_mp->elements)[ j ][ 1 ])/temp;
            }

            NRE(no_error_hp = find_convex_hull(no_error_temp_mp,
                                               DEFAULT_HULL_OPTIONS));

            no_error_hp_list[ i ] = no_error_hp;
        }

        EPE(plot_2D_hull_list(plot_id_list[ HULL_WITH_NO_ERROR_PLOT ],
                              num_data_hull_points, no_error_hp_list,
                              "S Constraint(s) w/o error"));

        for (i=0; i<num_data_hull_points; i++)
        {
            free_hull(no_error_hp_list[ i ]);
        }

        kjb_free(no_error_hp_list);
        free_matrix(no_error_temp_mp);
        free_hull(data_hp);
    }

    projected_gamut_hull_mp = canon_hp->vertex_mp;

    num_data_points = projected_data_mp->num_rows;

    if (num_data_points > 2)
    {
        int n, m, found;

        NRE(data_hp = find_convex_hull(projected_data_mp, 
                                       DEFAULT_HULL_OPTIONS));
        data_hull_vertex_mp = data_hp->vertex_mp;;

        count = 0;

        NRE(hull_error_box_mp = create_matrix(data_hull_vertex_mp->num_rows,4));

        for (m=0; m<data_hull_vertex_mp->num_rows; m++)
        {
            found = FALSE;

            for (n=0; n<projected_data_mp->num_rows; n++)
            {
                if (    (IS_ALMOST_EQUAL((data_hull_vertex_mp->elements)[m][0],
                                         (projected_data_mp->elements)[n][0],
                                         0.00001)) 
                    &&
                        (IS_ALMOST_EQUAL((data_hull_vertex_mp->elements)[m][1],
                                         (projected_data_mp->elements)[n][1],
                                         0.00001)))
                {
                    (hull_error_box_mp->elements)[count][0] =
                                               (error_box_mp->elements)[n][0];
                    (hull_error_box_mp->elements)[count][1] =
                                               (error_box_mp->elements)[n][1];
                    (hull_error_box_mp->elements)[count][2] =
                                               (error_box_mp->elements)[n][2];
                    (hull_error_box_mp->elements)[count][3] =
                                               (error_box_mp->elements)[n][3];

                    count++;
                    found = TRUE;

                    break;
                }
            }

            if ( ! found)
            {
                set_error("Not able to assign all hull points error.");
                return ERROR;
            }
        }
    }
    else
    {
        data_hp             = NULL;
        data_hull_vertex_mp = projected_data_mp;
        hull_error_box_mp   = error_box_mp;
    }

    num_data_hull_points = data_hull_vertex_mp->num_rows;

    NRE(temp_mp = create_matrix(4 * (projected_gamut_hull_mp->num_rows), 2));

    if (plot_id_list[ HULL_PLOT ] > 0)
    {
        NRE(facet_matrix_vector_list = create_matrix_vector_list(
                                                        num_data_hull_points));
    }

    i = 0;

    while ((result != ERROR) && (i<num_data_hull_points))
    {
        count = 0;

        min_r_div_b = (hull_error_box_mp->elements)[ i ][ 0 ];
        if (min_r_div_b < MIN_X_DIV_B) min_r_div_b = MIN_X_DIV_B;

        max_r_div_b = (hull_error_box_mp->elements)[ i ][ 1 ];
        if (max_r_div_b > MAX_X_DIV_B) max_r_div_b = MAX_X_DIV_B;

        min_g_div_b = (hull_error_box_mp->elements)[ i ][ 2 ];
        if (min_g_div_b < MIN_X_DIV_B) min_g_div_b = MIN_X_DIV_B;

        max_g_div_b = (hull_error_box_mp->elements)[ i ][ 3 ];
        if (max_g_div_b > MAX_X_DIV_B) max_g_div_b = MAX_X_DIV_B;

        for (j=0; j<projected_gamut_hull_mp->num_rows; j++)
        {
            c_r = (projected_gamut_hull_mp->elements)[ j ][ 0 ];

            if      (c_r < MIN_X_DIV_B) c_r = MIN_X_DIV_B;
            else if (c_r > MAX_X_DIV_B) c_r = MAX_X_DIV_B;

            c_g = (projected_gamut_hull_mp->elements)[ j ][ 1 ];

            if      (c_g < MIN_X_DIV_B) c_g = MIN_X_DIV_B;
            else if (c_g > MAX_X_DIV_B) c_g = MAX_X_DIV_B;

            (temp_mp->elements)[count][ 0 ] = c_r / min_r_div_b;
            (temp_mp->elements)[count][ 1 ] = c_g / min_g_div_b;
            count++;
            (temp_mp->elements)[count][ 0 ] = c_r / min_r_div_b;
            (temp_mp->elements)[count][ 1 ] = c_g / max_g_div_b;
            count++;
            (temp_mp->elements)[count][ 0 ] = c_r / max_r_div_b;
            (temp_mp->elements)[count][ 1 ] = c_g / min_g_div_b;
            count++;
            (temp_mp->elements)[count][ 0 ] = c_r / max_r_div_b;
            (temp_mp->elements)[count][ 1 ] = c_g / max_g_div_b;
            count++;
        }

        hp = find_convex_hull(temp_mp, DEFAULT_HULL_OPTIONS);

        if (hp == NULL)
        {
            result = ERROR; 
        }
        else if (plot_id_list[ HULL_PLOT ] > 0)
        {
            result = copy_matrix_vector(&(facet_matrix_vector_list[ i ]), 
                                       hp->facets);
        }

        if (result != ERROR) 
        {
            result = insert_into_queue(hull_list_head_ptr, 
                                       (Queue_element**)NULL, (void*)hp);
        }

        i++; 
    }

    if (plot_id_list[ HULL_PLOT ] > 0)
    {
        if (result != ERROR) 
        {
            result = kjb_sprintf(name, sizeof(name), "S Constraint(s)");
        }
        
        if (result != ERROR) 
        {
            EPE(plot_matrix_vector_list_cols(plot_id_list[ HULL_PLOT ], 
                                            num_data_hull_points,
                                            facet_matrix_vector_list, name));
        }

        free_matrix_vector_list(num_data_hull_points, facet_matrix_vector_list);
    }

    free_matrix(temp_mp);

    if (data_hp != NULL)    /* Then hull_error_box_mp was allocated. */
    {
        free_matrix(hull_error_box_mp);
        free_hull(data_hp);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int do_2D_point_surface_maps
(
    int*            plot_id_list,
    int*            data_point_nums,
    Hull*           canon_hp,
    Matrix*         projected_data_mp,
    Matrix*         error_box_mp,
    Queue_element** hull_list_head_ptr 
)
{
    Matrix*     projected_gamut_hull_mp;
    Matrix*     temp_mp;
    Matrix*     error_temp_mp;
    int         i, j, m, n;
    char        name[ 200 ];
    Hull*       hp;
    double        temp;
    int         num_data_points; 

    num_data_points = projected_data_mp->num_rows; 

    if (    (plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] > 0) 
         && (canon_hp != NULL))
    {
        int  plot_id = plot_id_list[ HULL_WITH_NO_ERROR_PLOT ];

        projected_gamut_hull_mp = canon_hp->vertex_mp;

        NRE(temp_mp = create_matrix(projected_gamut_hull_mp->num_rows, 2));

        for (i = 0; i<num_data_points; i++)
        {
            for (j = 0; j<projected_gamut_hull_mp->num_rows; j++)
            {
                temp = (projected_data_mp->elements)[ i ][ 0 ];

                if (temp < MIN_X_DIV_B) temp = MIN_X_DIV_B;

                (temp_mp->elements)[j][ 0 ] =
                        ((projected_gamut_hull_mp->elements)[ j ][ 0 ])/temp;

                temp = (projected_data_mp->elements)[ i ][ 1 ];

                if (temp < MIN_X_DIV_B) temp = MIN_X_DIV_B;

                (temp_mp->elements)[j][ 1 ] =
                        ((projected_gamut_hull_mp->elements)[ j ][ 1 ])/temp;

            }

            NRE(hp = find_convex_hull(temp_mp, DEFAULT_HULL_OPTIONS));

            ERE(kjb_sprintf(name, sizeof(name), "S Constraint %d w/o error", 
                            data_point_nums[ i ]));

            EPE(plot_matrix_vector_cols(plot_id, hp->facets, name));

            free_hull(hp);
        }

        free_matrix(temp_mp);
    }

    projected_gamut_hull_mp = canon_hp->vertex_mp;

    NRE(error_temp_mp = create_matrix(4*(projected_gamut_hull_mp->num_rows),2));

    for (i = 0; i<num_data_points; i++)
    {
        for (j = 0; j<projected_gamut_hull_mp->num_rows; j++)
        {
            for (m=0; m<2; m++)
            {
                for (n=0; n<2; n++)
                {
                    temp = (error_box_mp->elements)[ i ][ m ];

                    if (temp < MIN_X_DIV_B)
                    {
                        temp = MIN_X_DIV_B;
                    }

                    (error_temp_mp->elements)[4*j + 2*m + n][ 0 ] =
                            ((projected_gamut_hull_mp->elements)[j][0])/temp;

                    temp = (error_box_mp->elements)[ i ][ 2 + n];

                    if (temp < MIN_X_DIV_B)
                    {
                        temp = MIN_X_DIV_B;
                    }

                    (error_temp_mp->elements)[4*j + 2*m + n ][ 1 ] =
                           ((projected_gamut_hull_mp->elements)[ j ][ 1 ])/temp;
                }
            }
        }

        hp = find_convex_hull(error_temp_mp, DEFAULT_HULL_OPTIONS);

        if (hp == NULL) 
        {
            free_matrix(error_temp_mp);
            return ERROR;
        }

        if (plot_id_list[ HULL_PLOT ] > 0)
        {
            kjb_sprintf(name, sizeof(name), "S Constraint %d", 
                        data_point_nums[ i ]);

            EPE(plot_matrix_vector_cols(plot_id_list[ HULL_PLOT ], hp->facets,
                                       name));
        }

        ERE(insert_into_queue(hull_list_head_ptr, (Queue_element**)NULL, 
                              (void*)hp));
    }

    free_matrix(error_temp_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_2D_illum_maps
(
    int*         plot_id_list,
    Constraints* constraints_ptr,
    double*      illum_hull_x_max_ptr,
    double*      illum_hull_y_max_ptr 
)
{
    IMPORT double relative_sensor_error;
    Matrix*       projected_illum_mp        = NULL;
    Vector*       error_illum_2D_trans_r_vp = NULL;
    Vector*       error_illum_2D_trans_g_vp = NULL;
    Vector*       error_hull_x1_vp          = NULL;
    Vector*       error_hull_x2_vp          = NULL;
    Vector*       error_hull_y1_vp          = NULL;
    Vector*       error_hull_y2_vp          = NULL;
    double        canon_illum_R             = DBL_NOT_SET;
    double        canon_illum_G             = DBL_NOT_SET;
    double        canon_illum_B             = DBL_NOT_SET;
    Matrix*       projected_illum_error_box_mp    = NULL;
    Matrix*       projected_illum_error_points_mp = NULL;
    Hull*         illum_hp;
    Vector*       hull_bounds_vp = NULL;
    Vector*       canon_illum_vp = NULL;
    Matrix*       illum_mp       = NULL;
    int           result         = NO_ERROR;
    Queue_element** hull_list_head_ptr=&(constraints_ptr->illum_hull_list_head);

    /*
    // If there is no illumination hull or canon illum, then we still want to
    // carry on, so just return NO_ERROR.
    */
    if (    (get_sharp_illum_gamut(&illum_mp) == ERROR) 
         || (get_sharp_canon_illum(&canon_illum_vp) == ERROR)
       )
    {
        free_matrix(illum_mp);
        free_vector(canon_illum_vp); 
        return NO_ERROR;
    }

#ifdef TEST
    if (is_interactive()) TEST_PSO(("Doing 2D illum maps.\n"));
#endif

    if (result != ERROR)
    {
        Vector* min_vp = NULL;
        
        result = get_min_matrix_col_elements(&min_vp, illum_mp);

        if (result != ERROR) 
        {
            double min_illum_blue = (min_vp->elements)[ BLUE_INDEX ];

            if (min_illum_blue < MIN_ALLOWABLE_2D_ILLUM_VALUE)
            {
                set_error("Illum RGB contains blue values less than %1.f.",
                          MIN_ALLOWABLE_2D_ILLUM_VALUE);
                result = ERROR;
            }
            else if (min_illum_blue < MIN_USABLE_2D_ILLUM_VALUE)
            {
                char sharp_method_string[ 100 ];

                get_current_sharp_method_string(sharp_method_string,
                                                sizeof(sharp_method_string));

                warn_pso("Illum RGB contains blue values less than %.1f\n",
                         MIN_USABLE_2D_ILLUM_VALUE); 
                warn_pso("Min illum RGB value is %.2e (min allowable is %.1f).\n",
                         min_illum_blue, 
                         MIN_ALLOWABLE_2D_CANON_GAMUT_VALUE);
                warn_pso("Resetting those values to %.1f\n", 
                         MIN_USABLE_2D_ILLUM_VALUE); 
                warn_pso("Problem occured doing 2D gamut mapping with %q.\n\n", 
                         sharp_method_string);

                result = ow_min_thresh_matrix_col(illum_mp, BLUE_INDEX, 
                                                  MIN_USABLE_2D_ILLUM_VALUE);
            }
        }

        free_vector(min_vp);
    }

    if (result != ERROR) 
    {
        result = copy_vector(&(constraints_ptr->canon_illum_vp), 
                             canon_illum_vp);
    }

    if (result != ERROR) 
    {
        result = project_matrix(&projected_illum_mp, illum_mp, DIVIDE_BY_SUM);
    }

    if (result != ERROR) 
    {
        constraints_ptr->illum_hp = find_convex_hull(projected_illum_mp, 
                                                     DEFAULT_HULL_OPTIONS);
        if (constraints_ptr->illum_hp == NULL) result = ERROR; 
    }
                
    if (result != ERROR)
    {
        canon_illum_R = (canon_illum_vp->elements)[ RED_INDEX ];
        canon_illum_G = (canon_illum_vp->elements)[ GREEN_INDEX ];
        canon_illum_B = (canon_illum_vp->elements)[ BLUE_INDEX ];

        result = get_projection_matrix(&projected_illum_mp, illum_mp,
                                       DIVIDE_BY_BLUE, 
                                       TRUNCATE_INVALID_CHROMATICITY, 
                                       MAX_X_DIV_B, DBL_NOT_SET,
                                       relative_sensor_error,
                                       &projected_illum_error_box_mp,
                                       (Matrix**) NULL);
    }

    if (result != ERROR)
    {
        double min = min_matrix_element(projected_illum_mp);

        if (min < MIN_ILLUMINANT_CHROMATICITY_COMPONANT)
        {
            set_error("Illumination chromaticity componant less than %e found.",
                      MIN_ILLUMINANT_CHROMATICITY_COMPONANT);
            result = ERROR;
        }
    }

    if (result != ERROR)
    {
        projected_illum_error_points_mp = 
                               expand_error_box(projected_illum_error_box_mp);
        if (projected_illum_error_box_mp == NULL) result = ERROR; 
    }

    if (result != ERROR)
    {
        double min = min_matrix_element(projected_illum_error_points_mp);

        if (min < MIN_ILLUMINANT_CHROMATICITY_COMPONANT)
        {
            set_error("Illumination chromaticity componant less than %e found.",
                      MIN_ILLUMINANT_CHROMATICITY_COMPONANT);
            result = ERROR;
        }
    }

    if (result == ERROR) goto cleanup; 

    if (    (plot_id_list != NULL) 
         && (plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] > 0))
    {
        Vector* illum_2D_trans_r_vp = NULL;
        Vector* illum_2D_trans_g_vp = NULL;
        Vector* hull_x1_vp;
        Vector* hull_x2_vp;
        Vector* hull_y1_vp;
        Vector* hull_y2_vp;

        ERE(get_matrix_col(&illum_2D_trans_r_vp, projected_illum_mp,
                           RED_INDEX));

        ERE(get_matrix_col(&illum_2D_trans_g_vp, projected_illum_mp,
                           GREEN_INDEX));

        ERE(ow_multiply_vector_by_scalar(illum_2D_trans_r_vp, 
                                        canon_illum_B/canon_illum_R));
        ERE(ow_multiply_vector_by_scalar(illum_2D_trans_g_vp, 
                                        canon_illum_B/canon_illum_G));

        ERE(get_2D_hull_segments(illum_2D_trans_r_vp, illum_2D_trans_g_vp,
                                 &hull_x1_vp, &hull_x2_vp, &hull_y1_vp,
                                 &hull_y2_vp));

        ERE(do_illum_2D_inv_trans_gamut(plot_id_list,
                                        illum_2D_trans_r_vp, 
                                        illum_2D_trans_g_vp,
                                        hull_x1_vp, hull_x2_vp,
                                        hull_y1_vp, hull_y2_vp,
                                        (Queue_element**) NULL));

        free_vector(hull_x1_vp);
        free_vector(hull_x2_vp);
        free_vector(hull_y1_vp);
        free_vector(hull_y2_vp);

        free_vector(illum_2D_trans_r_vp);
        free_vector(illum_2D_trans_g_vp); 
    }

    ERE(get_matrix_col(&error_illum_2D_trans_r_vp,
                       projected_illum_error_points_mp, RED_INDEX));

    ERE(get_matrix_col(&error_illum_2D_trans_g_vp,
                       projected_illum_error_points_mp, GREEN_INDEX));

    ERE(ow_multiply_vector_by_scalar(error_illum_2D_trans_r_vp, 
                                     canon_illum_B/canon_illum_R));
    ERE(ow_multiply_vector_by_scalar(error_illum_2D_trans_g_vp, 
                                     canon_illum_B/canon_illum_G));

    /*
    // Another hack to fight memory leaks. 
    */
    if (    (ERROR == get_2D_hull_segments(error_illum_2D_trans_r_vp,
                                           error_illum_2D_trans_g_vp,
                                           &error_hull_x1_vp,
                                           &error_hull_x2_vp,
                                           &error_hull_y1_vp,
                                           &error_hull_y2_vp)
            )
         ||
            (ERROR == do_illum_2D_inv_trans_gamut(plot_id_list,
                                                  error_illum_2D_trans_r_vp,
                                                  error_illum_2D_trans_g_vp,
                                                  error_hull_x1_vp,
                                                  error_hull_x2_vp,
                                                  error_hull_y1_vp,
                                                  error_hull_y2_vp,
                                                  hull_list_head_ptr)
             )
        )
    {
        result = ERROR;
        goto cleanup;
    }

    if ((result != ERROR) && (plot_id_list[INTERSECTION_PLOT] > 0))
    {
        EPE(plot_projected_black_body_curve(plot_id_list[ INTERSECTION_PLOT],
                                            canon_illum_vp,
                                            2 /* dimension */));
    }

    illum_hp = (Hull*)((*hull_list_head_ptr)->contents);

    ERE(get_max_matrix_col_elements(&hull_bounds_vp, illum_hp->vertex_mp));

    *illum_hull_x_max_ptr = (hull_bounds_vp->elements)[ 0 ];
    *illum_hull_y_max_ptr = (hull_bounds_vp->elements)[ 1 ];

cleanup :

    free_vector(hull_bounds_vp);

    free_matrix(projected_illum_mp);
    free_matrix(projected_illum_error_box_mp);
    free_matrix(projected_illum_error_points_mp);

    free_vector(error_hull_x1_vp);
    free_vector(error_hull_x2_vp);
    free_vector(error_hull_y1_vp);
    free_vector(error_hull_y2_vp);

    free_vector(error_illum_2D_trans_r_vp);
    free_vector(error_illum_2D_trans_g_vp);

    free_matrix(illum_mp);
    free_vector(canon_illum_vp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_illum_2D_inv_trans_gamut
(
    int*            plot_id_list,
    Vector*         illum_2D_trans_r_vp,
    Vector*         illum_2D_trans_g_vp,
    Vector*         hull_x1_vp,
    Vector*         hull_x2_vp,
    Vector*         hull_y1_vp,
    Vector*         hull_y2_vp,
    Queue_element** hull_list_head_ptr 
)
{
    IMPORT int    plot_illumination_constraint; 
    int           num_facets;
    int           count;
    int           i;
    int           j;
    Vector*       illum_2D_inv_trans_r_vp  = NULL;
    Vector*       illum_2D_inv_trans_g_vp  = NULL;
    Matrix_vector* inverse_gamut_segments   = NULL;
    Matrix*       inverse_gamut_segment_mp = NULL;
    double          inv_x1;
    double          inv_x2;
    double          inv_y1;
    double          inv_y2;
    double          x_temp;
    double          y_temp;
    Matrix*       temp_mp;
    Hull*         hp;
    double          inv_x;
    double          inv_y;
    double          t;
    double          step;

    ERE(invert_vector(&illum_2D_inv_trans_r_vp, illum_2D_trans_r_vp));
    ERE(invert_vector(&illum_2D_inv_trans_g_vp, illum_2D_trans_g_vp));

    if (hull_list_head_ptr != NULL)     /* Call with error (standard). */
    {
        if (    (plot_id_list != NULL) 
             && (plot_id_list[ ILLUM_HULL_POINTS_PLOT ] > 0)
             && plot_illumination_constraint
           )
        {
            EPE(plot_points(plot_id_list[ ILLUM_HULL_POINTS_PLOT ],
                            illum_2D_inv_trans_r_vp, illum_2D_inv_trans_g_vp,
                            "Inverse maps"));

        }
    }
    else
    {
        if (    (plot_id_list != NULL) 
             && (plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] > 0)
             && (plot_id_list[ ILLUM_HULL_POINTS_PLOT ] > 0)
             && plot_illumination_constraint
           )
        {
            EPE(plot_points(plot_id_list[ HULL_WITH_NO_ERROR_PLOT ],
                            illum_2D_inv_trans_r_vp, illum_2D_inv_trans_g_vp,
                            "Inverse maps w/o error"));

        }
    }

    free_vector(illum_2D_inv_trans_r_vp);
    free_vector(illum_2D_inv_trans_g_vp);

    num_facets = hull_x1_vp->length;

    NRE(temp_mp = create_matrix(num_facets*(fs_gamut_2D_inversion_resolution + 1),
                                2));

    if (    (plot_id_list != NULL)
         && (plot_id_list[ ILLUM_INVERSE_HULL_PLOT ] > 0))
    {
        NRE(inverse_gamut_segments = create_matrix_vector(num_facets));
    }

    count = 0;

    for (i=0; i<num_facets; i++)
    {
        if (    (plot_id_list != NULL)
             && (plot_id_list[ ILLUM_INVERSE_HULL_PLOT ] > 0))
        {
            ERE(get_target_matrix(&(inverse_gamut_segments->elements[ i ]),
                                  fs_gamut_2D_inversion_resolution + 1, 2));
            inverse_gamut_segment_mp = inverse_gamut_segments->elements[ i ];
            
        }

        inv_x1 = (hull_x1_vp->elements) [i];
        inv_x2 = (hull_x2_vp->elements) [i];
        inv_y1 = (hull_y1_vp->elements) [i];
        inv_y2 = (hull_y2_vp->elements) [i];

        step = 1.0/fs_gamut_2D_inversion_resolution;
        t = 0.0;

        for (j=0; j<=fs_gamut_2D_inversion_resolution; j++)
        {
            inv_x = (1.0 - t) * inv_x1 + t * inv_x2;
            inv_y = (1.0 - t) * inv_y1 + t * inv_y2;
            x_temp = 1.0 / inv_x;
            y_temp = 1.0 / inv_y;

            (temp_mp->elements)[count][ 0 ] = x_temp;
            (temp_mp->elements)[count][ 1 ] = y_temp;

            if (    (plot_id_list != NULL) 
                 && (plot_id_list[ ILLUM_INVERSE_HULL_PLOT ] > 0))
            {
                (inverse_gamut_segment_mp->elements)[ j ][ 0 ] = x_temp;
                (inverse_gamut_segment_mp->elements)[ j ][ 1 ] = y_temp;
            }

            t += step;
            count++;
        }
    }

    if (    (plot_id_list != NULL) 
         && (plot_id_list[ ILLUM_INVERSE_HULL_PLOT ] > 0)
         && plot_illumination_constraint
       )
    {
        EPE(plot_multi_segment_curve(plot_id_list[ ILLUM_INVERSE_HULL_PLOT],
                                     inverse_gamut_segments, 
                                     "Inverse Hull"));
        free_matrix_vector(inverse_gamut_segments);
    }

    hp = find_convex_hull(temp_mp, DEFAULT_HULL_OPTIONS);

    if (hp == NULL) 
    {
        free_matrix(temp_mp);
        return ERROR;
    }

    if (hull_list_head_ptr != NULL)    /* Call with error (standard). */
    {
        if (    (plot_id_list != NULL)
             && (plot_id_list[ INTERSECTION_PLOT ] > 0)
             && plot_illumination_constraint
           )
        {
            EPE(plot_matrix_vector_cols(plot_id_list[ INTERSECTION_PLOT ], 
                                       hp->facets, "I Constraint"));
        }
    }
    else
    {
        if (    (plot_id_list != NULL)
             && (plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] > 0)
             && plot_illumination_constraint
           )
        {
            EPE(plot_matrix_vector_cols(plot_id_list[ HULL_WITH_NO_ERROR_PLOT ],
                                       hp->facets, "I Constraint w/o error"));
        }
    }

    if (hull_list_head_ptr != NULL) {   /* Call with error (standard). */
        ERE(insert_into_queue(hull_list_head_ptr, (Queue_element**)NULL, 
                              (void*)hp));
    }
    else
    {
        free_hull(hp);
    }

    free_matrix(temp_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_hull_segments
(
    Vector*  x_vp,
    Vector*  y_vp,
    Vector** hull_x1_vpp,
    Vector** hull_x2_vpp,
    Vector** hull_y1_vpp,
    Vector** hull_y2_vpp 
)
{
    Matrix*  vertex_mp;
    Vector*  hull_x1_vp;
    Vector*  hull_x2_vp;
    Vector*  hull_y1_vp;
    Vector*  hull_y2_vp;
    Matrix** facet_mvp;
    int      i;
    int      num_facets;
    Hull*    hp;

    NRE(vertex_mp = create_matrix(x_vp->length, 2));

    ERE(put_matrix_col(vertex_mp, x_vp, 0));
    ERE(put_matrix_col(vertex_mp, y_vp, 1));

    hp = find_convex_hull(vertex_mp, DEFAULT_HULL_OPTIONS);

    if (hp == NULL) 
    {
        free_matrix(vertex_mp);
        return ERROR;
    }

    num_facets = hp->num_facets;

    facet_mvp = hp->facets->elements;

    NRE(hull_x1_vp = create_vector(num_facets));
    NRE(hull_x2_vp = create_vector(num_facets));
    NRE(hull_y1_vp = create_vector(num_facets));
    NRE(hull_y2_vp = create_vector(num_facets));

    for (i=0; i<num_facets; i++)
    {
        (hull_x1_vp->elements)[i] = ((facet_mvp[i])->elements)[0][0];
        (hull_y1_vp->elements)[i] = ((facet_mvp[i])->elements)[0][1];
        (hull_x2_vp->elements)[i] = ((facet_mvp[i])->elements)[1][0];
        (hull_y2_vp->elements)[i] = ((facet_mvp[i])->elements)[1][1];
    }

    *hull_x1_vpp = hull_x1_vp;
    *hull_x2_vpp = hull_x2_vp;
    *hull_y1_vpp = hull_y1_vp;
    *hull_y2_vpp = hull_y2_vp;

    free_hull(hp);
    free_matrix(vertex_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_2D_plots
(
    int  canon_gamut_num,
    int  enable_plotting,
    int* plot_id_list 
)
{
    IMPORT int  prompt_for_plots;
    IMPORT int  print_plot_style;
    int         plot_request_list[ MAX_NUM_2D_PLOTS ];
    int         i;

    for (i=0; i<MAX_NUM_2D_PLOTS; i++)
    {
        plot_request_list[ i ] = FALSE;
    }

    if ((is_interactive()) && (enable_plotting))
    {
        if ( ! prompt_for_plots)
        {
            /*
            // Default behaviour is to only show the intersection plot (unless
            // we are not using X11 in which case don't show anything). All
            // others have been initialized to FALSE above.
            */

            if (X11_is_ok())
            {
                plot_request_list[ INTERSECTION_PLOT ] = TRUE;
            }
        }
        else
        {
            if (yes_no_query("Show illum hull and intersections? (y/n) "))
            {
                plot_request_list[ INTERSECTION_PLOT ] = TRUE;

                if (yes_no_query("Show all hulls? (y/n) "))
                {
                    plot_request_list[ HULL_PLOT] = TRUE;

                    if (yes_no_query("Show without error? (y/n) "))
                    {
                        plot_request_list[ HULL_WITH_NO_ERROR_PLOT ] = TRUE;
                    }
                }

                if (yes_no_query("Show inverse illum hull? (y/n) "))
                {
                    plot_request_list[ ILLUM_INVERSE_HULL_PLOT ] = TRUE;

                    if (yes_no_query("Show illum Hull Points? (y/n) "))
                    {
                        plot_request_list[ ILLUM_HULL_POINTS_PLOT ] = TRUE;
                    }
                }
            }
        }
    }

    for (i=0; i<MAX_NUM_2D_PLOTS; i++)
    {
        plot_id_list[ i ] = 0;
    }

    if (plot_request_list[ INTERSECTION_PLOT ])
    {
        char data_source[ MAX_FILE_NAME_SIZE ];
        char plot_title_buff[ 1000 ];
        char short_name[ 100 ];
        char sharp_method_string[ 100 ]; 
        int  plot_id;

        ERE(plot_id = plot_open());
        plot_id_list[ INTERSECTION_PLOT ] = plot_id;

        get_current_sharp_method_string(sharp_method_string, 
                                        sizeof(sharp_method_string)); 
        get_data_source(data_source, sizeof(data_source));
        BUFF_GEN_GET_LAST_TOKEN(data_source, short_name, "/");

        ERE(kjb_sprintf(plot_title_buff, sizeof(plot_title_buff),
                        "%d: 2D Diagonal Mappings [ %d ] ( %s ) [ %s ]",
                        plot_id, canon_gamut_num + 1, 
                        sharp_method_string, short_name));

        ERE(plot_set_title(plot_id_list[ INTERSECTION_PLOT ],
                           plot_title_buff, 0, 1));

        ERE(plot_set_x_legend(plot_id_list[ INTERSECTION_PLOT ],
                              "First Componant of Diagonal Transform"));

        if (print_plot_style)
        {
            ERE(plot_set_y_legend(plot_id_list[ INTERSECTION_PLOT ],
                                  "Second Componant of Diagonal Transform"));
        }
        else
        {
            ERE(plot_set_y_legend(plot_id_list[ INTERSECTION_PLOT ],
                                  "     Second Componant"));

            /*
            // Hack it for now (should be a routine to do this.)
            */
            EPE(send_command_to_plot(CURRENT_PLOT, "set size 0.7, 0.95"));
        }

        if (plot_request_list[ HULL_PLOT])
        {
            plot_id_list[ HULL_PLOT] = plot_id_list[ INTERSECTION_PLOT ];

            if (plot_request_list[ HULL_WITH_NO_ERROR_PLOT ])
            {
                plot_id_list[ HULL_WITH_NO_ERROR_PLOT ] =
                        plot_id_list[ INTERSECTION_PLOT ];
            }
        }

        if (plot_request_list[ ILLUM_INVERSE_HULL_PLOT ])
        {
            plot_id_list[ ILLUM_INVERSE_HULL_PLOT ] =
                    plot_id_list[ INTERSECTION_PLOT ];

            if (plot_request_list[ ILLUM_HULL_POINTS_PLOT ])
            {
                plot_id_list[ ILLUM_HULL_POINTS_PLOT ] =
                        plot_id_list[ INTERSECTION_PLOT ];
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int plot_2D_hull_list
(
    int         plot_id,
    int         num_hulls,
    Hull**      hp_list,
    const char* title 
)
{
    Hull*          hp;
    int            i;
    Matrix_vector** facet_matrix_vector_list;

    NRE(facet_matrix_vector_list = create_matrix_vector_list(num_hulls));

    for (i=0; i<num_hulls; i++)
    {
        hp = hp_list[ i ];
        ERE(copy_matrix_vector(&(facet_matrix_vector_list[ i ]), hp->facets));
    }

    ERE(plot_matrix_vector_list_cols(plot_id, num_hulls, 
                                     facet_matrix_vector_list, title));

    free_matrix_vector_list(num_hulls, facet_matrix_vector_list); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

