
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "illum_data.h"
#include "synthetic_data.h"
#include "grey.h"

/* -------------------------------------------------------------------------- */

#define AVE_DB_GREY_FILE      "ave_db_grey.rgb"

/* -------------------------------------------------------------------------- */

static Vector* fs_ave_db_grey_vp = NULL;
static char    fs_ave_db_grey_config_file_name[ MAX_FILE_NAME_SIZE ] = { '\0' };

/* -------------------------------------------------------------------------- */

static int write_grey
(
    const char*   directory,
    const char*   file_name,
    const Vector* grey_vp 
);

static void read_default_ave_db_grey_data(void);
static int read_ave_db_grey_from_file(const char* file_name);
static void free_ave_db_grey(void);

static int set_ave_db_grey_vector
(
    const Vector* grey_vp,
    const char*   config_file_name 
);

static int get_grey_appearance
(
    Vector**      grey_appearance_vpp,
    const Matrix* data_mp 
);

/* -------------------------------------------------------------------------- */

int set_grey_file(const char* option, const char* value)
{
    char lc_option[ 100 ]; 
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || match_pattern(lc_option, "ave-db-grey-file")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_ave_db_grey_vp == NULL) 
            {
                ERE(pso("No average database grey has been set.\n"));
            }
            else 
            {
                ERE(pso("Average database grey is (%.1f, %.1f, %.1f).\n",
                        fs_ave_db_grey_vp->elements[ 0 ], 
                        fs_ave_db_grey_vp->elements[ 1 ],
                        fs_ave_db_grey_vp->elements[ 2 ]));

                ERE(pso("Average database grey source is %s.\n",
                        fs_ave_db_grey_config_file_name)); 
            }
        }
        else
        {
            ERE(read_ave_db_grey_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_grey(const char* make_directive, const char* directory)
{
    Matrix*    canon_data_mp         = NULL;
    Vector*    grey_vp               = NULL;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    ! STRCMP_EQ(lc_make_directive, "all") 
         && ! STRCMP_EQ(lc_make_directive, "grey")
       )
    {
        return NOT_FOUND; 
    }

    verbose_pso(5, "Computing DB grey data.\n");

    ERE(get_synthetic_canon_data(&canon_data_mp));
    ERE(average_matrix_rows(&grey_vp, canon_data_mp));
    ERE(set_ave_db_grey_vector(grey_vp, "<make>"));  

    if ((directory != NULL) && (directory[ 0 ] != '\0'))
    {
        ERE(write_grey(directory, AVE_DB_GREY_FILE, grey_vp));
    }
        
    free_vector(grey_vp); 
    free_matrix(canon_data_mp); 

    verbose_pso(10, "Grey data computed.\n");

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_grey
(
    const char*   directory,
    const char*   file_name,
    const Vector* grey_vp 
)
{
    char  file_path[ MAX_FILE_NAME_SIZE ];

    BUFF_CPY(file_path, directory);
    BUFF_CAT(file_path, DIR_STR);
    BUFF_CAT(file_path, file_name);

    return write_row_vector(grey_vp, file_path); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void read_default_ave_db_grey_data(void)
{
    static int     first_time     = TRUE;

    if (first_time && (fs_ave_db_grey_vp == NULL))
    {
        first_time = FALSE; 
        verbose_pso(5, "Reading default average grey data.\n"); 

        /* Ignore the return, as it will fail if the file does not exist. */
        (void)read_ave_db_grey_from_file(AVE_DB_GREY_FILE);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_ave_db_grey_from_file(const char* file_name)
{
    Vector* grey_vp = NULL;
    char config_file_name[ MAX_FILE_NAME_SIZE ];
    int result; 

    ERE(read_vector_from_gamut_config_file(&grey_vp, (const char*)NULL, 
                                           file_name, "database grey",
                                           config_file_name,
                                           sizeof(config_file_name)));

    result = set_ave_db_grey_vector(grey_vp, config_file_name);

    free_vector(grey_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_ave_db_grey_vector
(
    const Vector* grey_vp,
    const char*   config_file_name 
)
{
    Vector* save_db_grey_vp = fs_ave_db_grey_vp;

    ERE(copy_vector(&fs_ave_db_grey_vp, grey_vp));
    BUFF_CPY(fs_ave_db_grey_config_file_name, config_file_name);

    if (save_db_grey_vp == NULL)
    {
        add_cleanup_function(free_ave_db_grey);
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void free_ave_db_grey(void)
{
    free_vector(fs_ave_db_grey_vp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_grey_world(const Matrix* data_mp, Standard_result* result_ptr)
{
    Vector* grey_appearance_vp         = NULL;
    Vector* illum_estimate_vp          = NULL;
    Vector* sharp_ave_db_grey_vp       = NULL;
    int     result                     = NO_ERROR;

    if (data_mp == NULL) return NO_ERROR; 

    /*
    // This is the simplest grey world case. The transform is estimated by the
    // the ratio of the average to the canonical divided by 2. Alternately, the
    // average estimates grey, and twice the average estimates the actual
    // illuminant. The second way is coded. If we are doing sharpening, the
    // data, and therefore the appearance vector is already sharp.
    */
    ERE(get_grey_appearance(&grey_appearance_vp, data_mp));
    ERE(multiply_vector_by_scalar(&illum_estimate_vp, grey_appearance_vp, 2.0));
        
    (void)get_map_to_canon_illum(&(result_ptr->grey_world_vp), 
                                 illum_estimate_vp);

    if (fs_ave_db_grey_vp == NULL)
    {
        read_default_ave_db_grey_data(); 
    }

    if (fs_ave_db_grey_vp != NULL) 
    {
        ASSERT(grey_appearance_vp != NULL); 

        result = sharpen_vector_data(&sharp_ave_db_grey_vp, fs_ave_db_grey_vp);

        /*
        // The mapping is simply the ratio of the appearance (average) under the
        // canonical to the test data. The canonical illuminant is implicit in
        // that the average grey is with respect to the canonical illuminant.
        // If we are doing sharpening, the data, and therefore the appearance
        // vector is already sharp. 
        */
        if (result != ERROR)
        {
            ERE(divide_vectors(&(result_ptr->db_grey_vp), 
                               sharp_ave_db_grey_vp, grey_appearance_vp));
        }
    }

    free_vector(grey_appearance_vp);
    free_vector(illum_estimate_vp);
    free_vector(sharp_ave_db_grey_vp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_grey_appearance
(
    Vector**      grey_appearance_vpp,
    const Matrix* data_mp 
)
{
    double    sum_r;
    double    sum_g;
    double    sum_b;
    double    ave_r;
    double    ave_g;
    double    ave_b;
    int     i;
    Vector* grey_appearance_vp;
    int     num_data_points;

    num_data_points = data_mp->num_rows;

    sum_r = sum_g = sum_b = 0.0;

    for (i=0; i<num_data_points; i++) 
    {
        sum_r += (data_mp->elements)[i][ RED_INDEX ];
        sum_g += (data_mp->elements)[i][ GREEN_INDEX ];
        sum_b += (data_mp->elements)[i][ BLUE_INDEX ];
    }

    ave_r = sum_r / (double)num_data_points;
    ave_g = sum_g / (double)num_data_points;
    ave_b = sum_b / (double)num_data_points;

    ERE(get_target_vector(grey_appearance_vpp, 3));
    grey_appearance_vp = *grey_appearance_vpp; 

    (grey_appearance_vp->elements)[ RED_INDEX ] = ave_r;
    (grey_appearance_vp->elements)[ GREEN_INDEX ] = ave_g;
    (grey_appearance_vp->elements)[ BLUE_INDEX ] = ave_b;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

