
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "spectra.h"
#include "illum_data.h"

/* -------------------------------------------------------------------------- */

#define ILLUM_GAMUT_FILE      "illum_gamut.rgb"
#define ACTUAL_ILLUM_FILE     "actual_illum.rgb"
#define CANON_ILLUM_FILE      "canon_illum.rgb"

/* -------------------------------------------------------------------------- */

static Matrix* fs_cached_illum_gamut_mp = NULL;
static char    fs_illum_gamut_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int     fs_illum_gamut_version   = 0;

static Vector* fs_cached_actual_illum_vp = NULL;
static char    fs_actual_illum_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int     fs_actual_illum_version  = 0;

static Vector* fs_cached_canon_illum_vp = NULL;
static char    fs_canon_illum_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int     fs_canon_illum_version    = 0;

static Vector* fs_cached_ave_illum_vp = NULL;

static Vector* fs_cached_canon_image_white_vp = NULL;
static char    fs_canon_image_white_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };

static Vector* fs_cached_actual_image_white_vp = NULL;
static char    fs_actual_image_white_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };

/* -------------------------------------------------------------------------- */

static int write_illum_gamut(const char* file_name);
static int read_illum_gamut_from_file(const char* file_name);
static int initialize_illum_gamut(void);

static int set_illum_gamut_matrix
(
    Matrix*     illum_gamut_mp,
    const char* config_file 
);

static int write_actual_illum(const char* file_name);
static int read_actual_illum_from_file(const char* file_name);

static int set_actual_illum_vector
(
    Vector*     actual_illum_vp,
    const char* config_file 
);

static int write_canon_illum(const char* file_name);
static int read_canon_illum_from_file(const char* file_name);

static int set_canon_illum_vector
(
    Vector*     canon_illum_vp,
    const char* config_file 
);

static int read_canon_image_white_from_file(const char* file_name);
static int read_actual_image_white_from_file(const char* file_name);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_illum_data_files(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "igf"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_illum_gamut_mp == NULL) 
            {
                ERE(pso("No illuminant gamut has been set.\n"));
            }
            else 
            {
                ERE(pso("Illuminant gamut source is %s.\n", 
                        fs_illum_gamut_config_file)); 
            }
        }
        else
        {
            ERE(read_illum_gamut_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "aif"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_actual_illum_vp == NULL) 
            {
                ERE(pso("No actual illuminant has been set.\n"));
            }
            else 
            {
                ERE(pso("Actual illuminant is (%.1f, %.1f, %.1f).\n", 
                        fs_cached_actual_illum_vp->elements[ 0 ], 
                        fs_cached_actual_illum_vp->elements[ 1 ],
                        fs_cached_actual_illum_vp->elements[ 2 ])); 

                ERE(pso("Actual illuminant source is %s.\n", 
                        fs_actual_illum_config_file)); 
            }
        }
        else
        {
            ERE(read_actual_illum_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "aiw"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_actual_image_white_vp == NULL) 
            {
                ERE(pso("No actual image white has been set.\n"));
            }
            else 
            {
                ERE(pso("Actual image white is (%.1f, %.1f, %.1f).\n", 
                        fs_cached_actual_image_white_vp->elements[ 0 ], 
                        fs_cached_actual_image_white_vp->elements[ 1 ],
                        fs_cached_actual_image_white_vp->elements[ 2 ])); 

                ERE(pso("Actual image white source is %s.\n", 
                        fs_actual_image_white_config_file)); 
            }
        }
        else
        {
            ERE(read_actual_image_white_from_file(value)); 
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "cif"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_canon_illum_vp == NULL) 
            {
                ERE(pso("No canonical illuminant has been set.\n"));
            }
            else 
            {
                ERE(pso("Canonical illuminant is (%.1f, %.1f, %.1f).\n", 
                        fs_cached_canon_illum_vp->elements[ 0 ], 
                        fs_cached_canon_illum_vp->elements[ 1 ],
                        fs_cached_canon_illum_vp->elements[ 2 ])); 

                ERE(pso("Canonical illuminant source is %s.\n", 
                        fs_canon_illum_config_file)); 
            }
        }
        else
        {
            ERE(read_canon_illum_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "ciw"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_canon_image_white_vp == NULL) 
            {
                ERE(pso("No canon image white has been set.\n"));
            }
            else 
            {
                ERE(pso("Canon image white is (%.1f, %.1f, %.1f).\n", 
                        fs_cached_canon_image_white_vp->elements[ 0 ], 
                        fs_cached_canon_image_white_vp->elements[ 1 ],
                        fs_cached_canon_image_white_vp->elements[ 2 ])); 

                ERE(pso("Canon image white source is %s.\n", 
                        fs_canon_image_white_config_file)); 
            }
        }
        else
        {
            ERE(read_canon_image_white_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_illum_gamut(const char* make_directive, const char* file_name)
{
    static int previous_illum_gamut_version   = NOT_SET;
    static int previous_illum_spectra_version = NOT_SET;
    static int previous_sensor_version        = NOT_SET;
    int        illum_spectra_version;
    int        result                         = NO_ERROR;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    ! STRCMP_EQ(lc_make_directive, "all") 
         && ! STRCMP_EQ(lc_make_directive, "gamut")
         && ! STRCMP_EQ(lc_make_directive, "illum")
         && ! STRCMP_EQ(lc_make_directive, "illum-gamut")
       )
    {
        return NOT_FOUND; 
    }

    if (    (fs_illum_gamut_version == previous_illum_gamut_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (illum_spectra_is_up_to_date(previous_illum_spectra_version))
       )
    {
        verbose_pso(10, "Illum gamut is up to date. Skipping make.\n");
        result = fs_illum_gamut_version; 
    }
    else 
    {
        Spectra* illum_sp       = NULL;
        Matrix*  illum_gamut_mp = NULL;

        verbose_pso(5, "Illum gamut is out of date. Remaking it.\n");

        ERE(illum_spectra_version = get_illum_spectra(&illum_sp));

        previous_sensor_version = get_RGB_data_from_spectra(&illum_gamut_mp, 
                                                            illum_sp,
                                                            (Spectra*)NULL);

        if (previous_sensor_version == ERROR) result = ERROR; 

        if (result != ERROR) 
        {
            int i;
            double max, factor;

            for (i=0; i<illum_gamut_mp->num_rows; i++) 
            {
                max = MAX_OF(MAX_OF(illum_gamut_mp->elements[ i ][ 0 ],
                                    illum_gamut_mp->elements[ i ][ 1 ]),
                             illum_gamut_mp->elements[ i ][ 2 ]);

                factor = 255.0 / max;

                illum_gamut_mp->elements[ i ][ 0 ] *= factor; 
                illum_gamut_mp->elements[ i ][ 1 ] *= factor; 
                illum_gamut_mp->elements[ i ][ 2 ] *= factor; 
            }

            result = set_illum_gamut_matrix(illum_gamut_mp, "<make>");  
        }

        if (result != ERROR) 
        {
            ASSERT(result == fs_illum_gamut_version);

            previous_illum_spectra_version = illum_spectra_version;
            previous_illum_gamut_version   = fs_illum_gamut_version;

            verbose_pso(10, "Illum gamut remade.\n");
        }

        free_matrix(illum_gamut_mp); 
        free_spectra(illum_sp); 
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_illum_gamut(file_name) == ERROR) result = ERROR;
    }
            
    return result;  /* ERROR or fs_illum_gamut_version. */
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_illum_gamut(const char* file_name)
{
    return write_matrix(fs_cached_illum_gamut_mp, file_name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_illum_gamut_from_file(const char* file_name)
{
    Matrix* illum_gamut_mp = NULL;
    char    config_file[ MAX_FILE_NAME_SIZE ];
    int     result = NO_ERROR;

    ERE(read_matrix_from_gamut_config_file(&illum_gamut_mp, (const char*)NULL, 
                                           file_name, "illum gamut", 
                                           config_file, sizeof(config_file)));

    if (illum_gamut_mp->num_cols != 3)
    {
        set_error("Illum gamut file must be rows of R, G, and B (three columns).");
        result = ERROR;
    }

    if (result == NO_ERROR)
    {
        /* Null illum_gamut_mp should be OK. */
        result = set_illum_gamut_matrix(illum_gamut_mp, config_file);
    }

    free_matrix(illum_gamut_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_illum_gamut_matrix
(
    Matrix*     new_illum_gamut_mp,
    const char* config_file 
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(copy_matrix(&fs_cached_illum_gamut_mp, new_illum_gamut_mp));
    BUFF_CPY(fs_illum_gamut_config_file, config_file);

    fs_illum_gamut_version++;

    return fs_illum_gamut_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_average_of_illum_gamut_data(Vector** target_vpp)
{
    static int     previous_illum_gamut_version = NOT_SET;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(make_illum_gamut("illum-gamut", "")); 

    if (    (fs_cached_ave_illum_vp == NULL)
         || (previous_illum_gamut_version != fs_illum_gamut_version)
       )
    {
        Matrix* data_mp = NULL;

        verbose_pso(5, "Average of illum data is out of date. Remaking it.\n");

        ERE(initialize_illum_gamut());      
        ERE(average_matrix_rows(&fs_cached_ave_illum_vp, fs_cached_illum_gamut_mp));
        ERE(ow_scale_vector_by_max(fs_cached_ave_illum_vp)); 
        ERE(ow_multiply_vector_by_scalar(fs_cached_ave_illum_vp, 255.0)); 
        free_matrix(data_mp); 

        previous_illum_gamut_version = fs_illum_gamut_version;
    }
    else 
    {
        verbose_pso(10, "Average of illum data is up to date. Skipping make.\n");
    }

    ERE(copy_vector(target_vpp, fs_cached_ave_illum_vp)); 

    return fs_illum_gamut_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_illum_gamut(Matrix** target_mpp)
{

    ERE(initialize_illum_gamut()); 
    ERE(copy_matrix(target_mpp, fs_cached_illum_gamut_mp));
    ERE(ow_sharpen_matrix_data(*target_mpp)); 

    return fs_illum_gamut_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_illum_gamut(void)
{

    if (fs_illum_gamut_version == 0)
    {
        ERE(read_illum_gamut_from_file(ILLUM_GAMUT_FILE));
    }

    if (fs_cached_illum_gamut_mp == NULL)
    {
        set_error("No illumination gamut data is available.");
        return ERROR;
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_actual_illum(const char* make_directive, const char* file_name)
{
    static int previous_sensor_version         = NOT_SET;
    static int previous_actual_illum_version   = NOT_SET;
    static int previous_actual_spectra_version = NOT_SET;
    int        actual_spectra_version;
    int        result;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    (*lc_make_directive != '\0') 
         && ! STRCMP_EQ(lc_make_directive, "all") 
         && ! STRCMP_EQ(lc_make_directive, "gamut") 
         && ! STRCMP_EQ(lc_make_directive, "actual")
         && ! STRCMP_EQ(lc_make_directive, "actual-illum")
       )
    {
        return NOT_FOUND; 
    }

    if (    (fs_actual_illum_version == previous_actual_illum_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (actual_spectra_is_up_to_date(previous_actual_spectra_version))
       )
    {
        verbose_pso(10, "Actual illum RGB is up to date. Skipping make.\n");
        result = fs_actual_illum_version; 
    }
    else 
    {
        Spectra* actual_illum_sp = NULL;
        Vector*  actual_illum_vp = NULL;

        verbose_pso(5, "Actual illum RGB is out of date. Remaking it.\n");

        ERE(actual_spectra_version = get_actual_spectra(&actual_illum_sp));

        previous_sensor_version = get_RGB_vector_from_spectra(&actual_illum_vp, 
                                                              actual_illum_sp, 
                                                              0);
        if (previous_sensor_version == ERROR) 
        {
            result = ERROR; 
        }
        else 
        {
            result = set_actual_illum_vector(actual_illum_vp, "<make>");  
        }

        if (result != ERROR)
        {
            ASSERT(result == fs_actual_illum_version); 

            previous_actual_illum_version   = fs_actual_illum_version;
            previous_actual_spectra_version = actual_spectra_version;

            verbose_pso(10, "Actual illum RGB remade.\n");
        }

        free_vector(actual_illum_vp); 
        free_spectra(actual_illum_sp); 
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_actual_illum(file_name) == ERROR) result = ERROR;
    }
            
    return result; /* Either ERROR or version. */ 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_actual_illum(const char* file_name)
{

    return write_row_vector(fs_cached_actual_illum_vp, file_name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_actual_illum_from_file(const char* file_name)
{
    Vector* actual_illum_vp = NULL;
    char config_file[ MAX_FILE_NAME_SIZE ];
    int result; 

    result = read_vector_from_gamut_config_file(&actual_illum_vp, 
                                                (const char*)NULL, 
                                                file_name, "actual illum",
                                                config_file, 
                                                sizeof(config_file));

    if ((result != ERROR) && (actual_illum_vp->length != 3))
    {
        set_error("The vector read from %s does not have length 3.",
                  config_file);
        result = ERROR;
    }

    /*
    // For now, set the magnitude, as there are many un-normalized data files,
    // and we have yet to come up with the need to have non-normalized data,
    // except in simulation (which is why we normalize here, and not in a lower
    // level routine!
    */
    if ((result != ERROR) && (actual_illum_vp != NULL)) 
    {
        result = ow_scale_vector_by_max(actual_illum_vp); 
    }

    if ((result != ERROR) && (actual_illum_vp != NULL)) 
    {
        result = ow_multiply_vector_by_scalar(actual_illum_vp, 255.0); 
    }

    if ((result != ERROR) && (actual_illum_vp != NULL)) 
    {
        result = set_actual_illum_vector(actual_illum_vp, config_file);
    }

    if (result != ERROR) 
    {
        result = set_actual_illum_vector(actual_illum_vp, config_file);
    }

    free_vector(actual_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_actual_image_white_from_file(const char* file_name)
{
    Vector* actual_image_white_vp             = NULL;
    char    config_file[ MAX_FILE_NAME_SIZE ];

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(read_vector_from_gamut_config_file(&actual_image_white_vp, 
                                           (const char*)NULL, file_name, 
                                           "actual image white",
                                           config_file, sizeof(config_file)));

    ERE(copy_vector(&fs_cached_actual_image_white_vp, actual_image_white_vp));
    BUFF_CPY(fs_actual_image_white_config_file, config_file);

    free_vector(actual_image_white_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_actual_illum_vector
(
    Vector*     new_actual_illum_vp,
    const char* config_file 
)
{
#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(copy_vector(&fs_cached_actual_illum_vp, new_actual_illum_vp));

    /* 
    // Changed for version 0.2.8
    //
    // This has been moved to the file read so that build from spectra can
    // artificially muck with the scaling. 
    //
    // ERE(ow_scale_vector_by_max(fs_cached_actual_illum_vp)); 
    // ERE(ow_multiply_vector_by_scalar(fs_cached_actual_illum_vp, 255.0)); 
    */

    BUFF_CPY(fs_actual_illum_config_file, config_file);

    fs_actual_illum_version++;

    return fs_actual_illum_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_actual_illum(Vector** target_vpp)
{
    
    ERE(get_actual_illum(target_vpp));
    ERE(ow_sharpen_vector_data(*target_vpp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_actual_illum(Vector** target_vpp)
{
    static int first_time = TRUE;

    kjb_clear_error();

    if (first_time && (fs_cached_actual_illum_vp == NULL))
    {
        first_time = FALSE; 
        ERE(read_actual_illum_from_file(ACTUAL_ILLUM_FILE));
    }

    if (fs_cached_actual_illum_vp == NULL)
    {
        insert_error("No test illumination RGB is available.");
        return ERROR;
    }
    else 
    {
        ERE(copy_vector(target_vpp, fs_cached_actual_illum_vp));
    }

    if (fs_cached_actual_image_white_vp != NULL) 
    {
        double factor;
        double actual_image_brightness = 0.0;
        double actual_illum_brightness = 0.0;
        int  i;

        db_rv(fs_cached_actual_illum_vp);
        db_rv(fs_cached_actual_image_white_vp);

        for (i=0; i<3; i++)
        {
            actual_illum_brightness += fs_cached_actual_illum_vp->elements[ i ];
            actual_image_brightness +=fs_cached_actual_image_white_vp->elements[i];
        }

        factor = actual_image_brightness / actual_illum_brightness;
        dbf(factor); 

        ERE(ow_multiply_vector_by_scalar(*target_vpp, factor));
        db_rv(*target_vpp); 
    }

    return fs_actual_illum_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_canon_illum(const char* make_directive, const char* file_name)
{
    static int previous_sensor_version        = NOT_SET;
    static int previous_canon_illum_version   = NOT_SET;
    static int previous_canon_spectra_version = NOT_SET;
    int        canon_spectra_version;
    int        result;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    (*lc_make_directive != '\0') 
         && ! STRCMP_EQ(lc_make_directive, "all") 
         && ! STRCMP_EQ(lc_make_directive, "gamut") 
         && ! STRCMP_EQ(lc_make_directive, "canon")
         && ! STRCMP_EQ(lc_make_directive, "canon-illum")
       )
    {
        return NOT_FOUND; 
    }

    if (    (fs_canon_illum_version == previous_canon_illum_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (canon_spectra_is_up_to_date(previous_canon_spectra_version))
       )
    {
        verbose_pso(10, "Canonical illum RGB is up to date. Skipping make.\n");
        result = fs_canon_illum_version; 
    }
    else 
    {
        Spectra* canon_illum_sp = NULL;
        Vector*  canon_illum_vp = NULL;

        verbose_pso(5, "Canonical illum RGB is out of date. Remaking it.\n");

        ERE(canon_spectra_version = get_canon_spectra(&canon_illum_sp));

        previous_sensor_version = get_RGB_vector_from_spectra(&canon_illum_vp, 
                                                              canon_illum_sp,0);
        if (previous_sensor_version == ERROR) 
        {
            result = ERROR; 
        }
        else 
        {
            result = set_canon_illum_vector(canon_illum_vp, "<make>");  
        } 

        if (result != ERROR)
        {
            ASSERT(result == fs_canon_illum_version); 

            previous_canon_illum_version   = fs_canon_illum_version;
            previous_canon_spectra_version = canon_spectra_version;

            verbose_pso(10, "Canonical illum RGB remade.\n");
        }

        free_spectra(canon_illum_sp); 
        free_vector(canon_illum_vp); 
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_canon_illum(file_name) == ERROR) result = ERROR; 
    }
            
    return result; /* Either ERROR or version. */ 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_canon_illum(const char* file_name)
{
    return write_row_vector(fs_cached_canon_illum_vp, file_name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_illum_from_file(const char* file_name)
{
    Vector* canon_illum_vp = NULL;
    char config_file[ MAX_FILE_NAME_SIZE ];
    int result; 

    result = read_vector_from_gamut_config_file(&canon_illum_vp, 
                                                (const char*)NULL, 
                                                file_name, "canon illum", 
                                                config_file, 
                                                sizeof(config_file));

    if ((result != ERROR) && (canon_illum_vp->length != 3))
    {
        set_error("The vector read from %s does not have length 3.",
                  config_file);
        result = ERROR;
    }

    /*
    // For now, set the magnitude, as there are many un-normalized data files,
    // and we have yet to come up with the need to have non-normalized data.
    */
    if ((result != ERROR) && (canon_illum_vp != NULL)) 
    {
        result = ow_scale_vector_by_max(canon_illum_vp); 
    }

    if ((result != ERROR) && (canon_illum_vp != NULL)) 
    {
        result = ow_multiply_vector_by_scalar(canon_illum_vp, 255.0); 
    }

    if (result != ERROR) 
    {
        result = set_canon_illum_vector(canon_illum_vp, config_file);
    }

    free_vector(canon_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_canon_illum_vector
(
    Vector*     new_canon_illum_vp,
    const char* config_file 
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(copy_vector(&fs_cached_canon_illum_vp, new_canon_illum_vp));

    /* 
    // Changed for version 0.2.8
    //
    // This has been moved to the file read to be consistent with
    // set_actual_illum_vector(). In the case of the canon illum vector, it
    // should not make any difference, since in the non-read case, i.e., make
    // from spectra, the scaling should be inherent in the spectra, and unlike
    // the actual illum case, we never muck with the scaling.
    //
    // ERE(ow_scale_vector_by_max(fs_cached_canon_illum_vp)); 
    // ERE(ow_multiply_vector_by_scalar(fs_cached_canon_illum_vp, 255.0)); 
    */

    BUFF_CPY(fs_canon_illum_config_file, config_file);

    fs_canon_illum_version++;

    return fs_canon_illum_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_image_white_from_file(const char* file_name)
{
    Vector*        canon_image_white_vp = NULL;
    char           config_file[ MAX_FILE_NAME_SIZE ];

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    ERE(read_vector_from_gamut_config_file(&canon_image_white_vp, 
                                           (const char*)NULL, file_name, 
                                           "canon image white",
                                           config_file, sizeof(config_file)));

    ERE(copy_vector(&fs_cached_canon_image_white_vp, canon_image_white_vp));
    BUFF_CPY(fs_canon_image_white_config_file, config_file);

    free_vector(canon_image_white_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_illum_rgb_from_diagonal_transform
(
    Vector** illum_rgb_vpp,
    Vector*  diag_vp 
)
{
    Vector* sharp_canon_illum_vp = NULL;
    Vector* sharp_illum_rgb_vp = NULL;
    int     result; 

    ERE(get_sharp_canon_illum(&sharp_canon_illum_vp));

    result = divide_vectors(&sharp_illum_rgb_vp, sharp_canon_illum_vp, 
                            diag_vp); 

    if (result != ERROR) 
    {
        result = unsharpen_vector_data(illum_rgb_vpp, sharp_illum_rgb_vp);
    }

    free_vector(sharp_illum_rgb_vp);
    free_vector(sharp_canon_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_map_to_canon_illum(Vector** map_vpp, const Vector* input_vp)
{
    Vector* canon_illum_vp         = NULL;
    int     result;

    ERE(get_sharp_canon_illum(&canon_illum_vp));

    result = divide_vectors(map_vpp, canon_illum_vp, input_vp);

    free_vector(canon_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_canon_illum(Vector** target_vpp)
{

    ERE(get_canon_illum(target_vpp));
    ERE(ow_sharpen_vector_data(*target_vpp)); 

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_canon_illum(Vector** target_vpp)
{

    if (fs_canon_illum_version == 0)
    {
        ERE(read_canon_illum_from_file(CANON_ILLUM_FILE));
    }

    if (fs_cached_canon_illum_vp == NULL)
    {
        set_error("No canonical illumination RGB is available.");
        return ERROR;
    }

    ERE(copy_vector(target_vpp, fs_cached_canon_illum_vp));

    /*
    // NEW STUFF, March 26, 01
    //
    // We should normalize the traget image, not how we compute it! Otherwise,
    // the RGB mapping results will be weighted more for brighter canonical
    // target images. 
    */

    return fs_canon_illum_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_normalize_data_by_canon_illum_brightness(Matrix* mp)
{

    if (    (fs_cached_canon_illum_vp != NULL) 
         && (fs_cached_canon_image_white_vp != NULL) 
       )
    {
        double factor;
        double canon_image_brightness;
        /*
        // NEW STUFF, March 26, 01
        //
        // Switch in thinking. We need to normalize the canonical image, not the
        // machinary for calculating it. 
        */
        canon_image_brightness = max_vector_element(fs_cached_canon_image_white_vp);

        factor = 255.0 / canon_image_brightness; 

        dbf(factor); 

        ERE(ow_multiply_matrix_by_scalar(mp, factor));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_vector(fs_cached_canon_illum_vp);
    free_matrix(fs_cached_illum_gamut_mp);
    free_vector(fs_cached_actual_illum_vp);
    free_vector(fs_cached_ave_illum_vp);
    free_vector(fs_cached_canon_image_white_vp);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

