
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "image_input.h"
#include "image_data.h"
#include "illum_data.h"

/* -------------------------------------------------------------------------- */

static KJB_image* fs_input_pp_ip                    = NULL;
static Matrix*    fs_cached_input_data_mp           = NULL;
static int        fs_image_data_version             = 0;

static Matrix*    fs_cached_canon_data_mp           = NULL;
static int*       fs_cached_invalid_registered      = NULL;

static int        fs_process_image_data_before_send = FALSE;

static int        fs_image_block_size               = 5;
static double     fs_valid_pixel_fraction           = 0.2;

/* Negative fs_variance_threshold means that it is not set and not used. */
static double     fs_variance_threshold             = -1.0;

static int        fs_variance_block_size            = -1;
static int        fs_segment_input_image            = FALSE;

/*
// Default value is lower than what you might think (i.e. 245, say) because it
// is applied after offset is removed. 
*/
static double     fs_max_image_data_value            = 235.0;

/* -------------------------------------------------------------------------- */

static int process_image_data(void);

static int process_image_data_guts
(
    const KJB_image* input_ip,
    const KJB_image* canon_ip,
    Matrix**         input_data_mpp,
    Matrix**         canon_data_mpp,
    int**            invalid_registered_ptr,
    KJB_image**      pp_ipp 
);

static int initialize_image_data(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_image_pp_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_int;
    int  temp_boolean_value;
    double temp_real;
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);
    
    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "ibs")
         || match_pattern(lc_option, "image-block-size")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("image-block-size = %d.\n", fs_image_block_size));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Image block size (ibs) is %d.\n", fs_image_block_size));
        }
        else
        {
            ERE(ss1pi(value, &temp_int)); 

            if (temp_int <= 0) 
            {
                set_error("Image block size must be positive."); 
                return ERROR; 
            }

            fs_image_block_size = temp_int;
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "var-block-size")) 
         || (match_pattern(lc_option, "vbs"))
         || (match_pattern(lc_option, "var-fraction-thresh"))
         || (match_pattern(lc_option, "vft"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (    (fs_variance_block_size > 0) 
                 && (fs_variance_threshold > 0.0)
               )
            {
                ERE(pso("Local variance block size (vbs) is %d.\n", 
                        fs_variance_block_size));
                ERE(pso("Local variance fraction threshold (vft) is %f\n", 
                    fs_variance_threshold));
            }
            else 
            {
                ERE(pso("Local variance pre-processing is not being used.\n"));
            }
        }
        else if (    (lc_option[ 0 ] == '\0') 
                  || (match_pattern(lc_option, "var-block-size")) 
                  || (match_pattern(lc_option, "vbs"))
                )
        {
            ERE(ss1pi(value, &temp_int)); 

            if (temp_int <= 0) 
            {
                set_error("Variance block size must be positive."); 
                return ERROR; 
            }

            fs_variance_block_size = temp_int;
        }
        else
        {
            if (is_no_value_word(lc_option))
            {
                fs_variance_threshold = -1.0;
            }
            else 
            {
                ERE(ss1d(value, &temp_real)); 

                if (temp_real > 1.0) 
                {
                    set_error("Variance fraction threshold must be at most 1.");
                    return ERROR;
                }

                fs_variance_threshold = temp_real;
            }
        }
        result = NO_ERROR; 
    }
    
    if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "valid-pixel-fraction"))
        || (match_pattern(lc_option, "vpf"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("valid-pixel-fraction = %3.0f%%.\n", 
                    100.0 * fs_valid_pixel_fraction));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The fraction of valid pixels required in a block "));
            ERE(pso("(vpf) is %3.0f%%.\n", 100.0 * fs_valid_pixel_fraction));
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 

            if (temp_real < 0.0) temp_real = 0.0; 
            fs_valid_pixel_fraction = temp_real / 100.0;

        }
        result = NO_ERROR; 
    }
    
        if (   (lc_option[ 0 ] == '\0') 
        || (match_pattern(lc_option, "max-image-data-value"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_max_image_data_value < 0.0)
            {
                ERE(pso("max-image-data-value = off\n"));
            }
            else 
            {
                ERE(pso("max-image-data-value =  %.1f\n",
                        fs_max_image_data_value));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_max_image_data_value < 0.0)
            {
                ERE(pso("No maximum data value is set.\n"));
            }
            else 
            {
                ERE(pso("The maximum data value used is %.1f.\n",
                        fs_max_image_data_value));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_max_image_data_value = DBL_NOT_SET;
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 
            fs_max_image_data_value = temp_real;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "segment-input-image"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("segment-input-image = %s\n",
                    fs_segment_input_image ? "t" : "f")); 
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Image segmentation %s used for data reduction.\n",
                    fs_segment_input_image ? "is" : "is not")); 
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_segment_input_image = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((result == NO_ERROR) && (value[ 0 ] != '\0'))
    {
        fs_process_image_data_before_send = TRUE; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_image_data_points(void)
{

    ERE(initialize_image_data()); 

    ASSERT(    (fs_cached_input_data_mp != NULL)
            && (fs_cached_input_data_mp->num_rows > 0)); 

    return fs_cached_input_data_mp->num_rows; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_image_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(initialize_image_data()); 

    ASSERT(    (fs_cached_input_data_mp != NULL)
            && (fs_cached_input_data_mp->num_rows > 0)); 

    ERE(select_data(target_mpp, fs_cached_input_data_mp, num_data_points, 
                    data_point_nums, (int*)NULL));

    return fs_image_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_registered_image_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(initialize_image_data()); 

    ASSERT(    (fs_cached_input_data_mp != NULL)
            && (fs_cached_input_data_mp->num_rows > 0)); 

    ERE(select_data(target_mpp, fs_cached_input_data_mp, num_data_points, 
                    data_point_nums, fs_cached_invalid_registered));

    return fs_image_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_registered_image_data(Matrix** target_mpp)
{

    ERE(initialize_image_data()); 

    ASSERT(    (fs_cached_input_data_mp != NULL)
            && (fs_cached_input_data_mp->num_rows > 0)); 

    ERE(copy_matrix_data(target_mpp, fs_cached_input_data_mp, 
                         fs_cached_invalid_registered));

    return fs_image_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_image_data(void)
{

    ERE(process_image_data());

    if (fs_cached_input_data_mp == NULL)
    {
        set_error("No image data has been read.");
        return ERROR;
    }
    else if (fs_cached_input_data_mp->num_rows == 0)
    {
        set_error("No good image data points.");
        return ERROR; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_image_canon_data(Matrix** target_mpp)
{

    ERE(process_image_data());

    /* Note: fs_cached_canon_data_mp may well be NULL. Then *target_mpp
    // becomes NULL also.
    */
    ERE(copy_matrix_data(target_mpp, fs_cached_canon_data_mp, 
                         fs_cached_invalid_registered));

    /*
    // NEW STUFF, March 26, 01
    */
    ERE(ow_normalize_data_by_canon_illum_brightness(*target_mpp)); 

    return fs_image_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_image_canon_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums 
)
{

    ERE(process_image_data());

    ERE(select_data(target_mpp, fs_cached_canon_data_mp, num_data_points, 
                    data_point_nums, fs_cached_invalid_registered));

    /*
    // NEW STUFF, March 26, 01
    */
    ERE(ow_normalize_data_by_canon_illum_brightness(*target_mpp)); 

    return fs_image_data_version; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_image_var_data_points(void)
{
    set_error("Image var data code not yet ready.");
    return ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_image_var_data
(
    Matrix** dummy_target_mpp,
    int      dummy_num_data_points,
    int*     dummy_data_point_nums,
    Matrix** dummy_var_data_mpp 
)
{

    set_error("Var image data not available yet.");
    return ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_registered_image_var_data
(
    Matrix** dummy_target_mpp,
    int      dummy_num_data_points,
    int*     dummy_data_point_nums 
)
{

    set_error("Var image data not available yet.");
    return ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_registered_image_var_data(Matrix** dummy_target_mpp)
{

    set_error("Var image data not available yet.");
    return ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

const KJB_image* get_pp_input_image(void)
{

    ERN(process_image_data());

    if (fs_input_pp_ip == NULL)
    {
        set_error("No preprocessed input image is available.");
    }

    return fs_input_pp_ip; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int process_image_data(void)
{
    static int cached_input_image_version = 0;
    static int cached_canon_image_version = 0;
    const KJB_image* input_ip;
    const KJB_image* canon_ip;
    int              result;
    int              input_image_version;
    int              canon_image_version;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(input_ip = get_input_image(&input_image_version));

    canon_ip = get_canon_image(&canon_image_version);

    if (    ( ! fs_process_image_data_before_send)
         && (input_image_version == cached_input_image_version)
         && (canon_image_version == cached_canon_image_version)
       )
    {
        verbose_pso(10,
                  "Skipping rebuild of image data because it is up to date.\n");
        return NO_ERROR;
    }
    else
    {
        verbose_pso(10, "Rebuilding image data because it is out of date.\n");
    }
        
    if (fs_input_pp_ip != NULL) fs_input_pp_ip->read_only = FALSE;

    result = process_image_data_guts(input_ip, canon_ip, 
                                     &fs_cached_input_data_mp,
                                     &fs_cached_canon_data_mp,
                                     &fs_cached_invalid_registered,
                                     &fs_input_pp_ip);

    if (fs_input_pp_ip != NULL) fs_input_pp_ip->read_only = TRUE;

    if (result != ERROR) 
    {
        fs_image_data_version++;

        cached_canon_image_version = canon_image_version;
        cached_input_image_version = input_image_version;

        fs_process_image_data_before_send = FALSE; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int process_image_data_guts
(
    const KJB_image* input_ip,
    const KJB_image* canon_ip,
    Matrix**         input_data_mpp,
    Matrix**         canon_data_mpp,
    int**            invalid_registered_ptr,
    KJB_image**      pp_ipp 
)
{
    KJB_image*       variance_threshold_ip = NULL;
    KJB_image*       ave_ip       = NULL;
    KJB_image*       ave_canon_ip = NULL;
    Matrix*          data_mp      = NULL;
    Matrix*          canon_mp     = NULL;
    int              result       = NO_ERROR;
    int              valid_canon_image;
    int              num_valid_pixels_required_in_block;
    int              num_pix;
    int*             invalid_registered = NULL;

    ASSERT(input_ip != NULL);

    init_cpu_time();
    init_real_time();
    interactive_pso("Computing image data.\n");

    num_pix = fs_image_block_size * fs_image_block_size; 
    num_valid_pixels_required_in_block = 
                           num_pix * fs_valid_pixel_fraction + EXTRA_FOR_ROUNDING;

    if (result != ERROR)
    {
        result = ave_image_where_uniform(&ave_ip, input_ip,
                                         fs_image_block_size, fs_image_block_size,
                                         num_valid_pixels_required_in_block,
                                         3.0);
    }

    if (result != ERROR)
    {
        result = kjb_copy_image(pp_ipp, ave_ip);
    }

    /* We are done with "ip", and so we are done with "variance_threshold_ip" */
    kjb_free_image(variance_threshold_ip);

    if ((result != ERROR) && (fs_max_image_data_value > 0.0))
    {
        result = mark_pixels_above_threshold(ave_ip,
                                             fs_max_image_data_value);
    }

    if ((result != ERROR) && (canon_ip != NULL))
    {
        if (is_same_size_image(canon_ip, input_ip))
        {
            if (fs_segment_input_image)
            {
                pso("Image segmentation is being used.\n");
                pso("Target (canonical) will be ignored.\n");
                valid_canon_image = FALSE;
            }
            else 
            {
                valid_canon_image = TRUE;
            }
        }
        else
        {
            pso("Image and target are different sizes.\n");
            pso("Target (canonical) will be ignored.\n");
            valid_canon_image = FALSE;
        }
    }
    else
    {
        valid_canon_image = FALSE;
    }

    if ((result != ERROR) && (valid_canon_image))
    {

        kjb_free_image(variance_threshold_ip);

        if (result != ERROR)
        {
            result = ave_image_where_uniform(&ave_canon_ip, canon_ip,
                                             fs_image_block_size,
                                             fs_image_block_size,
                                             num_valid_pixels_required_in_block,
                                             3.0);
        }

        if ((result != ERROR) && (fs_max_image_data_value > 0.0))
        {
            result = mark_pixels_above_threshold(ave_canon_ip,
                                                 fs_max_image_data_value);
        }

        if (result != ERROR)
        {
            result = get_registered_image_pixels(ave_ip, ave_canon_ip,
                                                 &data_mp, &canon_mp,
                                                 &invalid_registered);
        }

        kjb_free_image(ave_canon_ip);

    }
    else if ((result != ERROR) && (fs_segment_input_image))
    {
        Segmentation_t1* segmentation_ptr = NULL;

        verbose_pso(2, "Segmenting input image.\n");

        result = t1_segment_image(ave_ip, &segmentation_ptr);

        if (result != ERROR)
        {
            int num_segments = segmentation_ptr->num_segments; 
            int i; 

            if (result != ERROR)
            {
                result = get_target_matrix(&data_mp, num_segments, 3);  
            }

            if (result != ERROR)
            {
                for (i=0; i<num_segments; i++)
                {
                    Segment_t1* seg_ptr=segmentation_ptr->segments[i];

                    data_mp->elements[ i ][ 0 ] = seg_ptr->R_ave; 
                    data_mp->elements[ i ][ 1 ] = seg_ptr->G_ave; 
                    data_mp->elements[ i ][ 2 ] = seg_ptr->B_ave; 
                }
            }
        }

        t1_free_segmentation(segmentation_ptr); 
    }
    else if (result != ERROR)
    {
        result = get_image_pixels(ave_ip, &data_mp);
    }

    kjb_free_image(ave_ip);

    free_matrix(*canon_data_mpp);
    free_matrix(*input_data_mpp);
    kjb_free(*invalid_registered_ptr); 

    *input_data_mpp = data_mp;

    *canon_data_mpp = canon_mp;     /* NULL, unless canon_ip is not */
    *invalid_registered_ptr = invalid_registered; /* NULL, unless ... */

    if (result != ERROR)
    {
        interactive_pso("Image data computed (%ldms / %ldms).\n", 
                        get_cpu_time(), get_real_time());
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_matrix(fs_cached_input_data_mp);
    free_matrix(fs_cached_canon_data_mp);

    kjb_free(fs_cached_invalid_registered); 

    if (fs_input_pp_ip != NULL) fs_input_pp_ip->read_only = FALSE;

    kjb_free_image(fs_input_pp_ip);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

