
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin 
|  contributed the interface with the correlation code). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "initialize_error.h"

/* -------------------------------------------------------------------------- */

static void initialize_error_by_sharp(Error_by_sharp*, double);
static void initialize_error_by_algorithm(Error_by_algorithm*, double);
static void initialize_best_fit_error(Best_fit_error*, double);
static void initialize_standard_error(Standard_error*, double);
static void initialize_gamut_error(Gamut_error*, double);

/* Lindsay: Colour by Correlation: Feb 17/98 */
static void initialize_correlation_error
(
    Correlation_error* error_ptr,
    double             set_value 
);
/* End Lindsay */

#ifdef SIMPLE_TEMPLATE
static void initialize_simple_template_error
(
    Simple_template_error* error_ptr,
    double                 set_value 
);

#endif 

#ifdef TEMPLATE
static void initialize_template_error
(
    Template_error* error_ptr,
    double          set_value 
);
#endif 

#ifdef MULTI_TEMPLATE
static void initialize_multi_template_error
(
    Multi_template_error* error_ptr,
    double                set_value 
);

#endif 

static void initialize_input_data_error_by_algorithm
(
    Input_data_error_by_algorithm* ,
    double                          
);

static void initialize_gamut_input_data_error
(
    Gamut_input_data_error* ,
    double                   
);

/* -------------------------------------------------------------------------- */

void initialize_error
(
    Colour_constancy_error* error_ptr,
    double                  set_value 
)
{

    if (error_ptr == NULL) return;

    initialize_error_by_sharp(&(error_ptr->illum_angular),           set_value);
    initialize_error_by_sharp(&(error_ptr->illum_sum_chrom),         set_value);

    initialize_error_by_sharp(&(error_ptr->illum_rgb),               set_value);
    initialize_error_by_sharp(&(error_ptr->illum_lab),               set_value);

    initialize_error_by_sharp(&(error_ptr->illum_r_div_sum),         set_value);
    initialize_error_by_sharp(&(error_ptr->illum_g_div_sum),         set_value);
    initialize_error_by_sharp(&(error_ptr->illum_sum_lum),           set_value);

    initialize_error_by_sharp(&(error_ptr->world_map_blue_chrom_rms),set_value);
    initialize_error_by_sharp(&(error_ptr->world_map_sum_chrom_rms), set_value);
    initialize_error_by_sharp(&(error_ptr->world_map_3D_rms),        set_value);
    initialize_error_by_sharp(&(error_ptr->world_map_3D_lab),        set_value);

    initialize_error_by_sharp(&(error_ptr->scene_map_blue_chrom_rms),set_value);
    initialize_error_by_sharp(&(error_ptr->scene_map_sum_chrom_rms), set_value);
    initialize_error_by_sharp(&(error_ptr->scene_map_3D_rms),        set_value);
    initialize_error_by_sharp(&(error_ptr->scene_map_3D_lab),        set_value);

    initialize_error_by_sharp(&(error_ptr->cpu_time),                set_value);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_error_by_sharp
(
    Error_by_sharp* error_ptr,
    double          set_value 
)
{
    int i;

    for (i=0; i<NUM_SHARP_METHODS; i++) 
    {
        initialize_error_by_algorithm(&(error_ptr->sharp[ i ]), set_value);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_error_by_algorithm
(
    Error_by_algorithm* error_ptr,
    double              set_value 
)
{
    int i; 

    initialize_best_fit_error(&(error_ptr->best_fit), set_value);
    initialize_standard_error(&(error_ptr->standard), set_value);

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_error(&(error_ptr->gamut_2D[ i ]), set_value);
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_error(&(error_ptr->gamut_3D[ i ]), set_value);
    }

    initialize_correlation_error(&(error_ptr->correlation), set_value);

#ifdef SIMPLE_TEMPLATE
    initialize_simple_template_error(&(error_ptr->simple_template), set_value);
#endif 

#ifdef TEMPLATE
    initialize_template_error(&(error_ptr->template), set_value);
#endif 

#ifdef MULTI_TEMPLATE
    initialize_multi_template_error(&(error_ptr->multi_template), set_value);
#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_best_fit_error
(
    Best_fit_error* error_ptr,
    double          set_value 
)
{

    error_ptr->world_best_3D_fit      = set_value;
    error_ptr->world_best_2D_fit      = set_value;
    error_ptr->world_best_3D_diag_fit = set_value;
    error_ptr->world_best_2D_diag_fit = set_value;
    error_ptr->scene_best_3D_fit      = set_value;
    error_ptr->scene_best_2D_fit      = set_value;
    error_ptr->scene_best_3D_diag_fit = set_value;
    error_ptr->scene_best_2D_diag_fit = set_value;
    error_ptr->actual                 = set_value;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_standard_error
(
    Standard_error* error_ptr,
    double          set_value 
)
{

    error_ptr->nothing                = set_value;
    error_ptr->ave_illum              = set_value;
    error_ptr->grey_world             = set_value;
    error_ptr->db_grey_world          = set_value;
    error_ptr->scale_by_max           = set_value;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_gamut_error
(
    Gamut_error* error_ptr,
    double       set_value 
)
{
    int i;

    error_ptr->mv_surface                      = set_value;
    error_ptr->mv_surface_illum                = set_value;
    error_ptr->ha_surface                      = set_value;
    error_ptr->ha_surface_illum                = set_value;
    error_ptr->icsa_surface_illum              = set_value;

    error_ptr->wa_surface                      = set_value;
    error_ptr->wa_surface_illum                = set_value;

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        error_ptr->scia_surface_illum[ i ] = set_value;
        error_ptr->scca_surface_illum[ i ] = set_value;
    }

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        error_ptr->ha_surface_lum[         i ] = set_value;
        error_ptr->mv_surface_lum[         i ] = set_value;
        error_ptr->ha_surface_illum_lum[   i ] = set_value;
        error_ptr->mv_surface_illum_lum[   i ] = set_value;
        error_ptr->scla_surface_illum_lum[ i ] = set_value;
        error_ptr->lcsa_surface_illum_lum[ i ] = set_value;
        error_ptr->scia_surface_illum_lum[ i ] = set_value;
        error_ptr->scca_surface_illum_lum[ i ] = set_value;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay: Sept 30/98 */
static void initialize_correlation_error
(
    Correlation_error* error_ptr,
    double             set_value 
)
{
    int i; /* Modified April 14/99 */

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        error_ptr->correlation_01[ i ]                        = set_value;
        error_ptr->correlation_uniform_MAP_inv_dist_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MAP_gaussian_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MAP_fitted[ i ]        = set_value;
        error_ptr->correlation_gaussian_MAP[ i ]              = set_value;
        error_ptr->correlation_uniform_MLM_inv_dist_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MLM_gaussian_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MLM_fitted[ i ]        = set_value;
        error_ptr->correlation_gaussian_MLM[ i ]              = set_value;
        error_ptr->correlation_uniform_MMSE_inv_dist_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MMSE_gaussian_fill[ i ] = set_value;
        error_ptr->correlation_uniform_MMSE_fitted[ i ]        = set_value;
        error_ptr->correlation_gaussian_MMSE[ i ]              = set_value;
    }
}
/* End Lindsay: Sept 30/98 */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE
static void initialize_simple_template_error
(
    Simple_template_error* error_ptr,
    double                 set_value 
)
{

    error_ptr->alg  = set_value;
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE
static void initialize_template_error
(
    Template_error* error_ptr,
    double          set_value 
)
{

    error_ptr->first_alg  = set_value;
    error_ptr->second_alg = set_value;
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE
static void initialize_multi_template_error
(
    Multi_template_error* error_ptr,
    double                set_value 
)
{
    int i;

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        error_ptr->alg[ i ] = set_value;
    }
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_input_data_error
(
    Input_data_error* input_data_error_ptr,
    double            set_value 
)
{
    int i;

    if (input_data_error_ptr == NULL) return;

    for (i=0; i<NUM_SHARP_METHODS; i++) 
    {
        initialize_input_data_error_by_algorithm(
                                            &(input_data_error_ptr->sharp[ i ]),
                                            set_value);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_input_data_error_by_algorithm
(
    Input_data_error_by_algorithm* input_data_error_ptr,
    double                         set_value 
)
{
    int i; 

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_input_data_error(&(input_data_error_ptr->gamut_2D[i ]),
                                          set_value);
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_input_data_error(&(input_data_error_ptr->gamut_3D[ i]),
                                          set_value);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_gamut_input_data_error
(
    Gamut_input_data_error* gamut_input_data_error_ptr,
    double                  set_value 
)
{

    gamut_input_data_error_ptr->absolute_pixel_error = set_value;
    gamut_input_data_error_ptr->relative_pixel_error = set_value;
    gamut_input_data_error_ptr->diag_error           = set_value;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

