
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin 
|  contributed the interface with the correlation code). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "lib.h"
#include "version.h"

/* -------------------------------------------------------------------------- */

#define DATA_DIR "data"

#define ALWAYS_DO_STANDARD

/* -------------------------------------------------------------------------- */

static char fs_gamut_config_dir[ MAX_FILE_NAME_SIZE ] = { '\0' }; 
static char fs_default_cc_options[ 10000 ] = { '\0' };  

/* -------------------------------------------------------------------------- */

static void initialize_result_by_algorithm(Result_by_algorithm*);

static void initialize_best_fit_result(Best_fit_result* result_ptr);

static void initialize_standard_result(Standard_result* result_ptr);

static void initialize_gamut_result(Gamut_result* result_ptr);

/* Lindsay: Colour by Correlation - Feb 17/98 */
static void initialize_correlation_result(Correlation_result* result_ptr);
/* End Lindsay */

#ifdef SIMPLE_TEMPLATE
static void initialize_simple_template_result
(
    Simple_template_result* result_ptr 
);
#endif 

#ifdef TEMPLATE
static void initialize_template_result(Template_result* result_ptr);
#endif 

#ifdef MULTI_TEMPLATE
static void initialize_multi_template_result
(
    Multi_template_result* result_ptr 
);
#endif 

static void clear_result_by_algorithm_type
(
    Result_by_algorithm* result_ptr 
);

static void clear_best_fit_result(Best_fit_result* best_fit_ptr);

static void clear_standard_result(Standard_result* result_ptr);

static void clear_gamut_result(Gamut_result* result_ptr);

/* Lindsay: Feb 17/98 */
static void clear_correlation_result(Correlation_result* result_ptr);
/* End Lindsay */

#ifdef SIMPLE_TEMPLATE
static void clear_simple_template_result
(
    Simple_template_result* result_ptr 
);
#endif 

#ifdef TEMPLATE
static void clear_template_result(Template_result* result_ptr);
#endif 

#ifdef MULTI_TEMPLATE
static void clear_multi_template_result
(
    Multi_template_result* result_ptr 
);
#endif 

static int copy_map_vector
(
    Vector** target_vpp,
    Vector*  map_vp,
    char*    tag 
);

static int copy_map_matrix
(
    Matrix** target_mpp,
    Matrix*  map_mp,
    char*    tag 
);

static void set_default_gamut_config_dir(void);

/* -------------------------------------------------------------------------- */

int set_gamut_config_dir(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "config-dir"))
         || (match_pattern(lc_option, "gamut-config-dir"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_gamut_config_dir[ 0 ] == '\0')
            {
                set_default_gamut_config_dir();
            }

            if (fs_gamut_config_dir[ 0 ] == '\0') 
            {
                ERE(pso("No configuration directory is available.\n"));
            }
            else 
            {
                ERE(pso("Configuration directory is %s.\n", fs_gamut_config_dir));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_gamut_config_dir[ 0 ] = '\0';
            }
            else 
            {
                BUFF_CPY(fs_gamut_config_dir, value); 
            }
        }
        result = NO_ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_algorithm_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "cc-options"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Default cc-options are %q.\n", fs_default_cc_options));
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_default_cc_options[ 0 ] = '\0';
            }
            else 
            {
                BUFF_CPY(fs_default_cc_options, value);
            }
        }
        result = NO_ERROR; 
    }
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_cc_methods(char* arg, Colour_constancy_methods* cc_methods_ptr)
{
    static const char *cc_method_key_words[ ] =
    {
        "sm", "hi", 
        "2d", "3d", 
        "luminance", "luminance-factors", 
        "dull",
        "fixed-sharp-correction",
        "iterative-sharp-correction", 
        "synthetic-db-sharp-correction", 
        "synthetic-mip-db-sharp-correction", 
        "world-db-sharp-correction", 
        "scene-db-sharp-correction", 
        "fixed-sharp", 
        "iterative-sharp", 
        "synthetic-db-sharp", 
        "synthetic-mip-db-sharp", 
        "scene-db-sharp",
        "world-db-sharp", 
        "standard", "std", 
        "gamut", 
        "correlation",    /* Lindsay - Sept 30/98 */
        "corr",           /* Kobus - July 23/98 */
#ifdef SIMPLE_TEMPLATE
        "simple-template", 
#endif 
#ifdef TEMPLATE
        "template", 
#endif 
        "reset-specular", "dont-reset-specular", 
#ifdef MULTI_TEMPLATE
        "multi-template", 
#endif 
        "reset-specular", "dont-reset-specular", 
        "skip-on-error", "continue-on-error", 
        "print-algorithm-error", "dont-print-algorithm-error", 
        "print-error", "dont-print-error", 
        "plot", "no-plot", "report", "no-report", "debug"
    } ;

    int  key_words_found_list[ sizeof(cc_method_key_words)/sizeof(char*) ];
    int  num_key_words = sizeof(cc_method_key_words)/sizeof(char*);
    int  i, num_found_key_words;
    char lc_key_word[ 100 ];
    int  pass; 
    int  no_sharp_method_set; 

    cc_methods_ptr->do_luminance         = FALSE;
    cc_methods_ptr->do_luminance_factors = FALSE;
    cc_methods_ptr->do_2D                = FALSE;
    cc_methods_ptr->do_3D                = FALSE;
    cc_methods_ptr->do_standard          = FALSE;
    cc_methods_ptr->do_gamut             = FALSE;
    cc_methods_ptr->do_correlation       = FALSE;     /* Lindsay: Feb 17/98 */
#ifdef TEMPLATE
    cc_methods_ptr->do_template          = FALSE; 
#endif 
    cc_methods_ptr->reset_specular       = FALSE; 
#ifdef SIMPLE_TEMPLATE
    cc_methods_ptr->do_simple_template   = FALSE; 
#endif 
#ifdef MULTI_TEMPLATE
    cc_methods_ptr->do_multi_template    = FALSE; 
#endif 
    cc_methods_ptr->reset_specular       = FALSE; 
    cc_methods_ptr->hull_intersection    = TRUE;

    for (i=0; i<NUM_SHARP_METHODS; i++) 
    {
        cc_methods_ptr->sharp[ i ] = FALSE;
    }

    cc_methods_ptr->print_algorithm_error = TRUE;
    cc_methods_ptr->skip_on_error = FALSE;
    cc_methods_ptr->debug         = FALSE;
    /*
    // The following options have different defaults, depending on the command,
    // so we don't touch them.
    */
    cc_methods_ptr->plot_enable   = NOT_SET;
    cc_methods_ptr->report_result = NOT_SET;

    /*
    // Get options from the default string, and then from the input string. We
    // have to be careful here, as the routine ic_parse_key_words overrwrites
    // its input (this is a feature). Thus we can't do something as simple as
    // just concating the two strings (we want the overwritten result as it
    // contains the point specifications).
    */
    for (pass = 1; pass <= 2; pass++)
    {
        if (pass == 1)
        {
            char cc_options[ sizeof(fs_default_cc_options) ];

            BUFF_CPY(cc_options, fs_default_cc_options);

            ERE(num_found_key_words = ic_parse_key_words(cc_options,
                                                         num_key_words,
                                                         cc_method_key_words,
                                                         key_words_found_list,
                                                         (int*)NULL));
        }
        else 
        {
            ERE(num_found_key_words = ic_parse_key_words(arg, num_key_words,
                                                         cc_method_key_words,
                                                         key_words_found_list,
                                                         (int*)NULL));
        }

        /*
        // Extra work gets done here, since the comparison is essentially
        // already determined by [key_words_found_list[i]]. However, this is
        // resiliant to changes in cc_method_key_words.
        */

        for (i=0; i<num_found_key_words; i++)
        {
            EXTENDED_LC_BUFF_CPY(lc_key_word, 
                                 cc_method_key_words[key_words_found_list[i]]);

            if (STRCMP_EQ(lc_key_word, "hi"))
            {
                cc_methods_ptr->hull_intersection = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "luminance"))
            {
                cc_methods_ptr->do_luminance = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "luminance-factors"))
            {
                cc_methods_ptr->do_luminance_factors = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "2d"))
            {
                cc_methods_ptr->do_2D = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "3d"))
            {
                cc_methods_ptr->do_3D = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "gamut"))
            {
                cc_methods_ptr->do_gamut = TRUE;
            }
            else if (    (STRCMP_EQ(lc_key_word, "std"))
                      || (STRCMP_EQ(lc_key_word, "standard"))
                    )
            {
                cc_methods_ptr->do_standard = TRUE;
            }
            /* Lindsay: Sept 30/98  */
            else if (   (STRCMP_EQ(lc_key_word, "correlation"))
                     || (STRCMP_EQ(lc_key_word, "corr")) /* Kobus: July 23/99 */
                    )
            {
                cc_methods_ptr->do_correlation = TRUE;
            }
            /* End Lindsay: Sept 30/98 */
#ifdef SIMPLE_TEMPLATE
            else if (STRCMP_EQ(lc_key_word, "simple-template")) 
            {
                cc_methods_ptr->do_simple_template = TRUE;
            }
#endif 
#ifdef TEMPLATE
            else if (STRCMP_EQ(lc_key_word, "template"))
            {
                cc_methods_ptr->do_template = TRUE;
            }
#endif 
#ifdef MULTI_TEMPLATE
            else if (STRCMP_EQ(lc_key_word, "multi-template"))
            {
                cc_methods_ptr->do_multi_template = TRUE;
            }
#endif 
            else if (STRCMP_EQ(lc_key_word, "reset-specular"))
            {
                cc_methods_ptr->reset_specular = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "dont-reset-specular"))
            {
                cc_methods_ptr->reset_specular = FALSE;
            }
            else if (STRCMP_EQ(lc_key_word, "dull"))
            {
                cc_methods_ptr->sharp[ DONT_SHARPEN ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "fixed-sharp-correction"))
            {
                cc_methods_ptr->sharp[ FIXED_SHARP_CORRECTION ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "iterative-sharp-correction"))
            {
                cc_methods_ptr->sharp[ ITERATIVE_SHARP_CORRECTION ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "synthetic-db-sharp-correction"))
            {
                cc_methods_ptr->sharp[ SYNTHETIC_DB_SHARP_CORRECTION ] = TRUE; 
            }
            else if (STRCMP_EQ(lc_key_word, 
                               "synthetic-mip-db-sharp-correction")
                    )
            {
                cc_methods_ptr->sharp[ SYNTHETIC_MIP_DB_SHARP_CORRECTION ] = TRUE; 
            }
            else if (STRCMP_EQ(lc_key_word, "scene-db-sharp-correction"))
            {
                cc_methods_ptr->sharp[ SCENE_DB_SHARP_CORRECTION ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "world-db-sharp-correction"))
            {
                cc_methods_ptr->sharp[ WORLD_DB_SHARP_CORRECTION ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "fixed-sharp"))
            {
                cc_methods_ptr->sharp[ FIXED_SHARP ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "synthetic-db-sharp"))
            {
                cc_methods_ptr->sharp[ SYNTHETIC_DB_SHARP ] = TRUE; 
            }
            else if (STRCMP_EQ(lc_key_word, "synthetic-mip-db-sharp"))
            {
                cc_methods_ptr->sharp[ SYNTHETIC_MIP_DB_SHARP ] = TRUE; 
            }
            else if (STRCMP_EQ(lc_key_word, "scene-db-sharp"))
            {
                cc_methods_ptr->sharp[ SCENE_DB_SHARP ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "world-db-sharp"))
            {
                cc_methods_ptr->sharp[ WORLD_DB_SHARP ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "iterative-sharp"))
            {
                cc_methods_ptr->sharp[ ITERATIVE_SHARP ] = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "plot"))
            {
                cc_methods_ptr->plot_enable = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "no-plot"))
            {
                cc_methods_ptr->plot_enable = FALSE;
            }
            else if (STRCMP_EQ(lc_key_word, "report"))
            {
                cc_methods_ptr->report_result = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "no-report"))
            {
                cc_methods_ptr->report_result = FALSE;
            }
            else if (STRCMP_EQ(lc_key_word, "skip-on-error"))
            {
                cc_methods_ptr->skip_on_error = TRUE;
            }
            else if (STRCMP_EQ(lc_key_word, "continue-on-error"))
            {
                cc_methods_ptr->skip_on_error = FALSE;
            }
            else if (    (STRCMP_EQ(lc_key_word, "print-algorithm-error"))
                      || (STRCMP_EQ(lc_key_word, "print-error"))
                    )
            {
                cc_methods_ptr->print_algorithm_error = TRUE;
            }
            else if (    (STRCMP_EQ(lc_key_word, "dont-print-algorithm-error"))
                      || (STRCMP_EQ(lc_key_word, "dont-print-error"))
                    )
            {
                cc_methods_ptr->print_algorithm_error = FALSE;
            }
            else if (STRCMP_EQ(lc_key_word, "debug"))
            {
                cc_methods_ptr->debug = TRUE;
            }
        }
    }

    if (    (cc_methods_ptr->do_standard       == FALSE)
         && (cc_methods_ptr->do_gamut          == FALSE)
         && (cc_methods_ptr->do_correlation    == FALSE)/* Lindsay: Feb 17/98*/
#ifdef SIMPLE_TEMPLATE
         && (cc_methods_ptr->do_simple_template == FALSE)
#endif 
#ifdef TEMPLATE
         && (cc_methods_ptr->do_template       == FALSE)
#endif 
#ifdef MULTI_TEMPLATE
         && (cc_methods_ptr->do_multi_template == FALSE)
#endif 
         && (cc_methods_ptr->do_luminance_factors == FALSE)
         && (cc_methods_ptr->do_luminance         == FALSE)
       )
    {
        /* Defaults */

        cc_methods_ptr->do_standard       = TRUE;
        cc_methods_ptr->do_gamut          = TRUE;
        cc_methods_ptr->do_correlation    = TRUE; /* Lindsay: Feb 17/98 */
#ifdef SIMPLE_TEMPLATE
        /*
        // Set to TRUE unless algorithm group is slow or uninteresting. 
        */
        cc_methods_ptr->do_simple_template = TRUE;
#endif 
#ifdef TEMPLATE
        /*
        // Set to TRUE unless algorithm group is slow or uninteresting. 
        */
        cc_methods_ptr->do_template = TRUE;
#endif 
#ifdef MULTI_TEMPLATE
        /*
        // TSet to RUE unless algorithm group is slow or uninteresting. 
        */
        cc_methods_ptr->do_multi_template = TRUE;
#endif 
    }
    else if ((cc_methods_ptr->do_2D) || (cc_methods_ptr->do_3D))
    {
         cc_methods_ptr->do_gamut = TRUE;
    }

    if (    (cc_methods_ptr->do_luminance_factors == FALSE)
         && (cc_methods_ptr->do_luminance         == FALSE)
         && (cc_methods_ptr->do_2D                == FALSE)
         && (cc_methods_ptr->do_3D                == FALSE)
       )
    {
        cc_methods_ptr->do_2D = TRUE;
    }

#ifdef ALWAYS_DO_STANDARD
    /*
    // For now, always do standard. We play the song and dance above to make
    // this easy to change to the same symantics as the rest of the algorithms
    // simply by commenting out the following line. 
    */
    cc_methods_ptr->do_standard = TRUE; 
#endif 

    no_sharp_method_set = TRUE; 

    for (i=0; i<NUM_SHARP_METHODS; i++)
    {
        if (cc_methods_ptr->sharp[ i ]) no_sharp_method_set = FALSE;
    }

    if (no_sharp_method_set) 
    {
        /* Default case (without sharpenning only) */
        cc_methods_ptr->sharp[ DONT_SHARPEN ] = TRUE;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_constraints(Constraints* constraints_ptr)
{
    int i;

    constraints_ptr->illum_hull_list_head          = NULL;
    constraints_ptr->illum_hp                      = NULL;
    constraints_ptr->surface_hull_list_head        = NULL;
    
    constraints_ptr->canon_illum_vp                = NULL;

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        constraints_ptr->sum_luminance[ i ]        = DBL_NOT_SET;
    }
    
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_constraints(Constraints* constraints_ptr)
{

    if (constraints_ptr->illum_hull_list_head != NULL)
    {
        free_hull_list(&(constraints_ptr->illum_hull_list_head));
    }

    free_hull(constraints_ptr->illum_hp); 

    if (constraints_ptr->surface_hull_list_head != NULL)
    {
        free_hull_list(&(constraints_ptr->surface_hull_list_head));
    }

    free_vector(constraints_ptr->canon_illum_vp); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_hull_list(Queue_element** hull_list_head_ptr)
{
    Queue_element *save_cur_elem, *cur_elem;

    if (*hull_list_head_ptr != NULL)
    {
        cur_elem = *hull_list_head_ptr;

        while (cur_elem != NULL)
        {
            free_hull((Hull*)(cur_elem->contents));
            save_cur_elem = cur_elem;
            cur_elem = cur_elem->next;
            kjb_free(save_cur_elem);
        }

        *hull_list_head_ptr = NULL;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_cc_result(Colour_constancy_result* result_ptr)
{
    int i;

    for (i=0; i<NUM_SHARP_METHODS; i++) 
    {
        initialize_result_by_algorithm(&(result_ptr->sharp[ i ]));
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_result_by_algorithm
(
    Result_by_algorithm* result_ptr 
)
{
    int i;

    initialize_best_fit_result(&(result_ptr->best_fit));

    initialize_standard_result(&(result_ptr->standard));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_result(&(result_ptr->gamut_2D[ i ]));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        initialize_gamut_result(&(result_ptr->gamut_3D[ i ]));
    }

    initialize_correlation_result(&(result_ptr->correlation));  /* Lindsay */

#ifdef SIMPLE_TEMPLATE
    initialize_simple_template_result(&(result_ptr->simple_template));
#endif 

#ifdef TEMPLATE
    initialize_template_result(&(result_ptr->template));
#endif 

#ifdef MULTI_TEMPLATE
    initialize_multi_template_result(&(result_ptr->multi_template));
#endif 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_best_fit_result(Best_fit_result* result_ptr)
{

    result_ptr->world_best_2D_fit_mp      = NULL;
    result_ptr->world_best_3D_fit_mp      = NULL;
    result_ptr->world_best_3D_diag_fit_vp = NULL;
    result_ptr->world_best_2D_diag_fit_vp = NULL;
    result_ptr->scene_best_2D_fit_mp      = NULL;
    result_ptr->scene_best_3D_fit_mp      = NULL;
    result_ptr->scene_best_3D_diag_fit_vp = NULL;
    result_ptr->scene_best_2D_diag_fit_vp = NULL;
    result_ptr->actual_vp                 = NULL;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_standard_result(Standard_result* result_ptr)
{

    result_ptr->ave_illum_vp          = NULL;
    result_ptr->grey_world_vp         = NULL;
    result_ptr->db_grey_vp            = NULL;
    result_ptr->scale_by_max_vp            = NULL;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void initialize_gamut_result(Gamut_result* result_ptr)
{
    int i;

    result_ptr->mv_surface_vp                      = NULL;
    result_ptr->mv_surface_illum_vp                = NULL;
    result_ptr->ha_surface_vp                      = NULL;
    result_ptr->ha_surface_illum_vp                = NULL;
    result_ptr->icsa_surface_illum_vp              = NULL;
    result_ptr->wa_surface_vp                      = NULL;
    result_ptr->wa_surface_illum_vp                = NULL;

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        result_ptr->scia_surface_illum_vp[ i ] = NULL;
        result_ptr->scca_surface_illum_vp[ i ] = NULL;
    }

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        result_ptr->ha_surface_lum_vp[         i ] = NULL;
        result_ptr->mv_surface_lum_vp[         i ] = NULL;
        result_ptr->ha_surface_illum_lum_vp[   i ] = NULL;
        result_ptr->mv_surface_illum_lum_vp[   i ] = NULL;
        result_ptr->scla_surface_illum_lum_vp[ i ] = NULL;
        result_ptr->lcsa_surface_illum_lum_vp[ i ] = NULL;
        result_ptr->scia_surface_illum_lum_vp[ i ] = NULL;
        result_ptr->scca_surface_illum_lum_vp[ i ] = NULL;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay: Sept 30/98 */
static void initialize_correlation_result(Correlation_result* result_ptr)
{
    int i; /* Modified April 14/99 */

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        result_ptr->correlation_01_vp[                        i ] = NULL;
        result_ptr->correlation_gaussian_MAP_vp[              i ] = NULL;
        result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MAP_fitted_vp[        i ] = NULL;
        result_ptr->correlation_gaussian_MLM_vp[              i ] = NULL;
        result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MLM_fitted_vp[        i ] = NULL;
        result_ptr->correlation_gaussian_MMSE_vp[              i ] = NULL;
        result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ] = NULL;
        result_ptr->correlation_uniform_MMSE_fitted_vp[        i ] = NULL;
    }
}
/* End Lindsay */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static void initialize_simple_template_result
(
    Simple_template_result* result_ptr 
)
{

    result_ptr->alg_vp = NULL;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static void initialize_template_result(Template_result* result_ptr)
{

    result_ptr->first_alg_vp  = NULL;
    result_ptr->second_alg_vp = NULL;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE

static void initialize_multi_template_result
(
    Multi_template_result* result_ptr 
)
{
    int i; 

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        result_ptr->alg_vp[ i ]  = NULL;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_hull_result(Hull_result* result_ptr)
{
    int j; 

    result_ptr->surface_hp = NULL;
    result_ptr->surface_illum_hp = NULL;
    result_ptr->illum_constrained_surface_hp = NULL;
    
    for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
    {
        result_ptr->luminance_surface_hp[j ] = NULL;
        result_ptr->luminance_surface_illum_hp[ j ] = NULL;
        result_ptr->luminance_constrained_surface_hp[ j ] = NULL;
    }
    
    result_ptr->surface_illum_var_hp = NULL;
    result_ptr->scia_points_mp = NULL;
    result_ptr->scca_points_mp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void clear_result(Colour_constancy_result* result_ptr)
{
    int i;

    if (result_ptr == NULL) return;

    for (i=0; i<NUM_SHARP_METHODS; i++) 
    {
        clear_result_by_algorithm_type(&(result_ptr->sharp[ i ]));
    }

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void clear_result_by_algorithm_type
(
    Result_by_algorithm* result_ptr 
)
{
    int i; 

    clear_best_fit_result(&(result_ptr->best_fit));

    clear_standard_result(&(result_ptr->standard));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        clear_gamut_result(&(result_ptr->gamut_2D[ i ]));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        clear_gamut_result(&(result_ptr->gamut_3D[ i ]));
    }

    clear_correlation_result(&(result_ptr->correlation));       /* Lindsay */

#ifdef SIMPLE_TEMPLATE
    clear_simple_template_result(&(result_ptr->simple_template));
#endif 

#ifdef TEMPLATE
    clear_template_result(&(result_ptr->template));
#endif 

#ifdef MULTI_TEMPLATE
    clear_multi_template_result(&(result_ptr->multi_template));
#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void clear_best_fit_result(Best_fit_result* best_fit_ptr)
{

    free_matrix(best_fit_ptr->world_best_3D_fit_mp);
    best_fit_ptr->world_best_3D_fit_mp = NULL;

    free_matrix(best_fit_ptr->world_best_2D_fit_mp);
    best_fit_ptr->world_best_2D_fit_mp = NULL;

    free_vector(best_fit_ptr->world_best_3D_diag_fit_vp);
    best_fit_ptr->world_best_3D_diag_fit_vp = NULL;

    free_vector(best_fit_ptr->world_best_2D_diag_fit_vp);
    best_fit_ptr->world_best_2D_diag_fit_vp = NULL;

    free_matrix(best_fit_ptr->scene_best_3D_fit_mp);
    best_fit_ptr->scene_best_3D_fit_mp = NULL;

    free_matrix(best_fit_ptr->scene_best_2D_fit_mp);
    best_fit_ptr->scene_best_2D_fit_mp = NULL;

    free_vector(best_fit_ptr->scene_best_3D_diag_fit_vp);
    best_fit_ptr->scene_best_3D_diag_fit_vp = NULL;

    free_vector(best_fit_ptr->scene_best_2D_diag_fit_vp);
    best_fit_ptr->scene_best_2D_diag_fit_vp = NULL;

    free_vector(best_fit_ptr->actual_vp);
    best_fit_ptr->actual_vp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void clear_standard_result(Standard_result* result_ptr)
{

    if (result_ptr == NULL) return; 

    free_vector(result_ptr->ave_illum_vp);
    result_ptr->ave_illum_vp = NULL;

    free_vector(result_ptr->grey_world_vp);
    result_ptr->grey_world_vp = NULL;

    free_vector(result_ptr->db_grey_vp);
    result_ptr->db_grey_vp = NULL;

    free_vector(result_ptr->scale_by_max_vp);
    result_ptr->scale_by_max_vp = NULL;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void clear_gamut_result(Gamut_result* result_ptr)
{
    int i;

    if (result_ptr == NULL) return; 

    free_vector(result_ptr->mv_surface_vp);
    result_ptr->mv_surface_vp = NULL;

    free_vector(result_ptr->mv_surface_illum_vp);
    result_ptr->mv_surface_illum_vp = NULL;

    free_vector(result_ptr->ha_surface_vp);
    result_ptr->ha_surface_vp = NULL;

    free_vector(result_ptr->ha_surface_illum_vp);
    result_ptr->ha_surface_illum_vp = NULL;

    free_vector(result_ptr->icsa_surface_illum_vp);
    result_ptr->icsa_surface_illum_vp = NULL;

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        free_vector(result_ptr->scia_surface_illum_vp[ i ]);
        result_ptr->scia_surface_illum_vp[ i ] = NULL;

        free_vector(result_ptr->scca_surface_illum_vp[ i ]);
        result_ptr->scca_surface_illum_vp[ i ] = NULL;
    }

    free_vector(result_ptr->wa_surface_vp);
    result_ptr->wa_surface_vp = NULL;

    free_vector(result_ptr->wa_surface_illum_vp);
    result_ptr->wa_surface_illum_vp = NULL;

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        free_vector(result_ptr->ha_surface_lum_vp[ i ]);
        result_ptr->ha_surface_lum_vp[ i ] = NULL;

        free_vector(result_ptr->mv_surface_lum_vp[ i ]);
        result_ptr->mv_surface_lum_vp[ i ] = NULL;

        free_vector(result_ptr->ha_surface_illum_lum_vp[ i ]);
        result_ptr->ha_surface_illum_lum_vp[ i ] = NULL;

        free_vector(result_ptr->mv_surface_illum_lum_vp[ i ]);
        result_ptr->mv_surface_illum_lum_vp[ i ] = NULL;

        free_vector(result_ptr->scla_surface_illum_lum_vp[ i ]);
        result_ptr->scla_surface_illum_lum_vp[ i ] = NULL;

        free_vector(result_ptr->lcsa_surface_illum_lum_vp[ i ]);
        result_ptr->lcsa_surface_illum_lum_vp[ i ] = NULL;

        free_vector(result_ptr->scia_surface_illum_lum_vp[ i ]);
        result_ptr->scia_surface_illum_lum_vp[ i ] = NULL;

        free_vector(result_ptr->scca_surface_illum_lum_vp[ i ]);
        result_ptr->scca_surface_illum_lum_vp[ i ] = NULL;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void clear_correlation_result(Correlation_result* result_ptr)
{
    int i; /* Modified Apr 14/99 */

    if (result_ptr == NULL) return; 

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_01_vp[ i ]);
        result_ptr->correlation_01_vp[ i ] = NULL;
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_gaussian_MAP_vp[ i ]);
        result_ptr->correlation_gaussian_MAP_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ]);
        result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ]);
        result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MAP_fitted_vp[ i ]);
        result_ptr->correlation_uniform_MAP_fitted_vp[ i ] = NULL;
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_gaussian_MLM_vp[ i ]);
        result_ptr->correlation_gaussian_MLM_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ]);
        result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ]);
        result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MLM_fitted_vp[ i ]);
        result_ptr->correlation_uniform_MLM_fitted_vp[ i ] = NULL;
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_gaussian_MMSE_vp[ i ]);
        result_ptr->correlation_gaussian_MMSE_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ]);
        result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ]);
        result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ] = NULL;
    }
    
    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {    
        free_vector(result_ptr->correlation_uniform_MMSE_fitted_vp[ i ]);
        result_ptr->correlation_uniform_MMSE_fitted_vp[ i ] = NULL;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE
static void clear_simple_template_result(Simple_template_result* result_ptr)
{

    if (result_ptr == NULL) return; 

    free_vector(result_ptr->alg_vp);
    result_ptr->alg_vp = NULL;
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE
static void clear_template_result(Template_result* result_ptr)
{

    if (result_ptr == NULL) return; 

    free_vector(result_ptr->first_alg_vp);
    result_ptr->first_alg_vp = NULL;

    free_vector(result_ptr->second_alg_vp);
    result_ptr->second_alg_vp = NULL;

}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE
static void clear_multi_template_result
(
    Multi_template_result* result_ptr 
)
{
    int i; 

    if (result_ptr == NULL) return; 

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        free_vector(result_ptr->alg_vp[ i ]);
        result_ptr->alg_vp[ i ] = NULL;
    }
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void clear_hull_result(Hull_result* result_ptr)
{
    int j;

    if (result_ptr == NULL) return; 

    free_hull(result_ptr->surface_hp);
    result_ptr->surface_hp = NULL;

    free_hull(result_ptr->surface_illum_hp);
    result_ptr->surface_illum_hp =  NULL;

    free_hull(result_ptr->illum_constrained_surface_hp);
    result_ptr->illum_constrained_surface_hp =  NULL;

    for (j=0; j<MAX_NUM_LUMINANCE_METHODS; j++)
    {
        free_hull(result_ptr->luminance_constrained_surface_hp[ j ]);
        result_ptr->luminance_constrained_surface_hp[ j ] = NULL;

        free_hull(result_ptr->luminance_surface_hp[ j ]);
        result_ptr->luminance_surface_hp[ j ] = NULL;

        free_hull(result_ptr->luminance_surface_illum_hp[ j ]);
        result_ptr->luminance_surface_illum_hp[ j ] =  NULL;

    }

    free_hull(result_ptr->surface_illum_var_hp);
    result_ptr->surface_illum_var_hp = NULL;

    free_matrix(result_ptr->scia_points_mp);
    result_ptr->scia_points_mp = NULL;

    free_matrix(result_ptr->scca_points_mp);
    result_ptr->scca_points_mp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ok_to_increase_pixel_error
(
    int     prompt_to_increase_error_flag,
    double* absolute_pixel_error_ptr,
    double* relative_pixel_error_ptr 
)
{
    int    relative_error_is_being_used = FALSE;
    int    absolute_error_is_being_used = FALSE;
    char   prompt[ 200 ];
    int    int_pixel_error;
    double absolute_pixel_error = 0.0;
    double relative_pixel_error = 0.0;
    double old_absolute_pixel_error = DBL_NOT_SET;
    double old_relative_pixel_error = DBL_NOT_SET;
    int    increase_error_flag;

    if (*relative_pixel_error_ptr >= 0.0)
    {
        relative_error_is_being_used = TRUE;
        old_relative_pixel_error = 100.0 * (*relative_pixel_error_ptr);
    }
    
    if (*absolute_pixel_error_ptr >= 0.0)
    {
        absolute_error_is_being_used = TRUE;
        old_absolute_pixel_error = *absolute_pixel_error_ptr;
    }

    if (    ( ! relative_error_is_being_used )
         && ( ! absolute_error_is_being_used )
       )
    {
        return FALSE;
    }
     
    if (prompt_to_increase_error_flag)
    {
        if (*relative_pixel_error_ptr >= 0.0)
        {
            relative_pixel_error = 1.5 * old_relative_pixel_error;
            relative_pixel_error += EXTRA_FOR_ROUNDING;
            int_pixel_error = (int)(MAX_OF(relative_pixel_error, 1.0));
            int_pixel_error = ((int_pixel_error - 1) / 5 + 1) * 5;
            relative_pixel_error = int_pixel_error;
        }
        
        if (*absolute_pixel_error_ptr >= 0.0)
        {
            absolute_pixel_error = 1.5 * old_absolute_pixel_error;
            absolute_pixel_error += EXTRA_FOR_ROUNDING;
            int_pixel_error = (int)(MAX_OF(absolute_pixel_error, 1.0));
            int_pixel_error = ((int_pixel_error - 1) / 5 + 1) * 5;
            absolute_pixel_error = int_pixel_error;
        }

        if (    (relative_error_is_being_used)
             && (absolute_error_is_being_used)
           )
        {
            kjb_sprintf(prompt, sizeof(prompt),
                        "%s %.1f%% %s %.1f. Try %.1f%% and %.1f? (y/n) ",
                        "No solution with rpe ", old_relative_pixel_error, 
                        "and ape ", old_absolute_pixel_error, 
                        relative_pixel_error, absolute_pixel_error);
        }
        else if (relative_error_is_being_used)
        {
            kjb_sprintf(prompt, sizeof(prompt),
                        "%s %.1f%%. Try %.1f%%? (y/n) ",
                        "No solution with relative pixel error", 
                        old_relative_pixel_error, relative_pixel_error);
        }
        else if (absolute_error_is_being_used)
        {
            kjb_sprintf(prompt, sizeof(prompt),
                        "%s %.1f. Try %.1f? (y/n) ",
                        "No solution with absolute pixel error", 
                        old_absolute_pixel_error, absolute_pixel_error);
        }
        else 
        {
            SET_CANT_HAPPEN_BUG();
        }

        increase_error_flag = yes_no_query(prompt);
    }
    else
    {
        /*
        // If we are not prompting to increase the error, then
        // do it more conservatively to get a more accurate answer.
        */
        if (*relative_pixel_error_ptr >= 0.0)
        {
            relative_pixel_error = 1.0 +  old_relative_pixel_error;

            verbose_pso(8, "Increasing relative pixel error to %.1f%%\n",
                        relative_pixel_error); 
        }
        
        if (*absolute_pixel_error_ptr >= 0.0)
        {
            absolute_pixel_error = 1.0 + old_absolute_pixel_error;

            verbose_pso(8, "Increasing absolute pixel error to %.1f\n",
                        absolute_pixel_error); 
        }

        increase_error_flag = TRUE;
    }

    if (increase_error_flag)
    {
        if (relative_error_is_being_used)
        {
            *relative_pixel_error_ptr = relative_pixel_error / 100.0;
        }

        if (absolute_error_is_being_used)
        {
            *absolute_pixel_error_ptr = absolute_pixel_error;
        }
    }
    else
    {
        *relative_pixel_error_ptr = DBL_HALF_MOST_NEGATIVE;
        *absolute_pixel_error_ptr = DBL_HALF_MOST_NEGATIVE;
    }

    return increase_error_flag;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       read_real_from_gamut_config_file
 *                               
 * Reads a number from config file
 *
 * This routime essentially calls read_real_from_config_file() with the "dir"
 * parameter set to the gamut program configuration directory.  The default
 * configuration directory is currently a function of the version. It can be
 * changed by the user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * If the buffer config_file_name is not NULL, then the file name actually used
 * is copied into it. If it is used, then its size must be passed in via
 * config_file_name_size; 
 *
 * If file_name is the special name "off" (or "none" or "0"), then the above
 * does not apply. In this case, *value_ptr is set to DBL_NOT_SET, which is
 * guarenteed to be negative. 
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     read_real_from_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

int read_real_from_gamut_config_file
(
    double*     value_ptr,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir();
    }

    return read_real_from_config_file(value_ptr, env_var, fs_gamut_config_dir, 
                                      file_name, message_name, 
                                      config_file_name, 
                                      config_file_name_size);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       read_vector_from_gamut_config_file
 *                               
 * Reads a vector from config file
 *
 * This routime essentially calls read_vector_from_config_file() with the "dir"
 * parameter set to the gamut program configuration directory.  The default
 * configuration directory is currently a function of the version. It can be
 * changed by the user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * If the buffer config_file_name is not NULL, then the file name actually used
 * is copied into it. If it is used, then its size must be passed in via
 * config_file_name_size; 
 *
 * If file_name is the special name "off" (or "none" or "0"), then the above
 * does not apply. In this case, *result_vpp is freed and set to NULL. 
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     read_vector_from_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

int read_vector_from_gamut_config_file
(
    Vector**    result_vpp,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir();
    }

    return read_vector_from_config_file(result_vpp, env_var, 
                                        fs_gamut_config_dir, file_name, 
                                        message_name, config_file_name,
                                        config_file_name_size);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       read_matrix_from_gamut_config_file
 *                               
 * Reads a matrix from config file
 *
 * This routime essentially calls read_matrix_from_config_file() with the "dir"
 * parameter set to the gamut program configuration directory.  The default
 * configuration directory is currently a function of the version. It can be
 * changed by the user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * If the buffer config_file_name is not NULL, then the file name actually used
 * is copied into it. If it is used, then its size must be passed in via
 * config_file_name_size; 
 *
 * If file_name is the special name "off" (or "none" or "0"), then the above
 * does not apply. In this case, *result_mpp is freed and set to NULL. 
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     read_matrix_from_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

int read_matrix_from_gamut_config_file
(
    Matrix**    result_mpp,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir();
    }

    return read_matrix_from_config_file(result_mpp, env_var, 
                                        fs_gamut_config_dir, file_name, 
                                        message_name, config_file_name,
                                        config_file_name_size);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       read_spectra_from_gamut_config_file
 *                               
 * Reads spectra from config file
 *
 * This routime essentially calls read_spectra_from_config_file() with the "dir"
 * parameter set to the gamut program configuration directory.  The default
 * configuration directory is currently a function of the version. It can be
 * changed by the user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * If the buffer config_file_name is not NULL, then the file name actually used
 * is copied into it. If it is used, then its size must be passed in via
 * config_file_name_size; 
 *
 * If file_name is the special name "off" (or "none" or "0"), then the above
 * does not apply. In this case, *result_sp_ptr is freed and set to NULL. 
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     read_spectra_from_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

int read_spectra_from_gamut_config_file
(
    Spectra**   result_sp_ptr,
    const char* env_var,
    const char* file_name,
    const char* message_name,
    char*       config_file_name,
    size_t      config_file_name_size 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir();
    }

    return read_spectra_from_config_file(result_sp_ptr, env_var, 
                                         fs_gamut_config_dir, file_name, 
                                         message_name, config_file_name, 
                                         config_file_name_size);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       read_fl_db_from_gamut_config_file
 *                               
 * Reads fluorescent database from config file
 *
 * This routime essentially calls read_fl_db_from_config_file() with the "dir"
 * parameter set to the gamut program configuration directory.  The default
 * configuration directory is currently a function of the version. It can be
 * changed by the user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * If the buffer config_file_name is not NULL, then the file name actually used
 * is copied into it. If it is used, then its size must be passed in via
 * config_file_name_size; 
 *
 * If file_name is the special name "off" (or "none" or "0"), then the above
 * does not apply. In this case, *result_sp_ptr is freed and set to NULL. 
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     read_fl_db_from_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

int read_fl_db_from_gamut_config_file
(
    Fluorescent_database** result_fl_db_ptr_ptr,
    const char*            env_var,
    const char*            file_name,
    const char*            message_name,
    char*                  config_file_name,
    size_t                 config_file_name_size 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir();
    }

    return read_fl_db_from_config_file(result_fl_db_ptr_ptr, env_var, 
                                       fs_gamut_config_dir, file_name, 
                                       message_name, config_file_name, 
                                       config_file_name_size);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         open_gamut_config_file                               
 *                               
 * Opens a config file for gamut data
 *
 * This routime essentially calls open_config_file() with the "dir" parameter
 * set to the gamut program configuration directory.  The default configuration
 * directory is currently a function of the version. It can be changed by the
 * user with the "config-dir" option. 
 *
 * If env_var is not NULL or the empty string, then that string is tried as a
 * file name before any of the other locations are tried. 
 *
 * The parameter message_name can be used to specify a name for the
 * configuration file to be used in error and verbose output messages. If
 * message_name is NULL, then "configuration" is used.
 *
 * Index: gamut library, gamut I/O, gamut configuration files
 *
 * See also:
 *     open_config_file
 *
 * -----------------------------------------------------------------------------
*/ 

FILE* open_gamut_config_file
(
    const char* env_var,
    const char* file_name,
    const char* message_name 
)
{

    if (fs_gamut_config_dir[ 0 ] == '\0')
    {
        set_default_gamut_config_dir(); 
    }

    return open_config_file(env_var,  fs_gamut_config_dir, file_name, 
                            message_name);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void set_default_gamut_config_dir(void)
{

    kjb_sprintf(fs_gamut_config_dir, sizeof(fs_gamut_config_dir),
                "%s%sgamut-%d.%d.%d", DATA_DIR, DIR_STR, 
                MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL);

    if (check_config_sub_dir(fs_gamut_config_dir) != ERROR) return; 

    /* Try again without patch level. */

    kjb_sprintf(fs_gamut_config_dir, sizeof(fs_gamut_config_dir),
                "%s%sgamut-%d.%d", DATA_DIR, DIR_STR, 
                MAJOR_VERSION, MINOR_VERSION);

    if (check_config_sub_dir(fs_gamut_config_dir) != ERROR) return; 

    /* Try again without minor version */

    kjb_sprintf(fs_gamut_config_dir, sizeof(fs_gamut_config_dir),
                "%s%sgamut-%d", DATA_DIR, DIR_STR, MAJOR_VERSION);

    if (check_config_sub_dir(fs_gamut_config_dir) != ERROR) return; 

    /* Try again without version */

    kjb_sprintf(fs_gamut_config_dir, sizeof(fs_gamut_config_dir), "%s%sgamut", 
                DATA_DIR, DIR_STR);

    if (check_config_sub_dir(fs_gamut_config_dir) != ERROR) return; 

    /* OK, lets give up.  */

    fs_gamut_config_dir[ 0 ] = '\0'; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int select_data
(
    Matrix**      selected_mpp,
    const Matrix* data_mp,
    int           num_data_points,
    const int*    data_point_nums,
    const int*    invalid_data 
)
{
    int result = select_matrix_data(selected_mpp, data_mp, num_data_points,
                                    data_point_nums, invalid_data);

    if (result == ERROR)
    {
        add_error("No valid data for colour constancy processing.");
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifndef __C2MAN__

/*
// This routine takes the user input in "request" and determines the
// corresponding map needed to calculate the results.  It also sorts out error
// structure members. This is a bit of a hack, but it keeps all the parsing in
// the same place. Either of these two services can be ignored by making the
// relavent pointers NULL. 
*/

#endif 

int parse_result_request
(
    char*                    request,
    Colour_constancy_result* result_ptr,
    Vector**                 map_vpp,
    Matrix**                 map_mpp,
    Sharp_method*            sharp_method_ptr,
    Error_by_sharp*          error_ptr,
    double**                 error_item_ptr_ptr 
)
{
    char                   lc_request[ 100 ];
    char                   base_request[ 100 ];
    size_t                 request_len;
    int                    dimension                = NOT_SET;
    Result_by_algorithm*   result_alg_ptr = NULL;
    Best_fit_result*       best_fit_alg_ptr = NULL;
    Standard_result*       standard_alg_ptr = NULL;
    Gamut_result*          gamut_alg_ptr            = NULL;
    Correlation_result*    correlation_alg_ptr = NULL; /* Lindsay Feb 17/98 */
#ifdef SIMPLE_TEMPLATE
    Simple_template_result* simple_template_alg_ptr;
#endif
#ifdef TEMPLATE
    Template_result*       template_alg_ptr;
#endif
#ifdef MULTI_TEMPLATE
    Multi_template_result* multi_template_alg_ptr;
#endif
    Error_by_algorithm*    error_alg_ptr            = NULL;
    Best_fit_error*        best_fit_error_ptr       = NULL;
    Standard_error*        standard_error_ptr       = NULL;
    Gamut_error*           gamut_error_ptr          = NULL;
    Correlation_error*     correlation_error_ptr    = NULL; /* Lindsay */
#ifdef SIMPLE_TEMPLATE
    Simple_template_error*  simple_template_error_ptr       = NULL;
#endif 
#ifdef TEMPLATE
    Template_error*        template_error_ptr       = NULL;
#endif 
#ifdef MULTI_TEMPLATE
    Multi_template_error*  multi_template_error_ptr       = NULL;
#endif 
    int index, sub_index; 
    Sharp_method sharp_method = DONT_SHARPEN; 

    EXTENDED_LC_BUFF_CPY(lc_request, request);

    request_len = strlen(lc_request); 

    /*
    // FIX
    //
    // Parse some user method of designating sharp ( using # perhaps ??? )
    //
    if (    (request_len >3)
         && (lc_request[ request_len - 3 ] == '-')
         && (lc_request[ request_len - 2 ] == 's')
       )
    {
        *sharp_method_ptr = TRUE; 
        lc_request[ request_len - 3 ] = '\0';
        request_len -= 2;
    }
    else 
    {
        *sharp_method_ptr = FALSE; 
    }
    */
    if (sharp_method_ptr != NULL)
    {
        *sharp_method_ptr = sharp_method; 
    }

    if (    (request_len > 3) 
         && ((lc_request[ 0 ] == '2') || (lc_request[ 0 ] == '3'))
         && (lc_request[ 1 ] == 'd')
         && (lc_request[ 2 ] == '-')
       )
    {
        dimension = lc_request[ 0 ] - '0';
        BUFF_CPY(base_request, &(lc_request[ 3 ]));
    }
    else 
    {
        BUFF_CPY(base_request, lc_request);
    }

    if (result_ptr != NULL)
    {
        result_alg_ptr = &(result_ptr->sharp[ sharp_method ]);
        best_fit_alg_ptr = &(result_alg_ptr->best_fit); 

        standard_alg_ptr       = &(result_alg_ptr->standard);
        correlation_alg_ptr    = &(result_alg_ptr->correlation);   /* Lindsay */

#ifdef SIMPLE_TEMPLATE
        simple_template_alg_ptr = &(result_alg_ptr->simple_template);
#endif 
        
#ifdef TEMPLATE
        template_alg_ptr = &(result_alg_ptr->template);
#endif 

#ifdef MULTI_TEMPLATE
        multi_template_alg_ptr = &(result_alg_ptr->multi_template);
#endif 
    }

    if ((error_ptr != NULL) && (error_item_ptr_ptr != NULL))
    {
        error_alg_ptr = &(error_ptr->sharp[ sharp_method ]); 
    }

    if (error_alg_ptr != NULL) 
    {
        best_fit_error_ptr    = &(error_alg_ptr->best_fit);
        standard_error_ptr    = &(error_alg_ptr->standard);
        correlation_error_ptr = &(error_alg_ptr->correlation);
#ifdef SIMPLE_TEMPLATE
        simple_template_error_ptr = &(error_alg_ptr->simple_template);
#endif 
#ifdef TEMPLATE
        template_error_ptr = &(error_alg_ptr->template);
#endif 
#ifdef MULTI_TEMPLATE
        multi_template_error_ptr = &(error_alg_ptr->multi_template);
#endif 
    }

    if (dimension == 2) 
    {
        if (result_alg_ptr != NULL)
        {
            gamut_alg_ptr = &(result_alg_ptr->gamut_2D[ 0 ]);
        }

        if (error_alg_ptr != NULL) 
        {
            gamut_error_ptr = &(error_alg_ptr->gamut_2D[ 0 ]);
        }
    }
    else if (dimension == 3)
    {
        if (result_alg_ptr != NULL)
        {
            gamut_alg_ptr = &(result_alg_ptr->gamut_3D[ 0 ]);
        }

        if (error_alg_ptr != NULL) 
        {
            gamut_error_ptr = &(error_alg_ptr->gamut_3D[ 0 ]);
        }
    }

    ERE(parse_index(base_request, INT_DUMMY_ARG, (int*)NULL, (int*)NULL, 
                    &index, &sub_index));

    if (STRCMP_EQ(base_request, "actual")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                best_fit_alg_ptr->actual_vp,
                                request));
        }

        if (best_fit_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(best_fit_error_ptr->actual);
        }
    }
    else if (STRCMP_EQ(base_request, "ave-illum")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                standard_alg_ptr->ave_illum_vp,
                                request));
        }

        if (standard_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(standard_error_ptr->ave_illum);
        }
    }
    else if (STRCMP_EQ(base_request, "gw")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                standard_alg_ptr->grey_world_vp,
                                request));
        }

        if (standard_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(standard_error_ptr->grey_world);
        }
    }
    else if (STRCMP_EQ(base_request, "db-gw")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                standard_alg_ptr->db_grey_vp,
                                request));
        } 

        if (standard_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(standard_error_ptr->db_grey_world);
        }
    }
    else if (STRCMP_EQ(base_request, "scale-by-max")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                standard_alg_ptr->scale_by_max_vp,
                                request));
        } 

        if (standard_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(standard_error_ptr->scale_by_max);
        }
    }
    /* Lindsay - Sept 30/98. Modified April 14/99 */
    else if (STRCMP_EQ(base_request, "corr-01")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation 0/1"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_01_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(correlation_error_ptr->correlation_01[ index ]);
        }
    }
    else if (   STRCMP_EQ(base_request, "corr-uniform-map-inv-dist")
             || STRCMP_EQ(base_request, "corr-uniform-map-inv-dist-fill"))
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Inverse Distance)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                    &(correlation_error_ptr->correlation_uniform_MAP_inv_dist_fill[ index ]);
        }
    }
    else if (   STRCMP_EQ(base_request, "corr-uniform-map-gaussian")
             || STRCMP_EQ(base_request, "corr-uniform-map-gaussian-fill"))
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Gaussian Mask)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MAP_gaussian_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                   &(correlation_error_ptr->correlation_uniform_MAP_gaussian_fill[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-uniform-map-fitted")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Fitted)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MAP_fitted_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                &(correlation_error_ptr->correlation_uniform_MAP_fitted[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-gaussian-map")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Gaussian"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_gaussian_MAP_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr= &(correlation_error_ptr->correlation_gaussian_MAP[ index ]);
        }
    }
    else if (   STRCMP_EQ(base_request, "corr-uniform-mlm-inv-dist")
             || STRCMP_EQ(base_request, "corr-uniform-mlm-inv-dist-fill"))
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Inverse Distance)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                    &(correlation_error_ptr->correlation_uniform_MLM_inv_dist_fill[ index ]);
        }
    }
    else if (    (STRCMP_EQ(base_request, "corr-uniform-mlm-gaussian"))
              || (STRCMP_EQ(base_request, "corr-uniform-mlm-gaussian-fill"))
              || (STRCMP_EQ(base_request, "corr-uniform-mlm-gauss-fill"))
            ) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Gaussian Mask)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MLM_gaussian_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                   &(correlation_error_ptr->correlation_uniform_MLM_gaussian_fill[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-uniform-mlm-fitted")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Fitted)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MLM_fitted_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                &(correlation_error_ptr->correlation_uniform_MLM_fitted[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-gaussian-mlm")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Gaussian"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_gaussian_MLM_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr= &(correlation_error_ptr->correlation_gaussian_MLM[ index ]);
        }
    }
    else if (   STRCMP_EQ(base_request, "corr-uniform-mmse-inv-dist")
             || STRCMP_EQ(base_request, "corr-uniform-mmse-inv-dist-fill"))
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Inverse Distance)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                    &(correlation_error_ptr->correlation_uniform_MMSE_inv_dist_fill[ index ]);
        }
    }
    else if (    (STRCMP_EQ(base_request, "corr-uniform-mmse-gaussian"))
              || (STRCMP_EQ(base_request, "corr-uniform-mmse-gaussian-fill"))
              || (STRCMP_EQ(base_request, "corr-uniform-mmse-gauss-fill"))
            ) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Gaussian Mask)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                   &(correlation_error_ptr->correlation_uniform_MMSE_gaussian_fill[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-uniform-mmse-fitted")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Uniform (Fitted)"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_uniform_MMSE_fitted_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr =
                &(correlation_error_ptr->correlation_uniform_MMSE_fitted[ index ]);
        }
    }
    else if (STRCMP_EQ(base_request, "corr-gaussian-mmse")) 
    {
        ERE(check_index_range(index, NUM_CORMATS_PER_ALG, "Correlation Gaussian"));
        
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                correlation_alg_ptr->correlation_gaussian_MMSE_vp[ index ],
                                request));
        } 

        if (correlation_error_ptr != NULL)
        {
            *error_item_ptr_ptr= &(correlation_error_ptr->correlation_gaussian_MMSE[ index ]);
        }
    }
    /* End Lindsay Sept 30/98 */

#ifdef SIMPLE_TEMPLATE
    /*
    // Don't make the strings too long, or allow a second abreviated choice.
    // (The user has to type it relatively often).
    */
    else if (STRCMP_EQ(base_request, "simple-template-alg")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                simple_template_alg_ptr->alg_vp,
                                request));
        } 

        if (template_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(simple_template_error_ptr->alg);
        }
    }
#endif 
#ifdef TEMPLATE
    /*
    // Don't make the strings too long, or allow a second abreviated choice.
    // (The user has to type it relatively often).
    */
    else if (STRCMP_EQ(base_request, "template-first-alg")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                template_alg_ptr->first_alg_vp,
                                request));
        } 

        if (template_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(template_error_ptr->first_alg);
        }
    }
    /*
    // Don't make the strings too long, or allow a second abreviated choice.
    // (The user has to type it relatively often).
    */
    else if (STRCMP_EQ(base_request, "template-second-alg")) 
    {
        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                template_alg_ptr->second_alg_vp,
                                request));
        } 

        if (template_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(template_error_ptr->second_alg);
        }
    }
#endif 
#ifdef MULTI_TEMPLATE
    /*
    // Don't make the strings too long, or allow a second abreviated choice.
    // (The user has to type it relatively often).
    */
    else if (STRCMP_EQ(base_request, "multi-template-alg")) 
    {
        ERE(check_index_range(index, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                              "Multi-template"));

        if (map_vpp != NULL)
        {
            ERE(copy_map_vector(map_vpp,
                                multi_template_alg_ptr->alg_vp[ index ],
                                request));
        } 

        if (multi_template_error_ptr != NULL)
        {
            *error_item_ptr_ptr = &(multi_template_error_ptr->alg[ index ]);
        }
    }
#endif 
    else if (IS_SET(dimension))
    {
        /*
        // Backwards compatability. 
        */
        if (STRCMP_EQ(base_request, "s-mv")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].mv_surface_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].mv_surface);
            }
        }
        else if (STRCMP_EQ(base_request, "s-ha")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].ha_surface_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].ha_surface);
            }
        }
        else if (STRCMP_EQ(base_request, "s-wa")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].wa_surface_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].wa_surface);
            }
        }
        else if (STRCMP_EQ(base_request, "sl-ha")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));
            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].ha_surface_lum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].ha_surface_lum[  sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "sl-mv")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));
            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].mv_surface_lum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].mv_surface_lum[  sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "si-mv")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].mv_surface_illum_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].mv_surface_illum);
            }
        }
        else if (STRCMP_EQ(base_request, "si-ha")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].ha_surface_illum_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].ha_surface_illum);
            }
        }
        else if (STRCMP_EQ(base_request, "si-wa")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].wa_surface_illum_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].wa_surface_illum);
            }
        }
        else if (STRCMP_EQ(base_request, "si-icsa")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].icsa_surface_illum_vp,
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].icsa_surface_illum);
            }
        }
        else if (STRCMP_EQ(base_request, "si-scla")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].scla_surface_illum_lum_vp[sub_index],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr= &(gamut_error_ptr[ index ].scla_surface_illum_lum[sub_index]);
            }
        }
        else if (STRCMP_EQ(base_request, "sil-ha")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].ha_surface_illum_lum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].ha_surface_illum_lum[ sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "sil-mv")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].mv_surface_illum_lum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].mv_surface_illum_lum[ sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "sil-lcsa")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].lcsa_surface_illum_lum_vp[sub_index],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr= &(gamut_error_ptr[ index ].lcsa_surface_illum_lum[sub_index]);
            }
        }
        else if (STRCMP_EQ(base_request, "si-scia")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_SCA_WEIGHT_FUNCTIONS, 
                                  "Surface constrained average function"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].scia_surface_illum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].scia_surface_illum[ sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "si-scca")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_SCA_WEIGHT_FUNCTIONS, 
                                  "Surface constrained average function"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].scca_surface_illum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].scca_surface_illum[ sub_index ]);
            }
        }
        else if (STRCMP_EQ(base_request, "si-scia")) 
        {
            ERE(check_index_range(index, MAX_NUM_CANON_GAMUTS, "Canon gamut"));

            ERE(check_index_range(sub_index, MAX_NUM_LUMINANCE_METHODS, 
                                  "Luminance"));
            
            if (map_vpp != NULL)
            {
                ERE(copy_map_vector(map_vpp,
                                    gamut_alg_ptr[ index ].scia_surface_illum_vp[ sub_index ],
                                    request));
            } 

            if (gamut_error_ptr != NULL)
            {
                *error_item_ptr_ptr = &(gamut_error_ptr[ index ].scia_surface_illum[ sub_index ]);
            }
        }
        else if (    (STRCMP_EQ(base_request, "world-best-diag"))
                  || (STRCMP_EQ(base_request, "best-diag"))
                )
        {
            if (dimension == 2)
            {
                if (map_vpp != NULL)
                {
                    ERE(copy_map_vector(map_vpp,
                                        best_fit_alg_ptr->world_best_2D_diag_fit_vp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = 
                                     &(best_fit_error_ptr->world_best_2D_diag_fit);
                }
            }
            else 
            {
                if (map_vpp != NULL)
                {
                    ERE(copy_map_vector(map_vpp,
                                        best_fit_alg_ptr->world_best_3D_diag_fit_vp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr =
                                     &(best_fit_error_ptr->world_best_3D_diag_fit);
                }
            }
        }
        else if (    (STRCMP_EQ(base_request, "world-best"))
                  || (STRCMP_EQ(base_request, "best"))
                )
        {
            if (dimension == 2)
            {
                if (map_mpp != NULL)
                {
                    ERE(copy_map_matrix(map_mpp,
                                        best_fit_alg_ptr->world_best_2D_fit_mp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = &(best_fit_error_ptr->world_best_2D_fit);
                }
            }
            else 
            {
                if (map_mpp != NULL)
                {
                    ERE(copy_map_matrix(map_mpp,
                                        best_fit_alg_ptr->world_best_3D_fit_mp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = &(best_fit_error_ptr->world_best_3D_fit);
                }
            }
        }
        else if (STRCMP_EQ(base_request, "scene-best-diag"))
        {
            if (dimension == 2)
            {
                if (map_vpp != NULL)
                {
                    ERE(copy_map_vector(map_vpp,
                                        best_fit_alg_ptr->scene_best_2D_diag_fit_vp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = 
                                     &(best_fit_error_ptr->scene_best_2D_diag_fit);
                }
            }
            else 
            {
                if (map_vpp != NULL)
                {
                    ERE(copy_map_vector(map_vpp,
                                        best_fit_alg_ptr->scene_best_3D_diag_fit_vp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr =
                                     &(best_fit_error_ptr->scene_best_3D_diag_fit);
                }
            }
        }
        else if (STRCMP_EQ(base_request, "scene-best"))
        {
            if (dimension == 2)
            {
                if (map_mpp != NULL)
                {
                    ERE(copy_map_matrix(map_mpp,
                                        best_fit_alg_ptr->scene_best_2D_fit_mp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = &(best_fit_error_ptr->scene_best_2D_fit);
                }
            }
            else 
            {
                if (map_mpp != NULL)
                {
                    ERE(copy_map_matrix(map_mpp,
                                        best_fit_alg_ptr->scene_best_3D_fit_mp,
                                        request));
                } 

                if (best_fit_error_ptr != NULL)
                {
                    *error_item_ptr_ptr = &(best_fit_error_ptr->scene_best_3D_fit);
                }
            }
        }
        else  
        {
            set_error("%q is an invalid result designator.", request);
            return ERROR;
        }
    }
    else  
    {
        set_error("%q is an invalid result designator.", request);
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int copy_map_vector
(
    Vector** target_vpp,
    Vector*  map_vp,
    char*    tag 
)
{

    if (map_vp == NULL)
    {
        set_error("%s result is not available.", tag);
        return ERROR;
    }

    return copy_vector(target_vpp, map_vp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int copy_map_matrix
(
    Matrix** target_mpp,
    Matrix*  map_mp,
    char*    tag 
)
{

    if (map_mp == NULL)
    {
        set_error("%s result is not available.", tag);
        return ERROR;
    }

    return copy_matrix(target_mpp, map_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int remove_origin_from_data
(
    Matrix** target_mpp,
    Matrix*  source_mp,
    double   tolerance_arg 
)
{
    double tolerance = tolerance_arg;
    int  i, j;
    int  count;
    int  num_rows  = source_mp->num_rows;
    int  num_cols  = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    count = 0;

    for (i=0; i<num_rows; i++)
    {
        int keep = FALSE; 

        for (j=0; j<num_cols; j++)
        {
            if (source_mp->elements[i][j] > tolerance) 
            {
                keep = TRUE;
                break;
            }
        }

        if (keep)
        {
            for (j=0; j<num_cols; j++)
            {
                ((*target_mpp)->elements)[count][j] =
                                                    (source_mp->elements)[i][j];
            }
            count++;
        }
    }

    (*target_mpp)->num_rows = count;

    if (count == 0)
    {
        set_error("No points left after removing all the ones within %.3e ",
                  tolerance);
        cat_error("of the origin.");
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             parse_index
 *                               
 * Determines index from <str>:<index>{.<sub_index>}
 *
 * This routine determines index from <str>:<index>{.<sub_index>}. This routine
 * serves the purposes of several situations, and so is a bit convoluted.
 * However, it still seems better to have the code which interprets the user
 * input with respect to indexes all in one place. 
 * 
 * One issue is whether the string is part of a request or simply a state
 * query. The difference is that in the case of a state query, it usually makes
 * sense to be asking for all the items (i.e., we want the possible range)
 * whareas, in a request, it usually only makes sense to specify a particular
 * item. 
 * 
 * It is assumed that the range requests (first and last) are only relavent for
 * state queries, and they can be NULL if they are not relavent. 
 *
 * In most contexts, a sub index is not relavent, and in these cases, NULL can
 * be used. It is assumed that a sub index is only relavent for non-query cases. 
 * This is enforced. Specifically, if sub_index_ptr is not NULL, then the range
 * request pointers MUST be NULL. 
 * 
 * If we are processing a state query, then if the user does not specify an
 * index, then this is interpreted as asking for a range. In the case of a
 * non-query request, where a result is being asked for, or a value is being
 * set, then no index is interpreted as index being 0 (this helps with backwards
 * compatability). 
 *
 * If there is an index (but no sub_index), then all *index_ptr,
 * *first_index_ptr, and *second_index_ptr are set to that value.
 *
 * Regardless of the various possibilities, if there is a ":" in the input, this
 * routine destructively erases the ":" and all subsequent characters. 
 * 
 * It is important to note that this routine does not use max_index to error
 * check, as we are assuming that we are possibly parsing a string which is not
 * meant for the calling code. In other words, this routine answers the question
 * of what the values should be IF it WERE the case that we were parsing an
 * indexed option of interest. The prescence of an index does not guarentee
 * this.  Index ranges are checked using check_index_range(). 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure with an error message being
 *     set. This routine will fail if the the text after the colon is not a
 *     valid positive integer. 
 *
 * Index: gamut library
 *     
 * -----------------------------------------------------------------------------
*/ 

int parse_index
(
    char* str,
    int   max_index,
    int*  first_index_ptr,
    int*  second_index_ptr,
    int*  index_ptr,
    int*  sub_index_ptr 
)
{

    if (    (sub_index_ptr != NULL) 
         && ((first_index_ptr != NULL) || (second_index_ptr != NULL))
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    if (FIND_CHAR_YES(str, ':'))
    {
        char  index_str[ 100 ];
        char  num_buff[ 100 ];
        char  index_buff[ 100 ];
        char* index_buff_pos; 
        char* index_str_pos;

        gen_get_last_token(str, index_str, sizeof(index_str), ":");
        str[ strlen(str) - 1 ] = '\0';

        index_str_pos = index_str;

        BUFF_GEN_GET_TOKEN(&index_str_pos, index_buff, "|"); 
        index_buff_pos = index_buff;

        BUFF_GEN_GET_TOKEN(&index_buff_pos, num_buff, "."); 

        ERE(ss1pi(num_buff, index_ptr));
        (*index_ptr)--;

        if (sub_index_ptr == NULL)
        {
            if (first_index_ptr != NULL)
            {
                *first_index_ptr = *index_ptr;
            }

            if (second_index_ptr != NULL)
            {
                *second_index_ptr = *index_ptr;
            }
        }
        else 
        {
            if (BUFF_GEN_GET_TOKEN_OK(&index_buff_pos, num_buff, "."))
            {
                ERE(ss1pi(num_buff, sub_index_ptr));
                (*sub_index_ptr)--;
            }
            else 
            {
                *sub_index_ptr = 0;
            }
        }
    }
    else 
    {
        if (first_index_ptr != NULL)
        {
            *first_index_ptr = 0;
        }

        if (second_index_ptr != NULL)
        {
            *second_index_ptr = max_index - 1;
        }

        *index_ptr = 0; 

        if (sub_index_ptr != NULL)
        {
            *sub_index_ptr = 0; 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             check_index_range
 *                               
 * Checks that index is in range
 *
 * This routine checks that index is in range. If it is not, then an error
 * message is set, and ERROR is returned. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure with an error message being
 *     set. 
 *
 * Index: gamut library
 *     
 * -----------------------------------------------------------------------------
*/ 

int check_index_range(int index, int max_index, const char* title_str)
{

    if ((index < 0) || (index >= max_index))
    {
        set_error("%s index must be between 1 and %d.", title_str, max_index);
        return ERROR;
    }
    else 
    {
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int write_data_matrix
(
    Matrix* output_mp,
    char*   output_file,
    int     append_flag 
)
{
    int          result = NO_ERROR;
    Error_action save_error_action = get_error_action();

    if (append_flag) 
    {
        FILE *fp = kjb_fopen(output_file, "a");
        off_t byte_size; 

        if (fp == NULL) 
        {
            result = ERROR;
        }
        else if (fp_get_byte_size(fp, &byte_size) == ERROR)
        {
            result = ERROR; 
        }

        dbi(byte_size); 

        if ((result != ERROR) && (byte_size > 0))
        {
            result = kjb_fputs(fp, "\n#!eof\n\n"); 
        }

        if (result != ERROR) 
        {
            result = fp_write_matrix(output_mp, fp); 
        }

        if (result == ERROR) 
        {
            set_error_action(FORCE_ADD_ERROR_ON_ERROR);
        }

        if (kjb_fclose(fp) == ERROR)
        {
            result = ERROR; 
        }
        
        set_error_action(save_error_action);
    }
    else 
    {
        result = write_matrix(output_mp, output_file); 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

