
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"     /* Only safe as first include in a ".c" file. */

#include "i/i_convolve.h"

#define NORMALIZE_CONVOLUTIONS

static int get_2D_gaussian_mask 
(
    Matrix** mask_mpp,     /* Output gaussian smoothing mask. */
    int      mask_size,    /* Number of elements per axis. Must be odd. */
    double   sigma         /* Standard deviation in bin units.         */
);

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                           ow_gauss_convolve_image   
 *                               
 * Convolves image with Gaussian mask
 *
 * This routine convolves and image with a Gaussian mask, overwriting the image
 * in the process. The mask size is automatically chosen based on the value of
 * sigma. 
 *
 * Index: images, convolution, image transformation, image smoothing
 *
 * -----------------------------------------------------------------------------
*/ 

int ow_gauss_convolve_image(KJB_image* ip, double sigma)
{
    KJB_image* out_ip = NULL;
    int          num_rows = ip->num_rows;
    int          num_cols = ip->num_cols;
    int i, j;

    ERE(gauss_convolve_image(&out_ip, ip, sigma));

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            ip->pixels[ i ][ j ] = out_ip->pixels[ i ][ j ];
        }
    }

    kjb_free_image(out_ip);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =================================================================================================
 *                           gauss_convolve_image   
 *                               
 * Convolves image with Gaussian mask
 *
 * This routine convolves the image pointed to by in_ip with a Gaussian mask
 * with the specified sigma, putting the result into *out_ipp.  The mask
 * size is automatically chosen based on the value of sigma. 
 *
 * If *out_ipp is NULL, then an image of the appropriate size is created, if
 * it is the wrong size, then it is resized, and if it is the right size, the
 * storage is recycled. 
 *
 * Index: images, convolution, image transformation, image smoothing
 *
 * -----------------------------------------------------------------------------
*/ 

int gauss_convolve_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    double           sigma 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols, i, j, mi, mj, m, n;
    int        mask_width;
    int        mask_size;
    Matrix*    mask_mp    = NULL;

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    mask_width = 1.0 + 2.0 * sigma; 
    mask_size  = 1 + 2 * mask_width; 

    ERE(get_2D_gaussian_mask(&mask_mp, mask_size, sigma)); 

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            double r_sum      = 0.0;
            double g_sum      = 0.0;
            double b_sum      = 0.0;
            double weight_sum = 0.0;
            double weight; 

            for (mi = 0; mi < mask_size; mi++)
            {
                for (mj = 0; mj < mask_size; mj++)
                {
                    m = i + mi - mask_width;
                    n = j + mj - mask_width;

                    if (   (m >= 0) && (m < num_rows)
                        && (n >= 0) && (n < num_cols))
                    {
                        weight = mask_mp->elements[mi][mj];

                        r_sum += in_ip->pixels[m][n].r * weight; 
                        g_sum += in_ip->pixels[m][n].g * weight; 
                        b_sum += in_ip->pixels[m][n].b * weight; 

#ifdef NORMALIZE_CONVOLUTIONS 
                        weight_sum += weight;
#endif 
                    }
                }
            }
#ifdef NORMALIZE_CONVOLUTIONS 
            r_sum /= weight_sum; 
            g_sum /= weight_sum; 
            b_sum /= weight_sum; 
#endif 
            out_ip->pixels[i][j].r = r_sum; 
            out_ip->pixels[i][j].g = g_sum; 
            out_ip->pixels[i][j].b = b_sum; 

            out_ip->pixels[i][j].invalid = in_ip->pixels[i][j].invalid; 
        }
    }

    free_matrix(mask_mp);  

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                           gauss_sample_image   
 *                               
 * Takes gaussian samples of an image
 *
 * This routine reduces image resolution by taking averages with Gaussian
 * weights. 
 *
 * If *out_ipp is NULL, then an image of the appropriate size is created, if
 * it is the wrong size, then it is resized, and if it is the right size, the
 * storage is recycled. 
 *
 * Index: images, convolution, image transformation, image smoothing
 *
 * -----------------------------------------------------------------------------
*/ 

int gauss_sample_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    int              resolution,
    double           sigma 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols, oi, oj, i, j, mi, mj, m, n;
    int        num_out_rows;
    int        num_out_cols;
    int        mask_width;
    int        mask_size;
    Matrix*    mask_mp      = NULL;

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

    num_out_rows = num_rows / resolution;
    num_out_cols = num_cols / resolution;

    ERE(get_target_image(out_ipp, num_out_rows, num_out_cols)); 
    out_ip = *out_ipp; 

    /* EXTRA_FOR_ROUNDING added July 14, 2002. */
    mask_width = 1.0 + 2.0 * sigma + EXTRA_FOR_ROUNDING; 
    mask_size  = 1 + 2 * mask_width; 

    ERE(get_2D_gaussian_mask(&mask_mp, mask_size, sigma)); 

    i = resolution / 2;

    for (oi=0; oi<num_out_rows; oi++)
    {
        j = resolution / 2;
        
        for (oj=0; oj<num_out_cols; oj++)
        {
            double r_sum      = 0.0;
            double g_sum      = 0.0;
            double b_sum      = 0.0;
            double weight_sum = 0.0;
            double weight; 

            for (mi = 0; mi < mask_size; mi++)
            {
                for (mj = 0; mj < mask_size; mj++)
                {
                    m = i + mi - mask_width;
                    n = j + mj - mask_width;

                    if (   (m >= 0) && (m < num_rows)
                        && (n >= 0) && (n < num_cols)
                        /*
                        && ( ! in_ip->pixels[m][n].invalid.pixel)
                        */
                       )
                    {
                        weight = mask_mp->elements[mi][mj];

                        r_sum += in_ip->pixels[m][n].r * weight; 
                        g_sum += in_ip->pixels[m][n].g * weight; 
                        b_sum += in_ip->pixels[m][n].b * weight; 

                        weight_sum += weight;
                    }
                }
            }

            if (weight_sum > 0.0)
            {
                r_sum /= weight_sum; 
                g_sum /= weight_sum; 
                b_sum /= weight_sum; 

                out_ip->pixels[oi][oj].r = r_sum; 
                out_ip->pixels[oi][oj].g = g_sum; 
                out_ip->pixels[oi][oj].b = b_sum; 

                out_ip->pixels[oi][oj].invalid.pixel = VALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.r = VALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.g = VALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.b = VALID_PIXEL; 
            }
            else 
            {
                out_ip->pixels[oi][oj].r = in_ip->pixels[i][j].r; 
                out_ip->pixels[oi][oj].g = in_ip->pixels[i][j].g; 
                out_ip->pixels[oi][oj].b = in_ip->pixels[i][j].b; 

                out_ip->pixels[oi][oj].invalid.pixel = INVALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.r = INVALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.g = INVALID_PIXEL; 
                out_ip->pixels[oi][oj].invalid.b = INVALID_PIXEL; 
            }
        
            j += resolution; 
        }

        i += resolution; 
    }

    free_matrix(mask_mp);  

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_gaussian_mask
(
    Matrix** mask_mpp, /* Output gaussian smoothing mask.           */
    int      mask_size,/* Number of elements per axis. Must be odd. */
    double   sigma     /* Standard deviation in bin units.         */
)
{
    int r, c;
    int    mask_centre;
    double dist_sqr;
    double sigma_sqr;
    double gauss;
    double mask_sum;
    
    if ( (mask_size < 0 ) || IS_EVEN(mask_size) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    sigma_sqr = sigma * sigma;

    mask_centre = mask_size / 2;
    mask_sum  = 0.0;
    
    ERE(get_zero_matrix(mask_mpp, mask_size, mask_size));

    for (r = 0; r < mask_size; r++)
    {
        for (c = 0; c < mask_size; c++)
        {
            dist_sqr =(r - mask_centre)*(r - mask_centre) + 
                                            (c - mask_centre)*(c - mask_centre);

            gauss = exp( -0.5 * dist_sqr / sigma_sqr ); 
            (*mask_mpp)->elements[r][c] = gauss;
            mask_sum += gauss;
        }
    }

    ERE(ow_divide_matrix_by_scalar(*mask_mpp,  mask_sum));
   
    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

