
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"     /* Only safe as first include in a ".c" file. */
#include "i/i_draw.h"

/* -------------------------------------------------------------------------- */

#define NUM_POINTS_IN_LINE   2000

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                             image_draw_contour
 *                               
 * Draws a contour on an image
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_contour
(
    KJB_image* ip,
    Matrix*    mp,
    int        width,
    int        r,
    int        g,
    int        b 
)
{
    Vector* vp1    = NULL;
    Vector* vp2    = NULL;
    int     i;
    int     result = NO_ERROR;

    for (i = 0; i<mp->num_rows; i++)
    {
        if (result == ERROR) break;

        result = get_matrix_row(&vp1, mp, i);

        if (result != ERROR)
        {
             result = get_matrix_row(&vp2, mp, 
                                     ((i + 1) == mp->num_rows) ? 0 : i + 1);
        }

        if (result != ERROR)
        {
            image_draw_segment(ip, vp1, vp2, width, r, g, b); 
        }
    }

    free_vector(vp1); 
    free_vector(vp2); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_gradient
 *                               
 * Draws a line segment with gradient on an image
 *
 * This routine sets the pixels of "ip" in the segment joining p1_vp
 * and p2_vp to a shade of (r1,g1,b1) and (r2,g2,b2).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_gradient
(
    KJB_image* ip,
    Vector*    p1_vp,
    Vector*    p2_vp,
    int        width,
    int        r1,
    int        g1,
    int        b1,
    int        r2,
    int        g2,
    int        b2 
)
{
    double x       = p1_vp->elements[ 0 ];
    double y       = p1_vp->elements[ 1 ];
    double x_step  = (p2_vp->elements[ 0 ] - x) / (double)NUM_POINTS_IN_LINE;
    double y_step  = (p2_vp->elements[ 1 ] - y) / (double)NUM_POINTS_IN_LINE;
    int    i, ii, jj, di, dj, iii, jjj;
    float  f_r     = MIN_OF(255, MAX_OF(0, r1));
    float  f_g     = MIN_OF(255, MAX_OF(0, g1));
    float  f_b     = MIN_OF(255, MAX_OF(0, b1));
    float  r_range = MIN_OF(255, MAX_OF(0, r2)) - r1;
    float  g_range = MIN_OF(255, MAX_OF(0, g2)) - g1;
    float  b_range = MIN_OF(255, MAX_OF(0, b2)) - b1;
    float  r_step  = r_range / NUM_POINTS_IN_LINE;
    float  g_step  = g_range / NUM_POINTS_IN_LINE;
    float  b_step  = b_range / NUM_POINTS_IN_LINE;

    width--;
    width = MAX_OF(0, width); 

    if ((ip == NULL) || (p1_vp == NULL) || (p2_vp == NULL)) return NO_ERROR;

    for (i=0; i<NUM_POINTS_IN_LINE; i++)
    {
        ii = x + 0.5;
        jj = y + 0.5;

        for (di = -width; di < width + 1; di++)
        {
            iii = MAX_OF(0, MIN_OF(ip->num_rows - 1, ii + di)); 

            for (dj = -width; dj < width + 1; dj++)
            {
                jjj = MAX_OF(0, MIN_OF(ip->num_cols - 1, jj + dj)); 

                ip->pixels[ iii ][ jjj ].r = f_r;
                ip->pixels[ iii ][ jjj ].g = f_g;
                ip->pixels[ iii ][ jjj ].b = f_b;
                ip->pixels[ iii ][ jjj ].invalid.pixel = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.r = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.g = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.b = VALID_PIXEL;
            }
        }

        x += x_step;
        y += y_step;
        f_r += r_step;
        f_g += g_step;
        f_b += b_step;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_segment
 *                               
 * Draws a line segment on an image
 *
 * This routine sets the pixels of "ip" in the segment joining p1_vp
 * and p2_vp to (r,g,b).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_segment
(
    KJB_image* ip,
    Vector*    p1_vp,
    Vector*    p2_vp,
    int        width,
    int        r,
    int        g,
    int        b 
)
{
    double x      = p1_vp->elements[ 0 ];
    double y      = p1_vp->elements[ 1 ];
    double x_step = (p2_vp->elements[ 0 ] - x) / (double)NUM_POINTS_IN_LINE;
    double y_step = (p2_vp->elements[ 1 ] - y) / (double)NUM_POINTS_IN_LINE;
    int    i, ii, jj, di, dj, iii, jjj;
    float  f_r    = MAX_OF(0, r);
    float  f_g    = MAX_OF(0, g);
    float  f_b    = MAX_OF(0, b);

    if ((ip == NULL) || (p1_vp == NULL) || (p2_vp == NULL)) return NO_ERROR;

    width--;
    width = MAX_OF(0, width); 

    for (i=0; i<NUM_POINTS_IN_LINE; i++)
    {
        ii = x + 0.5;
        jj = y + 0.5;

        for (di = -width; di < width + 1; di++)
        {
            iii = MAX_OF(0, MIN_OF(ip->num_rows - 1, ii + di)); 

            for (dj = -width; dj < width + 1; dj++)
            {
                jjj = MAX_OF(0, MIN_OF(ip->num_cols - 1, jj + dj)); 

                ip->pixels[ iii ][ jjj ].r = f_r;
                ip->pixels[ iii ][ jjj ].g = f_g;
                ip->pixels[ iii ][ jjj ].b = f_b;
                ip->pixels[ iii ][ jjj ].invalid.r = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.g = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.b = VALID_PIXEL;
                ip->pixels[ iii ][ jjj ].invalid.pixel = VALID_PIXEL;
            }
        }

        x += x_step;
        y += y_step;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_segment_2
 *                               
 * Draws a line segment on an image
 *
 * This routine sets the pixels of "ip" in the segment joining (i_beg,
 * j_beg) and (i_end, j_end) to (r,g,b).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_segment_2
(
    KJB_image* ip,
    int        i_beg,
    int        j_beg,
    int        i_end,
    int        j_end,
    int        width,
    int        r,
    int        g,
    int        b 
)
{
    static double p1_elements[ 2 ];
    static Vector  p1    = {2, p1_elements};
    static Vector* p1_vp = &p1;
    static double  p2_elements[ 2 ];
    static Vector  p2    = {2, p2_elements};
    static Vector* p2_vp = &p2;

    p1_elements[ 0 ] = i_beg; 
    p1_elements[ 1 ] = j_beg; 
    p2_elements[ 0 ] = i_end; 
    p2_elements[ 1 ] = j_end; 

    return image_draw_segment(ip, p1_vp, p2_vp, width, r, g, b); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_points
 *                               
 * Draws a set of points on an image
 *
 * This routine sets the pixels of "ip" whose coordinates are in the matrix "mp"
 * to (r,g,b).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_points(KJB_image* ip, Matrix* mp, int r, int g, int b)
{
    int    num_rows = ip->num_rows;
    int    num_cols = ip->num_cols;
    int    i, j;
    int    count;
    float  f_r      = r;
    float  f_g      = g;
    float  f_b      = b;
    double x, y;

    if ((ip == NULL) || (mp == NULL)) return NO_ERROR;

    for (count=0; count<mp->num_rows; count++)
    {
        x = mp->elements[ count ][ 0 ];
        y = mp->elements[ count ][ 1 ];
        i = MAX_OF(0, MIN_OF(num_rows - 1, (int)(x + 0.5)));
        j = MAX_OF(0, MIN_OF(num_cols - 1, (int)(y + 0.5)));

        ip->pixels[ i ][ j ].r = f_r;
        ip->pixels[ i ][ j ].g = f_g;
        ip->pixels[ i ][ j ].b = f_b;
        ip->pixels[ i ][ j ].invalid.r = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.g = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.b = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.pixel = VALID_PIXEL;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_pixels
 *                               
 * Draws a set of pixels on an image
 *
 * This routine sets the pixels of "ip" whose coordinates are in "pixels"
 * to (r,g,b).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_pixels
(
    KJB_image*  ip,
    int         num_pixels,
    Pixel_info* pixels,
    int         r,
    int         g,
    int         b 
)
{
    int   i, j;
    int   count;
    float f_r   = r;
    float f_g   = g;
    float f_b   = b;

    if ((ip == NULL) || (pixels == NULL)) return NO_ERROR;

    for (count = 0; count < num_pixels; count++)
    {
        i = pixels[ count ].i;
        j = pixels[ count ].j;

        ip->pixels[ i ][ j ].r = f_r;
        ip->pixels[ i ][ j ].g = f_g;
        ip->pixels[ i ][ j ].b = f_b;
        ip->pixels[ i ][ j ].invalid.pixel = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.r = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.g = VALID_PIXEL;
        ip->pixels[ i ][ j ].invalid.b = VALID_PIXEL;
    }

    return NO_ERROR;
}
              
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_point
 *                               
 * Draws a point on an image
 *
 * This routine sets the pixels (i += width, j += width) of "ip" to (r,g,b).
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_point
(
    KJB_image* ip,
    int        i,
    int        j,
    int        width,
    int        r,
    int        g,
    int        b 
)
{
    int   di, dj;
    float f_r = r;
    float f_g = g;
    float f_b = b;

    if (ip == NULL) return NO_ERROR;

    for (di=-width; di<=width; di++)
    {
        for (dj=-width; dj<=width; dj++)
        {
            if (    (i + di >= 0)
                 && (j + dj >= 0) 
                 && (i + di < ip->num_rows)
                 && (j + dj < ip->num_cols)
               )
            {
                ip->pixels[ i + di ][ j + dj ].r = f_r;
                ip->pixels[ i + di ][ j + dj ].g = f_g;
                ip->pixels[ i + di ][ j + dj ].b = f_b;
                ip->pixels[ i + di ][ j + dj ].invalid.pixel = VALID_PIXEL;
                ip->pixels[ i + di ][ j + dj ].invalid.r = VALID_PIXEL;
                ip->pixels[ i + di ][ j + dj ].invalid.g = VALID_PIXEL;
                ip->pixels[ i + di ][ j + dj ].invalid.b = VALID_PIXEL;
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_image
 *                               
 * Draws an image onto an image
 *
 * This routine draws an image ip onto canvas_ip  at location (i, j), reducing
 * the size of ip by scale. The image is clipped if part (or all) of it lies
 * outsize canvas_ip. 
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_image
(
    KJB_image*       canvas_ip,
    const KJB_image* ip,
    int              i,
    int              j,
    int              scale 
)
{
    int di, dj; 
    int ii = i;

    if (ip == NULL) return NO_ERROR;

    for (di = 0; di  < ip->num_rows; di += scale)
    {
        int jj = j;

        for (dj = 0; dj < ip->num_cols; dj += scale)
        {
            if (    (ii >= 0)
                 && (jj >= 0)
                 && (ii < canvas_ip->num_rows)
                 && (jj < canvas_ip->num_cols)
               )
            {
                canvas_ip->pixels[ ii ][ jj ].r = ip->pixels[ di ][ dj ].r;
                canvas_ip->pixels[ ii ][ jj ].g = ip->pixels[ di ][ dj ].g;
                canvas_ip->pixels[ ii ][ jj ].b = ip->pixels[ di ][ dj ].b;
                canvas_ip->pixels[ ii ][ jj ].invalid.r = VALID_PIXEL; 
                canvas_ip->pixels[ ii ][ jj ].invalid.g = VALID_PIXEL; 
                canvas_ip->pixels[ ii ][ jj ].invalid.b = VALID_PIXEL; 
                canvas_ip->pixels[ ii ][ jj ].invalid.pixel = VALID_PIXEL; 
            }

            jj++;
        }

        ii++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_draw_rectange
 *                               
 * Draws a rectange on an image
 *
 * This routine draws a rectange of dimensions (width, height) at loaction
 * (i,j) onto an image.
 *
 * Index: images, image output
 *
 * -----------------------------------------------------------------------------
*/ 

int image_draw_rectangle
(
    KJB_image* ip,
    int        i,
    int        j,
    int        height,
    int        width,
    int        r,
    int        g,
    int        b 
)
{
    int   di, dj;
    float f_r = r;
    float f_g = g;
    float f_b = b;
    int   ii  = i;

    if (ip == NULL) return NO_ERROR;

    for (di=0; di<height; di++)
    {
        int jj = j; 

        for (dj=0; dj<width; dj++)
        {
            if (    (ii > 0)
                 && (jj > 0)
                 && (ii < ip->num_rows)
                 && (jj < ip->num_cols)
               )
            {
                ip->pixels[ ii ][ jj ].r = f_r;
                ip->pixels[ ii ][ jj ].g = f_g;
                ip->pixels[ ii ][ jj ].b = f_b;
                ip->pixels[ ii ][ jj ].invalid.pixel = VALID_PIXEL;
                ip->pixels[ ii ][ jj ].invalid.r = VALID_PIXEL;
                ip->pixels[ ii ][ jj ].invalid.g = VALID_PIXEL;
                ip->pixels[ ii ][ jj ].invalid.b = VALID_PIXEL;
            }

            jj++;
        }
        ii++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

