
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"     /* Only safe as first include in a ".c" file. */

#include "i/i_map.h"
#include "i/i_stat.h"

/* =============================================================================
 *                               pre_map_image   
 *                               
 *
 * Index: images, image mapping, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int pre_map_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols;
    Pixel*     in_pos;
    Pixel*     out_pos;
    int        i, j;
    Vector*    pixel_vp;
    Vector*    new_pixel_vp = NULL;

    if (    (map_mp->num_rows != 3) 
         || ((map_mp->num_cols != 3) && (map_mp->num_cols != 4)) 
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    NRE(pixel_vp = create_vector(map_mp->num_cols));

    for (i=0; i<num_rows; i++)
    {
        in_pos  = in_ip->pixels[ i ];
        out_pos = out_ip->pixels[ i ];

        for (j=0; j<num_cols; j++)
        {
            (pixel_vp->elements)[ 0 ] = in_pos->r;
            (pixel_vp->elements)[ 1 ] = in_pos->g;
            (pixel_vp->elements)[ 2 ] = in_pos->b;

            if (map_mp->num_cols == 4)
            {
                (pixel_vp->elements)[ 3 ] = 1.0;
            }

            ERE(multiply_matrix_and_vector(&new_pixel_vp, map_mp, pixel_vp));

            out_pos->r = (new_pixel_vp->elements)[ 0 ];
            out_pos->g = (new_pixel_vp->elements)[ 1 ];
            out_pos->b = (new_pixel_vp->elements)[ 2 ];

            out_pos->invalid = in_pos->invalid; 

            if (out_pos->r < FLT_ZERO) 
            {
                out_pos->r = FLT_ZERO; 
                out_pos->invalid.r |= TRUNCATED_PIXEL;
            }

            if (out_pos->g < FLT_ZERO) 
            {
                out_pos->g = FLT_ZERO; 
                out_pos->invalid.g |= TRUNCATED_PIXEL;
            }

            if (out_pos->b < FLT_ZERO) 
            {
                out_pos->b = FLT_ZERO; 
                out_pos->invalid.b |= TRUNCATED_PIXEL;
            }

            out_pos->invalid.pixel = 
                 out_pos->invalid.r | out_pos->invalid.g | out_pos->invalid.b;

            out_pos++;
            in_pos++;
        }
    }

    free_vector(pixel_vp);
    free_vector(new_pixel_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               post_map_image   
 *                               
 *
 * Index: images, image mapping, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int post_map_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols;
    Pixel*     in_pos;
    Pixel*     out_pos;
    int        i, j;
    Vector*    pixel_vp;
    Vector*    new_pixel_vp = NULL;

    if (    (map_mp->num_cols != 3) 
         || ((map_mp->num_rows != 3) && (map_mp->num_rows != 4)) 
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

#ifdef TEST
    if (map_mp->num_rows == 4)
    {
    }
#endif 

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    NRE(pixel_vp = create_vector(map_mp->num_rows));

    for (i=0; i<num_rows; i++)
    {
        in_pos  = in_ip->pixels[ i ];
        out_pos = out_ip->pixels[ i ];

        for (j=0; j<num_cols; j++)
        {
            (pixel_vp->elements)[ 0 ] = in_pos->r;
            (pixel_vp->elements)[ 1 ] = in_pos->g;
            (pixel_vp->elements)[ 2 ] = in_pos->b;

            if (map_mp->num_rows == 4)
            {
                (pixel_vp->elements)[ 3 ] = 1.0; 
            }

            ERE(multiply_vector_and_matrix(&new_pixel_vp, pixel_vp, map_mp));

            out_pos->r = (new_pixel_vp->elements)[ 0 ];
            out_pos->g = (new_pixel_vp->elements)[ 1 ];
            out_pos->b = (new_pixel_vp->elements)[ 2 ];

            out_pos->invalid = in_pos->invalid; 

            if (out_pos->r < FLT_ZERO) 
            {
                out_pos->r = FLT_ZERO; 
                out_pos->invalid.r |= TRUNCATED_PIXEL;
            }

            if (out_pos->g < FLT_ZERO) 
            {
                out_pos->g = FLT_ZERO; 
                out_pos->invalid.g |= TRUNCATED_PIXEL;
            }

            if (out_pos->b < FLT_ZERO) 
            {
                out_pos->b = FLT_ZERO; 
                out_pos->invalid.b |= TRUNCATED_PIXEL;
            }

            out_pos->invalid.pixel = 
                 out_pos->invalid.r | out_pos->invalid.g | out_pos->invalid.b;

            out_pos++;
            in_pos++;
        }
    }

    free_vector(pixel_vp);
    free_vector(new_pixel_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             post_map_projected_image    
 *                               
 *
 * Index: images, image mapping, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int post_map_projected_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols;
    Pixel*     in_pos;
    Pixel*     out_pos;
    int        i, j;
    double     red, blue, green;
    Vector*    pixel_vp;
    Vector*    new_pixel_vp = NULL;

    if ((map_mp->num_rows != 2) || (map_mp->num_cols != 2) )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    NRE(pixel_vp = create_vector(2));

    for (i=0; i<num_rows; i++)
    {
        in_pos  = in_ip->pixels[ i ];
        out_pos = out_ip->pixels[ i ];

        for (j=0; j<num_cols; j++)
        {
            red   = in_pos->r;
            green = in_pos->g;
            blue  = in_pos->b;

            if (blue < 1.0) blue = 1.0;

            (pixel_vp->elements)[ 0 ] = red / blue;
            (pixel_vp->elements)[ 1 ] = green / blue;

            ERE(multiply_vector_and_matrix(&new_pixel_vp, pixel_vp, map_mp));

            out_pos->r = blue * (new_pixel_vp->elements)[ 0 ];
            out_pos->g = blue * (new_pixel_vp->elements)[ 1 ];
            out_pos->b = blue;

            out_pos->invalid = in_pos->invalid; 

            out_pos->invalid.r |= out_pos->invalid.b;
            out_pos->invalid.g |= out_pos->invalid.b;

            if (out_pos->r < FLT_ZERO) 
            {
                out_pos->r = FLT_ZERO; 
                out_pos->invalid.r |= TRUNCATED_PIXEL;
            }

            if (out_pos->g < FLT_ZERO) 
            {
                out_pos->g = FLT_ZERO; 
                out_pos->invalid.g |= TRUNCATED_PIXEL;
            }

            if (out_pos->b < FLT_ZERO) 
            {
                out_pos->b = FLT_ZERO; 
                out_pos->invalid.b |= TRUNCATED_PIXEL;
            }

            out_pos->invalid.pixel = 
                 out_pos->invalid.r | out_pos->invalid.g | out_pos->invalid.b;

            out_pos++;
            in_pos++;
        }
    }

    free_vector(pixel_vp);
    free_vector(new_pixel_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          change_basis_post_map_image     
 *                               
 *
 * Index: images, image mapping, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int change_basis_post_map_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    Matrix*          change_basis_mp,
    Matrix*          map_mp 
)
{
    Matrix* temp_mp             = NULL;
    Matrix* inv_change_basis_mp = NULL;
    Matrix* prod_map_mp         = NULL;
    int     result;

    if (    (map_mp->num_rows != map_mp->num_cols) 
         || (change_basis_mp->num_rows != change_basis_mp->num_cols) 
         || (change_basis_mp->num_rows != map_mp->num_rows) 
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    ERE(get_matrix_inverse(&inv_change_basis_mp, change_basis_mp));

    result = multiply_matrices(&temp_mp, change_basis_mp, map_mp);

    if (result == ERROR) 
    {
        free_matrix(inv_change_basis_mp);
        return ERROR; 
    }

    result = multiply_matrices(&prod_map_mp, temp_mp, inv_change_basis_mp);

    if (result == ERROR) 
    {
        free_matrix(inv_change_basis_mp);
        free_matrix(temp_mp);
        return ERROR; 
    }

    result = post_map_image(out_ipp, in_ip, prod_map_mp);

    free_matrix(inv_change_basis_mp);
    free_matrix(temp_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                       change_basis_post_map_projected_image         
 *                               
 *
 *
 * Index: images, image mapping, image transformation 
 *
 * -----------------------------------------------------------------------------
*/ 

int change_basis_post_map_projected_image
(
    KJB_image**      out_ipp,
    const KJB_image* in_ip,
    Matrix*          change_basis_mp,
    Matrix*          map_mp 
)
{
    KJB_image* out_ip;
    Pixel*     in_pos;
    Pixel*     out_pos;
    int        i, j, num_rows, num_cols;
    double     red, blue, green;
    double     factor;
    Vector*    pixel_vp;
    Vector*    new_pixel_vp = NULL;
    Vector*    pr_pixel_vp;
    Vector*    pr_new_pixel_vp     = NULL;
    Matrix*    inv_change_basis_mp = NULL;
    Vector*    ave_rgb_vp          = NULL;

    num_rows = in_ip->num_rows;
    num_cols = in_ip->num_cols;

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    ERE(get_matrix_inverse(&inv_change_basis_mp, change_basis_mp));

    pixel_vp        = create_vector(3);
    pr_pixel_vp     = create_vector(2);

    if (    (get_ave_rgb(&ave_rgb_vp, in_ip) == ERROR)
         || (inv_change_basis_mp == NULL)  
         || (pixel_vp == NULL)    
         || (pr_pixel_vp == NULL) 
       )
    {
        free_vector(ave_rgb_vp);
        free_vector(pixel_vp);
        free_vector(pr_pixel_vp);
        free_matrix(inv_change_basis_mp);
        return ERROR; 
    }

    factor = (ave_rgb_vp->elements)[ 2 ];
    free_vector(ave_rgb_vp);

    for (i=0; i<num_rows; i++)
    {
        in_pos  = in_ip->pixels[ i ];
        out_pos = out_ip->pixels[ i ];

        for (j=0; j<num_cols; j++)
        {
            (pixel_vp->elements)[ 0 ] = in_pos->r;
            (pixel_vp->elements)[ 1 ] = in_pos->g;
            (pixel_vp->elements)[ 2 ] = in_pos->b;

            /*
            // If this fails, we leak memory. However, it should not fail, so we
            // won't deal with it.
            */
            ERE(multiply_vector_and_matrix(&new_pixel_vp, pixel_vp, 
                                           change_basis_mp));

            red   = (new_pixel_vp->elements)[ 0 ];
            green = (new_pixel_vp->elements)[ 1 ];
            blue  = (new_pixel_vp->elements)[ 2 ];

            if (blue < 1.0) blue = 1.0;

            (pr_pixel_vp->elements)[ 0 ] = red / blue;
            (pr_pixel_vp->elements)[ 1 ] = green / blue;

            /*
            // If this fails, we leak memory. However, it should not fail, so we
            // won't deal with it.
            */
            ERE(multiply_vector_and_matrix(&pr_new_pixel_vp, pr_pixel_vp, 
                                           map_mp));

            (pixel_vp->elements)[ 0 ] = (pr_new_pixel_vp->elements)[ 0 ];
            (pixel_vp->elements)[ 1 ] = (pr_new_pixel_vp->elements)[ 1 ];
            (pixel_vp->elements)[ 2 ] = 1.0;

            /*
            // If this fails, we leak memory. However, it should not fail, so we
            // won't deal with it.
            */
            ERE(multiply_vector_and_matrix(&new_pixel_vp, pixel_vp,
                                           inv_change_basis_mp));

            out_pos->r = factor * (new_pixel_vp->elements)[ 0 ];
            out_pos->g = factor * (new_pixel_vp->elements)[ 1 ];
            out_pos->b = factor * (new_pixel_vp->elements)[ 2 ];

            out_pos->invalid = in_pos->invalid; 

            out_pos->invalid.r |= out_pos->invalid.b;
            out_pos->invalid.g |= out_pos->invalid.b;

            if (out_pos->r < FLT_ZERO) 
            {
                out_pos->r = FLT_ZERO; 
                out_pos->invalid.r |= TRUNCATED_PIXEL;
            }

            if (out_pos->g < FLT_ZERO) 
            {
                out_pos->g = FLT_ZERO; 
                out_pos->invalid.g |= TRUNCATED_PIXEL;
            }

            if (out_pos->b < FLT_ZERO) 
            {
                out_pos->b = FLT_ZERO; 
                out_pos->invalid.b |= TRUNCATED_PIXEL;
            }

            out_pos->invalid.pixel = 
                 out_pos->invalid.r | out_pos->invalid.g | out_pos->invalid.b;

            in_pos++;
            out_pos++;
        }
    }

    free_vector(pr_pixel_vp);
    free_vector(pr_new_pixel_vp);
    free_vector(pixel_vp);
    free_vector(new_pixel_vp);
    free_matrix(inv_change_basis_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

