
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "i/i_gen_f.h"     /* Only safe as first include in a ".c" file. */ 
#include "i/i_matrix.h"

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                             rgb_matrix_vector_to_image
 *                             
 * Converts an array of 3 matrices to an image
 *
 * This routine makes an image from an array of 3 matrices, one each for R, G, 
 * and B. The result is put into (*out_ipp) which is created if it is NULL,
 * and resized if it is the wrong size. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an appropriate error
 *     message being set.
 *
 * Index: images, matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int rgb_matrix_vector_to_image(Matrix** mp_list, KJB_image** out_ipp)
{
    KJB_image* out_ip;
    int        num_rows, num_cols, i, j;
    Pixel*     out_ip_row_pos;
    double*    r_pos;
    double*    g_pos;
    double*    b_pos;

    num_rows = mp_list[ 0 ]->num_rows; 
    num_cols = mp_list[ 0 ]->num_cols; 

    if (    (num_rows < 1) 
         || (num_cols < 1)
         || (mp_list[ 1 ]->num_rows != num_rows)
         || (mp_list[ 2 ]->num_rows != num_rows)
         || (mp_list[ 1 ]->num_cols != num_cols)
         || (mp_list[ 2 ]->num_cols != num_cols)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    for (i=0; i<num_rows; i++)
    {
        out_ip_row_pos = out_ip->pixels[ i ];

        r_pos = mp_list[ 0 ]->elements[ i ];
        g_pos = mp_list[ 1 ]->elements[ i ];
        b_pos = mp_list[ 2 ]->elements[ i ];

        for (j=0; j<num_rows; j++)
        {
            out_ip_row_pos->r = *r_pos++;
            out_ip_row_pos->g = *g_pos++; 
            out_ip_row_pos->b = *b_pos++; 

            out_ip_row_pos->invalid.r     = VALID_PIXEL;
            out_ip_row_pos->invalid.g     = VALID_PIXEL;
            out_ip_row_pos->invalid.b     = VALID_PIXEL;
            out_ip_row_pos->invalid.pixel = VALID_PIXEL;

            out_ip_row_pos++;
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             image_to_rgb_matrix_vector
 *                             
 * Converts an image to an array of 3 matrices
 *
 * This routine converts an image to an array of 3 matrices, one each for R, G,
 * and B. If any of the matrices are NULL, then they are created. Also, if any
 * are the wrong size, they are resized. Otherwise, the storage is recycled as
 * is.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an appropriate error
 *     message being set.
 *
 * Index: images, matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int image_to_rgb_matrix_vector(const KJB_image* ip, Matrix** mp_list)
{
    Matrix** mp_list_pos;
    int      i, j, num_rows, num_cols;
    Pixel*   ip_row_pos;
    double*  r_pos;
    double*  g_pos;
    double*  b_pos;

    num_rows = ip->num_rows;
    num_cols = ip->num_cols;

    ASSERT(num_rows > 0);
    ASSERT(num_cols > 0);

    mp_list_pos = mp_list;

    /*
    // FIX ( Slight possibility of memory leak on failure. 
    */
    for (i=0; i<3; i++)
    {
        ERE(get_target_matrix(mp_list_pos, num_rows, num_cols));
        mp_list_pos++; 
    }

    for (i=0; i<num_rows; i++)
    {
        ip_row_pos = ip->pixels[ i ];

        r_pos = mp_list[ 0 ]->elements[ i ];
        g_pos = mp_list[ 1 ]->elements[ i ];
        b_pos = mp_list[ 2 ]->elements[ i ];

        for (j=0; j<num_cols; j++)
        {
            *r_pos++ = ip_row_pos->r;
            *g_pos++ = ip_row_pos->g;
            *b_pos++ = ip_row_pos->b;

            ip_row_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int rgb_matrices_to_image
(
    Matrix*     r_mp,
    Matrix*     g_mp,
    Matrix*     b_mp,
    KJB_image** out_ipp 
)
{
    KJB_image* out_ip;
    int        num_rows, num_cols, i, j;
    Pixel*     out_ip_row_pos;
    double*    r_pos;
    double*    g_pos;
    double*    b_pos;

    num_rows = r_mp->num_rows; 
    num_cols = r_mp->num_cols; 

    if (    (num_rows < 1) 
         || (num_cols < 1)
         || (g_mp->num_rows != num_rows)
         || (g_mp->num_rows != num_rows)
         || (b_mp->num_cols != num_cols)
         || (b_mp->num_cols != num_cols)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    ERE(get_target_image(out_ipp, num_rows, num_cols)); 
    out_ip = *out_ipp; 

    for (i=0; i < num_rows; i++)
    {
        out_ip_row_pos = out_ip->pixels[ i ];

        r_pos = r_mp->elements[ i ];
        g_pos = g_mp->elements[ i ];
        b_pos = b_mp->elements[ i ];

        for (j=0; j<num_rows; j++)
        {
            out_ip_row_pos->r = *r_pos++;
            out_ip_row_pos->g = *g_pos++; 
            out_ip_row_pos->b = *b_pos++; 

            out_ip_row_pos->invalid.r     = VALID_PIXEL;
            out_ip_row_pos->invalid.g     = VALID_PIXEL;
            out_ip_row_pos->invalid.b     = VALID_PIXEL;
            out_ip_row_pos->invalid.pixel = VALID_PIXEL;

            out_ip_row_pos++;
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int image_to_rgb_matrices
(
    const KJB_image* ip,
    Matrix**         r_mpp,
    Matrix**         g_mpp,
    Matrix**         b_mpp 
)
{
    Pixel*  ip_row_pos;
    double* r_pos;
    double* g_pos;
    double* b_pos;
    int     i, j, num_rows, num_cols;

    num_rows = ip->num_rows;
    num_cols = ip->num_cols;

    ASSERT(num_rows > 0);
    ASSERT(num_cols > 0);

    /*
    // FIX
    // 
    // Can leak memory if one of the matrices is NULL, and a subsequent
    // get_target_matrix fails. 
    */
    ERE(get_target_matrix(r_mpp, num_rows, num_cols));
    ERE(get_target_matrix(g_mpp, num_rows, num_cols));
    ERE(get_target_matrix(b_mpp, num_rows, num_cols));

    for (i=0; i<num_rows; i++)
    {
        ip_row_pos = ip->pixels[ i ];

        r_pos = (*r_mpp)->elements[ i ];
        g_pos = (*g_mpp)->elements[ i ];
        b_pos = (*b_mpp)->elements[ i ];

        for (j=0; j<num_cols; j++)
        {
            *r_pos++ = ip_row_pos->r;
            *g_pos++ = ip_row_pos->g;
            *b_pos++ = ip_row_pos->b;

            ip_row_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_rgb_matrices(Matrix* r_mp, Matrix* g_mp, Matrix* b_mp)
{

    free_matrix(r_mp);
    free_matrix(g_mp);
    free_matrix(b_mp);

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_rgb_matrix_vector(Matrix** mp_list)
{

    free_matrix(mp_list[ 0 ]);
    free_matrix(mp_list[ 1 ]);
    free_matrix(mp_list[ 2 ]);

}

