
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"     /* Only safe as first include in a ".c" file. */
#include "l/l_string.h"
#include "l/l_sys_str.h"
#include "l/l_sys_scan.h"
#include "l/l_debug.h"

/* -------------------------------------------------------------------------- */

#define NUM_BYTES_PER_LINE  32

/* -------------------------------------------------------------------------- */

int set_debug_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_int_value;
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "debug")
         || match_pattern(lc_option, "debug-level")
       )
    {
        IMPORT int kjb_debug_level;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Debug level is %d.\n", kjb_debug_level)); 
        }
        else
        {
            ERE(ss1i(value, &temp_int_value));
            kjb_debug_level = temp_int_value; 
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  hex_print
 *
 * Prints memory in hex. 
 * 
 * Prints "len" bytes in the region pointed to by "input_reg_ptr" on the file
 * pointed to by "fp" in HEX. Generally only used for debugging.
 *
 * Index: debugging 
 *
 * -----------------------------------------------------------------------------
 */

void hex_print(FILE* fp, void* input_reg_ptr, size_t len)
{
    size_t num_lines;
    size_t i;
    size_t j;
    size_t num_int_blocks_left;
    uint*  reg_ptr;

    if (len == 0) return;

    reg_ptr = (uint*)input_reg_ptr;

    num_lines = len/NUM_BYTES_PER_LINE; 

    for (i=0;i<num_lines; ++i) 
    {
        for (j=0; j< NUM_BYTES_PER_LINE/sizeof(int); ++j) 
        {
            kjb_fprintf(fp, " %08x", *reg_ptr); 
            reg_ptr++;
        }
        kjb_fprintf(fp, "\n");
    }

    num_int_blocks_left = (len % NUM_BYTES_PER_LINE) / sizeof(int);

    for (j=0; j<num_int_blocks_left; ++j) 
    {
        kjb_fprintf(fp, " %08x", *reg_ptr); 
        reg_ptr++;
    }
    kjb_fprintf(fp, "\n");
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                             get_status_string
 * -----------------------------------------------------------------------------
*/

void get_status_string(int status_num, char* buff, size_t buff_size)
{

    switch (status_num)
    {
        case NO_ERROR:
            kjb_strncpy(buff, "NO_ERROR", buff_size);
            break;
        case EOF: 
            kjb_strncpy(buff, "EOF", buff_size);             
            break;
        case INTERRUPTED:
            kjb_strncpy(buff, "INTERRUPTED", buff_size);
            break;
        case TRUNCATED:
            kjb_strncpy(buff, "TRUNCATED", buff_size);
            break;
        case WOULD_BLOCK:
            kjb_strncpy(buff, "WOULD_BLOCK", buff_size);
            break;
        case ERROR:
            kjb_strncpy(buff, "ERROR", buff_size);
            break;
        case FOUND:
            kjb_strncpy(buff, "FOUND", buff_size);
            break;
        case NOT_FOUND:
            kjb_strncpy(buff, "NOT_FOUND", buff_size);
            break;
        case NO_SOLUTION:
            kjb_strncpy(buff, "NO_SOLUTION", buff_size);
            break;
        case NOT_SET:
            kjb_strncpy(buff, "NOT_SET", buff_size);
            break;
        case BACKGROUND_READ:
            kjb_strncpy(buff, "BACKGROUND_READ", buff_size);
            break;
        case PROCESS_IS_DEAD: 
            kjb_strncpy(buff, "PROCESS_IS_DEAD", buff_size); 
            break;
        default: 
            kjb_sprintf(buff, buff_size, "%d", status_num);  
            break;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                 test_pso  
 *
 * Printing to stdout with <<TEST>> prefix
 *
 * This routine writes a formatted string to stdout with each lined prefixed
 * with <<TEST>>. Normally this routine is only used indirectly using the macro
 * TEST_PSO. Other behaviour is similar to pso. Specifically, it is similar to
 * kjb_fprintf with respect to extended formatting options and paging. 
 *
 * Macros: 
 *    This routine is normally accessed using the macro TEST_PSO which sets
 *    things up so that the printing only occurrs if TEST is defined. If TEST
 *    is not defined, then TEST_PSO becomes nothing (IE, it is removed by the
 *    pre-processor). TEST_PSO is used as follows: 
 * |    TEST_PSO(( <format>, arg1, arg2, ....)); // Note DOUBLE parens!
 *
 * Returns:
 *    Return is void, as it is normally used with the TEST_PSO macro. 
 *
 * See also: 
 *    pso, kjb_fprintf
 *
 * Index: debugging, TEST
 *
 * -----------------------------------------------------------------------------
*/

/*PRINTFLIKE1*/
void test_pso(const char* format_str, ...)
{
    static int begining_of_line   = TRUE; 
    int        result;
    va_list    ap;
    char       buff[ 100000 ];
    char*      buff_pos; 

    va_start(ap, format_str); 

    result = kjb_vsprintf(buff, sizeof(buff), format_str, ap); 

    va_end(ap);

    if (result == ERROR)
    {
        kjb_print_error();
        set_bug("Error processing TEST error message in test_pso");
        return;
    }

    buff_pos = buff;

    /*CONSTCOND*/
    while (TRUE)
    {
        if (begining_of_line)
        {
            pso("<<TEST>> ");
            begining_of_line = FALSE;
        }

        while (*buff_pos != '\n')
        {
            if (*buff_pos == '\0') break;

            kjb_putc(*buff_pos);
            result++; 
            buff_pos++; 
        }

        if (*buff_pos == '\0') break;

        kjb_putc(*buff_pos);
        buff_pos++;
        begining_of_line = TRUE;
        result++;

        if (*buff_pos == '\0') break;
    }

    kjb_flush();
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

