
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "l/l_2D_arr.h" 
#include "l/l_int_matrix.h" 
#include "l/l_int_vector.h" 

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                           same_int_matrix_dimensions                        
 *                               
 * Returns TRUE if two matrices have the same dimensions
 *
 * This routine returns TRUE if two matrices have the same dimensions and FALSE
 * otherwise. If either matrix is NULL, then TRUE is returned if both are NULL,
 * and FALSE otherwise.
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int same_int_matrix_dimensions(const Int_matrix* first_mp, 
                               const Int_matrix* second_mp)
{
    if ((first_mp == NULL) || (second_mp == NULL))
    {
        return (first_mp == second_mp);
    }
    else if ((first_mp->num_rows != second_mp->num_rows))
    {
        return FALSE;
    }
    else if ((first_mp->num_cols != second_mp->num_cols))
    {
        return FALSE;
    }
    else
    {
        return TRUE;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                           get_zero_int_matrix                        
 *                               
 * Gets a zero integer matrix
 *
 * This routine is identical to get_target_int_matrix(), except that the
 * int_matrix is set to zero (regardless of whether it is created or reused). 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int get_zero_int_matrix(Int_matrix** target_mpp, int num_rows, int num_cols)
{

    ERE(get_target_int_matrix(target_mpp, num_rows, num_cols));
    
    return ow_zero_int_matrix(*target_mpp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         get_initialized_int_matrix                        
 *                               
 * Gets an initialized target int_matrix
 *
 * This routine is identical to get_target_int_matrix(), except that the
 * int_matrix elements are set to a specified initial value (regardless of
 * whether it is created or reused).
 *
 * Index : integer matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int get_initialized_int_matrix(Int_matrix** target_mpp, 
                               int num_rows, int num_cols, int initial_value)
{

    ERE(get_target_int_matrix(target_mpp, num_rows, num_cols));
    
    return ow_set_int_matrix(*target_mpp, initial_value);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_target_int_matrix                        
 *                               
 * Gets target int_matrix
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of integer matrices. If *target_mpp is NULL, then this
 * routine creates the int_matrix. If it is not null, and it is the right size,
 * then this routine does nothing. If it is the wrong size, then it is resized.
 *
 * In order to be memory efficient, we free before resizing. If the resizing
 * fails, then the original contents of the *target_mpp will be lost.
 * However, target_mpp->elements will be set to NULL, so *target_mpp can
 * be safely sent to free_matrix(). 
 *
 * See also:
 *    Int_matrix
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/ 

#ifdef TRACK_MEMORY_ALLOCATION

int debug_get_target_int_matrix(Int_matrix** target_mpp, 
                                int num_rows, int num_cols, 
                                const char* file_name, int line_number)
{
    Int_matrix* out_mp = *target_mpp; 

    if ((num_rows < 0) || (num_cols < 0))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if ((num_rows == 0) || (num_cols == 0))
    {
    }

    if (out_mp == NULL)
    {
        NRE(out_mp = debug_create_int_matrix(num_rows, num_cols, file_name, 
                                             line_number));
        *target_mpp = out_mp; 
    }
    else if ((out_mp->num_rows == num_rows) && (out_mp->num_cols == num_cols))
    {
        /*EMPTY*/
        ; /* Do nothing */
    }
    else
    {
        free_2D_int_array(out_mp->elements);

        if ((num_cols > 0) && (num_rows > 0))
        {
            NRE((out_mp->elements = debug_allocate_2D_int_array(num_rows,
                                                                num_cols,
                                                                file_name, 
                                                                line_number)));
        }
        else
        {
            out_mp->elements = NULL; 
        }

        out_mp->num_rows = num_rows;
        out_mp->num_cols = num_cols; 
    }

    return NO_ERROR; 
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

int get_target_int_matrix(Int_matrix** target_mpp, int num_rows, int num_cols)
{
    Int_matrix* out_mp = *target_mpp; 

    if ((num_rows < 0) || (num_cols < 0))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (out_mp == NULL)
    {
        NRE(out_mp = create_int_matrix(num_rows, num_cols));
        *target_mpp = out_mp; 
    }
    else if ((out_mp->num_rows == num_rows) && (out_mp->num_cols == num_cols))
    {
        /*EMPTY*/
        ; /* Do nothing */
    }
    else
    {
        free_2D_int_array(out_mp->elements);

        if ((num_cols > 0) && (num_rows > 0))
        {
            NRE((out_mp->elements = allocate_2D_int_array(num_rows, num_cols)));
        }
        else
        {
            out_mp->elements = NULL; 
        }

        out_mp->num_rows = num_rows;
        out_mp->num_cols = num_cols; 
    }

    return NO_ERROR; 
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_int_matrix 
 *                               
 * Creates a int_matrix 
 *
 * This routine creates a int_matrix and returns a pointer P to it. The number
 * of rows is P->num_rows, the number of columns is P->num_cols, and the
 * elements (of type "int") may be referenced by (P->elements)[ row ][ col ]. 
 *
 * The documenation for the Int_matrix type (see "Int_matrix") should be
 * consulted for CRITICAL information on the internal structure of the returned
 * int_matrix.
 * 
 * If TEST is #defined (unix only) then this routine is #define'd to be
 * debug_create_int_matrix, which is the version available in the developement
 * library. In developement code memory is tracked so that memory leaks can be
 * found more easily. Furthermore, all memory freed is checked that it was
 * allocated by an L library routine. 
 *
 * The routine free_int_matrix should be used to dispose of the storage once it
 * is no longer needed.
 *
 * Warning:
 *    The user of this library must take care with assumptions about the
 *    structure of the array embedded in the int_matrix type. See the
 *    documentation for "Int_matrix" for more details. 
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    One success it returns a pointer to the int_matrix.
 *
 * See also:
 *    Int_matrix
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

Int_matrix* debug_create_int_matrix(int num_rows, int num_cols, 
                                    const char* file_name, int line_number)
{
    Int_matrix* mp;
       
    if ((num_rows < 0) || (num_cols < 0))
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }

    if ((num_rows == 0) || (num_cols == 0))
    {
    }

    /* 
    //  Use debug_kjb_malloc as opposed to macros to pass down file_name
    //  and line_number. 
    */

    NRN(mp = DEBUG_TYPE_MALLOC(Int_matrix, file_name, line_number));

    if ((num_rows > 0) && (num_cols > 0))
    {
        mp->elements = debug_allocate_2D_int_array(num_rows, num_cols,
                                                   file_name, 
                                                   line_number);
        if (mp->elements == NULL) 
        {
            free_int_matrix(mp);
            return NULL;
        }
    }
    else 
    {
        mp->elements = NULL; 
    }
 
    mp->num_rows = num_rows;
    mp->num_cols = num_cols;
    
    return mp;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

Int_matrix* create_int_matrix(int num_rows, int num_cols)
{
    Int_matrix* mp;
       
    NRN(mp = TYPE_MALLOC(Int_matrix));
 
    if ((num_rows > 0) && (num_cols > 0))
    {
        mp->elements = allocate_2D_int_array(num_rows, num_cols);

        if (mp->elements == NULL) 
        {
            free_int_matrix(mp);
            return NULL;
        }
    }
    else 
    {
        mp->elements = NULL; 
    }
 
    mp->num_rows = num_rows;
    mp->num_cols = num_cols;
    
    return mp;
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_int_matrix
 *                               
 * Frees a the space associated with a int_matrix.
 *
 * This routine frees the storage associated with a int_matrix obtained from
 * create_vector (perhaps indirectrly). If the argument is NULL, then this
 * routine returns safely without doing anything.
 *
 * See also:
 *    Int_matrix
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/

void free_int_matrix(Int_matrix* mp)
{

    if (mp != NULL) 
    {
        if (mp->elements != NULL) 
        {
#ifdef DISABLE
#ifdef TRACK_MEMORY_ALLOCATION
            check_initialization(mp->elements[ 0 ], mp->num_rows * mp->num_cols,
                                 sizeof(int)); 
#endif 
#endif 

            free_2D_int_array(mp->elements); 
        }
      
        kjb_free(mp);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              create_zero_int_matrix
 *
 * Creates a int_matrix with all elements zero
 *
 * This routine returns a pointer to a zero int_matrix of the requested size. 
 *
 * Returns:
 *     A  pointer to the int_matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
 */

Int_matrix* create_zero_int_matrix(int num_rows, int num_cols)
{
    Int_matrix* output_mp; 

    NRN(output_mp = create_int_matrix(num_rows, num_cols)); 

    if (ow_zero_int_matrix(output_mp) == ERROR)
    {
        free_int_matrix(output_mp); 
        return NULL;
    }

    return output_mp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  ow_zero_int_matrix
 *
 * Sets all elements of a int_matrix to zero
 *
 * This routine sets all elements of a int_matrix to zero. The int_matrix must
 * have already been created. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure. Failure is only possible if the
 *     int_matrix is NULL or malformed, and this is treated as a bug. 
 *
 * Index : integer matrices
 *
 * -----------------------------------------------------------------------------
 */

int ow_zero_int_matrix(Int_matrix* mp)
{

    return ow_set_int_matrix(mp, 0); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  ow_set_int_matrix
 *
 * Sets all elements of a int_matrix to the specified value
 *
 * This routine sets all elements of a int_matrix to the specified value. The
 * int_matrix must have already been created. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure. Failure is only possible if the
 *     int_matrix is NULL or malformed, and this is treated as a bug. 
 *
 * Index : integer matrices
 *
 * -----------------------------------------------------------------------------
 */

int ow_set_int_matrix(Int_matrix* mp, int set_val)
{
    int   i;
    int   j;
    int* row_pos;

    if (mp->num_cols == 0) return NO_ERROR; 

    for (i=0; i<mp->num_rows; i++) 
    {
        row_pos = (mp->elements)[ i ]; 

        for (j=0; j<mp->num_cols; j++) 
        {
            *row_pos++ = set_val; 
        }
    }
 
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                     get_target_int_matrix_vector
 *                               
 * Gets a target integer matrix vector
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of matrix vectors. If *target_mvpp is
 * NULL, then this routine creates the matrix matrix. If it is not null, and it
 * is the right size, then this routine does nothing. If it is the wrong size,
 * then it is resized. 
 *
 * The routine free_matrix_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

int get_target_int_matrix_vector(Int_matrix_vector **mvpp, int length)
{

    if (    (*mvpp != NULL)
         && ((*mvpp)->length == length) 
       )
    {
        return NO_ERROR; 
    }

    free_int_matrix_vector(*mvpp);
    NRE(*mvpp = create_int_matrix_vector(length));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Int_matrix_vector* create_int_matrix_vector(int length)
{
    Int_matrix_vector* int_mvp;
    int           i;

    NRN(int_mvp = TYPE_MALLOC(Int_matrix_vector));
    int_mvp->length = length; 
    NRN(int_mvp->elements = N_TYPE_MALLOC(Int_matrix*, length));

    for (i=0; i<length; i++)
    {
        int_mvp->elements[ i ] = NULL;
    }

    return int_mvp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_int_matrix_vector(Int_matrix_vector *int_mvp)
{
    int         count;
    int         i;
    Int_matrix** mp_array_pos;

    if (int_mvp == NULL) return;

    mp_array_pos = int_mvp->elements;
    count = int_mvp->length; 

    for (i=0; i<count; i++) 
    {
        free_int_matrix(*mp_array_pos);
        mp_array_pos++;
    }

    kjb_free(int_mvp->elements);
    kjb_free(int_mvp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Int_matrix_vector** create_int_matrix_vector_list(int count)
{
    Int_matrix_vector** int_matrix_vector_list;
    int            i;

    NRN(int_matrix_vector_list = N_TYPE_MALLOC(Int_matrix_vector*, count));

    for (i=0; i<count; i++)
    {
        int_matrix_vector_list[ i ] = NULL;
    }

    return int_matrix_vector_list;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_int_matrix_vector_list(int count, 
                                 Int_matrix_vector **int_matrix_vector_list)
{
    int i;

    if (int_matrix_vector_list == NULL) return;

    for (i=0; i<count; i++) 
    {
        free_int_matrix_vector(int_matrix_vector_list[ i ]);
    }

    kjb_free(int_matrix_vector_list);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                        concat_int_matrices_vertically 
 *
 * Concatenates integer matrices verically                             
 *
 * This routine takes an array of integer matrices and forms one int_matrix
 * consisting of those integer matrices stacked on top of each other. 
 *
 * The first argument is the adress of the target int_matrix. If the target
 * int_matrix itself is null, then a int_matrix of the appropriate size is
 * created. If the target int_matrix is the wrong size, it is resized. Finally,
 * if it is the right size, then the storage is recycled, as is. 
 *
 * The int_matrix array may contain any number of NULL integer matrices. If there
 * are only NULL integer matrices, or if num_integer_matrices is zero, then the
 * target int_matrix is freed and set to NULL. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int concat_int_matrices_vertically(Int_matrix** mpp, int num_int_matrices, 
                                   const Int_matrix** int_matrix_list)
{
    int i, j, k, count;
    int num_rows = 0;
    int num_cols = NOT_SET; 
    Int_matrix* mp; 

    if (num_int_matrices < 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (num_int_matrices == 0)
    {
        free_int_matrix(*mpp);
        *mpp = NULL;
        return NO_ERROR;
    }

    for (i=0; i<num_int_matrices; i++)
    {
        if (int_matrix_list[ i ] != NULL)
        {
            if (num_cols == NOT_SET)
            {
                num_cols = (int_matrix_list[ i ])->num_cols;
            }
            else if ((int_matrix_list[ i ])->num_cols != num_cols)
            {
                SET_ARGUMENT_BUG();
                return ERROR;
            }
            num_rows += ((int_matrix_list[ i ])->num_rows);
        }
    }

    if (num_rows == 0)
    {
        free_int_matrix(*mpp);
        *mpp = NULL;
        return NO_ERROR;
    }

    ERE(get_target_int_matrix(mpp, num_rows, num_cols));
    mp = *mpp; 

    count = 0;

    while (count < num_rows)
    {
        for (k=0; k<num_int_matrices; k++)
        {
            if (int_matrix_list[ k ] != NULL)
            {
                for (i=0; i<(int_matrix_list[ k ])->num_rows; i++)
                {
                    for (j=0; j<num_cols; j++)
                    {
                        mp->elements[ count ][ j ] = 
                                     (int_matrix_list[ k ])->elements[ i ][ j ];
                    }
                    count++;
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                               create_int_matrix_copy
 *
 * Creates a copy of a integer array
 *
 * This routine creates a copy of a integer array and returns a pointer to it. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then NULL is returned
 *    with and error message being set. Otherwise a pointer to a newly created
 *    copy of the input integer array is returned. 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/

Int_matrix* create_int_matrix_copy(const Int_matrix* source_mp)
{
    Int_matrix* target_mp = NULL;

    ERN(copy_int_matrix(&target_mp, source_mp));

    return target_mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                copy_int_matrix
 *                               
 * Copies a integer array
 *
 * This routine copies the integer array into the one pointed to by
 * target_mpp. If the integer array pointed to by target_mpp is NULL,
 * then a integer array of the appropriate size is created. If it exists, but
 * is the wrong size, then it is recycled. Otherwise, the storage is recycled. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

int debug_copy_int_matrix(Int_matrix** target_mpp, const Int_matrix* source_mp,
                          const char* file_name, int line_number)
{
    int   i;
    int   j;
    int   num_rows; 
    int   num_cols; 
    int* target_pos;
    int* source_pos;

    if (source_mp == NULL) 
    {
        free_int_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(debug_get_target_int_matrix(target_mpp, num_rows, num_cols, file_name, 
                                line_number));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos++ = *source_pos++;
        }
    }

    return NO_ERROR;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

int copy_int_matrix(Int_matrix** target_mpp, const Int_matrix* source_mp)
{
    int   i;
    int   j;
    int   num_rows; 
    int   num_cols; 
    int* target_pos;
    int* source_pos;

    if (source_mp == NULL) 
    {
        free_int_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_int_matrix(target_mpp, num_rows, num_cols));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos++ = *source_pos++;
        }
    }

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            ow_copy_int_matrix
 *                               
 * Copies an integer matrix to an offset position of a target integer matrix
 *
 * This routine copies source_mp into target_mp, offset by (target_row_offset,
 * target_col_offset). 
 *
 * Returns :
 *    If the routine fails (can really happen unless there is an argument
 *    error), then ERROR is returned with and error message being set.
 *    Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int ow_copy_int_matrix(Int_matrix* target_mp, 
                       int target_row_offset, int target_col_offset, 
                       const Int_matrix* source_mp)
{
    int  i, j, num_rows, num_cols;
    int* target_pos;
    int* source_pos;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    if (target_row_offset + num_rows > target_mp->num_rows)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (target_col_offset + num_cols > target_mp->num_cols)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        source_pos = source_mp->elements[ i ];
        target_pos = target_mp->elements[ target_row_offset + i ] + target_col_offset;
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            copy_int_matrix_block
 *                               
 * Copies a matrix block to target matrix
 *
 * This routine copies a block of source_mp into *target_mpp which is created or
 * resized if necessary.
 *
 * Returns :
 *    If the routine fails then ERROR is returned with and error message being
 *    set.  Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int copy_int_matrix_block
(
    Int_matrix**      target_mpp,
    const Int_matrix* source_mp,
    int               source_row_offset,
    int               source_col_offset,
    int               num_rows,
    int               num_cols 
)
{

    ERE(get_target_int_matrix(target_mpp, num_rows, num_cols));

    ERE(ow_copy_int_matrix_block(*target_mpp, 0, 0, 
                                 source_mp, 
                                 source_row_offset, source_col_offset,
                                 num_rows, num_cols));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            ow_copy_int_matrix_block
 *                               
 * Copies a integer matrix block to an offset position of a target integer matrix
 *
 * This routine copies a block of source_mp into target_mp, offset by
 * (target_row_offset, target_col_offset). 
 *
 * Returns :
 *    If the routine fails (can really happen unless there is an argument
 *    error), then ERROR is returned with and error message being set.
 *    Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int ow_copy_int_matrix_block(Int_matrix* target_mp, 
                             int target_row_offset, int target_col_offset, 
                             const Int_matrix* source_mp, int source_row_offset, 
                             int source_col_offset, int num_rows, int num_cols)
{
    int i, j;
    int* target_pos;
    int* source_pos;

    if (    (target_row_offset > 0)
         || (source_row_offset > 0)
       )
    {
    }

    if (    (target_row_offset + num_rows > target_mp->num_rows)
         || (target_col_offset + num_cols > target_mp->num_cols)
         || (source_row_offset + num_rows > source_mp->num_rows)
         || (source_col_offset + num_cols > source_mp->num_cols)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        source_pos = source_mp->elements[ source_row_offset + i ] + source_col_offset;
        target_pos = target_mp->elements[ target_row_offset + i ] + target_col_offset;
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int copy_int_matrix_vector(Int_matrix_vector **out_int_mvpp, 
                           const Int_matrix_vector *in_int_mvp)
{
    Int_matrix** out_mvp;
    Int_matrix** in_mvp;
    int         num_int_matrices;
    int         i;

    if (in_int_mvp == NULL)
    {
        free_int_matrix_vector(*out_int_mvpp);
        *out_int_mvpp = NULL;
        return NO_ERROR;
    }

    num_int_matrices = in_int_mvp->length;

    if (    (*out_int_mvpp != NULL)
         && ((*out_int_mvpp)->length != num_int_matrices)
       )
    {
        free_int_matrix_vector(*out_int_mvpp);
        *out_int_mvpp = NULL;
    }

    if (*out_int_mvpp == NULL)
    {
        NRE(*out_int_mvpp = create_int_matrix_vector(num_int_matrices));
    }

    out_mvp = (*out_int_mvpp)->elements;
    in_mvp  = in_int_mvp->elements;

    for (i=0; i<num_int_matrices; i++) 
    {
        ERE(copy_int_matrix(&(out_mvp[ i ]), in_mvp[ i ]));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                copy_int_matrix_row
 *                               
 * Replaces a matrix row with a row from another
 * 
 * This routine replaces row "target_row_num" of the matrix "target_mp" with the
 * contents of "source_row_num" of "source_mp". 
 * 
 * Note : 
 *     Matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int copy_int_matrix_row(Int_matrix* target_mp, int target_row_num, 
                        const Int_matrix* source_mp, int source_row_num)
{

    int len, i;
    int* target_row_pos;
    int* source_row_pos; 

    if (    (source_row_num < 0) || (source_row_num >= source_mp->num_rows)
         || (target_row_num < 0) || (target_row_num >= target_mp->num_rows)
         || (source_mp->num_cols != target_mp->num_cols)
       ) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = source_mp->num_cols; 

    source_row_pos = source_mp->elements[ source_row_num ];
    target_row_pos = target_mp->elements[ target_row_num ];

    for (i=0; i<len; i++)
    {
        *target_row_pos = *source_row_pos; 
        target_row_pos++;
        source_row_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_int_matrix_row
 *                               
 * Returns an integer matrix row
 * 
 * This routine gets row "row_num" from the integer matrix "mp" and puts it
 * into the integer vector pointed to by vpp. If *vpp is NULL, then the vector
 * is created.  If it is the wrong length, then it is resized. Otherwise, the
 * storage is reused.
 * 
 * Note : 
 *     Int_matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_int_matrix_row(Int_vector** vpp, const Int_matrix* mp, int row_num)
{
    int  len, i;
    int* row_pos;
    int* vect_pos;

    if ((row_num < 0) || (row_num >= mp->num_rows)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_cols; 

    ERE(get_target_int_vector(vpp, len)); 

    row_pos  = mp->elements[ row_num ];
    vect_pos = (*vpp)->elements;

    for (i=0; i<len; i++)
    {
        *vect_pos++ = *row_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_int_matrix_col
 * 
 * Returns an integer matrix column
 * 
 * This routine gets column "col_num" from the integer matrix "mp" and puts it
 * into the integer vector pointed to by vpp. If *vpp is NULL, then the vector
 * is created.  If it is the wrong length, then it is resized. Otherwise, the
 * storage is reused.
 * 
 * Note : 
 *     Int_matrix column counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_int_matrix_col(Int_vector** vpp, const Int_matrix* mp, int col_num)
{
    int  len, i;
    int* vect_pos;

    if ((col_num < 0) || (col_num >= mp->num_cols)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_rows; 

    ERE(get_target_int_vector(vpp, len)); 

    vect_pos = (*vpp)->elements;

    for (i=0; i<len; i++)
    {
        *vect_pos++ = mp->elements[ i ][ col_num ]; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                put_int_matrix_row
 *                               
 * Replaces an integer matrix row with the contents of an integer vector
 * 
 * This routine replaces row "row_num" of the integer matrix "mp" with the
 * contents of the integer vector vp. 
 * 
 * Note : 
 *     Int_matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int put_int_matrix_row(Int_matrix* mp, const Int_vector* vp, int row_num)
{
    int  len, i;
    int* row_pos;
    int* vect_pos;

    if (    (row_num < 0) || (row_num >= mp->num_rows)
         || (mp->num_cols != vp->length)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = vp->length; 

    row_pos = mp->elements[ row_num ];
    vect_pos = vp->elements;

    for (i=0; i<len; i++)
    {
        *row_pos = *vect_pos; 
        row_pos++;
        vect_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                put_int_matrix_col
 *                               
 * Replaces an integer matrix column with the contents of an integer vector
 * 
 * This routine replaces column "col_num" of the integer matrix "mp" with the
 * contents of integer vector vp. 
 * 
 * Note : 
 *     Int_matrix column counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int put_int_matrix_col(Int_matrix* mp, const Int_vector* vp, int col_num)
{
    int   len;
    int   i;
    int* vect_pos;

    if (    (col_num < 0) || (col_num >= mp->num_cols)
         || (mp->num_rows != vp->length)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = vp->length; 

    vect_pos = vp->elements;

    for (i=0; i<len; i++)
    {
        mp->elements[ i ][ col_num ] = *vect_pos; 
        vect_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              get_int_transpose
 *
 * Gets the transposes an integer matrix
 *
 * This routine produces an integer matrix which is the transpose of the
 * argument.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then an integer matrix of the appropriate size is created.
 * If the target matrix is the wrong size, it is resized. Finally, if it is the
 * right size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

int get_int_transpose(Int_matrix** target_mpp, const Int_matrix* mp)
{
    Int_matrix* target_mp;
    int     i;
    int     j;

    ERE(get_target_int_matrix(target_mpp, mp->num_cols, mp->num_rows)); 
    target_mp = *target_mpp; 

    for (i = 0; i < mp->num_rows; i++) 
    {
        for (j = 0; j < mp->num_cols; j++) 
        {
            (target_mp->elements)[j][i] = (mp->elements)[i][j];
         }
     }
     
     return NO_ERROR; 
 }
       
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_min_int_matrix_element(const Int_matrix* mp, 
                               int* min_ptr, int* min_i_ptr, int* min_j_ptr)
{
    int    i;
    int    j;
    int   min;
    int** row_ptr;
    int*  element_ptr;
    int    min_i, min_j;

    if (    (mp == NULL)
         || (mp->num_rows == 0) 
         || (mp->num_cols == 0)
       )
     {
         /*
         // This could be made to simply set an error message and return error,
         // but for now, treat it as a bug.
         */
         SET_ARGUMENT_BUG();
         return ERROR;
     }

    min = mp->elements[0][0]; 
    min_i = 0; 
    min_j = 0; 

    row_ptr =  mp->elements;   

    for (i = 0; i<mp->num_rows; i++) 
    {
        element_ptr = *row_ptr; 

        for (j = 0; j<mp->num_cols; j++) 
        {
            if (*element_ptr < min) 
            {
                min = *element_ptr; 
                min_i = i; 
                min_j = j; 
            }
            element_ptr++;
        }
        row_ptr++; 
    }

    if (min_ptr != NULL) *min_ptr = min;
    if (min_i_ptr != NULL) *min_i_ptr = min_i;
    if (min_j_ptr != NULL) *min_j_ptr = min_j;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_max_int_matrix_element(const Int_matrix* mp, 
                               int* max_ptr, int* max_i_ptr, int* max_j_ptr)
{
    int    i;
    int    j;
    int   max;
    int** row_ptr;
    int*  element_ptr;
    int    max_i, max_j;

    if (    (mp == NULL)
         || (mp->num_rows == 0) 
         || (mp->num_cols == 0)
       )
     {
         /*
         // This could be made to simply set an error message and return error,
         // but for now, treat it as a bug.
         */
         SET_ARGUMENT_BUG();
         return ERROR;
     }

    max = mp->elements[0][0]; 
    max_i = 0; 
    max_j = 0; 

    row_ptr =  mp->elements;   

    for (i = 0; i<mp->num_rows; i++) 
    {
        element_ptr = *row_ptr; 

        for (j = 0; j<mp->num_cols; j++) 
        {
            if (*element_ptr > max) 
            {
                max = *element_ptr; 
                max_i = i; 
                max_j = j; 
            }
            element_ptr++;
        }
        row_ptr++; 
    }

    if (max_ptr != NULL) *max_ptr = max;
    if (max_i_ptr != NULL) *max_i_ptr = max_i;
    if (max_j_ptr != NULL) *max_j_ptr = max_j;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

