
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef L_INT_ARRAY_INCLUDED
#define L_INT_ARRAY_INCLUDED

#include "l/l_def.h"
#include "l/l_int_vector.h"

/* =============================================================================
 *                                   Int_matrix 
 * 
 * Int_matrix type
 *
 * This type is the int_matrix type for the KJB library and is used by many
 * routines. It stores a int_matrix with num_rows rows and num_cols cols. If "mp"
 * is a pointer to the type Int_matrix, then mp->elements[ row ][ col ] accesses
 * the element (row, col), and mp->elements[ row ] accesses the row'th row.
 * Note that counting starts at 0.
 *
 * The integer array rows may or may not be stored consecutively in memory.
 * Because of this, and because some routines may (most don't) take advantage
 * of knowlege of the internal structure for performance, it is important NOT
 * to swap rows by swapping pointers -- the elements should be copied. Even if
 * no routines make use of the internal structutre, swapping rows by swapping
 * pointers will make it so that the memory does not get free'd properly.
 *
 * Warning:
 *    As described above, the user of this library must take care with
 *    assumptions about the structure of the array embedded in the int_matrix
 *    type.
 *
 * See also:
 *    KJB_INIT(), create_int_matrix, free_int_matrix, allocate_2D_int_matrix
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/

typedef struct Int_matrix 
{   
    int   num_rows;
    int   num_cols;
    int** elements;
}   
Int_matrix;

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                   Int_matrix_vector 
 * 
 * Type for an array of integer matrices
 *
 * This type is used in the KJB library for arrays of integer matrices. 
 *
 * Index: integer matrices
 *
 * -----------------------------------------------------------------------------
*/
 
typedef struct Int_matrix_vector
{
    int          length;
    Int_matrix** elements;
}
Int_matrix_vector;

/* -------------------------------------------------------------------------- */

int same_int_matrix_dimensions
(
    const Int_matrix* first_mp,
    const Int_matrix* second_mp 
);

int get_zero_int_matrix
(
    Int_matrix** target_mpp,
    int          num_rows,
    int          num_cols 
);

int get_initialized_int_matrix
(
    Int_matrix** target_mpp,
    int          num_rows,
    int          num_cols,
    int          initial_value 
);

#ifdef TRACK_MEMORY_ALLOCATION

#   define create_int_matrix(x,y) debug_create_int_matrix(x,y, __FILE__, __LINE__)
 
#   define get_target_int_matrix(x, y, z)   \
                        debug_get_target_int_matrix(x, y, z, __FILE__, __LINE__)

    Int_matrix* debug_create_int_matrix
    (
        int         rows,
        int         cols,
        const char* file_name,
        int         line_number 
    );

    int debug_get_target_int_matrix
    (
        Int_matrix** target_mpp,
        int          num_rows,
        int          num_cols,
        const char*  file_name,
        int          line_number 
    );

#else 
    Int_matrix* create_int_matrix(int rows, int cols);

    int get_target_int_matrix
    (
        Int_matrix** target_mpp,
        int          num_rows,
        int          num_cols 
    );
#endif

void free_int_matrix(Int_matrix* mp);

Int_matrix* create_zero_int_matrix(int num_rows, int num_cols);

int ow_zero_int_matrix(Int_matrix*);

int ow_set_int_matrix(Int_matrix*, int);

int get_target_int_matrix_vector(Int_matrix_vector** mvpp, int length);

Int_matrix_vector* create_int_matrix_vector(int count);

void free_int_matrix_vector(Int_matrix_vector*);

Int_matrix_vector** create_int_matrix_vector_list(int count);

void free_int_matrix_vector_list(int count, Int_matrix_vector** int_mvpp);

int concat_int_matrices_vertically
(
    Int_matrix**      mpp,
    int               num_matrices,
    const Int_matrix* int_matrix_list[] 
);

Int_matrix* create_int_matrix_copy(const Int_matrix* source_mp);

#ifdef TRACK_MEMORY_ALLOCATION
#    define copy_int_matrix(x,y) debug_copy_int_matrix(x,y, __FILE__, __LINE__)

    int debug_copy_int_matrix
    (
        Int_matrix**      target_mpp,
        const Int_matrix* source_mp,
        const char*       file_name,
        int               line_number 
    );
#else
    int copy_int_matrix
    (
        Int_matrix**      target_mpp,
        const Int_matrix* source_mp 
    );
#endif 

int copy_int_matrix_block
(
    Int_matrix**      target_mpp,
    const Int_matrix* source_mp,
    int               source_row_offset,
    int               source_col_offset,
    int               num_rows,
    int               num_cols 
);

int ow_copy_int_matrix_block
(
    Int_matrix*       target_mp,
    int               target_row_offset,
    int               target_col_offset,
    const Int_matrix* source_mp,
    int               source_row_offset,
    int               source_col_offset,
    int               num_rows,
    int               num_cols 
);

int ow_copy_int_matrix
(
    Int_matrix*       target_mp,
    int               target_row_offset,
    int               target_col_offset,
    const Int_matrix* source_mp 
);

int copy_int_matrix_vector
(
    Int_matrix_vector**      out_int_mvpp,
    const Int_matrix_vector* in_int_mvp 
);

int copy_int_matrix_row
(
    Int_matrix*       target_mp,
    int               target_row_num,
    const Int_matrix* source_mp,
    int               source_row_num 
);

int get_int_matrix_row
(
    Int_vector**      vpp,
    const Int_matrix* mp,
    int               row_num 
);

int get_int_matrix_col
(
    Int_vector**      vpp,
    const Int_matrix* mp,
    int               col_num 
);

int put_int_matrix_row(Int_matrix* mp, const Int_vector* vp, int row_num);

int put_int_matrix_col(Int_matrix* mp, const Int_vector* vp, int col_num);

int get_int_transpose(Int_matrix** target_mpp, const Int_matrix* mp);

int get_min_int_matrix_element
(
    const Int_matrix* mp,
    int*              min_ptr,
    int*              min_i_ptr,
    int*              min_j_ptr 
);

int get_max_int_matrix_element
(
    const Int_matrix* mp,
    int*              max_ptr,
    int*              max_i_ptr,
    int*              max_j_ptr 
);

#endif

