
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 

#include "l/l_io.h" 
#include "l/l_parse.h" 
#include "l/l_sys_scan.h" 
#include "l/l_sort.h" 
#include "l/l_int_vector.h" 

/* -------------------------------------------------------------------------- */

static int descend_compare_indexed_int_vector_elements
(
    const void* first_ptr,
    const void* second_ptr 
);

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                          get_zero_int_vector                        
 *                               
 * Gets zero integer list 
 *
 * This routine is similar to get_target_int_vector(3), except that the integer
 * list elements are set to zero. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/ 

int get_zero_int_vector(Int_vector** vpp, int len)
{

    ERE(get_target_int_vector(vpp, len));

    return ow_zero_int_vector(*vpp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          get_initialized_int_vector                        
 *                               
 * Gets initialzed integer list 
 *
 * This routine is similar to get_target_int_vector(3), except that the elements
 * of the integer list are set to the specitied value. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/ 

int get_initialized_int_vector(Int_vector** vpp, int len, int initial_value)
{

    ERE(get_target_int_vector(vpp, len));
    ERE(ow_set_int_vector(*vpp, initial_value));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef FREE_BEFORE_RESIZE

/* =============================================================================
 *                                get_target_int_vector                        
 *                               
 * Gets target integer list 
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of integer list.. If *target_vpp is NULL, then this
 * routine creates the integer list. If it is not null, and it is the right
 * size, then this routine does nothing. If it is the wrong size, then it is
 * resized. 
 *
 * In order to be memory efficient, we free before resizing. If the resizing
 * fails, then the original contents of the *target_vpp will be lost.
 * However, (*target_vpp)->elements will be set to NULL, so (*target_vpp)
 * can be safely sent to free_int_vector(). 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/ 

#else

/* =============================================================================
 *                                get_target_int_vector                        
 *                               
 * Gets target integer list 
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of integer list.. If *target_vpp is NULL, then this
 * routine creates the integer list. If it is not null, and it is the right
 * size, then this routine does nothing. If it is the wrong size, then it is
 * resized. 
 *
 * If an actual resize is needed, then a new integer list of the required size
 * is first created. If the creation is successful, then the old integer list
 * is free'd.  The reason is that if the new allocation fails, a calling
 * application should still be able to the old integer list. The alternate is
 * to free the old integer list first.  This is more memory efficient. A more
 * sophisticated alternative is to free the old integer list if it can be
 * deterimined that the subsequent allocation will succeed. Although such
 * approaches have merit, it is expected that resizing will occur infrequently
 * enough that it is not worth implementing them. Thus the simplest method
 * with good semantics under most conditions has been used. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/ 

#endif 

#ifdef TRACK_MEMORY_ALLOCATION
int debug_get_target_int_vector(Int_vector** target_vpp, int length, 
                                const char* file_name, int line_number)
{
    Int_vector* vp = *target_vpp; 

    if (length < 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (vp == NULL)
    {
        NRE(vp = debug_create_int_vector(length, file_name, line_number));
        *target_vpp = vp; 
    }
    else if (length == vp->length)
    {
        /*EMPTY*/
        ; /* Do nothing */ 
    }
    else if (length < vp->length)
    {
        vp->length = length;
    }
    else
    {
        kjb_free(vp->elements); 

        if (length > 0)
        {
            NRE(vp->elements = DEBUG_INT_MALLOC(length, file_name, 
                                                line_number)); 
        }
        else
        {
            vp->elements = NULL; 
        }

        vp->length = length;
    }

    return NO_ERROR;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

int get_target_int_vector(Int_vector** target_vpp, int length)
{
    Int_vector* vp = *target_vpp; 

    if (length < 0)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (vp == NULL)
    {
        NRE(vp = create_int_vector(length));
        *target_vpp = vp; 
    }
    else if (length == vp->length)
    {
        /*EMPTY*/
        ; /* Do nothing */ 
    }
    else if (length < vp->length)
    {
        vp->length = length;
    }
    else
    {
        kjb_free(vp->elements); 

        if (length > 0)
        {
            NRE(vp->elements = INT_MALLOC(length));
        }
        else
        {
            vp->elements = NULL; 
        }

        vp->length = length;
    }

    return NO_ERROR;
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_int_vector
 *                               
 * Creates an integer list
 *
 * This routine creates an integer list and returns a pointer to it.
 *
 * If TEST is #defined (unix only) then this routine is #define'd to be
 * debug_create_int_vector, which is the version available in the developement
 * library. In developement code memory is tracked so that memory leaks can be
 * found more easily. Furthermore, all memory freed is checked that it was
 * allocated by an L library routine. 
 *
 * The routine free_int_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    One success it returns a pointer to the integer list.
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

Int_vector* debug_create_int_vector(int len, 
                                    const char* file_name, int line_number)
{
    Int_vector* vp;
   
    if (len < 0) 
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }

    /* 
    //  Use debug_kjb_malloc as opposed to macros to pass down file_name
    //  and line_number. 
    */
    NRN(vp = DEBUG_TYPE_MALLOC(Int_vector, file_name, line_number)); 
   
    if (len > 0)
    {
        /* 
        //  Use debug_kjb_malloc as opposed to macros to pass down file_name
        //  and line_number. 
        */
        vp->elements = DEBUG_INT_MALLOC(len, file_name, line_number);

        if (vp->elements == NULL) 
        {
            kjb_free(vp);
            return NULL;
        }   
    }
    else
    {
        vp->elements = NULL; 
    }
   
    vp->length = len;

    return vp;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

Int_vector* create_int_vector(int len)
{
    Int_vector* vp;
    
    if (len < 0) 
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }
 
    NRN(vp = TYPE_MALLOC(Int_vector));
    
    vp->elements = N_TYPE_MALLOC(int, len);

    if (len > 0)
    {
        if (vp->elements == NULL) 
        {
           kjb_free(vp);
           return NULL;
        }   
    }
    else
    {
        vp->elements = NULL; 
    }
    
    vp->length = len;
 
    return vp;
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_int_vector
 *                               
 * Frees the space associated with a integer list
 *
 * This routine frees the storage associated with an integer list obtained from
 * create_int_vector (perhaps indirectly). If the argument is NULL, then this
 * routine returns safely without doing anything.
 *
 * Acknowlegement:
 *    This routine grew out of one written by Ian Harder. However, it is 
 *    important to note that this routine is somewhat different than any
 *    of Ian's routines. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
*/

void free_int_vector(Int_vector* vp)
{

    if (vp != NULL)
    {
        if (vp->elements != NULL) 
        {
#ifdef DISABLE 
#ifdef TRACK_MEMORY_ALLOCATION

            check_initialization(vp->elements, vp->length, sizeof(int));
#endif 
#endif 
            kjb_free(vp->elements);
        }
    }

    kjb_free(vp);

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Int_vector* create_zero_int_vector(int length)
{
    Int_vector* vp;

    NRN(vp = create_int_vector(length));

    if (ow_zero_int_vector(vp) == ERROR)
    {
        free_int_vector(vp);
        return NULL;
    }

    return vp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_zero_int_vector(Int_vector* vp)
{

    return ow_set_int_vector(vp, 0);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_set_int_vector(Int_vector* vp, int set_val)
{
    int i;
    int len; 
    int* vec_pos; 

    if ((vp == NULL) || (vp->length < 0)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    len = vp->length; 
    vec_pos = vp->elements;

    for (i=0; i<len; i++) 
    {
        *vec_pos = set_val;
        vec_pos++; 
    }
 
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ascend_sort_int_vector(Int_vector* vp)
{

    ERE(int_sort(vp->elements, vp->length, sizeof(int), 0,
                 USE_CURRENT_ATN_HANDLING));

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Indexed_int_vector* create_indexed_int_vector(Int_vector* vp)
{
    Indexed_int_vector* result_ptr; 
    int len, i;

    NRN(result_ptr = TYPE_MALLOC(Indexed_int_vector));

    len = vp->length;

    result_ptr->length = len;
    result_ptr->elements = N_TYPE_MALLOC(Indexed_int_vector_element, len);

    if (result_ptr->elements == NULL)
    {
        kjb_free(result_ptr);
        return NULL;
    }

    for (i=0; i<len; i++)
    {
        result_ptr->elements[i].index = i;
        result_ptr->elements[i].element = vp->elements[ i ];
    }

    return result_ptr;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_indexed_int_vector(Indexed_int_vector *vp)
{

    if (vp != NULL)
    {
        kjb_free(vp->elements);
        kjb_free(vp);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int descend_sort_indexed_int_vector(Indexed_int_vector *vp)
{

    ERE(kjb_sort(vp->elements, vp->length, sizeof(Indexed_int_vector_element),
                 descend_compare_indexed_int_vector_elements, 
                 USE_CURRENT_ATN_HANDLING));

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int descend_compare_indexed_int_vector_elements(const void* first_ptr, 
                                                       const void* second_ptr)
{
    const Indexed_int_vector_element *elem1_ptr = (const Indexed_int_vector_element*)first_ptr;
    const Indexed_int_vector_element *elem2_ptr = (const Indexed_int_vector_element*)second_ptr;

    if (elem2_ptr->element > elem1_ptr->element)
    {
        return FIRST_ITEM_GREATER;
    }
    else if (elem2_ptr->element < elem1_ptr->element)
    {
        return SECOND_ITEM_GREATER;
    }
    else 
    {
        return EQUAL_ITEMS;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           copy_int_vector_vector
 *
 * Copies an integer vector vector
 *
 * This routine copies the integer vector pvector ointed to by source_vvp to
 * the integer vector pointed to by *target_vp. If *target_vp is NULL, then it
 * is created. If it is the wrong size, it is resized. 
 *
 * Returns:
 *    On success, this routine returns a pointer to a newly created integer
 *    list which is a copy of the input integer list. On failure, it returns
 *    NULL, and sets an error message. Currently this routine can only fail if
 *    storage allocation fails. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
 */

int copy_int_vector_vector
(
    Int_vector_vector** target_vvpp, 
    const Int_vector_vector* source_vvp
)
{
    int   i;
    int   length; 

    if (source_vvp == NULL)
    {
        free_int_vector_vector(*target_vvpp);
        *target_vvpp = NULL;
        return NO_ERROR;
    }

    length = source_vvp->length;

    ERE(get_target_int_vector_vector(target_vvpp, length));

    for (i=0; i<length; i++)
    {
        ERE(copy_int_vector(&((*target_vvpp)->elements[ i ]),
                            source_vvp->elements[ i ])); 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           create_int_vector_copy
 *
 * Creates a copy of an integer vector
 *
 * This routine creates a copy of the input integer vector, and returns a
 * pointer to it.
 *
 * Returns:
 *    On success, this routine returns a pointer to a newly created integer
 *    list which is a copy of the input integer list. On failure, it returns
 *    NULL, and sets an error message. Currently this routine can only fail if
 *    storage allocation fails. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
 */

Int_vector* create_int_vector_copy(Int_vector* source_vp)
{
    Int_vector* target_vp = NULL;

    ERN(copy_int_vector(&target_vp, source_vp));

    return target_vp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           copy_int_vector
 *
 * Copies an integer vector
 *
 * This routine copies the integer vector pointed to by source_vp to the
 * integer vector pointed to by *target_vp. If *target_vp is NULL, then it is
 * created. If it is the wrong size, it is resized. 
 *
 * Returns:
 *    On success, this routine returns a pointer to a newly created integer
 *    list which is a copy of the input integer list. On failure, it returns
 *    NULL, and sets an error message. Currently this routine can only fail if
 *    storage allocation fails. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
 */

int copy_int_vector(Int_vector** target_vpp, const Int_vector* source_vp)
{
    int   i;
    int   length; 
    int* target_pos;
    int* source_pos;

    if (source_vp == NULL)
    {
        free_int_vector(*target_vpp);
        *target_vpp = NULL;
        return NO_ERROR;
    }

    length = source_vp->length;

    ERE(get_target_int_vector(target_vpp, length));

    source_pos = source_vp->elements;
    target_pos = (*target_vpp)->elements;

    for (i=0; i<length; i++)
    {
        *target_pos++ = *source_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           copy_int_vector_section
 *
 * Copies an integer list
 *
 * This routine copies part of the integer vector pointed to by source_vp to
 * the integer vector poited to by *target_vp. "length" elements are copied
 * starting with "first_elem". If *target_vp is NULL, then it is created. If it
 * is the wrong size, it is resized. 
 *
 * Returns:
 *    On success, this routine returns a pointer to a newly created integer
 *    list which is a copy of the input integer list. On failure, it returns
 *    NULL, and sets an error message. Currently this routine can only fail if
 *    storage allocation fails. 
 *
 * Index: integer vectors
 *
 * -----------------------------------------------------------------------------
 */

int copy_int_vector_section(Int_vector** target_vpp, 
                            const Int_vector* source_vp,
                            int first_elem, int length)
{
    int   i;
    int* target_pos;
    int* source_pos;

    if (source_vp == NULL)
    {
        free_int_vector(*target_vpp);
        *target_vpp = NULL;
        return NO_ERROR;
    }

    if (    (first_elem < 0)
         || (first_elem + length > source_vp->length)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR; 
    }

    ERE(get_target_int_vector(target_vpp, length));

    source_pos = source_vp->elements + first_elem;
    target_pos = (*target_vpp)->elements;

    for (i=0; i<length; i++)
    {
        *target_pos++ = *source_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               read_int_vector
 *                    
 * Reads data from a file specified by name into aninteger vector
 * 
 * This routine reads data contained in a file specified by its name into an
 * integer vector. The length of the new vector is determined by the number of
 * data elements contained in the input file. The integer vector pointed to by
 * *vpp is created or resized as necessary.
 * 
 * "file_name" points to a character array specifying the name of the file
 * to read the data from. If the "file_name" argument is NULL, or the first
 * character is null, input is expected from STDIN instead of a file.
 * 
 * Returns:
 *     NO_ERROR on success and ERROR on failure, with an appropriate error
 *     message being set. 
 * 
 * See Also:
 *     fp_read_int_vector, fp_read_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int read_int_vector(Int_vector** vpp, const char* file_name)
{
    FILE* fp;
    int   result;

    if ((file_name == NULL) || (*file_name == '\0'))
    {
        fp = stdin;
    }
    else 
    {
        NRE(fp = kjb_fopen(file_name, "r")); 
    }

    result = fp_read_int_vector(vpp, fp);

    /*
    //  The file pointer based read routines can return EOF because they are
    //  used as building blocks. But for name based read routines, we expect at
    //  least one vector. 
    */
    if (result == EOF) 
    {
        set_error("Unable to read an integer matrix file from %F.", fp); 
        result = ERROR; 
    }

    if ((file_name != NULL) && (*file_name != '\0'))
    {
        (void)kjb_fclose(fp);  /* Ignore return--only reading. */
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                           fp_read_int_vector
 * 
 * Reads data from an input stream into a integer vector
 * 
 * This routine reads data contained in a file specified by a pointer to FILE
 * into an integer vector. The length of the new vector is determined by the
 * number of data elements contained in the input file.  The vector *vpp is
 * created or resized as necessary. 
 * 
 * "vpp" is a pointer to a integer vector pointer.
 *
 * "fp" points to a FILE object that corresponds to the stream to read from.
 * 
 * Returns:
 *     NO_ERROR on success and ERROR on failure, with an appropriate error
 *     message being set. 
 * 
 * See Also:
 *     read_int_vector, fp_read_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int fp_read_int_vector(Int_vector** vpp, FILE* fp)
{
    Int_vector* vp;
    int     num_elements;
    char    line[ LARGE_IO_BUFF_SIZE ];
    char    element_buff[ 200 ];
    char*   line_pos;
    int     i;
    int     scan_res;

    ERE(num_elements = gen_count_tokens_in_file(fp, " ,\t"));

    if (num_elements == 0)
    {
        set_error("Expecting at least one number in %F.", fp);
        return ERROR;
    }

    ERE(get_target_int_vector(vpp, num_elements)); 
    vp = *vpp; 

    i = 0; 

    while (i<vp->length) 
    {
        Size_result read_res;

        ERE(read_res = BUFF_GET_REAL_LINE(fp, line)); 

        if (read_res == EOF) 
        {
            set_error("Unexpected end of file reading vector from %F.", fp); 
            return ERROR; 
        }

        line_pos = line; 

        while (BUFF_GEN_GET_TOKEN_OK((const char**)&line_pos, element_buff, 
                                     " ,\t")) 
        {
            if (i < vp->length) 
            {
                scan_res = ss1i(element_buff, (vp->elements)+i); 

                if (scan_res == ERROR) 
                {
                    insert_error("Error reading integer from %F.",
                                 fp); 
                    return ERROR; 
                }

                i++;
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          write_col_int_vector
 *                    
 * Writes an integer vector as a column to a file specified by name
 * 
 * This routine writes data contained in a Int_vector object to a file specified
 * by its name. The vector is output as a column of data.
 * 
 * "file_name" points to a character array specifying the name of the file for
 * writing. If the "file_name" argument is NULL, or the first character is
 * null, output is directed to STDOUT.
 *
 * "vp" points to the Int_vector object whose contents are to be written.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     fp_write_col_int_vector, write_row_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int write_col_int_vector(const Int_vector* vp, const char* file_name)
{
    FILE* fp;
    int   result; 

    if (file_name == NULL) 
    {
        fp = stdout;
    }
    else 
    {
        NRE(fp = kjb_fopen(file_name, "w")); 
    }

    result = fp_write_col_int_vector(vp, fp);

    if (file_name != NULL) 
    {
        Error_action save_error_action = get_error_action();

        if (result == ERROR) 
        {
            set_error_action(FORCE_ADD_ERROR_ON_ERROR);
        }

        if (kjb_fclose(fp) == ERROR)
        {
            result = ERROR;
        }

        set_error_action(save_error_action); 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          fp_write_col_int_vector
 *                    
 * Writes an integer vector as a column to an output stream
 * 
 * This routine writes data contained in a Int_vector object to a file specified
 * by a pointer to FILE. The vector is output as a column of data.
 * 
 * "fp" points to a FILE object associated with the output stream.
 * "vp" points to the Int_vector object whose contents are to be written.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     write_col_int_vector, fp_write_row_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int fp_write_col_int_vector(const Int_vector* vp, FILE* fp)
{
    int   i;

    for (i=0; i<vp->length; i++) 
    {
        ERE(kjb_fprintf(fp, "%d\n", vp->elements[i]));
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          fp_write_col_int_vector_with_title
 *                    
 * Writes an integer vector as a column with a title
 * 
 * This routine writes data in an integer vector with a title to the file
 * pointed to by "fp". The vector is output as a column of data. This routine
 * is largely designed for debugging, and is often called with the macro dbi_cv. 
 * 
 * "fp" points to a character array specifying the name of the file to write
 * data ti. If fp is NULL, output is directed to STDOUT.  "vp" points to the
 * Int_vector object whose contents are to be written.  "title" points to a
 * title string which is printed before the vector. "title" may be NULL, in
 * which case, nothing is printed. If the vector itself is NULL, then "NULL" is
 * printed instead of any data.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     fp_write_row_int_vector_with_title, db_icv, db_irv
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int fp_write_col_int_vector_with_title(const Int_vector* vp, FILE* fp, 
                                       const char* title)
{
    int   i;

    if (vp == NULL) 
    {
        if (title != NULL) 
        {
            ERE(kjb_fprintf(fp, title));
            ERE(kjb_fprintf(fp, ": "));
        }
        ERE(kjb_fprintf(fp, "NULL\n"));

        return NO_ERROR;
     }

    if (title != NULL) 
    {
        ERE(kjb_fprintf(fp, title));
        ERE(kjb_fprintf(fp, "\n"));
    }

    for (i=0; i<vp->length; i++) 
    {
        ERE(kjb_fprintf(fp, "%d\n", vp->elements[i]));
    }

    ERE(kjb_fprintf(fp, "\n"));

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          write_row_int_vector
 *                    
 * Writes an integer vector as a rowumn to a file specified by name
 * 
 * This routine writes data contained in a Int_vector object to a file specified
 * by its name. The vector is output as a rowumn of data.
 * 
 * "file_name" points to a character array specifying the name of the file for
 * writing. If the "file_name" argument is NULL, or the first character is
 * null, output is directed to STDOUT.
 *
 * "vp" points to the Int_vector object whose contents are to be written.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     fp_write_row_int_vector, write_row_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int write_row_int_vector(const Int_vector* vp, const char* file_name)
{
    FILE* fp;
    int   result; 

    if (file_name == NULL) 
    {
        fp = stdout;
    }
    else 
    {
        NRE(fp = kjb_fopen(file_name, "w")); 
    }

    result = fp_write_row_int_vector(vp, fp);

    if (file_name != NULL) 
    {
        Error_action save_error_action = get_error_action();

        if (result == ERROR) 
        {
            set_error_action(FORCE_ADD_ERROR_ON_ERROR);
        }

        if (kjb_fclose(fp) == ERROR)
        {
            result = ERROR;
        }

        set_error_action(save_error_action); 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          fp_write_row_int_vector
 *                    
 * Writes an integer vector as a row to an output stream
 * 
 * This routine writes data contained in a Int_vector object to a file specified
 * by a pointer to FILE. The vector is output as a rowumn of data.
 * 
 * "fp" points to a FILE object associated with the output stream.
 * "vp" points to the Int_vector object whose contents are to be written.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     write_row_int_vector, fp_write_row_int_vector
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int fp_write_row_int_vector(const Int_vector* vp, FILE* fp)
{
    int   i;

    for (i=0; i<vp->length; i++) 
    {
        ERE(kjb_fprintf(fp, "%d ", vp->elements[i]));
    }

    ERE(kjb_fprintf(fp, "\n")); 

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          fp_write_row_int_vector_with_title
 *                    
 * Writes an integer vector as a row with a title
 * 
 * This routine writes data in an integer vector with a title to the file
 * pointed to by "fp". The vector is output as a row of data. This routine
 * is largely designed for debugging, and is often called with the macro dbi_cv. 
 * 
 * "fp" points to a character array specifying the name of the file to write
 * data ti. If fp is NULL, output is directed to STDOUT.  "vp" points to the
 * Int_vector object whose contents are to be written.  "title" points to a
 * title string which is printed before the vector. "title" may be NULL, in
 * which case, nothing is printed. If the vector itself is NULL, then "NULL" is
 * printed instead of any data.
 * 
 * Returns:
 *     NO_ERROR if successful, ERROR on failure.
 * 
 * See Also:
 *     fp_write_col_int_vector_with_title, db_icv, db_irv
 * 
 * Index: I/O, vectors, vector I/O
 *
 * -----------------------------------------------------------------------------
*/ 

int fp_write_row_int_vector_with_title(const Int_vector* vp, FILE* fp, 
                                    const char* title)
{
    int   i;

    if (vp == NULL) 
    {
        if (title != NULL) 
        {
            ERE(kjb_fprintf(fp, title));
            ERE(kjb_fprintf(fp, ": "));
        }
        ERE(kjb_fprintf(fp, "NULL\n"));

        return NO_ERROR;
     }

    if (title != NULL) 
    {
        ERE(kjb_fprintf(fp, title));
        ERE(kjb_fprintf(fp, "\n"));
    }

    for (i=0; i<vp->length; i++) 
    {
        if (i != 0) 
        {
            ERE(kjb_fputs(fp, "  "));
        }

        ERE(kjb_fprintf(fp, "%d", vp->elements[i]));
    }

    ERE(kjb_fprintf(fp, "\n"));

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             get_target_int_vector_vector
 *                               
 * Gets a target integer vector vector
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of integer vector arrays. If *target_vvpp is NULL, then
 * this routine creates the vector vector. If it is not null, and it is the
 * right size, then this routine does nothing. If it is the wrong size, then it
 * is resized. 
 *
 * The routine free_int_vector_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Index: memory allocation, arrays, vectors
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

int get_target_int_vector_vector(Int_vector_vector **target_vvpp, int count)
{

    free_int_vector_vector(*target_vvpp);
    NRE(*target_vvpp = create_int_vector_vector(count));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_int_vector_vector
 *                               
 * Creates an integer vector vector
 *
 * This routine creates a vector array of size "count" which is a type for a
 * list of vectors.  All vector pointers are set to NULL. 
 *
 * This routine sets all the pointers to NULL. 
 *
 * The routine free_int_vector_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Index: memory allocation, arrays, vectors
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

Int_vector_vector* create_int_vector_vector(int count)
{
    Int_vector_vector* vvp;
    int           i;

    NRN(vvp = TYPE_MALLOC(Int_vector_vector));
    vvp->length = count; 
    NRN(vvp->elements = N_TYPE_MALLOC(Int_vector*, count));

    for (i=0; i<count; i++)
    {
        vvp->elements[ i ] = NULL;
    }

    return vvp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_int_vector_vector
 *                               
 * Frees the storage in an integer vector vector
 *
 * This routine frees the storate in a vector array. 
 *
 * Index: memory allocation, arrays, vectors
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

void free_int_vector_vector(Int_vector_vector *vvp)
{
    int count, i;
    Int_vector** vp_pos;

    if (vvp == NULL) return;

    vp_pos = vvp->elements;
    count = vvp->length; 

    for (i=0; i<count; i++) 
    {
        free_int_vector(*vp_pos);
        vp_pos++;
    }

    kjb_free(vvp->elements);
    kjb_free(vvp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int concat_int_vectors(Int_vector** vpp, int num_vectors, 
                       const Int_vector** vectors)
{
    int  i;
    int  length     = 0;
    int* target_pos;

    for (i = 0; i < num_vectors; i++)
    {
        if (vectors[ i ] != NULL)
        {
            length += vectors[ i ]->length;
        }
    }

    ERE(get_target_int_vector(vpp, length));
    target_pos = (*vpp)->elements;

    for (i = 0; i < num_vectors; i++)
    {
        if (vectors[ i ] != NULL)
        {
            const Int_vector* source_vp     = vectors[ i ];
            int               source_length = source_vp->length;
            int*              source_pos    = source_vp->elements;
            int               j;

            for (j = 0; j < source_length; j++)
            {
                *target_pos = *source_pos; 
                target_pos++;
                source_pos++;
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                       int_vector_linear_search
 *
 * Finds an integer in an integer vector
 *
 * This routine finds an integer in the integer vector "vp" with a linear
 * search.
 *
 * Returns:
 *    On success, this routine returns an index into the array. If the element
 *    is not found, then it returns NOT_FOUND. 
 * 
 * -----------------------------------------------------------------------------
 */

int int_vector_linear_search(const Int_vector* vp, int search_elem)
{
    int* array = vp->elements;
    int num_elements = vp->length;

    return int_linear_search(array, num_elements, search_elem);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

