
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "l/l_gen_f.h"     /* Only safe as first include in a ".c" file. */ 
#include "l/l_string.h"
#include "l/l_parse.h"
#include "l/l_sys_scan.h"
#include "l/l_sys_str.h"
#include "l/l_verbose.h"

/* =============================================================================
 *                              kjb_sprintf                          
 *                               
 * Similar routine to sprintf, except that it checks for buffer overrun, and 
 * the extra formatting items described in kjb_fprintf are avaliable. 
 *
 * Warning:
 *    The second parameter of this routine is the size of the buffer. This is 
 *    DIFFERENT than sprintf!
 * 
 * Note:
 *    The return type "Size_result" is an int except when size_t is 16 bits, in 
 *    which case it is a long. 
 * 
 * Returns:
 *    The number of characters written to the buffer. 
 *
 * See also:
 *    kjb_fprintf
 *
 * Index: strings
 *
 * -----------------------------------------------------------------------------
*/ 

/*PRINTFLIKE3*/
Size_result kjb_sprintf(char* buff, size_t max_len, const char* format_str, ...)
{
    int     result;
    va_list ap;

    va_start(ap, format_str); 

    result = kjb_vsprintf(buff, max_len, format_str, ap);

    va_end(ap);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                                        
 *                               
 *
 *
 * -----------------------------------------------------------------------------
*/ 

/*
 *  Note: This routine should always leave buff in a reaonable state (eg
 *        possibly a null terminated string) as it may be called from
 *        kjb_error.
 *
 *  FIX   We should probably be looking at sprintf's return.
 */

Size_result kjb_vsprintf
(
    char*       buff,
    size_t      max_len, 
    const char* format_str, 
    va_list     ap
)
{
    long         save_max_len; 
    char*        new_format_str_pos; 
    int          format_item; 
    char         type_char; 
    int          num_precision_args; 
    int          precision_arg1; 
    int          precision_arg2; 
    int          short_flag; 
    int          long_flag; 
#ifdef LONG_DOUBLE_OK
    int          long_double_flag; 
#endif
    char         new_format_str[ MAX_FORMAT_STRING_SIZE ]; 
    char         trunc_buff[ 100 ]; 
    int          justify_flag; 
    int          field_width_flag; 
    int          comma_flag; 
    int          precision_flag; 
    int          field_width_arg_flag; 
    int          field_width; 
    size_t       field_size; 
    int          temp_field_width; 
    char*        buff_pos; 
    size_t       temp_buff_len; 
    char         temp_buff[ MAX_FORMAT_STRING_SIZE ]; 
    const char*  temp_buff_pos; 

    if ((buff == NULL) || (format_str == NULL) || (max_len < 1))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

#ifdef TEST
    if (max_len == 4)
    {
        warn_pso("Suspicious buffer length (4) at line %d of %s.\n",
                 __LINE__, __FILE__);
        kjb_abort(); 
    }
#endif 

    buff[ 0 ] =  '\0'; 
    buff_pos = buff; 
    max_len--;
    save_max_len = max_len; 

    while (*format_str != '\0') 
    {
        format_item = FALSE; 

        while ((!format_item) &&
               (*format_str != '\0') &&
               (max_len > 0))
        { 
            if (*format_str == '%') 
            {
                format_item = TRUE;
            }
            else 
            {
                *buff_pos++ = *format_str++;
                max_len--;
            }
        }

        *buff_pos = '\0';

        if (format_item)     /* Can't be at string end if true */ 
        {
            if (max_len == 0) 
            {
                *buff_pos = '\0'; 
                SET_BUFFER_OVERFLOW_BUG();
                return ERROR;
            }

            short_flag = FALSE;
            long_flag = FALSE;
#ifdef LONG_DOUBLE_OK
            long_double_flag = FALSE; 
#endif
            num_precision_args = 0; 
            justify_flag = field_width_flag = precision_flag = FALSE;
            comma_flag = field_width_arg_flag = FALSE;

            new_format_str_pos = new_format_str;
            *new_format_str_pos = *format_str; 
            format_str++; 
            new_format_str_pos++; 

            /* 
             * FIX: Currently we accept stuff like "-+  -  +". Not serious,
             *      as this stuff just gets passed down to sprintf anyway. 
             *      BUT we DONT look at sprintf's return! 
             */
            while (    (*format_str == '-') 
                    || (*format_str == ' ') 
                    || (*format_str == '0') 
                    || (*format_str == '#') 
                    || (*format_str == '+')  
                    || (*format_str == ',') 
                  )
            {
                if (*format_str == '-') 
                {
                    justify_flag = TRUE;
                }

                if (*format_str == ',') 
                {
                    comma_flag = TRUE;
                }
                else 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                }

                format_str++;
            }

            if (*format_str == '*') 
            {
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;

                num_precision_args++;
                field_width_flag = TRUE; 
                field_width_arg_flag = TRUE; 
            }
            else 
            {
                if (isdigit((int)(*format_str))) 
                {
                    field_width_flag = TRUE; 
                }

                while (isdigit((int)(*format_str))) 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                    format_str++;
                }
            } 

            if (*format_str == '.') 
            {
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;

                if (*format_str == '*') 
                {
                    *new_format_str_pos = *format_str; 
                    new_format_str_pos++;
                    format_str++;

                    num_precision_args++;
                    precision_flag = TRUE;
                }
                else 
                {
                    if (isdigit((int)(*format_str))) 
                    {
                        precision_flag = TRUE; 
                    }

                    while (isdigit((int)(*format_str))) 
                    {
                        *new_format_str_pos = *format_str; 
                        new_format_str_pos++;
                        format_str++;
                    }
                }
            }

            if (*format_str == 'l') 
            {
                long_flag = TRUE;
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;
            }
            else if (*format_str == 'h') 
            {
                short_flag = TRUE; 
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
                format_str++;
            }
            else if (*format_str == 'L') 
            {
#ifdef LONG_DOUBLE_OK
                long_double_flag = TRUE; 
                *new_format_str_pos = *format_str; 
                new_format_str_pos++;
#endif
                format_str++;
            }

            type_char = *format_str; 

            if (*format_str != '\0') 
            {
                format_str++;
            }

            if (type_char == '\0') 
            {
                *buff_pos = '\0'; 
                SET_FORMAT_STRING_BUG();
                return ERROR;
            } 

            if ((type_char == 'd') || (type_char == 'i'))
            {
                *new_format_str_pos = type_char;
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
                    /*
                    // Untested since changing "short"to "int" due to gcc
                    // complaining that there may be a problem because
                    // short gets promoted to int. Fair enough, but I don't
                    // know whether this is correct solution across all
                    // platforms?
                    */
                    if (short_flag) 
                    {
                        sprintf(temp_buff, 
                                new_format_str,
                                va_arg(ap, int)); 
                    }
                    else if (long_flag) 
                    {
                        sprintf(temp_buff, 
                                new_format_str,
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                va_arg(ap, int)); 
                    }
                } 
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    /*
                    // Untested since changing "short"to "int" due to gcc
                    // complaining that there may be a problem because
                    // short gets promoted to int. Fair enough, but I don't
                    // know whether this is correct solution across all
                    // platforms?
                    */
                    if (short_flag) 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, int)); 
                    }
                    else if (long_flag) 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, int)); 
                    }
                } 
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, int)); 
                    }
                    else if (long_flag) 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, int)); 
                    }
                } 
                else 
                {
                    *buff_pos = '\0'; 
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                /*
                // FIX 
                //
                // Currently, commas will not be inserted into anything other
                // than non-justified positive numbers. Note that commas do not
                // properly work with precision arguments!
                */
                if (comma_flag && all_digits(temp_buff))
                {
                    char buff_copy[ 100 ];
                    char* buff_copy_pos = buff_copy;
                    int  len = strlen(temp_buff);
                    int first_chunk = (len - 1) % 3 + 1;
                    int i; 

                    i = 0;

                    while (i<first_chunk)
                    {
                        *buff_copy_pos = temp_buff[ i ];
                        buff_copy_pos++;
                        i++;
                    } 

                    while (i<len)
                    {
                        *buff_copy_pos++ = ',';
                        *buff_copy_pos++ = temp_buff[ i++ ];
                        *buff_copy_pos++ = temp_buff[ i++ ];
                        *buff_copy_pos++ = temp_buff[ i++ ];
                    }

                    *buff_copy_pos = '\0';

                    BUFF_CPY(temp_buff, buff_copy);
                }
                    
                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            } 
            else if ( (type_char == 'o') || (type_char == 'x') || 
                      (type_char == 'u') || (type_char == 'X') ) 
            {
                *new_format_str_pos = type_char;
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(temp_buff, 
                                new_format_str,
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

                    if (short_flag) 
                    {
                        /*
                        // Untested since changing "short"to "int" due to gcc
                        // complaining that there may be a problem because
                        // short gets promoted to int. Fair enough, but I don't
                        // know whether this is correct solution across all
                        // platforms?
                        */
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned int)); 
                    }
                    else if (long_flag)
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned long)); 
                    }
                    else 
                    {
                        sprintf(temp_buff,
                                new_format_str,
                                precision_arg1, 
                                precision_arg2, 
                                va_arg(ap, unsigned int)); 
                    }
                } 
                else 
                {
                    *buff_pos = '\0'; 
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                /*
                // FIX 
                //
                // Currently, commas will not be inserted into anything other
                // than non-justified positive numbers. Note that commas do not
                // properly work with precision arguments!
                */
                if ((type_char=='u') && comma_flag && all_digits(temp_buff))
                {
                    char buff_copy[ 100 ];
                    char* buff_copy_pos = buff_copy;
                    int  len = strlen(temp_buff);
                    int first_chunk = (len - 1) % 3 + 1;
                    int i; 

                    i = 0;

                    while (i<first_chunk)
                    {
                        *buff_copy_pos = temp_buff[ i ];
                        buff_copy_pos++;
                        i++;
                    } 

                    while (i<len)
                    {
                        *buff_copy_pos++ = ',';
                        *buff_copy_pos++ = temp_buff[ i++ ];
                        *buff_copy_pos++ = temp_buff[ i++ ];
                        *buff_copy_pos++ = temp_buff[ i++ ];
                    }

                    *buff_copy_pos = '\0';

                    BUFF_CPY(temp_buff, buff_copy);
                }
                    
                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            } 
            else if ((type_char == 'e') || (type_char == 'f') ||
                     (type_char == 'g') || (type_char == 'E') ||
                     (type_char == 'G'))
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (num_precision_args == 0) 
                {
#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(temp_buff, new_format_str, 
                                va_arg(ap, long_double));
                    }
                    else 
                    {
                        sprintf(temp_buff, new_format_str, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(temp_buff, new_format_str,
                            va_arg(ap, double));
#endif
                }
                else if (num_precision_args == 1) 
                {
                    precision_arg1 = va_arg(ap, int); 

#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(temp_buff, new_format_str,
                                precision_arg1, va_arg(ap, long_double));
                    }
                    else 
                    {
                        sprintf(temp_buff, new_format_str, 
                                precision_arg1, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(temp_buff, new_format_str, precision_arg1, 
                            va_arg(ap, double));
#endif
                }
                else if (num_precision_args == 2) 
                {
                    precision_arg1 = va_arg(ap, int); 
                    precision_arg2 = va_arg(ap, int); 

#ifdef LONG_DOUBLE_OK
                    if (long_double_flag) 
                    {
                        sprintf(temp_buff, new_format_str,
                                precision_arg1, precision_arg2, 
                                va_arg(ap, long_double));
                    }
                    else 
                    {
                        sprintf(temp_buff, new_format_str,
                                precision_arg1, precision_arg2, 
                                va_arg(ap, double));
                    }
#else
                    sprintf(temp_buff, new_format_str,
                            precision_arg1, precision_arg2, 
                            va_arg(ap, double));
#endif
                }
                else 
                {
                    *buff_pos = '\0'; 
                    SET_CANT_HAPPEN_BUG();
                    return ERROR;
                }

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 's') 
            {
                if ((justify_flag) || (precision_flag) || (field_width_flag)) 
                {
                    *new_format_str_pos = type_char; 
                    new_format_str_pos++;
                    *new_format_str_pos = '\0';

                    /*
                     *  Overflow is possible: FIX 
                     */
                    if (num_precision_args == 0) 
                    {
                        sprintf(temp_buff, new_format_str, 
                                va_arg(ap, char*));
                    }
                    else if (num_precision_args == 1) 
                    {
                        precision_arg1 = va_arg(ap, int); 

                        sprintf(temp_buff, new_format_str, 
                                precision_arg1, va_arg(ap, char*));
                    } 
                    else if (num_precision_args == 2) 
                    {
                        precision_arg1 = va_arg(ap, int); 
                        precision_arg2 = va_arg(ap, int); 

                        sprintf(temp_buff, new_format_str, 
                                precision_arg1, precision_arg2, 
                                va_arg(ap, char*));
                    }
                    else 
                    {
                        *buff_pos = '\0'; 
                        SET_CANT_HAPPEN_BUG();
                        return ERROR;
                    }

                    temp_buff_pos = temp_buff;
                }
                else 
                {
                    temp_buff_pos = va_arg(ap, char *);
                }

                temp_buff_len = strlen(temp_buff_pos);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff_pos); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'S') 
            {

#ifdef HAVE_STRERROR
                const char* sys_mess = strerror(errno);
#else

#ifndef errno  /* GNU defines errno as a macro, leading to warning messages. */
                IMPORT int   errno; 
#endif 
                IMPORT int   sys_nerr; 
                IMPORT SYS_ERRLIST_TYPE sys_errlist[]; 
#endif 

#ifdef HAVE_STRERROR
                if (sys_mess != NULL) 
#else
                if (    (errno != NO_SYSTEM_ERROR) 
                     && (errno < sys_nerr)
                   )
#endif 
                {
                    temp_buff_pos = "\nSystem error message is: "; 
                    temp_buff_len = strlen(temp_buff_pos);

                    if (temp_buff_len > max_len)
                    {
                        *buff_pos = '\0'; 
                        SET_BUFFER_OVERFLOW_BUG();
                        return ERROR;
                    }

                    strcpy(buff_pos, temp_buff_pos); 
                    buff_pos += temp_buff_len; 
                    max_len -= temp_buff_len;

#ifdef HAVE_STRERROR
                    temp_buff_pos = sys_mess; 
#else
                    temp_buff_pos = sys_errlist[errno];
#endif 

                    temp_buff_len = strlen(temp_buff_pos);

                    if (temp_buff_len > max_len)
                    {
                        *buff_pos = '\0'; 
                        SET_BUFFER_OVERFLOW_BUG();
                        return ERROR;
                    }

                    strcpy(buff_pos, temp_buff_pos); 
                    buff_pos += temp_buff_len; 
                    max_len -= temp_buff_len;
                }
            }
            else if ((type_char == 'q') || (type_char == 't')) 
            {
                field_width = sizeof(trunc_buff); 

                if (num_precision_args == 1)
                {
                    temp_field_width = va_arg(ap, int); 

                    if (temp_field_width < field_width)
                    {
                        field_width = temp_field_width;
                    }
                }
                else if (field_width_flag)
                {
                    *new_format_str_pos = '\0';     

                    if (ss1pi(new_format_str + 1, &temp_field_width) == ERROR)
                    {
                        SET_FORMAT_STRING_BUG();
                        return ERROR; 
                    }

                    if (temp_field_width < field_width)
                    {
                        field_width = temp_field_width;
                    }
                }

                if (type_char == 't') 
                {
                    if (field_width < 5) 
                    {
                        field_size = 5;
                    }
                    else field_size = field_width;

                    str_trunc_cpy(trunc_buff, va_arg(ap, char *), field_size); 
                }
                else 
                {
                    if (field_width < 7) 
                    {
                        field_size = 7;
                    }
                    else field_size = field_width;

                    trunc_quote_cpy(trunc_buff, va_arg(ap, char *), field_size);
                }

                temp_buff_len = strlen(trunc_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, trunc_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'D') 
            {
                temp_buff[ 0 ] = '\0';
                BUFF_GET_USER_FD_NAME(va_arg(ap, int), temp_buff);
                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'F') 
            {
                temp_buff[ 0 ] = '\0';
                BUFF_GET_USER_FD_NAME(fileno(va_arg(ap, FILE *)), temp_buff);
                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'P') 
            {
                temp_buff[ 0 ] = '\0';
                BUFF_GET_FD_NAME(fileno(va_arg(ap, FILE *)), temp_buff);
                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'R') 
            {
                temp_buff[ 0 ] = '\0';

                get_status_string(va_arg(ap, int), temp_buff, 
                                  sizeof(temp_buff));

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'c') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';

                if (field_width_arg_flag) 
                {
                    precision_arg1 = va_arg(ap, int); 

                    sprintf(temp_buff, new_format_str,
                            precision_arg1, va_arg(ap, int));
                }
                else 
                {
                    sprintf(temp_buff, new_format_str,
                            va_arg(ap, int));
                }

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'n') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';
                sprintf(temp_buff, new_format_str, va_arg(ap, int*));

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else if (type_char == 'p') 
            {
                *new_format_str_pos = type_char;
                new_format_str_pos++;
                *new_format_str_pos = '\0';
                sprintf(temp_buff, new_format_str, va_arg(ap, void*)); 

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            } 
            else if (type_char == '%') 
            {
                *new_format_str_pos = type_char; 
                new_format_str_pos++;
                *new_format_str_pos = '\0';
                sprintf(temp_buff, new_format_str); 

                temp_buff_len = strlen(temp_buff);

                if (temp_buff_len > max_len)
                {
                    *buff_pos = '\0'; 
                    SET_BUFFER_OVERFLOW_BUG();
                    return ERROR;
                }

                strcpy(buff_pos, temp_buff); 
                buff_pos += temp_buff_len; 
                max_len -= temp_buff_len;
            }
            else 
            {
                *buff_pos = '\0'; 
                SET_FORMAT_STRING_BUG();
                return ERROR;
            } 
        }
    }

    *buff_pos = '\0'; 

    return (save_max_len - max_len);
}

