
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 

#include "m/m_vec_metric.h" 
#include "m/m_error.h" 

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                         check_same_matrix_dimensions
 *                               
 * Checks that two matrices have the same dimensions
 *
 * This routine checks that two matrices have the same dimensions. It is
 * designed for the situation where differing dimension is indicative of a
 * programming problem. It the matrices do not have the same dimension, then the
 * routine set_bug is called with a message string constructed from the
 * routine_name argument. 
 *
 * Returns: 
 *    NO_ERROR if the matrices have the same dimensions, and ERROR otherwise.
 *    Depending on the bug handler in place, this routine may not return at all. 
 *
 * Index: error handling, debugging 
 *
 * -----------------------------------------------------------------------------
*/ 

int check_same_matrix_dimensions
(
    const Matrix* in_mp,
    const Matrix* out_mp,
    const char*   routine_name 
)
{
    int num_in_rows, num_out_rows;
    int num_in_cols, num_out_cols;

    if (in_mp == NULL) 
    {
        set_bug("Null input matrix sent to %s.", routine_name);
        return ERROR; 
    }

    if (out_mp == NULL) 
    {
        set_bug("Null output matrix sent to %s.", routine_name);
        return ERROR; 
    }

    num_in_rows = in_mp->num_rows;
    num_out_rows = out_mp->num_rows;

    if (num_in_rows != num_out_rows) 
    {
        set_bug("Mismatch in row counts in %s.", routine_name);
        return ERROR; 
    }
        
    num_in_cols = in_mp->num_cols;
    num_out_cols = out_mp->num_cols;

    if (num_in_cols != num_out_cols) 
    {
        set_bug("Mismatch in row counts in %s.", routine_name);
        return ERROR; 
    }
        
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         check_same_vector_lengths
 *                               
 * Checks that two vectors have the same length
 *
 * This routine checks that two vectors have the same length. It is designed
 * for the situation where differing length is indicative of a programming
 * problem. It the vectors do not have the same length, then the routine
 * set_bug is called with a message string constructed from the routine_name
 * argument. 
 *
 * Returns: 
 *    NO_ERROR if the vectors are the same length, and ERROR otherwise.
 *    Depending on the bug handler in place, this routine may not return at all. 
 *
 * Index:
 *    error handling, debugging 
 *
 * -----------------------------------------------------------------------------
*/ 

int check_same_vector_lengths
(
    const Vector* in_vp,
    const Vector* out_vp,
    const char*   routine_name 
)
{
    int in_length, out_length;

    if (in_vp == NULL) 
    {
        set_bug("Null input vector sent to %s.", routine_name);
        return ERROR; 
    }

    if (out_vp == NULL) 
    {
        set_bug("Null output vector sent to %s.", routine_name);
        return ERROR; 
    }

    in_length = in_vp->length;
    out_length = out_vp->length;

    if (in_length != out_length) 
    {
        set_bug("Mismatch in vector length in %s.", routine_name);
        return ERROR; 
    }
        
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_expanded_error_box(Matrix** expanded_mpp, const Matrix* mp)
{
    int count, i, j, k;
    int num_box_rows;
    int num_box_cols;
    int num_expanded_rows;
    int expansion_factor;
    int dim;
    Matrix* expanded_mp;
    int temp; 

    num_box_rows = mp->num_rows;
    num_box_cols = mp->num_cols;

    /*
    // The input to this routine is a matrix with twice as many columns as the
    // dimension because each original point (X1, X2, ... ) became 
    // (X1-error, X1+error, X2-error, X2+error, 
    */
    dim = num_box_cols/2;

    expansion_factor = (1 << dim);
    num_expanded_rows = num_box_rows * expansion_factor;

    ERE(get_target_matrix(expanded_mpp, num_expanded_rows, dim)); 
    expanded_mp = *expanded_mpp; 

    count = 0;

    for (i=0; i<num_box_rows; i++) 
    {
        for (j=0; j<expansion_factor; j++) 
        {
            for (k=0; k<dim; k++) 
            {
                temp = j >> k;

                if (IS_EVEN(temp)) 
                {
                    (expanded_mp->elements)[count][k] = (mp->elements)[i][2*k];
                }
                else 
                {
                    (expanded_mp->elements)[count][k] =(mp->elements)[i][2*k+1];
                }
            }
            count++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Matrix* expand_error_box(const Matrix* mp)
{
    int count, i, j, k;
    int num_box_rows;
    int num_box_cols;
    int num_expanded_rows;
    int expansion_factor;
    int dim;
    Matrix* expanded_mp;
    int temp; 

    num_box_rows = mp->num_rows;
    num_box_cols = mp->num_cols;

    /*
    // The input to this routine is a matrix with twice as many columns as the
    // dimension because each original point (X1, X2, ... ) became 
    // (X1-error, X1+error, X2-error, X2+error, 
    */
    dim = num_box_cols/2;

    expansion_factor = (1 << dim);
    num_expanded_rows = num_box_rows * expansion_factor;

    NRN(expanded_mp = create_matrix(num_expanded_rows, dim)); 

    count = 0;

    for (i=0; i<num_box_rows; i++) 
    {
        for (j=0; j<expansion_factor; j++) 
        {
            for (k=0; k<dim; k++) 
            {
                temp = j >> k;

                if (IS_EVEN(temp)) 
                {
                    (expanded_mp->elements)[count][k] = (mp->elements)[i][2*k];
                }
                else 
                {
                    (expanded_mp->elements)[count][k] =(mp->elements)[i][2*k+1];
                }
            }
            count++;
        }
    }

    return expanded_mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Matrix* create_error_points(const Matrix* mp, double error)
{
    Matrix* error_box_mp, *new_mp;
    int i,j;

    NRN(error_box_mp = create_matrix(mp->num_rows, 2*(mp->num_cols)));

    for (i=0; i<mp->num_rows; i++) 
    {
        for (j=0; j<mp->num_cols; j++) 
        {
            (error_box_mp->elements)[i][ 2*j ] = (mp->elements)[i][j] + error;
            (error_box_mp->elements)[i][2*j+1] = (mp->elements)[i][j] - error;
        }
    }

    NRN(new_mp = expand_error_box(error_box_mp));

    free_matrix(error_box_mp);

    return new_mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             get_points_with_error
 *                               
 * Computes points with added error
 *
 * This routine adds error to matrix rows. If abs_err is greater than 0, abs_err
 * absolute error is added. If rel_err is greater than 0, then rel_err relative
 * error is added. If both are positive, then both are used, with the absolute
 * error being added first. Finally, if min_valid_value is greater than or equal
 * to zero, the points computed will be truncated at that value.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine will fail if storage allocation fails. 
 *    
 * Index: data error
 *
 * -----------------------------------------------------------------------------
*/ 

int get_points_with_error
(
    Matrix**      result_mpp,
    const Matrix* mp,
    double        abs_err,
    double        rel_err,
    double        min_valid_value 
)
{
    Matrix* error_box_mp;
    int i,j;
    int result; 

    NRE(error_box_mp = create_matrix(mp->num_rows, 2*(mp->num_cols)));

    for (i=0; i<mp->num_rows; i++) 
    {
        for (j=0; j<mp->num_cols; j++) 
        {
            (error_box_mp->elements)[i][ 2*j ] = (mp->elements)[i][j];
            
            (error_box_mp->elements)[i][2*j+1] = (mp->elements)[i][j];

            if (abs_err > 0.0)
            {
                (error_box_mp->elements)[i][ 2*j ] += abs_err;
                (error_box_mp->elements)[i][2*j+1] -= abs_err;
            }

            if (rel_err > 0.0)
            {
                (error_box_mp->elements)[i][ 2*j ]   *= (1.0 + rel_err); 
                (error_box_mp->elements)[i][ 2*j+1 ] *= (1.0 - rel_err); 
            }

            if (min_valid_value >= 0.0)
            {
                (error_box_mp->elements)[i][ 2*j ] =
                    MAX_OF((error_box_mp->elements)[i][ 2*j ], min_valid_value);
            }
        }
    }

    result = get_expanded_error_box(result_mpp, error_box_mp);

    free_matrix(error_box_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST
    void set_divide_by_zero_error(int line, const char* file)
{

      if (is_interactive())
      {
          set_bug("Division by zero on line %d of file %s.",line, file);
      }
      else 
      {
          set_error("Division by zero on line %d of file %s.",line, file);
      }
  }

#else 

    void set_divide_by_zero_error(void)
{

      set_error("Division by zero.");
  }

#endif 
   
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST
    void set_overflow_error(int line, const char* file)
{

      if (is_interactive())
      {
          set_bug("Arithmetic overflow detected on line %d of file %s.",
                  line, file);
      }
      else 
      {
          set_error("Arithmetic overflow detected on line %d of file %s.",
                    line, file);
      }
  }

#else 

    void set_overflow_error(void)
{

      set_error("Arithmetic overflow detected.");
  }

#endif 
   
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int check_quadratic_constraints
(
    const Vector* result_vp,
    const Matrix* le_constraint_mp,
    const Vector* le_constraint_col_vp,
    const Vector* lb_constraint_vp,
    const Vector* ub_constraint_vp 
)
{
    Vector* constraint_result_vp = NULL;
    int     result               = NO_ERROR;
    double    relative_tolerance   = 0.05;
    double    absolute_tolerance   = 0.005;

    if (    (ub_constraint_vp != NULL) 
         && (ub_constraint_vp->length > 0)
         && ( ! (first_vector_is_less(result_vp, ub_constraint_vp, 
                                      relative_tolerance, absolute_tolerance)))

       )
    {
        set_error("Upper-bound constraint not satisfied.");
        add_error("Relative tolerance is %e and absolute tolerance is %e.",
                  relative_tolerance, absolute_tolerance);
        return ERROR;
    }

    if (    (lb_constraint_vp != NULL)
         && (lb_constraint_vp->length > 0)
         && ( ! (first_vector_is_less(lb_constraint_vp, result_vp, 
                                      relative_tolerance, absolute_tolerance)))
       )
    {
        set_error("Lower-bound constraint not satisfied.");
        add_error("Relative tolerance is %e and absolute tolerance is %e.",
                  relative_tolerance, absolute_tolerance);
        return ERROR;
    }

    if ((le_constraint_mp != NULL) && (le_constraint_mp->num_rows > 0))
    {
        ERE(multiply_matrix_and_vector(&constraint_result_vp, le_constraint_mp,
                                       result_vp));

        if ( ! (first_vector_is_less(constraint_result_vp, le_constraint_col_vp,
                                     relative_tolerance, absolute_tolerance))
           )
        {
            set_error("Inequality constraint not satisfied.");
            add_error("Relative tolerance is %e and absolute tolerance is %e.",
                      relative_tolerance, absolute_tolerance);
            result = ERROR;
        }
    }

    free_vector(constraint_result_vp);

    return result;
}
   
