
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "m/m_mat_arith.h" 

/* -------------------------------------------------------------------------- */

/*
 * =============================================================================
 *                              get_vector_outer_product 
 *
 * Computes outer product of two vectors. 
 *
 * This routine computes theouter product of two vectors. The two vectors must
 * have the same dimensions.
 * 
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: vectors, vector arithmetic, matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int get_vector_outer_product(Matrix** mpp, const Vector* vp1, const Vector* vp2)
{
    int i, j, len; 

    ERE(check_same_vector_lengths(vp1, vp2, "get_vector_outer_product")); 
    len = vp1->length; 
    ERE(get_target_matrix(mpp, len, len));

    for (i=0; i<len; i++)
    {
        for (j=0; j<len; j++)
        {
            (*mpp)->elements[ i ][ j ] = 
                                        vp1->elements[ i ] * vp2->elements[ j ];
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              add_matrices 
 *
 * Performs element-wise addition of two matrices. 
 *
 * This routine performs element-wise addition of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. The second
 * matrix is added to each block in the first matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int add_matrices(Matrix** target_mpp, const Matrix* first_mp, 
                 const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     block_i;
    int     block_j;
    int     target_i;
    int     target_j;
    int     num_rows;
    int     num_cols;
    int     num_second_rows;
    int     num_second_cols;
    int     row_factor;
    int     col_factor; 
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't be added to %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    target_mp = *target_mpp; 

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    target_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] + 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              ow_add_matrices 
 *
 * Performs element-wise addition of two matrices. 
 *
 * This routine performs element-wise addition of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. The second
 * matrix is added to each block in the first matrix. The first matrix is thus
 * overwritten. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_matrices(Matrix* first_mp, const Matrix* second_mp)
{
    int i, j;
    int block_i;
    int block_j;
    int target_i;
    int target_j;
    int num_rows, num_cols;
    int num_second_rows;
    int num_second_cols;
    int row_factor;
    int col_factor;
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't be added to %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    first_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] + 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              subtract_matrices 
 *
 * Performs element-wise subtraction of two matrices. 
 *
 * This routine performs element-wise subtraction of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. The second
 * matrix is subtracted from each block in the first matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int subtract_matrices(Matrix** target_mpp, const Matrix* first_mp, 
                      const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     block_i;
    int     block_j;
    int     target_i;
    int     target_j;
    int     num_rows;
    int     num_cols;
    int     num_second_rows;
    int     num_second_cols;
    int     row_factor;
    int     col_factor; 
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't be subtracted from %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    target_mp = *target_mpp; 

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    target_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] - 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              ow_subtract_matrices 
 *
 * Performs element-wise subtraction of two matrices. 
 *
 * This routine performs element-wise subtraction of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. The second
 * matrix is subtracted from each block in the first matrix. The first matrix
 * is thus over-written. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_matrices(Matrix* first_mp, const Matrix* second_mp)
{
    int i, j;
    int block_i;
    int block_j;
    int target_i;
    int target_j;
    int num_rows, num_cols;
    int num_second_rows;
    int num_second_cols;
    int row_factor;
    int col_factor;
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't be subtracted from %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    first_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] - 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          multiply_matrices_ew 
 *
 * Performs element-wise multiplication of two matrices. 
 *
 * This routine performs element-wise multiplication of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. Each block
 * in the first matrix with dimensions of the second is multipled elementwise
 * by the second. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrices_ew(Matrix** target_mpp, const Matrix* first_mp, 
                         const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     block_i;
    int     block_j;
    int     target_i;
    int     target_j;
    int     num_rows;
    int     num_cols;
    int     num_second_rows;
    int     num_second_cols;
    int     row_factor;
    int     col_factor; 
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't multiply %d by %d.", 
                  num_rows, num_cols, num_second_rows, 
                  num_second_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    target_mp = *target_mpp; 

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    target_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] * 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                      ow_multiply_matrices_ew 
 *
 * Performs element-wise multiplication of two matrices. 
 *
 * This routine performs element-wise multiplication of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. Each block
 * in the first matrix with dimensions of the second is multipled elementwise
 * by the second. Thus the first matrix is over-written. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrices_ew(Matrix* first_mp, const Matrix* second_mp)
{
    int i, j;
    int block_i;
    int block_j;
    int target_i;
    int target_j;
    int num_rows, num_cols;
    int num_second_rows;
    int num_second_cols;
    int row_factor;
    int col_factor;
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't multiply %d by %d.", 
                  num_rows, num_cols, num_second_rows, 
                  num_second_cols);
        return ERROR;
    }

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    first_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] * 
                                                 second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          divide_matrices_ew 
 *
 * Performs element-wise division of two matrices. 
 *
 * This routine performs element-wise division of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. Each block
 * of the first matrix is divided by the second matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine will fail if any of the matrix elements are too close
 *     to zero, or if memory allocation fails. The routine will also fail due to
 *     dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()).
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int divide_matrices_ew(Matrix** target_mpp, const Matrix* first_mp, 
                       const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     block_i;
    int     block_j;
    int     target_i;
    int     target_j;
    int     num_rows;
    int     num_cols;
    int     num_second_rows;
    int     num_second_cols;
    int     row_factor;
    int     col_factor; 
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't divide %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    target_mp = *target_mpp; 

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    double divisor;

                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    divisor = second_mp->elements[ i ][ j ];

#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
                    if (IS_ZERO_DBL(divisor))
                    {
                        SET_DIVIDE_BY_ZERO_ERROR();
                        return ERROR; 
                    }
#else 
                    /* 
                    // Exact test is OK, because case of small divisor is
                    // handled below.
                    */
                    if (divisor == 0.0)
                    {
                        SET_DIVIDE_BY_ZERO_ERROR();
                        return ERROR; 
                    }
#endif 

                    target_mp->elements[ target_i ][ target_j ] =
                          first_mp->elements[ target_i ][ target_j ] / divisor;

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
                    if (    (target_mp->elements[ target_i ][ target_j ] >
                                                      DBL_MOST_POSITIVE) 
                         || (target_mp->elements[ target_i ][ target_j ] < 
                                                      DBL_MOST_NEGATIVE)
                       )
                    {
                        SET_OVERFLOW_ERROR();
                        return ERROR; 
                    }
#endif 

                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_divide_matrices_ew 
 *
 * Performs element-wise division of two matrices. 
 *
 * This routine performs element-wise division of two matrices. The second
 * matrix must fit into the first an integral number of times. For example, if
 * the first matrix has dimenions 6x3, then the second can have dimensions 6x3
 * (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT 2x2, 5x1, etc. Each block
 * of the first matrix is divided by the second matrix. The first matrix is
 * thus over-written. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine will fail if any of the matrix elements are too close
 *     to zero, or if memory allocation fails. The routine will also fail due to
 *     dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()).
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrices_ew(Matrix* first_mp, const Matrix* second_mp)
{
    int i, j;
    int block_i;
    int block_j;
    int target_i;
    int target_j;
    int num_rows, num_cols;
    int num_second_rows;
    int num_second_cols;
    int row_factor;
    int col_factor;
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't divide %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    double divisor;

                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    divisor = second_mp->elements[ i ][ j ];

#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
                    if (IS_ZERO_DBL(divisor))
                    {
                        SET_DIVIDE_BY_ZERO_ERROR();
                        return ERROR; 
                    }
#else 
                    /* 
                    // Exact test is OK, because case of small divisor is
                    // handled below.
                    */
                    if (divisor == 0.0)
                    {
                        SET_DIVIDE_BY_ZERO_ERROR();
                        return ERROR; 
                    }
#endif 

                    first_mp->elements[ target_i ][ target_j ] =
                          first_mp->elements[ target_i ][ target_j ] / divisor;

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
                    if (    (second_mp->elements[ target_i ][ target_j ] >
                                                      DBL_MOST_POSITIVE) 
                         || (second_mp->elements[ target_i ][ target_j ] < 
                                                      DBL_MOST_NEGATIVE)
                       )
                    {
                        SET_OVERFLOW_ERROR();
                        return ERROR; 
                    }
#endif 

                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              ow_add_matrix_times_scalar 
 *
 * Performs element-wise addition of a matrix and a matrix times scalar. 
 *
 * This routine performs element-wise addition of a matrix and a matrix times a
 * scalar. The second matrix must fit into the first an integral number of
 * times. For example, if the first matrix has dimenions 6x3, then the second
 * can have dimensions 6x3 (the common case), 6x1, 3x3, 2x3, 1x6, etc, but NOT
 * 2x2, 5x1, etc. The second matrix is added to each block in the first matrix.
 * The first matrix is thus overwritten. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_matrix_times_scalar(Matrix* first_mp, const Matrix* second_mp, 
                               double scalar)
{
    int i, j;
    int block_i;
    int block_j;
    int target_i;
    int target_j;
    int num_rows, num_cols;
    int num_second_rows;
    int num_second_cols;
    int row_factor;
    int col_factor;
    
    num_rows = first_mp->num_rows;
    num_cols = first_mp->num_cols;

    num_second_rows = second_mp->num_rows;
    num_second_cols = second_mp->num_cols;

    row_factor = num_rows / num_second_rows;
    col_factor = num_cols / num_second_cols;

    if (    (num_rows != row_factor * num_second_rows)
         || (num_cols != col_factor * num_second_cols)
       )
    {
        set_error("%d by %d can't be added to %d by %d.", 
                  num_second_rows, num_second_cols,
                  num_rows, num_cols);
        return ERROR;
    }

    for (block_i=0; block_i<row_factor; block_i++)
    {
        for (block_j=0; block_j<col_factor; block_j++)
        {
            for (i=0; i<num_second_rows; i++) 
            {
                for (j=0; j<num_second_cols; j++) 
                {
                    target_i = block_i * num_second_rows + i;
                    target_j = block_j * num_second_cols + j;

                    first_mp->elements[ target_i ][ target_j ] =
                        first_mp->elements[ target_i ][ target_j ] + 
                                         scalar * second_mp->elements[ i ][ j ];
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          ow_add_int_matrix_to_matrix
 *                               
 * Adds an integer matrix to a matrix 
 *
 * This routine addes the integer matrix source_mp to the matrix
 * target_mp.
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

void ow_add_int_matrix_to_matrix(Matrix* target_mp, const Int_matrix* source_mp)
{
    int     i, j, num_rows, num_cols;
    double* target_pos;
    int* source_pos;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = (target_mp->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos += (double)(*source_pos);
            target_pos++;
            source_pos++; 
        }
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          multiply_matrices
 *
 * Multiplies two matrices. 
 *
 * This routine multiplies two matrices. The matrices must be compatable for
 * muliplication, or ERROR is returned. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrices(Matrix** target_mpp, const Matrix* first_mp, 
                      const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     k;
    int     num_rows;
    int     num_cols;
    int     length; 
    double*   first_pos; 
    double**  second_elements;
    
    if (first_mp->num_cols != second_mp->num_rows)
    {
        set_error("Incompatable dimensions for matrix multiplication.");
        add_error("Attempt to multiply a %d by %d with a %d by %d.",
                  first_mp->num_rows, first_mp->num_cols, 
                  second_mp->num_rows, second_mp->num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, first_mp->num_rows, 
                          second_mp->num_cols));

    target_mp = *target_mpp; 

    second_elements = second_mp->elements; 

    num_rows = target_mp->num_rows;
    num_cols = target_mp->num_cols;
    length   = first_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        for (j=0; j<num_cols; j++) 
        {
            long_double sum = 0.0; /* long_double only has extra precision 
                                      where it is supported in hardware. */

            first_pos  = (first_mp->elements)[ i ]; 

            for (k=0; k<length; k++)
            {
                sum += ((*first_pos) * (second_elements)[k][j]);
                first_pos++; 
            }

            target_mp->elements[ i ][ j ] = sum; 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        multiply_by_transpose      
 *
 * Multiplies one matrix by the transpose of the second.
 *
 * This routine multiplies one matrix by the transpose of the second. The
 * matrices must be compatable for muliplication, or ERROR is returned. 
 *
 * It is slightly faster to multiply by the transpose, if it is available.
 * Hence, if the operation required is multiplication by the transpose, then it
 * is much better to use this routine, as opposed to getting the transpose, and
 * then multiplying. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_by_transpose(Matrix** target_mpp, const Matrix* first_mp, 
                          const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     k;
    int     num_rows;
    int     num_cols;
    int     length; 
    
    if (first_mp->num_cols != second_mp->num_cols)
    {
        set_error("Incompatable dimensions for matrix multiplication.");
        add_error("Attempt to multiply %d by %d with the transpose of %d by %d.",
                  first_mp->num_rows, first_mp->num_cols, 
                  second_mp->num_rows, second_mp->num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, first_mp->num_rows, 
                          second_mp->num_rows));

    target_mp = *target_mpp; 

    num_rows = target_mp->num_rows;
    num_cols = target_mp->num_cols;
    length   = first_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        for (j=0; j<num_cols; j++) 
        {
            long_double sum = 0.0; /* long_double only has extra precision 
                                      where it is supported in hardware. */
            double* first_pos;
            double* second_pos;

            first_pos  = (first_mp->elements)[ i ]; 
            second_pos = (second_mp->elements)[ j ]; 

            for (k=0; k<length; k++)
            {
                sum += (*first_pos) * (*second_pos);
                first_pos++; 
                second_pos++; 
            }

            target_mp->elements[ i ][ j ] = sum; 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                       multiply_with_transpose      
 *
 * Multiplies the transpose of a matrix by the second.
 *
 * This routine multiplies the transpose of the first matrix argument by the
 * second matrix. The matrices must be compatable for muliplication, or ERROR
 * is returned. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_with_transpose(Matrix** target_mpp, const Matrix* first_mp, 
                            const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     k;
    int     num_rows;
    int     num_cols;
    int     length; 
    
    if (first_mp->num_rows != second_mp->num_rows)
    {
        set_error("Incompatable dimensions for matrix multiplication.");
        add_error("Attempt to multiply the transpose of %d by %d with %d by %d.",
                  first_mp->num_rows, first_mp->num_cols, 
                  second_mp->num_rows, second_mp->num_cols);
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, first_mp->num_cols, 
                          second_mp->num_cols));

    target_mp = *target_mpp; 

    num_rows = target_mp->num_rows;
    num_cols = target_mp->num_cols;
    length   = first_mp->num_rows;

    for (i=0; i<num_rows; i++) 
    {
        for (j=0; j<num_cols; j++) 
        {
            long_double sum = 0.0; /* long_double only has extra precision 
                                      where it is supported in hardware. */

            for (k=0; k<length; k++)
            {
                double first_elem;
                double second_elem;

                first_elem = first_mp->elements[ k ][ i ];
                second_elem = second_mp->elements[ k ][ j ]; 
                    
                sum += first_elem * second_elem;
            }

            target_mp->elements[ i ][ j ] = sum; 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         invert_matrix_elements 
 *
 * Computes the element-wise reciprocal of a matrix
 *
 * This routine computes the element-wise reciprocal of a matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Thos routine will fail if any of the matrix elements are too close
 *     to zero, or if memory allocation fails. 
 * 
 * See also:
 *     ow_invert_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int invert_matrix_elements(Matrix** target_mpp, const Matrix* source_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*source_row_pos_ptr))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else 
            /* 
            // Exact test is OK, because case of small divisor is
            // handled below.
            */
            if (*source_row_pos_ptr == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            (*target_row_pos_ptr) = 1.0 / (*source_row_pos_ptr);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*target_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*target_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         square_matrix_elements 
 *
 * Computes the element-wise square of a matrix
 *
 * This routine computes the element-wise square of a matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Thos routine will fail if any of the matrix elements are too close
 *     to zero, or if memory allocation fails. 
 * 
 * See also:
 *     ow_square_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int square_matrix_elements(Matrix** target_mpp, const Matrix* source_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) 
                                                    * (*source_row_pos_ptr);

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         exp_matrix_elements 
 *
 * Computes exp(x) for matrix elements
 *
 * This routine computes exp(x) for matrix elements. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Thos routine will fail if any of the matrix elements are too close
 *     to zero, or if memory allocation fails. 
 * 
 * See also:
 *     ow_exp_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int exp_matrix_elements(Matrix** target_mpp, const Matrix* source_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = exp(*source_row_pos_ptr);

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         log_matrix_elements 
 *
 * Computes log(x) for matrix elements
 *
 * This routine computes log(x) for matrix elements. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. If the matrix element is 0.0, then
 * the value of the argument log_zero is used.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Thos routine will fail if any of the matrix elements are negative,
 *     or if memory allocation fails. 
 * 
 * See also:
 *     ow_log_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int log_matrix_elements(Matrix** target_mpp, const Matrix* source_mp, 
                        double log_zero)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            double x = *source_row_pos_ptr;

            if (x < 0.0)
            {
                set_error("Attempt to take the log of a negative number.");
                add_error("(Element (%d, %d) of a %d x %d matrix.",
                          i + 1, j + 1, num_rows, num_cols); 
                return NO_ERROR; 
            }
            else if (x == 0.0)
            {
                (*target_row_pos_ptr) = log_zero; 
            }
            else 
            {
                (*target_row_pos_ptr) = log(x);
            }

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         sqrt_matrix_elements 
 *
 * Computes sqrt(x) for matrix elements
 *
 * This routine computes sqrt(x) for matrix elements. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. If the matrix element is 0.0, then
 * the value of the argument log_zero is used.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Thos routine will fail if any of the matrix elements are negative,
 *     or if memory allocation fails. 
 * 
 * See also:
 *     ow_log_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int sqrt_matrix_elements(Matrix** target_mpp, const Matrix* source_mp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            double x = *source_row_pos_ptr;

            if (x < 0.0)
            {
                set_error("Attempt to take the squar root of a negative number.");
                add_error("(Element (%d, %d) of a %d x %d matrix.",
                          i + 1, j + 1, num_rows, num_cols); 
                return NO_ERROR; 
            }
            else 
            {
                (*target_row_pos_ptr) = sqrt(x);
            }

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          add_scalar_to_matrix 
 *
 * Adds a scalar to each element of a matrix
 *
 * This routine adds a scalar to each element of a matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory 
 *     allocation failure. 
 * 
 * See also:
 *     ow_add_scalar_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int add_scalar_to_matrix(Matrix** target_mpp, const Matrix* source_mp, 
                         double scalar)
{
    Matrix* target_mp;
    int    i;
    int    j;
    int    num_rows;
    int    num_cols;
    double** source_row_ptr;
    double** target_row_ptr;
    double*  source_row_pos_ptr;
    double*  target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) + scalar;
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        subtract_scalar_from_matrix  
 *
 * Subtracts a scalar to each element of a matrix
 *
 * This routine subtracts a scalar to each element of a matrix. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory 
 *     allocation failure. 
 * 
 * See also:
 *     ow_subtract_scalar_from_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int subtract_scalar_from_matrix(Matrix** target_mpp, const Matrix* source_mp, 
                                double scalar)
{
    Matrix* target_mp;
    int     i, j, num_rows, num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) - scalar;
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         multiply_matrix_by_scalar
 *
 * Multiplies each element of a matrix by a scalar
 *
 * This routine multiplies each element of a matrix by a scalar. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory 
 *     allocation failure. 
 * 
 * See also:
 *     ow_multiply_matrix_by_scalar(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrix_by_scalar(Matrix** target_mpp, const Matrix* source_mp, 
                              double scalar)
{
    Matrix* target_mp;
    int    i, j, num_rows, num_cols;
    double** source_row_ptr;
    double** target_row_ptr;
    double*  source_row_pos_ptr;
    double*  target_row_pos_ptr;
    
    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) * scalar;
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         divide_matrix_by_scalar
 *
 * Divides each element of a matrix by a scalar
 *
 * This routine divides each element of a matrix by a scalar. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. ERROR will be returned if the scalar is too close to zero.   set.
 *     This routine will fail if the scalar is too close to zero or if memory
 *     allocation failure. 
 * 
 * See also:
 *     ow_divide_matrix_by_scalar(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int divide_matrix_by_scalar(Matrix** target_mpp, const Matrix* source_mp, 
                            double scalar)
{
#ifdef SLOW
    Matrix* target_mp;
    int     i, j, num_rows, num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;

#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
    if (IS_ZERO_DBL(scalar)) 
    {
        SET_DIVIDE_BY_ZERO_ERROR();
        return ERROR;
    }
#else
    if (scalar == 0.0) 
    {
        SET_DIVIDE_BY_ZERO_ERROR();
        return ERROR;
    }
#endif 

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);
        target_row_pos_ptr = (*target_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            *target_row_pos_ptr = (*source_row_pos_ptr) / scalar;

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*target_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*target_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;

#else

    if (ABS_OF(scalar) < 10.0 * DBL_MIN)
    {
        SET_DIVIDE_BY_ZERO_ERROR(); 
        return ERROR; 
    }
    else
    {
        return multiply_matrix_by_scalar(target_mpp, source_mp, 1.0 / scalar);
    }

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_invert_matrix_elements 
 *
 * Computes the element-wise reciprocal of a matrix
 *
 * This routine replaces a matrix by its element-wise reciprocal. 
 *
 * Note:
 *     This routine overwrites its input. See invert_matrix_elements(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails if any of the matrix elements are too close to
 *     zero.
 * 
 * See also:
 *     invert_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_invert_matrix_elements(Matrix* source_mp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*source_row_pos_ptr))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else
            /*
            // Exact compare is OK  because overflow is handled separately.
            */
            if (*source_row_pos_ptr == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            (*source_row_pos_ptr) = 1.0 / (*source_row_pos_ptr);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*source_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*source_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_square_matrix_elements 
 *
 * Computes the element-wise square of a matrix
 *
 * This routine replaces a matrix by its element-wise square. 
 *
 * Note:
 *     This routine overwrites its input. See square_matrix_elements(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails if any of the matrix elements are too close to
 *     zero.
 * 
 * See also:
 *     square_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_square_matrix_elements(Matrix* source_mp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*source_row_pos_ptr) = (*source_row_pos_ptr)*(*source_row_pos_ptr);

            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_sqrt_matrix_elements(Matrix* source_mp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            if (*source_row_pos_ptr < 0.0)
            {
                set_error("Attempt to take the square root of negative number.");
                add_error("(Element (%d, %d) of a %d x %d matrix.",
                          i + 1, j + 1, num_rows, num_cols); 
                return ERROR;
            }
            else 
            {
                *source_row_pos_ptr = sqrt(*source_row_pos_ptr);
            }

            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_exp_matrix_elements 
 *
 * Computes exp(x) of matrix elements
 *
 * This routine computes exp(x) of matrix elements
 *
 * Note:
 *     This routine overwrites its input. See exp_matrix_elements(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails if any of the matrix elements are too close to
 *     zero.
 * 
 * See also:
 *     exp_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_exp_matrix_elements(Matrix* source_mp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*source_row_pos_ptr) = exp(*source_row_pos_ptr);

            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_log_matrix_elements 
 *
 * Computes log(x) of matrix elements
 *
 * This routine computes log(x) of matrix elements. If the element is zero, then
 * the argument log_zero is used instead. If the element is negative, routine
 * fails. 
 *
 * Note:
 *     This routine overwrites its input. See log_matrix_elements(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails if any of the matrix elements are negative.
 * 
 * See also:
 *     log_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_log_matrix_elements(Matrix* source_mp, double log_zero)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            double x = *source_row_pos_ptr;

            if (x < 0.0)
            {
                set_error("Attempt to take the log of negative number.");
                add_error("(Element (%d, %d) of a %d x %d matrix.",
                          i + 1, j + 1, num_rows, num_cols); 
                return ERROR;
            }
            else if (x == 0.0)
            {
                (*source_row_pos_ptr) = log_zero;
            }
            else 
            {
                (*source_row_pos_ptr) = log(x);
            }

            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         ow_add_scalar_to_matrix 
 *
 * Adds a scalar to each element of a matrix
 *
 * This routine adds a scalar to each element of a matrix. 
 *
 * Note:
 *     This routine overwrites its input. See add_scalar_to_matrix(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine cannot fail (gracefully!). 
 * 
 * See also:
 *     add_scalar_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_scalar_to_matrix(Matrix* source_mp, double scalar)
{
    int    i, j, num_rows, num_cols;
    double** source_row_ptr;
    double*  source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*source_row_pos_ptr) = (*source_row_pos_ptr) + scalar;
            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_subtract_scalar_from_matrix  
 *
 * Subtracts a scalar to each element of a matrix
 *
 * This routine subtracts a scalar to each element of a matrix. 
 *
 * Note:
 *     This routine overwrites its input. See subtract_scalar_from_matrix(3) for
 *     an alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine cannot fail (gracefully!). 
 * 
 * See also:
 *     subtract_scalar_from_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_scalar_from_matrix(Matrix* source_mp, double scalar)
{
    int i, j, num_rows, num_cols;
    double** source_row_ptr;
    double*  source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*source_row_pos_ptr) = (*source_row_pos_ptr) - scalar;
            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_multiply_matrix_by_scalar
 *
 * Multiplies each element of a matrix by a scalar
 *
 * This routine multiplies each element of a matrix by a scalar. 
 *
 * Note:
 *     This routine overwrites its input. See multiply_matrix_by_scalar(3) for
 *     an alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine cannot fail (gracefully!). 
 * 
 * See also:
 *     multiply_matrix_by_scalar(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrix_by_scalar(Matrix* source_mp, double scalar)
{
    int    i, j, num_rows, num_cols;
    double** source_row_ptr;
    double*  source_row_pos_ptr;
    
    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = (*source_row_ptr);

        for (j=0; j<num_cols; j++) 
        {
            (*source_row_pos_ptr) = (*source_row_pos_ptr) * scalar;
            source_row_pos_ptr++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_divide_matrix_by_scalar
 *
 * Divides each element of a matrix by a scalar
 *
 * This routine divides each element of a matrix by a scalar. 
 *
 * Note:
 *     This routine overwrites its input. See divide_matrix_by_scalar(3) for an
 *     alternate routine which does not. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine cannot fail (gracefully!). 
 * 
 * See also:
 *     divide_matrix_by_scalar(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrix_by_scalar(Matrix* source_mp, double scalar)
{

    if (ABS_OF(scalar) < 10.0 * DBL_MIN)
    {
        SET_DIVIDE_BY_ZERO_ERROR(); 
        return ERROR; 
    }
    else
    {
        return ow_multiply_matrix_by_scalar(source_mp, 1.0 / scalar);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                       multiply_vector_and_matrix
 *
 * Multiplies a vector by a matrix
 *
 * This routine multiplies a vector by a matrix. The vector is used as a row
 * vector. The dimensions must be appropriate for muliplication, or ERROR is
 * returned. 
 *
 * The first argument is the adress of the target vector. If the target vector
 * itself is null, then a vector of the appropriate size is created. If the
 * target vector is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: vectors, matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_vector_and_matrix(Vector** output_vpp, const Vector* input_vp, 
                               const Matrix* input_mp)
{
    Vector* output_vp;
    int i;
    int j; 
    double sum;  

    if (input_vp->length != input_mp->num_rows) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_vector(output_vpp, input_mp->num_cols));
    output_vp = *output_vpp; 

    for (j=0; j<output_vp->length; j++) 
    {
        sum = 0.0;

        for (i=0; i<input_mp->num_rows; i++) 
        {
            sum += ((input_mp->elements)[i][j]) *
                                      ((input_vp->elements)[i]); 
        }

        (output_vp->elements)[j] = sum;
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        multiply_matrix_and_vector
 *
 * Multiplies a matrix by a vector
 *
 * This routine multiplies a matrix by a vector. The vector is used as a column
 * vector. The dimensions must be appropriate for muliplication, or ERROR is
 * returned. 
 *
 * The first argument is the adress of the target vector. If the target vector
 * itself is null, then a vector of the appropriate size is created. If the
 * target vector is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: vectors, matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrix_and_vector(Vector** output_vpp, const Matrix* input_mp, 
                               const Vector* input_vp)
{
    Vector*   output_vp;
    int       i;
    int       j;
    double      sum;
    double*     vec_pos; 
    double*     mat_row_pos; 

    if (input_vp->length != input_mp->num_cols) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_vector(output_vpp, input_mp->num_rows));
    output_vp = *output_vpp; 

    for (i = output_vp->length; --i>=0; ) 
    {
        sum = 0.0;
        vec_pos = input_vp->elements;
        mat_row_pos = input_mp->elements[i];

        for (j = input_mp->num_cols; --j>=0; ) 
        {
            sum += (*mat_row_pos) * (*vec_pos); 
            mat_row_pos++;
            vec_pos++; 
        }

        (output_vp->elements)[i] = sum;
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         multiply_matrix_rows
 *
 * Multiplies each row of a matrix by each row of another
 *
 * This routine multiplies each row of a matrix by each row of another. The
 * number of columns in the two matrices must be the same.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrix_rows(Matrix** target_mpp, const Matrix* first_mp, 
                         const Matrix* second_mp)
{
    Matrix* target_mp;
    int     i, j, k;
    int     first_num_rows;
    int     second_num_rows;
    int     num_rows, num_cols;
    double**  first_row_ptr;
    double*   first_row_pos_ptr;
    double**  second_row_ptr;
    double*   second_row_pos_ptr;
    double**  target_row_ptr;
    double*   target_row_pos_ptr;
    
    if (first_mp->num_cols != second_mp->num_cols)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    first_num_rows = first_mp->num_rows; 
    second_num_rows = second_mp->num_rows; 

    num_rows = first_num_rows * second_num_rows; 
    num_cols = first_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    target_mp = *target_mpp;

    first_row_ptr = first_mp->elements;
    target_row_ptr = target_mp->elements;

    for (i=0; i<first_num_rows; i++) 
    {
        second_row_ptr = second_mp->elements;

        for (j=0; j<second_num_rows; j++) 
        {
            first_row_pos_ptr = *first_row_ptr;
            second_row_pos_ptr = *second_row_ptr;

            target_row_pos_ptr = *target_row_ptr;

            for (k=0; k<num_cols; k++) 
            {
                *target_row_pos_ptr = (*first_row_pos_ptr) * 
                                                         (*second_row_pos_ptr);
                first_row_pos_ptr++;
                second_row_pos_ptr++;
                target_row_pos_ptr++;
            }

           target_row_ptr++;  
           second_row_ptr++;  
        }

        first_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        add_row_vector_to_matrix 
 *
 * Adds a vector to each row of a matrix
 *
 * This routine adds a vector to each row of a matrix. The length of the
 * vector must match the number of columns in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_add_row_vector_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int add_row_vector_to_matrix(Matrix** target_mpp, const Matrix* source_mp, 
                             const Vector* vp)
{
    Matrix* target_mp;
    int    i;
    int    j;
    int    num_rows;
    int    num_cols;
    double** source_row_ptr;
    double** target_row_ptr;
    double*  source_row_pos_ptr;
    double*  target_row_pos_ptr;
    double*  vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *target_row_pos_ptr = (*source_row_pos_ptr) + (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
            vp_pos++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                      subtract_row_vector_from_matrix 
 *
 * Subtracts a vector from each row of a matrix
 *
 * This routine subtracts a vector to each row of a matrix. The length of the
 * vector must match the number of columns in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_subtract_row_vector_from_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int subtract_row_vector_from_matrix(Matrix** target_mpp, 
                                    const Matrix* source_mp,
                                    const Vector* vp)
{
    Matrix* target_mp;
    int    i;
    int    j;
    int    num_rows;
    int    num_cols;
    double** source_row_ptr;
    double** target_row_ptr;
    double*  source_row_pos_ptr;
    double*  target_row_pos_ptr;
    double*  vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *target_row_pos_ptr = (*source_row_pos_ptr) - (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
            vp_pos++;
        }
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           multiply_matrix_by_row_vector_ew 
 *
 * Multiplies each row of a matrix by a vector
 *
 * This routine multiplies each row of a matrix by a vector. The length of the
 * vector must match the number of columns in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_multiply_matrix_by_row_vector_ew(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrix_by_row_vector_ew(Matrix** target_mpp, 
                                     const Matrix* source_mp, 
                                     const Vector* vp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *target_row_pos_ptr = (*source_row_pos_ptr) * (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
            vp_pos++;
        }

        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         divide_matrix_by_row_vector 
 *
 * Divides each row of a matrix by a vector
 *
 * This routine divides each row of a matrix by a vector. The length of the
 * vector must match the number of columns in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine will fail if any of the matrix elements are too close
 *     to zero.  The routine will also fail due to dimension mismatch, but this
 *     is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_divide_matrix_by_row_vector(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int divide_matrix_by_row_vector(Matrix** target_mpp, 
                                const Matrix* source_mp,
                                const Vector* vp)
{
#ifdef SLOW
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
#else 
    Vector* inv_vp = NULL; 
    int result; 
#endif 
    
#ifdef SLOW

    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*vp_pos))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else
            /*
            // Exact compare is OK because overflow is handled below.
            */
            if (*vp_pos == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            *target_row_pos_ptr = (*source_row_pos_ptr) / (*vp_pos);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*target_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*target_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
            target_row_pos_ptr++;
            vp_pos++;
        }

        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;

#else

    result = invert_vector(&inv_vp, vp);

    if (result != ERROR)
    {
        result = multiply_matrix_by_row_vector_ew(target_mpp, source_mp, inv_vp);
    }

    free_vector(inv_vp); 

    return result; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          ow_add_row_vector_to_matrix 
 *
 * Adds a vector to each row of a matrix. 
 *
 * This routine adds a vector to each row of the input matrix. The length of the
 * vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     add_row_vector_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_row_vector_to_matrix(Matrix* source_mp, const Vector* vp)
{
    int    i;
    int    j;
    int    num_rows;
    int    num_cols;
    double** source_row_ptr;
    double*  source_row_pos_ptr;
    double*  vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr += (*vp_pos);
            source_row_pos_ptr++;
            vp_pos++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         ow_subtract_row_vector_from_matrix 
 *
 * Subtracts a vector to each row of a matrix. 
 *
 * This routine subtracts a vector to each row of the input matrix. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     subtract_vector_to_matrix_rows(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_row_vector_from_matrix(Matrix* source_mp, const Vector* vp)
{
    int    i;
    int    j;
    int    num_rows;
    int    num_cols;
    double** source_row_ptr;
    double*  source_row_pos_ptr;
    double*  vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr -= (*vp_pos);
            source_row_pos_ptr++;
            vp_pos++;
        }
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                        ow_multiply_matrix_by_row_vector_ew
 *
 * Multiplies each row of a matrix by a vector
 *
 * This routine multiplies each row of the input matrix by a vector. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     multiply_matrix_by_row_vector_ew(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrix_by_row_vector_ew(Matrix* source_mp, const Vector* vp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr *= (*vp_pos);
            source_row_pos_ptr++;
            vp_pos++;
        }

        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_divide_matrix_by_row_vector
 *
 * Divides each row of a matrix by a vector
 *
 * This routine divides each row of the input matrix by a vector. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * See also:
 *     divide_matrix_by_row_vector(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrix_by_row_vector(Matrix* mp, const Vector* vp)
{
#ifdef SLOW
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  row_ptr;
    double*   row_pos_ptr;
    double*   vp_pos;
#else
    Vector* inv_vp = NULL; 
    int result; 
#endif 
    
#ifdef SLOW
    if (mp->num_cols != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_ptr = mp->elements;

    num_rows = mp->num_rows;
    num_cols = mp->num_cols;

    for (i=0; i<num_rows; i++) 
    {
        row_pos_ptr = *row_ptr;
        vp_pos = vp->elements;

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*vp_pos))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else
            /*
            // Exact compare is OK because overflow is handled below.
            */
            if (*vp_pos == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            *row_pos_ptr /= *vp_pos;

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            row_pos_ptr++;
            vp_pos++;
        }

        row_ptr++;
    }

    return NO_ERROR;

#else

    result = invert_vector(&inv_vp, vp);

    if (result != ERROR)
    {
        result = ow_multiply_matrix_by_row_vector_ew(mp, inv_vp);
    }

    free_vector(inv_vp); 

    return result; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_add_vector_to_matrix_row
 *
 * Adds a vector to a specified row of a matrix
 *
 * This routine adds a vector to a specified row of the input matrix. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_vector_to_matrix_row(Matrix* mp, const Vector* vp, int row)
{
    int     j;
    int     num_cols;
    double*   row_pos;
    double*   vp_pos;
    
    if (    (mp->num_cols != vp->length)
         || (row < 0)
         || (row >= mp->num_rows)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;
    vp_pos = vp->elements;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos += *vp_pos;

        row_pos++;
        vp_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_subtract_vector_from_matrix_row
 *
 * Subtracs a vector from the specified row of a matrix
 *
 * This routine adds a vector to a specified row of the input matrix. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_vector_from_matrix_row(Matrix* mp, const Vector* vp, int row)
{
    int     j;
    int     num_cols;
    double*   row_pos;
    double*   vp_pos;
    
    if (    (mp->num_cols != vp->length)
         || (row < 0)
         || (row >= mp->num_rows)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;
    vp_pos = vp->elements;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos -= *vp_pos;

        row_pos++;
        vp_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_multiply_matrix_row_by_vector
 *
 * Multiplies a specified row of a matrix by a vector
 *
 * This routine multiplies a specified row of a matrix by a vector. The length
 * of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrix_row_by_vector(Matrix* mp, const Vector* vp, int row)
{
    int     j;
    int     num_cols;
    double*   row_pos;
    double*   vp_pos;
    
    if (    (mp->num_cols != vp->length)
         || (row < 0)
         || (row >= mp->num_rows)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;
    vp_pos = vp->elements;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos *= *vp_pos;

        row_pos++;
        vp_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_divide_matrix_row_by_vector
 *
 * Divides a specified row of a matrix by a vector
 *
 * This routine divides a specified row of the input matrix by a vector. The
 * length of the vector must match the number of columns in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrix_row_by_vector(Matrix* mp, const Vector* vp, int row)
{
#ifdef SLOW
    int     j;
    int     num_cols;
    double*   row_pos;
    double*   vp_pos;
#else 
    Vector* inv_vp = NULL; 
    int result; 
#endif 

#ifdef SLOW
    if (    (mp->num_cols != vp->length)
         || (row < 0)
         || (row >= mp->num_rows)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;
    vp_pos = vp->elements;

    for (j=0; j<num_cols; j++) 
    {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
        if (IS_ZERO_DBL(*vp_pos))
        {
            SET_DIVIDE_BY_ZERO_ERROR();
            return ERROR; 
        }
#else
        /*
        // Exact compare is OK because overflow is handled below.
        */
        if (*vp_pos == 0.0)
        {
            SET_DIVIDE_BY_ZERO_ERROR();
            return ERROR; 
        }
#endif 

        *row_pos /= *vp_pos;

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
        if (    (*row_pos > DBL_MOST_POSITIVE) 
             || (*row_pos < DBL_MOST_NEGATIVE)
           )
        {
            SET_OVERFLOW_ERROR();
            return ERROR; 
        }
#endif 

        row_pos++;
        vp_pos++;
    }

    return NO_ERROR;

#else

    result = invert_vector(&inv_vp, vp);

    if (result != ERROR)
    {
        result = ow_multiply_matrix_row_by_vector(mp, inv_vp, row);
    }

    free_vector(inv_vp); 

    return result; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_add_scalar_to_matrix_row
 *
 * Adds a scalar to a specified row of a matrix
 *
 * This routine adds a scalar to a specified row of the input matrix. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_scalar_to_matrix_row(Matrix* mp, double scalar, int row)
{
    int   j;
    int   num_cols;
    double* row_pos;
    
    if ((row < 0) || (row >= mp->num_rows))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos += scalar; 
        row_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_subtract_scalar_from_matrix_row
 *
 * Subtracts a scalar from a specified row of a matrix
 *
 * This routine subtracts a scalar from a specified row of the input matrix.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_scalar_from_matrix_row(Matrix* mp, double scalar, int row)
{
    int   j;
    int   num_cols;
    double* row_pos;
    
    if ((row < 0) || (row >= mp->num_rows))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos -= scalar; 
        row_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_multiply_matrix_row_by_scalar
 *
 * Multiplies a specified row of a matrix by a scalar
 *
 * This routine multiplies the specified row of a matrix by a scalar.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrix_row_by_scalar(Matrix* mp, double scalar, int row)
{
    int   j;
    int   num_cols;
    double* row_pos;
    
    if ((row < 0) || (row >= mp->num_rows))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;

    for (j=0; j<num_cols; j++) 
    {
        *row_pos *= scalar; 
        row_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_divide_matrix_row_by_scalar
 *
 * Divides a specified row of a matrix by a scalar
 *
 * This routine divides a specified row of the input matrix by a scalar.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails If any of the matrix elements are too close to
 *     zero. 
 * 
 * See also:
 *     divide_matrix_by_row_vector(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrix_row_by_scalar(Matrix* mp, double scalar, int row)
{
#ifdef SLOW
    int   j;
    int   num_cols;
    double* row_pos;
    
    if ((row < 0) || (row >= mp->num_rows))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
    if (IS_ZERO_DBL(scalar)
    {
        SET_DIVIDE_BY_ZERO_ERROR();
        return ERROR; 
    }
#else
    /*
    // Exact compare is OK because overflow is handled below.
    */
    if (scalar == 0.0)
    {
        SET_DIVIDE_BY_ZERO_ERROR();
        return ERROR; 
    }
#endif 

    row_pos = *(mp->elements + row);
    num_cols = mp->num_cols;

    for (j=0; j<num_cols; j++) 
    {

        ASSERT_IS_NUMBER_DBL(*row_pos);
        ASSERT_IS_FINITE_DBL(*row_pos);

        ASSERT_IS_NUMBER_DBL(scalar);
        ASSERT_IS_FINITE_DBL(scalar);
        ASSERT_IS_GREATER_DBL(scalar, 10.9 * DBL_MIN); 

        *row_pos /= scalar; 

        ASSERT_IS_NUMBER_DBL(*row_pos);
        ASSERT_IS_FINITE_DBL(*row_pos);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
        if (    (*row_pos > DBL_MOST_POSITIVE) 
             || (*row_pos < DBL_MOST_NEGATIVE)
           )
        {
            SET_OVERFLOW_ERROR();
            return ERROR; 
        }
#endif 

        row_pos++;
    }

    return NO_ERROR;

#else

    if (ABS_OF(scalar) < 10.0 * DBL_MIN)
    {
        SET_DIVIDE_BY_ZERO_ERROR(); 
        return ERROR; 
    }
    else
    {
        return ow_multiply_matrix_row_by_scalar(mp, 1.0 / scalar, row);
    }

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         add_col_vector_to_matrix 
 *
 * Adds a vector to each column of a matrix
 *
 * This routine adds a vector to each column of a matrix. The length of the
 * vector must match the number of rows in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_add_col_vector_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int add_col_vector_to_matrix(Matrix** target_mpp, 
                             const Matrix* source_mp, 
                             const Vector* vp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) + (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                             ow_add_col_vector_to_matrix 
 *
 * Adds a vector to each column of a matrix. 
 *
 * This routine adds a vector to each column of the input matrix. The length of
 * the vector must match the number of rows in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     add_col_vector_to_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_add_col_vector_to_matrix(Matrix* source_mp, const Vector* vp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr += (*vp_pos);
            source_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                       subtract_col_vector_from_matrix 
 *
 * Subtracts a vector to each column of a matrix
 *
 * This routine subtracts a vector to each column of a matrix. The length of the
 * vector must match the number of rows in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_subtract_col_vector_from_matrix(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int subtract_col_vector_from_matrix(Matrix** target_mpp, 
                                    const Matrix* source_mp,
                                    const Vector* vp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) - (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          ow_subtract_col_vector_from_matrix 
 *
 * Subtracts a vector to each column of a matrix. 
 *
 * This routine subtracts a vector to each column of the input matrix. The
 * length of the vector must match the number of rows in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     subtract_vector_to_matrix_cols(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_col_vector_from_matrix(Matrix* source_mp, const Vector* vp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr -= (*vp_pos);
            source_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                         multiply_matrix_by_col_vector_ew 
 *
 * Multiplies each column of a matrix by a vector
 *
 * This routine multiplies each column of a matrix by a vector. The length of
 * the vector must match the number of rows in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can only fail (gracefully!) due to memory
 *     allocation failure. The routine will also fail due to dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     ow_multiply_matrix_by_col_vector_ew(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int multiply_matrix_by_col_vector_ew(Matrix** target_mpp, 
                                     const Matrix* source_mp, 
                                     const Vector* vp)
{
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            (*target_row_pos_ptr) = (*source_row_pos_ptr) * (*vp_pos);
            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                      ow_multiply_matrix_by_col_vector_ew
 *
 * Multiplies each column of a matrix by a vector
 *
 * This routine multiplies each column of the input matrix by a vector. The
 * length of the vector must match the number of rows in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. Currenly this routine can't fail gracefully.  The routine will fail
 *     due to dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     multiply_matrix_by_col_vector_ew(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_multiply_matrix_by_col_vector_ew(Matrix* source_mp, const Vector* vp)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    double*   vp_pos;
    
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
            *source_row_pos_ptr *= (*vp_pos);
            source_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          divide_matrix_by_col_vector 
 *
 * Divides each column of a matrix by a vector
 *
 * This routine divides each column of a matrix by a vector. The length of the
 * vector must match the number of rows in the matrix.  
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine fails if any of the matrix elements are too close to
 *     zero, or if memory allocation fails. The routine will also fail due to
 *     dimension mismatch, but this is currently treated as a bug (see
 *     set_bug()). 
 * 
 * See also:
 *     ow_divide_matrix_by_col_vector(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int divide_matrix_by_col_vector(Matrix** target_mpp, 
                                const Matrix* source_mp, 
                                const Vector* vp)
{
#ifdef SLOW
    Matrix* target_mp;
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double**  target_row_ptr;
    double*   source_row_pos_ptr;
    double*   target_row_pos_ptr;
    double*   vp_pos;
#else 
    Vector* inv_vp = NULL; 
    int result; 
#endif 
    
#ifdef SLOW
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_target_matrix(target_mpp, source_mp->num_rows,
                          source_mp->num_cols));
    target_mp = *target_mpp;

    source_row_ptr = source_mp->elements;
    target_row_ptr = target_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;
        target_row_pos_ptr = *target_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*vp_pos))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else
            /*
            // Exact compare is OK because overflow is handled below.
            */
            if (*vp_pos == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            (*target_row_pos_ptr) = (*source_row_pos_ptr) / (*vp_pos);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*target_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*target_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
            target_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
        target_row_ptr++;
    }

    return NO_ERROR;

#else

    result = invert_vector(&inv_vp, vp);

    if (result != ERROR)
    {
        result = multiply_matrix_by_col_vector_ew(target_mpp, source_mp, inv_vp);
    }

    free_vector(inv_vp); 

    return result; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                           ow_divide_matrix_by_col_vector
 *
 * Divides each column of a matrix by a vector
 *
 * This routine divides each column of the input matrix by a vector. The length
 * of the vector must match the number of rows in the matrix.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set. This routine will fail if any of the matrix elements are too close
 *     to zero, This routine will also fail if there is a dimension mismatch,
 *     but this is currently treated as a bug (see set_bug()). 
 * 
 * See also:
 *     divide_matrix_by_col_vector(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_divide_matrix_by_col_vector(Matrix* source_mp, const Vector* vp)
{
#ifdef SLOW
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    double**  source_row_ptr;
    double*   source_row_pos_ptr;
    double*   vp_pos;
#else 
    Vector* inv_vp = NULL; 
    int result; 
#endif 
    
#ifdef SLOW
    if (source_mp->num_rows != vp->length)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    source_row_ptr = source_mp->elements;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;
    vp_pos = vp->elements;

    for (i=0; i<num_rows; i++) 
    {
        source_row_pos_ptr = *source_row_ptr;

        for (j=0; j<num_cols; j++) 
        {
#ifdef USE_IS_ZERO_DBL_FOR_DIVIDE_BY_ZERO
            if (IS_ZERO_DBL(*vp_pos))
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#else
            /*
            // Exact compare is OK because overflow is handled below.
            */
            if (*vp_pos == 0.0)
            {
                SET_DIVIDE_BY_ZERO_ERROR();
                return ERROR; 
            }
#endif 

            *source_row_pos_ptr /= (*vp_pos);

#ifdef USE_COMPARE_WITH_REAL_MAX_FOR_DIVIDE_BY_ZERO
            if (    (*source_row_pos_ptr > DBL_MOST_POSITIVE) 
                 || (*source_row_pos_ptr < DBL_MOST_NEGATIVE)
               )
            {
                SET_OVERFLOW_ERROR();
                return ERROR; 
            }
#endif 

            source_row_pos_ptr++;
        }

        vp_pos++;
        source_row_ptr++;
    }

    return NO_ERROR;

#else

    result = invert_vector(&inv_vp, vp);

    if (result != ERROR)
    {
        result = ow_multiply_matrix_by_col_vector_ew(source_mp, inv_vp);
    }

    free_vector(inv_vp); 

    return result; 

#endif 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              sum_matrix_elements
 *
 * Adds the elements of a matrix
 *
 * This routine returns the sum of the elements of the matrix pointed to by mp  
 *
 * Returns:
 *    The sum of the elements of the matrix pointed to by mp. 
 * 
 * See also:
 *     average_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

double sum_matrix_elements(const Matrix* mp)
{
    int    i;
    int    j;
    double   sum;
    double** row_ptr;
    double*  element_ptr;

    if (mp->num_rows == 0) return 0.0;
    if (mp->num_cols == 0) return 0.0;

    sum = 0.0; 

    row_ptr =  mp->elements;   

    for (i = 0; i<mp->num_rows; i++) 
    {
        element_ptr = *row_ptr; 

        for (j = 0; j<mp->num_cols; j++) 
        {
            sum += *element_ptr; 
            element_ptr++;
        }
        row_ptr++; 
    }

     return sum; 
 }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                          sum_matrix_row_elements
 *
 * Returns the sum of a specified matrix row
 *
 * This routine returns the sum of the elements of a specified row of the
 * matrix pointed to by mp  
 *
 * Returns:
 *    The sum of the elements of the specified row of the matrix pointed to by
 *    mp. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

double sum_matrix_row_elements(const Matrix* mp, int row)
{
    int   j;
    double  sum         = 0.0;
    double* element_ptr = *(mp->elements + row);

    for (j = 0; j<mp->num_cols; j++) 
    {
        sum += *element_ptr; 
        element_ptr++;
    }

     return sum; 
 }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              sum_matrix_col_elements
 *
 * Returns the sum of a specified matrix column
 *
 * This routine returns the sum of the elements of a specified column of the
 * matrix pointed to by mp  
 *
 * Returns:
 *    The sum of the elements of the specified column of the matrix pointed to
 *    by mp. 
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

double sum_matrix_col_elements(const Matrix* mp, int row)
{
    int  i;
    double sum = 0.0;

    for (i = 0; i<mp->num_rows; i++) 
    {
        sum += mp->elements[ row ][ i ]; 
    }

     return sum; 
 }

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              average_matrix_elements
 *
 * Averages the elements of a matrix
 *
 * This routine returns the average of the elements of the matrix pointed to by
 * mp  
 *
 * Returns:
 *    The average of the elements of the matrix pointed to by mp. 
 * 
 * See also:
 *     sum_matrix_elements(3)
 *     
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

double average_matrix_elements(const Matrix* mp)
{

    if (mp->num_rows == 0) return 0.0;
    if (mp->num_cols == 0) return 0.0;

    return sum_matrix_elements(mp) / ((mp->num_rows) * (mp->num_cols));
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                       ow_subtract_identity_matrix 
 *
 * Subtracts the identity matrix from a square matrix. 
 *
 * This routine subtracts the identity matrix from a square matrix. The matrix
 * is over-written. If the matrix is not square, this is treated as a bug.  
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int ow_subtract_identity_matrix(Matrix* mp)
{
    int num_rows = mp->num_rows;
    int num_cols = mp->num_cols;
    int i; 

    if (num_cols != num_rows)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        mp->elements[ i ][ i ] -= 1.0;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            do_matrix_recomposition
 *
 * Computes A*diag*B
 *
 * This routine multiplies a matrix by a a diagonal matrix stored as a vector,
 * and then multiplies the result be a third matrix. The matrices must be
 * compatable for muliplication, or ERROR is returned. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int do_matrix_recomposition(Matrix** target_mpp, 
                            const Matrix* first_mp, 
                            const Vector* vp, 
                            const Matrix* second_mp)
{
    Matrix* temp_target_mp = NULL;
    int     result = NO_ERROR;
    
    ERE(multiply_matrix_by_col_vector_ew(&temp_target_mp, first_mp, vp));

    if (multiply_matrices(target_mpp, temp_target_mp, second_mp) == ERROR)
    {
        result = ERROR;
    }

    free_matrix(temp_target_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            do_matrix_recomposition_2
 *
 * Computes A*diag*B'
 *
 * This routine multiplies a matrix by a a diagonal matrix stored as a vector
 * by a matrix transpose. The matrices must be compatable for muliplication, or
 * ERROR is returned. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: matrices, matrix arithmetic
 *
 * -----------------------------------------------------------------------------
 */

int do_matrix_recomposition_2
(
    Matrix**      target_mpp, 
    const Matrix* first_mp, 
    const Vector* vp, 
    const Matrix* second_mp
)
{
    Matrix* temp_target_mp = NULL;
    int     result = NO_ERROR;
    
    ERE(multiply_matrix_by_col_vector_ew(&temp_target_mp, first_mp, vp));

    if (multiply_by_transpose(target_mpp, temp_target_mp, second_mp) == ERROR)
    {
        result = ERROR;
    }

    free_matrix(temp_target_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

