
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "m/m_mat_basic.h" 
#include <math.h>

/* -------------------------------------------------------------------------- */

/*
 * =============================================================================
 *                               create_matrix_copy
 *
 * Creates a copy of a matrix
 *
 * This routine creates a copy of a matrix and returns a pointer to it. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then NULL is returned
 *    with and error message being set. Otherwise a pointer to a newly created
 *    copy of the input matrix is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

Matrix* create_matrix_copy(const Matrix* source_mp)
{
    Matrix* target_mp = NULL;

    ERN(copy_matrix(&target_mp, source_mp));

    return target_mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                copy_matrix
 *                               
 * Copies a matrix
 *
 * This routine copies the matrix into the one pointed to by target_mpp. If
 * the matrix pointed to by target_mpp is NULL, then a matrix of the
 * appropriate size is created. If it exists, but is the wrong size, then it is
 * recycled. Otherwise, the storage is recycled. If the source matrix is NULL,
 * then the target matrix becomes NULL also, and any storage associated with it
 * is freed. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

int debug_copy_matrix
(
    Matrix**      target_mpp,
    const Matrix* source_mp, 
    const char*   file_name, 
    int           line_number
)
{
    int   i;
    int   j;
    int   num_rows; 
    int   num_cols; 
    double* target_pos;
    double* source_pos;

    if (source_mp == NULL) 
    {
        free_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(debug_get_target_matrix(target_mpp, num_rows, num_cols, file_name, 
                                line_number));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

int copy_matrix(Matrix** target_mpp, const Matrix* source_mp)
{
    int   i;
    int   j;
    int   num_rows; 
    int   num_cols; 
    double* target_pos;
    double* source_pos;

    if (source_mp == NULL) 
    {
        free_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            copy_matrix_block
 *                               
 * Copies a matrix block to target matrix
 *
 * This routine copies a block of source_mp into *target_mpp which is created or
 * resized if necessary.
 *
 * Returns :
 *    If the routine fails then ERROR is returned with and error message being
 *    set.  Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int copy_matrix_block
(
    Matrix**      target_mpp,
    const Matrix* source_mp, 
    int           source_row_offset,
    int           source_col_offset,
    int           num_rows,
    int           num_cols
)
{

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    ERE(ow_copy_matrix_block(*target_mpp, 0, 0, 
                             source_mp, source_row_offset, source_col_offset,
                             num_rows, num_cols));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            ow_copy_matrix_block
 *                               
 * Copies a matrix block to an offset position of a target matrix
 *
 * This routine copies a block of source_mp into target_mp, offset by
 * (target_row_offset, target_col_offset). 
 *
 * Returns :
 *    If the routine fails (can really happen unless there is an argument
 *    error), then ERROR is returned with and error message being set.
 *    Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int ow_copy_matrix_block
(
    Matrix*       target_mp,
    int           target_row_offset, 
    int           target_col_offset, 
    const Matrix* source_mp, 
    int           source_row_offset, 
    int           source_col_offset,
    int           num_rows,
    int           num_cols
)
{
    int   i;
    int   j;
    double* target_pos;
    double* source_pos;

    if (    (target_row_offset + num_rows > target_mp->num_rows)
         || (target_col_offset + num_cols > target_mp->num_cols)
         || (source_row_offset + num_rows > source_mp->num_rows)
         || (source_col_offset + num_cols > source_mp->num_cols)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        source_pos = source_mp->elements[ source_row_offset + i ] + source_col_offset;
        target_pos = target_mp->elements[ target_row_offset + i ] + target_col_offset;
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            copy_matrix_block_2
 *                               
 * Copies a matrix block to target matrix
 *
 * This routine copies a block of source_mp into *target_mpp which is created or
 * resized if necessary. It has further functionality from copy_matrix_block, in
 * that the copying can jump rows/cols in the source matrix. 
 *
 * Returns :
 *    If the routine fails then ERROR is returned with and error message being
 *    set.  Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int copy_matrix_block_2
(
    Matrix**      target_mpp,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           source_row_step,
    int           source_col_step,
    int           num_rows,
    int           num_cols 
)
{

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    ERE(ow_copy_matrix_block_2(*target_mpp, 0, 0, 1, 1,
                               source_mp, source_row_offset, source_col_offset,
                               source_row_step, source_col_step, 
                               num_rows, num_cols));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            ow_copy_matrix_block_2
 *                               
 * Copies a matrix block to an offset position of a target matrix
 *
 * This routine copies a block of source_mp into target_mp, offset by
 * (target_row_offset, target_col_offset). If source_row_step or source_col_step
 * are greater than 1, then the copying jumps rows/columns in the source matrix.
 * If target_row_step or target_col_step are greater than one, then the copying
 * jumps rows/columns in the target matrix. 
 *
 * Returns :
 *    If the routine fails (can really happen unless there is an argument
 *    error), then ERROR is returned with and error message being set.
 *    Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int ow_copy_matrix_block_2
(
    Matrix*       target_mp,
    int           target_row_offset,
    int           target_col_offset,
    int           target_row_step,
    int           target_col_step,
    const Matrix* source_mp,
    int           source_row_offset,
    int           source_col_offset,
    int           source_row_step,
    int           source_col_step,
    int           num_rows,
    int           num_cols 
)
{
    int   i;
    int   j;
    double* target_pos;
    double* source_pos;

    /*
    dbi(target_mp->num_rows);
    dbi(target_mp->num_cols);
    dbi(target_row_offset); 
    dbi(target_col_offset); 
    dbi(target_row_step); 
    dbi(target_col_step); 

    dbi(source_mp->num_rows);
    dbi(source_mp->num_cols);
    dbi(source_row_offset); 
    dbi(source_col_offset); 
    dbi(source_row_step); 
    dbi(source_col_step); 

    dbi(num_rows); 
    dbi(num_cols); 
    */

    if (source_row_step < 1) source_row_step = 1; 
    if (source_col_step < 1) source_col_step = 1; 
    if (target_row_step < 1) target_row_step = 1; 
    if (target_col_step < 1) target_col_step = 1; 

    if (    (target_row_offset > 0)
         || (source_row_offset > 0)
       )
    {
    }

    if (    (target_row_offset + target_row_step * (num_rows - 1) >= target_mp->num_rows)
         || (target_col_offset + target_col_step * (num_cols - 1) >= target_mp->num_cols)
         || (source_row_offset + source_row_step * (num_rows - 1) >= source_mp->num_rows)
         || (source_col_offset + source_col_step * (num_cols - 1) >= source_mp->num_cols)
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        source_pos = source_mp->elements[ source_row_offset + i * source_row_step ] + source_col_offset;
        target_pos = target_mp->elements[ target_row_offset + i * target_row_step ] + target_col_offset;
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos += target_col_step;
            source_pos += source_col_step;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            ow_copy_matrix
 *                               
 * Copies a matrix to an offset position of a target matrix
 *
 * This routine copies source_mp into target_mp, offset by (target_row_offset,
 * target_col_offset). 
 *
 * Returns :
 *    If the routine fails (can really happen unless there is an argument
 *    error), then ERROR is returned with and error message being set.
 *    Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int ow_copy_matrix
(
    Matrix*       target_mp,
    int           target_row_offset,
    int           target_col_offset,
    const Matrix* source_mp 
)
{
    int   i;
    int   j;
    int   num_rows; 
    int   num_cols; 
    double* target_pos;
    double* source_pos;

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    if (target_row_offset + num_rows > target_mp->num_rows)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (target_col_offset + num_cols > target_mp->num_cols)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        source_pos = source_mp->elements[ i ];
        target_pos = target_mp->elements[ target_row_offset + i ] + target_col_offset;
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = *source_pos;
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int copy_matrix_data
(
    Matrix**      target_mpp,
    const Matrix* mp,
    const int*    invalid_data 
)
{

    if ((invalid_data == NULL) || (mp == NULL))
    {
        return copy_matrix(target_mpp, mp);
    }
    else 
    {
        int   count      = 0;
        int   i, j;
        int   num_rows   = mp->num_rows;
        int   num_cols   = mp->num_cols;
        double* target_pos;
        double* source_pos;

        ERE(get_target_matrix(target_mpp, num_rows, num_cols));

        for (i=0; i<num_rows; i++)
        {
            if ( ! invalid_data[ i ])
            {
                source_pos = (mp->elements)[ i ];
                target_pos = ((*target_mpp)->elements)[ count ];
               
                for (j=0; j<num_cols; j++)
                {
                    *target_pos++ = *source_pos++;
                }

                count++;
            }
        }

        if (count == 0)
        {
            set_error("No valid data.");
            free_matrix(*target_mpp);
            *target_mpp = NULL;
            return ERROR;
        }
        else
        {
            (*target_mpp)->num_rows = count;
            return NO_ERROR;
        }
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          copy_int_matrix_to_matrix
 *                               
 * Copies an integer matrix to a matrix 
 *
 * This routine copies the integer matrix source_mp to the one pointed to by
 * target_mpp. It is thus used mostly for caching conversion results for
 * performance. If the matrix pointed to by target_mpp is NULL, then a matrix
 * of the appropriate size is created. If it exists, but is the wrong size,
 * then it is recycled. Otherwise, the storage is recycled. If the source
 * matrix is NULL, then the target matrix becomes NULL also, and any storage
 * associated with it is freed. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int copy_int_matrix_to_matrix
(
    Matrix**          target_mpp,
    const Int_matrix* source_mp 
)
{
    int     i, j, num_rows, num_cols;
    double* target_pos;
    int* source_pos;

    if (source_mp == NULL) 
    {
        free_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos++ = *source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                          copy_matrix_to_int_matrix
 *                               
 * Copies a matrix to an integer matrix 
 *
 * This routine copies the matrix source_mp to the integer matrix pointed to by
 * target_mpp. It is thus used mostly for caching conversion results for
 * performance. If the matrix pointed to by target_mpp is NULL, then a matrix
 * of the appropriate size is created. If it exists, but is the wrong size,
 * then it is recycled. Otherwise, the storage is recycled. If the source
 * matrix is NULL, then the target matrix becomes NULL also, and any storage
 * associated with it is freed. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int copy_matrix_to_int_matrix
(
    Int_matrix**  target_mpp,
    const Matrix* source_mp 
)
{
    int     i;
    int     j;
    int     num_rows;
    int     num_cols;
    int*    target_pos;
    double* source_pos;

    if (source_mp == NULL) 
    {
        free_int_matrix(*target_mpp);
        *target_mpp = NULL;
        return NO_ERROR;
    }

    num_rows = source_mp->num_rows;
    num_cols = source_mp->num_cols;

    ERE(get_target_int_matrix(target_mpp, num_rows, num_cols));

    for (i=0; i<num_rows; i++)
    {
        source_pos = (source_mp->elements)[ i ];
        target_pos = ((*target_mpp)->elements)[ i ];
       
        for (j=0; j<num_cols; j++)
        {
            *target_pos = ROUND_DBL_TO_INT(*source_pos);
            target_pos++;
            source_pos++;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int copy_matrix_vector
(
    Matrix_vector**      out_mvpp,
    const Matrix_vector* in_mvp 
)
{
    Matrix**      out_mp2; 
    Matrix**      in_mp2; 
    int           num_matrices;
    int           i;

    if (in_mvp == NULL)
    {
        free_matrix_vector(*out_mvpp);
        *out_mvpp = NULL;
        return NO_ERROR;
    }

    num_matrices = in_mvp->length;

    if (    (*out_mvpp != NULL)
         && ((*out_mvpp)->length != num_matrices)
       )
    {
        free_matrix_vector(*out_mvpp);
        *out_mvpp = NULL;
    }

    if (*out_mvpp == NULL)
    {
        NRE(*out_mvpp = create_matrix_vector(num_matrices));
    }

    out_mp2 = (*out_mvpp)->elements;
    in_mp2  = in_mvp->elements;

    for (i=0; i<num_matrices; i++) 
    {
        ERE(copy_matrix(&(out_mp2[ i ]), in_mp2[ i ]));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int concat_matrix_vectors
(
    Matrix_vector**       mvpp,
    int                   num_vectors,
    const Matrix_vector** matrix_vectors 
)
{
    int i, length = 0;
    Matrix** target_pos; 

    for (i = 0; i < num_vectors; i++)
    {
        if (matrix_vectors[ i ] != NULL)
        {
            length += matrix_vectors[ i ]->length;
        }
    }

    ERE(get_target_matrix_vector(mvpp, length));
    target_pos = (*mvpp)->elements;

    for (i = 0; i < num_vectors; i++)
    {
        const Matrix_vector* source_mvp    = matrix_vectors[ i ];

        if (source_mvp != NULL)
        {
            int                  source_length = source_mvp->length;
            Matrix**             source_pos    = source_mvp->elements;
            int                  j;

            for (j = 0; j < source_length; j++)
            {
                ERE(copy_matrix(target_pos, *source_pos));
                target_pos++;
                source_pos++;
            }
        }
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int mp_row_get_indexed_vector
(
    Indexed_vector** vpp,
    Matrix*          init_mp,
    int              row 
)
{
    Indexed_vector_element* vp_pos;
    int length = init_mp->num_cols; 
    double* mp_row_pos = init_mp->elements[ row ]; 
    int             i;

    ERE(get_target_indexed_vector(vpp, length));

    vp_pos = (*vpp)->elements; 

    for (i=0; i<length; i++)
    {
        vp_pos->index = i;
        vp_pos->element = *mp_row_pos;
        vp_pos++; 
        mp_row_pos++; 
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_random_matrix_row
 *                               
 * Gets a random matrix row
 * 
 * This routine gets a random row from the matrix "mp" and puts it into the
 * vector pointed to by vpp. If *vpp is NULL, then the vector is created.
 * If it is the wrong length, then it is resized. Otherwise, the storage is
 * reused. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise the index of the row selected
 *    is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_random_matrix_row(Vector** vpp, const Matrix* mp)
{
    int index = (int)(kjb_rand_2() * mp->num_rows); 

    if (index == mp->num_rows) index--; 

    ERE(get_matrix_row(vpp, mp, index));

    return index;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_matrix_row
 *                               
 * Returns a matrix row
 * 
 * This routine gets row "row_num" from the matrix "mp" and puts it into the
 * vector pointed to by vpp. If *vpp is NULL, then the vector is created.
 * If it is the wrong length, then it is resized. Otherwise, the storage is
 * reused.
 * 
 * Note : 
 *     Matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_matrix_row(Vector** vpp, const Matrix* mp, int row_num)
{
    int     len;
    int     i;
    double*   row_pos;
    double*   vect_pos;

    if ((row_num < 0) || (row_num >= mp->num_rows)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_cols; 

    ERE(get_target_vector(vpp, len)); 

    row_pos  = mp->elements[ row_num ];
    vect_pos = (*vpp)->elements;

    for (i=0; i<len; i++)
    {
        *vect_pos++ = *row_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                remove_matrix_row
 *                               
 * Removes a matrix row
 * 
 * This routine removes row "row_num" from the matrix "mp" and optionally puts
 * it into the vector pointed to by vpp (if vpp is not NULL). If *vpp
 * is NULL, then the vector is created.  If it is the wrong length, then it is
 * resized. Otherwise, the storage is reused.
 * 
 * Note: 
 *     Matrix row counting begins with 0. 
 *
 * Note
 *    Currently, trying to remove the only row of matrix is treated as a bug.
 *    This may change with future versions. 
 *
 * Returns:
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int remove_matrix_row(Vector** vpp, Matrix* mp, int row_num)
{
    int len, i, j;
    int num_rows = mp->num_rows;

    /*
    // FIX 
    //
    // Currently, lets not deal with single row matrices. Once we implement
    // resziable matrices, then this restriction should be relaxed.
    */
    if ((row_num < 0) || (row_num >= num_rows) || (num_rows <= 1)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_cols; 

    if (vpp != NULL)
    {
        ERE(get_target_vector(vpp, len)); 
    }

    num_rows--;

    for (j=0; j<len; j++)
    {
        if (vpp != NULL)
        {
            (*vpp)->elements[ j ] = mp->elements[ row_num ][ j ];
        }

        for (i=row_num; i<num_rows; i++)
        {
            mp->elements[ i ][ j ] = mp->elements[ i + 1 ][ j ];
        }
    }

    mp->num_rows = num_rows;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                put_matrix_row
 *                               
 * Replaces a matrix row with the contents of a vector
 * 
 * This routine replaces row "row_num" of the matrix "mp" with the contents of 
 * vector vp. 
 * 
 * Note : 
 *     Matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int put_matrix_row(Matrix* mp, const Vector* vp, int row_num)
{

    int len, i;
    double* row_pos;
    double* vect_pos; 

    if (    (row_num < 0) || (row_num >= mp->num_rows)
         || (mp->num_cols != vp->length)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = vp->length; 

    row_pos = mp->elements[ row_num ];
    vect_pos = vp->elements;

    for (i=0; i<len; i++)
    {
        *row_pos++ = *vect_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                copy_matrix_row
 *                               
 * Replaces a matrix row with a row from another
 * 
 * This routine replaces row "target_row_num" of the matrix "target_mp" with the
 * contents of "source_row_num" of "source_mp". 
 * 
 * Note : 
 *     Matrix row counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int copy_matrix_row
(
    Matrix*       target_mp,
    int           target_row_num,
    const Matrix* source_mp,
    int           source_row_num 
)
{

    int len, i;
    double* target_row_pos;
    double* source_row_pos; 

    if (    (source_row_num < 0) || (source_row_num >= source_mp->num_rows)
         || (target_row_num < 0) || (target_row_num >= target_mp->num_rows)
         || (source_mp->num_cols != target_mp->num_cols)
       ) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = source_mp->num_cols; 

    source_row_pos = source_mp->elements[ source_row_num ];
    target_row_pos = target_mp->elements[ target_row_num ];

    for (i=0; i<len; i++)
    {
        *target_row_pos = *source_row_pos; 
        target_row_pos++;
        source_row_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_random_matrix_col
 *                               
 * Gets a random matrix column
 * 
 * This routine gets a random column from the matrix "mp" and puts it into the
 * vector pointed to by vpp. If *vpp is NULL, then the vector is created.
 * If it is the wrong length, then it is resized. Otherwise, the storage is
 * reused. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise the index of the column
 *    selected is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_random_matrix_col(Vector** vpp, const Matrix* mp)
{
    int index = (int)(kjb_rand_2() * mp->num_cols); 

    ERE(get_matrix_col(vpp, mp, index));

    return index;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_matrix_col
 * 
 * Returns a matrix column
 * 
 * This routine gets column "col_num" from the matrix "mp" and puts it into the
 * vector pointed to by vpp. If *vpp is NULL, then the vector is created.
 * If it is the wrong length, then it is resized. Otherwise, the storage is
 * reused.
 * 
 * Note : 
 *     Matrix column counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int get_matrix_col(Vector** vpp, const Matrix* mp, int col_num)
{
    int   len;
    int   i;
    double* vect_pos;

    if ((col_num < 0) || (col_num >= mp->num_cols)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_rows; 

    ERE(get_target_vector(vpp, len)); 

    vect_pos = (*vpp)->elements;

    for (i=0; i<len; i++)
    {
        *vect_pos++ = mp->elements[ i ][ col_num ]; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                remove_matrix_col
 *                               
 * Removes a matrix column
 * 
 * This routine removes column "col_num" from the matrix "mp" and optionally
 * puts it into the vector pointed to by vpp (if vpp is not NULL). If
 * *vpp is NULL, then the vector is created.  If it is the wrong length,
 * then it is resized. Otherwise, the storage is reused.
 * 
 * Note: 
 *     Matrix column counting begins with 0. 
 *
 * Note
 *    Currently, trying to remove the only column of matrix is treated as a
 *    bug.  This may change with future versions. 
 *
 * Returns:
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int remove_matrix_col(Vector** vpp, Matrix* mp, int col_num)
{
    int len, i, j;
    int num_cols = mp->num_cols;

    /*
    // FIX 
    //
    // Currently, lets not deal with single column matrices. Once we implement
    // resziable matrices, then this restriction should be relaxed.
    */
    if ((col_num < 0) || (col_num >= num_cols) || (num_cols <= 1)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = mp->num_cols; 

    if (vpp != NULL)
    {
        ERE(get_target_vector(vpp, len)); 
    }

    num_cols--;

    for (i=0; i<len; i++)
    {
        if (vpp != NULL)
        {
            (*vpp)->elements[ i ] = mp->elements[ i ][ col_num ];
        }

        for (j=col_num; j<num_cols; j++)
        {
            mp->elements[ i ][ j ] = mp->elements[ i ][ j + 1 ];
        }
    }

    mp->num_cols = num_cols;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                put_matrix_col
 *                               
 * Replaces a matrix column with the contents of a vector
 * 
 * This routine replaces column "col_num" of the matrix "mp" with the contents
 * of vector vp. 
 * 
 * Note : 
 *     Matrix column counting begins with 0. 
 *
 * Returns :
 *    If the routine fails (due to storage allocation), then ERROR is returned
 *    with and error message being set. Otherwise NO_ERROR is returned. 
 * 
 * Index: matrices, vectors
 *
 * -----------------------------------------------------------------------------
*/

int put_matrix_col(Matrix* mp, const Vector* vp, int col_num)
{
    int   len;
    int   i;
    double* vect_pos;

    if (    (col_num < 0) || (col_num >= mp->num_cols)
         || (mp->num_rows != vp->length)) 
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    len = vp->length; 

    vect_pos = vp->elements;

    for (i=0; i<len; i++)
    {
        mp->elements[ i ][ col_num ] = *vect_pos++; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int swap_matrix_rows(Matrix* mp, int row1, int row2)
{
    double  temp;
    double* row1_pos;
    double* row2_pos;
    int   i;
    int   num_cols;

    num_cols = mp->num_cols;
    row1_pos = (mp->elements)[row1];
    row2_pos = (mp->elements)[row2];
        
    for (i=0; i<num_cols; i++) 
    {
        temp = *row1_pos;
        *row1_pos = *row2_pos;
        *row2_pos = temp;

        row1_pos++;
        row2_pos++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int swap_matrix_cols(Matrix* mp, int col1, int col2)
{
    int  num_rows;
    int  i;
    double temp;

    num_rows = mp->num_rows;

    for (i=0; i<num_rows; i++) 
    {
        temp = (mp->elements)[i][col1];
        (mp->elements)[i][col1] = (mp->elements)[i][col2];
        (mp->elements)[i][col2] = temp;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int select_matrix_data
(
    Matrix**      selected_mpp,
    const Matrix* data_mp,
    int           num_data_points,
    const int*    data_point_nums,
    const int*    invalid_data 
)
{
    Matrix* selected_mp;
    int     num_cols, i, j;
    int     count = 0;

    if ((data_mp == NULL) || (data_mp->num_rows == 0))
    {
        set_error("No valid input data to select data from.");
        return ERROR;
    }
    else if (num_data_points <= 0) 
    {
        return copy_matrix_data(selected_mpp, data_mp, invalid_data);
    }

    num_cols = data_mp->num_cols; 

    ERE(get_target_matrix(selected_mpp, num_data_points, num_cols));
    selected_mp = *selected_mpp; 

    for (i=0; i<num_data_points; i++)
    {
        int index = data_point_nums[ i ] - 1;

        if (index >= data_mp->num_rows)
        {
            set_bug("Index exceeds matrix dimension in select_matrix_data().");
            return ERROR;
        }

        if ((invalid_data == NULL) || (! invalid_data[ i ]))
        {
            for (j=0; j<num_cols; j++)
            {
                (selected_mp->elements)[ count ][ j ] = 
                                             (data_mp->elements)[ index ][ j ];
            }
            count++;
        }
    }

    selected_mp->num_rows = count; 

    if (count == 0)
    {
        set_error("No valid input data to select data from.\n");
        return ERROR; 
    }
    else 
    {
        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                              vector_is_matrix_row
 *                               
 *
 *
 * -----------------------------------------------------------------------------
*/ 

int vector_is_matrix_row(const Matrix* mp, const Vector* vp)
{
    int i, j;
    int num_rows = mp->num_rows;
    int num_cols = mp->num_cols;
    int this_row; 

    if (vp->length != num_cols)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    for (i=0; i<num_rows; i++)
    {
        this_row = TRUE;

        for (j=0; j<num_cols; j++)
        {
            if ( ! IS_EQUAL_DBL(mp->elements[ i ][ j ], vp->elements[ j ]))
            {
                this_row = FALSE;
                break;
            }
        }

        if (this_row) return TRUE;
    }

    return FALSE;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST

void debug_verify_matrix_vector
(
    const Matrix_vector* mvp,
    const char*          file_name,
    int                  line_number 
)
{
    int i, length;

    if (mvp == NULL) return;

    length = mvp->length;

    for (i = 0; i < length; i++)
    {
        debug_verify_matrix(mvp->elements[ i], file_name, line_number);
    }
}
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEST
void debug_verify_matrix
(
    const Matrix* mp,
    const char*   file_name,
    int           line_number 
)
{
    int i,j, num_rows, num_cols;

    if (mp == NULL) return;

    num_rows = mp->num_rows;
    num_cols = mp->num_cols;

    for (i = 0; i < num_rows; i++)
    {
        for (j = 0; j < num_cols; j++)
        {
            if (    (isnand(mp->elements[ i ][ j ]))
                 || ( ! finite(mp->elements[ i ][ j ]))
               )
            {
                warn_pso("Matrix verifcation failed.\n");
                warn_pso("Element (%d, %d) is not ", i, j);

                if (isnand(mp->elements[ i ][ j ]))
                {
                    warn_pso("a number.\n");
                }
                else 
                {
                    warn_pso("finite.\n");
                }

                warn_pso("Bits are %x, %x\n", 
                         *((int*)&(mp->elements[ i ][ j ])),
                         *(1 + ((int*)&(mp->elements[ i ][ j ]))));
                warn_pso("Setting it to zero\n");
                mp->elements[ i ][ j ] = 0.0;
                warn_pso("Verification called from line %d of %s.\n",
                         line_number, file_name);
            }
        }
    }
}

#endif 

