
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "m/m_matrix.h" 

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                           get_zero_matrix                        
 *                               
 * Gets a zero target matrix
 *
 * This routine is identical to get_target_matrix(), except that the matrix is
 * set to zero (regardless of whether it is created or reused). 
 *
 * Index : matrices
 *
 * -----------------------------------------------------------------------------
*/ 

#ifdef TRACK_MEMORY_ALLOCATION

int debug_get_zero_matrix
(
    Matrix**    target_mpp,
    int         num_rows,
    int         num_cols,
    const char* file_name,
    int         line_number 
)
{

    ERE(debug_get_target_matrix(target_mpp, num_rows, num_cols, 
                                file_name, line_number));
    
    return ow_zero_matrix(*target_mpp);
}

#else

int get_zero_matrix(Matrix** target_mpp, int num_rows, int num_cols)
{

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    
    return ow_zero_matrix(*target_mpp);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                           get_unity_matrix                        
 *                               
 * Gets a target matrix initialized to all ones.
 *
 * This routine is identical to get_target_matrix(), except that the matrix is
 * set to ones (regardless of whether it is created or reused). 
 *
 * Index : matrices
 *
 * -----------------------------------------------------------------------------
*/ 

#ifdef TRACK_MEMORY_ALLOCATION

int debug_get_unity_matrix
(
    Matrix**    target_mpp,
    int         num_rows,
    int         num_cols,
    const char* file_name,
    int         line_number 
)
{

    ERE(debug_get_target_matrix(target_mpp, num_rows, num_cols, 
                                file_name, line_number));
    
    return ow_set_matrix(*target_mpp, 1.0);
}

#else

int get_unity_matrix(Matrix** target_mpp, int num_rows, int num_cols)
{

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    
    return ow_set_matrix(*target_mpp, 1.0);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         get_initialized_matrix                        
 *                               
 * Gets an initialized target matrix
 *
 * This routine is identical to get_target_matrix(), except that the matrix
 * elements are set to a specified initial value (regardless of whether it is
 * created or reused). 
 *
 * Index : matrices
 *
 * -----------------------------------------------------------------------------
*/ 

#ifdef TRACK_MEMORY_ALLOCATION

int debug_get_initialized_matrix
(
    Matrix**    target_mpp,
    int         num_rows,
    int         num_cols,
    double      initial_value,
    const char* file_name,
    int         line_number 
)
{

    ERE(debug_get_target_matrix(target_mpp, num_rows, num_cols, 
                                file_name, line_number));
    
    return ow_set_matrix(*target_mpp, initial_value);
}

#else

int get_initialized_matrix
(
    Matrix** target_mpp,
    int      num_rows,
    int      num_cols,
    double   initial_value 
)
{

    ERE(get_target_matrix(target_mpp, num_rows, num_cols));
    
    return ow_set_matrix(*target_mpp, initial_value);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                get_target_matrix                        
 *                               
 * Gets target matrix
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of matrices. If *target_mpp is NULL, then this routine
 * creates the matrix. If it is not null, and it is the right size, then this
 * routine does nothing. If it is the wrong size, then it is resized. 
 *
 * In order to be memory efficient, we free before resizing. If the resizing
 * fails, then the original contents of the *target_mpp will be lost.  However,
 * target_mpp->elements will be set to NULL, so *target_mpp can be safely sent
 * to free_matrix().  Note that this is in fact the convention throughout the
 * KJB library--if destruction on failure is a problem (usually when *target_mpp
 * is global)--then work on a copy!  
 *
 * See also:
 *    Matrix
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/ 

/*
// FIX 
//
// We need to upgrade this routine so that the storage can be recycled, but this
// is tricky because we have to fix both the array of row pointers.
//
// Actually, this only makes sense if we store the amount of storage as an extra
// variable. Then the storage gets bumped up to a max and stays there. If we
// simply downsize, then on a upsize we have to free anyway. Hence this is not
// as useful. 
*/

#ifdef TRACK_MEMORY_ALLOCATION

int debug_get_target_matrix
(
    Matrix**    target_mpp,
    int         num_rows,
    int         num_cols,
    const char* file_name,
    int         line_number 
)
{
    Matrix* out_mp = *target_mpp; 

    if ((num_rows < 0) || (num_cols < 0))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (out_mp == NULL)
    {
        NRE(out_mp = DEBUG_create_matrix(num_rows, num_cols, file_name, 
                                         line_number));
        *target_mpp = out_mp; 
    }
    else if ((out_mp->num_rows == num_rows) && (out_mp->num_cols == num_cols))
    {
        /*EMPTY*/
        ; /* Do nothing */
    }
    else
    {
        free_2D_double_array(out_mp->elements);

        if ((num_rows > 0) && (num_cols > 0))
        {
            NRE((out_mp->elements = debug_allocate_2D_double_array(num_rows,
                                                                 num_cols,
                                                                 file_name, 
                                                                 line_number)));
        }
        else
        {
            out_mp->elements = NULL; 
        }
        
        out_mp->num_rows = num_rows;
        out_mp->num_cols = num_cols; 
    }

    return NO_ERROR; 
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

int get_target_matrix(Matrix** target_mpp, int num_rows, int num_cols)
{
    Matrix* out_mp = *target_mpp; 

    if ((num_rows < 0) || (num_cols < 0))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (out_mp == NULL)
    {
        NRE(out_mp = create_matrix(num_rows, num_cols));
        *target_mpp = out_mp; 
    }
    else if ((out_mp->num_rows == num_rows) && (out_mp->num_cols == num_cols))
    {
        /*EMPTY*/
        ; /* Do nothing */
    }
    else
    {
        free_2D_double_array(out_mp->elements);

        if ((num_rows > 0) && (num_cols > 0))
        {
            NRE((out_mp->elements = allocate_2D_double_array(num_rows, 
                                                           num_cols)));
        }
        else
        {
            out_mp->elements = NULL; 
        }
        
        out_mp->num_rows = num_rows;
        out_mp->num_cols = num_cols; 
    }

    return NO_ERROR; 
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_matrix 
 *                               
 * Creates a matrix 
 *
 * This routine creates a matrix and returns a pointer P to it. The number 
 * of rows is P->num_rows, the number of columns is P->num_cols, and the
 * elements (of type "double") may be referenced by (P->elements)[ row ][ col ]. 
 * Type "double" is either float or double, depending on whether the single or
 * double precision library is being used. (Usually double precision).
 *
 * The documenation for the Matrix type (see "Matrix") should be consulted for
 * CRITICAL information on the internal structure of the returned matrix.
 * 
 * If TEST is #defined (unix only) then this routine is #define'd to be
 * DEBUG_create_matrix, which is the version available in the developement
 * library. In developement code memory is tracked so that memory leaks can be
 * found more easily. Furthermore, all memory freed is checked that it was
 * allocated by an L library routine. 
 *
 * The routine free_matrix should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Warning:
 *    The user of this library must take care with assumptions about the
 *    structure of the array embedded in the matrix type. See the documentation
 *    for "Matrix" for more details. 
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    One success it returns a pointer to the matrix.
 *
 * See also:
 *    Matrix
 *
 * Acknowlegement:
 *    This routine grew out of one written by Ian Harder. However, it is 
 *    important to note that this routine is somewhat different than any
 *    of Ian's routines. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

Matrix* DEBUG_create_matrix
(
    int         num_rows,
    int         num_cols,
    const char* file_name,
    int         line_number 
)
{
    Matrix* mp;
       
    /* 
    //  Use debug_kjb_malloc as opposed to macros to pass down file_name
    //  and line_number. 
    */

    NRN(mp = DEBUG_TYPE_MALLOC(Matrix, file_name, line_number));

    if ((num_rows > 0) && (num_cols > 0))
    {
        mp->elements = debug_allocate_2D_double_array(num_rows, num_cols, file_name, 
                                                    line_number);
     
        if (mp->elements == NULL) 
        {
            free_matrix(mp);
            return NULL;
        }
    }
    else
    {
        mp->elements = NULL; 
    }
        
    mp->num_rows = num_rows;
    mp->num_cols = num_cols;
    
    return mp;
}

        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

Matrix* create_matrix(int num_rows, int num_cols)
{
    Matrix* mp;
       
    NRN(mp = TYPE_MALLOC(Matrix));
 
    if ((num_rows > 0) && (num_cols > 0))
    {
        mp->elements = allocate_2D_double_array(num_rows, num_cols);
     
        if (mp->elements == NULL) 
        {
            free_matrix(mp);
            return NULL;
        }
    }
    else
    {
        mp->elements = NULL; 
    }
        
    mp->num_rows = num_rows;
    mp->num_cols = num_cols;
    
    return mp;
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_matrix
 *                               
 * Frees a the space associated with a matrix.
 *
 * This routine frees the storage associated with a matrix obtained from
 * create_vector (perhaps indirectrly). If the argument is NULL, then this
 * routine returns safely without doing anything.
 *
 * See also:
 *    Matrix
 *
 * Acknowlegement:
 *    This routine grew out of one written by Ian Harder. However, it is 
 *    important to note that this routine is somewhat different than any
 *    of Ian's routines. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

void free_matrix(Matrix* mp)
{

    if (mp != NULL) 
    {
        if (mp->elements != NULL) 
        {
#ifdef DISABLE
#ifdef TRACK_MEMORY_ALLOCATION
            check_initialization(mp->elements[ 0 ], mp->num_rows * mp->num_cols,
                                 sizeof(double)); 
#endif 
#endif 

            free_2D_double_array(mp->elements); 
        }
      
        kjb_free(mp);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              create_zero_matrix
 *
 * Creates a matrix with all elements zero
 *
 * This routine returns a pointer to a zero matrix of the requested size. 
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

Matrix* create_zero_matrix(int num_rows, int num_cols)
{
    Matrix* output_mp; 

    NRN(output_mp = create_matrix(num_rows, num_cols)); 

    if (ow_zero_matrix(output_mp) == ERROR)
    {
        free_matrix(output_mp); 
        return NULL;
    }

    return output_mp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  create_identity_matrix
 *
 * Creates an identity matrix of the specified size
 *
 * This routine returns a pointer to an identity matrix of the specified size. 
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

Matrix* create_identity_matrix(int size)
{
    Matrix* output_mp;
    int i; 

    NRN(output_mp = create_zero_matrix(size, size));

    for (i=0; i<size; i++) 
    {
        (output_mp->elements)[i][i] = 1.0; 
    }

    return output_mp; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  get_identity_matrix
 *
 * Gets an identity matrix of the specified size
 *
 * This routine puts an identity matrix of the specified size into
 * *output_mpp. If *output_mpp is NULL, then the matrix is created. If it
 * is the wrong sized, then it is resized. Otherwise, the storage is recycled.
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure.  
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

int get_identity_matrix(Matrix** output_mpp, int size)
{
    Matrix* output_mp;
    int i; 

    ERE(get_zero_matrix(output_mpp, size, size));
    output_mp = *output_mpp;

    for (i=0; i<size; i++) 
    {
        (output_mp->elements)[i][i] = 1.0; 
    }

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  ow_zero_matrix
 *
 * Sets all elements of a matrix to zero
 *
 * This routine sets all elements of a matrix to zero. The matrix must have
 * already been created. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure. Failure is only possible if the
 *     matrix is NULL or malformed, and this is treated as a bug. 
 *
 * Index : matrices
 *
 * -----------------------------------------------------------------------------
 */

int ow_zero_matrix(Matrix* mp)
{

    /* 
    // Previously I used memset on the pointer to the first element, but this
    // is a bug because we can temporarily reset the rows and cols of the
    // matrix to work on a subset. Until we implement the concept of a max num
    // rows and cols, we cannot use this kind of trick to speed things up!
    //
    // If we explictly implement subsetting, then we would do a memset on the
    // num_rows * max_num_cols. 
    */

    return ow_set_matrix(mp, 0.0); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                                  ow_set_matrix
 *
 * Sets all elements of a matrix to the specified value
 *
 * This routine sets all elements of a matrix to the specified value. The matrix
 * must have already been created. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure. Failure is only possible if the
 *     matrix is NULL or malformed, and this is treated as a bug. 
 *
 * Index : matrices
 *
 * -----------------------------------------------------------------------------
 */

int ow_set_matrix(Matrix* mp, double set_val)
{
    int i, j;
    double* row_pos; 
    double  real_set_val = set_val; 

    for (i=0; i<mp->num_rows; i++) 
    {
        row_pos = (mp->elements)[ i ]; 

        for (j=0; j<mp->num_cols; j++) 
        {
            *row_pos = real_set_val; 
            row_pos++; 
        }
    }
 
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              get_transpose
 *
 * Gets the transposes a matrix
 *
 * This routine produces a matrix which is the transpose of the argument.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

int get_transpose(Matrix** target_mpp, const Matrix* mp)
{
    Matrix* target_mp;
    int     i;
    int     j;

    ERE(get_target_matrix(target_mpp, mp->num_cols, mp->num_rows)); 
    target_mp = *target_mpp; 

    for (i = 0; i < mp->num_rows; i++) 
    {
        for (j = 0; j < mp->num_cols; j++) 
        {
            (target_mp->elements)[j][i] = (mp->elements)[i][j];
         }
     }
     
     return NO_ERROR; 
 }
       
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              create_transpose
 *
 * Creates a matrix which is the transpose of the argument
 *
 * This routine creates a matrix which is the transpose of the argument.
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

Matrix* create_transpose(const Matrix* mp)
{
    Matrix* t_mp;
    int i, j; 

    NRN(t_mp = create_matrix(mp->num_cols, mp->num_rows)); 

    for (i = 0; i < mp->num_rows; i++) 
    {
        for (j = 0; j < mp->num_cols; j++) 
        {
            (t_mp->elements)[j][i] = (mp->elements)[i][j];
         }
     }
     
     return t_mp; 
 }
       
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              get_diagonal_matrix
 *
 * Gets a diagonal matrix from a vector
 *
 * This routine creates a diagonal matrix from a vector provided as the
 * argument. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

int get_diagonal_matrix(Matrix** mpp, Vector* vp)
{
    Matrix* d_mp;
    int     i;

    ERE(get_zero_matrix(mpp, vp->length, vp->length)); 
    d_mp = *mpp; 

    for (i = 0; i < vp->length; i++) 
    {
        (d_mp->elements)[i][i] = (vp->elements)[i];
     }
     
     return NO_ERROR; 
 }
       
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              create_diagonal_matrix
 *
 * Creates a diagonal matrix from a vector
 *
 * This routine creates a diagonal matrix from a vector provided as the
 * argument.
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
 */

Matrix* create_diagonal_matrix(const Vector* vp)
{
    Matrix* d_mp;
    int i; 

    NRN(d_mp = create_zero_matrix(vp->length, vp->length)); 

    for (i = 0; i < vp->length; i++) 
    {
        (d_mp->elements)[i][i] = (vp->elements)[i];
     }
     
     return d_mp; 
 }
       
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             get_target_matrix_vector
 *                               
 * Gets a target matrix vector
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of matrix vectors. If *target_mvpp is
 * NULL, then this routine creates the matrix vector. If it is not null, and it
 * is the right size, then this routine does nothing. If it is the wrong size,
 * then it is resized. 
 *
 * The routine free_matrix_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

int get_target_matrix_vector(Matrix_vector** mvpp, int length)
{

    if (    (*mvpp != NULL)
         && ((*mvpp)->length == length) 
       )
    {
        return NO_ERROR; 
    }

    free_matrix_vector(*mvpp);
    NRE(*mvpp = create_matrix_vector(length));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_matrix_vector
 *                               
 * Creates a matrix vector
 *
 * This routine creates a matrix array of size "length" which is a type for a
 * list of matrixs.  All matrix pointers are set to NULL. 
 *
 * This routine sets all the pointers to NULL. 
 *
 * The routine free_matrix_vector should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

Matrix_vector* create_matrix_vector(int length)
{
    Matrix_vector* mvp;
    int           i;

    NRN(mvp = TYPE_MALLOC(Matrix_vector));
    mvp->length = length; 
    NRN(mvp->elements = N_TYPE_MALLOC(Matrix*, length));

    for (i=0; i<length; i++)
    {
        mvp->elements[ i ] = NULL;
    }

    return mvp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_matrix_vector
 *                               
 * Frees the storage in a matrix vector
 *
 * This routine frees the storage in a matrix array. 
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

void free_matrix_vector(Matrix_vector* mvp)
{
    int count, i;
    Matrix** mp_array_pos;

    if (mvp == NULL) return;

    mp_array_pos = mvp->elements;
    count = mvp->length; 

    for (i=0; i<count; i++) 
    {
        free_matrix(*mp_array_pos);
        mp_array_pos++;
    }

    kjb_free(mvp->elements);
    kjb_free(mvp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         matrix_vectors_are_comparable 
 *                               
 * Returns whether two matrix vectors have the same dimensions
 *
 * This routine returns whether two matrix vectors have the same dimensions. It
 * returns TRUE if both arguments are NULL, or if they are matrix vectors of the
 * same length with each paor of corresponding component matrices are of the
 * same dimension or simultaneously NULL. 
 *
 * Returns: 
 *    TRUE if the matrix vectors are of the same length and all the sub matrices
 *    are the same dimensions (or are similtaneously NULL). 
 *
 * Index:
 *    error handling, debugging 
 *
 * -----------------------------------------------------------------------------
*/ 

int matrix_vectors_are_comparable
(
    const Matrix_vector* in_mvp,
    const Matrix_vector* out_mvp 
)
{
    int i, length; 

    if ((in_mvp == NULL) && (out_mvp == NULL))
    {
        return TRUE; 
    }

    if (    ((in_mvp == NULL) && (out_mvp != NULL))
         || ((out_mvp == NULL) && (in_mvp != NULL))
         || (in_mvp->length != out_mvp->length)
       )
    {
        return FALSE;
    }

    length = in_mvp->length; 

    for (i = 0; i < length; i++)
    {
        if (    (in_mvp->elements[ i ]->num_rows != out_mvp->elements[ i ]->num_rows)
             || (in_mvp->elements[ i ]->num_cols != out_mvp->elements[ i ]->num_cols)
           )
        {
            return FALSE;
        }
    }

    return TRUE;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             get_target_matrix_vector_vector
 *                               
 * Gets a target matrix vector vector
 *
 * This routine implements the creation/over-writing semantics used in the KJB
 * library in the case of matrix vector vectors. If *target_mvvpp is
 * NULL, then this routine creates the object. If it is not null, and it
 * is the right size, then this routine does nothing. If it is the wrong size,
 * then it is resized. 
 *
 * The routine free_matrix_vector_vector should be used to dispose of the
 * storage once it is no longer needed.
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

int get_target_matrix_vector_vector(Matrix_vector_vector** mvvpp, int length)
{

    if (    (*mvvpp != NULL)
         && ((*mvvpp)->length == length) 
       )
    {
        return NO_ERROR; 
    }

    free_matrix_vector_vector(*mvvpp);
    NRE(*mvvpp = create_matrix_vector_vector(length));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             create_matrix_vector_vector
 *                               
 * Creates a matrix vector vector
 *
 * This routine creates a vector of vectors of matrices array of size "length.
 * All matrix vector pointers are set to NULL. 
 *
 * The routine free_matrix_vector_vector should be used to dispose of the
 * storage once it is no longer needed.
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

Matrix_vector_vector* create_matrix_vector_vector(int length)
{
    Matrix_vector_vector* mvvp;
    int           i;

    NRN(mvvp = TYPE_MALLOC(Matrix_vector_vector));
    mvvp->length = length; 
    NRN(mvvp->elements = N_TYPE_MALLOC(Matrix_vector*, length));

    for (i=0; i<length; i++)
    {
        mvvp->elements[ i ] = NULL;
    }

    return mvvp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_matrix_vecto_vector
 *                               
 * Frees the storage in a matrix vector vector
 *
 * This routine frees the storage in a matrix vector vector object. 
 *
 * Index: memory allocation, arrays, matrixs
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

void free_matrix_vector_vector(Matrix_vector_vector* mvvp)
{
    int count, i;
    Matrix_vector** mvpp;

    if (mvvp == NULL) return;

    mvpp = mvvp->elements;
    count = mvvp->length; 

    for (i=0; i<count; i++) 
    {
        free_matrix_vector(*mvpp);
        mvpp++;
    }

    kjb_free(mvvp->elements);
    kjb_free(mvvp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Matrix_vector** create_matrix_vector_list(int count)
{
    Matrix_vector** matrix_vector_list;
    int            i;

    NRN(matrix_vector_list = N_TYPE_MALLOC(Matrix_vector*, count));

    for (i=0; i<count; i++)
    {
        matrix_vector_list[ i ] = NULL;
    }

    return matrix_vector_list;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_matrix_vector_list
(
    int             count,
    Matrix_vector** matrix_vector_list 
)
{
    int i;

    if (matrix_vector_list == NULL) return;

    for (i=0; i<count; i++) 
    {
        free_matrix_vector(matrix_vector_list[ i ]);
    }

    kjb_free(matrix_vector_list);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                            concat_matrices_vertically 
 *
 * Concatenates matrices verically                             
 *
 * This routine takes an array of matrices and forms one matrix consisting of
 * those matrices stacked on top of each other. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * The matrix array may contain any number of NULL matrices. If there are only
 * NULL matrices, or if num_matrices is zero, then the target matrix is freed
 * and set to NULL. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int concat_matrices_vertically
(
    Matrix**       mpp,
    int            num_matrices,
    const Matrix** matrix_list 
)
{
    int i, j, k, count;
    int num_rows = 0;
    int num_cols = NOT_SET; 
    Matrix* mp; 

    if (num_matrices <= 0)
    {
        free_matrix(*mpp);
        *mpp = NULL;
        return NO_ERROR;
    }

    for (i=0; i<num_matrices; i++)
    {
        if (matrix_list[ i ] != NULL)
        {
            if (num_cols == NOT_SET)
            {
                num_cols = (matrix_list[ i ])->num_cols;
            }
            else if ((matrix_list[ i ])->num_cols != num_cols)
            {
                SET_ARGUMENT_BUG();
                return ERROR;
            }
            num_rows += ((matrix_list[ i ])->num_rows);
        }
    }

    if (num_rows == 0)
    {
        free_matrix(*mpp);
        *mpp = NULL;
        return NO_ERROR;
    }

    ERE(get_target_matrix(mpp, num_rows, num_cols));
    mp = *mpp; 

    count = 0;

    while (count < num_rows)
    {
        for (k=0; k<num_matrices; k++)
        {
            if (matrix_list[ k ] != NULL)
            {
                for (i=0; i<(matrix_list[ k ])->num_rows; i++)
                {
                    for (j=0; j<num_cols; j++)
                    {
                        mp->elements[ count ][ j ] = 
                                       (matrix_list[ k ])->elements[ i ][ j ];
                    }
                    count++;
                }
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                         get_matrix_from_matrix_vector 
 *
 * Concatenates the matrices in a matrix vector                             
 *
 * This routine takes the matrices in a matrix vectorand forms one matrix
 * consisting of those matrices stacked on top of each other. 
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * The matrix array may contain any number of NULL matrices. If there are only
 * NULL matrices, or if num_matrices is zero, then the target matrix is freed
 * and set to NULL. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/ 

int get_matrix_from_matrix_vector
(
    Matrix**       mpp,
    const Matrix_vector* mvp
)
{
    int i;
    int num_rows = 0;
    int num_cols = NOT_SET; 
    Matrix* mp; 

    if (mvp == NULL)
    {
        free_matrix(*mpp); 
        *mpp = NULL;
        return NO_ERROR;
    }

    for (i=0; i<mvp->length; i++)
    {
        if (mvp->elements[ i ] != NULL)
        {
            if (num_cols == NOT_SET)
            {
                num_cols = (mvp->elements[ i ])->num_cols;
            }
            else if ((mvp->elements[ i ])->num_cols != num_cols)
            {
                SET_ARGUMENT_BUG();
                return ERROR;
            }
            num_rows += ((mvp->elements[ i ])->num_rows);
        }
    }

    if (num_rows == 0)
    {
        free_matrix(*mpp);
        *mpp = NULL;
        return NO_ERROR;
    }

    ERE(get_target_matrix(mpp, num_rows, num_cols)); 
    mp = *mpp; 

    num_rows = 0; 

    for (i=0; i<mvp->length; i++)
    {
        if (mvp->elements[ i ] != NULL)
        {
            ERE(ow_copy_matrix(mp, num_rows, 0, mvp->elements[ i ]));
            num_rows += ((mvp->elements[ i ])->num_rows);
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                create_random_matrix
 *                               
 * Creates a random matrix
 *
 * This routine creates a matrix of the specified dimensions, and fills it with
 * random values between 0.0 and 1.0. The routine kjb_rand is used for the
 * random numbers.
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

Matrix* create_random_matrix(int num_rows, int num_cols)
{
    Matrix* mp;
    double*   elem_ptr; 
    int    i, j; 

    NRN(mp = create_matrix(num_rows, num_cols));

    for (i=0; i<num_rows; i++) 
    {
        elem_ptr = (mp->elements)[ i ];
    
        for (j=0; j<num_cols; j++)
        {
            *elem_ptr++ = kjb_rand();
        }
    }

    return mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                create_random_matrix_2
 *                               
 * Creates a random matrix
 *
 * This routine creates a matrix of the specified dimensions, and fills it with
 * random values between 0.0 and 1.0. The routine kjb_rand_2() is used for the
 * random numbers. 
 *
 * Returns:
 *     A  pointer to the matrix on success and NULL on failure (with an error
 *     message being set.) This routine will only fail if storage allocation
 *     fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

Matrix* create_random_matrix_2(int num_rows, int num_cols)
{
    Matrix* mp;
    double*   elem_ptr; 
    int    i, j; 

    NRN(mp = create_matrix(num_rows, num_cols));

    for (i=0; i<num_rows; i++) 
    {
        elem_ptr = (mp->elements)[ i ];
    
        for (j=0; j<num_cols; j++)
        {
            *elem_ptr++ = kjb_rand_2();
        }
    }

    return mp;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                              get_random_matrix
 *                               
 * Gets a random matrix 
 *
 * This routine gets a matrix of the specified dimensions, and fills it with
 * random values between 0.0 and 1.0. The routine kjb_rand() is used.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int get_random_matrix(Matrix** target_mpp, int num_rows, int num_cols)
{
    Matrix* mp;
    double*   elem_ptr; 
    int    i, j; 

    ERE(get_target_matrix(target_mpp,num_rows, num_cols)); 
    mp = *target_mpp; 

    for (i=0; i<num_rows; i++) 
    {
        elem_ptr = (mp->elements)[ i ];
    
        for (j=0; j<num_cols; j++)
        {
            *elem_ptr++ = kjb_rand();
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                              get_random_matrix_2
 *                               
 * Gets a random matrix 
 *
 * This routine gets a matrix of the specified dimensions, and fills it with
 * random values between 0.0 and 1.0. The routine kjb_rand_2 is used.
 *
 * The first argument is the adress of the target matrix. If the target matrix
 * itself is null, then a matrix of the appropriate size is created. If the
 * target matrix is the wrong size, it is resized. Finally, if it is the right
 * size, then the storage is recycled, as is. 
 *
 * Returns:
 *     NO_ERROR on success and ERROR on failure This routine will only fail if
 *     storage allocation fails. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

int get_random_matrix_2(Matrix** target_mpp, int num_rows, int num_cols)
{
    Matrix* mp;
    double*   elem_ptr; 
    int    i, j; 

    ERE(get_target_matrix(target_mpp,num_rows, num_cols)); 
    mp = *target_mpp; 

    for (i=0; i<num_rows; i++) 
    {
        elem_ptr = (mp->elements)[ i ];
    
        for (j=0; j<num_cols; j++)
        {
            *elem_ptr++ = kjb_rand_2();
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_perturb_matrix(Matrix* mp, double fraction)
{
    int i, j, num_rows, num_cols;

    num_rows = mp->num_rows;
    num_cols = mp->num_cols;

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            mp->elements[ i ][ j ] *= (1.0 + fraction * (kjb_rand_2() - 0.5)); 
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             allocate_2D_mp_array
 *                               
 * Allocates a 2D array of matrix pointers
 *
 * This routine returns a pointer P to a two-dimensional array of matrix
 * pointers. P can be de-referenced by P[row][col] to obtain the storage of a
 * Matrix*. P[0] points to a contiguous area of memory which contains the
 * entire array. P[1] is a short-cut pointer to the first element of the second
 * row, and so on. Thus the array can be accessed sequentually starting at
 * P[0], or explicitly by row and column. (Note that this is not the common way
 * of setting up a two-dimensional array--see below). 
 *
 * This routine sets all the pointers to NULL. 
 *
 * If TEST is #defined (unix only) then this routine is #define'd to be
 * debug_allocate_2D_mp_array, which is the version available in the
 * development library. In development code, memory is tracked so that memory
 * leaks can be found more easily. Furthermore, all memory free'd is checked
 * that it was allocated by a KJB library routine. Finally, memory is checked
 * for overuns. 
 *
 * The routine free_2D_mp_array should be used to dispose of the storage once
 * it is no longer needed.
 *
 * Note:
 *    Naming convention--this routine is called "allocate_2D_mp_array" rather
 *    than "create_2D_mp_array" or "create_matrix_matrix" because we are not
 *    creating an abstract data type, but rather just allocating raw storage.
 *    For many puposes this routine is a hack waiting for a
 *    create_matrix_matrix routine to be written, but it has uses outside of
 *    this as well. 
 *
 * Warning:
 *    The structure of the returned array is somewhat different than a more
 *    common form of a two-dimensional array in "C" where each row is allocated
 *    separately. Here the storage area is contiguous. This allows for certain
 *    operations to be done quickly, but note the following IMPORTANT point:
 *    num_rows cannot be swapped by simply swapping row pointers! 
 *
 * Index: memory allocation, arrays, matrices
 *
 * Returns: 
 *    On error, this routine returns NULL, with an error message being set.  
 *    On success it returns a pointer to the array.
 *
 * -----------------------------------------------------------------------------
*/

#ifdef TRACK_MEMORY_ALLOCATION

Matrix*** debug_allocate_2D_mp_array(int num_rows, int num_cols, 
                                     const char* file_name, int line_number)
{
    Matrix***   array;
    Matrix***   array_pos;
    Matrix**    col_ptr;
    int         i, j;
    Malloc_size num_bytes;

    if ((num_rows <= 0) || (num_cols <= 0))
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }

    /* 
    //  Use debug_kjb_malloc as opposed to macros to pass down file_name
    //  and line_number. 
    */
    num_bytes = num_rows * sizeof(Matrix**);
    NRN(array = (Matrix ***)debug_kjb_malloc(num_bytes, file_name, line_number));

    num_bytes = num_rows * num_cols * sizeof(Matrix*);
    col_ptr = (Matrix**)debug_kjb_malloc(num_bytes, file_name, line_number);

    if (col_ptr == NULL)
    {
        kjb_free(array);
        return NULL;
    }

    array_pos = array; 

    for(i=0; i<num_rows; i++) 
    {
        *array_pos = col_ptr;
        array_pos++;
        col_ptr += num_cols; 
    }

    for(i=0; i<num_rows; i++) 
    {
        for (j=0; j<num_cols; j++)
        {
            array[ i ][ j ] = NULL;
        }
    }
        
    return array; 
}
        /*  ==>                               /\                              */
        /*  ==>  Development code above       ||                              */
#else   /* ------------------------------------------------------------------ */
        /*  ==>  Production code below        ||                              */
        /*  ==>                               \/                              */

Matrix*** allocate_2D_mp_array(int num_rows, int num_cols)
{
    Matrix*** array;
    Matrix*** array_pos;
    Matrix**  col_ptr;
    int       i, j;

    if ((num_rows <= 0) || (num_cols <= 0))
    {
        SET_ARGUMENT_BUG();
        return NULL;
    }

    NRN(array = N_TYPE_MALLOC(Matrix**, num_rows));

    col_ptr = N_TYPE_MALLOC(Matrix*, num_rows * num_cols);

    if (col_ptr == NULL)
    {
        kjb_free(array);
        return NULL;
    }

    array_pos = array; 

    for(i=0; i<num_rows; i++) 
    {
        *array_pos = col_ptr;
        array_pos++;
        col_ptr += num_cols; 
    }

    for(i=0; i<num_rows; i++) 
    {
        for (j=0; j<num_cols; j++)
        {
            array[ i ][ j ] = NULL;
        }
    }
        
    return array; 
}

#endif    /*   #ifdef TRACK_MEMORY_ALLOCATION ... #else ...   */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_2D_mp_array
 *                               
 * Frees an array from allocate_2D_mp_array
 *
 * This routine frees the storage obtained from allocate_2D_mp_array. If the
 * argument is NULL, then this routine returns safely without doing anything.
 *
 * -----------------------------------------------------------------------------
*/

void free_2D_mp_array(Matrix*** array)
{
    
    if (array == NULL) return;

    kjb_free(*array);
    kjb_free(array);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                             free_2D_mp_array_and_matrices
 *                               
 * Frees a 2D array of matrices
 *
 * This routine frees the storage obtained from allocate_2D_mp_array and the
 * matrices pointed to by the pointers. If the argument is NULL, then this
 * routine returns safely without doing anything.
 *
 * -----------------------------------------------------------------------------
*/

void free_2D_mp_array_and_matrices
(
    Matrix*** array,
    int       num_rows,
    int       num_cols 
)
{
    int i, j; 
    
    if (array == NULL) return;

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            free_matrix(array[ i ][ j ]);
        }
    }

    kjb_free(*array);
    kjb_free(array);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

