
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#ifndef M_MATRIX_INCLUDED
#define M_MATRIX_INCLUDED

#include "l/l_gen.h"
#include "m/m_vector.h"

/* =============================================================================
 *                                   Matrix 
 * 
 * Matrix type
 *
 * This type is the matrix type for the KJB library and is used by many
 * routines. It stores a matrix with num_rows rows and num_cols cols. If "mp"
 * is a pointer to the type Matrix, then mp->elements[ row ][ col ] accesses
 * the element (row, col), and mp->elements[ row ] accesses the row'th row.
 * Note that counting starts at 0. The elements of the matrix are doubles.
 *
 * The matrix rows may or may not be stored consecutively in memory. Because of
 * this, and because some routines may (most don't) take advantage of knowlege
 * of the internal structure for performance, it is important NOT to swap rows
 * by swapping pointers -- the elements should be copied. 
 *
 * Warning:
 *    As described above, the user of this library must take care with
 *    assumptions about the structure of the array embedded in the matrix type.
 *
 * See also:
 *    KJB_INIT(), get_target_matrix, create_matrix, free_matrix, allocate_2D_double_array
 *
 * Acknowlegement:
 *    This type grew out of one written by Ian Harder. However, it is important
 *    to note that this type is somewhat different than Ian's Matrix type. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/

typedef struct Matrix 
{   
    int    num_rows;
    int    num_cols;
    double **elements;
}   
Matrix;

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                                   Matrix_vector 
 * 
 * Type for an array of matrices
 *
 * This type is used in the KJB library for arrays of matrices. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/
 
typedef struct Matrix_vector
{
    int      length;
    Matrix** elements;
}
Matrix_vector;

/* -------------------------------------------------------------------------- */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* =============================================================================
 *                               Matrix_vector_vector 
 * 
 * Type for an array of arrays of matrices
 *
 * This type is used in the KJB library for arrays of arrays of matrices. 
 *
 * Index: matrices
 *
 * -----------------------------------------------------------------------------
*/
 
typedef struct Matrix_vector_vector
{
    int             length;
    Matrix_vector** elements;
}
Matrix_vector_vector;

/* -------------------------------------------------------------------------- */

#ifdef TRACK_MEMORY_ALLOCATION

#   define get_zero_matrix(x, y, z)   \
                            debug_get_zero_matrix(x, y, z, __FILE__, __LINE__)

#   define get_unity_matrix(x, y, z)   \
                            debug_get_unity_matrix(x, y, z, __FILE__, __LINE__)

#   define get_initialized_matrix(w, x, y, z)   \
                  debug_get_initialized_matrix(w, x, y, z, __FILE__, __LINE__)

#   define get_target_matrix(x, y, z)   \
                            debug_get_target_matrix(x, y, z, __FILE__, __LINE__)

#   define create_matrix(x,y)  DEBUG_create_matrix(x,y, __FILE__, __LINE__)
 
#   define allocate_2D_mp_array(x, y) \
                 debug_allocate_2D_mp_array((x), (y), __FILE__, __LINE__)

    int debug_get_zero_matrix
    (
        Matrix**    target_mpp,
        int         num_rows,
        int         num_cols,
        const char* file_name,
        int         line_number 
    );

    int debug_get_unity_matrix
    (
        Matrix**    target_mpp,
        int         num_rows,
        int         num_cols,
        const char* file_name,
        int         line_number 
    );

    int debug_get_initialized_matrix
    (
        Matrix**    target_mpp,
        int         num_rows,
        int         num_cols,
        double      initial_value,
        const char* file_name,
        int         line_number 
    );

    int debug_get_target_matrix
    (
        Matrix**    target_mpp,
        int         num_rows,
        int         num_cols,
        const char* file_name,
        int         line_number 
    );

    Matrix* DEBUG_create_matrix
    (
        int         rows,
        int         cols,
        const char* file_name,
        int         line_number 
    );

    Matrix*** debug_allocate_2D_mp_array(int, int, const char*, int);
#else 
    int get_zero_matrix(Matrix** target_mpp, int num_rows, int num_cols);

    int get_unity_matrix(Matrix** target_mpp, int num_rows, int num_cols);

    int get_initialized_matrix
    (
        Matrix** target_mpp,
        int      num_rows,
        int      num_cols,
        double   initial_value 
    );

    Matrix* create_matrix(int rows, int cols);

    int get_target_matrix
    (
        Matrix** out_arg_mp,
        int      num_rows,
        int      num_cols 
    );
    Matrix*** allocate_2D_mp_array(int, int);
#endif

void free_matrix(Matrix* mp);
Matrix* create_identity_matrix(int);

Matrix* create_zero_matrix(int num_rows, int num_cols);

int get_identity_matrix(Matrix** output_mpp, int size);

int ow_zero_matrix(Matrix*);

int ow_set_matrix(Matrix*, double);

int get_transpose(Matrix** target_mpp, const Matrix* mp);

Matrix* create_transpose(const Matrix*);

int get_diagonal_matrix(Matrix** mp, Vector* vp);

Matrix* create_diagonal_matrix(const Vector*);

int get_target_matrix_vector(Matrix_vector** mvpp, int count);

int matrix_vectors_are_comparable
(
    const Matrix_vector* in_mvp,
    const Matrix_vector* out_mvp 
);

Matrix_vector* create_matrix_vector(int count);

void free_matrix_vector(Matrix_vector*);

int                   get_target_matrix_vector_vector
(
    Matrix_vector_vector** mvvpp,
    int                    length 
);

Matrix_vector_vector* create_matrix_vector_vector(int length);

void                  free_matrix_vector_vector      
(
    Matrix_vector_vector* mvvp 
);
Matrix_vector** create_matrix_vector_list(int count);

void free_matrix_vector_list
(
    int             count,
    Matrix_vector** matrix_vector_list 
);

int concat_matrices_vertically
(
    Matrix**       mpp,
    int            num_matrices,
    const Matrix** matrix_vector 
);

int get_matrix_from_matrix_vector(Matrix** mpp, const Matrix_vector* mvp);

int get_random_matrix(Matrix** target_mpp, int num_rows, int num_cols);

int get_random_matrix_2(Matrix** target_mpp, int num_rows, int num_cols);

Matrix* create_random_matrix(int, int);

Matrix* create_random_matrix_2(int num_rows, int num_cols);

int ow_perturb_matrix(Matrix* mp, double fraction);

void free_2D_mp_array(Matrix***);

void free_2D_mp_array_and_matrices
(
    Matrix*** array,
    int       num_rows,
    int       num_cols 
);

#endif

