
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_def.h"

#include "m/m_random.h"

/* -------------------------------------------------------------------------- */

#define GAUSSIAN_NOISE_LUT_SIZE   5000
#define ROOT_ONE_HALF_PIE         1.253314137 
#define GAUSSIAN_NOISE_LUT_JUMP   (ROOT_ONE_HALF_PIE / GAUSSIAN_NOISE_LUT_SIZE)
#define GAUSSIAN_NOISE_LUT_RESOLUTION 100.0 

/* -------------------------------------------------------------------------- */

static double* fs_gaussian_distribution_lut = NULL; 

/* Will be at least one less than the table size. */
static int   fs_gaussian_distribution_lut_size;  

/* -------------------------------------------------------------------------- */

static int initialize_gaussian_distribution_lut(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void prepare_memory_cleanup(void);
    static void free_gaussian_distribution_lut(void);
#endif 

/* -------------------------------------------------------------------------- */

/* =============================================================================
 *                          gauss_rand                                        
 *                               
 * Returns a Gaussian distributed, random number 
 *
 * This routine returns an approximately  Gaussian distributed, random number,
 * with mean zero and variance one. 
 *
 * Index: random
 *
 * See also:
 *     kjb_rand
 *
 * -----------------------------------------------------------------------------
*/ 

double gauss_rand(void)
{
    double uniform_random_number;
    double gaussian_random_number;
    int    negative_noise         = FALSE;
    double a, x, y;
    int    index;

    EPETE(initialize_gaussian_distribution_lut()); 

    /* 
    // We use kjb_rand, as opposed to kjb_rand_2, to make this routine part of
    // the first random stream. 
    */
    uniform_random_number = kjb_rand();

    if (uniform_random_number >= 0.5)
    {
        negative_noise = TRUE; 
        uniform_random_number -= 0.5;
    }

    index = (int)(uniform_random_number * 2.0 * fs_gaussian_distribution_lut_size);

    if ((index < 0) || (index >= fs_gaussian_distribution_lut_size))
    {
        SET_CANT_HAPPEN_BUG();
        kjb_exit(EXIT_FAILURE); 
    }

    ASSERT(fs_gaussian_distribution_lut_size < GAUSSIAN_NOISE_LUT_SIZE); 

    x = fs_gaussian_distribution_lut[ index ];
    y = fs_gaussian_distribution_lut[ index + 1];
    /* 
    // We use kjb_rand, as opposed to kjb_rand_2, to make this routine part of
    // the first random stream. 
    */
    a = kjb_rand();

    gaussian_random_number = a * x + (1.0 - a) * y;

    if (negative_noise) 
    {
        gaussian_random_number *= -1.0;   
    }

    return gaussian_random_number; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_gaussian_distribution_lut(void)
{
    long_double    sum      = 0.0;
    int            count    = 0;
    double         x        = 0.0;
    double         step     = GAUSSIAN_NOISE_LUT_JUMP / 
                                                  GAUSSIAN_NOISE_LUT_RESOLUTION;
    double         dist_val = 1.0;
    int            num_its  = 0;

    if (fs_gaussian_distribution_lut != NULL) return NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    NRE(fs_gaussian_distribution_lut = DBL_MALLOC(GAUSSIAN_NOISE_LUT_SIZE)); 

    fs_gaussian_distribution_lut[ 0 ] = 0.0; 

    /*CONSTCOND*/
    while (TRUE) 
    {
        if ((num_its > 10000) || (count >= GAUSSIAN_NOISE_LUT_SIZE - 1))
        {
            break; 
        }

        if (sum > GAUSSIAN_NOISE_LUT_JUMP)
        {
            sum -= GAUSSIAN_NOISE_LUT_JUMP; 
            count++; 

            fs_gaussian_distribution_lut[ count ] = x; 
            step = GAUSSIAN_NOISE_LUT_JUMP / dist_val / 
                                                  GAUSSIAN_NOISE_LUT_RESOLUTION;
            num_its = 0; 
        }

        x += step; 
        dist_val = exp(-x*x/2.0); 
        sum += dist_val * step; 

        num_its ++; 
    }

    fs_gaussian_distribution_lut_size = count; 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_gaussian_distribution_lut);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_gaussian_distribution_lut(void)
{
    kjb_free(fs_gaussian_distribution_lut);
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

