
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "m/m_vec_stat.h" 
#include "m/m_vec_metric.h" 
#include "m/m_stat.h" 

/* -------------------------------------------------------------------------- */

/*
// CHECK
//
// Perhaps these should be long_double, not double. (But currently, long_double
// is normally double because quad precision is not normally supported in
// hardware on your basic machine. 
*/
static double fs_data_sum         = 0.0;
static double fs_data_min;
static double fs_data_max;
static double fs_data_squared_sum = 0.0;
static int    fs_num_data_points  = NOT_SET;
static int    fs_dirty_stat_data  = FALSE;

/* -------------------------------------------------------------------------- */

static int get_2D_gaussian_mask
(
    Matrix** mask_mpp,  /* Output gaussian smoothing mask.  */
    int      mask_size, /* Number of elements per axis (odd) */
    double   sigma      /* Standard deviation in bin units. */              
);

/* -------------------------------------------------------------------------- */

int add_data_point(double data_value)
{

    fs_num_data_points++;
    fs_data_sum += data_value;
    fs_data_squared_sum += (data_value * data_value); 

    if (fs_num_data_points == 1) 
    {
        fs_data_min = fs_data_max = data_value; 
    }
    else 
    {
        if      (data_value < fs_data_min) fs_data_min = data_value; 
        else if (data_value > fs_data_max) fs_data_max = data_value; 
    }

    fs_dirty_stat_data = TRUE; 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Stat* get_data_stats(void)
{
    Stat*  stat_ptr;
    double temp;

    if (fs_num_data_points == 0) 
    {
        set_bug("Call to get_data_stats before add_data_point.");
        return NULL;
    }
   
    NRN(stat_ptr = TYPE_MALLOC(Stat)); 

    stat_ptr->n = fs_num_data_points; 

    stat_ptr->mean = (fs_data_sum / (double)fs_num_data_points); 

    if (fs_num_data_points == 1) 
    {
        stat_ptr->stdev = DBL_NOT_SET; 
    }
    else 
    {
        temp = fs_data_squared_sum;
        temp -= fs_data_sum * fs_data_sum / fs_num_data_points; 
        temp /= (fs_num_data_points - 1);
        stat_ptr->stdev = sqrt(temp);
    }

    stat_ptr->min = fs_data_min;
    stat_ptr->max = fs_data_max; 

    fs_dirty_stat_data = FALSE; 

    return stat_ptr;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int clear_data_stats(void)
{
    if (fs_dirty_stat_data)
    {
        set_bug("Attempt to clear data stats with dirty stat data.");
        return ERROR;
    }

    fs_num_data_points  = 0;
    fs_data_sum         = 0.0;
    fs_data_squared_sum = 0.0;
   
    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void cleanup_data_stats(void)
{
    fs_dirty_stat_data = FALSE; 
   
    EPE(clear_data_stats()); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_correlation_coefficient
(
    Vector* first_vp,
    Vector* second_vp,
    double* result_ptr 
)
{
    double    first_mean           = get_vector_mean(first_vp);
    double    second_mean          = get_vector_mean(second_vp);
    Vector* first_minus_mean_vp  = NULL;
    Vector* second_minus_mean_vp = NULL;
    int     result               = NO_ERROR;

    if (result != ERROR) 
    {
        result = copy_vector(&first_minus_mean_vp, first_vp);
    }

    if (result != ERROR) 
    {
        result = copy_vector(&second_minus_mean_vp, second_vp);
    }

    if (result != ERROR) 
    {
        result = ow_subtract_scalar_from_vector(first_minus_mean_vp, 
                                                first_mean);
    }

    if (result != ERROR) 
    {
        result = ow_subtract_scalar_from_vector(second_minus_mean_vp, 
                                                second_mean);
    }

    if (result != ERROR) 
    {
        result = ow_scale_vector_by_magnitude(first_minus_mean_vp);
    }

    if (result != ERROR) 
    {
        result = ow_scale_vector_by_magnitude(second_minus_mean_vp);
    }

    if (result != ERROR) 
    {
        result = get_dot_product(first_minus_mean_vp, second_minus_mean_vp,
                                 result_ptr);
    }

    free_vector(first_minus_mean_vp);
    free_vector(second_minus_mean_vp);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int gauss_convolve_matrix
(
    Matrix**      out_mpp,
    const Matrix* in_mp,
    double        sigma 
)
{
    Matrix* out_mp;
    int          num_rows, num_cols, i, j, mi, mj, m, n;
    int          mask_width;
    int          mask_size;
    Matrix*      mask_mp = NULL;

    num_rows = in_mp->num_rows;
    num_cols = in_mp->num_cols;

    ERE(get_target_matrix(out_mpp, num_rows, num_cols)); 
    out_mp = *out_mpp; 

    mask_width = (int)(1.0 + 2.0 * sigma); 
    mask_size  = 1 + 2 * mask_width; 

    ERE(get_2D_gaussian_mask(&mask_mp, mask_size, sigma)); 

    for (i=0; i<num_rows; i++)
    {
        for (j=0; j<num_cols; j++)
        {
            double sum      = 0.0;
#ifdef NORMALIZE_CONVOLUTIONS
            double weight_sum = 0.0;
#endif 
            double weight; 

            for (mi = 0; mi < mask_size; mi++)
            {
                for (mj = 0; mj < mask_size; mj++)
                {
                    m = i + mi - mask_width;
                    n = j + mj - mask_width;

                    if (   (m >= 0) && (m < num_rows)
                        && (n >= 0) && (n < num_cols))
                    {
                        weight = mask_mp->elements[mi][mj];
                        sum += in_mp->elements[m][n] * weight; 

#ifdef NORMALIZE_CONVOLUTIONS 
                        weight_sum += weight;
#endif 
                    }
                }
            }
#ifdef NORMALIZE_CONVOLUTIONS 
            sum /= weight_sum; 
#endif 
            out_mp->elements[i][j] = sum; 
        }
    }

    free_matrix(mask_mp);  

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_gaussian_mask
(
    Matrix** mask_mpp /* Output gaussian smoothing mask.           */,
    int mask_size     /* Number of elements per axis. Must be odd. */, 
    double sigma      /* Standard deviation in bin units.         */
)
{
    int r, c;
    int    mask_centre;
    double dist_sqr; 
    double sigma_sqr; 
    double gauss;
    double   mask_sum;
    
    if ( (mask_size < 0 ) || IS_EVEN(mask_size) )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    sigma_sqr = sigma * sigma;

    mask_centre = mask_size / 2;
    mask_sum  = 0.0;
    
    ERE(get_zero_matrix(mask_mpp, mask_size, mask_size));

    for (r = 0; r < mask_size; r++)
    {
        for (c = 0; c < mask_size; c++)
        {
            dist_sqr =(r - mask_centre)*(r - mask_centre) + 
                                            (c - mask_centre)*(c - mask_centre);

            gauss = exp( -0.5 * dist_sqr / sigma_sqr ); 
            (*mask_mpp)->elements[r][c] = gauss;
            mask_sum += gauss;
        }
    }

    ERE(ow_divide_matrix_by_scalar(*mask_mpp,  mask_sum));
   
    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

int get_gaussian_density
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       prob_ptr 
)
{
    double    log_prob;

    ERE(get_log_gaussian_density(x_vp, u_vp, var_vp, &log_prob));

    *prob_ptr = exp(log_prob);

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_log_gaussian_density
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       log_prob_ptr 
)
{
    double dist_sqrd;
    double log_var_prod = 0.0;
    int    dim          = var_vp->length;
    int    i;
    double log_prob;

    ERE(get_malhalanobis_distance_sqrd(x_vp, u_vp, var_vp, &dist_sqrd));

    for (i = 0; i < dim; i++)
    {
        log_var_prod += log(var_vp->elements[ i ]); 
    }

    log_prob = -0.5 * dist_sqrd - (log_var_prod / 2.0) - (((double)dim) / 2.0) * log(2.0 * M_PI); 

    *log_prob_ptr = log_prob;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       dist_ptr 
)
{
    double    dist_sqrd;

    ERE(get_malhalanobis_distance_sqrd(x_vp, u_vp, var_vp, &dist_sqrd));

    *dist_ptr = sqrt(dist_sqrd);

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance_sqrd
(
    const Vector* x_vp,
    const Vector* u_vp,
    const Vector* var_vp,
    double*       dist_sqrd_ptr 
)
{
    Vector* d_vp      = NULL;
    Vector* temp_vp   = NULL;
    double    dist_sqrd;
    int     result;

    result = subtract_vectors(&d_vp, x_vp, u_vp);

    if (result != ERROR)
    {
        result = divide_vectors(&temp_vp, d_vp, var_vp);
    }

    if (result != ERROR)
    {
        result = get_dot_product(d_vp, temp_vp, &dist_sqrd);
    }

    if (dist_sqrd < 0.0)
    {
        db_rv(var_vp); 
    }

    ASSERT(dist_sqrd >= 0.0);

    free_vector(temp_vp);
    free_vector(d_vp);

    if (result != ERROR)
    {
        *dist_sqrd_ptr = dist_sqrd;
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_gaussian_density_2(const Vector* x_vp, const Vector* u_vp, 
                           const Vector* var_vp,
                           double max_norm_var,
                           double* prob_ptr)
{
    double log_prob;

    ERE(get_log_gaussian_density_2(x_vp, u_vp, var_vp, max_norm_var,
                                   &log_prob));

    *prob_ptr = exp(log_prob);

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_log_gaussian_density_2(const Vector* x_vp, 
                               const Vector* u_vp, 
                               const Vector* var_vp, 
                               double max_norm_var, 
                               double* log_prob_ptr)
{
    double dist_sqrd;
    double log_var_prod = 0.0;
    int    dim          = var_vp->length;
    int    i;
    double log_prob;

    ERE(get_malhalanobis_distance_sqrd_2(x_vp, u_vp, var_vp, 
                                         max_norm_var,
                                         &dist_sqrd));

    for (i = 0; i < dim; i++)
    {
        log_var_prod += log(var_vp->elements[ i ]); 
    }

    log_prob = -0.5 * dist_sqrd - (log_var_prod / 2.0) - (((double)dim) / 2.0) * log(2.0 * M_PI); 

    *log_prob_ptr = log_prob;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance_2(const Vector* x_vp, const Vector* u_vp, 
                                const Vector* var_vp, double max_norm_var, 
                                double* dist_ptr)
{
    double    dist_sqrd;

    ERE(get_malhalanobis_distance_sqrd_2(x_vp, u_vp, var_vp, 
                                         max_norm_var, &dist_sqrd));

    *dist_ptr = sqrt(dist_sqrd);

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance_sqrd_2(const Vector* x_vp, 
                                     const Vector* u_vp, 
                                     const Vector* var_vp,
                                     double max_norm_var, 
                                     double* dist_sqrd_ptr)
{
    Vector* d_vp      = NULL;
    Vector* temp_vp   = NULL;
    double  dist_sqrd = DBL_NOT_SET;
    int     result;

    result = subtract_vectors(&d_vp, x_vp, u_vp);

    if (result != ERROR)
    {
        result = divide_vectors(&temp_vp, d_vp, var_vp);
    }

    if (result != ERROR)
    {
        result = ow_multiply_vectors(temp_vp, d_vp);
    }

    if ((result != ERROR) && (max_norm_var > 0.0))
    {
        result = ow_max_thresh_vector(temp_vp, max_norm_var);
    }

    if (result != ERROR)
    {
        dist_sqrd = sum_vector_elements(temp_vp);

        if (dist_sqrd < 0.0)
        {
            db_rv(var_vp); 
        }

        ASSERT(dist_sqrd >= 0.0);
    }

    free_vector(temp_vp);
    free_vector(d_vp);

    if (result != ERROR)
    {
        *dist_sqrd_ptr = dist_sqrd;
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_gaussian_density_3(const Vector* x_vp, const Vector* u_vp, 
                           const Vector* var_vp,
                           double max_norm_var,
                           double* prob_ptr)
{
    double log_prob;

    ERE(get_log_gaussian_density_3(x_vp, u_vp, var_vp, max_norm_var,
                                   &log_prob));

    *prob_ptr = exp(log_prob);

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_log_gaussian_density_3(const Vector* x_vp, 
                               const Vector* u_vp, 
                               const Vector* var_vp, 
                               double max_norm_var, 
                               double* log_prob_ptr)
{
    double dist_sqrd;
    double log_var_prod = 0.0;
    int    max_dim = x_vp->length; 
    int    dim;
    int    i;
    double log_prob;

    ERE(dim = get_malhalanobis_distance_sqrd_3(x_vp, u_vp, var_vp, 
                                               max_norm_var,
                                               &dist_sqrd));

    for (i = 0; i < max_dim; i++)
    {
        if (var_vp->elements[ i ] > 0.0)
        {
            log_var_prod += log(var_vp->elements[ i ]); 
        }
    }

    log_prob = -0.5 * dist_sqrd - (log_var_prod / 2.0) - (((double)dim) / 2.0) * log(2.0 * M_PI); 

    *log_prob_ptr = log_prob;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance_3(const Vector* x_vp, const Vector* u_vp, 
                                const Vector* var_vp, double max_norm_var, 
                                double* dist_ptr)
{
    double    dist_sqrd;
    int dim; 

    ERE(dim = get_malhalanobis_distance_sqrd_3(x_vp, u_vp, var_vp, 
                                               max_norm_var, &dist_sqrd));

    *dist_ptr = sqrt(dist_sqrd);

    return dim; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_malhalanobis_distance_sqrd_3(const Vector* x_in_vp, 
                                     const Vector* u_in_vp, 
                                     const Vector* var_in_vp,
                                     double max_norm_var, 
                                     double* dist_sqrd_ptr)
{
    Vector* d_vp      = NULL;
    Vector* temp_vp   = NULL;
    double  dist_sqrd = DBL_NOT_SET;
    int     result;
    int     dim = 0; 
    Vector* x_vp = NULL; 
    Vector* u_vp = NULL; 
    Vector* var_vp = NULL; 
    int     max_dim = x_in_vp->length; 
    int     i; 

    for (i = 0; i < max_dim; i++)
    {
        if (var_in_vp->elements[ i ] > 0.0) 
        {
            dim++;
        }
    }

    ERE(get_target_vector(&x_vp, dim)); 
    ERE(get_target_vector(&u_vp, dim)); 
    ERE(get_target_vector(&var_vp, dim)); 

    dim = 0;

    for (i = 0; i < max_dim; i++)
    {
        if (var_in_vp->elements[ i ] > 0.0) 
        {
            x_vp->elements[ dim ] = x_in_vp->elements[ i ];
            u_vp->elements[ dim ] = u_in_vp->elements[ i ];
            var_vp->elements[ dim ] = var_in_vp->elements[ i ];

            dim++; 
        }
    }

    result = subtract_vectors(&d_vp, x_vp, u_vp);

    if (result != ERROR)
    {
        result = divide_vectors(&temp_vp, d_vp, var_vp);
    }

    if (result != ERROR)
    {
        result = ow_multiply_vectors(temp_vp, d_vp);
    }

    if ((result != ERROR) && (max_norm_var > 0.0))
    {
        result = ow_max_thresh_vector(temp_vp, max_norm_var);
    }

    if (result != ERROR)
    {
        dist_sqrd = sum_vector_elements(temp_vp);

        if (dist_sqrd < 0.0)
        {
            db_rv(var_vp); 
        }

        ASSERT(dist_sqrd >= 0.0);
    }

    free_vector(temp_vp);
    free_vector(d_vp);
    free_vector(x_vp); 
    free_vector(u_vp); 
    free_vector(var_vp); 

    if (result != ERROR)
    {
        *dist_sqrd_ptr = dist_sqrd;
    }

    if (result == ERROR)
    {
        return ERROR; 
    }
    else 
    {
        return dim; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int compute_mean_and_stdev(double* mean_ptr, 
                           double* stdev_ptr, 
                           int num_data_points, 
                           double data_sum, 
                           double data_sqrd_sum)
{
    double temp; 

    if (mean_ptr != NULL)
    {
        if (num_data_points < 1)
        {
            set_error("Attempt to compute mean of %d data points.",
                      num_data_points);
            return ERROR;
        }

        ASSERT_IS_NUMBER_DBL(data_sum);
        ASSERT_IS_FINITE_DBL(data_sum);

        *mean_ptr = data_sum / (double)num_data_points;
    }

    if (stdev_ptr != NULL)
    {
        if (num_data_points < 2)
        {
            set_error("Attempt to compute stdev with %d data points.",
                      num_data_points);
            return ERROR;
        }

        ASSERT_IS_NUMBER_DBL(data_sum);
        ASSERT_IS_FINITE_DBL(data_sum);
        ASSERT_IS_NUMBER_DBL(data_sqrd_sum);
        ASSERT_IS_FINITE_DBL(data_sqrd_sum);

        temp = data_sqrd_sum;
        temp -= data_sum * data_sum / (double)num_data_points; 
        temp /= (num_data_points - 1.0);

        ASSERT_IS_NUMBER_DBL(temp);
        ASSERT_IS_FINITE_DBL(temp);

        if (temp < 0.0) 
        {
            if (ABS_OF(temp / data_sqrd_sum) > 1000.0 * DBL_EPSILON)
            {
                warn_pso("Possible precision or other problem.\n");
                dbe(data_sum); 
                dbe(data_sqrd_sum); 
                dbi(num_data_points);
                dbe(temp); 
            }

            *stdev_ptr = 0.0;
        }
        else
        {
            *stdev_ptr = sqrt(temp);
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// We ignore degrees of freedom for now.
*/
int compute_weighted_mean_and_stdev(double* mean_ptr, double* stdev_ptr, 
                                    double total_weight, double data_sum, 
                                    double data_sqrd_sum)
{
    double temp; 

    if (mean_ptr != NULL)
    {
        *mean_ptr = data_sum / total_weight;
    }

    if (stdev_ptr != NULL)
    {
        ASSERT_IS_NUMBER_DBL(data_sum);
        ASSERT_IS_FINITE_DBL(data_sum);
        ASSERT_IS_NUMBER_DBL(data_sqrd_sum);
        ASSERT_IS_FINITE_DBL(data_sqrd_sum);

        temp = data_sqrd_sum;
        temp -= data_sum * data_sum / total_weight; 
        temp /= total_weight;

        ASSERT_IS_NUMBER_DBL(temp);
        ASSERT_IS_FINITE_DBL(temp);

        if (temp < 0.0) 
        {
            if (ABS_OF(temp / data_sqrd_sum) > 1000.0 * DBL_EPSILON)
            {
                warn_pso("Possible precision or other problem.\n");
                dbe(data_sum); 
                dbe(data_sqrd_sum); 
                dbe(total_weight);
                dbe(temp); 
            }

            *stdev_ptr = 0.0;
        }
        else
        {
            *stdev_ptr = sqrt(temp);
        }
    }

    return NO_ERROR;
}

