
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "m/m_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "m/m_vec_stat.h" 

/* -------------------------------------------------------------------------- */

static Vector* fs_data_sum_vp            = NULL;
static Vector* fs_data_min_vp            = NULL;
static Vector* fs_data_max_vp            = NULL;
static Vector* fs_data_squared_sum_vp    = NULL;
static int     fs_num_vector_data_points = NOT_SET;
static int     fs_dirty_vector_stat_data = FALSE; 

/* -------------------------------------------------------------------------- */

static int init_vector_data_analysis(int length);

/* -------------------------------------------------------------------------- */

double get_vector_mean(const Vector* vp)
{

    return sum_vector_elements(vp) / vp->length;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Stat* get_vector_stats(const Vector* vp)
{
    int len, i; 
    double* pos; 

    ERN(clear_data_stats());

    len  = vp->length; 
    pos = vp->elements; 

    for (i=0; i<len; i++) 
    {
	  ERN(add_data_point(*pos)); 
	  pos++;
	 }

    return get_data_stats();
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int init_vector_data_analysis(int length)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(cleanup_vector_stats);
        first_time = FALSE;
    }

    fs_num_vector_data_points = 0; 

    NRE(fs_data_sum_vp = create_zero_vector(length)); 
    NRE(fs_data_min_vp = create_zero_vector(length)); 
    NRE(fs_data_max_vp = create_zero_vector(length)); 
    NRE(fs_data_squared_sum_vp = create_zero_vector(length)); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int add_vector_data_point(const Vector* vp)
{
    int     i;
    int     length;
    double* pos;
    double* sum_pos;
    double* squared_sum_pos;
    double* min_pos;
    double* max_pos;

    if (fs_num_vector_data_points == NOT_SET) 
    {
        ERE(init_vector_data_analysis(vp->length)); 
    }
    else if (vp->length != fs_data_sum_vp->length) 
    {
        set_bug("Incorrect vector length in add_vector_data_point.");
        return ERROR; 
    }

    length = vp->length; 
    pos = vp->elements; 

    sum_pos = fs_data_sum_vp->elements; 
    min_pos = fs_data_min_vp->elements; 
    max_pos = fs_data_max_vp->elements; 
    squared_sum_pos = fs_data_squared_sum_vp->elements; 

    fs_num_vector_data_points++;

    for (i=0; i<length; i++) 
    {
        (*sum_pos) += (*pos); 

        if (fs_num_vector_data_points == 1) 
        {
            *min_pos = *pos; 
            *max_pos = *pos; 
        }
        else 
        {
            if (*pos < *min_pos) *min_pos = *pos; 
            if (*pos > *max_pos) *max_pos = *pos; 
        }

        (*squared_sum_pos) += (*pos) * (*pos); 

        pos++;
        sum_pos++;
        squared_sum_pos++;
        min_pos++;
        max_pos++; 
    }

    fs_dirty_vector_stat_data = TRUE; 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Vector_stat* get_vector_data_stats(void)
{
    Vector_stat* vector_stat_ptr;
    double       temp;
    int          i;
    int          length;
    double*      mean_pos;
    double*      stdev_pos;
    double*      sum_pos;
    double*      squared_sum_pos;

    if (fs_num_vector_data_points == NOT_SET) 
    {
        set_bug("Call to get_data_stats before init_vector_data_analysis.");
        return NULL;
    }
   
    if (fs_num_vector_data_points == 0) 
    {
        set_bug("Call to get_vector_data_stats before add_vector_data_point.");
        return NULL;
    }
   
    if (fs_num_vector_data_points == 1) 
    {
        /*
        // CHECK
        // 
        // Should this be a bug? 
        */
        set_bug("Call to get_vector_data_stats for only one point.");
        return NULL;
    }
   
    NRN(vector_stat_ptr = TYPE_MALLOC(Vector_stat)); 

    length = fs_data_sum_vp->length; 

    NRN(vector_stat_ptr->mean_vp = create_vector(length)); 
    NRN(vector_stat_ptr->min_vp = create_vector_copy(fs_data_min_vp));
    NRN(vector_stat_ptr->max_vp = create_vector_copy(fs_data_max_vp));
    NRN(vector_stat_ptr->stdev_vp = create_vector(length)); 

    vector_stat_ptr->n = fs_num_vector_data_points; 

    mean_pos = vector_stat_ptr->mean_vp->elements; 
    stdev_pos = vector_stat_ptr->stdev_vp->elements; 

    sum_pos = fs_data_sum_vp->elements; 
    squared_sum_pos = fs_data_squared_sum_vp->elements; 

    for (i=0; i<length; i++) 
    {
        *mean_pos = (*sum_pos) / ((double) fs_num_vector_data_points); 
        temp = *squared_sum_pos;
        temp -= (*sum_pos) * (*sum_pos) / fs_num_vector_data_points;
        temp /= (fs_num_vector_data_points - 1);

        if (temp < 0.0) 
        {
            if (temp < -DBL_EPSILON) 
            {
                /*
                // FIX
                // 
                // Perhaps it is time we make this a bug ? Perhpas set 
                // the threshold on the basis of lenght?
                */

                dbe(temp); 
                temp = 0.0; 
            }
            temp = 0.0;
        }

        *stdev_pos = sqrt(temp); 

        mean_pos++;
        stdev_pos++;
        sum_pos++;
        squared_sum_pos++;
    }

    fs_dirty_vector_stat_data = FALSE; 

    return vector_stat_ptr;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_vector_stat(Vector_stat* vector_stat_ptr)
{
    
    if (vector_stat_ptr != NULL)
    {
        free_vector(vector_stat_ptr->mean_vp); 
        free_vector(vector_stat_ptr->min_vp); 
        free_vector(vector_stat_ptr->max_vp); 
        free_vector(vector_stat_ptr->stdev_vp); 
       
        kjb_free(vector_stat_ptr);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int clear_vector_stats(void)
{
    if (fs_dirty_vector_stat_data)
    {
        set_bug("Attempt to clear vector stats with dirty stat data.");
        return ERROR;
    }

    fs_num_vector_data_points = NOT_SET; 

    free_vector(fs_data_sum_vp); 
    free_vector(fs_data_min_vp); 
    free_vector(fs_data_max_vp); 
    free_vector(fs_data_squared_sum_vp); 

    fs_data_sum_vp = NULL;
    fs_data_min_vp = NULL;
    fs_data_max_vp = NULL;
    fs_data_squared_sum_vp = NULL;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void cleanup_vector_stats(void)
{
    fs_dirty_vector_stat_data = FALSE; 

    EPE(clear_vector_stats());
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

