
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "n/n_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "n/n_lapack.h" 

/* -------------------------------------------------------------------------- */

#ifndef HAVE_LAPACK 
/* -----------------------------------------------------------------------------
|                              no LAPACK 
|                                  ||   
|                                 \||/  
|                                  \/
*/
static void set_dont_have_lapack_error(void)
{
    set_error("Operation failed because the program was built without ");
    cat_error("the LAPACK interface.");
    add_error("(Appropriate re-compiling is needed to fix this.)");
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no LAPACK        
----------------------------------------------------------------------------- */
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int have_lapack(void)
{
#ifdef HAVE_LAPACK
    return TRUE;
#else 
    return FALSE;
#endif 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef HAVE_LAPACK

#ifdef HPUX 
#    define dgeev_ dgeev
#endif 

/* -----------------------------------------------------------------------------
|                              HAVE_LAPACK
|                                  ||   
|                                 \||/  
|                                  \/
*/

void dgeev_
(
    const char* ,
    const char* ,
    kjb_int32*  N_ptr,
    double*     A,
    kjb_int32*  LDA_ptr,
    double*     WR,
    double*     WI,
    double*     VL,
    kjb_int32*  LDVL_ptr,
    double*     VR,
    kjb_int32*  LDVR_ptr,
    double*     WORK,
    kjb_int32*  LWORK_ptr,
    kjb_int32*  INFO_ptr 
);
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_LAPACK        
----------------------------------------------------------------------------- */

/* =============================================================================
 *                              diagonalize
 *                               
 * Diagonalizes a matrix 
 *
 * Given a matrix A, this routine computes A = inv(E)*D*E, where D is diagonal,
 * and contains the eigenvalues of A. If A is symmetric, then the eigenvalues
 * are double. If A is not symmetric, then the eigenvalues may, or may not be
 * double. Currently, this routine only returns the double parts of the
 * eigenvalues, and prints a warning message if any of the imaginary components
 * of the eigenvalues exceed a certain threshold. 
 *
 * The matrix of eigenvectors is put into the matrix pointed to by *E_mpp,
 * which is created or resized as needed. The diagonal matrix of eigenvalues is
 * put into the vector pointed to by *D_vpp, which is also created or
 * resized as needed. 
 *
 * Note:
 *     This routine requires that this file was compiled with HAVE_LAPACK
 *     #defined. Also, the executable must be liked with the lapack library. If
 *     HAVE_LAPACK is not #defined, then this routine returns ERROR, and sets an
 *     error mesage to this effect. The routine have_lapack() may be used to
 *     determine if this module was compiled with HAVE_LAPACK. 
 *
 * Note:
 *     The underlying fortran routine computes both the double and imaginary
 *     components of all values, and, if needed, this routine could be modified
 *     to create one which returns the imaginary components.
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an appropriate error
 *     message being set. 
 *
 * Index: diagonalization, eigenvalues, eigenvectors, matrix decomposition
 *
 * -----------------------------------------------------------------------------
*/ 

int diagonalize(const Matrix* mp, Matrix** E_mpp, Vector** D_vpp)
{
    kjb_int32   i;
    kjb_int32   j;
    kjb_int32   index;
    kjb_int32   N;
    kjb_int32   INFO;
    kjb_int32   LDA;
    kjb_int32   LWORK;
    kjb_int32   LDVL;
    kjb_int32   LDVR;
    double* A;
    double* WR;
    double* WI;
    double* VL;
    double* VR;
    double* WORK;
    Matrix* E_mp     = NULL;
    Vector* D_vp     = NULL;
    double    max;
    Vector* E_col_vp = NULL;

    NRE(A = get_fortran_1D_dp_array_from_matrix(mp));

    N = mp->num_rows;
    LDVR = LDA = N;
    LDVL = 1;

    NRE(WR = DBL_MALLOC(N));
    NRE(WI = DBL_MALLOC(N));
    NRE(VR = DBL_MALLOC(N * N));
    NRE(VL = DBL_MALLOC(1));

    LWORK = 20 * N; 
    NRE(WORK = DBL_MALLOC(LWORK));  

    dgeev_("N", "V", &N, A, &LDA, WR, WI, VL, &LDVL, VR, &LDVR, WORK, &LWORK,
           &INFO);

    if (INFO == 0) 
    {
        NRE(E_mp = put_fortran_1D_dp_array_into_matrix(N, N, VR)); 

        NRE(D_vp = put_1D_dp_array_into_vector(N, WR)); 

        for (i=0; i<N; i++) 
        {
            if (fabs(WI[ i ]) > 0.00001) 
            {
                warn_pso("Eigenvalue with imaginary componant %e found.\n",
                         WI[ i ]);
            }
        }

        ERE(get_target_matrix(E_mpp, N, N));
        ERE(get_target_vector(D_vpp, N));

        for (i=0; i<N; i++) 
        {
            index = 0;
            max = D_vp->elements[ 0 ];;

            for (j=1; j<N; j++) 
            {
                if (D_vp->elements[ j ] > max) 
                {
                    index = j;
                    max = (D_vp->elements)[ j ];
                }
             }

            ERE(get_matrix_col(&E_col_vp, E_mp, index));
            ERE(put_matrix_col(*E_mpp, E_col_vp, i));
            ((*D_vpp)->elements)[ i ] = (D_vp->elements)[ index ];
            (D_vp->elements)[ index ] = DBL_MOST_NEGATIVE;
        }
     }
    else 
    {
        *E_mpp = NULL;
        *D_vpp = NULL;
    }

    free_vector(E_col_vp);
    free_vector(D_vp);
    free_matrix(E_mp);

    kjb_free(A);
    kjb_free(WR);
    kjb_free(WI);
    kjb_free(VR);
    kjb_free(VL);
    kjb_free(WORK);

    if (INFO == 0) 
    {
        return NO_ERROR;
    }
    else 
    {
        set_error("Unable to diagonalize matrix.");
        return ERROR; 
    }
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_LAPACK        
----------------------------------------------------------------------------- */
#else 
/* -----------------------------------------------------------------------------
|                              no LAPACK 
|                                  ||   
|                                 \||/  
|                                  \/
*/
/*ARGSUSED*/
int diagonalize
(
    const Matrix* dummy_mp,
    Matrix**      dummy_E_mpp,
    Vector**      dummy_D_vpp 
)
{

    set_dont_have_lapack_error();
    return ERROR;
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no LAPACK        
----------------------------------------------------------------------------- */
#endif

/* =============================================================================
 *                             do_lapack_svd
 *                               
 *
 *
 * -----------------------------------------------------------------------------
*/ 

#ifdef HAVE_LAPACK
/* -----------------------------------------------------------------------------
|                              HAVE_LAPACK
|                                  ||   
|                                 \||/  
|                                  \/
*/

#ifdef HPUX 
#    define dgesvd_ dgesvd
#endif 

void dgesvd_
(
    const char* JOBU_ptr,
    const char* JOBVT_ptr,
    kjb_int32*  M_ptr,
    kjb_int32*  N_ptr,
    double*     A_ptr,
    kjb_int32*  LDA_ptr,
    double*     S_ptr,
    double*     U_ptr,
    kjb_int32*  LDU_ptr,
    double*     VT_ptr,
    kjb_int32*  LDVT_ptr,
    double*     WORK_ptr,
    kjb_int32*  LWORK_ptr,
    kjb_int32*  INFO_ptr 
);

/*
//       SUBROUTINE DGESVD( JOBU, JOBVT, M, N, A, LDA, S, U, LDU, VT, LDVT,
//      $                   WORK, LWORK, INFO )
// *
// *  -- LAPACK driver routine (version 2.0) --
// *     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
// *     Courant Institute, Argonne National Lab, and Rice University
// *     September 30, 1994
// *
// *     .. Scalar Arguments ..
//       CHARACTER          JOBU, JOBVT
//       INTEGER            INFO, LDA, LDU, LDVT, LWORK, M, N
// *     ..
// *     .. Array Arguments ..
//       DOUBLE PRECISION   A( LDA, * ), S( * ), U( LDU, * ),
//      $                   VT( LDVT, * ), WORK( * )
// *     ..
// *
// *  Purpose
// *  =======
// *
// *  DGESVD computes the singular value decomposition (SVD) of a double
// *  M-by-N matrix A, optionally computing the left and/or right singular
// *  vectors. The SVD is written
// *
// *       A = U * SIGMA * transpose(V)
// *
// *  where SIGMA is an M-by-N matrix which is zero except for its
// *  min(m,n) diagonal elements, U is an M-by-M orthogonal matrix, and
// *  V is an N-by-N orthogonal matrix.  The diagonal elements of SIGMA
// *  are the singular values of A; they are double and non-negative, and
// *  are returned in descending order.  The first min(m,n) columns of
// *  U and V are the left and right singular vectors of A.
// *
// *  Note that the routine returns V**T, not V.
// *
// *  Arguments
// *  =========
// *
// *  JOBU    (input) CHARACTER*1
// *          Specifies options for computing all or part of the matrix U:
// *          = 'A':  all M columns of U are returned in array U:
// *          = 'S':  the first min(m,n) columns of U (the left singular
// *                  vectors) are returned in the array U;
// *          = 'O':  the first min(m,n) columns of U (the left singular
// *                  vectors) are overwritten on the array A;
// *          = 'N':  no columns of U (no left singular vectors) are
// *                  computed.
// *
// *  JOBVT   (input) CHARACTER*1
// *          Specifies options for computing all or part of the matrix
// *          V**T:
// *          = 'A':  all N rows of V**T are returned in the array VT;
// *          = 'S':  the first min(m,n) rows of V**T (the right singular
// *                  vectors) are returned in the array VT;
// *          = 'O':  the first min(m,n) rows of V**T (the right singular
// *                  vectors) are overwritten on the array A;
// *          = 'N':  no rows of V**T (no right singular vectors) are
// *                  computed.
// *
// *          JOBVT and JOBU cannot both be 'O'.
// *
// *  M       (input) INTEGER
// *          The number of rows of the input matrix A.  M >= 0.
// *
// *  N       (input) INTEGER
// *          The number of columns of the input matrix A.  N >= 0.
// *
// *  A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
// *          On entry, the M-by-N matrix A.
// *          On exit,
// *          if JOBU = 'O',  A is overwritten with the first min(m,n)
// *                          columns of U (the left singular vectors,
// *                          stored columnwise);
// *          if JOBVT = 'O', A is overwritten with the first min(m,n)
// *                          rows of V**T (the right singular vectors,
// *                          stored rowwise);
// *          if JOBU .ne. 'O' and JOBVT .ne. 'O', the contents of A
// *                          are destroyed.
// *
// *  LDA     (input) INTEGER
// *          The leading dimension of the array A.  LDA >= max(1,M).
// *
// *  S       (output) DOUBLE PRECISION array, dimension (min(M,N))
// *          The singular values of A, sorted so that S(i) >= S(i+1).
// *
// *  U       (output) DOUBLE PRECISION array, dimension (LDU,UCOL)
// *          (LDU,M) if JOBU = 'A' or (LDU,min(M,N)) if JOBU = 'S'.
// *          If JOBU = 'A', U contains the M-by-M orthogonal matrix U;
// *          if JOBU = 'S', U contains the first min(m,n) columns of U
// *          (the left singular vectors, stored columnwise);
// *          if JOBU = 'N' or 'O', U is not referenced.
// *
// *  LDU     (input) INTEGER
// *          The leading dimension of the array U.  LDU >= 1; if
// *          JOBU = 'S' or 'A', LDU >= M.
// *
// *  VT      (output) DOUBLE PRECISION array, dimension (LDVT,N)
// *          If JOBVT = 'A', VT contains the N-by-N orthogonal matrix
// *          V**T;
// *          if JOBVT = 'S', VT contains the first min(m,n) rows of
// *          V**T (the right singular vectors, stored rowwise);
// *          if JOBVT = 'N' or 'O', VT is not referenced.
// *
// *  LDVT    (input) INTEGER
// *          The leading dimension of the array VT.  LDVT >= 1; if
// *          JOBVT = 'A', LDVT >= N; if JOBVT = 'S', LDVT >= min(M,N).
// *
// *  WORK    (workspace/output) DOUBLE PRECISION array, dimension (LWORK)
// *          On exit, if INFO = 0, WORK(1) returns the optimal LWORK;
// *          if INFO > 0, WORK(2:MIN(M,N)) contains the unconverged
// *          superdiagonal elements of an upper bidiagonal matrix B
// *          whose diagonal is in S (not necessarily sorted). B
// *          satisfies A = U * B * VT, so it has the same singular values
// *          as A, and singular vectors related by U and VT.
// *
// *  LWORK   (input) INTEGER
// *          The dimension of the array WORK. LWORK >= 1.
// *          LWORK >= MAX(3*MIN(M,N)+MAX(M,N),5*MIN(M,N)-4).
// *          For good performance, LWORK should generally be larger.
// *
// *  INFO    (output) INTEGER
// *          = 0:  successful exit.
// *          < 0:  if INFO = -i, the i-th argument had an illegal value.
// *          > 0:  if DBDSQR did not converge, INFO specifies how many
// *                superdiagonals of an intermediate bidiagonal form B
// *                did not converge to zero. See the description of WORK
// *                above for details.
// *
// *  =====================================================================
*/

int do_lapack_svd
(
    const Matrix* a_mp,
    Matrix**      u_mpp,
    Vector**      d_vpp,
    Matrix**      vt_mpp 
)
{
    kjb_int32   M, N;
    kjb_int32   INFO;
    kjb_int32   LDA;
    kjb_int32   LWORK;
    kjb_int32   LDU;
    kjb_int32   LDVT;
    double* A;
    double* S;
    double* U;
    double* VT;
    double* WORK;
    int     result = NO_ERROR; 

    M = a_mp->num_rows;
    N = a_mp->num_cols;
    LDA = M;
    LDVT = N;
    LDU = M;
    LWORK = MAX_OF(3*MIN_OF(M,N) + MAX_OF(M,N), 5 * MIN_OF(M,N) - 4); 
    /* Docs say LWORK should be bigger for good performance, but they don't say
    // how much bigger! Lets make it 3 times bigger. 
    */
    LWORK *= 3;

    A = get_fortran_1D_dp_array_from_matrix(a_mp);
    S = DBL_MALLOC(MIN_OF(M, N));
    U = DBL_MALLOC(M * MIN_OF(M,N));
    VT = DBL_MALLOC(N * N);
    WORK = DBL_MALLOC(LWORK);  

    if (    (A == NULL) || (S == NULL) || (U == NULL) || (VT == NULL) 
         || (WORK == NULL)
       )
    {
        result = ERROR; 
    }

    if (result != ERROR)
    {
        dgesvd_("S", "A", &M, &N, A, &LDA, S, U, &LDU, VT, &LDVT, WORK, &LWORK,
               &INFO);

        if (INFO < 0) 
        {
            set_bug("Problem with parameter %ld of LAPACK routine dgesvd.", 
                    -(long)INFO);
            result = ERROR;
        }
        else if (INFO > 0) 
        {
            set_error("SVD computation did not converge."); 
            result = ERROR; 
        }
    }

    if ((result != ERROR) && (u_mpp != NULL)) 
    {
        free_matrix(*u_mpp);
        *u_mpp = put_fortran_1D_dp_array_into_matrix(M, MIN_OF(M, N), U); 
        if (*u_mpp == NULL) result = ERROR; 
    }

    if ((result != ERROR) && (d_vpp != NULL))
    {
        free_vector(*d_vpp);
        *d_vpp = put_1D_dp_array_into_vector(MIN_OF(M,N), S); 
        if (*d_vpp == NULL) result = ERROR; 
    }

    if ((result != ERROR) && (vt_mpp != NULL)) 
    {
        free_matrix(*vt_mpp);
        *vt_mpp = put_fortran_1D_dp_array_into_matrix(N, N, VT);
        if (*vt_mpp == NULL) result = ERROR; 
    }

    kjb_free(A);
    kjb_free(S);
    kjb_free(VT);
    kjb_free(U);
    kjb_free(WORK);

    if (result == ERROR)
    {
        add_error("Problem occured using LAPACK routine DGESVD for SVD.");
    }

    return result;
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              HAVE_LAPACK        
----------------------------------------------------------------------------- */
#else 
/* -----------------------------------------------------------------------------
|                              no LAPACK 
|                                  ||   
|                                 \||/  
|                                  \/
*/
/*ARGSUSED*/
int do_lapack_svd
(
    const Matrix* dummy_a_mp,
    Matrix**      dummy_u_mpp,
    Vector**      dummy_d_vpp,
    Matrix**      dummy_vt_mpp 
)
{

    set_dont_have_lapack_error();
    return ERROR;
}
/*
|                                  /\    
|                                 /||\  
|                                  ||
|                              no LAPACK        
----------------------------------------------------------------------------- */
#endif

