
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "p/p_gen_f.h"      /*  Only safe if first #include in a ".c" file  */ 
#include "p/p_plot.h" 

/* -------------------------------------------------------------------------- */

#define MAX_MAX_NUM_PLOTS            1000
#define DEFAULT_MAX_NUM_PLOTS         100 
#define PLOT_COMMAND_BUFF_SIZE      10000
#define SET_PLOT_RANGE(x)        (x>-1e29)

/* -------------------------------------------------------------------------- */

typedef struct Plot 
{   
    int    dim;
    int    plot_pid;
    int    plot_id;
    int    plot_write_des;
    int    plot_read_des;
    int    num_temp_files;
    double x_min;
    double x_max;
    double y_min;
    double y_max;
    double z_min;
    double z_max;
    void*  attached_data_ptr;
    void   (*attached_data_disposal_fn)(void *);
}    
Plot;  

/* -------------------------------------------------------------------------- */

static int            fs_colour_plot_flag        = TRUE;
static int            fs_max_plot_num            = 0;
static int            fs_current_plot            = NOT_SET;

static int            fs_max_num_plots           = DEFAULT_MAX_NUM_PLOTS;

static Plot**         fs_plot_info_array         = NULL;
static int            fs_display_plot_flag       = TRUE;

static Queue_element* fs_display_plot_flag_stack_head = NULL;

#ifdef TRACK_MEMORY_ALLOCATION
    static int        fs_display_plot_flag_stack_first_use = TRUE; 
#endif 

/* -------------------------------------------------------------------------- */

static int plot_open_with_dimension
(
    int dim,
    int x_size,
    int y_size,
    int x_tlc,
    int y_tlc 
);

static Plot* pp_plot_open
(
    int plot_num,
    int x_size,
    int y_size,
    int x_tlc,
    int y_tlc,
    int dim 
);

static Plot* get_plot_ptr(int plot_num);

static int pp_plot_points
(
    Plot*         plot_ptr,
    const Vector* x_vp,
    const Vector* y_vp,
    const char*   name 
);

static int pp_plot_segments
(
    Plot*         plot_ptr,
    const Vector* x1_vp,
    const Vector* y1_vp,
    const Vector* x2_vp,
    const Vector* y2_vp,
    const char*   name 
);

static int pp_plot_update(Plot* plot_ptr);
static int pp_plot_clear(Plot* plot_ptr);
static int pp_plot_clear_data(Plot* plot_ptr);
static int pp_plot_clear_plot(Plot* plot_ptr);
static int pp_plot_close(Plot* plot_ptr);

static int pp_plot_write(Plot* plot_ptr, const char* buff);

static FILE* open_plot_temp_file(int plot_id, const char* file_name);

static int get_plot_temp_file_path
(
    char*       file_path_buff,
    size_t      file_path_buff_size,
    int         plot_id,
    const char* file_name 
);

static int get_plot_temp_file_dir
(
    char*  dir_buff,
    size_t dir_buff_size,
    int    plot_id 
);

static int get_plot_temp_file_name
(
    char*  file_name_buff,
    size_t file_name_buff_size,
    int    file_num 
);

static int plot_remove_data_files(Plot* plot_ptr);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_display_plot_flag_stack(void);
#endif 

/* -------------------------------------------------------------------------- */

int set_colour_plot(void)
{

    fs_colour_plot_flag = TRUE;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_monochrome_plot(void)
{

    fs_colour_plot_flag = FALSE;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_max_num_plots(int max_num_plots)
{
    
    if (max_num_plots < 1)
    {
        set_error("Max num plots must be at least one.");
        return ERROR;
    }
    else if (max_num_plots > MAX_MAX_NUM_PLOTS)
    {
        set_error("Maximum value of max num plots is %d.", MAX_MAX_NUM_PLOTS);
        return ERROR;
    }
    else if ((fs_max_plot_num > 0) && (max_num_plots < fs_max_plot_num))
    {
        set_error("The maximum number of open plots (%d) has exceeded %d.",
                  fs_max_plot_num, max_num_plots);
        add_error("The max number of plots is not changed.");
        return ERROR; 
    }

    fs_max_num_plots = max_num_plots;

    return NO_ERROR; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            push_display_plot_flag 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int push_display_plot_flag(int flag)
{
    int* save_flag_ptr;

    NRE(save_flag_ptr = INT_MALLOC(1));
    *save_flag_ptr = fs_display_plot_flag;

#ifdef TRACK_MEMORY_ALLOCATION
    if (fs_display_plot_flag_stack_first_use) 
    {
        fs_display_plot_flag_stack_first_use = FALSE; 
        add_cleanup_function(free_display_plot_flag_stack);
    }
#endif 

    ERE(insert_into_queue(&fs_display_plot_flag_stack_head, 
                          (Queue_element**)NULL, 
                          (void*)save_flag_ptr));

    fs_display_plot_flag = flag;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            pop_display_plot_flag 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int pop_display_plot_flag(void)
{
    Queue_element* cur_elem; 

    if (fs_display_plot_flag_stack_head == NULL) 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;  
    }

    NRE(cur_elem = remove_first_element(&fs_display_plot_flag_stack_head,
                                        (Queue_element**)NULL));

    fs_display_plot_flag = *((int*)(cur_elem->contents));

    free_queue_element(cur_elem, kjb_free); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* -------------------------------------------------------------------------- */

/*
 * =============================================================================
 *                            plot_open 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_open(void)
{
    return plot_open_with_dimension(2, NOT_SET, NOT_SET, NOT_SET, NOT_SET);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_open3 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_open3(void)
{
    return plot_open_with_dimension(3, NOT_SET, NOT_SET, NOT_SET, NOT_SET);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            special_plot_open
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int special_plot_open
(
    int x_size,
    int y_size,
    int x_tlc,
    int y_tlc,
    int dim 
)
{
    return plot_open_with_dimension(x_size, y_size, x_tlc, y_tlc, dim); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      plot_open_with_dimension
 * -----------------------------------------------------------------------------
 */

static int plot_open_with_dimension
(
    int dim,
    int x_size,
    int y_size,
    int x_tlc,
    int y_tlc 
)
{
    static int victim   = 0;
    Plot*      plot_ptr;
    int        i;
    int        plot_num = NOT_SET;

    if (fs_max_num_plots < 1) 
    {
        set_bug("Max number of plots is not positive.");
        return ERROR;
    }
    else if (fs_max_num_plots > MAX_MAX_NUM_PLOTS)
    {
        SET_CANT_HAPPEN_BUG();
    }

    if (fs_plot_info_array == NULL) 
    {
        SKIP_HEAP_CHECK_2();
        NRE(fs_plot_info_array = N_TYPE_MALLOC(Plot*, MAX_MAX_NUM_PLOTS));
        CONTINUE_HEAP_CHECK_2();

        for (i=0; i<MAX_MAX_NUM_PLOTS; i++) 
        {
            fs_plot_info_array[ i ] = NULL;
        }
    }

    for (i=0; i < fs_max_num_plots; i++) 
    {
        if (fs_plot_info_array[ i ] == NULL)
        {
            plot_num = i;
            break;
        }
    }

    if (plot_num == NOT_SET)
    {
        plot_num = victim % fs_max_num_plots;
        ERE(plot_close(plot_num));
        victim++;
    }

    fs_current_plot = plot_num;

    fs_max_plot_num = MAX_OF(fs_max_plot_num, plot_num);

    if (x_size <= 0) x_size = 800;
    if (y_size <= 0) y_size = 600;
    if (x_tlc  <= 0) x_tlc  = 20+120*(plot_num%4);
    if (y_tlc  <= 0) y_tlc  = 20+120*(plot_num/4);

    NRE(plot_ptr = pp_plot_open(plot_num, x_size, y_size, x_tlc, y_tlc, dim));

    fs_plot_info_array[ plot_num ] = plot_ptr;

    return plot_ptr->plot_id;
}
   
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static Plot* pp_plot_open
(
    int plot_num,
    int x_size,
    int y_size,
    int x_tlc,
    int y_tlc,
    int dim 
)
{
    static int first_time         = TRUE;
    int        plot_pipe[2];
    int        error_pipe[2];
    int        plot_pid;
    char       dir_buff[ MAX_FILE_NAME_SIZE ];

    if (first_time)
    {
        add_cleanup_function(plot_close_all);
        first_time = FALSE; 
    }

    ERN(get_plot_temp_file_dir(dir_buff, sizeof(dir_buff), plot_num + 1));
    ERN(kjb_mkdir(dir_buff)); 

    if ((pipe(plot_pipe) == EOF) || (pipe(error_pipe) == EOF))
    {
        set_error("Creation of pipe for a gnuplot process failed.%S");
        return NULL;
    }

    plot_pid = kjb_fork();

    if (plot_pid < 0) 
    {
        set_error("Fork of gnuplot process failed.%S");
        return NULL;
    }

    if (IS_CHILD(plot_pid)) 
    {
        char system_command[ 256 ]; 

        if (close(plot_pipe[WRITE_END]) == -1) 
        {
            set_bug("Can't close write end of plot pipe.%S");
            return NULL;
        }

        if (dup2(plot_pipe[READ_END],fileno(stdin)) == -1)
        {
            set_bug("Can't duplicate read end of plot pipe as stdin.%S");
            return NULL;
        }

        if (close(error_pipe[READ_END]) == -1)
        {
            set_bug("Can't close read end of plot error pipe.%S");
            return NULL;
        }

        if (dup2(error_pipe[WRITE_END],fileno(stderr)) == -1)
        {
            set_bug("Can't duplicate write end of plot error pipe as stderr.");
            return NULL;
        }
        
        if (X11_is_ok()) 
        {
            if (fs_colour_plot_flag) 
            {
                ERN(kjb_sprintf(system_command, sizeof(system_command), 
                                "gnuplot -geometry %dx%d+%d+%d",
                                x_size, y_size, x_tlc, y_tlc)); 
            }
            else 
            {
                ERN(kjb_sprintf(system_command, sizeof(system_command), 
                                "gnuplot -geometry %dx%d+%d+%d -mono",
                                x_size, y_size, x_tlc, y_tlc)); 
            }
        }
        else 
        {
            if (fs_colour_plot_flag) 
            {
                ERN(kjb_sprintf(system_command, sizeof(system_command),
                                "gnuplot"));
            }
            else 
            {
                ERN(kjb_sprintf(system_command, sizeof(system_command),
                                "gnuplot -mono"));
            }
        }

        ERN(kjb_exec(system_command)); 

        /*NOTREACHED*/

        return NULL;   /* Keep error checkers happy. */
    }
    else  /* Parent */
    {
        char  plot_command_buff [ PLOT_COMMAND_BUFF_SIZE ];
        int   plot_write_des;
        int   plot_read_des;
        Plot* plot_ptr;
        int   i, check_child_res = NOT_SET;

        NRN(plot_ptr = TYPE_MALLOC(Plot)); 
        plot_ptr->num_temp_files = 0; 

        if (close(plot_pipe[ READ_END ]) == -1) 
        {
            set_bug("Can't close read end of plot pipe.%S");
            return NULL;
        }

        if (close(error_pipe[ WRITE_END ]) == -1)
        {
            set_bug("Can't close write end of plot error pipe.%S");
            return NULL;
        }

        i = 0; 
        while (    (i<50) 
                && ((check_child_res = check_child(plot_pid)) != NO_ERROR)
              )
        {
            dbm("wait"); 
            nap(200);
            i++;
        }

        if (check_child_res == NOT_SET)
        {
            SET_CANT_HAPPEN_BUG(); 
            return NULL;
        }
        else if (check_child_res != NO_ERROR)
        {
            set_error("Gnuplot process does not seem to be starting.");
            return NULL;
        }

        plot_write_des = plot_pipe[ WRITE_END ]; 
        plot_read_des  = error_pipe[ READ_END ]; 

        set_no_blocking(plot_read_des); 

        plot_ptr->plot_write_des = plot_write_des;
        plot_ptr->plot_read_des = plot_read_des;
        plot_ptr->plot_id = plot_num + 1;
        plot_ptr->plot_pid = plot_pid;
        plot_ptr->attached_data_ptr = NULL;
        plot_ptr->attached_data_disposal_fn = NULL;
        plot_ptr->x_min = 0.0;
        plot_ptr->x_max = 0.0;
        plot_ptr->y_min = 0.0;
        plot_ptr->y_max = 0.0;
        plot_ptr->z_min = 0.0;
        plot_ptr->z_max = 0.0;
        plot_ptr->dim = dim;

        ERN(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                        "cd \"%s\"\n", dir_buff)); 

        ERN(pp_plot_write(plot_ptr, plot_command_buff));

        ERN(pp_plot_write(plot_ptr, "set key\n"));
   
        if ((! X11_is_ok() ) || (! is_interactive()) || (! fs_display_plot_flag))
        {
            ERN(pp_plot_write(plot_ptr, "set terminal dumb\n"));
            ERN(pp_plot_write(plot_ptr, "set output \'/dev/null\'\n"));
        }
   
        ERN(pp_plot_write(plot_ptr, "set data style line\n"));
        ERN(pp_plot_write(plot_ptr, "set autoscale\n"));

        return plot_ptr;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_set_attached_data
(
    int   plot_id,
    void* data_ptr,
    void  (*data_disposal_fn) (void* ) 
)
{
    Plot* plot_ptr; 

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (    (plot_ptr->attached_data_ptr != NULL)
         && (plot_ptr->attached_data_disposal_fn != NULL)
       )
    {
        (*(plot_ptr->attached_data_disposal_fn))(plot_ptr->attached_data_ptr); 
    }

    plot_ptr->attached_data_ptr = data_ptr;
    plot_ptr->attached_data_disposal_fn = data_disposal_fn;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_get_attached_data(int plot_id, void** data_ptr_ptr)
{
    Plot* plot_ptr; 

    NRE(plot_ptr = get_plot_ptr(plot_id));

    *data_ptr_ptr = plot_ptr->attached_data_ptr;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      get_plot_ptr
 * -----------------------------------------------------------------------------
 */

static Plot* get_plot_ptr(int plot_id)
{
    Plot* plot_ptr;

    if (plot_id == CURRENT_PLOT) 
    {
        if (fs_current_plot < 0) 
        {
            set_error("No current plot."); 
            return NULL;
        }

        plot_ptr = fs_plot_info_array[ fs_current_plot ];

        if (plot_ptr == NULL)
        {
            set_error("The current plot no longer exists.");
        }

        return plot_ptr; 
    }
    else if ((plot_id <= 0) || (plot_id > fs_max_plot_num + 1))
    {
        set_error("%d is an invalid plot number.", plot_id);
        add_error("Currently valid plot numbers are 1 through %d.", 
                  1 + fs_max_plot_num);
        add_error("(Not all of these plots necessarily currenly exist.)"); 
        return NULL;
    }
    else 
    {
        plot_ptr = fs_plot_info_array[ plot_id - 1 ];

        if (plot_ptr == NULL)
        {
            set_error("Plot %d no longer exists.", plot_id);
        }

        return plot_ptr; 
    }

    /*NOTREACHED*/
}
  
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              save_plot 
 *
 * Saves a plot in postscript or EPS form. 
 *
 * This routine is used to save plots in postscript or EPS form. If the filename
 * ends in ".eps", then EPS is used; otherwise PS is used. If the filename is
 * not specified, then "plot.ps" is used. 
 *
 * Due to ugliness with gnuplot, the colours are not exactly what you may
 * expect. I have implemented a hack to help things a little, but there still is
 * room for problems. One problem is that under X-windows, the colours are
 * cycled past 7, but with PS, they are cycled past 8. The colours do not match,
 * and of course the best colour selection for paper is not that same as it is
 * for the screen. Nonetheless, some of the programs which use this routine rely
 * on the fact that the first three colours are red, green, and blue, and hence
 * the hack forces this. 
 *
 * An obvious extension to this routine is to be able to specify the colours,
 * but I have not done this yet. 
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

/*
// Proposed change: Add width and height parms.
//
// To specify a different size, we must change the bounding box, the translation
// and scale parameters in the EPS file. 
*/
int save_plot(int plot_id, const char* file_name)
{
  
    /*
    // This specifies the translation of the nine colours used for postscript to
    // the desired colours. 
    */
    static const struct 
    { 
        const char* in; 
        const char* out;
    }
    substitute_strings[] = {
                                { "0 1 0",             "1 0 0"    }, 
                                { "0 0 1",             "0 1 0"    },
                                { "1 0 0",             "0 0 1"    }, 
                                { "1 0 1",             "1 .3 0"   }, 
                                { "0 1 1",             "1 0 1"    },
                                { "1 1 0",             "0 1 1"    }, 
                                { "0 0 0",             ".2 .2 .3" }, 
                                { "1 0\\.3 0",         ".8 1 0"   }, 
                                { "0\\.5 0\\.5 0\\.5", "1 .2 .3"  }    };
    const int num_substitute_strings = sizeof(substitute_strings)/
                                                sizeof(substitute_strings[ 0 ]);
    char  temp_file_name_path[ MAX_FILE_NAME_SIZE ]; 
    char  file_name_path[ MAX_FILE_NAME_SIZE ];
    char  fixed_file_name_path[ MAX_FILE_NAME_SIZE ]; 
    char  fixed_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  root_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  suffix[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ 2000 ];
    Plot* plot_ptr;
    int   use_eps = FALSE;
    char  fix_command[ 5000 ]; 
    char  fix_substitute[ 500 ]; 
    int   i; 
    FILE* fp; 
    char  temp_name[ MAX_FILE_NAME_SIZE ]; 
    int   lock_wait_count = 0; 

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if ((file_name == NULL) || (*file_name == '\0')) 
    {
        file_name = "plot.ps";
    }

    if (strlen(file_name) > 4) 
    {
        const char* file_name_pos;

        file_name_pos = file_name + strlen(file_name) - 4;
        
        if (STRCMP_EQ(file_name_pos, ".eps")) 
        {
            use_eps = TRUE;
        }
    }

    if (FIND_CHAR_YES(file_name, '.'))
    {
        BUFF_CPY(root_file_name, file_name); 
        BUFF_GEN_GET_LAST_TOKEN(root_file_name, suffix, ".");
        root_file_name[ strlen(root_file_name) - 1 ] = '\0'; 

        BUFF_CPY(fixed_file_name, root_file_name);
        BUFF_CAT(fixed_file_name, "-fixed.");
        BUFF_CAT(fixed_file_name, suffix);
    }
    else 
    {
        BUFF_CPY(fixed_file_name, file_name);
        BUFF_CAT(fixed_file_name, "-fixed");
    }

    NRE(fp = kjb_fopen(file_name, "w"));
    ERE(kjb_fclose(fp)); 

    NRE(fp = kjb_fopen(fixed_file_name, "w"));
    ERE(kjb_fclose(fp)); 

    BUFF_REAL_PATH(file_name, file_name_path);
    BUFF_REAL_PATH(fixed_file_name, fixed_file_name_path);

    if (use_eps) 
    {
        pso("Using EPS\n");

        ERE(pp_plot_write(plot_ptr, 
                       "set terminal postscript eps color \"Times-Roman\"\n"));
    }
    else 
    {
        ERE(pp_plot_write(plot_ptr, 
                       "set terminal postscript color \"Times-Roman\"\n"));
     }

    ERE(get_temp_file_name(temp_file_name_path, sizeof(temp_file_name_path)));

    /*
    pso("Plot output file (must be full path): %s\n", file_name_path);
    pso("Plot fixed output file: %s\n", fixed_file_name_path);
    */

    BUFF_CPY(plot_command_buff, "set output \"");
    BUFF_CAT(plot_command_buff, temp_file_name_path);
    BUFF_CAT(plot_command_buff, "\"\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));
    ERE(pp_plot_write(plot_ptr, "replot\n"));

    BUFF_GET_TEMP_FILE_NAME(temp_name); 

    dbs(temp_name); 

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "!touch %s\n", temp_name));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    nap(100);  /* Get off the CPU. */

    while (    (lock_wait_count < 200) 
            && ( ! is_file(temp_name))
          )
    {
        nap(100); 
        lock_wait_count++; 
    }

    nap(100); 

    EPE(kjb_unlink(temp_name)); 

    BUFF_CPY(fix_command, "/bin/cp "); 
    BUFF_CAT(fix_command, temp_file_name_path); 
    BUFF_CAT(fix_command, " "); 
    BUFF_CAT(fix_command, file_name_path); 

    ERE(kjb_system(fix_command)); 

    fix_command[ 0 ] = '\0'; 

    for (i=0; i<num_substitute_strings; i++)
    {
        if (i != 0) BUFF_CAT(fix_command, " | ");
        BUFF_CAT(fix_command, "sed "); 

        ERE(kjb_sprintf(fix_substitute, sizeof(fix_substitute),
                        " 's/LT%d\\(.*\\)%s DL/LT%d\\1%s DL/' ", 
                        i, substitute_strings[ i ].in, 
                        i, substitute_strings[ i ].out)); 
        
        BUFF_CAT(fix_command, fix_substitute);

        if (i == 0) BUFF_CAT(fix_command, file_name_path);
    }

    BUFF_CAT(fix_command, " > ");
    BUFF_CAT(fix_command, fixed_file_name_path); 

    verbose_pso(200, "\n%s\n\n", fix_command); 

    ERE(kjb_system(fix_command)); 

    EPE(kjb_unlink(temp_file_name_path)); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                              save_plot_as_pbm 
 *
 * Saves a plot in PBM form
 *
 * Returns:
 *     NO_ERROR on success, and ERROR on failure, with an error message being
 *     set.
 * 
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int save_plot_as_pbm(int plot_id, const char* file_name)
{
    char  plot_command_buff[ 2000 ];
    Plot* plot_ptr;
    char  temp_name[ MAX_FILE_NAME_SIZE ]; 
    int   lock_wait_count = 0; 

    NRE(plot_ptr = get_plot_ptr(plot_id));

    BUFF_CPY(plot_command_buff, "set output \"");
    BUFF_CAT(plot_command_buff, file_name);
    BUFF_CAT(plot_command_buff, "\"\n");

    ERE(pp_plot_write(plot_ptr, "set terminal pbm small color\n"));

    ERE(pp_plot_write(plot_ptr, plot_command_buff));
    ERE(pp_plot_write(plot_ptr, "replot\n"));

    if ((! X11_is_ok()) || (! is_interactive()) || ( ! fs_display_plot_flag))
    {
        /*
        // This seems to cause problems, perhaps due to a bug in gnuplot?
        // We can do without it for most situations. 
        // 
        ERE(pp_plot_write(plot_ptr, "set terminal dumb\n"));
        ERE(pp_plot_write(plot_ptr, "set output \'/dev/null\'\n"));
        ERE(pp_plot_write(plot_ptr, "replot\n"));
        */
    }
    else 
    {
        ERE(pp_plot_write(plot_ptr, "set terminal x11\n"));
        ERE(pp_plot_write(plot_ptr, "set output\n"));
        ERE(pp_plot_write(plot_ptr, "replot\n"));
    }

    BUFF_GET_TEMP_FILE_NAME(temp_name); 

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "!touch %s\n", temp_name));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    nap(100);  /* Get off the CPU. */

    lock_wait_count = 0;

    while (    (lock_wait_count < 100) 
            && ( ! is_file(temp_name))
          )
    {
        if (kjb_isatty(fileno(stdout)))
        {
            char mess[ 1000 ];

            ERE(kjb_sprintf(mess, sizeof(mess), 
                            "Waiting for lock file (%04d)\r",
                            lock_wait_count));
            term_puts(mess); 
        }
        nap(200); 
        lock_wait_count++; 
    }

    EPE(kjb_unlink(temp_name)); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            send_command_to_plot 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int send_command_to_plot(int plot_id, const char* command)
{
    Plot* plot_ptr;
    char  plot_command_buff [ 10000 ];  /* Must be big enough for big cut and
                                        // paste for debuging.
                                        */

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (command == NULL) 
    {
        set_bug("Null command argument passed to send_command_to_current_plot");
        return ERROR; 
    }

    BUFF_CPY(plot_command_buff, command);
    BUFF_CAT(plot_command_buff, "\n");
    ERE(pp_plot_write(plot_ptr, plot_command_buff));
    ERE(pp_plot_update(plot_ptr));
            
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_set_title 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_set_title
(
    int         plot_id,
    const char* title,
    int         x_offset,
    int         y_offset 
)
{
    char plot_command_buff [ 256 ];
    char offset_string[ 100 ];

    BUFF_CPY(plot_command_buff, "set title \"");

    if (title != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, title);
        char_for_char_translate(temp_buff, '"', '\''); 

        BUFF_CAT(plot_command_buff, temp_buff);
    }
    else 
    {
        BUFF_CAT(plot_command_buff, "Untitled");
    }

    ERE(kjb_sprintf(offset_string, sizeof(offset_string), "%d,%d", 
                    x_offset, y_offset));

    BUFF_CAT(plot_command_buff, "\"  ");
    BUFF_CAT(plot_command_buff, offset_string);
    BUFF_CAT(plot_command_buff, " \n");

    ERE(plot_write(plot_id, plot_command_buff));
            
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_set_x_legend 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_set_x_legend(int plot_id, const char* legend)
{
    char plot_command_buff [ 256 ];

    BUFF_CPY(plot_command_buff, "set xlabel \"");

    if (legend != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, legend);
        char_for_char_translate(temp_buff, '"', '\''); 

        BUFF_CAT(plot_command_buff, temp_buff);
    }
    else 
    {
        BUFF_CAT(plot_command_buff, "");
    }

    BUFF_CAT(plot_command_buff, "\"");
    BUFF_CAT(plot_command_buff, " 0,-1 \n");

    ERE(plot_write(plot_id, plot_command_buff));
            
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_set_y_legend 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_set_y_legend(int plot_id, const char* legend)
{
    char  plot_command_buff [ 256 ];

    BUFF_CPY(plot_command_buff, "set ylabel \"");

    if (legend != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, legend);
        char_for_char_translate(temp_buff, '"', '\''); 

        BUFF_CAT(plot_command_buff, temp_buff);
    }
    else 
    {
        BUFF_CAT(plot_command_buff, "");
    }

    BUFF_CAT(plot_command_buff, "\"");
    BUFF_CAT(plot_command_buff, " -1,0 \n");

    ERE(plot_write(plot_id, plot_command_buff));
            
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_add_label 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_add_label(int plot_id, char* label, double x, double y)
{
    char plot_command_buff[ 256 ];
    char coord_buff[ 256 ];
    char temp_buff [ 256 ];

    ERE(kjb_sprintf(coord_buff, sizeof(coord_buff), "%f,%f", x, y)); 
    BUFF_CPY(plot_command_buff, "set label \""); 

    BUFF_CPY(temp_buff, label);
    char_for_char_translate(temp_buff, '"', '\''); 

    BUFF_CAT(plot_command_buff, temp_buff);

    BUFF_CAT(plot_command_buff, "\" at "); 
    BUFF_CAT(plot_command_buff, coord_buff); 
    BUFF_CAT(plot_command_buff, "\n"); 

    ERE(plot_write(plot_id, plot_command_buff));
    ERE(plot_update(plot_id));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_set_range 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_set_range
(
    int    plot_id,
    double x_min,
    double x_max,
    double y_min,
    double y_max 
)
{
    char plot_command_buff[ 256 ];
    char num_buff[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if ((SET_PLOT_RANGE(x_min)) || (SET_PLOT_RANGE(x_max))) 
    {
        BUFF_CPY(plot_command_buff, "set xrange ["); 

        if (SET_PLOT_RANGE(x_min)) 
        {
            if (plot_ptr->x_min < x_min) 
            {
                plot_ptr->x_min = x_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(x_max)) 
        {
            if (plot_ptr->x_max > x_max) 
            {
                plot_ptr->x_max = x_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] \n");

        ERE(pp_plot_write(plot_ptr, plot_command_buff));
    }

    if ((SET_PLOT_RANGE(y_min)) || (SET_PLOT_RANGE(y_max))) 
    {
        BUFF_CPY(plot_command_buff, "set yrange ["); 

        if (SET_PLOT_RANGE(y_min)) 
        {
            if (plot_ptr->y_min < y_min) 
            {
                plot_ptr->y_min = y_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(y_max)) 
        {
            if (plot_ptr->y_max > y_max) 
            {
                plot_ptr->y_max = y_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] \n");

        ERE(pp_plot_write(plot_ptr, plot_command_buff));
    }

    if (plot_ptr->num_temp_files >  1) 
    {
        ERE(pp_plot_update(plot_ptr));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_set_range3 
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_set_range3
(
    int    plot_id,
    double x_min,
    double x_max,
    double y_min,
    double y_max,
    double z_min,
    double z_max 
)
{
    char  plot_command_buff[ 256 ];
    char  num_buff[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim < 3)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if ((SET_PLOT_RANGE(x_min)) || (SET_PLOT_RANGE(x_max))) 
    {
        BUFF_CPY(plot_command_buff, "set xrange ["); 

        if (SET_PLOT_RANGE(x_min)) 
        {
            if (plot_ptr->x_min < x_min) 
            {
                plot_ptr->x_min = x_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(x_max)) 
        {
            if (plot_ptr->x_max > x_max) 
            {
                plot_ptr->x_max = x_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] \n");

        ERE(pp_plot_write(plot_ptr, plot_command_buff));
    }

    if ((SET_PLOT_RANGE(y_min)) || (SET_PLOT_RANGE(y_max))) 
    {
        BUFF_CPY(plot_command_buff, "set yrange ["); 

        if (SET_PLOT_RANGE(y_min)) 
        {
            if (plot_ptr->y_min < y_min) 
            {
                plot_ptr->y_min = y_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(y_max)) 
        {
            if (plot_ptr->y_max > y_max) 
            {
                plot_ptr->y_max = y_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] \n");
    }

    if ((SET_PLOT_RANGE(z_min)) || (SET_PLOT_RANGE(z_max))) 
    {
        BUFF_CPY(plot_command_buff, "set yrange ["); 

        if (SET_PLOT_RANGE(z_min)) 
        {
            if (plot_ptr->z_min < z_min) 
            {
                plot_ptr->z_min = z_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",z_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(z_max)) 
        {
            if (plot_ptr->z_max > z_max) 
            {
                plot_ptr->z_max = z_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",z_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] \n");

        ERE(pp_plot_write(plot_ptr, plot_command_buff));
    }

    if (plot_ptr->num_temp_files >  1) 
    {
        ERE(pp_plot_update(plot_ptr));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef NOT_USED_YET 

/*
 * =============================================================================
 *                            plot_function_string
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_function_string
(
    int         plot_id,
    double      x_min,
    double      x_max,
    double      y_min,
    double      y_max,
    const char* function_string 
)
{
    char plot_command_buff[ 256 ];
    char num_buff[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    if ((SET_PLOT_RANGE(x_min)) || (SET_PLOT_RANGE(x_max))) 
    {
        BUFF_CAT(plot_command_buff, " [x="); 

        if (SET_PLOT_RANGE(x_min)) 
        {
            if (plot_ptr->x_min < x_min) 
            {
                plot_ptr->x_min = x_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(x_max)) 
        {
            if (plot_ptr->x_max > x_max) 
            {
                plot_ptr->x_max = x_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",x_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] ");
    }

    if ((SET_PLOT_RANGE(y_min)) || (SET_PLOT_RANGE(y_max))) 
    {
        BUFF_CAT(plot_command_buff, " [y="); 

        if (SET_PLOT_RANGE(y_min)) 
        {
            if (plot_ptr->y_min < y_min) 
            {
                plot_ptr->y_min = y_min;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_min));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, ":");

        if (SET_PLOT_RANGE(y_max)) 
        {
            if (plot_ptr->y_max > y_max) 
            {
                plot_ptr->y_max = y_max;
            }

            ERE(kjb_sprintf(num_buff, sizeof(num_buff), "%f",y_max));
            BUFF_CAT(plot_command_buff, num_buff);
        }

        BUFF_CAT(plot_command_buff, "] ");
    }

    BUFF_CAT(plot_command_buff, function_string);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));
    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_histogram
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_histogram
(
    int           plot_id,
    const Vector* vp,
    int           num_bins,
    const char*   name 
)
{
    double  min, max;
    double  step;
    double  range;
    double  d_num_bins;
    Vector* hist_vp    = NULL;
    int     i, j;
    int     result;

    ERE(get_min_vector_element(vp, &min));
    ERE(get_max_vector_element(vp, &max));

    d_num_bins = num_bins;
    range = max - min;
    step  = range / d_num_bins; 

    dbe(range);
    dbe(min);
    dbe(max);
    dbi(num_bins); 
    dbe(d_num_bins); 

    if (step < 1.0e2 * DBL_MIN)
    {
        set_error("Bin size is too small for histogram plot.");
        return ERROR;
    }

    ERE(get_zero_vector(&hist_vp, num_bins)); 

    for (i = 0; i < vp->length; i++)
    {
        j = d_num_bins * (vp->elements[ i ] - min) / range;

        if (j >= num_bins)  
        {
            j = num_bins - 1;
        }
        else if (j < 0)
        {
            j = 0; 
        }

        (hist_vp->elements[ j ])++;
    }

    result = plot_vector(plot_id, hist_vp, min, step, name); 

    free_vector(hist_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_vector
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_vector
(
    int           plot_id,
    const Vector* vp,
    double        x_offset,
    double        x_step,
    const char*   name 
)
{
    FILE* temp_fp; 
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_num_str[ 256 ]; 
    char  plot_command_buff [ PLOT_COMMAND_BUFF_SIZE ]; 
    int   i; 
    double  x_min; 
    double  x_max; 
    double  temp_x_min; 
    double  temp_x_max; 
    char  line_type_str[ 256 ]; 
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    for (i=0; i<vp->length; i++) 
    {
        ERE(kjb_fprintf(temp_fp,"%e %e\n",x_offset+i*x_step,
                        (vp->elements)[i]));
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = x_offset;
    temp_x_max = x_offset + x_step * vp->length;

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = 0.9 * temp_x_min;
        x_max = 1.1 * temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max; 

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if ((name == NULL) || (*name == '\0')) 
    {
        ERE(kjb_sprintf(plot_num_str, sizeof(plot_num_str), "%d", 
                        (plot_ptr->num_temp_files)));
        BUFF_CAT(plot_command_buff, plot_num_str);
    }
    else 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d",
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_point_list
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_point_list(int plot_id, const Matrix* point_mp, char** names)
{
    FILE*   temp_fp;
    char    temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char    plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double    x_min;
    double    x_max;
    double    temp_x_min;
    double    temp_x_max;
    char    point_type[ 256 ];
    int     i;
    int     num_points;
    Vector* max_vp = NULL;
    Vector* min_vp = NULL;
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    num_points = point_mp->num_rows;

    if (num_points == 0) return NO_ERROR;

    /*
    // Memory leak on failure is OK, because we only expect to fail on
    // allocation failure. 
    */
    ERE(get_min_matrix_col_elements(&min_vp, point_mp));
    ERE(get_max_matrix_col_elements(&max_vp, point_mp));

    temp_x_min = 0.9 * (min_vp->elements)[ 0 ];
    temp_x_max = 1.1 * (max_vp->elements)[ 0 ];

    free_vector(min_vp);
    free_vector(max_vp);

    if (plot_ptr->num_temp_files == 0) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    for (i=0; i<num_points; i++) 
    {
        if (plot_ptr->num_temp_files == 0) 
        {
            BUFF_CPY(plot_command_buff, "plot "); 
        }
        else 
        {
            BUFF_CPY(plot_command_buff, "replot "); 
        }

        (plot_ptr->num_temp_files)++; 

        ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                    plot_ptr->num_temp_files));

        NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, 
                                          temp_file_name));

        ERE(kjb_fprintf(temp_fp,"%e %e\n",
                        (point_mp->elements)[i][0], 
                        (point_mp->elements)[i][1])); 

        ERE(kjb_fclose(temp_fp)); 

        BUFF_CAT(plot_command_buff, "\""); 
        BUFF_CAT(plot_command_buff, temp_file_name); 
        BUFF_CAT(plot_command_buff, "\" t \""); 

        if (names != NULL) 
        {
            BUFF_CAT(plot_command_buff, names[i]);
        }
        
        BUFF_CAT(plot_command_buff, "\" w points ");
        ERE(kjb_sprintf(point_type, sizeof(point_type), "%d\n", 
                        plot_ptr->num_temp_files));
        BUFF_CAT(plot_command_buff, point_type);

        ERE(pp_plot_write(plot_ptr, plot_command_buff));
    }

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_vector_point
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_vector_point
(
    int           plot_id,
    const Vector* point_vp,
    const char*   name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double  x;
    double  y;
    double  x_min;
    double  x_max;
    double  temp_x_min;
    double  temp_x_max;
    char  point_type[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    x = (point_vp->elements)[ 0 ];
    y = (point_vp->elements)[ 1 ];

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    ERE(kjb_fprintf(temp_fp,"%e %e\n",x,y)); 

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = 0.9 * x;
    temp_x_max = 1.1 * y;

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }
    
    BUFF_CAT(plot_command_buff, "\" w points ");
    ERE(kjb_sprintf(point_type, sizeof(point_type), "%d", 
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, point_type);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_point
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_point(int plot_id, double x, double y, const char* name)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double  x_min;
    double  x_max;
    double  temp_x_min;
    double  temp_x_max;
    char  point_type[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    ERE(kjb_fprintf(temp_fp,"%e %e\n",x,y)); 

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = 0.9 * x;
    temp_x_max =1.1 *  y;

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff),
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff[ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }
    
    BUFF_CAT(plot_command_buff, "\" w points ");
    ERE(kjb_sprintf(point_type, sizeof(point_type), "%d", 
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, point_type);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_row_points
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_row_points
(
    int           plot_id,
    const Matrix* mp,
    const char*   name 
)
{
    Vector* x_vp = NULL;
    Vector* y_vp = NULL;
    int result; 
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (mp->num_cols != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    result = get_matrix_col(&x_vp, mp, 0);

    if (result != ERROR)
    {
        result = get_matrix_col(&y_vp, mp, 1);
    }

    if (result != ERROR)
    {
        result = pp_plot_points(plot_ptr, x_vp, y_vp, name);
    }

    free_vector(x_vp);
    free_vector(y_vp); 

    return result; 
}
    
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_points
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_points
(
    int           plot_id,
    const Vector* x_vp,
    const Vector* y_vp,
    const char*   name 
)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_points(plot_ptr, x_vp, y_vp, name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_points
 * -----------------------------------------------------------------------------
 */

static int pp_plot_points
(
    Plot*         plot_ptr,
    const Vector* x_vp,
    const Vector* y_vp,
    const char*   name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int   i;
    double  x_min;
    double  x_max;
    double  temp_x_min;
    double  temp_x_max;
    char  point_type[ 256 ];

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (x_vp->length != y_vp->length) 
    {
        set_bug("X and Y vectors have different lengths in plot_points."); 
        return ERROR; 
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    for (i=0; i<x_vp->length; i++) 
    {
        ERE(kjb_fprintf(temp_fp,"%e %e\n",(x_vp->elements)[i], 
                        (y_vp->elements)[i]));
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = 0.9 * min_vector_element(x_vp);
    temp_x_max = 1.1 * max_vector_element(x_vp);

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }
    
    BUFF_CAT(plot_command_buff, "\" w points ");
    ERE(kjb_sprintf(point_type, sizeof(point_type), "%d", 
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, point_type);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_curve
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_curve
(
    int           plot_id,
    const Vector* x_vp,
    const Vector* y_vp,
    const char*   name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int   i;
    double  x_min;
    double  x_max;
    double  temp_x_min;
    double  temp_x_max;
    char  line_type_str[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (x_vp->length != y_vp->length) 
    {
        set_bug("X and Y vectors have different lengths in plot_curve."); 
        return ERROR; 
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    for (i=0; i<x_vp->length; i++) 
    {
        ERE(kjb_fprintf(temp_fp,"%e %e\n",(x_vp->elements)[i], 
                        (y_vp->elements)[i]));
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = 0.9 * min_vector_element(x_vp);
    temp_x_max = 1.1 * max_vector_element(x_vp);

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d",
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_line
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_line
(
    int         plot_id,
    double      x1,
    double      y1,
    double      x2,
    double      y2,
    const char* name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ]; 
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double  x_min;
    double  x_max;
    double  temp_x_min;
    double  temp_x_max;
    /*
    // char  line_type_str[ 256 ];
    */
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));
    ERE(kjb_fprintf(temp_fp,"%e %e\n%e %e\n", x1, y1, x2, y2));  
    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = 0.9 * MIN_OF(x1, x2);
    temp_x_max = 1.1 * MAX_OF(x1, y2); 

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    /*
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d",
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    */
    BUFF_CAT(plot_command_buff, "0"); 
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_multi_segment_curve
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_multi_segment_curve
(
    int                  plot_id,
    const Matrix_vector* segments,
    const char*          name 
)
{
    int     num_segments;
    FILE*   temp_fp;
    char    temp_file_name[ MAX_FILE_NAME_SIZE ];
    char    plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double    x;
    double    y;
    double    x_min;
    double    x_max;
    double    temp_x_min;
    double    temp_x_max;
    int     i;
    int     count;
    int     num_rows;
    Matrix* segment_mp;
    char    line_type_str[ 256 ];
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    temp_x_max = temp_x_min = segments->elements[ 0 ]->elements[ 0 ][ 0 ];

    num_segments = segments->length;

    for (count=0; count<num_segments; count++) 
    {
        segment_mp = segments->elements[ count ];

        num_rows = segment_mp->num_rows;

        for (i=0; i<num_rows; i++) 
        {
            x = (segment_mp->elements)[i][ 0 ];
            y = (segment_mp->elements)[i][ 1 ];

            if (x < temp_x_min) 
            {
                temp_x_min = x;
            }

            if (x > temp_x_max) 
            {
                temp_x_max = x;
            }

            kjb_fprintf(temp_fp,"%e %e\n",x, y); 
        }
        kjb_fprintf(temp_fp, "\n");
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min *= 0.9; 
    temp_x_max *= 1.1; 

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d",
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_vector_list_cols
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_vector_list_cols
(
    int             plot_id,
    int             list_len,
    Matrix_vector** matrix_vector_list,
    const char*     name 
)
{
    int      count;
    int      i;
    int      j;
    Vector*  x1_vp;
    Vector*  x2_vp;
    Vector*  y1_vp;
    Vector*  y2_vp;
    int      num_segments;
    Matrix** mp_array;
    int      result; 
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    num_segments = 0;

    for (i=0; i<list_len; i++) 
    {
        num_segments += matrix_vector_list[ i ]->length;
    }

    NRE(x1_vp = create_vector(num_segments));
    NRE(x2_vp = create_vector(num_segments));
    NRE(y1_vp = create_vector(num_segments));
    NRE(y2_vp = create_vector(num_segments));

    count = 0;

    for (i=0; i<list_len; i++) 
    {
        mp_array     = (matrix_vector_list[ i ])->elements;
        num_segments = (matrix_vector_list[ i ])->length;

        for (j=0; j<num_segments; j++) 
        {
            (x1_vp->elements)[count] = ((mp_array[ j ])->elements)[0][0];
            (y1_vp->elements)[count] = ((mp_array[ j ])->elements)[0][1];
            (x2_vp->elements)[count] = ((mp_array[ j ])->elements)[1][0];
            (y2_vp->elements)[count] = ((mp_array[ j ])->elements)[1][1];

            count++;
        }
    }

    result = pp_plot_segments(plot_ptr, x1_vp, y1_vp, x2_vp, y2_vp, name);

    free_vector(x1_vp);
    free_vector(x2_vp);
    free_vector(y1_vp);
    free_vector(y2_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_vector_cols
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_vector_cols
(
    int                  plot_id,
    const Matrix_vector* matrix_vector,
    const char*          name 
)
{
    int      num_segments;
    int      i;
    Matrix** mp_array;
    Vector*  x1_vp;
    Vector*  x2_vp;
    Vector*  y1_vp;
    Vector*  y2_vp;
    int      result; 
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    num_segments = matrix_vector->length;
    mp_array =     matrix_vector->elements;

    NRE(x1_vp = create_vector(num_segments));
    NRE(x2_vp = create_vector(num_segments));
    NRE(y1_vp = create_vector(num_segments));
    NRE(y2_vp = create_vector(num_segments));

    for (i=0; i<num_segments; i++) 
    {
        (x1_vp->elements)[i] = ((mp_array[ i ])->elements)[0][0];
        (y1_vp->elements)[i] = ((mp_array[ i ])->elements)[0][1];
        (x2_vp->elements)[i] = ((mp_array[ i ])->elements)[1][0];
        (y2_vp->elements)[i] = ((mp_array[ i ])->elements)[1][1];
    }

    result = pp_plot_segments(plot_ptr,x1_vp, y1_vp, x2_vp, y2_vp, name);

    free_vector(x1_vp);
    free_vector(x2_vp);
    free_vector(y1_vp);
    free_vector(y2_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_segments
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_segments
(
    int           plot_id,
    const Vector* x1_vp,
    const Vector* y1_vp,
    const Vector* x2_vp,
    const Vector* y2_vp,
    const char*   name 
)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_segments(plot_ptr, x1_vp, y1_vp, x2_vp, y2_vp, name);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_segments
 * -----------------------------------------------------------------------------
 */

static int pp_plot_segments
(
    Plot*         plot_ptr,
    const Vector* x1_vp,
    const Vector* y1_vp,
    const Vector* x2_vp,
    const Vector* y2_vp,
    const char*   name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ];
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    double  x_min;
    double  x_max;
    double  temp_x1_min;
    double  temp_x1_max;
    double  temp_x2_min;
    double  temp_x2_max;
    double  temp_x_min;
    double  temp_x_max;
    int   num_segments;
    int   seg_count;
    char  line_type_str[ 256 ];

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (x1_vp->length != x2_vp->length) 
    {
        set_bug("X1 and X2 vectors have different lengths in pp_plot_segments."); 
        return ERROR; 
    }
    if (x1_vp->length != y1_vp->length) 
    {
        set_bug("X1 and Y1 vectors have different lengths in pp_plot_segments."); 
        return ERROR; 
    }
    if (y1_vp->length != y2_vp->length) 
    {
        set_bug("Y1 and Y2 vectors have different lengths in pp_plot_segments."); 
        return ERROR; 
    }

    num_segments = x1_vp->length;

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    for (seg_count=0; seg_count<num_segments; seg_count++) 
    {
        ERE(kjb_fprintf(temp_fp,"%e %e\n",(x1_vp->elements)[seg_count], 
                        (y1_vp->elements)[seg_count]));
        ERE(kjb_fprintf(temp_fp,"%e %e\n\n",(x2_vp->elements)[seg_count], 
                        (y2_vp->elements)[seg_count]));
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x1_min = min_vector_element(x1_vp);
    temp_x1_max = max_vector_element(x1_vp);
    temp_x2_min = min_vector_element(x2_vp);
    temp_x2_max = max_vector_element(x2_vp);

    temp_x_min = MIN_OF(temp_x1_min, temp_x2_min);
    temp_x_max = MAX_OF(temp_x1_max, temp_x2_max); 

    temp_x_min *= 0.9; 
    temp_x_max *= 1.1; 

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max)); 

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max;

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d",
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_cols
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_cols
(
    int           plot_id,
    const Matrix* mp,
    double        x_offset,
    double        x_step,
    const char**  name_list,
    const int*    line_types 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ];
    char  plot_num_str[ 256 ];
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int    i, j;
    double x_min;
    double x_max;
    double temp_x_min;
    double temp_x_max;
    int    line_type;
    char   line_type_str[ 256 ];
    Plot*  plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    temp_x_min = x_offset;
    temp_x_max = x_offset + x_step * mp->num_rows;

    if (plot_ptr->num_temp_files == 0) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff),
                    "set xrange [%f:%f]\n", x_min, x_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max; 

    if (plot_ptr->num_temp_files == 0) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    for (j=0; j < mp->num_cols; j++) 
    {
        (plot_ptr->num_temp_files)++; 

        ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                    plot_ptr->num_temp_files));

        NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

        for (i=0; i<mp->num_rows; i++) 
        {
            ERE(kjb_fprintf(temp_fp,"%e %e\n",x_offset+i*x_step,
                            (mp->elements)[ i ][ j ]));
        }

        ERE(kjb_fclose(temp_fp)); 

        ERE(kjb_sprintf(temp_file_name, sizeof(temp_file_name), "%d",
                        plot_ptr->num_temp_files));

        BUFF_CAT(plot_command_buff, "\""); 
        BUFF_CAT(plot_command_buff, temp_file_name); 
        BUFF_CAT(plot_command_buff, "\" t \""); 

        if ((name_list == NULL) || (name_list[ j ] == NULL))
        {
            ERE(kjb_sprintf(plot_num_str, sizeof(plot_num_str), "%d", 
                            plot_ptr->num_temp_files));
            BUFF_CAT(plot_command_buff, plot_num_str);
        }
        else 
        {
            BUFF_CAT(plot_command_buff, name_list[ j ]); 
        }

        BUFF_CAT(plot_command_buff, "\" w lines ");

        if (line_types == NULL) 
        {
            line_type = plot_ptr->num_temp_files;
        }
        else 
        {
            line_type = line_types[ j ];
        }

        ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d", line_type));
        BUFF_CAT(plot_command_buff, line_type_str);

        if (j == (mp->num_cols - 1)) 
        {
            BUFF_CAT(plot_command_buff, "\n");
        }
        else 
        {
            BUFF_CAT(plot_command_buff, ",");
        }
    }

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_rows
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_rows
(
    int           plot_id,
    const Matrix* mp,
    double        x_offset,
    double        x_step,
    const char**  name_list,
    const int*    line_types 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ];
    char  plot_num_str[ 256 ];
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int    i, j;
    double x_min;
    double x_max;
    double temp_x_min;
    double temp_x_max;
    int    line_type;
    char   line_type_str[ 256 ];
    Plot*  plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    temp_x_min = x_offset;
    temp_x_max = x_offset + x_step * mp->num_cols;

    if (plot_ptr->num_temp_files == 0) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max; 

    if (plot_ptr->num_temp_files == 0) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    for (i=0; i < mp->num_rows; i++) 
    {
        (plot_ptr->num_temp_files) ++; 

        ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                    plot_ptr->num_temp_files));

        NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, 
                                          temp_file_name));

        for (j=0; j<mp->num_cols; j++) 
        {
            ERE(kjb_fprintf(temp_fp,"%e %e\n",x_offset+j*x_step,
                            (mp->elements)[ i ][ j ]));
        }

        ERE(kjb_fclose(temp_fp)); 

        BUFF_CAT(plot_command_buff, "\""); 
        BUFF_CAT(plot_command_buff, temp_file_name); 
        BUFF_CAT(plot_command_buff, "\" t \""); 

        if ((name_list == NULL) || (name_list[ i ] == NULL))
        {
            ERE(kjb_sprintf(plot_num_str, sizeof(plot_num_str), "%d", 
                            plot_ptr->num_temp_files));
            BUFF_CAT(plot_command_buff, plot_num_str);
        }
        else 
        {
            BUFF_CAT(plot_command_buff, name_list[ i ]); 
        }

        BUFF_CAT(plot_command_buff, "\" w lines ");

        if (line_types == NULL) 
        {
            line_type = plot_ptr->num_temp_files;
        }
        else 
        {
            line_type = line_types[ i ];
        }

        ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d", line_type));
        BUFF_CAT(plot_command_buff, line_type_str);

        if (i == (mp->num_rows - 1)) 
        {
            BUFF_CAT(plot_command_buff, "\n");
        }
        else 
        {
            BUFF_CAT(plot_command_buff, ",");
        }
    }

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_multi_matrix_rows
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_multi_matrix_rows
(
    int           plot_id,
    const Matrix* mp,
    double        x_offset,
    double        x_step,
    const char*   name 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ];
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int    i, j;
    double x_min;
    double x_max;
    double temp_x_min;
    double temp_x_max;
    char   line_type_str[ 256 ];
    Plot*  plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 2)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    (plot_ptr->num_temp_files)++; 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, temp_file_name));

    for (i=0; i < mp->num_rows; i++) 
    {
        for (j=0; j<mp->num_cols; j++) 
        {
            ERE(kjb_fprintf(temp_fp, "%e %e\n", x_offset+j*x_step,
                            (mp->elements)[i][j]));
        }
        ERE(kjb_fprintf(temp_fp,"\n\n"));
    }

    ERE(kjb_fclose(temp_fp)); 

    temp_x_min = x_offset;
    temp_x_max = x_offset + x_step * mp->num_cols;

    if (plot_ptr->num_temp_files == 1) 
    {
        x_min = temp_x_min;
        x_max = temp_x_max; 
    }
    else 
    {
        x_min = MIN_OF(temp_x_min, plot_ptr->x_min);
        x_max = MAX_OF(temp_x_max, plot_ptr->x_max);
    }

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->x_min = x_min;
    plot_ptr->x_max = x_max; 

    if (plot_ptr->num_temp_files == 1) 
    {
        BUFF_CPY(plot_command_buff, "plot "); 
    }
    else 
    {
        BUFF_CPY(plot_command_buff, "replot "); 
    }

    BUFF_CAT(plot_command_buff, "\""); 
    BUFF_CAT(plot_command_buff, temp_file_name); 
    BUFF_CAT(plot_command_buff, "\" t \""); 

    if (name != NULL) 
    {
        char temp_buff [ 256 ];

        BUFF_CPY(temp_buff, name);
        char_for_char_translate(temp_buff, '"', '\''); 
        BUFF_CAT(plot_command_buff, temp_buff);
    }

    BUFF_CAT(plot_command_buff, "\" w lines ");
    ERE(kjb_sprintf(line_type_str, sizeof(line_type_str), "%d", 
                    plot_ptr->num_temp_files));
    BUFF_CAT(plot_command_buff, line_type_str);
    BUFF_CAT(plot_command_buff, "\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_matrix_values
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_matrix_values
(
    int           plot_id,
    const Matrix* mp,
    double        x_min,
    double        x_max,
    double        y_min,
    double        y_max 
)
{
    FILE* temp_fp;
    char  temp_file_name[ MAX_FILE_NAME_SIZE ];
    char  plot_command_buff[ PLOT_COMMAND_BUFF_SIZE ];
    int   i, j;
    double  x_step;
    double  y_step;
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    if (plot_ptr->dim != 3)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if ((mp->num_rows < 2) || (mp->num_cols < 2))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    x_step = (x_max - x_min) / (mp->num_rows - 1);
    y_step = (y_max - y_min) / (mp->num_cols - 1);

    ERE(pp_plot_write(plot_ptr, "set parametric\n"));
    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set xrange [%f:%f]\n", x_min, x_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(kjb_sprintf(plot_command_buff, sizeof(plot_command_buff), 
                    "set yrange [%f:%f]\n", y_min, y_max));
    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    plot_ptr->num_temp_files = 1;

    BUFF_CPY(plot_command_buff, "splot '"); 

    ERE(get_plot_temp_file_name(temp_file_name, sizeof(temp_file_name),
                                plot_ptr->num_temp_files));

    NRE(temp_fp = open_plot_temp_file(plot_ptr->plot_id, 
                                      temp_file_name));

    for (i=0; i < mp->num_rows; i++) 
    {
        for (j=0; j<mp->num_cols; j++) 
        {
            ERE(kjb_fprintf(temp_fp,"%e %e %e\n",
                            x_min + (i * x_step), y_min + (j * y_step),
                            (mp->elements)[ i ][ j ]));
        }
        ERE(kjb_fprintf(temp_fp,"\n")); 
    }

    ERE(kjb_fclose(temp_fp)); 

    BUFF_CAT(plot_command_buff, temp_file_name);
    BUFF_CAT(plot_command_buff, "' notitle with lines\n");

    ERE(pp_plot_write(plot_ptr, plot_command_buff));

    ERE(pp_plot_update(plot_ptr));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_update
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_update(int plot_id)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_update(plot_ptr);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_update
 * -----------------------------------------------------------------------------
 */

static int pp_plot_update(Plot* plot_ptr)
{
    const char* plot_command;

    if (plot_ptr->num_temp_files > 0) 
    {
        plot_command = "replot\n";
        ERE(pp_plot_write(plot_ptr, plot_command));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_clear
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_clear(int plot_id)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_clear(plot_ptr);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_clear
 * -----------------------------------------------------------------------------
 */

static int pp_plot_clear(Plot* plot_ptr)
{
    int   result = NO_ERROR;

    if (pp_plot_clear_plot(plot_ptr) == ERROR)
    {
        result = ERROR;
    }

    if (pp_plot_clear_data(plot_ptr) == ERROR)
    {
        result = ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      pp_plot_clear_data
 * -----------------------------------------------------------------------------
 */

static int pp_plot_clear_data(Plot* plot_ptr)
{
    int   result;

    result = plot_remove_data_files(plot_ptr); 

    if (    (plot_ptr->attached_data_ptr != NULL) 
         && (plot_ptr->attached_data_disposal_fn != NULL)
       )
    {
        (*(plot_ptr->attached_data_disposal_fn))(plot_ptr->attached_data_ptr); 
    }

    plot_ptr->attached_data_ptr = NULL; 
    plot_ptr->attached_data_disposal_fn = NULL; 

    plot_ptr->num_temp_files = 0; 
    plot_ptr->x_min = 0.0;
    plot_ptr->x_max = 0.0;
    plot_ptr->y_min = 0.0;
    plot_ptr->y_max = 0.0;

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      pp_plot_clear_plot
 * -----------------------------------------------------------------------------
 */

static int pp_plot_clear_plot(Plot* plot_ptr)
{

    return pp_plot_write(plot_ptr, "set autoscale\nclear\n");
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_close
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_close(int plot_id)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_close(plot_ptr);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_close
 * -----------------------------------------------------------------------------
 */

static int pp_plot_close(Plot* plot_ptr)
{
    IMPORT int  kjb_debug_level;
    const char* plot_command;
    int         plot_index;
    int         result;
    char        dir_buff[ MAX_FILE_NAME_SIZE ];
    
    plot_command = "quit\n";

    /*
    // Can't return in case of error as it might be due to process having
    // disappeared. In this case we still need to be able to remove the files!
    */
    result = pp_plot_write(plot_ptr, plot_command);

    if ((result == ERROR) && (kjb_debug_level > 0))
    {
        kjb_print_error(); 
    }

    /*
    // Give it some time to exit, and then make sure it is dead.
    */
    nap(50);
    terminate_child_process(plot_ptr->plot_pid);

    result = pp_plot_clear_data(plot_ptr); 

    if (get_plot_temp_file_dir(dir_buff, 
                               sizeof(dir_buff), 
                               plot_ptr->plot_id ) == ERROR ) 
    {
        result = ERROR;
    }
    else if (kjb_rmdir(dir_buff) == ERROR)
    {
        result = ERROR;
    }

    plot_index = plot_ptr->plot_id - 1;

    ASSERT(plot_ptr == fs_plot_info_array[ plot_index ]);

    fs_plot_info_array[ plot_index ] = NULL;

    close(plot_ptr->plot_read_des);
    close(plot_ptr->plot_write_des);

    kjb_free(plot_ptr);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_close_all
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

void plot_close_all(void)
{
    Plot*          plot_ptr;
    int            count;
    
    if (fs_plot_info_array != NULL) 
    {
        for (count = 0; count < fs_max_num_plots; count++) 
        {
            plot_ptr = fs_plot_info_array[ count ];

            if (plot_ptr != NULL) 
            {
                EPE(pp_plot_close(plot_ptr));
            }
        }

        fs_max_plot_num = 0; 

        kjb_free(fs_plot_info_array);
        fs_plot_info_array = NULL; 
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 *                            plot_write
 *
 * Index: plotting
 *
 * -----------------------------------------------------------------------------
 */

int plot_write(int plot_id, const char* buff)
{
    Plot* plot_ptr;

    NRE(plot_ptr = get_plot_ptr(plot_id));

    return pp_plot_write(plot_ptr, buff);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                     pp_plot_write
 * -----------------------------------------------------------------------------
 */

static int pp_plot_write(Plot* plot_ptr, const char* buff)
{
    size_t len;
    long   result;
    char   read_buff[ 50000 ];
    int    read_res;
    int    child_status;
 
    /*
    // Have to check that the plot process exists. This check is needed for 
    // for every read due to a number of uglinesses that are beyond our control.
    */
    child_status = check_child(plot_ptr->plot_pid); 

    if (child_status == PROCESS_IS_DEAD)
    {
#ifdef TEST
        dbw();
        dbs(buff); 
        EPE(kjb_system("ps -eaf")); 
        dbw(); 
#endif 
        set_error("Plot process (pid=%d) is no longer alive.",
                  plot_ptr->plot_pid);
        add_error("(If it never existed, this could be a program problem.)");
        return ERROR;
    }

    /*
    // Next swallow any leftover messages for us, especially ones that say that
    // the process has died. 
    */
    read_buff[ 0 ] = '\0';

    /*CONSTCOND*/
    while ( TRUE )
    {
        ERE(read_res = BUFF_DGET_LINE(plot_ptr->plot_read_des, read_buff)); 

        if (read_res > 0)
        {
            if (    (STRCMP_EQ(read_buff, "Broken Pipe")) 
                 || (HEAD_CMP_EQ(read_buff, "XIO:"))
               )
            {
                set_error("Failure writing to plot pipe (process %d).",
                          plot_ptr->plot_pid);
                add_error("It looks like that process is dead. ");
                cat_error("Killing it to make sure."); 
                terminate_child_process(plot_ptr->plot_pid);
                return ERROR; 
            }
            else 
            {
                verbose_pso(1, read_buff);
                verbose_pso(1, "\n"); 
            }
        }
        else break;
    }

    /*
    // OK, the plot process seems to be alive, so continue with the write.
    */
    len = strlen(buff);
 
    verbose_pso(200, buff); 

    result = safe_pipe_write(plot_ptr->plot_write_des, buff, len);

    ASSERT(len <= INT32_MAX);

    if (result != (long)len)
    {
        insert_error("Failure writing to plot pipe (process %d).",
                     plot_ptr->plot_pid);
        return ERROR; 
    }

#ifdef TEST
    /*
    // Get any messages if they are available after 50 ms. to make debugging
    // easier.
    */

    nap(50);

    /*CONSTCOND*/
    while ( TRUE )
    {
        ERE(read_res = BUFF_DGET_LINE(plot_ptr->plot_read_des, read_buff)); 

        if (read_res > 0)
        {
            if (    (STRCMP_EQ(read_buff, "Broken Pipe")) 
                 || (HEAD_CMP_EQ(read_buff, "XIO:"))
               )
            {
                set_error("Failure writing to plot pipe (process %d).",
                          plot_ptr->plot_pid);
                add_error("It looks like that process is dead. ");
                cat_error("Killing it to make sure."); 
                terminate_child_process(plot_ptr->plot_pid);
                return ERROR; 
            }
            else 
            {
                verbose_pso(1, read_buff);
                verbose_pso(1, "\n"); 
            }
        }
        else break;
    }
#endif 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      open_plot_temp_file
 * -----------------------------------------------------------------------------
 */

static FILE* open_plot_temp_file(int plot_id, const char* file_name)
{
    char file_path[ MAX_FILE_NAME_SIZE ];
  
    ERN(get_plot_temp_file_path(file_path, sizeof(file_path), plot_id, 
                                file_name));

    return kjb_fopen(file_path, "w");
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      get_plot_temp_file_path
 * -----------------------------------------------------------------------------
 */

static int get_plot_temp_file_path
(
    char*       file_path_buff,
    size_t      file_path_buff_size,
    int         plot_id,
    const char* file_name 
)
{
    char dir_buff[ MAX_FILE_NAME_SIZE ];
  
    ERE(get_plot_temp_file_dir(dir_buff, sizeof(dir_buff), plot_id));  

    ERE(kjb_sprintf(file_path_buff, file_path_buff_size, "%s/%s", dir_buff,
                    file_name)); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      get_plot_temp_file_dir
 * -----------------------------------------------------------------------------
 */

static int get_plot_temp_file_dir
(
    char*  dir_buff,
    size_t dir_buff_size,
    int    plot_id 
)
{
    char  user_id[ 256 ];
    char user_dir_buff[ MAX_FILE_NAME_SIZE ];

    ERE(get_user_id(user_id, sizeof(user_id))); 

    ERE(kjb_sprintf(user_dir_buff, sizeof(user_dir_buff), "%s%s%s", 
                    TEMP_DIR, DIR_STR, user_id)); 

    if (get_path_type(user_dir_buff) != PATH_IS_DIRECTORY)
    {
        ERE(kjb_mkdir(user_dir_buff)); 
    }

    ERE(kjb_sprintf(dir_buff, dir_buff_size, "%s%s%ld-%d-plots", 
                    user_dir_buff, DIR_STR, (long)MY_PID, plot_id)); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      get_plot_temp_file_name
 * -----------------------------------------------------------------------------
 */

static int get_plot_temp_file_name
(
    char*  file_name_buff,
    size_t file_name_buff_size,
    int    file_num 
)
{
    
    return kjb_sprintf(file_name_buff, file_name_buff_size, "%d", file_num);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
 * =============================================================================
 * STATIC                      plot_remove_data_files
 * -----------------------------------------------------------------------------
 */

static int plot_remove_data_files(Plot* plot_ptr)
{
    char  temp_file_path[ MAX_FILE_NAME_SIZE ];
    char  temp_file_name[ 100 ]; 
    int   i;
    int   temp_result; 
    int   result = NO_ERROR;

    for (i=1; i <= plot_ptr->num_temp_files; i++) 
    {
        EPE(temp_result = get_plot_temp_file_name(temp_file_name, 
                                                  sizeof(temp_file_name), i));

        if (temp_result != ERROR)
        {
            EPE(temp_result = get_plot_temp_file_path(temp_file_path, 
                                                      sizeof(temp_file_path),
                                                      plot_ptr->plot_id,
                                                      temp_file_name));
        }

        if (temp_result != ERROR)
        {
            EPE(temp_result = kjb_unlink(temp_file_path));
        }

        if (temp_result == ERROR) result = ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
/*
 * =============================================================================
 * STATIC                      free_display_plot_flag_stack
 * -----------------------------------------------------------------------------
 */

static void free_display_plot_flag_stack(void)
{
    /* Dont free fs_plot_info_array here, as it is free'd in plot_close_all.*/

    free_queue(&fs_display_plot_flag_stack_head, (Queue_element**)NULL,
               kjb_free); 
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

