
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#ifdef MULTI_TEMPLATE /* Controlled in def.h */

#include "spectra.h"
#include "illum_data.h"
#include "multi_template.h"

/* -------------------------------------------------------------------------- */

/*
// Symbols which able/disable various bits of sample code.
*/

/* 
#define CONFIG_DATA_ACCESS_METHOD_A
*/
#define CONFIG_DATA_ACCESS_METHOD_B 

#define HAVE_DEFAULT_DATA

/* -------------------------------------------------------------------------- */

#ifdef HAVE_DEFAULT_DATA
#    define MULTI_TEMPLATE_DATA_FILE_ONE   "multi_template.rgb" 
#endif 

/* -------------------------------------------------------------------------- */

/*
// These two or three vars are for a single configuration data item. For some
// purposes, only the first two will be used. Note that since these variables
// are declared as static, they are initialized to 0. 
*/

static Matrix* fs_cached_multi_template_data_mp[ MAX_NUM_MULTI_TEMPLATE_ALGS];
static char    fs_multi_template_data_config_file[ MAX_NUM_MULTI_TEMPLATE_ALGS]
                                                         [ MAX_FILE_NAME_SIZE ];
static int     fs_multi_template_data_version[ MAX_NUM_MULTI_TEMPLATE_ALGS];

/*
// Thise vars are for more standard (i.e. non-configuration) options. 
//
// In this example, the first option applies to all variants, but the second
// and third option are potentially different for each variant. 
// 
// IMPORTANT: Most options wil NOT have multi-variants. Specifically, most
// options that modify builds need not be multi-variant, except one, which, at
// a minimum, implements an enable. 
//
// Under the current paradigm, the user will normally do something like this. 
//     <set thing 1 options>
//     make thing:1
//     <set thing 2 options>
//     make thing:2
//     <set thing 3 options>
//     make thing:3
//
// Thus the options that distinguish thing 1 from thing 2 need not be
// multi-variant, as the effect is mirrored in the objects that are made (or
// read from disk). 
//
// We will assume also that the default value for fs_multi_template_real_option
// is zero, which is how a static array gets initialized, just to keep the
// example simple. If these assumptions don't work, then more coding will be
// required.  Note that it would be bad style to simply initialize each member
// of fs_multi_template_real_option, because it is quite likely that someone
// will change MAX_NUM_MULTI_TEMPLATE_ALGS without checking for this kind of
// problem.  I suppose that you could add a comment where
// MAX_NUM_MULTI_TEMPLATE_ALGS is defined, but there are more elegant
// solutions. 
// 
// In this example, we will treat fs_multi_template_integer_option as an option
// which is used only in building. Furthermore, we will assume that a positive
// number is required for the build, and that specifying such a positive number
// is how the user is going to request that a build with that number will take
// place. In otherwords, it implements an enable, as well as possibly other
// stuff. 
*/
static int  fs_multi_template_boolean_option = TRUE;  /* Default value */
static int  fs_multi_template_integer_option[ MAX_NUM_MULTI_TEMPLATE_ALGS];
static double fs_multi_template_real_option[ MAX_NUM_MULTI_TEMPLATE_ALGS];

/* -------------------------------------------------------------------------- */

static int make_multi_template_data
(
    const int   index,
    const char* file_name 
);

static int write_multi_template_data(int index, const char* file_name);

static int read_multi_template_data_from_file
(
    int         index,
    const char* file_name 
);

static int set_multi_template_data_matrix
(
    int           index,
    const Matrix* template_data_mp,
    const char*   config_file 
);

#ifdef CONFIG_DATA_ACCESS_METHOD_A 
    static int get_multi_template_data(int index, Matrix** target_mpp);
#endif 

#ifdef CONFIG_DATA_ACCESS_METHOD_B
    static int initialize_multi_template_data(int index);
#endif 

static void invalidate_multi_template_data(int index);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

/*
// This routine must return one of NO_ERROR, ERROR, and NOT_FOUND, as
// appropriate. 
*/
int set_multi_template_options(const char* option, const char* value)
{
    int  temp_boolean_value;
    int  temp_int;
    double temp_real;
    char lc_option[ 100 ];
    int  result             = NOT_FOUND;
    int  i, index;
    int  first_index;
    int  second_index;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    /*
    // All options which are the same for all variants go first. 
    */

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "multi-template-boolean"))
         /* Abreviated or other spellings follow. */
         || (match_pattern(lc_option, "multi-template-boolean-alternate"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            /* Here we echo the long form the option exactly as it is typed. */
            ERE(pso("multi-template-boolean = %s\n", 
                    fs_multi_template_boolean_option ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            /* Here we echo the option in english. */
            ERE(pso("Multi-template boolean option %s set.\n", 
                    fs_multi_template_boolean_option ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_multi_template_boolean_option = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
    
    /*
    // We now handle options which are variant specific. The user will enter
    // options followed by a ":" followed by a variant number. The user starts
    // counting at one, we start at zero. The meaning of having no index at all
    // is different, depending on whether we are setting or echoing. We reduce
    // the amount of code by preparing for all cases first, even though some of
    // what we are preparing may not be used. For echoing, we use the
    // first_index and last_index, and for setting we use index. When echoing, 
    // "index" gets recycled as a counter. 
    */
    ERE(parse_index(lc_option, MAX_NUM_MULTI_TEMPLATE_ALGS,
                    &first_index, &second_index, &index, (int*)NULL));

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "multi-template-rgb-file"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                              "Multi-template algorithm"));

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_multi_template_data_config_file[ i ];

                ERE(pso("multi-template-rgb-file:%d = %s\n",
                        i + 1, *file_name ? file_name : "off"));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                char* file_name = fs_multi_template_data_config_file[ i ];

                ERE(pso("Multi-template %d rgb data source is %s\n",
                        i + 1, *file_name ? file_name : "<none>"));
            }
        }
        else
        {
            ERE(read_multi_template_data_from_file(index, value)); 
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "multi-template-integer"))
         /* Abreviated or other spellings follow. */
         || (match_pattern(lc_option, "multi-template-integer-alternate"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                              "Multi-template algorithm"));

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                /* 
                // Here we echo the long form the option exactly as it is typed.
                */
                ERE(pso("multi-template-integer:%d = %d\n", 
                        i + 1, fs_multi_template_integer_option[ i ]));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                /* Here we echo the option in english. */
                ERE(pso("Multi-template integer option %d is %d.\n", 
                        i + 1, fs_multi_template_integer_option[ i ]));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                temp_int = NOT_SET;
            }
            else 
            {
                ERE(ss1pi(value, &temp_int)); 
            }

            /*
            // Recall that for our example, fs_multi_template_integer_option being
            // set to non-zero means that we are interested in building the data
            // if a make where to be requested. Since it thus partly an enable,
            // it may be the case that setting it to zero or less should act as
            // a DISABLE. 
            //
            // Note that unlike the case illustrated in template.c, we do not
            // invalidate when the option value changes, becase here we
            // illustrate the second way of dealing with this, namely, the make
            // routine tracks the value. However, we invalidate if the user has
            // set the option "off" in order that the effect is immendiate. 
            */
            if (temp_int <= 0)
            {
                invalidate_multi_template_data(index);
            }
            else 
            {
                fs_multi_template_integer_option[ index ] = temp_int;
            }
        }
        result = NO_ERROR; 
    }
    
    if (   (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "multi-template-double"))
         /* Abreviated or other spellings follow. */
         || (match_pattern(lc_option, "multi-template-double-alternate"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                              "Multi-template algorithm"));

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                /*
                // Here we echo the long form the option exactly as it is typed.
                */
                ERE(pso("multi-template-double:%d = %5.2f\n", 
                        i + 1, fs_multi_template_real_option[ i ]));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                /* Here we echo the option in english. */
                ERE(pso("Multi-template double option %d is %5.2f.\n", 
                        i + 1, fs_multi_template_real_option[ i ]));
            }
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 
            fs_multi_template_real_option[ index ] = temp_real;
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// If your proposed algorithm group does not have any data which is (easily)
// buildable from spectra, then it is not necessary to copy this routine, and
// the static helper routine.
*/

int make_multi_template
(
    const char* make_directive,
    const char* file_name 
)
{
    char lc_make_directive[ 100 ];
    int  result = NOT_FOUND; 
    int  first_index, second_index; 
    int  index; 

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    /*
    // Since we have an enable implemented indirectly through
    // fs_multi_template_integer_option, we ask parse_index for a range, because a
    // user request that has no index should be interpreted as making them all.
    // However, if we did not have a enable implemented, then we would set the
    // first_index_ptr and second_index_ptr args to NULL, forcing parse index to
    // interpret no index as a request for the first variant.
    */
    ERE(parse_index(lc_make_directive, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                    &first_index, &second_index, 
                    &index, (int*)NULL));

    if (    (STRCMP_EQ(lc_make_directive, "all"))
         || (STRCMP_EQ(lc_make_directive, "multi-template"))
       )
    {
        ERE(check_index_range(index, MAX_NUM_MULTI_TEMPLATE_ALGS, 
                              "Multi-template"));

        if (    (second_index > first_index) 
             && (file_name != NULL)
             && (*file_name != '\0')
           )
        {
            set_error("An index must be specified when multi template make has a target file.");
            return ERROR;
        }

        for (index = first_index; index <= second_index; index++)
        {
            ERE(make_multi_template_data(index, file_name));
        }

        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// In this example, we take advantage of the caching already present to only
// build the data if it is out of data with respect to the spectra files. It is
// possible simply to build regardless (i.e. whenever this routine is called),
// but when we run large simulations this routine may be called many times, and
// so the time saved may add up to something. Since we already have a "template"
// for building only when needed, it seems to make sense to use it.
*/
static int make_multi_template_data(int index, const char* file_name)
{
    static int previous_template_data_version[ MAX_NUM_MULTI_TEMPLATE_ALGS ];
    static int previous_world_version[ MAX_NUM_MULTI_TEMPLATE_ALGS ];
    static int first_time  = TRUE;
    int        result      = NO_ERROR;

    if (first_time)
    {
        int i;

        for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
        {
            previous_world_version[ i ] = NOT_SET;
            previous_template_data_version[ i ] = NOT_SET;
        }
        first_time = FALSE;
    }

    /*
    // Normally, we don't want to built MAX_NUM_MULTI_TEMPLATE_ALGS things,
    // only the ones the user requested. For our example, we are assuming
    // that the user expresses interest through
    // fs_multi_template_integer_option.
    */
    if (fs_multi_template_integer_option[ index ] <= 0) return NO_ERROR; 

    if (    (fs_multi_template_data_version[ index ] ==
                                        previous_template_data_version[ index ])
         && (world_is_up_to_date(previous_world_version[ index ]))
       )
    {
        verbose_pso(10, "Multi-template %d data is up to date. ",
                    index + 1); 
        verbose_pso(10, "Skipping make.\n"); 
    }
    else 
    {
        Matrix* data_mp         = NULL;
        int     i;
        int     num_illum_spectra; 
        int     world_version;

        verbose_pso(5, "Multi-template data is out of date. Remaking it.\n"); 

        /*
        // The code here will depend on the nature of the algorithm. 
        //
        // Many of our algorithms take a look at the world under each
        // illuminant, and work out something to be matched, indexed by that
        // illuminant. It is recomended that certain routines are used in this
        // scenario because then we use the same world builder for all code, and
        // things like specularity modeling and fluorescent modeling are all
        // free. Simularly, if we were training on a random selection of
        // illuminants, then the recomended approach would be to request a
        // upgrade for get_synthetic_RGB() which would take a special value for
        // the illuminant to indicate that a random one is needed. 
        */
        result = get_num_illum_spectra(&num_illum_spectra); 

        if (result != ERROR)
        {
            for (i=0; i<num_illum_spectra; i++)
            {
                world_version = get_synthetic_RGB(&data_mp, i);

                if (world_version == ERROR) 
                {
                    result = ERROR;
                    break; 
                }

                /*
                // Do something!
                */
            }
        }

        if (result != ERROR)
        {
            /* 
            // This call updates fs_multi_template_data_version[ index].  For
            // illustrative purposes I have stuck in data_mp as an argument,
            // BUT the actual argument will be something entirely different
            */
            result = set_multi_template_data_matrix(index, data_mp, 
                                                    "<make>");  
        }

        if (result != ERROR) 
        {
            previous_world_version[ index ] = world_version;
            previous_template_data_version[ index ] = 
                                       fs_multi_template_data_version[ index ];
            
            verbose_pso(10, "Multi-template data remade.\n");
        }

        free_matrix(data_mp); 
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        result = write_multi_template_data(index, file_name);
    }
            
    /*
    // Unlike the corresponding example in template.c, we will not return
    // version information, since this routine can affect the version
    // information of more than one algorithm. It is quite unlikely that we
    // would want to write a module which caches the results of this routine,
    // but if this were the case, then we would need some additional code. 
    */

    return result;   /* Either ERROR or NO_ERROR */
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// If your proposed algorithm group does not have any data which is (easily)
// buildable from spectra, then it is not necessary to copy this code.
*/
static int write_multi_template_data(int index, const char* file_name)
{
    return write_matrix(fs_cached_multi_template_data_mp[ index ], file_name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_multi_template_data_from_file
(
    int         index,
    const char* file_name 
)
{
    Matrix* template_data_mp = NULL;
    char    config_file[ MAX_FILE_NAME_SIZE ];
    int     result;

    ERE(read_matrix_from_gamut_config_file(&template_data_mp, 
                                           (const char*)NULL, 
                                           file_name, "template rgb",
                                           config_file, sizeof(config_file)));

    /* 
    // The code that follows must be able to deal with the case that
    // template_data_mp was set to NULL by the above routine (this is not an
    // error). In the case of the this example, set_multi_template_data_matrix
    // can handle a template_data_mp set to NULL, so we are OK. 
    */
    result = set_multi_template_data_matrix(index, template_data_mp,
                                            config_file);

    free_matrix(template_data_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_multi_template_data_matrix
(
    int           index,
    const Matrix* new_template_data_mp,
    const char*   config_file 
)
{

    /*
    // It is OK if new_template_data_mp is NULL. 
    */
    ERE(copy_matrix(&(fs_cached_multi_template_data_mp[ index ]),
                    new_template_data_mp));
    BUFF_CPY(fs_multi_template_data_config_file[ index ], config_file);

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    (fs_multi_template_data_version[ index ])++;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// If we don't need to export the data out of here (the usual case), then we may
// not want to use this routine to "get" the static data because this then
// requires an extra copy (no double big deal unless the data is big). For this,
// or perhaps style reasons, you may wish to use CONFIG_DATA_ACCESS_METHOD_B
// below -- note that the key thing here is that we have to impose certain
// semantics in the read/calculation of the configuration data. (There are three
// ways the data can come to exist: default file read, user specified file read,
// and make). 
//
// Note that this routine implements the same "default config data" semantics as
// does initialize_multi_template_data(). However, note that we EITHER have this
// routine (case CONFIG_DATA_ACCESS_METHOD_A) or the initialzation routine (case
// CONFIG_DATA_ACCESS_METHOD_B). Thus, the semantics are implemented in exactly
// one place. 
*/

#ifdef CONFIG_DATA_ACCESS_METHOD_A 
static int get_multi_template_data(int index, Matrix** target_mpp)
{
#ifdef HAVE_DEFAULT_DATA
    static int first_time = TRUE; 
#endif 

    /*
    // Often with a multi-variation algorithm, it will not make much sense to
    // have default files. However, if we want a default for the first 
    // variant (for example), then we would implement it like so. 
    */
#ifdef HAVE_DEFAULT_DATA

    /*
    // Need a clear because we want to do an "insert_error()" without 
    // necessarily having a preceding set_error(); Usually a clear is not
    // needed.
    */
    kjb_clear_error();

    if (first_time && (fs_cached_multi_template_data_mp[ index ] == NULL))
    {
        first_time = FALSE; 

        if (index == 0)
        {
            ERE(read_multi_template_data_from_file(index, MULTI_TEMPLATE_DATA_FILE_ONE));
        }
    }
#endif 

    if (fs_cached_multi_template_data_mp[ index ] == NULL)
    {
#ifdef HAVE_DEFAULT_DATA
        insert_error("No template %d data is available.", index + 1);
#else
        add_error("No template %d data is available.", index + 1);
#endif 
        return ERROR;
    }

    ERE(copy_matrix(target_mpp, fs_cached_multi_template_data_mp[ index ]));

    return NO_ERROR; 
}
#endif  /* Method  A  */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine implements the "default config data" semantics in the case of
// CONFIG_DATA_ACCESS_METHOD_B. The same semantics are implements for case
// CONFIG_DATA_ACCESS_METHOD_A in the get_multi_template_data() routine. Note
// that we only have one or the other of these routines. 
*/

#ifdef CONFIG_DATA_ACCESS_METHOD_B
static int initialize_multi_template_data(int index)
{
#ifdef HAVE_DEFAULT_DATA
    static int     first_time = TRUE; 
#endif 

    /*
    // Often with a multi-variation algorithm, it will not make much sense to
    // have default files. However, if we want a default for the first variant,
    // then we would implement it like so. 
    */
#ifdef HAVE_DEFAULT_DATA
    /*
    // Need a clear because we want to do an "insert_error()" without 
    // necessarily having a preceding set_error(); Usually a clear is not
    // needed.
    */
    kjb_clear_error();

    if (first_time && (fs_cached_multi_template_data_mp[ index ] == NULL))
    {
        first_time = FALSE; 

        if (index == 0)
        {
            ERE(read_multi_template_data_from_file(index, 
                                                 MULTI_TEMPLATE_DATA_FILE_ONE));
        }
    }
#endif 

    if (fs_cached_multi_template_data_mp[ index ] == NULL)
    {
#ifdef HAVE_DEFAULT_DATA
        insert_error("No multi-template data %d is available.", index + 1);
#else
        add_error("No multi-template data %d is available.", index + 1);
#endif 
        return ERROR;
    }

    return NO_ERROR; 
}
#endif  /* Method B */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_multi_template
(
    const Matrix*          data_mp,
    Multi_template_result* result_ptr,
    Multi_template_error*  cpu_time_ptr 
)
{
#ifdef CONFIG_DATA_ACCESS_METHOD_A
    Matrix* local_template_data_mp = NULL;
#endif 
    Vector* grey_result_vp         = NULL;
    Vector* alg_result_vp          = NULL;
    Vector* map_result_vp          = NULL;
    int     i;
    int     result                 = NO_ERROR;
    long    cpu_time_in_ms;

    /* 
    // We use this for the toy algorithm below. It is not likely to be part of
    // the double algorithm.
    */
    ERE(average_matrix_rows(&grey_result_vp, data_mp)); 

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        if (result == ERROR) break; 

#ifdef CONFIG_DATA_ACCESS_METHOD_A
        if (get_multi_template_data(i, &local_template_data_mp) != ERROR)
#endif 
#ifdef CONFIG_DATA_ACCESS_METHOD_B
        if (initialize_multi_template_data(i) != ERROR) 
#endif 
        {
            init_cpu_time();

#ifdef CONFIG_DATA_ACCESS_METHOD_B
            ASSERT(fs_cached_multi_template_data_mp[ i ] != NULL); 
#endif 
            /*
            // Quite likely the guts of the implementation would be another
            // routine. Also, it would use the matrices local_template_data_mp,
            // or fs_cached_multi_template_data_mp[ i ]. 
            //
            // Invent a grey world type algorithm for testing. We can't just
            // call the double grey world, as this code is tied together into a
            // module. We have calculated a grey world result above, so we will
            // make it greener by 10% plus 20% times the index.
            */

            result = copy_vector(&alg_result_vp, grey_result_vp);

            if (result != ERROR) 
            {
                cpu_time_in_ms = get_cpu_time();
                cpu_time_ptr->alg[ i ] = (double)cpu_time_in_ms / 
                                                      MILLI_SECONDS_PER_SECOND;

                /* Make it redder by 10% plus 20% times the index. */
                alg_result_vp->elements[ 0 ] *= (1.1 + i * 0.2);

                /* 
                // If the algorithm computes rg-chromaticity, then we must
                // convert it to a diagonal map as follows. If the algorithm
                // computes something else then some other conversion may be in
                // order. It
                */
                result = get_map_to_canon_illum(&map_result_vp, alg_result_vp);
            }

            if (result != ERROR)
            {
                /*
                // If the algorithm is meant to deliver 2-dimensional results
                // (i.e, it is an algorithm for chromaticity), then we reduce
                // the dimension here. Since this step is the last one, we put
                // the result into the final location. If the steps are
                // different because we are doing something other than
                // rg-chromaticity, then the final result will be updated some
                // place else.
                */
                result = get_projection_vector(&(result_ptr->alg_vp[ i ]),
                                               map_result_vp, DIVIDE_BY_BLUE, 
                                               MAX_X_DIV_B);
            }
        }
    }

#ifdef CONFIG_DATA_ACCESS_METHOD_A
    free_matrix(local_template_data_mp);
#endif 
    free_vector(alg_result_vp);
    free_vector(grey_result_vp);
    free_vector(map_result_vp); 

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void invalidate_multi_template_data(int index)
{
    free_matrix(fs_cached_multi_template_data_mp[ index ]);
    fs_cached_multi_template_data_mp[ index ] = NULL;
    (fs_multi_template_data_version[ index ])++;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    int i; 

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        free_matrix(fs_cached_multi_template_data_mp[ i ]);
    }
}
    
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#endif /* MULTI_TEMPLATE */

