
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin 
|  contributed the interface with the correlation code). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "plot.h" 

/* 
// Kobus : Moved this from gamut_2D.c so it can be used by 3D also. 
*/

/***
****  BROCK - begin blackbody additions
***/
#include "black_body_spectrum.h"
/***
****  BROCK - end blackbody additions
***/

/* -------------------------------------------------------------------------- */

static int fs_plot_scca_points              = TRUE;
static int fs_plot_scia_points              = TRUE;
static int fs_plot_black_body               = FALSE;
static int fs_plot_intersection_of_surface_constraints = TRUE;
static int fs_plot_luminance_constraints    = TRUE;
static int fs_plot_3D_best_fit              = FALSE;
static int fs_plot_2D_best_fit              = FALSE;
static int fs_plot_actual                   = TRUE;
static int fs_plot_gw                       = TRUE;
static int fs_plot_hull_average             = TRUE;
static int fs_plot_2D_max                   = FALSE;
static int fs_plot_3D_max                   = TRUE;
static int fs_plot_luminance                = FALSE;
static int fs_plot_scale_by_max                  = TRUE;
static int fs_plot_weighted_hull_average    = FALSE;
static int fs_plot_correlation              = TRUE;   /* Lindsay */

#ifdef SIMPLE_TEMPLATE
static int fs_plot_simple_template          = TRUE;
#endif

#ifdef TEMPLATE
static int fs_plot_template                 = TRUE;
#endif

#ifdef MULTI_TEMPLATE
static int fs_plot_multi_template           = TRUE;
#endif

static Projection_method fs_projection_method_for_3d_plots = DIVIDE_BY_BLUE;

/* -------------------------------------------------------------------------- */

static Projection_method parse_projection_method_request
(
    const char* request 
);

static int get_projection_method_str
(
    Projection_method projection_method_option_value,
    char*             buff,
    size_t            buff_size 
);

static int get_plot_points
(
    Matrix**             point_plot_mpp,
    char*                point_plot_names[MAX_NUM_PLOT_POINTS],
    int                  canon_gamut_num,
    Result_by_algorithm* result_ptr,
    int                  dim 
);

static int add_plot_point(int dim, Matrix* mp, Vector* vp, int i);

/* -------------------------------------------------------------------------- */

int process_plot_command(char* line_pos)
{
    char  command[ 200 ];
    char  plot_num_str[ 200 ];
    char* save_line_pos;
    int   plot_num;

    trim_beg(&line_pos);

    if (isdigit((int)(*line_pos)))
    {
        if (! BUFF_GET_TOKEN_OK(&line_pos, plot_num_str))
        {
            set_error("Missing rest of plot number description.");
            return ERROR;
        }

        ERE(ss1pi(plot_num_str, &plot_num));
    }
    else
    {
        plot_num = CURRENT_PLOT;
    }

    trim_beg(&line_pos);

    save_line_pos = line_pos;

    if (! BUFF_GET_TOKEN_OK(&line_pos, command))
    {
        set_error("Either a command or \"save\" is required.");
        return ERROR;
    }

    trim_beg(&line_pos);

    if (IC_STRCMP_EQ(command, "save"))
    {
        ERE(save_plot(plot_num, line_pos));
    }
    else if (IC_STRCMP_EQ(command, "image"))
    {
        ERE(write_plot_image(plot_num, line_pos, 1));
    }
    else
    {
        ERE(send_command_to_plot(plot_num, save_line_pos));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_plot_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  temp_int;
    int  temp_boolean_value;
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "pps"))
         || (match_pattern(lc_option, "plot-print-style"))
       )
    {
        IMPORT int print_plot_style;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-print-style = %s\n", print_plot_style ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Plotting style is for %s output.\n",
                    print_plot_style ? "printed" : "display"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            print_plot_style = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "pfp"))
         || (match_pattern(lc_option, "prompt-for-plots"))
       )
    {
        IMPORT int prompt_for_plots;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("prompt-for-plots = %s\n", prompt_for_plots ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (prompt_for_plots)
            {
                ERE(pso("The program prompts for the desired plots.\n"));
            }
            else
            {
                ERE(pso("A single multi-purpose plot is displayed.\n"));
            }
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            prompt_for_plots = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') && (value[ 0 ] == '\0'))
    {
        ERE(pso("\n"));
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-black-body"))
         || (match_pattern(lc_option, "plot-bb"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-black-body = %s\n", fs_plot_black_body ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The projected black body curve %s plotted.\n",
                    fs_plot_black_body ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_black_body = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-illumination-constraints")))
    {
        IMPORT int plot_illumination_constraint;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-illumination-constraint = %s\n", 
                    plot_illumination_constraint ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Illumination constraint %s plotted.\n",
                    plot_illumination_constraint ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            plot_illumination_constraint = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, 
                           "plot-intersection-of-surface-constraints"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-intersection-of-surface-constraints = %s\n", 
                    fs_plot_intersection_of_surface_constraints ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Intersection of surface constraints %s plotted.\n",
                    fs_plot_intersection_of_surface_constraints ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_intersection_of_surface_constraints = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-scia-points")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-scia-points = %s\n", 
                    fs_plot_scia_points ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained illum ava points %s plotted.\n",
                    fs_plot_scia_points ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_scia_points = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-scca-points")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-scca-points = %s\n", 
                    fs_plot_scca_points ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Surface constrained illum ava points %s plotted.\n",
                    fs_plot_scca_points ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_scca_points = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-luminance-constraints")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-luminance-constraints = %s\n", 
                    fs_plot_luminance_constraints ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Luminance constraints %s plotted.\n",
                    fs_plot_luminance_constraints ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_luminance_constraints = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-3d-best-fit"))
         || (match_pattern(lc_option, "plot-best-3d-fit"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-3d-best-fit = %s\n", 
                    fs_plot_3D_best_fit ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 3D fit results %s plotted.\n",
                    fs_plot_3D_best_fit ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_3D_best_fit = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-2d-best-fit"))
         || (match_pattern(lc_option, "plot-best-2d-fit"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-2d-best-fit = %s\n", 
                    fs_plot_2D_best_fit ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Best 2D fit results %s plotted.\n",
                    fs_plot_2D_best_fit ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_2D_best_fit = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "plot-actual")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-actual = %s\n", fs_plot_actual ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Actual illuminant result %s plotted.\n", 
                     fs_plot_actual ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_actual = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "plot-scale-by-max")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-scale-by-max = %s\n", 
                    fs_plot_scale_by_max ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Scale by max results %s plotted.\n", 
                    fs_plot_scale_by_max ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_scale_by_max = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "plot-gw")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-gw = %s\n", fs_plot_gw ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Grey world results %s plotted.\n", 
                    fs_plot_gw ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_gw = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-ha"))
         || (match_pattern(lc_option, "plot-hull-ave"))
         || (match_pattern(lc_option, "plot-hull-average"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-ha = %s\n", fs_plot_hull_average ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Hull averages %s plotted.\n", 
                    fs_plot_hull_average ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_hull_average = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-3d-mv"))
         || (match_pattern(lc_option, "plot-3d-max"))
         || (match_pattern(lc_option, "plot-3d-max-vol"))
         || (match_pattern(lc_option, "plot-3d-max-volume"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-3d-max = %s\n", fs_plot_3D_max ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D Max volume results %s plotted.\n", 
                    fs_plot_3D_max ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_3D_max = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-2d-mv"))
         || (match_pattern(lc_option, "plot-2d-max"))
         || (match_pattern(lc_option, "plot-2d-max-vol"))
         || (match_pattern(lc_option, "plot-2d-max-volume"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-2d-max = %s\n", fs_plot_2D_max ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("2D Max volume results %s plotted.\n", 
                    fs_plot_2D_max ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_2D_max = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "plot-luminance"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-luminance = %s\n", fs_plot_luminance ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Luminance results %s plotted.\n", 
                    fs_plot_luminance ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_luminance = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    if ((lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "plot-wa")))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-wa = %s\n", fs_plot_weighted_hull_average ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Weighted hull average results %s plotted.\n",
                    fs_plot_weighted_hull_average ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_weighted_hull_average = temp_boolean_value;
        }
        result = NO_ERROR; 
    }

    /* Lindsay: Colour by Correlation - Sept 30/98 */
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-corr"))
         || (match_pattern(lc_option, "plot-correlation"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-correlation = %s\n",
                    fs_plot_correlation ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Correlation results %s plotted.\n", 
                    fs_plot_correlation ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_correlation = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
    /* End Lindsay */

#ifdef SIMPLE_TEMPLATE
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-simple-template"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-simple-template = %s\n", 
                    fs_plot_simple_template ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Simple template result %s plotted.\n", 
                    fs_plot_simple_template ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_simple_template = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
#endif 

#ifdef TEMPLATE
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-template"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-template = %s\n", fs_plot_template ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Template results %s plotted.\n", 
                    fs_plot_template ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_template = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
#endif 

#ifdef MULTI_TEMPLATE
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "plot-multi-template"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("plot-multi-template = %s\n", 
                    fs_plot_multi_template ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Multi template results %s plotted.\n", 
                    fs_plot_multi_template ? "are" : "are not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_plot_multi_template = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
#endif 

    /*
    // This should actually be a KJB library option. 
    */
    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "max-num-plots"))
         || (match_pattern(lc_option, "mnp"))
       )
    {
        if ((value != NULL) && (value[ 0 ] != '\0') && (value[ 0 ] != '?')) 
        {
            ERE(ss1i(value, &temp_int)); 
            ERE(set_max_num_plots(temp_int)); 
            result = NO_ERROR; 
        }
    }

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "3d-plot-projection-method"))
         || (match_pattern(lc_option, "plot-3d-projection-method"))
       )
    {
        Projection_method temp_projection_method; 

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            char method_str[ 100 ];

            ERE(get_projection_method_str(fs_projection_method_for_3d_plots,
                                          method_str, sizeof(method_str)));
            ERE(pso("3d-plot-projection-method = %s\n", method_str));
        }
        else if (value[ 0 ] == '\0')
        {
            char method_str[ 100 ];

            ERE(get_projection_method_str(fs_projection_method_for_3d_plots,
                                          method_str, sizeof(method_str)));
            char_for_char_translate(method_str, '-', ' ');
            ERE(pso("The projection method for 3D plots is %s.\n", 
                    method_str));
        }
        else
        {
            temp_projection_method = parse_projection_method_request(value);

            if (temp_projection_method == INVALID_PROJECTION_METHOD)
            {
                return ERROR;
            }

            fs_projection_method_for_3d_plots = temp_projection_method;
            result = NO_ERROR;
        }
        result = NO_ERROR; 
    }

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static Projection_method parse_projection_method_request
(
    const char* request 
)
{

    if (IC_STRCMP_EQ(request, "divide-by-red"))
    {
        return DIVIDE_BY_RED;
    }
    else if (IC_STRCMP_EQ(request, "divide-by-green"))
    {
        return DIVIDE_BY_GREEN;
    }
    else if (IC_STRCMP_EQ(request, "divide-by-blue"))
    {
        return DIVIDE_BY_BLUE;
    }
    else if (IC_STRCMP_EQ(request, "divide-by-sum"))
    {
        return DIVIDE_BY_SUM;
    }
    else if (IC_STRCMP_EQ(request, "onto-rg-plane"))
    {
        return ONTO_RG_PLANE;
    }
    else if (IC_STRCMP_EQ(request, "onto-rb-plane"))
    {
        return ONTO_RB_PLANE;
    }
    else if (IC_STRCMP_EQ(request, "onto-gb-plane"))
    {
        return ONTO_GB_PLANE;
    }
    else 
    {
        set_error("%q is an invalid projection method.", request);
        return INVALID_PROJECTION_METHOD;
    }

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_projection_method_str
(
    Projection_method projection_method_option_value,
    char*             buff,
    size_t            buff_size 
)
{
    const char* projection_method_option_str;

    if (projection_method_option_value == DIVIDE_BY_RED)
    {
        projection_method_option_str = "divide by red";
    }
    else if (projection_method_option_value == DIVIDE_BY_GREEN)
    {
        projection_method_option_str = "divide by green";
    }
    else if (projection_method_option_value == DIVIDE_BY_BLUE)
    {
        projection_method_option_str = "divide by blue";
    }
    else if (projection_method_option_value == DIVIDE_BY_SUM)
    {
        projection_method_option_str = "divide by sum";
    }
    else if (projection_method_option_value == ONTO_RG_PLANE)
    {
        projection_method_option_str = "onto RG plane";
    }
    else if (projection_method_option_value == ONTO_RB_PLANE)
    {
        projection_method_option_str = "onto RB plane";
    }
    else if (projection_method_option_value == ONTO_GB_PLANE)
    {
        projection_method_option_str = "onto GB plane";
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    kjb_strncpy(buff, projection_method_option_str, buff_size); 

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_2D_result
(
    int                  plot_id,
    int                  plot_all_hulls_flag,
    int                  canon_gamut_num,
    Result_by_algorithm* result_ptr,
    Hull_result*         hull_result_ptr,
    double               illum_hull_x_max,
    double               illum_hull_y_max 
)
{
    IMPORT int print_plot_style;
    IMPORT int prompt_for_plots;
    char*      point_plot_names[ MAX_NUM_PLOT_POINTS ];
    Matrix*    vertex_mp;
    Matrix*    point_plot_mp                           = NULL;
    int        i;

    if (hull_result_ptr->surface_hp != NULL)
    {
        if (    ( ! (print_plot_style && plot_all_hulls_flag))
             && (fs_plot_intersection_of_surface_constraints)
           )
        {
            Hull* surface_hp = hull_result_ptr->surface_hp;

            ERE(plot_matrix_vector_cols(plot_id, surface_hp->facets,
                                       "/\\ of S constraints"));
        }
    }

    if (!print_plot_style)
    {
        if (hull_result_ptr->surface_illum_hp != NULL)
        {
            Hull* surface_illum_hp = hull_result_ptr->surface_illum_hp;

            ERE(plot_matrix_vector_cols(plot_id, surface_illum_hp->facets,
                                      "/\\ I and S constraints"));
        }

        if (hull_result_ptr->surface_illum_var_hp != NULL)
        {
            Hull* surface_illum_var_hp = hull_result_ptr->surface_illum_var_hp;

            ERE(plot_matrix_vector_cols(plot_id, surface_illum_var_hp->facets,
                                 "/\\ of V, I, and S constraints"));
        }
    }

    for (i=0; i<MAX_NUM_PLOT_POINTS; i++)
    {
        point_plot_names[ i ] = NULL;
    }

    ERE(get_plot_points(&point_plot_mp, point_plot_names, canon_gamut_num,
                        result_ptr, 2));
    ERE(plot_point_list(plot_id, point_plot_mp, point_plot_names));

    for (i=0; i<MAX_NUM_PLOT_POINTS; i++)
    {
        kjb_free(point_plot_names[ i ]); 
    }

    if ( ! prompt_for_plots)
    {
        if (    (illum_hull_x_max > 0.0)
             && (illum_hull_y_max > 0.0)
           )
        {
            char key_command[ 100 ];
            double x_max = 1.5 * illum_hull_x_max;
            double y_max = 1.5 * illum_hull_y_max;

            ERE(plot_set_range(plot_id, 0.0, x_max, 0.0, y_max));
            ERE(kjb_sprintf(key_command, sizeof(key_command), 
                            "set key %f,%f\n", 1.35 * x_max, y_max));
            ERE(plot_write(plot_id, key_command));
            ERE(plot_update(plot_id));
        }
    }
    else
    {
        if ((hull_result_ptr->surface_hp != NULL) &&
            (yes_no_query("Scale for surface intersection? (y/n) ")))
        {
            Vector* hull_bounds_vp = NULL;
            char key_command[ 100 ];
            double x_max;
            double y_max;

            vertex_mp = (hull_result_ptr->surface_hp)->vertex_mp;

            ERE(get_max_matrix_col_elements(&hull_bounds_vp, vertex_mp));

            x_max = 1.5 * (hull_bounds_vp->elements)[ 0 ];
            y_max = 1.5 * (hull_bounds_vp->elements)[ 1 ]; 

            EPE(plot_set_range(plot_id, 0.0, x_max, 0.0, y_max));
            EPE(kjb_sprintf(key_command, sizeof(key_command), 
                            "set key %f,%f\n", 1.35 * x_max, y_max));
            EPE(plot_write(plot_id, key_command));
            EPE(plot_update(plot_id));

            if ((hull_result_ptr->surface_illum_hp != NULL) &&
                (yes_no_query("Scale for surface illum intersection? (y/n) ")))
            {
                vertex_mp = (hull_result_ptr->surface_illum_hp)->vertex_mp;

                ERE(get_max_matrix_col_elements(&hull_bounds_vp, vertex_mp));

                x_max = 1.5 * (hull_bounds_vp->elements)[ 0 ];
                y_max = 1.5 * (hull_bounds_vp->elements)[ 1 ]; 

                EPE(plot_set_range(plot_id, 0.0, x_max, 0.0, y_max));
                EPE(kjb_sprintf(key_command, sizeof(key_command), 
                                "set key %f,%f\n", 1.35 * x_max, y_max));
                EPE(plot_write(plot_id, key_command));
                EPE(plot_update(plot_id));
            }

            free_vector(hull_bounds_vp);
        }
    }

    free_matrix(point_plot_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_3D_result
(
    int                  plot_id,
    int                  plot_all_hulls_flag,
    int                  canon_gamut_num,
    Result_by_algorithm* alg_result_ptr,
    Hull_result*         hull_result_ptr 
)
{
    IMPORT int     print_plot_style;
    IMPORT int     prompt_for_plots;
    char*          point_plot_names[ MAX_NUM_PLOT_POINTS ];
    Matrix*        point_plot_mp                           = NULL;
    Vector*        surface_hull_max_vp                     = NULL;
    Vector*        surface_illum_hull_max_vp               = NULL;
    Vector*        illum_constrained_surface_max_vp        = NULL;
    int            result                                  = NO_ERROR;
    double         max_x;
    double         max_y;
    int            i;

    if (plot_id <= 0) return NO_ERROR;

    if (hull_result_ptr->surface_hp != NULL)
    {
        if (    ( ! (print_plot_style && plot_all_hulls_flag))
             && (fs_plot_intersection_of_surface_constraints)
           )
        {
            result = plot_3D_hull_projection(plot_id,
                                             hull_result_ptr->surface_hp,
                                             "/\\ of S Constraints",
                                             (Vector**)NULL, 
                                             &surface_hull_max_vp);
        }
    }

    if (    (result != ERROR)
         && (! print_plot_style)
         && (hull_result_ptr->surface_illum_hp != NULL)
       )
    {
        result = plot_3D_hull_projection(plot_id,
                                         hull_result_ptr->surface_illum_hp,
                                        "/\\ of S and I Constraints",
                                        (Vector**)NULL, 
                                        &surface_illum_hull_max_vp);
    }

    if (    (result != ERROR)
         && (hull_result_ptr->illum_constrained_surface_hp != NULL)
       )
    {
        result = plot_3D_hull_projection(plot_id,
                                  hull_result_ptr->illum_constrained_surface_hp,
                                  "Illum constrained S Constraints",
                                  (Vector**)NULL,
                                  &illum_constrained_surface_max_vp);
    }

    if (    (result != ERROR)
         && fs_plot_luminance_constraints
       )
    {
        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            Hull* hp = hull_result_ptr->luminance_surface_illum_hp[ i ];

            if (hp != NULL)
            {
                char title[ 100 ];

                result = kjb_sprintf(title, sizeof(title), 
                                 "/\\ of S, I, and L:%d Constraints",
                                     i + 1);

                if (result == ERROR) break; 

                result = plot_3D_hull_projection(plot_id, hp,
                                                 title, (Vector**)NULL, 
                                                 (Vector**)NULL);

                if (result == ERROR) break; 
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            Hull* hp = hull_result_ptr->luminance_constrained_surface_hp[ i ];

            if (hp != NULL)
            {
                char title[ 100 ];

                result = kjb_sprintf(title, sizeof(title), 
                                 "L:%d and I constrained S", i + 1);

                if (result == ERROR) break; 

                result = plot_3D_hull_projection(plot_id, hp,
                                                 title, (Vector**)NULL, 
                                                 (Vector**)NULL);

                if (result == ERROR) break; 
            }
        }
    }

    if (    (result != ERROR)
         && ( ! print_plot_style) 
         && (hull_result_ptr->surface_illum_var_hp != NULL)
       )
    {
        result = plot_3D_hull_projection(plot_id,
                                         hull_result_ptr->surface_illum_var_hp,
                                       "/\\ S, I, and V Constraints",
                                        (Vector**)NULL, (Vector**)NULL);
    }

    if (   (fs_plot_scia_points)
        && (hull_result_ptr->scia_points_mp != NULL)
        && (hull_result_ptr->scia_points_mp->num_rows > 0)
       )
    {
        EPE(plot_data_projection(plot_id, hull_result_ptr->scia_points_mp, 
                                 "scia points"));
    }

    if (   (fs_plot_scca_points)
        && (hull_result_ptr->scca_points_mp != NULL)
        && (hull_result_ptr->scca_points_mp->num_rows > 0)
       )
    {
        EPE(plot_data_projection(plot_id, hull_result_ptr->scca_points_mp, 
                                 "scca points"));
    }

    for (i=0; i<MAX_NUM_PLOT_POINTS; i++)
    {
        point_plot_names[ i ] = NULL;
    }

    if (result != ERROR) 
    {
        result = get_plot_points(&point_plot_mp, point_plot_names, 
                                 canon_gamut_num, alg_result_ptr, 3);
    }

    if (result != ERROR) 
    {
        result = plot_point_list(plot_id, point_plot_mp, point_plot_names);
    }

    for (i=0; i<MAX_NUM_PLOT_POINTS; i++)
    {
        kjb_free(point_plot_names[ i ]); 
    }

    if (result != ERROR) 
    {   
        char key_command[ 100 ];

        if ( ! prompt_for_plots)
        {
            Vector* hull_max_vp = NULL; 

            if (surface_illum_hull_max_vp != NULL)
            {
                hull_max_vp = surface_illum_hull_max_vp;
            }
            else if (surface_hull_max_vp != NULL)
            {
                hull_max_vp = surface_hull_max_vp;
            }

            if (hull_max_vp != NULL) 
            {
                max_x = 1.5 * hull_max_vp->elements[ 0 ];
                max_y = 1.5 * hull_max_vp->elements[ 1 ];

                EPE(plot_set_range(plot_id, 0.0, max_x, 0.0, max_y)); 
                EPE(kjb_sprintf(key_command, sizeof(key_command), 
                                "set key %f,%f\n", 1.35 * max_x, 1.1 * max_y));
                EPE(plot_write(plot_id, key_command));
                EPE(plot_update(plot_id));
            }
        }
        else
        {
            if (    (surface_hull_max_vp != NULL) 
                 && (yes_no_query("Scale for surface intersection? (y/n) ")))
            {     
                max_x = 1.5 * surface_hull_max_vp->elements[ 0 ];
                max_y = 1.5 * surface_hull_max_vp->elements[ 1 ];

                EPE(plot_set_range( plot_id, 0.0, max_x, 0.0, max_y)); 
                EPE(kjb_sprintf(key_command, sizeof(key_command), 
                                "set key %f,%f\n", 1.35 * max_x, 1.1 * max_y));
                EPE(plot_write(plot_id, key_command));
                EPE(plot_update(plot_id));
            }

            if (    (surface_illum_hull_max_vp != NULL) 
                 && (yes_no_query("Scale for SI intersection? (y/n) "))
               )
            {     
                max_x = 1.5 * surface_illum_hull_max_vp->elements[ 0 ];
                max_y = 1.5 * surface_illum_hull_max_vp->elements[ 1 ];

                EPE(plot_set_range(plot_id, 0.0, max_x, 0.0, max_y)); 
                EPE(kjb_sprintf(key_command, sizeof(key_command), 
                                "set key %f,%f\n", 1.35 * max_x, 1.1 * max_y));
                EPE(plot_write(plot_id, key_command));
                EPE(plot_update(plot_id));
            }

            if (    (illum_constrained_surface_max_vp != NULL) 
                 && (yes_no_query("Scale for illum constrained S intersection? (y/n) "))
               )
            {     
                max_x = illum_constrained_surface_max_vp->elements[0 ];
                max_x *= 1.5;
                max_y = illum_constrained_surface_max_vp->elements[1 ];
                max_y *= 1.5; 

                EPE(plot_set_range(plot_id, 0.0, max_x, 0.0, max_y)); 
                EPE(kjb_sprintf(key_command, sizeof(key_command), 
                                "set key %f,%f\n", 1.35 * max_x, 1.1 * max_y));
                EPE(plot_write(plot_id, key_command));
                EPE(plot_update(plot_id));
            }
        }
    }

    free_vector(surface_hull_max_vp); 
    free_vector(surface_illum_hull_max_vp); 
    free_vector(illum_constrained_surface_max_vp); 
    free_matrix(point_plot_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_data_projection(int plot_id, Matrix* rgb_mp, const char* title)
{
    Invalid_chromaticity_action invalid_chromaticity_action;
    Matrix*                     temp_mp                        = NULL;
    int                         result;
    Vector*                     x_vp = NULL; 
    Vector*                     y_vp = NULL; 

    if (fs_projection_method_for_3d_plots == DIVIDE_BY_SUM)
    {
        invalid_chromaticity_action = SKIP_INVALID_CHROMATICITY;
    }
    else 
    {
        invalid_chromaticity_action = TRUNCATE_INVALID_CHROMATICITY;
    }
                
    result = get_projection_matrix(&temp_mp, rgb_mp, 
                                   fs_projection_method_for_3d_plots, 
                                   invalid_chromaticity_action, 
                                   MAX_PROJECTION_COORDINATE,
                                   DBL_DUMMY_ARG, DBL_DUMMY_ARG,
                                   (Matrix**)NULL, (Matrix**)NULL);

    if (result != ERROR) 
    {
        result = get_matrix_col(&x_vp, temp_mp, 0);
    }

    if (result != ERROR) 
    {
        result = get_matrix_col(&y_vp, temp_mp, 1);
    }

    if (result != ERROR) 
    {
        result = plot_points(plot_id, x_vp, y_vp, title);
    }

    free_matrix(temp_mp);
    free_vector(x_vp);
    free_vector(y_vp); 

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_3D_hull_projection
(
    int         plot_id,
    Hull*       hp,
    const char* title,
    Vector**    min_vpp,
    Vector**    max_vpp 
)
{
    Invalid_chromaticity_action invalid_chromaticity_action;
    Hull*                       temp_hp                        = NULL;
    Matrix*                     trunc_mp                       = NULL;
    Matrix*                     temp_mp                        = NULL;
    int                         result;

    if (    (fs_projection_method_for_3d_plots == DIVIDE_BY_RED)
         || (fs_projection_method_for_3d_plots == DIVIDE_BY_GREEN)
         || (fs_projection_method_for_3d_plots == DIVIDE_BY_BLUE)
         || (fs_projection_method_for_3d_plots == DIVIDE_BY_SUM)
       )
    {
        result = remove_origin_from_data(&trunc_mp, hp->vertex_mp,
                                         MIN_RGB / (2.0 * MAX_RGB));
    }
    else 
    {
        result = copy_matrix(&trunc_mp, hp->vertex_mp);
    }

    if (fs_projection_method_for_3d_plots == DIVIDE_BY_SUM)
    {
        invalid_chromaticity_action = SKIP_INVALID_CHROMATICITY;
    }
    else 
    {
        invalid_chromaticity_action = TRUNCATE_INVALID_CHROMATICITY;
    }
                
    if (result != ERROR) 
    {
        result = get_projection_matrix(&temp_mp, trunc_mp, 
                                       fs_projection_method_for_3d_plots, 
                                       invalid_chromaticity_action, 
                                       MAX_PROJECTION_COORDINATE,
                                       DBL_DUMMY_ARG, DBL_DUMMY_ARG,
                                       (Matrix**)NULL, (Matrix**)NULL);
    }

    if (result != ERROR) 
    {
        temp_hp = find_convex_hull(temp_mp, DEFAULT_HULL_OPTIONS);

        if (temp_hp == NULL) result = ERROR;
    }

    if (result != ERROR) 
    {
        result = find_hull_bounds(temp_hp, min_vpp, max_vpp);
    }

    if (result != ERROR) 
    {
        result = plot_matrix_vector_cols(plot_id, temp_hp->facets, title);
    }

    free_hull(temp_hp);
    free_matrix(trunc_mp);
    free_matrix(temp_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int plot_3D_hull_list_projection
(
    int         plot_id,
    int         num_hulls,
    Hull**      hp_list,
    const char* title 
)
{
    Invalid_chromaticity_action invalid_chromaticity_action;
    Hull*                       hp;
    Hull*                       temp_hp = NULL;
    int                         i;
    Matrix_vector**              facet_matrix_vector_list;
    int                         result   = NO_ERROR;
    Matrix*                     temp_mp  = NULL;
    Matrix*                     trunc_mp = NULL;

    NRE(facet_matrix_vector_list = create_matrix_vector_list(num_hulls));

    i = 0; 

    if (fs_projection_method_for_3d_plots == DIVIDE_BY_SUM)
    {
        invalid_chromaticity_action = SKIP_INVALID_CHROMATICITY;
    }
    else 
    {
        invalid_chromaticity_action = TRUNCATE_INVALID_CHROMATICITY;
    }
                
    while ((result != ERROR) && (i<num_hulls))
    {
        hp = hp_list[ i ];

        if (    (fs_projection_method_for_3d_plots == DIVIDE_BY_RED)
             || (fs_projection_method_for_3d_plots == DIVIDE_BY_GREEN)
             || (fs_projection_method_for_3d_plots == DIVIDE_BY_BLUE)
             || (fs_projection_method_for_3d_plots == DIVIDE_BY_SUM)
           )
        {
            result = remove_origin_from_data(&trunc_mp, hp->vertex_mp,
                                             MIN_RGB / (2.0 * MAX_RGB));
        }
        else 
        {
            result = copy_matrix(&trunc_mp, hp->vertex_mp);
        }

        if (result != ERROR) 
        {
            result = get_projection_matrix(&temp_mp, trunc_mp,
                                           fs_projection_method_for_3d_plots,
                                           invalid_chromaticity_action, 
                                           MAX_PROJECTION_COORDINATE,
                                           DBL_DUMMY_ARG, DBL_DUMMY_ARG,
                                           (Matrix**)NULL, (Matrix**)NULL);
        }

        if (result != ERROR) 
        {
            temp_hp = find_convex_hull(temp_mp, DEFAULT_HULL_OPTIONS);
            if (temp_mp == NULL) result = ERROR;
        }

        if (result != ERROR) 
        {
            result = copy_matrix_vector(&(facet_matrix_vector_list[ i ]),
                                       temp_hp->facets);
        }

        free_hull(temp_hp);

        i++; 
    }

    if (result != ERROR) 
    {
        result = plot_matrix_vector_list_cols(plot_id, num_hulls, 
                                             facet_matrix_vector_list, title);
    }

    free_matrix_vector_list(num_hulls, facet_matrix_vector_list);
    free_matrix(trunc_mp);
    free_matrix(temp_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_plot_points
(
    Matrix**             point_plot_mpp,
    char**               point_names,
    int                  canon_gamut_num,
    Result_by_algorithm* result_ptr,
    int                  dim 
)
{
    Best_fit_result*       best_fit_ptr; 
    Correlation_result*    corr_result_ptr;           /* Lindsay Feb 17/98 */
#ifdef SIMPLE_TEMPLATE
    Simple_template_result* simple_template_result_ptr;
#endif 
#ifdef TEMPLATE
    Template_result*       template_result_ptr;
#endif 
#ifdef MULTI_TEMPLATE
    Multi_template_result* multi_template_result_ptr;
#endif 
    Gamut_result*          gamut_result_ptr;
    Standard_result*       standard_result_ptr;
    int                    count;
    Matrix*                point_plot_mp;

    if (dim == 2)
    {
        gamut_result_ptr = &(result_ptr->gamut_2D[ canon_gamut_num ]);
    }
    else if (dim == 3)
    {
        gamut_result_ptr = &(result_ptr->gamut_3D[ canon_gamut_num ]);
    }
    else 
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    best_fit_ptr              = &(result_ptr->best_fit); 
    standard_result_ptr       = &(result_ptr->standard);
    corr_result_ptr           = &(result_ptr->correlation); /*Lindsay Feb17*/

#ifdef SIMPLE_TEMPLATE
    simple_template_result_ptr = &(result_ptr->simple_template);
#endif 

#ifdef TEMPLATE
    template_result_ptr       = &(result_ptr->template);
#endif 

#ifdef MULTI_TEMPLATE
    multi_template_result_ptr = &(result_ptr->multi_template);
#endif 

    ERE(get_target_matrix(point_plot_mpp, MAX_NUM_PLOT_POINTS, 2));

    point_plot_mp = *point_plot_mpp;

    count = 0;

    if (    (best_fit_ptr->world_best_2D_diag_fit_vp != NULL)
         && (fs_plot_2D_best_fit)
       )

    {
        ERE(add_plot_point(dim, point_plot_mp,
                           best_fit_ptr->world_best_2D_diag_fit_vp,
                           count));
        point_names[ count ] = kjb_strdup("World best 2D D-Map");
        count++;
    }

    if (    (best_fit_ptr->world_best_3D_diag_fit_vp != NULL) 
         && (fs_plot_3D_best_fit)
       )
    {
        ERE(add_plot_point(dim, point_plot_mp, 
                           best_fit_ptr->world_best_3D_diag_fit_vp,
                           count));
        point_names[ count ] = kjb_strdup("World best 3D D-Map");
        count++;
    }

    if ((best_fit_ptr->actual_vp != NULL) && fs_plot_actual)
    {
        ERE(add_plot_point(dim, point_plot_mp, best_fit_ptr->actual_vp,
                           count));
        point_names[ count ] = kjb_strdup("Illum Diag Map");
        count++;
    }

    if ((standard_result_ptr->grey_world_vp != NULL) && fs_plot_gw)
    {
        ERE(add_plot_point(dim, point_plot_mp,
                           standard_result_ptr->grey_world_vp,
                           count));
        point_names[ count ] = kjb_strdup("Grey World");
        count++;
    }

#ifdef DISABLED
    if ((standard_result_ptr->rms_grey_world_vp != NULL) && fs_plot_gw)
    {
        ERE(add_plot_point(dim, point_plot_mp,
                           standard_result_ptr->rms_grey_world_vp,
                           count));
        point_names[ count ] = kjb_strdup("RMS Grey World");
        count++;
    }
#endif  

    if ((standard_result_ptr->db_grey_vp != NULL) && fs_plot_gw)
    {
        ERE(add_plot_point(dim, point_plot_mp,
                           standard_result_ptr->db_grey_vp,
                           count));
        point_names[ count ] = kjb_strdup("DB Grey World");
        count++;
    }

    if ((standard_result_ptr->scale_by_max_vp != NULL) && fs_plot_scale_by_max)
    {
        ERE(add_plot_point(dim, point_plot_mp, standard_result_ptr->scale_by_max_vp,
                           count));
        point_names[ count ] = kjb_strdup("Scale by max");
        count++;
    }

    if (    (gamut_result_ptr->mv_surface_vp != NULL)
         && (     (    fs_plot_2D_max
                    && (gamut_result_ptr->mv_surface_vp->length == 2)
                  )
               ||
                  (    fs_plot_3D_max
                    && (gamut_result_ptr->mv_surface_vp->length == 3)
                  )
            )
       )
    {
        ERE(add_plot_point(dim, point_plot_mp, gamut_result_ptr->mv_surface_vp,
                           count));
        point_names[ count ] = kjb_strdup("Max Volume, S");
        count++;
    }

    if (    (gamut_result_ptr->mv_surface_illum_vp != NULL)
         && (     (    fs_plot_2D_max
                    && (gamut_result_ptr->mv_surface_illum_vp->length == 2)
                  )
               ||
                  (    fs_plot_3D_max
                    && (gamut_result_ptr->mv_surface_illum_vp->length == 3)
                  )
            )
       )
    {
        ERE(add_plot_point(dim, point_plot_mp, 
                           gamut_result_ptr->mv_surface_illum_vp,
                           count));
        point_names[ count ] = kjb_strdup("Max Volume, S and I");
        count++;
    }

    if (fs_plot_hull_average && (gamut_result_ptr->ha_surface_vp != NULL))
    {
        ERE(add_plot_point(dim, point_plot_mp, gamut_result_ptr->ha_surface_vp,
                           count));
        point_names[ count ] = kjb_strdup("Hull Ave, S");
        count++;
    }

    if (fs_plot_hull_average && (gamut_result_ptr->ha_surface_illum_vp != NULL))
    {
        ERE(add_plot_point(dim, point_plot_mp, 
                           gamut_result_ptr->ha_surface_illum_vp,
                           count));
        point_names[ count ] = kjb_strdup("Hull Ave, S and I");
        count++;
    }

    if (fs_plot_hull_average && (gamut_result_ptr->icsa_surface_illum_vp != NULL))
    {
        ERE(add_plot_point(dim, point_plot_mp, 
                           gamut_result_ptr->icsa_surface_illum_vp,
                           count));
        point_names[ count ] = kjb_strdup("ICS Hull Ave");
        count++;
    }

    if (fs_plot_luminance)
    {
        int  i;
        char title[ 100 ];

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->scla_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), "SC Lum:%d Ave", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->scla_surface_illum_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->lcsa_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), "I+Lum:%d S Ave", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->lcsa_surface_illum_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }
    }

    if (fs_plot_hull_average)
    {    
        int  i;
        char title[ 100 ];

        for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
        {
            if (gamut_result_ptr->scia_surface_illum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "SC Illum Ave [ %d ]", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                                   gamut_result_ptr->scia_surface_illum_vp[ i ],
                                   count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
        {
            if (gamut_result_ptr->scca_surface_illum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "SC Chrom Ave [ %d ]", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                                   gamut_result_ptr->scca_surface_illum_vp[ i ],
                                   count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }
    }

    if (fs_plot_weighted_hull_average && (gamut_result_ptr->wa_surface_vp != NULL))
    {
        ERE(add_plot_point(dim, point_plot_mp, gamut_result_ptr->wa_surface_vp,
                           count));
        point_names[ count ] = kjb_strdup("W Ave, S");
        count++;
    }

    if (    fs_plot_weighted_hull_average
         && (gamut_result_ptr->wa_surface_illum_vp != NULL)
       )
    {
        ERE(add_plot_point(dim, point_plot_mp, 
                           gamut_result_ptr->wa_surface_illum_vp,
                           count));
        point_names[ count ] = kjb_strdup("Weighted Ave, S and I");
        count++;
    }

    if (fs_plot_luminance)
    {
        int  i;
        char title[ 100 ];

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->ha_surface_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), "Luminance:%d, S, ave",
                                i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->ha_surface_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->mv_surface_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), "Luminance:%d, S, mv", 
                                i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->mv_surface_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->ha_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "Luminance:%d, S and I, mv", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->ha_surface_illum_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->mv_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "Luminance:%d, S and I, mv", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                               gamut_result_ptr->mv_surface_illum_lum_vp[ i ],
                               count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->scia_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "SIL SCIA [ %d ]", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                                   gamut_result_ptr->scia_surface_illum_lum_vp[ i ],
                                   count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }

        for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
        {
            if (gamut_result_ptr->scca_surface_illum_lum_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(title, sizeof(title), 
                                "SIL SCCA [ %d ]", i + 1));

                ERE(add_plot_point(dim, point_plot_mp, 
                                   gamut_result_ptr->scca_surface_illum_lum_vp[ i ],
                                   count));
                point_names[ count ] = kjb_strdup(title);
                count++;
            }
        }
    }

    /* Lindsay - Sept 30/98 */
    if (fs_plot_correlation) 
    {
        int i;            /* Modified April 14/99 */
        char buff[ 100 ];
        
        for (i=0; i<NUM_CORMATS_PER_ALG; i++)
        {    
            if (corr_result_ptr->correlation_01_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (0/1) [ %d ]", i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_01_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Inv. Distance, MAP) [ %d ]",
                                i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Inv. Distance, MLM) [ %d ]",
                                i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Inv. Distance, MMSE) [ %d ]",
                                i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Gaussian Fill, MAP) [ %d ]",
                                i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Gaussian Fill, MLM) [ %d ]",
                                i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ]
                != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Uniform, Gaussian Fill, MMSE) [ %d ]",
                                i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ],
                                   count));

                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MAP_fitted_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Fitted, MAP) [ %d ]", i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MAP_fitted_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MLM_fitted_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Fitted, MLM) [ %d ]", i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MLM_fitted_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_uniform_MMSE_fitted_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Fitted, MMSE) [ %d ]", i+1));
            
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_uniform_MMSE_fitted_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_gaussian_MAP_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Gaussian, MAP) [ %d ]", i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_gaussian_MAP_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_gaussian_MLM_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Gaussian, MLM) [ %d ]", i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_gaussian_MLM_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }

            if (corr_result_ptr->correlation_gaussian_MMSE_vp[ i ] != NULL)
            {
                ERE(kjb_sprintf(buff, sizeof(buff),
                                "C-by-C (Gaussian, MMSE) [ %d ]", i+1));
                
                ERE(add_plot_point(dim, point_plot_mp, 
                                   corr_result_ptr->correlation_gaussian_MMSE_vp[ i ],
                                   count));
                
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }
        }
    }
    /* End Lindsay */

#ifdef SIMPLE_TEMPLATE
    if (fs_plot_simple_template) 
    {
        if (simple_template_result_ptr->alg_vp != NULL)
        {
            ERE(add_plot_point(dim, point_plot_mp, 
                               simple_template_result_ptr->alg_vp,
                               count));
            point_names[ count ] = kjb_strdup("Simple template alg");
            count++;
        }
    }
#endif 

#ifdef TEMPLATE
    if (fs_plot_template) 
    {
        if (template_result_ptr->first_alg_vp != NULL)
        {
            ERE(add_plot_point(dim, point_plot_mp, 
                               template_result_ptr->first_alg_vp,
                               count));
            point_names[ count ] = kjb_strdup("Template alg 1");
            count++;
        }

        if (template_result_ptr->second_alg_vp != NULL)
        {
            ERE(add_plot_point(dim, point_plot_mp, 
                               template_result_ptr->second_alg_vp,
                               count));
            point_names[ count ] = kjb_strdup("Template alg 2");
            count++;
        }
    }
#endif 

#ifdef MULTI_TEMPLATE
    if (fs_plot_multi_template) 
    {
        int i; 

        for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
        {
            if (multi_template_result_ptr->alg_vp[ i ] != NULL)
            {
                char buff[ 100 ];

                ERE(kjb_sprintf(buff, sizeof(buff), 
                                "Multi template alg [ %d ]", i + 1));
                ERE(add_plot_point(dim, point_plot_mp, 
                                   multi_template_result_ptr->alg_vp[ i ],
                                   count));
                point_names[ count ] = kjb_strdup(buff);
                count++;
            }
        }
    }
#endif 

    point_plot_mp->num_rows = count;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int add_plot_point(int dim, Matrix* mp, Vector* vp, int i)
{
    Vector* projected_vp = NULL;
    Vector* back_projected_vp = NULL;
    Vector* derived_vp; 
    int     result = NO_ERROR;

    if (    (mp->num_cols != 2)
         || ((dim != 2) && (dim != 3))
         || ((vp->length != 2) && (vp->length != 3))
       )
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (dim == 2) 
    {
        if (vp->length == 2)
        {
            result = put_matrix_row(mp, vp, i);
        }
        else if (vp->length == 3)
        {
            result = get_projection_vector(&projected_vp, vp, DIVIDE_BY_BLUE,
                                           MAX_PROJECTION_COORDINATE);

            if (result != ERROR) 
                result = put_matrix_row(mp, projected_vp, i);
        }
        else 
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR; 
        }
    }
    else if (dim == 3) 
    {
        if (vp->length == 2)
        {
            result = back_project_vector(&back_projected_vp, vp, 
                                         DIVIDE_BY_BLUE);
            derived_vp = back_projected_vp;
        }
        else 
        {
            derived_vp = vp;
        }

        if (result != ERROR) 
        {
            result = get_projection_vector(&projected_vp, derived_vp, 
                                           fs_projection_method_for_3d_plots,
                                           MAX_PROJECTION_COORDINATE);
        }

        if (result != ERROR) 
             result = put_matrix_row(mp, projected_vp, i);
    }   
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    free_vector(projected_vp);
    free_vector(back_projected_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* 
// Kobus : Moved this from gamut_2D.c and made it into a sub-routine so it can
// be used by 3D also. 
*/

/***
****  BROCK - begin blackbody additions
***/

int plot_projected_black_body_curve
(
    int     plot_id,
    Vector* canon_illum_vp,
    int     dim 
)
{
    double mintemp = 2000.0;     /* Starting colour temperature */
    double maxtemp=20000.0;      /* Ending colour temperature (inclusive) */
    double steptemp = 50.0;      /* colour temperature interval */
    double temp;                 /* Color temperature index */
    int plot_length;             /* number of colour temperatures to plot */
    Vector* locus_x_vp;          /* xvalues to plot */
    Vector* locus_y_vp;          /* yvalues to plot */
    char locus_label[1024];      /* Label for legend. */

    double *spectrum = NULL;     /* Hold results from black_body_spectrum */
    int out_nelm;                /* # of elements in spectrum */

    Vector *spectrum_vp;         /* Blackbody spectrum  */
    int spectrum_length;         /* Length of spectrum vector */
    double minwave = 380.0;      /* Starting spectrum wavelength, nm */
    double maxwave = 780.0;      /* Ending wavelength, nm (inclusive) */
    double stepwave = 4.0;       /* Wavelength interval, nm */

    double r_b_canon;   /* Numerator for chromaticity ratios. */
    double g_b_canon;   /* Numerator for chromaticity ratios. */
    double   R1, G1, B1;  
    int i, current;

    if (   ( ! fs_plot_black_body )    
        || ((dim == 3) && (fs_projection_method_for_3d_plots != DIVIDE_BY_BLUE))
       )
    {
        return NO_ERROR; 
    }

    /*** Calculate vector lengths. ***/

    plot_length = floor( ((maxtemp - mintemp) / steptemp) + 1.0 );
    spectrum_length = floor( ((maxwave - minwave) / stepwave) + 1.0 );

    /*** Allocate vectors for later use. ***/

    NRE(spectrum_vp = create_vector(spectrum_length));
    NRE(locus_x_vp  = create_vector(plot_length));
    NRE(locus_y_vp  = create_vector(plot_length));

    /*** 
    **** calculate chromaticity ratio numerators 
    ****  (constants dependent on canonical illuminant)
    ***/

    r_b_canon = canon_illum_vp->elements[0] / canon_illum_vp->elements[ 2 ];
    g_b_canon = canon_illum_vp->elements[1] / canon_illum_vp->elements[ 2 ];

    /*** 
    **** Loop over colour temperatures 
    ***/

    for (temp=mintemp, current=0; temp<=maxtemp; temp+=steptemp, current++)
    {

       /*** Get black-body spectrum ***/

       if (black_body_spectrum( temp, minwave, maxwave, stepwave,
                                &spectrum, &out_nelm, 1 ) == NULL)
       {
           set_error( "do_2D_illum_maps: black_body_spectrum failed.\n" );
           return( ERROR );
       }

       /*** Convert returned array to kobus-flavour vector ***/

       spectrum_vp->length = out_nelm;
       for (i=0; i<out_nelm; i++)
          spectrum_vp->elements[i] = spectrum[i];

       /*** 
       **** Integrate spectrum with camera sensors to get R,G,B.
       ***/
    
       if (get_RGB_from_spectrum( spectrum_vp, minwave, stepwave, 
                                   &R1, &G1, &B1)  == ERROR)
       {
           set_error( "do_2D_illum_maps: get_RGB_from_spectrum failed.\n" );
           return( ERROR );
       }

        /*
        // Kobus : Add in sharpening. 
        */
        ERE(sharpen_rgb(&R1, &G1, &B1, R1, G1, B1));
        /* End Kobus. */

        /*** Convert RGB to flavour of chromaticity space used in plot ***/
       
        locus_x_vp->elements[current] = r_b_canon / (R1 / B1) ;
        locus_y_vp->elements[current] = g_b_canon / (G1 / B1) ;
    }

    if (current != plot_length)
    {
        p_stderr("do_2D_illum_maps: WARNING - \n final plotlength: ");
        p_stderr("%d, calculated: %d\n", current, plot_length );
    }

    /*** Plot the black-body curve ***/

    /* Kobus : Changed to kjb_sprintf */
    ERE(kjb_sprintf(locus_label, sizeof(locus_label), 
                    "Blackbody curve: %5.0d -> %5.0d K\n",
                     (int)mintemp, (int)maxtemp));

    EPE(plot_curve(plot_id, locus_x_vp, locus_y_vp, locus_label));
 
    /*** Free allocated vectors ***/

    kjb_free(spectrum);  /* Kobus : Changed free to kjb_free. */

    free_vector(spectrum_vp);
    free_vector(locus_x_vp);
    free_vector(locus_y_vp);

    return NO_ERROR; 
}

/***
****  BROCK - end blackbody additions
***/

