
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#ifdef TEMPLATE /* Controlled in def.h */

#include "illum_data.h"
#include "spectra.h"
#include "template.h"

/* -------------------------------------------------------------------------- */

/*
// Symbols which able/disable various bits of sample code.
*/

/* 
#define CONFIG_DATA_ACCESS_METHOD_A
*/
#define CONFIG_DATA_ACCESS_METHOD_B 

/* -------------------------------------------------------------------------- */

#define TEMPLATE_DATA_FILE    "template.rgb"      /* Default value */

/* -------------------------------------------------------------------------- */

/*
// These two or three vars are for a single configuration data item. For some
// purposes, only the first two will be used. 
*/
static Matrix* fs_cached_template_data_mp                         = NULL;
static char    fs_template_data_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int     fs_template_data_version                           = 0;

/*
// Thise vars are for more standard (i.e. non-configuration) options. 
*/
static int  fs_template_boolean_option = TRUE;  /* Default value */
static int  fs_template_integer_option = 8;     /* Default value */
static double fs_template_real_option    = 2.5;   /* Default value */

/* -------------------------------------------------------------------------- */

static int make_template_data(const char* file_name);
static int write_template_data(const char* file_name);
static int read_template_data_from_file(const char* file_name);

static int set_template_data_matrix
(
    const Matrix* template_data_mp,
    const char*   config_file 
);

#ifdef CONFIG_DATA_ACCESS_METHOD_A 
    static int get_template_data(Matrix** target_mpp);
#endif 

#ifdef CONFIG_DATA_ACCESS_METHOD_B
    static int initialize_template_data(void);
#endif 

static int do_template_algorithm_one
(
    const Matrix* data_mp,
    Vector**      result_vpp 
);

static int do_template_algorithm_two
(
    const Matrix* data_mp,
    Vector**      result_vpp 
);

static void invalidate_template_data(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif 

/* -------------------------------------------------------------------------- */

/*
// This routine must return one of NO_ERROR, ERROR, and NOT_FOUND, as
// appropriate. 
*/
int set_template_options(const char* option, const char* value)
{
    int  temp_boolean_value;
    int  temp_int;
    double temp_real;
    char lc_option[ 100 ];
    int  result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "template-rgb-file"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("template-rgb-file = %s\n",
                    fs_template_data_config_file[ 0 ] ? fs_template_data_config_file
                                                   : "off")); 
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_template_data_config_file[ 0 ]  == '\0') 
            {
                ERE(pso("No template rgb been set.\n"));
            }
            else 
            {
                ERE(pso("Template rgb source is %s\n", 
                        fs_template_data_config_file)); 
            }
        }
        else
        {
            ERE(read_template_data_from_file(value)); 
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "template-boolean"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            /* Here we echo the long form the option exactly as it is typed. */
            ERE(pso("template-boolean = %s\n", 
                    fs_template_boolean_option ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            /* Here we echo the option in english. */
            ERE(pso("Template boolean option %s set.\n", 
                    fs_template_boolean_option ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_template_boolean_option = temp_boolean_value;
        }
        result = NO_ERROR; 
    }
    
    if (    (lc_option[ 0 ] == '\0') 
         || (match_pattern(lc_option, "template-integer"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            /* Here we echo the long form the option exactly as it is typed. */
            ERE(pso("template-integer = %d\n", fs_template_integer_option));
        }
        else if (value[ 0 ] == '\0')
        {
            /* Here we echo the option in english. */
            ERE(pso("Template integer option is %d.\n", 
                    fs_template_integer_option));
        }
        else
        {
            ERE(ss1pi(value, &temp_int)); 

            if (temp_int != fs_template_integer_option)
            {
                fs_template_integer_option = temp_int;
                /*
                // This is an example to show one way of dealing with the
                // case that a change in the integer option actually makes the
                // template data stale--which would be the case if the make used the
                // value of the option. Another way would be to have the make
                // routine cache the value of the integer option at the time of the
                // build.
                */
                invalidate_template_data();
            }
        }
        result = NO_ERROR; 
    }
    
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "template-double"))
        || (match_pattern(lc_option, "abbreviation-for-template-double"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            /* Here we echo the long form the option exactly as it is typed. */
            ERE(pso("template-double = %5.2f\n", fs_template_real_option));
        }
        else if (value[ 0 ] == '\0')
        {
            /* Here we echo the option in english. */
            ERE(pso("Template double option is %5.2f.\n", fs_template_real_option));
        }
        else
        {
            ERE(ss1d(value, &temp_real)); 
            fs_template_real_option = temp_real;
        }
        result = NO_ERROR; 
    }

    return result; 
}    

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// If your proposed algorithm group does not have any data which is (easily)
// buildable from spectra, then it is not necessary to copy this routine, and
// the static helper routine.
*/

int make_template(const char* make_directive, const char* file_name)
{
    char lc_make_directive[ 100 ];
    int  result = NOT_FOUND; 

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all"); 
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    (STRCMP_EQ(lc_make_directive, "all"))
         || (STRCMP_EQ(lc_make_directive, "template"))
       )
    {
        ERE(make_template_data(file_name));
        result = NO_ERROR; 
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// In this example, we take advantage of the caching already present to only
// build the data if it is out of data with respect to the spectra files. It is
// possible simply to build regardless (i.e. whenever this routine is called),
// but when we run large simulations this routine may be called many times, and
// so the time saved may add up to something. Since we already have a "template"
// for building only when needed, it seems to make sense to use it.
*/
static int make_template_data(const char* file_name)
{
    static int previous_template_data_version = NOT_SET;
    static int previous_world_version         = NOT_SET;
    int        result                         = NO_ERROR;

    if (    (fs_template_data_version == previous_template_data_version)
         && (world_is_up_to_date(previous_world_version))
       )
    {
        verbose_pso(10, "Template data is up to date. Skipping make.\n"); 
    }
    else 
    {
        Matrix* data_mp         = NULL;
        int     i;
        int     num_illum_spectra; 
        int     world_version;

        verbose_pso(5, "Template data is out of date. Remaking it.\n"); 

        /*
        // The code here will depend on the nature of the algorithm. 
        //
        // Many of our algorithms take a look at the world under each
        // illuminant, and work out something to be matched, indexed by that
        // illuminant. It is recomended that certain routines are used in this
        // scenario because then we use the same world builder for all code, and
        // things like specularity modeling and fluorescent modeling are all
        // free. Simularly, if we were training on a random selection of
        // illuminants, then the recomended approach would be to request a
        // upgrade for get_synthetic_RGB() which would take a special value for
        // the illuminant to indicate that a random one is needed. 
        */
        result = get_num_illum_spectra(&num_illum_spectra); 

        if (result != ERROR)
        {
            for (i=0; i<num_illum_spectra; i++)
            {
                world_version = get_synthetic_RGB(&data_mp, i);

                if (world_version == ERROR) 
                {
                    result = ERROR;
                    break; 
                }

                /*
                // Do something!
                */
            }
        }

        if (result != ERROR)
        {
            /* 
            // This call updates fs_template_data_version.  For illustrative
            // purposes I have stuck in data_mp as an argument, BUT the actual
            // argument will be something entirely different
            */
            result = set_template_data_matrix(data_mp, "<make>");  
        }

        if (result != ERROR) 
        {
            previous_world_version = world_version;
            previous_template_data_version = fs_template_data_version;
            
            verbose_pso(10, "Template data remade.\n");
        }

        free_matrix(data_mp); 
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        result = write_template_data(file_name);
    }
        
    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// If your proposed algorithm group does not have any data which is (easily)
// buildable from spectra, then it is not necessary to copy this code.
*/
static int write_template_data(const char* file_name)
{

    return write_matrix(fs_cached_template_data_mp, file_name); 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_template_data_from_file(const char* file_name)
{
    Matrix* template_data_mp = NULL;
    char    config_file[ MAX_FILE_NAME_SIZE ];
    int     result;

    ERE(read_matrix_from_gamut_config_file(&template_data_mp, 
                                           (const char*)NULL, 
                                           file_name, "template rgb",
                                           config_file, sizeof(config_file)));

    /* 
    // The code that follows must be able to deal with the case that
    // template_data_mp was set to NULL by the above routine (this is not an
    // error). In the case of the this example, set_template_data_matrix can
    // handle a template_data_mp set to NULL, so we are OK. 
    */
    result = set_template_data_matrix(template_data_mp, config_file);

    free_matrix(template_data_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_template_data_matrix
(
    const Matrix* new_template_data_mp,
    const char*   config_file 
)
{

    /*
    // It is OK if new_template_data_mp is NULL. 
    */
    ERE(copy_matrix(&fs_cached_template_data_mp, new_template_data_mp));
    BUFF_CPY(fs_template_data_config_file, config_file);

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif 

    fs_template_data_version++;

    return fs_template_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This may be non-static if we want to export the data out of here. This would
// be used if the algorithm(s) were complex enough that we put the configuration
// stuff in one file to serve several other files. 
//
// If we don't need to export the data out of here, then we may not want to use
// this routine to "get" the static data because this then requires an extra
// copy (no double big deal unless the data is big). For this, or perhaps style
// reasons, you may wish to use CONFIG_DATA_ACCESS_METHOD_B below -- note that
// the key thing here is that we have to impose certain semantics in the
// read/calculation of the configuration data. (There are three ways the data
// can come to exist: default file read, user specified file read, and make). 
//
// Note that this routine implements the same "default config data" semantics as
// does initialize_multi_template_data(). However, note that we EITHER have this
// routine (case CONFIG_DATA_ACCESS_METHOD_A) or the initialzation routine
// (case CONFIG_DATA_ACCESS_METHOD_B). Thus, the semantics are implemented in
// exactly one place. 
//
*/

#ifdef CONFIG_DATA_ACCESS_METHOD_A 

static int get_template_data(Matrix** target_mpp)
{
    static int first_time = TRUE;

    /*
    // Need a clear because we want to do an "insert_error()" without 
    // necessarily having a preceding set_error(); Usually a clear is not
    // needed.
    */
    kjb_clear_error();

    if (first_time && (fs_cached_template_data_mp == NULL))
    {
        first_time = FALSE; 
        ERE(read_template_data_from_file(TEMPLATE_DATA_FILE));
    }

    if (fs_cached_template_data_mp == NULL)
    {
        insert_error("No template data is available.");
        return ERROR;
    }

    ERE(copy_matrix(target_mpp, fs_cached_template_data_mp));

    return NO_ERROR; 
}

#endif  /* Method  A  */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine implements the "default config data" semantics in the case of
// CONFIG_DATA_ACCESS_METHOD_B. The same semantics are implements for case
// CONFIG_DATA_ACCESS_METHOD_A in the get_multi_template_data() routine. Note
// that we only have one or the other of these routines. 
*/

#ifdef CONFIG_DATA_ACCESS_METHOD_B

static int initialize_template_data(void)
{
    static int     first_time = TRUE; 

    /*
    // Need a clear because we want to do an "insert_error()" without 
    // necessarily having a preceding set_error(); Usually a clear is not
    // needed.
    */
    kjb_clear_error();

    if (first_time && (fs_cached_template_data_mp == NULL))
    {
        first_time = FALSE; 
        ERE(read_template_data_from_file(TEMPLATE_DATA_FILE));
    }

    if (fs_cached_template_data_mp == NULL)
    {
        insert_error("No template data is available.");
        return ERROR;
    }

    return NO_ERROR; 
}

#endif  /* Method B */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_template
(
    const Matrix*    data_mp,
    Template_result* template_result_ptr,
    Template_error*  cpu_time_ptr 
)
{
    long    cpu_time_in_ms;
    Vector* illum_rgb_vp   = NULL;
    Vector* rgb_map_vp     = NULL;
    int     result;

    /* ----------------------  Do first template alg  ---------------- */

    init_cpu_time();

    result = do_template_algorithm_one(data_mp, &illum_rgb_vp); 

    if ((result != ERROR) && (result != NOT_AVAILABLE)) 
    {
        cpu_time_in_ms = get_cpu_time(); 

        cpu_time_ptr->first_alg = ((double)cpu_time_in_ms) / 
                                             ((double)MILLI_SECONDS_PER_SECOND) ;

        result = get_map_to_canon_illum(&rgb_map_vp, illum_rgb_vp);
    }

    if ((result != ERROR) && (result != NOT_AVAILABLE))
    {
        result = get_projection_vector(&(template_result_ptr->first_alg_vp), 
                                       rgb_map_vp, DIVIDE_BY_BLUE, MAX_X_DIV_B);
    }

    if (result == ERROR) 
    {
        free_vector(rgb_map_vp);
        free_vector(illum_rgb_vp);
        return ERROR;
    }

    /* -------------------  Do second template alg  ------------------ */

    init_cpu_time();

    result = do_template_algorithm_two(data_mp, &illum_rgb_vp); 

    if ((result != ERROR) && (result != NOT_AVAILABLE)) 
    {
        cpu_time_in_ms = get_cpu_time(); 

        cpu_time_ptr->second_alg = ((double)cpu_time_in_ms) / 
                                             ((double)MILLI_SECONDS_PER_SECOND) ;

        result = get_map_to_canon_illum(&rgb_map_vp, illum_rgb_vp);
    }

    if ((result != ERROR) && (result != NOT_AVAILABLE))
    {
        result = get_projection_vector(&(template_result_ptr->second_alg_vp), 
                                       rgb_map_vp, DIVIDE_BY_BLUE, 
                                       MAX_X_DIV_B);
    }

    if (result == NOT_AVAILABLE) result = NO_ERROR; 

    /* ------------------------------------------------------------------ */

    free_vector(rgb_map_vp);
    free_vector(illum_rgb_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine should return one of NO_ERROR, ERROR, and NOT_AVAILABLE. It is
// somewhat arbitrary where one draws the line between NOT_AVAILABLE and ERROR. 
*/
static int do_template_algorithm_one
(
    const Matrix* data_mp,
    Vector**      result_vpp 
)
{

    /*
    // Add your first algorithm here.  
    */
#ifdef CONFIG_DATA_ACCESS_METHOD_A
    Matrix* local_template_data = NULL;

    ERE(get_template_data(&local_template_data));

    /* Use local_template_data. To be continued .....  */

    free_matrix(local_template_data);
#endif 

#ifdef CONFIG_DATA_ACCESS_METHOD_B

    ERE(initialize_template_data());

    ASSERT(fs_cached_template_data_mp != NULL); 

    /* Use fs_cached_template_data_mp. To be continued .....  */

    /* DON'T free fs_cached_template_data_mp */
#endif 

    /* 
    // Invent a grey world type algorithm for testing. We can't just call the
    // double grey world, as this code is tied together into a module. 
    */
    ERE(average_matrix_rows(result_vpp, data_mp)); 

    /*
    // Make it a little bit unlike the grey world, again for testing purposes
    // (Make it greener by 10%).
    */
    ((*result_vpp)->elements)[ 1 ] *= 1.1;

    return NO_ERROR; 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine should return one of NO_ERROR, ERROR, and NOT_AVAILABLE. It is
// somewhat arbitrary where one draws the line between NOT_AVAILABLE and ERROR. 
*/

static int do_template_algorithm_two
(
    const Matrix* data_mp,
    Vector**      result_vpp 
)
{

    /*
    // Add your second algorithm here, if you have one. Otherwise kill this
    // code. If you need a third, then add it! 
    */
#ifdef CONFIG_DATA_ACCESS_METHOD_A
    Matrix* local_template_data = NULL;

    ERE(get_template_data(&local_template_data));

    /* Use local_template_data. To be continued .....  */

    free_matrix(local_template_data);
#endif 

#ifdef CONFIG_DATA_ACCESS_METHOD_B

    ASSERT(fs_cached_template_data_mp != NULL); 

    ERE(initialize_template_data());

    /* Use fs_cached_template_data_mp. To be continued .....  */

    /* DON'T free fs_cached_template_data_mp */
#endif 

    /* 
    // Invent a grey world type algorithm for testing. We can't just call the
    // double grey world, as this code is tied together into a module. 
    */
    ERE(average_matrix_rows(result_vpp, data_mp)); 

    /*
    // Make it a little bit unlike the grey world, again for testing purposes
    // (Make less blue by 20%).
    */
    ((*result_vpp)->elements)[ 2 ] /= 1.2;

    return NO_ERROR; 

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void invalidate_template_data(void)
{

    free_matrix(fs_cached_template_data_mp);
    fs_cached_template_data_mp = NULL;
    fs_template_data_version++;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE; 

    if (first_time) 
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{

    free_matrix(fs_cached_template_data_mp);
}
    
#endif 

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#endif  /* TEMPLATE */

