
/* =========================================================================== *
|  
|  Copyright (c) 1994-2002 by Kobus Barnard (author). 
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen_f.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "compute_output.h"
#include "image_data.h"
#include "data.h"
#include "sharp.h"
#include "write.h"

/* -------------------------------------------------------------------------- */

static int write_output_image
(
    char*                    arg,
    char*                    output_file,
    Colour_constancy_result* result_ptr 
);

static int write_derived_data(char* output_file, int append_flag);

static int write_registered_data(char* output_file, int append_flag);

static int write_canon_data(char* output_file, int append_flag);

static int write_maps
(
    char*                    dummy_output_file,
    Colour_constancy_result* result_ptr 
);

/* -------------------------------------------------------------------------- */

int process_write_command(char* arg, Colour_constancy_result* result_ptr)
{ 
    char command[ 200 ];
    char lc_command[ 200 ];
    char request[ 200 ];
    char output_file[ MAX_FILE_NAME_SIZE ];
    int  append_flag = FALSE;

    output_file [ 0 ] = '\0';

    /* Null result can be meaninfull. */
    BUFF_GEN_GET_TOKEN(&arg, request, ">"); 

    if (*arg == '>')
    {
        arg++;

        if (*arg == '>')
        {       
            append_flag = TRUE; 
            arg++;
        }

        if ( ! BUFF_GET_TOKEN_OK(&arg, output_file)) 
        {
            set_error("A filename is required after the \">\".");
            return ERROR;
        }
    }

    arg = request; 

    if ( ! BUFF_GET_TOKEN_OK(&arg, command))
    {
        set_error("No write directive.");
        return ERROR;
    }

    EXTENDED_LC_BUFF_CPY(lc_command, command);

    trim_beg(&arg); 
    trim_end(arg); 

    if (STRCMP_EQ(lc_command, "data"))
    {
        set_error("Request \"data\" is obsolete.");
        add_error("    Use either \"derived-data\" or \"registered-data\".");
        add_error("    \"derived-data\" includes aperture, ");
        cat_error("clipping, and clustering.");
        add_error("    \"registered-data\" is registered with canon data.");
        return ERROR; 
    }
    else if (STRCMP_EQ(lc_command, "derived-data"))
    {
        if (arg[ 0 ] != '\0')
        {
            pso("%q is ignored.\n", arg);
        }
        return write_derived_data(output_file, append_flag);
    }
    else if (    (STRCMP_EQ(lc_command, "reg-data"))
              || (STRCMP_EQ(lc_command, "registered-data"))
            )
    {
        if (arg[ 0 ] != '\0')
        {
            pso("%q is ignored.\n", arg);
        }
        return write_registered_data(output_file, append_flag);
    }
    else if (STRCMP_EQ(lc_command, "canon-data"))
    {
        if (arg[ 0 ] != '\0')
        {
            pso("%q is ignored.\n", arg);
        }
        return write_canon_data(output_file, append_flag);
    }
    else if (STRCMP_EQ(lc_command, "maps"))
    {
        if (arg[ 0 ] != '\0')
        {
            pso("%q is ignored.\n", arg);
        }
        return write_maps(output_file, result_ptr);
    }
    else if (STRCMP_EQ(lc_command, "image"))
    {
        return write_output_image(arg, output_file, result_ptr);
    }
    else if (    (STRCMP_EQ(lc_command, "sharp"))
              || (STRCMP_EQ(lc_command, "sharpening"))
              || (STRCMP_EQ(lc_command, "sharpening-transforms"))
            )
    {
        return write_sharpening_transform(arg, output_file, append_flag);
    }
    else if (    (STRCMP_EQ(lc_command, "sharp-sensors"))
              || (STRCMP_EQ(lc_command, "sharpened-sensors"))
            )
    {
        dbs(lc_command); 
        dbs(arg); 
        return write_sharp_sensors(arg, output_file, append_flag);
    }
    else 
    {
        /*
        // Default behaviour : The "image" is implicit. 
        */
        return write_output_image(command, output_file, result_ptr);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_output_image
(
    char*                    arg,
    char*                    output_file,
    Colour_constancy_result* result_ptr 
)
{ 
    Image_map_data* image_map_data_ptr;
    char            request[ 200 ];
    KJB_image*    ip      = NULL;
    int             result;

    if ( ! BUFF_GET_TOKEN_OK(&arg, request))
    {
        set_error("No write image request specified.");
        return ERROR;
    }

    if (*output_file == '\0')
    {
        set_error("No file to write result image to is specified.");
        add_error("A file name is required.");
        return ERROR;
    }

    trim_beg(&arg);

    if (*arg != '\0')
    {
        pso("Extra input %q is ingnored.", arg);
    }

    if (IC_STRCMP_EQ(request, "pp"))
    {
        const KJB_image* pp_ip; 

        NRE(pp_ip = get_pp_input_image());
        return kjb_write_image(pp_ip, output_file);
    }

    NRE(image_map_data_ptr = create_image_map_data());

    result = parse_image_request(request, result_ptr, 
                                 FALSE,   /* reduction_flag */ 
                                 FALSE,   /* illumination_flag */ 
                                 FALSE,   /* chromaticity image */
                                 image_map_data_ptr,
                                 (Error_by_sharp*)NULL,
                                 (double**)NULL);
                                 
    if (result != ERROR)
    {
        result = compute_result_image(&ip, image_map_data_ptr);
    }

    if (result != ERROR)
    {
        result = kjb_write_image(ip, output_file);
    }

    if (result == ERROR)
    {
        insert_error("Unable to process write request %q.", request);
    }

    kjb_free_image(ip);
    free_image_map_data(image_map_data_ptr);

    return result;
}
        
/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_derived_data(char* output_file, int append_flag)
{
    Matrix* data_mp = NULL;
    int     result;

    result = get_selected_derived_data(&data_mp, 
                                  0, (int*)NULL); 

    if (result == ERROR) return ERROR; 

    if (data_mp == NULL) 
    {
        set_error("No derived data is available."); 
        return ERROR;
    }

    result = write_data_matrix(data_mp, output_file, append_flag); 

    free_matrix(data_mp);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_registered_data(char* output_file, int append_flag)
{
    Matrix* registered_data_mp = NULL;
    int     result;

    ERE(get_registered_data(&registered_data_mp)); 

    if (registered_data_mp == NULL) 
    {
        set_error("No registered data is available."); 
        return ERROR;
    }

    result = write_data_matrix(registered_data_mp, output_file, append_flag); 

    free_matrix(registered_data_mp);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_canon_data(char* output_file, int append_flag)
{
    Matrix* canon_data_mp = NULL;
    int     result;

    ERE(get_canon_data(&canon_data_mp)); 

    if (canon_data_mp == NULL) 
    {
        set_error("No canonical data is available."); 
        return ERROR;
    }

    result = write_data_matrix(canon_data_mp, output_file, append_flag); 

    free_matrix(canon_data_mp);

    return result; 
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int write_maps
(
    char*                    dummy_output_file,
    Colour_constancy_result* result_ptr 
)
{

    db_mat(result_ptr->sharp[ DONT_SHARPEN ].best_fit.world_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ DONT_SHARPEN ].best_fit.world_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].best_fit.world_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].best_fit.world_best_2D_diag_fit_vp);
    db_mat(result_ptr->sharp[ DONT_SHARPEN ].best_fit.scene_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ DONT_SHARPEN ].best_fit.scene_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].best_fit.scene_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].best_fit.scene_best_2D_diag_fit_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].best_fit.actual_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].standard.grey_world_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].standard.db_grey_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].standard.scale_by_max_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].mv_surface_illum_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].mv_surface_illum_vp);

    db_mat(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.world_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.world_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.world_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.world_best_2D_diag_fit_vp);
    db_mat(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.scene_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.scene_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.scene_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.scene_best_2D_diag_fit_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].best_fit.actual_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].standard.grey_world_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].standard.db_grey_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].standard.scale_by_max_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_2D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_2D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_2D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_2D[ 0 ].mv_surface_illum_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_3D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_3D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_3D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ SYNTHETIC_DB_SHARP ].gamut_3D[ 0 ].mv_surface_illum_vp);

    db_mat(result_ptr->sharp[ FIXED_SHARP ].best_fit.world_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ FIXED_SHARP ].best_fit.world_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].best_fit.world_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].best_fit.world_best_2D_diag_fit_vp);
    db_mat(result_ptr->sharp[ FIXED_SHARP ].best_fit.scene_best_2D_fit_mp);
    db_mat(result_ptr->sharp[ FIXED_SHARP ].best_fit.scene_best_3D_fit_mp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].best_fit.scene_best_3D_diag_fit_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].best_fit.scene_best_2D_diag_fit_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].best_fit.actual_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].standard.grey_world_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].standard.db_grey_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].standard.scale_by_max_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_2D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_2D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_2D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_2D[ 0 ].mv_surface_illum_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_3D[ 0 ].ha_surface_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_3D[ 0 ].ha_surface_illum_vp);

    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_3D[ 0 ].mv_surface_vp);
    db_rv(result_ptr->sharp[ FIXED_SHARP ].gamut_3D[ 0 ].mv_surface_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

