
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "compute_error.h"
#include "correlation.h"

#include "data.h"
#include "do_best_fit.h"
#include "do_cc.h"
#include "do_error_report.h"
#include "do_other_cc.h"

#include "gamut_2D.h"
#include "gamut_3D.h"
#include "illum_data.h"
#include "initialize_error.h"
#include "luminance.h"

#include "output_error.h"
#include "spectra.h"
#include "stat.h"

#ifdef SIMPLE_TEMPLATE
#include "simple_template.h"
#endif

#ifdef TEMPLATE
#include "template.h"
#endif

#ifdef MULTI_TEMPLATE
#include "multi_template.h"
#endif

/* ------------------------------------------------------------------------- */

static int do_cc_algorithms
(
    Colour_constancy_methods*      cc_methods_ptr,
    Sharp_method                   sharp_method,
    int                            prompt_to_increase_error_flag,
    int                            num_data_points,
    int*                           data_point_nums,
    Matrix*                        data_mp,
    Result_by_algorithm*           cc_result_ptr,
    Error_by_algorithm*            cpu_type_ptr,
    Input_data_error_by_algorithm* input_data_error_ptr
);

/* -------------------------------------------------------------------------- */

int do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    int                       num_data_points,
    int*                      data_point_nums,
    Colour_constancy_result*  cc_result_ptr
)
{
    int                    status           = NO_ERROR;
    Matrix*                selected_data_mp = NULL;
    Colour_constancy_error cc_error;
    Colour_constancy_error dummy_cc_error_two;
    Input_data_error       input_data_error;
    Input_data_error       dummy_input_data_error_two;
    int                    sharp_method;
    int                    need_dull_result;
    Vector*                illum_est_for_sharp_vp = NULL;

    if (cc_methods_ptr->reset_specular)
    {
        invalidate_reflectance_with_spec();
    }

    if (num_data_points <= 40)
    {
        int i;

        for (i=0; i<num_data_points; i++)
        {
            if (i % 10 == 0)
            {
                verbose_pso(4, "    ");
            }
            if (((i + 1) == num_data_points) || ((i + 1) % 10 == 0))
            {
                verbose_pso(4, "%d\n", data_point_nums[ i ]);
            }
            else
            {
                verbose_pso(4, "%d, ", data_point_nums[ i ]);
            }
        }
    }

    initialize_error(&cc_error, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&dummy_cc_error_two, DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&input_data_error, DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&dummy_input_data_error_two,
                                DBL_HALF_MOST_NEGATIVE);

    clear_result(cc_result_ptr);

    if (    (cc_methods_ptr->sharp[ ITERATIVE_SHARP ])
         || (cc_methods_ptr->sharp[ ITERATIVE_SHARP_CORRECTION ])
       )
    {
        need_dull_result = TRUE;
    }
    else
    {
        need_dull_result = FALSE;
    }

    for (sharp_method=DONT_SHARPEN; sharp_method<FIXED_SHARP; sharp_method++)
    {
        if (cc_methods_ptr->sharp[ sharp_method ]) need_dull_result = TRUE;
    }

    if (need_dull_result)
    {
        interactive_pso("Doing colour constancy without sharpening.\n");

        status = set_sharpening(DONT_SHARPEN, 0, (int*)NULL, (Vector*)NULL);

        if (status != ERROR)
        {
            status = get_selected_derived_data(&selected_data_mp,
                                               num_data_points,
                                               data_point_nums);

        }

        if (status != ERROR)
        {
            status = do_cc_algorithms(cc_methods_ptr,
                                      DONT_SHARPEN, /* sharp_method */
                                     prompt_to_increase_error_flag,
                                     num_data_points, data_point_nums,
                                     selected_data_mp,
                                     &(cc_result_ptr->sharp[ DONT_SHARPEN ]),
                                     &(cc_error.cpu_time.sharp[ DONT_SHARPEN ]),
                                     &(input_data_error.sharp[ DONT_SHARPEN ]));
        }

        if (    (status != ERROR)
             &&
                (    (cc_methods_ptr->sharp[ ITERATIVE_SHARP ])
                  || (cc_methods_ptr->sharp[ ITERATIVE_SHARP_CORRECTION ])
                )
           )
        {
            Vector* diag_vp;

#define USE_AVE_ILLUM

#ifdef USE_AVE_ILLUM
            if (cc_result_ptr->sharp[ DONT_SHARPEN ].standard.ave_illum_vp != NULL)
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].standard.ave_illum_vp;
            }
            else
#endif
#ifdef CHEAT_FOR_TESTING
            if (cc_result_ptr->sharp[ DONT_SHARPEN ].best_fit.actual_vp != NULL)
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].best_fit.actual_vp;
            }
            else
#endif
            if (cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].ha_surface_illum_vp != NULL)
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].ha_surface_illum_vp;
            }
            else if (cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].mv_surface_illum_vp != NULL)
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ 0 ].mv_surface_illum_vp;
            }
            else if (cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].ha_surface_illum_vp != NULL)
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ 0 ].ha_surface_illum_vp;
            }
            else
            {
                diag_vp = cc_result_ptr->sharp[ DONT_SHARPEN ].standard.grey_world_vp;
            }

            status = get_illum_rgb_from_diagonal_transform(
                                                       &illum_est_for_sharp_vp,
                                                       diag_vp);

            db_rv(illum_est_for_sharp_vp);
        }

        if (status != ERROR)
        {
            for (sharp_method=DONT_SHARPEN; sharp_method<FIXED_SHARP; sharp_method++)
            {
                if ((status != ERROR) && (cc_methods_ptr->sharp[ sharp_method]))
                {
                    status = set_sharpening((Sharp_method)sharp_method, num_data_points,
                                            data_point_nums,
                                            illum_est_for_sharp_vp);

                    if (status != ERROR)
                    {
                        char sharp_method_string[ 100 ];

                        get_sharp_method_string((Sharp_method)sharp_method,
                                                sharp_method_string,
                                               sizeof(sharp_method_string));

                        interactive_pso("Calculating %s error.\n",
                                        sharp_method_string);

                        /*
                        // We are calculating the result of either not
                        // sharpening, or of sharp-correction, both which used
                        // the dull result to compute the error,
                        */
                        status = compute_error(num_data_points,
                                               data_point_nums,
                                               (Sharp_method)sharp_method,
                                        &(cc_result_ptr->sharp[ DONT_SHARPEN ]),
                                        &cc_error);

                        if ((status != ERROR) && (sharp_method == DONT_SHARPEN))
                        {
                            status = write_error_output_files(&cc_error);
                        }
                    }
                }
            }
        }
        else
        {
            add_error("Skipping error calculations due to errors.");
        }
    }

    /*
    // Do the sharp correction sharp methods.
    */
    sharp_method = FIXED_SHARP_CORRECTION;

    while ((status != ERROR) && (sharp_method < FIXED_SHARP))
    {
        if (cc_methods_ptr->sharp[ sharp_method ])
        {
            char sharp_method_string[ 100 ];

            status = set_sharpening((Sharp_method)sharp_method, num_data_points,
                                    data_point_nums, illum_est_for_sharp_vp);

            if (status != ERROR)
            {
                get_sharp_method_string((Sharp_method)sharp_method, sharp_method_string,
                                        sizeof(sharp_method_string));

                status = get_selected_derived_data(&selected_data_mp,
                                                   num_data_points,
                                                   data_point_nums);
            }

            if (status != ERROR)
            {
                interactive_pso("Calculating maps for %s sharpening.\n",
                                sharp_method_string);

                status = get_fixed_sharp_maps_by_algorithm(
                                    &(cc_result_ptr->sharp[ sharp_method ]),
                                    &(cc_result_ptr->sharp[ DONT_SHARPEN ]));
            }

            if (status != ERROR)
            {
                interactive_pso("Calculating %s error.\n", sharp_method_string);

                status = compute_error(num_data_points, data_point_nums,
                                       (Sharp_method)sharp_method,
                                       &(cc_result_ptr->sharp[ sharp_method ]),
                                       &cc_error);
            }
            else
            {
                add_error("Skipping error calculations due to errors.");
            }
        }
        sharp_method++;
    }

    while ((status != ERROR) && (sharp_method < NUM_SHARP_METHODS))
    {
        if (cc_methods_ptr->sharp[ sharp_method ])
        {
            char sharp_method_string[ 100 ];

            get_sharp_method_string((Sharp_method)sharp_method, sharp_method_string,
                                    sizeof(sharp_method_string));

            interactive_pso("Doing colour constancy with %s sharpening.\n",
                            sharp_method_string);

            status = set_sharpening((Sharp_method)sharp_method, num_data_points,
                                    data_point_nums, illum_est_for_sharp_vp);

            if (status != ERROR)
            {
                get_sharp_method_string((Sharp_method)sharp_method,
                                        sharp_method_string,
                                        sizeof(sharp_method_string));

                status = get_selected_derived_data(&selected_data_mp,
                                                   num_data_points,
                                                   data_point_nums);
            }

            if (status != ERROR)
            {
                dbs(sharp_method_string);

                status = do_cc_algorithms(cc_methods_ptr,
                                          (Sharp_method)sharp_method,
                                          prompt_to_increase_error_flag,
                                          num_data_points, data_point_nums,
                                          selected_data_mp,
                                &(cc_result_ptr->sharp[ sharp_method ]),
                                &(cc_error.cpu_time.sharp[ sharp_method ]),
                                &(input_data_error.sharp[ sharp_method ]));
            }

            if (status != ERROR)
            {
                interactive_pso("Calculating %s error.\n", sharp_method_string);

                status = compute_error(num_data_points, data_point_nums,
                                       (Sharp_method)sharp_method,
                                       &(cc_result_ptr->sharp[ sharp_method ]),
                                       &cc_error);
            }
            else
            {
                add_error("Skipping error calculations due to errors.");
            }
        }
        sharp_method++;
    }

    if ((status != ERROR) && (cc_methods_ptr->report_result))
    {
        interactive_pso("Doing error report.\n\n");

        status = do_error_report(&cc_error, &dummy_cc_error_two,
                                 &input_data_error, &dummy_input_data_error_two,
                                 (const char*)NULL, (const char*)NULL,
                                 (const char*)NULL);

        if ((status != ERROR) && (cc_methods_ptr->do_luminance))
        {
            status = do_luminance_error_report();
        }

        if ((status != ERROR) && (cc_methods_ptr->do_luminance_factors))
        {
            status = do_luminance_factor_report();
        }
    }

    if (status != ERROR)
    {
        status = update_cumulative_error(cc_methods_ptr->sharp, &cc_error);

        if (status != ERROR)
        {
            status = update_cumulative_input_data_error(cc_methods_ptr->sharp,
                                                        &input_data_error);
        }
    }
    else
    {
        add_error("Skipping update of cumulative error due to errors.");
    }

    free_matrix(selected_data_mp);
    free_vector(illum_est_for_sharp_vp);

    /*
    // Reset sharpening to no sharpening in order not to affect things like
    // routines which write out the data used.
    */
    if (set_sharpening(DONT_SHARPEN, 0, (int*)NULL, (Vector*)NULL) == ERROR)
    {
        status = ERROR;
    }

    return status;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int do_cc_algorithms
(
    Colour_constancy_methods*      cc_methods_ptr,
    Sharp_method                   sharp_method,
    int                            prompt_to_increase_error_flag,
    int                            num_data_points,
    int*                           data_point_nums,
    Matrix*                        data_mp,
    Result_by_algorithm*           cc_result_ptr,
    Error_by_algorithm*            cpu_time_ptr,
    Input_data_error_by_algorithm* input_data_error_ptr
)
{
    IMPORT volatile int io_atn_flag;
    IMPORT volatile int halt_all_output;
    IMPORT int          kjb_debug_level;
    int                 cc_status;

    verbose_pso(4, "Using %d data points.\n", data_mp->num_rows);

    if (kjb_debug_level >= 3)
    {
        int save_num_rows = data_mp->num_rows;

        data_mp->num_rows = MIN_OF(10, data_mp->num_rows);
        db_mat(data_mp);
        data_mp->num_rows = save_num_rows;
    }

    cc_status = do_best_fit(&(cc_result_ptr->best_fit), num_data_points,
                            data_point_nums);

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (cc_status == ERROR)
    {
        if (cc_methods_ptr->skip_on_error)
        {
            return ERROR;
        }
        else if (cc_methods_ptr->print_algorithm_error)
        {
            kjb_print_error();
            cc_status = NO_ERROR;
        }
    }

    if (cc_methods_ptr->do_luminance)
    {
        interactive_pso("Doing luminance algorithms.\n");

        cc_status = do_luminance(data_mp);
    }

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (cc_status == ERROR)
    {
        if (cc_methods_ptr->skip_on_error)
        {
            return ERROR;
        }
        else if (cc_methods_ptr->print_algorithm_error)
        {
            kjb_print_error();
            cc_status = NO_ERROR;
        }
    }

    if (cc_methods_ptr->do_luminance_factors)
    {
        interactive_pso("Doing luminance factors.\n");

        cc_status = do_luminance_factors(data_mp);
    }

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (cc_status == ERROR)
    {
        if (cc_methods_ptr->skip_on_error)
        {
            return ERROR;
        }
        else if (cc_methods_ptr->print_algorithm_error)
        {
            kjb_print_error();
            cc_status = NO_ERROR;
        }
    }

    /*
    // Do all other algorithms before gamut mapping so that the plots that
    // are done as part of gamut mapping can use the other results.
    */

    if (cc_methods_ptr->do_standard)
    {
        interactive_pso("Doing standard colour constancy algorithms.\n");

        cc_status = do_standard(data_mp, &(cc_result_ptr->standard));

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }

/* Lindsay: Feb 17/98 */
    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_correlation)
         && (sharp_method == DONT_SHARPEN)
       )
    {
        interactive_pso("Doing 2D colour by correlation algorithms.\n");

        cc_status = do_correlation(cc_methods_ptr, data_mp,
                                   prompt_to_increase_error_flag,
                                   &(cc_result_ptr->correlation),
                                   &(cpu_time_ptr->correlation));

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }
/* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_simple_template)
         && (sharp_method == DONT_SHARPEN)
       )
    {
        interactive_pso("Doing simple template algorithm.\n");

        cc_status = do_simple_template(data_mp,
                                      &(cc_result_ptr->simple_template),
                                      &(cpu_time_ptr->simple_template));

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }
#endif

#ifdef TEMPLATE
    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_template)
         && (sharp_method == DONT_SHARPEN)
       )
    {
        interactive_pso("Doing template algorithms.\n");

        cc_status = do_template(data_mp, &(cc_result_ptr->template),
                                &(cpu_time_ptr->template));

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }
#endif

#ifdef MULTI_TEMPLATE
    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_multi_template)
         && (sharp_method == DONT_SHARPEN)
       )
    {
        interactive_pso("Doing multi template algorithms.\n");

        cc_status = do_multi_template(data_mp,
                                      &(cc_result_ptr->multi_template),
                                      &(cpu_time_ptr->multi_template));

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }
#endif

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_gamut)
         && (cc_methods_ptr->do_2D)
         && (    (cc_methods_ptr->hull_intersection)
            )
       )
    {
        interactive_pso("Doing 2D gamut mapping algorithms.\n");

        cc_status = do_gamut_2D(cc_methods_ptr,
                                prompt_to_increase_error_flag,
                                data_point_nums, data_mp,
                                cc_result_ptr,
                                cpu_time_ptr->gamut_2D,
                                input_data_error_ptr->gamut_2D);

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    if (    (cc_methods_ptr->do_gamut)
         && (cc_methods_ptr->do_3D)
         && (    (cc_methods_ptr->hull_intersection)
            )
       )
    {
        interactive_pso("Doing 3D gamut mapping algorithms.\n");

        cc_status = do_gamut_3D(cc_methods_ptr,
                                prompt_to_increase_error_flag,
                                data_point_nums, data_mp,
                                cc_result_ptr,
                                cpu_time_ptr->gamut_3D,
                                input_data_error_ptr->gamut_3D);

        if (cc_status == ERROR)
        {
            if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
            {
                return ERROR;
            }
            else if (cc_methods_ptr->print_algorithm_error)
            {
                kjb_print_error();
                cc_status = NO_ERROR;
            }
        }
    }

    if (io_atn_flag)
    {
        set_error("Processing interrupted.");
        halt_all_output = FALSE;
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

