
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter. 
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "cc.h"
#include "display.h"
#include "generate.h"
#include "lib.h"
#include "make.h"
#include "plot.h"
#include "report.h"
#include "reset.h"
#include "set.h"
#include "stat.h"
#include "version.h"
#include "write.h"

#include "image_input.h"

/* -------------------------------------------------------------------------- */

static const char* fs_default_options[ ] =
{
    "disable-dir-open=t",
    "mrv=5", "mgv=5", "mbv=10", "msv=25",
    "gamma=on", "image-stripping=on",
    "ciof=off", "cvof=off", "linearization=on", "rip=t",
    "debug=0",
    NULL
} ;

/* -------------------------------------------------------------------------- */

int main(int argc, char* argv[])
{
    const char**            cur_option;
    char                    line[ 10000 ];
    char*                   line_pos;
    char                    word[ 200 ];
    char                    lc_word[ 200 ];
    Colour_constancy_result cc_result;
    int                     get_res;
    unsigned long           heap_memory_used;
    unsigned long           previous_heap_memory_used = 0;

    EPE(kjb_init());

    set_atn_trap(confirm_exit_sig_fn, DONT_RESTART_AFTER_SIGNAL);
    EPETE(print_version());

#ifdef HELP_FILE   /* Normally not defined, and we just go with the default. */
    set_help_file(HELP_FILE);
#endif

    /*
    // EPETE(set_monochrome_plot());
    */

    cur_option = fs_default_options;

    while (*cur_option != NULL)
    {
        if (process_set_command(*cur_option) == ERROR)
        {
            insert_error("Processing of hard coded default option %q failed.",
                         *cur_option);
            kjb_print_error();
        }
        cur_option++;
    }

    /*
    // Not ablsolutely necessary, but more efficient to do it in advance. Also,
    // if we do not do it in advance, we can get bogus memory leak reports, as
    // then the implicit fork will occur when memory has been allocated, and we
    // don't neccessarily arrange for the free by the child process.
    */
    if (is_interactive())
    {
        EPE(create_image_display());
    }

    /*
    // If there were any arguments to the program, then treat them as
    // initial input.
    */
    if (argc > 1)
    {
        char* prog_args;

        NPETE(prog_args = unparse_prog_args(argc, argv));
        EPETE(add_line_to_input_stream(prog_args));
        kjb_free(prog_args);
    }

    initialize_cc_result(&cc_result);
    initialize_cumulative_error();

    while ((get_res = get_command_text("gamut", line, sizeof(line))) != EOF)
    {
        /*
        // Paging behaviour may have been temporarily changed by the user. For
        // each command, reset to the (user settable) default.
        */
        kjb_use_default_paging();

        line_pos = line;

        if (get_res == ERROR)
        {
            kjb_print_error();
        }
        else if (get_res < 0)
        {
            /*EMPTY*/
            ; /* Do nothing. */
        }
        else if (BUFF_GET_TOKEN_OK(&line_pos, word))
        {
            EXTENDED_LC_BUFF_CPY(lc_word, word);

            if (STRCMP_EQ(lc_word, "cc"))
            {
                EPE(process_cc_command(line_pos, &cc_result));
            }
            else if (STRCMP_EQ(lc_word, "2d"))
            {
                EPE(process_cc_command(line, &cc_result));
            }
            else if (STRCMP_EQ(lc_word, "3d"))
            {
                EPE(process_cc_command(line, &cc_result));
            }
            else if (    (STRCMP_EQ(lc_word, "d"))
                      || (STRCMP_EQ(lc_word, "display"))
                    )
            {
                EPE(process_display_command(line_pos, &cc_result));
            }
            else if (STRCMP_EQ(lc_word, "plot"))
            {
                EPE(process_plot_command(line_pos));
            }
            else if (STRCMP_EQ(lc_word, "reset"))
            {
                EPE(process_reset_command(line_pos));
            }
            else if (STRCMP_EQ(lc_word, "report"))
            {
                EPE(process_report_command(line_pos));
            }
            else if (    (STRCMP_EQ(lc_word, "s"))
                      || (STRCMP_EQ(lc_word, "set"))
                    )
            {
                EPE(process_set_command(line_pos));
            }
            else if (    (STRCMP_EQ(lc_word, "w"))
                      || (STRCMP_EQ(lc_word, "write"))
                    )
            {
                EPE(process_write_command(line_pos, &cc_result));
            }
            else if (STRCMP_EQ(lc_word, "version"))
            {
                EPE(print_version());
            }
            else if (    (STRCMP_EQ(lc_word, "g"))
                      || (STRCMP_EQ(lc_word, "gen"))
                      || (STRCMP_EQ(lc_word, "generate"))
                    )
            {
                EPE(process_generate_command(line_pos));
            }
            else if (    (STRCMP_EQ(lc_word, "m"))
                      || (STRCMP_EQ(lc_word, "make"))
                    )
            {
                EPE(process_make_command(line_pos));
            }
            else if (    (STRCMP_EQ(lc_word, "heap"))
                      || (STRCMP_EQ(lc_word, "memory"))
                    )
            {
                if (memory_used(&heap_memory_used) == ERROR)
                {
                    kjb_print_error();
                }
                else
                {
                    EPE(pso("\nHeap Memory Used: %,lu bytes\n\n",
                            heap_memory_used));
                }
            }
            else
            {
                p_stderr("%q is an invalid command.\n", word);
            }
        }

        if (    ((memory_used(&heap_memory_used)) != ERROR)
             && (heap_memory_used > previous_heap_memory_used + 1000000)
           )
        {
            if (is_interactive())
            {
                interactive_pso("\n");
                interactive_pso("Heap use has increased to %,lu bytes\n\n",
                                heap_memory_used);
            }
            else
            {
                verbose_pso(2, "Heap use has increased to %,lu bytes\n",
                            heap_memory_used);
            }
            previous_heap_memory_used = heap_memory_used;
        }

        EPE(kjb_fflush(stdout));
    }

#ifdef TRACK_MEMORY_ALLOCATION
    /*
    // Try to free as much as possible to make memory leakage more apparant.
    */

    clear_result(&cc_result);
#endif

    unset_atn_trap();

    kjb_cleanup(); /* Not necessary on most systems, but doing it twice is OK.*/

    return EXIT_SUCCESS;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

