
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "do_gamut_cc.h"
#include "gamut_alg_input.h"
#include "canon_gamut.h"
#include "illum_data.h"
#include "lib.h"
#include "luminance.h"
#include "plot.h"

#include "gamut_3D.h"

/* -------------------------------------------------------------------------- */

#define MIN_USABLE_3D_ILLUM_VALUE            (0.1)
#define MIN_ALLOWABLE_3D_ILLUM_VALUE         (0.01)

#define MAX_NUM_3D_PLOTS                  5

#define PR_ILLUM_HULL_PLOT                0
#define PR_INTERSECTION_PLOT              1
#define PR_HULL_PLOT                      2
#define PR_NO_ERROR_HULL_PLOT             3
#define PR_DATA_PLOT                      4

/* -------------------------------------------------------------------------- */

static int fs_gamut_3D_inversion_resolution   = 50;
static int fs_always_have_zero_in_canon_gamut = TRUE;
static int fs_use_approximate_3D_error_hull   = FALSE;
static int fs_illum_constraint_error          = FALSE;
static char fs_illum_constraint_error_message[ 1000 ];

/* -------------------------------------------------------------------------- */

static int do_gamut_3D_variant
(
    int                       canon_gamut_num,
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr
);

static int do_3D_maps
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       plot_id_list[ MAX_NUM_3D_PLOTS ],
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    double                    absolute_pixel_error,
    double                    relative_pixel_error,
    Gamut_result*             result_ptr,
    Hull_result*              hull_result_ptr,
    Gamut_error*              cpu_time_ptr
);

static int do_3D_surface_maps
(
    Matrix*         canon_gamut_mp,
    int*            plot_id_list,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
);

static int do_3D_hull_surface_maps
(
    int             plot_id_list[ MAX_NUM_3D_PLOTS ],
    Matrix*         canon_gamut_hull_mp,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
);

static int do_3D_point_surface_maps
(
    int             plot_id_list[ MAX_NUM_3D_PLOTS ],
    Matrix*         canon_gamut_hull_mp,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
);

static int do_3D_illum_maps
(
    int             plot_id_list[ MAX_NUM_3D_PLOTS ],
    double          size,
    Queue_element** hull_list_head_ptr,
    Hull**          inverse_hull_ptr
);

static int do_illum_3D_inv_trans_gamut
(
    int             plot_id_list[ MAX_NUM_3D_PLOTS ],
    Hull*           illum_2D_map_hp,
    double          max_sum,
    Queue_element** hull_list_head_ptr
);

static int initialize_3D_plots
(
    int         canon_gamut_num,
    int         enable_plotting,
    int         plot_id_list[ MAX_NUM_3D_PLOTS]
);

/* -------------------------------------------------------------------------- */

EXTERN_C int set_3D_gamut_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "gamut-3d-inversion-resolution"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("gamut-3d-inversion-resolution = %d\n",
                    fs_gamut_3D_inversion_resolution));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Gamut 3D inversion resolution %d.\n",
                    fs_gamut_3D_inversion_resolution));
        }
        else
        {
            int temp_int;

            ERE(ss1pi(value, &temp_int));

            if (temp_int <= 0)
            {
                set_error("Gamut inversion resolution must be positive.");
                return ERROR;
            }
            fs_gamut_3D_inversion_resolution = temp_int;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "3d-canon-data-includes-zero"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("3d-canon-data-includes-zero = %s\n",
                    fs_always_have_zero_in_canon_gamut ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D canon hulls %s include zero.\n",
                    fs_always_have_zero_in_canon_gamut ? "always" : "do not necessarily"));
        }
        else
        {
            int temp_boolean;

            ERE(temp_boolean = get_boolean_value(value));
            fs_always_have_zero_in_canon_gamut = temp_boolean;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "use-approximate-3d-error-hull"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("use-approximate-3d-error-hull = %s\n",
                    fs_use_approximate_3D_error_hull ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The effect of data error on hulls %s approximated.\n",
                    fs_use_approximate_3D_error_hull ? "is" : "is not"));
        }
        else
        {
            int temp_boolean;

            ERE(temp_boolean = get_boolean_value(value));
            fs_use_approximate_3D_error_hull = temp_boolean;
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_gamut_3D
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr
)
{
    Matrix* canon_gamut_mp = NULL;
    int     i;
    int     result = NO_ERROR;
    char sharp_method_string[ 100 ];

    get_current_sharp_method_string(sharp_method_string,
                                    sizeof(sharp_method_string));

#ifdef TEST
    interactive_pso("Dong 3D gamut mapping with %s.\n",
                    sharp_method_string);
#endif

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        IMPORT volatile int io_atn_flag;

        if (result == ERROR) break;

        result = get_sharp_canon_gamut(&canon_gamut_mp, i);

        if ((result == NO_ERROR) && (canon_gamut_mp != NULL))
        {
            result = do_gamut_3D_variant(i, cc_methods_ptr,
                                         prompt_to_increase_error_flag,
                                         canon_gamut_mp,
                                         data_point_nums, data_mp,
                                         result_ptr,
                                         &(cpu_time_ptr[ i ]),
                                         &(input_data_error_ptr[ i ]));

            /*
            // HACK
            //
            // If result is ERROR, then basically no 3D algorithm worked, which
            // is pretty rare. It is also possible that the 3D algorithms which
            // rely on the illumination contraint failed, and if this is the
            // case, we record this event in the file static variable
            // fs_illum_constraint_error, and save the message in
            // fs_illum_constraint_error_message.
            */
            if (fs_illum_constraint_error)
            {
                set_error(fs_illum_constraint_error_message);
                add_error("This only affects algorithms requiring the illumination constraint.");
                add_error("Reporting was deferred until after solutions were chosen where possible.");
                result = ERROR;
            }

            if (result == ERROR)
            {
                add_error("Problem with 3D gamut mapping variant %d with %q.",
                          i + 1, sharp_method_string);

                if ((cc_methods_ptr->skip_on_error) || (io_atn_flag))
                {
                    /*EMPTY*/
                    ; /* Do nothing. */
                }
                else if (cc_methods_ptr->print_algorithm_error)
                {
                    kjb_print_error();
                    result = NO_ERROR;
                }
            }
        }
    }

    free_matrix(canon_gamut_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_gamut_3D_variant
(
    int                       canon_gamut_num,
    Colour_constancy_methods* cc_methods_ptr,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    Result_by_algorithm*      result_ptr,
    Gamut_error*              cpu_time_ptr,
    Gamut_input_data_error*   input_data_error_ptr
)
{
    IMPORT double min_absolute_pixel_error;
    IMPORT double min_relative_pixel_error;
    IMPORT Negative_data_action negative_3D_data_action;
    int           plot_id_list[ MAX_NUM_3D_PLOTS ];
    Hull_result   hull_result;
    double        absolute_pixel_error;
    double        relative_pixel_error;
    int           map_res                           = NO_ERROR;
    int           try_count                               = 0;
    Matrix*       derived_data_mp                   = NULL;

    initialize_hull_result(&hull_result);

    EPE(initialize_3D_plots(canon_gamut_num, cc_methods_ptr->plot_enable,
                            plot_id_list));

    if (process_negative_data(&derived_data_mp, data_mp,
                              negative_3D_data_action,
                              "input data")
        == ERROR)
    {
        free_matrix(derived_data_mp);
        return ERROR;
    }

    absolute_pixel_error = min_absolute_pixel_error;
    relative_pixel_error = min_relative_pixel_error;

    while (try_count < 20)
    {
        try_count++;

        map_res = do_3D_maps(cc_methods_ptr,
                             plot_id_list,
                             prompt_to_increase_error_flag,
                             canon_gamut_mp,
                             data_point_nums, derived_data_mp,
                             absolute_pixel_error, relative_pixel_error,
                             &(result_ptr->gamut_3D[ canon_gamut_num]),
                             &hull_result,
                             cpu_time_ptr);

        if (    (map_res == NO_SOLUTION)
             && (ok_to_increase_pixel_error(prompt_to_increase_error_flag,
                                            &absolute_pixel_error,
                                            &relative_pixel_error))
           )
        {
#ifdef TEST
            interactive_pso("Freeing hull result .... ");
#endif
            clear_hull_result(&hull_result);
#ifdef TEST
            interactive_pso("done.\n");
#endif

            if (plot_id_list[ PR_INTERSECTION_PLOT ] > 0)
            {
                EPE(plot_clear(plot_id_list[ PR_INTERSECTION_PLOT ]));
            }
        }
        else
        {
            break;
        }
    }

    if (map_res == NO_SOLUTION)
    {
        set_error("Unable to find a 3D gamut solution after %d tries.",
                  try_count);

        if (absolute_pixel_error > 0.0)
        {
            add_error("Greatest absolute pixel error used was %.1f.",
                      absolute_pixel_error);
        }
        if (relative_pixel_error > 0.0)
        {
            add_error("Greatest pixel error used was  %.1f%%.",
                      (100.0 * relative_pixel_error));
        }

        map_res = ERROR;
    }

    if (    (map_res == NO_ERROR)
         && (plot_id_list[ PR_INTERSECTION_PLOT ] > 0)
       )
    {
        EPE(plot_3D_result(plot_id_list[PR_INTERSECTION_PLOT ],
                           (plot_id_list[ PR_HULL_PLOT ] > 0),
                           canon_gamut_num, result_ptr, &hull_result));
    }

#ifdef TEST
    interactive_pso("Freeing hull result .... ");
#endif
    clear_hull_result(&hull_result);
#ifdef TEST
    interactive_pso("done.\n");
#endif

    if (map_res == NO_ERROR)
    {
        input_data_error_ptr->absolute_pixel_error = absolute_pixel_error;
        input_data_error_ptr->relative_pixel_error = relative_pixel_error;
    }
    else
    {
        input_data_error_ptr->absolute_pixel_error = DBL_HALF_MOST_NEGATIVE;
        input_data_error_ptr->relative_pixel_error = DBL_HALF_MOST_NEGATIVE;
    }

    free_matrix(derived_data_mp);

    return map_res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_maps
(
    Colour_constancy_methods* cc_methods_ptr,
    int*                      plot_id_list,
    int                       prompt_to_increase_error_flag,
    Matrix*                   canon_gamut_mp,
    int*                      data_point_nums,
    Matrix*                   data_mp,
    double                    absolute_pixel_error,
    double                    relative_pixel_error,
    Gamut_result*             result_ptr,
    Hull_result*              hull_result_ptr,
    Gamut_error*              cpu_time_ptr
)
{
    Constraints constraints;
    double      size                   = 0.0;
    int         res;
    Vector*     canon_illum_vp         = NULL;

    fs_illum_constraint_error = FALSE;

    initialize_constraints(&constraints);

    res = get_sharp_canon_illum(&canon_illum_vp);

    if (res != ERROR)
    {
        res = copy_vector(&(constraints.canon_illum_vp), canon_illum_vp);
    }

    res = get_sum_luminance(data_mp, constraints.sum_luminance);

    if (res != ERROR)
    {
        res = do_3D_surface_maps(canon_gamut_mp,
                                 plot_id_list, data_point_nums, data_mp,
                                 absolute_pixel_error, relative_pixel_error,
                                 &size, &(constraints.surface_hull_list_head));
    }

    if (res == NO_ERROR)
    {
        /*
        // In theory this could be calculated once unless one of a certain
        // small number of things changed (illum RGB, canon illum RGB, hull
        // intersection resolution, sharpening, and ?). However, the code is
        // simpler (at least without major changes) if we just re-calculate
        // this. If we wanted to optimize the code, then the illuminant hull
        // should be calculated just once (until something relavent changes).
        //
        // NOTE: If this change is made, it will be necessary to compute
        //       both sharp and dull illuminant hulls. (And 2D/3D). As
        //       discussed in the TODO file, it would probably be best to
        //       create an environment/cache structure for this hull and
        //       the canonical RGB hull.
        //
        // Note that there is always only one illum hull even though we are
        // using a queue. This is to remain consistent with the other two hull
        // sets. (I'm not sure it this is a win). To change this, code in
        // cc_alg.c will also have to be changed.
        */
        res = do_3D_illum_maps(plot_id_list, 10.0 * size,
                               &(constraints.illum_hull_list_head),
                               &(constraints.illum_hp));

#ifdef TEST
        if (res == ERROR)
        {
            ASSERT(constraints.illum_hull_list_head == NULL);
            ASSERT(constraints.illum_hp == NULL);
        }
#endif

        if ((res == ERROR) && ( ! cc_methods_ptr->skip_on_error))
        {
            kjb_get_error(fs_illum_constraint_error_message,
                          sizeof(fs_illum_constraint_error_message));

            fs_illum_constraint_error = TRUE;
            res = NO_ERROR;
        }
    }

    free_vector(canon_illum_vp);

    if (res != ERROR)
    {
#ifdef TEST
        interactive_pso("Choosing solutions for 3D gamut mapping.\n");
#endif

        res = do_gamut_colour_constancy(cc_methods_ptr,
                                        prompt_to_increase_error_flag,
                                        &constraints,
                                        result_ptr, hull_result_ptr,
                                        cpu_time_ptr);
    }

#ifdef TEST
    interactive_pso("Freeing constraints .... ");
#endif

    free_constraints(&constraints);

#ifdef TEST
    interactive_pso("done.\n");
#endif

    return res;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_surface_maps
(
    Matrix*         canon_gamut_mp,
    int*            plot_id_list,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
)
{
    IMPORT int  use_data_hull;
    IMPORT double relative_sensor_error;
    Hull*       canon_hp;
    Hull*       expanded_canon_hp;
    int         result;

#ifdef TEST
    if (is_interactive()) TEST_PSE(("Doing 3D surface maps.\n"));
#endif

    NRE(canon_hp = find_convex_hull(canon_gamut_mp, DEFAULT_HULL_OPTIONS));

    if (relative_sensor_error > DBL_EPSILON)
    {
        expanded_canon_hp = expand_hull(canon_hp, relative_sensor_error,
                                        DEFAULT_HULL_OPTIONS);

        free_hull(canon_hp);

        if (expanded_canon_hp == NULL)
        {
            return ERROR;
        }
        else
        {
            canon_hp = expanded_canon_hp;
        }
    }

    if (plot_id_list[ PR_DATA_PLOT ] > 0)
    {
        int     plot_id    = plot_id_list[ PR_DATA_PLOT ];
        Matrix* all_data_mp = NULL;
        Vector* hull_max_vp = NULL;
        double  max_x       = DBL_NOT_SET;
        double  max_y       = DBL_NOT_SET;
        int plot_result;

        if (fs_always_have_zero_in_canon_gamut)
        {
            Matrix* canon_hull_point_sources[ 2 ];
            Matrix* zero_mp = NULL;
            Matrix* canon_hull_points_with_zero_mp = NULL;
            Hull*   canon_with_zero_hp = NULL;

            EPE(get_zero_matrix(&zero_mp, 1, 3));

            if (zero_mp != NULL)
            {
                canon_hull_point_sources[ 0 ] = zero_mp;
                canon_hull_point_sources[ 1 ] = canon_hp->vertex_mp;

                EPE(concat_matrices_vertically(&canon_hull_points_with_zero_mp,
                                               2, canon_hull_point_sources));
            }

            if (canon_hull_points_with_zero_mp != NULL)
            {
                NPE(canon_with_zero_hp = find_convex_hull(
                                                canon_hull_points_with_zero_mp,
                                                DEFAULT_HULL_OPTIONS));
            }

            if (canon_with_zero_hp != NULL)
            {
                EPE(plot_3D_hull_projection(plot_id, canon_with_zero_hp,
                                   "Canonical Gamut Hull (with zero)",
                                   (Vector**)NULL, &hull_max_vp));
                max_x = hull_max_vp->elements[ 0 ] ;
                max_y = hull_max_vp->elements[ 1 ] ;
            }

            free_matrix(zero_mp);
            free_matrix(canon_hull_points_with_zero_mp);
            free_hull(canon_with_zero_hp);
        }
        else
        {
            EPE(plot_3D_hull_projection(plot_id, canon_hp,
                               "Convex Hull of Canonical Gamut (no zero added)",
                               (Vector**)NULL, &hull_max_vp));
            max_x = hull_max_vp->elements[ 0 ] ;
            max_y = hull_max_vp->elements[ 1 ] ;
        }

        if ((data_mp != NULL) && (data_mp->num_rows > 3))
        {
            Hull* data_hp;

            NPE(data_hp = find_convex_hull(data_mp, DEFAULT_HULL_OPTIONS));

            if (data_hp != NULL)
            {
                EPE(plot_3D_hull_projection(plot_id, data_hp,
                                            "Convex Hull of Data",
                                            (Vector**)NULL, &hull_max_vp));
                if (hull_max_vp->elements[ 0 ] > max_x)
                {
                    max_x = hull_max_vp->elements[ 0 ];
                }

                if (hull_max_vp->elements[ 1 ] > max_y)
                {
                    max_y = hull_max_vp->elements[ 1 ];
                }
            }

            free_hull(data_hp);
        }

        plot_result = get_selected_derived_data(&all_data_mp,
                                               0, (int*)NULL);

        if (plot_result == ERROR)
        {
            kjb_print_error();
        }

        if ((all_data_mp != NULL) && (all_data_mp->num_rows > 3))
        {
            Hull* data_hp;

            NPE(data_hp = find_convex_hull(all_data_mp, DEFAULT_HULL_OPTIONS));

            if (data_hp != NULL)
            {
                EPE(plot_3D_hull_projection(plot_id, data_hp,
                                            "Convex Hull of All Data",
                                            (Vector**)NULL, (Vector**)NULL));
            }

            free_hull(data_hp);
        }

        if (data_mp != NULL)
        {
            EPE(plot_data_projection(plot_id, data_mp, "Data Points"));
        }

        free_matrix(all_data_mp);

        if ((max_x > 0.0) && (max_y > 0.0))
        {
            char key_command[ 100 ];

            EPE(plot_set_range(plot_id, 0.0, max_x, 0.0, max_y));
            EPE(kjb_sprintf(key_command, sizeof(key_command),
                            "set key %f,%f\n", 1.45 * max_x, 0.8 * max_y));
            EPE(plot_write(plot_id, key_command));
            EPE(plot_update(plot_id));
        }

        free_vector(hull_max_vp);
    }

    if (use_data_hull)
    {
        result = do_3D_hull_surface_maps(plot_id_list,
                                         canon_hp->vertex_mp, data_mp,
                                         absolute_pixel_error,
                                         relative_pixel_error,
                                         size_ptr, hull_list_head_ptr);
    }
    else
    {
        result = do_3D_point_surface_maps(plot_id_list,
                                          canon_hp->vertex_mp,
                                          data_point_nums, data_mp,
                                          absolute_pixel_error,
                                          relative_pixel_error,
                                          size_ptr, hull_list_head_ptr);
    }

    free_hull(canon_hp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_hull_surface_maps
(
    int*            plot_id_list,
    Matrix*         canon_gamut_hull_mp,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
)
{
    Matrix* error_mp;
    Matrix* no_error_mp          = NULL;
    Hull*   data_hp;
    Hull*   no_error_hp;
    Hull*   hp                   = NULL;/* I don't think this value gets use. */
    Matrix* data_hull_mp;
    int     num_data_hull_points;
    Hull**  hp_list              = NULL;
    Hull**  no_error_hp_list     = NULL;
    unsigned long   hull_options;
    int     i;
    int     j;
    double    max_sum;
    int     count;
    int     result = NO_ERROR;
    int     do_error;

    /*
    // General notes for this routine:
    //
    // We don't bother skipping finding the hulls in the no error case, even
    // though it is not really needed. However, speeding up the code in this way
    // would requre too much re-writing.
    //
    // Also, this module is a bit ugly due to history, but I think I will pass
    // on cleaning it up for now.
    */

    if (data_mp->num_rows > 3)
    {
        data_hp = find_convex_hull(data_mp, DEFAULT_HULL_OPTIONS);

        if (data_hp == NULL)
        {
            add_error("Unable to get convex hull of data.");
            return ERROR;
        }

        num_data_hull_points = data_hp->vertex_mp->num_rows;
        data_hull_mp = data_hp->vertex_mp;
    }
    else
    {
        data_hp = NULL;
        data_hull_mp = data_mp;
        num_data_hull_points = data_mp->num_rows;
    }

    max_sum = 0;

    if (    (absolute_pixel_error > 0.0)
         || (relative_pixel_error > 0.0)
       )
    {
        do_error = TRUE;
    }
    else
    {
        do_error = FALSE;
    }

    /*
    // Dimension error_mp large enough for a cube for each point plus one
    // point for pure black.
    */
    if (fs_always_have_zero_in_canon_gamut)
    {
        NRE(error_mp = create_matrix(8*(canon_gamut_hull_mp->num_rows)+1, 3));
    }
    else
    {
        NRE(error_mp = create_matrix(8*(canon_gamut_hull_mp->num_rows), 3));
    }

    if (plot_id_list[ PR_HULL_PLOT ] > 0)
    {
        NRE(hp_list = N_TYPE_MALLOC(Hull*, num_data_hull_points));
    }

    if (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
    {
        NRE(no_error_hp_list = N_TYPE_MALLOC(Hull*, num_data_hull_points));

        for (i=0; i<num_data_hull_points; i++)
        {
            no_error_hp_list[ i ] = NULL;
        }
    }

    if (    (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
         || (fs_use_approximate_3D_error_hull && do_error)
       )
    {
        if (fs_always_have_zero_in_canon_gamut)
        {
            NRE(no_error_mp = create_matrix(canon_gamut_hull_mp->num_rows + 1,
                                            3));
        }
        else
        {
            NRE(no_error_mp = create_matrix(canon_gamut_hull_mp->num_rows, 3));
        }
    }

    for (i = 0; i<num_data_hull_points; i++)
    {
        double r = MAX_OF(0.1, (data_hull_mp->elements)[ i ][ 0 ]);
        double g = MAX_OF(0.1, (data_hull_mp->elements)[ i ][ 1 ]);
        double b = MAX_OF(0.1, (data_hull_mp->elements)[ i ][ 2 ]);
        double max_r = r, max_g = g, max_b = b, min_r = r, min_g = g, min_b = b;

        if (result == ERROR) break;

        /*
        // Originally we handled the possibility that one of the two error types
        // is zero by falling through. However, qhull does not like repeated
        // data, so we safeguard by doing the non-error case separately.
        */

        if (relative_pixel_error > 0.0)
        {
            min_r *= (1.0 - relative_pixel_error);
            max_r *= (1.0 + relative_pixel_error);
            min_g *= (1.0 - relative_pixel_error);
            max_g *= (1.0 + relative_pixel_error);
            min_b *= (1.0 - relative_pixel_error);
            max_b *= (1.0 + relative_pixel_error);
        }

        if (absolute_pixel_error > 0.0)
        {
            min_r -= absolute_pixel_error;
            max_r += absolute_pixel_error;
            min_g -= absolute_pixel_error;
            max_g += absolute_pixel_error;
            min_b -= absolute_pixel_error;
            max_b += absolute_pixel_error;
        }

        /*
        // We can't allow min_r, etc to be zero (or less). Given good
        // pre-processing, or good synthetic data, min_r etc should be well
        // above zero. However, with large error, we may need to threshold.
        */
        if (min_r < 1.0) min_r = 1.0;
        if (min_g < 1.0) min_g = 1.0;
        if (min_b < 1.0) min_b = 1.0;

        count = 0;

        for (j = 0; j<canon_gamut_hull_mp->num_rows; j++)
        {
            double r_c = (canon_gamut_hull_mp->elements)[ j ][ 0 ];
            double g_c = (canon_gamut_hull_mp->elements)[ j ][ 1 ];
            double b_c = (canon_gamut_hull_mp->elements)[ j ][ 2 ];
            int  k, m, n;
            double x, y, z;

            if (    (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
                 || (fs_use_approximate_3D_error_hull && do_error)
               )
            {
                (no_error_mp->elements)[ j ][0] = r_c / r;
                (no_error_mp->elements)[ j ][1] = g_c / g;
                (no_error_mp->elements)[ j ][2] = b_c / b;
            }

            if ( ! do_error)
            {
                (error_mp->elements)[ count ][0] = x = r_c / r;
                (error_mp->elements)[ count ][1] = y = g_c / g;
                (error_mp->elements)[ count ][2] = z = b_c / b;

                if ((x + y + z) > max_sum) max_sum = (x + y + z);

                count++;
            }
            else
            {
                for (k=0; k<2; k++)
                {
                    for (m=0; m<2; m++)
                    {
                        for (n=0; n<2; n++)
                        {
                            x = r_c / ((IS_EVEN(n)) ? min_r : max_r);
                            y = g_c / ((IS_EVEN(m)) ? min_g : max_g);
                            z = b_c / ((IS_EVEN(k)) ? min_b : max_b);

                            (error_mp->elements)[ count ][ 0 ] = x;
                            (error_mp->elements)[ count ][ 1 ] = y;
                            (error_mp->elements)[ count ][ 2 ] = z;

                            if ((x + y + z) > max_sum) max_sum = (x + y + z);

                            count++;
                        }
                    }
                }
            }
        }

        /* In the !do_error case, error_mp will not be full. */
        error_mp->num_rows = count;

        if ((result != ERROR) && (fs_use_approximate_3D_error_hull))
        {
            Matrix* approximate_error_mp = NULL;

            if (fs_always_have_zero_in_canon_gamut)
            {
                (no_error_mp->num_rows)--;
            }

            no_error_hp = find_convex_hull(no_error_mp, DEFAULT_HULL_OPTIONS);

            if (fs_always_have_zero_in_canon_gamut)
            {
                (no_error_mp->num_rows)++;
            }

            if (no_error_hp == NULL)
            {
                result = ERROR;
            }
            else
            {
                result = get_approximate_error_hull_data(&approximate_error_mp,
                                                         error_mp,
                                                         8, no_error_hp);
            }

            if (result != ERROR)
            {
                result = copy_matrix(&error_mp, approximate_error_mp);
            }

            free_matrix(approximate_error_mp);
            free_hull(no_error_hp);
        }

        if ((result != ERROR) && (fs_always_have_zero_in_canon_gamut))
        {
            Matrix* with_zero_error_mp = NULL;
            int     num_rows           = error_mp->num_rows;

            if (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
            {
                (no_error_mp->elements)[ j ][ 0 ] = 0.0;
                (no_error_mp->elements)[ j ][ 1 ] = 0.0;
                (no_error_mp->elements)[ j ][ 2 ] = 0.0;
            }

            result = get_target_matrix(&with_zero_error_mp, num_rows + 1, 3);

            if (result != ERROR)
            {
                (with_zero_error_mp->num_rows)--;
                result = copy_matrix(&with_zero_error_mp, error_mp);
                (with_zero_error_mp->num_rows)++;

                (with_zero_error_mp->elements)[ num_rows ][ 0 ] = 0.0;
                (with_zero_error_mp->elements)[ num_rows ][ 1 ] = 0.0;
                (with_zero_error_mp->elements)[ num_rows ][ 2 ] = 0.0;
            }

            if (result != ERROR)
            {
                result = copy_matrix(&error_mp, with_zero_error_mp);
            }

            free_matrix(with_zero_error_mp);
        }

        if (result != ERROR)
        {
            hull_options = DEFAULT_HULL_OPTIONS;

            hp = find_convex_hull(error_mp, hull_options);

            if (hp == NULL)
            {
                add_error("Unable to find hull of transforms with error.");
                result = ERROR;
            }
            else
            {
                result = insert_into_queue(hull_list_head_ptr,
                                           (Queue_element**)NULL, (void*)hp);
            }
        }

        if ((result != ERROR) && (plot_id_list[ PR_HULL_PLOT ] > 0))
        {
            hp_list[ i ] = hp;
        }

        if (    (result != ERROR)
             && (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0))
        {
            no_error_hp = find_convex_hull(no_error_mp, DEFAULT_HULL_OPTIONS);

            if (no_error_hp == NULL)
            {
                result = ERROR;
            }
            else
            {
                no_error_hp_list[ i ] = no_error_hp;
            }
        }

    }

    if ((result != ERROR) && (plot_id_list[ PR_HULL_PLOT ] > 0))
    {
        EPE(plot_3D_hull_list_projection(plot_id_list[ PR_HULL_PLOT ],
                                         num_data_hull_points, hp_list,
                                         "S Constraint(s)"));
    }

    if (plot_id_list[ PR_HULL_PLOT ] > 0)
    {
        kjb_free(hp_list);
    }

    if ((result != ERROR) && (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0))
    {
        EPE(plot_3D_hull_list_projection(plot_id_list[ PR_NO_ERROR_HULL_PLOT ],
                                         num_data_hull_points, no_error_hp_list,
                                         "S Constraint(s) w/o Error"));
    }

    if (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
    {
        for (i=0; i<num_data_hull_points; i++)
        {
            free_hull(no_error_hp_list[ i ]);
        }
        kjb_free(no_error_hp_list);
    }

    free_matrix(no_error_mp);
    free_hull(data_hp);
    free_matrix(error_mp);

    *size_ptr = max_sum;

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_point_surface_maps
(
    int*            plot_id_list,
    Matrix*         canon_gamut_hull_mp,
    int*            data_point_nums,
    Matrix*         data_mp,
    double          absolute_pixel_error,
    double          relative_pixel_error,
    double*         size_ptr,
    Queue_element** hull_list_head_ptr
)
{
    Matrix* error_mp;
    int     i;
    int     j;
    double    max_sum      = 0.0;
    Hull*   hp;
    unsigned long   hull_options;
    int     result       = NO_ERROR;

    /*
    // General notes for this routine:
    //
    // We don't bother handling the potentially faster no error case
    // differently, as the point surface method is slow anyway, and is only
    // used for debugging, checking, getting specialized plots, etc.
    //
    // I have not bothered  implementing plotting of no-error hulls.
    //
    // Also, this module is a bit ugly due to history, but I think I will pass
    // on cleaning it up for now.
    */

    /*
    // Dimension error_mp large enough for a cube for each point plus one
    // point for pure black.
    */
    if (fs_always_have_zero_in_canon_gamut)
    {
        NRE(error_mp = create_matrix(8*(canon_gamut_hull_mp->num_rows)+1, 3));
    }
    else
    {
        NRE(error_mp = create_matrix(8*(canon_gamut_hull_mp->num_rows), 3));
    }

    for (i = 0; i<data_mp->num_rows; i++)
    {
        int  count = 0;
        double max_r, max_g, max_b, min_r, min_g, min_b;
        int  dont_do_error = FALSE;

        min_r = max_r = MAX_OF(0.1, (data_mp->elements)[ i ][ 0 ]);
        min_g = max_g = MAX_OF(0.1, (data_mp->elements)[ i ][ 1 ]);
        min_b = max_b = MAX_OF(0.1, (data_mp->elements)[ i ][ 2 ]);

        /*
        // Originally we handled the possibility that one of the two error types
        // is zero by falling through. However, qhull does not like reapeated
        // data, so we safeguard by doing the non-error case separately.
        */

        if (    (absolute_pixel_error <= 0.0)
             && (relative_pixel_error <= 0.0)
           )
        {
            dont_do_error = TRUE;
        }
        else
        {
            if (relative_pixel_error > 0.0)
            {
                min_r *= (1.0 - relative_pixel_error);
                max_r *= (1.0 + relative_pixel_error);
                min_g *= (1.0 - relative_pixel_error);
                max_g *= (1.0 + relative_pixel_error);
                min_b *= (1.0 - relative_pixel_error);
                max_b *= (1.0 + relative_pixel_error);
            }

            if (absolute_pixel_error > 0.0)
            {
                min_r -= absolute_pixel_error;
                max_r += absolute_pixel_error;
                min_g -= absolute_pixel_error;
                max_g += absolute_pixel_error;
                min_b -= absolute_pixel_error;
                max_b += absolute_pixel_error;
            }
        }

        /*
        // We can't have min_r, etc zero (or less). Give good pre-processing,
        // or good synthetic data, min_r etc should be well above zero.
        // However, with large error, we may need to threshold.
        */
        if (min_r < 1.0) min_r = 1.0;
        if (min_g < 1.0) min_g = 1.0;
        if (min_b < 1.0) min_b = 1.0;

        for (j = 0; j<canon_gamut_hull_mp->num_rows; j++)
        {
            double r_c = (canon_gamut_hull_mp->elements)[ j ][ 0 ];
            double g_c = (canon_gamut_hull_mp->elements)[ j ][ 1 ];
            double b_c = (canon_gamut_hull_mp->elements)[ j ][ 2 ];
            int  k, m, n;
            double x, y, z;

            if (dont_do_error)
            {
                (error_mp->elements)[ count ][ 0 ] = x = r_c / min_r;
                (error_mp->elements)[ count ][ 1 ] = y = g_c / min_g;
                (error_mp->elements)[ count ][ 2 ] = z = b_c / min_b;

                if ((x + y + z) > max_sum) max_sum = (x + y + z);

                count++;
            }
            else
            {
                for (k=0; k<2; k++)
                {
                    for (m=0; m<2; m++)
                    {
                        for (n=0; n<2; n++)
                        {
                            x = r_c / ((IS_EVEN(n)) ? min_r : max_r);
                            y = g_c / ((IS_EVEN(m)) ? min_g : max_g);
                            z = b_c / ((IS_EVEN(k)) ? min_b : max_b);

                            (error_mp->elements)[ count ][ 0 ] = x;
                            (error_mp->elements)[ count ][ 1 ] = y;
                            (error_mp->elements)[ count ][ 2 ] = z;

                            if ((x + y + z) > max_sum) max_sum = (x + y + z);

                            count++;
                        }
                    }
                }
            }

            error_mp->num_rows = count;

            if (fs_always_have_zero_in_canon_gamut)
            {
                (error_mp->elements)[ count ][ 0 ] = 0.0;
                (error_mp->elements)[ count ][ 1 ] = 0.0;
                (error_mp->elements)[ count ][ 2 ] = 0.0;

                error_mp->num_rows = count + 1;
            }
        }

        hull_options = DEFAULT_HULL_OPTIONS;

        hp = find_convex_hull(error_mp, hull_options);

        if (hp == NULL)
        {
            result = ERROR;
        }
        else if (plot_id_list[ PR_HULL_PLOT ] > 0)
        {
            char title[ 200 ];

            result = kjb_sprintf(title, sizeof(title), "Surface %d Hull",
                                 data_point_nums[ i ]);

            if (result != ERROR)
            {
                EPE(plot_3D_hull_projection(plot_id_list[ PR_HULL_PLOT ],
                                            hp, title,
                                            (Vector**)NULL,
                                            (Vector**)NULL));
            }
        }

        if (result != ERROR)
        {
            result = insert_into_queue(hull_list_head_ptr,
                                       (Queue_element**)NULL, (void*)hp);
        }
    }

    free_matrix(error_mp);

    *size_ptr = max_sum;

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_3D_illum_maps
(
    int*            plot_id_list,
    double          size,
    Queue_element** hull_list_head_ptr,
    Hull**          illum_hp_ptr
)
{
    IMPORT double relative_sensor_error;
    Matrix*       illum_3D_map_mp       = NULL;
    Matrix*       illum_2D_map_mp       = NULL;
    Matrix*       illum_2D_map_error_points_mp = NULL;
    Matrix*       projected_illum_mp = NULL;
    Hull*         illum_2D_map_hp    = NULL;
    double        min;
    Vector*       canon_illum_vp = NULL;
    Matrix*       illum_mp       = NULL;
    int           result         = NO_ERROR;

    /*
    // If there is no illumination hull or canon illum, then we still want to
    // carry on, so just return NO_ERROR.
    */
    if (    (get_sharp_illum_gamut(&illum_mp) == ERROR)
         || (get_sharp_canon_illum(&canon_illum_vp) == ERROR)
       )
    {
        free_matrix(illum_mp);
        free_vector(canon_illum_vp);
        return NO_ERROR;
    }

#ifdef TEST
    if (is_interactive()) TEST_PSE(("Doing 3D illum maps.\n"));
#endif

    if (result != ERROR)
    {

        min = min_matrix_element(illum_mp);

        dbf(max_matrix_element(illum_mp));

        if (min < MIN_ALLOWABLE_3D_ILLUM_VALUE)
        {
            set_error("Illum RGB contains values less than %.1e.",
                      MIN_ALLOWABLE_3D_ILLUM_VALUE);
            cat_error(" (Min RGB is %.1e.)", min);
            result = ERROR;
        }
        else if (min < MIN_USABLE_3D_ILLUM_VALUE)
        {
            char sharp_method_string[ 100 ];

            get_current_sharp_method_string(sharp_method_string,
                                            sizeof(sharp_method_string));

            warn_pso("Illum RGB contains values less than %.1f.\n",
                     MIN_USABLE_3D_ILLUM_VALUE);
            warn_pso("Min value is %.2e (min allowable is %.1f).\n",
                     min, MIN_ALLOWABLE_3D_ILLUM_VALUE);
            warn_pso("Resetting those values to %.1f.\n",
                     MIN_USABLE_3D_ILLUM_VALUE);
            warn_pso("Problem occured doing 3D gamut mapping with %q.\n\n",
                     sharp_method_string);

            result = ow_min_thresh_matrix(illum_mp, MIN_USABLE_3D_ILLUM_VALUE);
        }
    }

    if (result != ERROR)
    {
        result = divide_matrix_by_row_vector(&illum_3D_map_mp, illum_mp,
                                              canon_illum_vp);
    }

    if (result != ERROR)
    {
        result = get_projection_matrix(&illum_2D_map_mp, illum_3D_map_mp,
                                       DIVIDE_BY_SUM, SKIP_INVALID_CHROMATICITY,
                                       5.0 * MAX_X_DIV_SUM, DBL_NOT_SET,
                                       relative_sensor_error, (Matrix **)NULL,
                                       &illum_2D_map_error_points_mp);
        dbf(min_matrix_element(illum_2D_map_mp));
        dbf(max_matrix_element(illum_2D_map_mp));

    }

    if (result != ERROR)
    {
        illum_2D_map_hp = find_convex_hull(illum_2D_map_error_points_mp,
                                           DEFAULT_HULL_OPTIONS);
        if (illum_2D_map_hp == NULL)
        {
            add_error("Problem occured finding 2D illum hull for 3D gamut algorithms.");
            result = ERROR;
        }
    }

    if (result != ERROR)
    {
        result = do_illum_3D_inv_trans_gamut(plot_id_list,
                                             illum_2D_map_hp,
                                             size, hull_list_head_ptr);
    }

    free_hull(illum_2D_map_hp);

    if (result != ERROR)
    {
        result = project_matrix(&projected_illum_mp, illum_mp, DIVIDE_BY_SUM);

        if (result == ERROR)
        {
            db_mat(projected_illum_mp);
        }
    }

    if (result != ERROR)
    {
        *illum_hp_ptr = find_convex_hull(projected_illum_mp,
                                         DEFAULT_HULL_OPTIONS);
        if (*illum_hp_ptr == NULL) result = ERROR;
    }

    if ((result != ERROR) && (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0))
    {
        Hull* no_error_illum_2D_map_hp = find_convex_hull(illum_2D_map_mp,
                                                          DEFAULT_HULL_OPTIONS);

        if (no_error_illum_2D_map_hp == NULL)
        {
            result = ERROR;
        }
        else
        {
            result = do_illum_3D_inv_trans_gamut(plot_id_list,
                                                 no_error_illum_2D_map_hp,
                                                 size, (Queue_element **)NULL);
        }

        free_hull(no_error_illum_2D_map_hp);
    }

    if ((result != ERROR) && (plot_id_list[ PR_INTERSECTION_PLOT] > 0))
    {
        EPE(plot_projected_black_body_curve(plot_id_list[PR_INTERSECTION_PLOT],
                                            canon_illum_vp,
                                            3 /* dimension */));
    }

    free_matrix(illum_3D_map_mp);
    free_matrix(illum_2D_map_mp);
    free_matrix(projected_illum_mp);
    free_matrix(illum_2D_map_error_points_mp);

    free_matrix(illum_mp);
    free_vector(canon_illum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int do_illum_3D_inv_trans_gamut
(
    int*            plot_id_list,
    Hull*           illum_2D_map_hp,
    double          max_sum,
    Queue_element** hull_list_head_ptr
)
{
    IMPORT int plot_illumination_constraint;
    Matrix*    illum_3D_inv_map;
    Matrix**   facet_mvp;
    Matrix*    facet_mp;
    int        i, j;
    int        num_facets;
    Matrix*    temp_mp;
    Matrix*    proj_inv_mp      = NULL;
    Matrix*    proj_inv_hull_mp;
    int        temp_row;
    double       x, y, z, x1, x2, y1, y2;
    Hull*      hp;
    int        plot_id;
    int        num_proj_inv_rows;

    num_facets = illum_2D_map_hp->num_facets;

    facet_mvp = illum_2D_map_hp->facets->elements;

    NRE(temp_mp = create_matrix((1+fs_gamut_3D_inversion_resolution)*num_facets,
                                3));

    temp_row = 0;

    for (i=0; i<num_facets; i++)
    {
        facet_mp = *facet_mvp;

        x1 = (facet_mp->elements)[0][0];
        x2 = (facet_mp->elements)[1][0];
        y1 = (facet_mp->elements)[0][1];
        y2 = (facet_mp->elements)[1][1];

        for (j=0; j<=fs_gamut_3D_inversion_resolution; j++)
        {
            x = (j*x1 + (fs_gamut_3D_inversion_resolution-j)*x2)/
                                                 fs_gamut_3D_inversion_resolution;

            y = (j*y1 + (fs_gamut_3D_inversion_resolution-j)*y2)/
                                                 fs_gamut_3D_inversion_resolution;

            (temp_mp->elements)[temp_row][0] = x;
            (temp_mp->elements)[temp_row][1] = y;
            (temp_mp->elements)[temp_row][2] = 1 - x - y;

            temp_row++;
        }

        facet_mvp++;
    }

    ERE(ow_invert_matrix_elements(temp_mp));

    ERE(get_projection_matrix(&proj_inv_mp, temp_mp, DIVIDE_BY_SUM,
                              SKIP_INVALID_CHROMATICITY, 5.0 * MAX_X_DIV_SUM,
                              DBL_NOT_SET, DBL_NOT_SET,
                              (Matrix**)NULL, (Matrix**)NULL));

    if (plot_id_list[PR_ILLUM_HULL_PLOT] > 0)
    {
        Matrix_vector* inverse_gamut_segments;
        Matrix*       inverse_gamut_segment_mp;
        int           count;

        plot_id = plot_id_list[ PR_ILLUM_HULL_PLOT ];

        NRE(inverse_gamut_segments = create_matrix_vector(num_facets));

        count = 0;

        for (i=0; i<num_facets; i++)
        {
            ERE(get_target_matrix(&(inverse_gamut_segments->elements[ i ]),
                                  fs_gamut_3D_inversion_resolution + 1, 2));

            inverse_gamut_segment_mp = inverse_gamut_segments->elements[ i ];

            for (j=0; j<fs_gamut_3D_inversion_resolution + 1; j++)
            {
                (inverse_gamut_segment_mp->elements)[j][0] =
                                           (proj_inv_mp->elements)[ count ][0];
                (inverse_gamut_segment_mp->elements)[j][1] =
                                           (proj_inv_mp->elements)[ count ][1];
                count++;
            }
        }

        if (plot_illumination_constraint)
        {
            if (hull_list_head_ptr == NULL)
            {
                EPE(plot_multi_segment_curve(plot_id, inverse_gamut_segments,
                                             "Inverse Hull Without Error"));
            }
            else
            {
                EPE(plot_multi_segment_curve(plot_id, inverse_gamut_segments,
                                             "Inverse Hull With Error"));
            }
        }

        free_matrix_vector(inverse_gamut_segments);
    }

    hp = find_convex_hull(proj_inv_mp, DEFAULT_HULL_OPTIONS);
    free_matrix(proj_inv_mp);

    if (hp == NULL)
    {
        free_matrix(temp_mp);
        return ERROR;
    }

    if (plot_id_list[ PR_ILLUM_HULL_PLOT ] > 0)
    {
        plot_id = plot_id_list[ PR_ILLUM_HULL_PLOT ];

        if (plot_illumination_constraint)
        {
            if (hull_list_head_ptr == NULL)
            {
                EPE(plot_matrix_vector_cols(plot_id, hp->facets,
                                           "Without Error Convex Estimate"));
            }
            else
            {
                EPE(plot_matrix_vector_cols(plot_id, hp->facets,
                                           "With Error Convex Estimate"));
            }
        }
    }

    proj_inv_hull_mp = hp->vertex_mp;
    num_proj_inv_rows = proj_inv_hull_mp->num_rows;

    /* If we want to include the origin. */
    NRE(illum_3D_inv_map = create_matrix(num_proj_inv_rows + 1, 3));

    /*
    // If we want to truncate.
    //
    // NRE(illum_3D_inv_map = create_matrix(2*num_proj_inv_rows, 3));
    */

    for (i=0; i < num_proj_inv_rows; i++)
    {
        x = (proj_inv_hull_mp->elements)[i][0];
        (illum_3D_inv_map->elements)[i][0] = x;

        y = (proj_inv_hull_mp->elements)[i][1];
        (illum_3D_inv_map->elements)[i][1] = y;

        z = 1.0 - x - y;
        (illum_3D_inv_map->elements)[i][2] = z;

        /*
        // Truncate !!!
        //
        // (illum_3D_inv_map->elements)[i + num_proj_inv_rows][ 0 ] = x/10000;
        // (illum_3D_inv_map->elements)[i + num_proj_inv_rows][ 1 ] = y/10000;
        // (illum_3D_inv_map->elements)[i + num_proj_inv_rows][ 2 ] = z/10000;
        */
    }

    free_hull(hp);

    /* If we want to include the origin. */
    (illum_3D_inv_map->elements)[ num_proj_inv_rows ][0] = 0.0;
    (illum_3D_inv_map->elements)[ num_proj_inv_rows ][1] = 0.0;
    (illum_3D_inv_map->elements)[ num_proj_inv_rows ][2] = 0.0;

    ERE(ow_multiply_matrix_by_scalar(illum_3D_inv_map, max_sum));

    hp = find_convex_hull(illum_3D_inv_map, DEFAULT_HULL_OPTIONS);

    if (hp == NULL)
    {
        free_matrix(temp_mp);
        free_matrix(illum_3D_inv_map);
        return ERROR;
    }

    if (plot_illumination_constraint)
    {
        if (hull_list_head_ptr == NULL)
        {
            if (plot_id_list[ PR_NO_ERROR_HULL_PLOT ] > 0)
            {
                EPE(plot_3D_hull_projection(
                                        plot_id_list[ PR_NO_ERROR_HULL_PLOT ],
                                            hp, "I Constraint W/O Error",
                                            (Vector**)NULL,
                                            (Vector**)NULL));
            }
        }
        else
        {
            if (plot_id_list[ PR_INTERSECTION_PLOT ] > 0)
            {
                EPE(plot_3D_hull_projection(
                                        plot_id_list[ PR_INTERSECTION_PLOT ],
                                            hp, "I Constraint",
                                            (Vector**)NULL,
                                            (Vector**)NULL));
            }
        }
    }

    if (hull_list_head_ptr != NULL)
    {
        ERE(insert_into_queue(hull_list_head_ptr, (Queue_element**)NULL,
                              (void*)hp));
    }
    else
    {
        free_hull(hp);
    }

    free_matrix(temp_mp);
    free_matrix(illum_3D_inv_map);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_3D_plots
(
    int         canon_gamut_num,
    int         enable_plotting,
    int*        plot_id_list
)
{
    IMPORT int prompt_for_plots;
    IMPORT int print_plot_style;
    int        plot_request_list[ MAX_NUM_3D_PLOTS ];
    int        i;

    for (i=0; i<MAX_NUM_3D_PLOTS; i++)
    {
        plot_request_list[ i ] = FALSE;
    }

    if ((is_interactive()) && (enable_plotting))
    {
        if ( ! prompt_for_plots)
        {
            /*
            // Default behaviour is to only show the intersection plot (unless
            // we are not using X11 in which case don't show anything). All
            // others have been initialized to FALSE above.
            */

            if (X11_is_ok())
            {
                plot_request_list[ PR_INTERSECTION_PLOT ] = TRUE;
            }
        }
        else
        {

            if (yes_no_query("Show projected hulls? (y/n) "))
            {
                 plot_request_list[ PR_DATA_PLOT ] = TRUE;
            }

            if (yes_no_query("Show projected hull interections? (y/n) "))
            {
                plot_request_list[ PR_INTERSECTION_PLOT ] = TRUE;

                if (yes_no_query("Show all hulls? (y/n) "))
                {
                    plot_request_list[ PR_HULL_PLOT ]= TRUE;

                    if (yes_no_query("Show without error? (y/n) "))
                    {
                        plot_request_list[ PR_NO_ERROR_HULL_PLOT ]= TRUE;
                    }
                }
            }

            if (yes_no_query("Show illumination gamut approximation? (y/n) "))
            {
                plot_request_list[ PR_ILLUM_HULL_PLOT ] = TRUE;
            }
        }
    }

    for (i=0; i<MAX_NUM_3D_PLOTS; i++)
    {
        plot_id_list[ i ] = 0;
    }

    if (plot_request_list[ PR_DATA_PLOT ] )
    {
        char  data_source[ MAX_FILE_NAME_SIZE ];
        char  plot_title_buff[ 1000 ];
        char  short_name[ 100 ];
        char  sharp_method_string[ 100 ];
        int   plot_id;

        ERE(plot_id = plot_open());
        plot_id_list[ PR_DATA_PLOT ] = plot_id;

        get_current_sharp_method_string(sharp_method_string,
                                        sizeof(sharp_method_string));
        get_data_source(data_source, sizeof(data_source));
        BUFF_GEN_GET_LAST_TOKEN(data_source, short_name, "/");

        ERE(kjb_sprintf(plot_title_buff, sizeof(plot_title_buff),
                        "%d: Projected 3D Hulls [ %d ] ( %s ) [ %s ]",
                        plot_id, canon_gamut_num + 1,
                        sharp_method_string, short_name));

        ERE(plot_set_title(plot_id_list[ PR_DATA_PLOT ],
                           plot_title_buff, 0, 1));

        ERE(plot_set_x_legend(plot_id_list[ PR_DATA_PLOT ],
                              "First Coordinate"));

        if (print_plot_style)
        {
            ERE(plot_set_y_legend(plot_id_list[ PR_DATA_PLOT ],
                                  "Second Coordinate"));
        }
        else
        {
            ERE(plot_set_y_legend(plot_id_list[ PR_DATA_PLOT ],
                                  "     Second Componant"));

            /*
            // Hack it for now (should be a routine to do this.)
            */
            ERE(send_command_to_plot(CURRENT_PLOT, "set size 0.7, 0.95"));
        }
    }

    if (plot_request_list[ PR_INTERSECTION_PLOT ] )
    {
        char  data_source[ MAX_FILE_NAME_SIZE ];
        char  plot_title_buff[ 1000 ];
        char  short_name[ 100 ];
        char  sharp_method_string[ 100 ];
        int   plot_id;

        ERE(plot_id = plot_open());
        plot_id_list[ PR_INTERSECTION_PLOT ] = plot_id;

        get_current_sharp_method_string(sharp_method_string,
                                        sizeof(sharp_method_string));
        get_data_source(data_source, sizeof(data_source));
        BUFF_GEN_GET_LAST_TOKEN(data_source, short_name, "/");

        ERE(kjb_sprintf(plot_title_buff, sizeof(plot_title_buff),
                        "%d: Projected 3D Diagonal Mappings [ %d ] ( %s ) [ %s ]",
                        plot_id, canon_gamut_num + 1,
                        sharp_method_string, short_name));

        ERE(plot_set_title(plot_id_list[ PR_INTERSECTION_PLOT ],
                           plot_title_buff, 0, 1));

        ERE(plot_set_x_legend(plot_id_list[ PR_INTERSECTION_PLOT ],
                              "First Componant of Diagonal Transform"));

        if (print_plot_style)
        {
            ERE(plot_set_y_legend(plot_id_list[ PR_INTERSECTION_PLOT ],
                                  "Second Componant of Diagonal Transform"));
        }
        else
        {
            ERE(plot_set_y_legend(plot_id_list[ PR_INTERSECTION_PLOT ],
                                  "     Second Componant"));

            /*
            // Hack it for now (should be a routine to do this.)
            */
            ERE(send_command_to_plot(CURRENT_PLOT, "set size 0.7, 0.95"));
        }

        if (plot_request_list[ PR_HULL_PLOT])
        {
            plot_id_list[ PR_HULL_PLOT] =
                    plot_id_list[ PR_INTERSECTION_PLOT ];

            if (plot_request_list[ PR_NO_ERROR_HULL_PLOT ])
            {
                plot_id_list[ PR_NO_ERROR_HULL_PLOT ] =
                        plot_id_list[ PR_INTERSECTION_PLOT ];
            }
        }

        if (plot_request_list[ PR_ILLUM_HULL_PLOT ])
        {
            ERE(plot_id_list[ PR_ILLUM_HULL_PLOT ] = plot_open());
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

