
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "data.h"
#include "image_input.h"

/* -------------------------------------------------------------------------- */

static KJB_image* fs_input_ip = NULL;
static char       fs_input_image_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int        fs_input_image_version = 0;

static KJB_image* fs_canon_ip = NULL;
static char       fs_canon_image_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static int        fs_canon_image_version = 0;

static KJB_image* fs_reduced_input_ip = NULL;
static KJB_image* fs_reduced_canon_ip = NULL;

/* -------------------------------------------------------------------------- */

static int check_image_file
(
    const char* proposed_file_name,
    char*       resolved_file_name,
    size_t      resolved_file_name_size
);

static int read_image_file
(
    char*       file_name,
    KJB_image** ipp,
    int*        version_ptr,
    KJB_image** reduced_ipp
);

static void invalidate_input_images(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif

/* -------------------------------------------------------------------------- */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

EXTERN_C int set_image_files(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "idf"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_input_image_config_file[ 0 ] == '\0')
            {
                ERE(pso("No input image file has been set.\n"));
            }
            else
            {
                ERE(pso("Input image file is %s\n", fs_input_image_config_file));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_input_image_config_file[ 0 ] = '\0';
            }
            else
            {
                /*
                // This copies value into fs_input_image_config_file if it is OK
                */
                ERE(check_image_file(value, fs_input_image_config_file,
                                     sizeof(fs_input_image_config_file)));
                ERE(force_image_data());
            }

            if (fs_input_ip != NULL) fs_input_ip->read_only = FALSE;
            kjb_free_image(fs_input_ip);
            fs_input_ip = NULL;
            kjb_free_image(fs_reduced_input_ip);
            fs_reduced_input_ip = NULL;
            fs_input_image_version++;
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "cidf"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_canon_image_config_file[ 0 ] == '\0')
            {
                ERE(pso("No canonical image file has been set.\n"));
            }
            else
            {
                ERE(pso("Canonical image file is %s\n",
                        fs_canon_image_config_file));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_canon_image_config_file[ 0 ] = '\0';
            }
            else
            {
                /*
                // This copies value into fs_input_image_config_file if it is OK
                */
                ERE(check_image_file(value, fs_canon_image_config_file,
                                     sizeof(fs_canon_image_config_file)));
            }

            if (fs_canon_ip != NULL) fs_canon_ip->read_only = FALSE;
            kjb_free_image(fs_canon_ip);
            fs_canon_ip = NULL;
            kjb_free_image(fs_reduced_canon_ip);
            fs_reduced_canon_ip = NULL;
            fs_canon_image_version++;

        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int check_image_file
(
    const char* proposed_file_name,
    char*       resolved_file_name,
    size_t      resolved_file_name_size
)
{
    int result;

    result = get_config_file((const char*)NULL, (const char*)NULL,
                             proposed_file_name, "image data file",
                             resolved_file_name, resolved_file_name_size);

    if (result == ERROR)
    {
        char proposed_file_name_gz[ MAX_FILE_NAME_SIZE ];

        BUFF_CPY(proposed_file_name_gz, proposed_file_name);
        BUFF_CAT(proposed_file_name_gz, ".gz");

        result = get_config_file((const char*)NULL, (const char*)NULL,
                                 proposed_file_name_gz, "image data file",
                                 resolved_file_name, resolved_file_name_size);

        if (result != ERROR)
        {
            ERE(verbose_pso(10, "Adding \".gz\" to %q.\n",
                            proposed_file_name));
        }
    }

    /*
    // A bid of hack to get the error message for the non gzipped file name.
    // We proceed as if this could succeed in order to maintain consistency on
    // the very small chance that it does!
    */
    if (result == ERROR)
    {
        result = get_config_file((const char*)NULL, (const char*)NULL,
                                 proposed_file_name,
                                 "image data file",
                                 resolved_file_name,
                                 resolved_file_name_size);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void get_image_data_source(char* source_buff, size_t max_len)
{
    kjb_strncpy(source_buff, fs_input_image_config_file, max_len);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_image_file
(
    char*       file_name,
    KJB_image** ipp,
    int*        version_ptr,
    KJB_image** reduced_ipp
)
{
    static int first_time = TRUE;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    if (first_time)
    {
        set_image_data_invalidation_fn(invalidate_input_images);
        first_time = FALSE;
    }

    if (*ipp != NULL) (*ipp)->read_only = FALSE;
    ERE(kjb_read_image(ipp, file_name));
    if (*ipp != NULL) (*ipp)->read_only = TRUE;

    (*version_ptr)++;

    kjb_free_image(*reduced_ipp);
    *reduced_ipp = NULL;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

const KJB_image* get_input_image(int* version_ptr)
{

    if (fs_input_ip == NULL)
    {
        kjb_clear_error();

        if (fs_input_image_config_file[ 0 ] == '\0')
        {
            set_error("No input image has been set.");
        }
        else if (STRCMP_EQ(fs_input_image_config_file, "<capture>"))
        {
            set_error("Captured input image has been invalidated.");
            add_error("It must be recaptured.");
        }
        else
        {
            ERN(read_image_file(fs_input_image_config_file, &fs_input_ip,
                                &fs_input_image_version, &fs_reduced_input_ip));
        }
    }

    if (version_ptr != NULL) *version_ptr = fs_input_image_version;

    return fs_input_ip;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int input_image_is_up_to_date(int proposed_current_version)
{
    return (fs_input_image_version == proposed_current_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

const KJB_image* get_canon_image(int* version_ptr)
{

    if (fs_canon_ip == NULL)
    {
        kjb_clear_error();

        if (fs_canon_image_config_file[ 0 ] == '\0')
        {
            set_error("No canon image has been set.");
        }
        else if (STRCMP_EQ(fs_canon_image_config_file, "<capture>"))
        {
            set_error("Captured canon image has been invalidated.");
            add_error("It must be recaptured.");
        }
        else
        {
            ERN(read_image_file(fs_canon_image_config_file, &fs_canon_ip,
                                &fs_canon_image_version, &fs_reduced_canon_ip));
        }
    }

    if (version_ptr != NULL ) *version_ptr = fs_canon_image_version;

    return fs_canon_ip;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int canon_image_is_up_to_date(int proposed_current_version)
{
    return (fs_canon_image_version == proposed_current_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

const KJB_image* get_reduced_input_image(void)
{
    const KJB_image* full_input_ip;

    if (fs_reduced_input_ip == NULL)
    {
        NRN(full_input_ip = get_input_image((int*)NULL));

        init_cpu_time();
        init_real_time();

        interactive_pso("Reducing input image.\n");

        ERN(ave_image(&fs_reduced_input_ip, full_input_ip, 2, 2));

        interactive_pso("Input image reduced (%ldms / %ldms).\n",
                        get_cpu_time(), get_real_time());
    }

    return fs_reduced_input_ip;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

const KJB_image* get_reduced_canon_image(void)
{
    const KJB_image*  full_canon_ip;

    if (fs_reduced_canon_ip == NULL)
    {
        NRN(full_canon_ip = get_canon_image((int*)NULL));

        if (is_interactive())
        {
            init_cpu_time();
            init_real_time();

            pso("Reducing canon image.\n");
        }

        ERN(ave_image(&fs_reduced_canon_ip, full_canon_ip, 2, 2));

        if (is_interactive())
        {
            pso("Canon image reduced (%ldms / %ldms).\n", get_cpu_time(),
                get_real_time());
        }
    }

    return fs_reduced_canon_ip;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static void invalidate_input_images(void)
{

    if ((fs_input_ip != NULL) || (fs_canon_ip != NULL))
    {
        verbose_pso(30, "Invalidating input images.\n");
    }

    if (fs_input_ip != NULL) fs_input_ip->read_only = FALSE;
    kjb_free_image(fs_input_ip);
    fs_input_ip = NULL;

    if (fs_canon_ip != NULL) fs_canon_ip->read_only = FALSE;
    kjb_free_image(fs_canon_ip);
    fs_canon_ip = NULL;

    if (fs_reduced_input_ip != NULL) fs_reduced_input_ip->read_only = FALSE;
    kjb_free_image(fs_reduced_input_ip);
    fs_reduced_input_ip = NULL;

    if (fs_reduced_canon_ip != NULL) fs_reduced_canon_ip->read_only = FALSE;
    kjb_free_image(fs_reduced_canon_ip);
    fs_reduced_canon_ip = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{

    if (fs_input_ip         != NULL) fs_input_ip->read_only         = FALSE;
    if (fs_canon_ip         != NULL) fs_canon_ip->read_only         = FALSE;
    if (fs_reduced_input_ip != NULL) fs_reduced_input_ip->read_only = FALSE;
    if (fs_reduced_canon_ip != NULL) fs_reduced_canon_ip->read_only = FALSE;

    kjb_free_image(fs_input_ip);
    kjb_free_image(fs_canon_ip);
    kjb_free_image(fs_reduced_input_ip);
    kjb_free_image(fs_reduced_canon_ip);
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

