/* =============================================================================
 * The functions in this file are meant to offer a number of ways to homogenize,
 * (normalize), vectors. All functions add error messages as appropriate, but
 * leave error-handling to the caller of the function. Vectors with a z-value of
 * zero will cause errors.
 *
 * Function Summary:
 *    ow_normalize_3D_vector - changes the values of the vector directly,
 *        leaving z = 1.0. Returns ERROR or NO_ERROR as appropriate.
 *    normalize_3D_vector - places the normalized values into a given Vector.
 *    Returns ERROR or NO_ERROR as appropriate.
 *
 * Author: Andrew Emmott (aemmott)
 *
 * -------------------------------------------------------------------------- */

/*Includes*/
#include <g/g_homogeneous_point.h>

#ifdef __cplusplus
extern "C" {
#endif
/* =============================================================================
 *                                 ow_normalize_3D_vector
 *
 * Normalizes the values in the given Vector.
 *
 * Changes the values of the vector directly, leaving z = 1.0. Returns ERROR or
 * NO_ERROR as appropriate.
 *
 * Returns:
 *   ERROR if the z-value = 0.0.
 *   NO_ERROR otherwise.
 *
 * Related:
 *    normalize_3D_vector
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: normalize, normalization, homogeneous
 * -------------------------------------------------------------------------- */
int ow_normalize_3D_vector
(
    Vector *vector_to_normalize /*The Vector to be changed.*/
)
{
    int z_index;
    double z_value;
    int i;

    /*Error Checking*/
    /*Bad or null vector.*/
    if (vector_to_normalize == NULL || vector_to_normalize->length < 1)
    {
        G_HOMOGENEOUS_POINT_NULL_ERROR;
        return ERROR;
    }

    z_index = vector_to_normalize->length - 1;
    z_value = vector_to_normalize->elements[z_index];
    /*z-value of 0*/
    if (fabs(z_value) < DBL_EPSILON)
    {
        G_HOMOGENEOUS_POINT_Z_ERROR;
        return ERROR;
    }
    /*End Error Checking*/

    /*Divide all other elements by the z-value.*/
    for (i = 0; i < z_index; i++)
    {
        vector_to_normalize->elements[i] /= z_value;
    }
    vector_to_normalize->elements[z_index] = 1.0;

    return NO_ERROR;
}

/* =============================================================================
 *                                 normalize_3D_vector
 *
 * Normalizes the values in a vector and places them in a new vector. 
 *
 * Takes a Vector pointer that contains the values to normalize, and a Vector**
 * that is used to allocate a new Vector of the same size. The normalized values
 * are placed inside the new Vector. If the Vector** points to an existing
 * Vector, it will be completely overwritten, including its size if necessary.
 * Returns ERROR or NO_ERROR as appropriate.
 *
 * Returns:
 *   ERROR if the z-value = 0.0.
 *   ERROR if the call to get_target_vector fails.
 *   NO_ERROR otherwise.
 *
 * Related:
 *    ow_normalize_3D_vector
 *
 * Author:
 *    Andrew Emmott (aemmott)
 *
 * Documenter:
 *    Andrew Emmott (aemmott)
 *
 * Index: normalize, normalization, homogeneous
 * -------------------------------------------------------------------------- */
int normalize_3D_vector
(
    Vector **normalized_vector_address, /*The Vector pointer address to set with the normalized values.*/
    const Vector *vector_to_normalize /*The Vector whose values to normalize.*/
)
{
    Vector *new_vector;
    int z_index;
    double z_value;
    int i;

    /*Error Checking*/
    /*Bad or null Vector.*/
    if (vector_to_normalize == NULL || vector_to_normalize->length < 1)
    {
        G_HOMOGENEOUS_POINT_NULL_ERROR;
        return ERROR;
    }

    z_index = vector_to_normalize->length - 1;
    z_value = vector_to_normalize->elements[z_index];
    
    /*z-value of 0.*/
    if (fabs(z_value) < DBL_EPSILON)
    {
        G_HOMOGENEOUS_POINT_Z_ERROR;
        return ERROR;
    }

    /*Allocation failure.*/
    if(get_target_vector(normalized_vector_address, (z_index+1))==ERROR)
    {
        G_HOMOGENEOUS_POINT_ALLOC_ERROR;
        return ERROR;
    }
    /*End Error Checking*/

    /*Insert the new values, divided by the z-value.*/
    new_vector = *normalized_vector_address;
    for (i = 0; i < z_index; i++)
    {
        new_vector->elements[i] = vector_to_normalize->elements[i]/z_value;
    }
    new_vector->elements[z_index] = 1.0;

    return NO_ERROR;
}

#ifdef __cplusplus
}
#endif
