
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "luminance.h"
#include "stat.h"
#include "illum_data.h"

/* -------------------------------------------------------------------------- */

static int print_luminance_result
(
    const char* name,
    double      result_one,
    double      result_two,
    int         dry_run
);

static int write_luminance_factor
(
    const char* directory,
    const char* file_name,
    double      luminance
);

static int get_actual_luminance
(
    const Matrix* __attribute__((unused)) dummy_data_mp,
    double*       luminance_estimate_ptr
);

static int get_ave_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
);

static int get_median_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
);

static int get_max_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
);

/* -------------------------------------------------------------------------- */

typedef struct Luminance_algorithm
{
    const char* name;
    char        config_file[ MAX_FILE_NAME_SIZE ];
    int         (*luminance_fn)(const Matrix* data_mp, double* luminance_estimate_ptr);
    double      alg_factor;
    double      factor;
    int         factor_count;
    long_double factor_sum;
    long_double factor_sum_sqr;
    int         grand_total_factor_count;
    long_double grand_total_factor_sum;
    long_double grand_total_factor_sum_sqr;
    double      error;
    int         error_count;
    long_double error_sum;
    long_double error_max;
    long_double error_sum_sqr;
    int         grand_total_error_count;
    long_double grand_total_error_sum;
    long_double grand_total_error_max;
    long_double grand_total_error_sum_sqr;
}
Luminance_algorithm;

/* -------------------------------------------------------------------------- */

static Luminance_algorithm fs_luminance_algorithms[ ] =
{
    {
        "actual-luminance", { '\0' },
        get_actual_luminance,
        1.0 , DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "ave-luminance", { '\0' },
        get_ave_luminance,
        3.5, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#ifdef DISABLE
    {
        "ave-ni-luminance", { '\0' },
        get_ave_ni_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#endif
    {
        "median-luminance", { '\0' },
        get_median_luminance,
        9.3, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#ifdef DISABLE
    {
        "median-ni-luminance", { '\0' },
        get_median_ni_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#endif
    {
        "max-luminance", { '\0' },
        get_max_luminance,
        1.75, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#ifdef DISABLE
    {
        "max-ni-luminance", { '\0' },
        get_max_ni_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "cluster-ave-luminance", { '\0' },
        get_cluster_ave_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "cluster-ave-ni-luminance", { '\0' },
        get_cluster_ave_ni_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "ave-grey-luminance", { '\0' },
        get_ave_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "ave-ni-grey-luminance", { '\0' },
        get_ave_ni_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "median-grey-luminance", { '\0' },
        get_median_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "median-ni-grey-luminance", { '\0' },
        get_median_ni_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "max-grey-luminance", { '\0' },
        get_max_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "max-ni-grey-luminance", { '\0' },
        get_max_ni_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "cluster-ave-grey-luminance", { '\0' },
        get_cluster_ave_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
    {
        "cluster-ave-ni-grey-luminance", { '\0' },
        get_cluster_ave_ni_grey_luminance,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    },
#endif
    {
        NULL, { '\0' }, NULL,
        DBL_NOT_SET, DBL_NOT_SET, 0, 0.0, 0.0, 0, 0.0, 0.0,
        DBL_NOT_SET, 0, 0.0, 0.0, 0.0, 0, 0.0, 0.0, 0.0
    }
};

static int fs_current_luminance_algorithms[ MAX_NUM_LUMINANCE_METHODS ];

/* -------------------------------------------------------------------------- */

EXTERN_C int set_luminance_options(const char* option, const char* value)
{
    Luminance_algorithm* luminance_algorithm_ptr;
    char                 lc_option[ 100 ];
    int                  result = NOT_FOUND;
    char                 file_option[ 100 ];
    int                  i;
    int                  index;
    int                  first_index;
    int                  second_index;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        BUFF_CPY(file_option, luminance_algorithm_ptr->name);
        BUFF_CAT(file_option, "-file");

        if (    (lc_option[ 0 ] == '\0')
             || (match_pattern(lc_option, file_option))
           )
        {
            if (value == NULL)
            {
                return NO_ERROR;
            }
            else if (value[ 0 ] == '?')
            {
                ERE(pso("Under construction\n"));
            }
            else if (value[ 0 ] == '\0')
            {
                if (IS_EQUAL_DBL(luminance_algorithm_ptr->factor, NOT_SET))
                {
                    ERE(pso("No luminance factor for %s is set.\n",
                            luminance_algorithm_ptr->name));
                }
                else
                {
                    ERE(pso("Luminance factor for %s is %.1f. \n",
                            luminance_algorithm_ptr->name,
                            luminance_algorithm_ptr->factor));
                    ERE(pso("Luminance factor source for %s is:%s%s\n",
                            luminance_algorithm_ptr->name,
                            (strlen(luminance_algorithm_ptr->config_file)>30) ?
                                "\n    " : " ",
                            luminance_algorithm_ptr->config_file));
                }
            }
            else
            {
                ERE(read_real_from_gamut_config_file(
                                 &(luminance_algorithm_ptr->factor),
                                 (char*)NULL, value, file_option,
                                 luminance_algorithm_ptr->config_file,
                                 sizeof(luminance_algorithm_ptr->config_file)));
            }
            result = NO_ERROR;
        }

        luminance_algorithm_ptr++;
    }

    ERE(parse_index(lc_option, MAX_NUM_LUMINANCE_METHODS,
                    &first_index, &second_index, &index, (int*)NULL));

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "luminance-algorithm"))
         || (match_pattern(lc_option, "luminance-alg"))
         || (match_pattern(lc_option, "lum-alg"))
       )
    {
        /*
        // This code must only be executed if there is an option match, and thus
        // must be copied inside the code for each option.
        */
        ERE(check_index_range(index, MAX_NUM_LUMINANCE_METHODS,
                              "Luminance method"));
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            for (i = first_index; i <= second_index; i++)
            {
                int cur_alg = fs_current_luminance_algorithms[ i ];

                if (cur_alg)
                {
                    ERE(pso("lum-alg:%d = %s\n", i + 1,
                            fs_luminance_algorithms[ cur_alg - 1 ].name));
                }
                else
                {
                    ERE(pso("lum-alg:%d = off\n", i + 1));
                }
            }
        }
        else if (value[ 0 ] == '\0')
        {
            for (i = first_index; i <= second_index; i++)
            {
                int cur_alg = fs_current_luminance_algorithms[ i ];

                if (cur_alg)
                {
                    ERE(pso("Luminance algorithm %d is %s.\n",
                            i + 1,
                            fs_luminance_algorithms[ cur_alg - 1 ].name));
                }
                else
                {
                    ERE(pso("No luminance algorithm %d.\n", i + 1));
                }
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_current_luminance_algorithms[ index ] = 0;
            }
            else
            {
                int count = 0;

                luminance_algorithm_ptr = fs_luminance_algorithms;
                result = NOT_FOUND;

                while (luminance_algorithm_ptr->name != NULL)
                {
                    if (STRCMP_EQ(value, luminance_algorithm_ptr->name))
                    {
                        /*
                        // We start counting at 1 because we are using initial zeros
                        // to indicate no value.
                        */
                        fs_current_luminance_algorithms[ index ] = count + 1;
                        result = NO_ERROR;
                        break;
                    }
                    luminance_algorithm_ptr++;
                    count++;
                }

                if (result == NOT_FOUND)
                {
                    set_error("%q is an invalid luminance algorithm.", value);
                    return ERROR;
                }
            }
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_luminance(const Matrix* data_mp)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;
    Vector* actual_illum_vp = NULL;
    double    actual_luminance;
    double    actual_luminance_estimate;
    double    alg_luminance_estimate;
    double    error;
    double    error_sqr;

    ERE(get_actual_illum(&actual_illum_vp));
    actual_luminance = sum_vector_elements(actual_illum_vp);

    while (luminance_algorithm_ptr->name != NULL)
    {
        ERE((luminance_algorithm_ptr->luminance_fn)(data_mp,
                                                    &alg_luminance_estimate));
        if (luminance_algorithm_ptr->alg_factor < 0.0)
        {
            ERE(read_real_from_gamut_config_file(
                                &(luminance_algorithm_ptr->alg_factor),
                                (const char*)NULL,
                                luminance_algorithm_ptr->name,
                                (const char*)NULL,
                                luminance_algorithm_ptr->config_file,
                                sizeof(luminance_algorithm_ptr->config_file)));
        }

        if (luminance_algorithm_ptr->alg_factor > 0.0)
        {
            actual_luminance_estimate =
                   alg_luminance_estimate * luminance_algorithm_ptr->alg_factor;
            error = ABS_OF(actual_luminance - actual_luminance_estimate);
            error /= actual_luminance;  /* Do relative error. */
            error_sqr = error*error;

            (luminance_algorithm_ptr->error) = error;
            (luminance_algorithm_ptr->error_sum) += error;

            if (error > luminance_algorithm_ptr->error_max)
            {
                luminance_algorithm_ptr->error_max = error;
            }

            (luminance_algorithm_ptr->error_sum_sqr) += error_sqr;
            (luminance_algorithm_ptr->error_count)++;

            (luminance_algorithm_ptr->grand_total_error_sum) += error;

            if (luminance_algorithm_ptr->grand_total_error_max > error)
            {
                luminance_algorithm_ptr->grand_total_error_max = error;
            }

            (luminance_algorithm_ptr->grand_total_error_sum_sqr) += error_sqr;
            (luminance_algorithm_ptr->grand_total_error_count)++;
        }

        luminance_algorithm_ptr++;
    }

    free_vector(actual_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_current_luminance_error_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        if (luminance_algorithm_ptr->error >= 0.0)
        {
            ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                        luminance_algorithm_ptr->error,
                                        DBL_NOT_SET, dry_run));
        }

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_cumulative_luminance_error_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        int count = luminance_algorithm_ptr->error_count;

        if (count > 0)
        {
            double error_stat_one;
            double error_stat_two;

            ERE(calculate_cumulative_error_stats((double)count, DBL_NOT_SET,
                               (double)luminance_algorithm_ptr->error_sum,
                               (double)luminance_algorithm_ptr->error_max,
                               (double)luminance_algorithm_ptr->error_sum_sqr,
                               &error_stat_one, &error_stat_two));

            ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                       error_stat_one, error_stat_two,
                                       dry_run));
        }

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int reset_cumulative_luminance_error(void)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        luminance_algorithm_ptr->error_count   = 0;
        luminance_algorithm_ptr->error_sum     = 0.0;
        luminance_algorithm_ptr->error_max     = 0.0;
        luminance_algorithm_ptr->error_sum_sqr = 0.0;

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_grand_total_luminance_error_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        int count = luminance_algorithm_ptr->grand_total_error_count;

        if (count > 0)
        {
            double error_stat_one;
            double error_stat_two;

            ERE(calculate_cumulative_error_stats((double)count, DBL_NOT_SET,
                     (double)luminance_algorithm_ptr->grand_total_error_sum,
                     (double)luminance_algorithm_ptr->grand_total_error_max,
                     (double)luminance_algorithm_ptr->grand_total_error_sum_sqr,
                     &error_stat_one, &error_stat_two));

            ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                       error_stat_one, error_stat_two,
                                       dry_run));
        }

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_luminance_factors(const Matrix* data_mp)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;
    Vector* actual_illum_vp = NULL;
    double    actual_luminance;
    double    alg_luminance_estimate;
    double    factor;
    double    factor_sqr;

    ERE(get_actual_illum(&actual_illum_vp));
    actual_luminance = sum_vector_elements(actual_illum_vp);

    while (luminance_algorithm_ptr->name != NULL)
    {
        ERE((luminance_algorithm_ptr->luminance_fn)(data_mp,
                                                    &alg_luminance_estimate));
        if (alg_luminance_estimate > 0.0)
        {
            factor = actual_luminance / alg_luminance_estimate;
            factor_sqr = factor*factor;

            (luminance_algorithm_ptr->factor) = factor;
            (luminance_algorithm_ptr->factor_sum) += factor;
            (luminance_algorithm_ptr->factor_sum_sqr) += factor*factor;
            (luminance_algorithm_ptr->factor_count)++;
            (luminance_algorithm_ptr->grand_total_factor_sum) += factor;
            (luminance_algorithm_ptr->grand_total_factor_sum_sqr) += factor_sqr;
            (luminance_algorithm_ptr->grand_total_factor_count)++;
        }

        luminance_algorithm_ptr++;
    }

    free_vector(actual_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_current_luminance_factor_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                   luminance_algorithm_ptr->factor,
                                   DBL_NOT_SET, dry_run));

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_cumulative_luminance_factor_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        int count = luminance_algorithm_ptr->factor_count;

        if (count > 1)
        {
            double sum_sqr        = luminance_algorithm_ptr->factor_sum_sqr;
            double sum            = luminance_algorithm_ptr->factor_sum;
            double mean           = sum / count;
            double temp           = (sum_sqr - (sum*sum)/count) / (count - 1) ;
            double stdev          = sqrt(temp);
            double relative_stdev = stdev / mean;

            ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                        relative_stdev, DBL_NOT_SET, dry_run));
        }

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int reset_cumulative_luminance_factor(void)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        luminance_algorithm_ptr->factor_count   = 0;
        luminance_algorithm_ptr->factor_sum     = 0.0;
        luminance_algorithm_ptr->factor_sum_sqr = 0.0;

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_grand_total_luminance_factor_report(int dry_run)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;

    while (luminance_algorithm_ptr->name != NULL)
    {
        int count = luminance_algorithm_ptr->grand_total_factor_count;

        if (count > 1)
        {
            double sum_sqr= luminance_algorithm_ptr->grand_total_factor_sum_sqr;
            double sum    = luminance_algorithm_ptr->grand_total_factor_sum;
            double mean           = sum / count;
            double temp           = (sum_sqr - (sum*sum)/count) / (count - 1) ;
            double stdev          = sqrt(temp);
            double relative_stdev = stdev / mean;

            ERE(print_luminance_result(luminance_algorithm_ptr->name,
                                       relative_stdev, DBL_NOT_SET, dry_run));
        }

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int print_luminance_result
(
    const char* name,
    double      result_one,
    double      result_two,
    int         dry_run
)
{

    if (result_one >= 0.0)
    {
        char name_buff[ 100 ];
        unsigned int len       = strlen(name);
        unsigned int left_over = len - 3 * (len/3);
        int room_left = 57 - left_over - len; /* Divisible by 3 */
        int i;

        /*
        // See comment in do_error_report as to why we do this ugly hack.
        */
        if (dry_run) return ERROR;

        BUFF_CPY(name_buff, name);
        char_for_char_translate(name_buff, '-', ' ');
        name_buff[ 0 ] = extended_toupper(name_buff[ 0 ]);

        ERE(pso(name_buff));
        ERE(print_blanks(stdout, left_over));

        for (i=0; i<room_left/3; i++)
        {
            ERE(pso(".  "));
        }

        if (IS_INTEGRAL_DBL(result_one))
        {
            ERE(pso("%4d     ", (int)result_one));
        }
        else
        {
            ERE(pso("%7.3f", result_one));
        }

        if (result_two >= 0.0)
        {
            if (IS_INTEGRAL_DBL(result_two))
            {
                ERE(pso(" %4d", (int)result_two));
            }
            else
            {
                ERE(pso("   %7.3f", result_two));
            }
        }

        ERE(pso("\n"));

    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_luminance(const char* make_directive, const char* directory)
{
    Luminance_algorithm* luminance_algorithm_ptr = fs_luminance_algorithms;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all");
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    ! STRCMP_EQ(lc_make_directive, "all")
         && ! STRCMP_EQ(lc_make_directive, "luminance")
       )
    {
        return NOT_FOUND;
    }

    while (luminance_algorithm_ptr->name != NULL)
    {
        long_double sum   = luminance_algorithm_ptr->factor_sum;
        int       count = luminance_algorithm_ptr->factor_count;

        if (sum > 0.0)
        {
            luminance_algorithm_ptr->alg_factor = sum / count;

            dbi(count);
            dbf(luminance_algorithm_ptr->alg_factor);

            if ((directory != NULL) && (*directory != '\0'))
            {
                ERE(write_luminance_factor(directory,
                                    luminance_algorithm_ptr->name,
                                    luminance_algorithm_ptr->factor));
            }
        }

#ifdef OLD_AND_BUGGY
        /*
        // Perhaps need to figure out cleaner way in general!
        */
        else
        {
            luminance_algorithm_ptr->alg_factor = NOT_SET;
        }
#endif

        luminance_algorithm_ptr++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int write_luminance_factor
(
    const char* directory,
    const char* file_name,
    double      factor
)
{
    FILE* fp;
    char  file_path[ MAX_FILE_NAME_SIZE ];

    ERE(kjb_mkdir(directory));

    BUFF_CPY(file_path, directory);

    if (last_char(file_path) != DIR_CHAR)
    {
        BUFF_CAT(file_path, DIR_STR);
    }

    BUFF_CAT(file_path, file_name);

    NRE(fp = kjb_fopen(file_path, "w"));
    ERE(kjb_fprintf(fp, "%7.3f\n", factor));
    ERE(kjb_fclose(fp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sum_luminance(const Matrix* data_mp, double* luminance_estimates)
{
    int (*luminance_fn)(const Matrix*, double*);
    double alg_factor;
    int    i;
    int    cur_alg;

    for (i = 0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        cur_alg = fs_current_luminance_algorithms[ i ];

        if (cur_alg)
        {

            luminance_fn = fs_luminance_algorithms[ cur_alg - 1 ].luminance_fn;

            ERE((*luminance_fn)(data_mp, &(luminance_estimates[ i ])));
            alg_factor = fs_luminance_algorithms[ cur_alg - 1 ].alg_factor;
            luminance_estimates[ i ] *= alg_factor;
        }
        else
        {
            luminance_estimates[ i ] = DBL_NOT_SET;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
static int get_actual_luminance
(
    const Matrix* __attribute__((unused)) dummy_data_mp,
    double*       luminance_estimate_ptr
)
{
    Vector* actual_illum_vp = NULL;

    ERE(get_sharp_actual_illum(&actual_illum_vp));
    *luminance_estimate_ptr = sum_vector_elements(actual_illum_vp);
    free_vector(actual_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_ave_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
)
{
    Vector* ave_vp = NULL;

    ERE(average_matrix_rows(&ave_vp, data_mp));
    *luminance_estimate_ptr = sum_vector_elements(ave_vp);
    free_vector(ave_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_median_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
)
{
    Vector* sum_vp = NULL;
    int result;

    ERE(sum_matrix_cols(&sum_vp, data_mp));

    result = ascend_sort_vector(sum_vp);

    if (result != ERROR)
    {
        *luminance_estimate_ptr = (sum_vp->elements[ (sum_vp->length)/2 ]) /3.0;
    }

    free_vector(sum_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_max_luminance
(
    const Matrix* data_mp,
    double*       luminance_estimate_ptr
)
{
    Vector* max_vp = NULL;

    ERE(get_max_matrix_col_elements(&max_vp, data_mp));
    *luminance_estimate_ptr = sum_vector_elements(max_vp);
    free_vector(max_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

