
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "canon_gamut.h"
#include "compute_error.h"
#include "compute_output.h"
#include "correlation.h"    /* Lindsay */

#include "data.h"
#include "do_best_fit.h"
#include "do_error_report.h"
#include "do_gamut_cc.h"

#include "file_data.h"
#include "gamut_3D.h"
#include "gamut_2D.h"
#include "gamut_alg_input.h"

#include "grey.h"

#include "illum_data.h"
#include "image_data.h"
#include "image_input.h"
#include "lib.h"
#include "luminance.h"

#include "output_error.h"
#include "plot.h"

#include "set.h"
#include "spectra.h"
#include "stat.h"
#include "synthetic_data.h"

#ifdef TEMPLATE
#    include "template.h"
#endif

#ifdef MULTI_TEMPLATE
#    include "multi_template.h"
#endif

/* For declared function pointers. Generally harmless. */
#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------- */

#define MAX_NUM_SET_FN  100

/* -------------------------------------------------------------------------- */

static int process_option(const char* option, const char* value);

/* -------------------------------------------------------------------------- */

int process_set_command(const char* arg)
{

    return process_option_string(arg, process_option);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int process_option(const char* option, const char* value)
{
    int  result = NOT_FOUND;
    int  temp_result;
    int  (*set_fn_list[ MAX_NUM_SET_FN ])(const char* opt, const char* val);

    set_fn_list[ 0 ] = kjb_l_set;
    set_fn_list[ 1 ] = kjb_m_set;
    set_fn_list[ 2 ] = kjb_n_set;
    set_fn_list[ 3 ] = kjb_h_set;
    set_fn_list[ 4 ] = kjb_c_set;
    set_fn_list[ 5 ] = kjb_s_set;
    set_fn_list[ 6 ] = set_qhull_options;
    set_fn_list[ 7 ] = set_video_options;
    set_fn_list[ 8 ] = t1_set_segmentation_options;
    set_fn_list[ 9 ] = kjb_i_set;
    ERE(temp_result = call_set_fn(10, set_fn_list, "Low Level Options",
                                      option, value));

    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_gamut_config_dir;
    set_fn_list[ 1 ] = set_spectra_options;
    set_fn_list[ 2 ] = set_illum_data_files;
    set_fn_list[ 3 ] = set_grey_file;
    set_fn_list[ 4 ] = set_luminance_options;
    ERE(temp_result = call_set_fn(5, set_fn_list,
                                      "Configuration data files",
                                       option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_synthetic_data_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Synthetic data options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_data_files;
    ERE(temp_result = call_set_fn(1, set_fn_list, "RGB Data Files",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_image_files;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Image Data Files",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_data_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Data Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_image_pp_options;
    ERE(temp_result =
        call_set_fn(1, set_fn_list,
                     "Image Preprocessing Options (see also low level options)",
                        option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_output_image_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Image Output Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_algorithm_options;
    set_fn_list[ 1 ] = set_sharp_file;
    set_fn_list[ 2 ] = set_best_fit_options;
    ERE(temp_result = call_set_fn(3, set_fn_list, "Algorithm Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_gamut_options;
    set_fn_list[ 1 ] = set_gamut_alg_input_options;
    set_fn_list[ 2 ] = set_2D_gamut_options;
    set_fn_list[ 3 ] = set_3D_gamut_options;
    set_fn_list[ 4 ] = set_canon_gamut_options;
    ERE(temp_result = call_set_fn(5, set_fn_list, "Gamut Mapping Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    /* Lindsay: Colour by Correlation - Feb 17/98 */
    set_fn_list[ 0 ] = set_correlation_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Correlation Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;
    /* End Lindsay: Feb 17/98 */

#ifdef TEMPLATE
    set_fn_list[ 0 ] = set_template_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Template Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;
#endif

#ifdef MULTI_TEMPLATE
    set_fn_list[ 0 ] = set_multi_template_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Multi template Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;
#endif

    set_fn_list[ 0 ] = set_compute_error_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Error Computing Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_stat_options;
    set_fn_list[ 1 ] = set_report_options;
    ERE(temp_result = call_set_fn(2, set_fn_list, "Report Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_output_error_options;
    ERE(temp_result = call_set_fn(1, set_fn_list, "Error Output Options",
                                      option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    set_fn_list[ 0 ] = set_plot_options;
    ERE(temp_result = call_set_fn(1, set_fn_list,
                              "Plot Options (only relavent for interative use)",
                               option, value));
    if (temp_result == NO_ERROR) result = NO_ERROR;

    if (result == NOT_FOUND)
    {
        set_error("%q is an invalid option.", option);
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef __cplusplus
}
#endif

