
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "do_error_report.h"
#include "initialize_error.h"
#include "stat.h"

/* ======================================================================---- */

typedef enum Error_stat
{
    INVALID_ERROR_STAT = ERROR,
    NO_ERROR_STAT,
    COUNT_ERROR_STAT,
    DEFAULT_TO_DULL_COUNT_ERROR_STAT,
    MAX_ERROR_STAT,
    MEAN_ERROR_STAT,
    RMS_ERROR_STAT,
    STDEV_ERROR_STAT
}
Error_stat;

/* ======================================================================---- */

static Error_stat fs_error_stat_one_request = RMS_ERROR_STAT;
static Error_stat fs_error_stat_two_request = MAX_ERROR_STAT;

static int fs_default_to_dull_error = FALSE;

static int                    fs_cumulative_input_count = 0;

static Colour_constancy_error fs_cumulative_error_count;
static Colour_constancy_error fs_cumulative_default_to_dull_count;
static Colour_constancy_error fs_cumulative_error_sum;
static Colour_constancy_error fs_cumulative_error_max;
static Colour_constancy_error fs_cumulative_error_sum_sqr;

static Input_data_error       fs_cumulative_input_data_error_count;
static Input_data_error       fs_cumulative_input_data_default_to_dull_count;
static Input_data_error       fs_cumulative_input_data_error_sum;
static Input_data_error       fs_cumulative_input_data_error_max;
static Input_data_error       fs_cumulative_input_data_error_sum_sqr;

static int                    fs_grand_input_count = 0;

static Colour_constancy_error fs_grand_error_count;
static Colour_constancy_error fs_grand_default_to_dull_count;
static Colour_constancy_error fs_grand_error_sum;
static Colour_constancy_error fs_grand_error_max;
static Colour_constancy_error fs_grand_error_sum_sqr;

static Input_data_error       fs_grand_input_data_error_count;
static Input_data_error       fs_grand_input_data_default_to_dull_count;
static Input_data_error       fs_grand_input_data_error_sum;
static Input_data_error       fs_grand_input_data_error_max;
static Input_data_error       fs_grand_input_data_error_sum_sqr;

/* ======================================================================---- */

static Error_stat parse_error_stat_request(const char* request);

static int get_error_stat_heading_str
(
    Error_stat error_stat_option_value,
    char*      buff,
    size_t     buff_size
);

static int print_error_stat_option
(
    const char* id_str,
    Error_stat  error_stat_option_value
);

static int update_cumulative_input_data_error_by_sharp
(
    Sharp_method      sharp_method,
    Input_data_error* error_ptr,
    Input_data_error* cum_input_data_error_count_ptr,
    Input_data_error* cum_input_data_default_to_dull_count_ptr,
    Input_data_error* cum_input_data_error_sum_ptr,
    Input_data_error* cum_input_data_error_max_ptr,
    Input_data_error* cum_input_data_error_sum_sqr_ptr
);

static int update_cumulative_gamut_input_data_error
(
    Gamut_input_data_error* error_ptr,
    Gamut_input_data_error* dull_error_ptr,
    Gamut_input_data_error* count_ptr,
    Gamut_input_data_error* default_to_dull_count_ptr,
    Gamut_input_data_error* cumulative_error_sum_ptr,
    Gamut_input_data_error* max_ptr,
    Gamut_input_data_error* sum_sqr_ptr
);

static int update_cumulative_error_by_sharp
(
    Sharp_method            sharp_method,
    Colour_constancy_error* error_ptr,
    Colour_constancy_error* cum_error_count_ptr,
    Colour_constancy_error* cum_default_to_dull_count_ptr,
    Colour_constancy_error* cum_error_sum_ptr,
    Colour_constancy_error* cum_error_max_ptr,
    Colour_constancy_error* cum_error_sum_sqr_ptr
);

static int update_cumulative_error_by_algorithm
(
    Error_by_algorithm* error_ptr,
    Error_by_algorithm* dull_error_ptr,
    Error_by_algorithm* count_ptr,
    Error_by_algorithm* default_to_dull_count_ptr,
    Error_by_algorithm* cumulative_error_sum_ptr,
    Error_by_algorithm* max_ptr,
    Error_by_algorithm* sum_sqr_ptr
);

static int update_cumulative_best_fit_error
(
    Best_fit_error* error_ptr,
    Best_fit_error* dull_error_ptr,
    Best_fit_error* count_ptr,
    Best_fit_error* default_to_dull_count_ptr,
    Best_fit_error* cumulative_error_sum_ptr,
    Best_fit_error* max_ptr,
    Best_fit_error* sum_sqr_ptr
);

static int update_cumulative_standard_error
(
    Standard_error* error_ptr,
    Standard_error* dull_error_ptr,
    Standard_error* count_ptr,
    Standard_error* default_to_dull_count_ptr,
    Standard_error* cumulative_error_sum_ptr,
    Standard_error* max_ptr,
    Standard_error* sum_sqr_ptr
);

static int update_cumulative_gamut_error
(
    Gamut_error* error_ptr,
    Gamut_error* dull_error_ptr,
    Gamut_error* count_ptr,
    Gamut_error* default_to_dull_count_ptr,
    Gamut_error* cumulative_error_sum_ptr,
    Gamut_error* max_ptr,
    Gamut_error* sum_sqr_ptr
);

static int update_cumulative_correlation_error
(
    Correlation_error* count_ptr,
    Correlation_error* default_to_dull_count_ptr,
    Correlation_error* cumulative_error_sum_ptr,
    Correlation_error* max_ptr,
    Correlation_error* sum_sqr_ptr,
    Correlation_error* error_ptr,
    Correlation_error* dull_error_ptr
);

#ifdef SIMPLE_TEMPLATE
static int update_cumulative_simple_template_error
(
    Simple_template_error* error_ptr,
    Simple_template_error* dull_error_ptr,
    Simple_template_error* count_ptr,
    Simple_template_error* default_to_dull_count_ptr,
    Simple_template_error* cumulative_error_sum_ptr,
    Simple_template_error* max_ptr,
    Simple_template_error* sum_sqr_ptr
);

#endif

#ifdef TEMPLATE
static int update_cumulative_template_error
(
    Template_error* error_ptr,
    Template_error* dull_error_ptr,
    Template_error* count_ptr,
    Template_error* default_to_dull_count_ptr,
    Template_error* cumulative_error_sum_ptr,
    Template_error* max_ptr,
    Template_error* sum_sqr_ptr
);

#endif

#ifdef MULTI_TEMPLATE
static int update_cumulative_multi_template_error
(
    Multi_template_error* error_ptr,
    Multi_template_error* dull_error_ptr,
    Multi_template_error* count_ptr,
    Multi_template_error* default_to_dull_count_ptr,
    Multi_template_error* cumulative_error_sum_ptr,
    Multi_template_error* max_ptr,
    Multi_template_error* sum_sqr_ptr
);

#endif

static int update_cumulative_error_item
(
    double  error,
    double  dull_error,
    double* count_ptr,
    double* default_to_dull_count_ptr,
    double* cumulative_error_sum_ptr,
    double* max_ptr,
    double* sum_sqr_ptr
);

static int calculate_cumulative_gamut_input_data_error
(
    Gamut_input_data_error* count_ptr,
    Gamut_input_data_error* default_to_dull_count_ptr,
    Gamut_input_data_error* cumulative_error_sum_ptr,
    Gamut_input_data_error* max_ptr,
    Gamut_input_data_error* sum_sqr_ptr,
    Gamut_input_data_error* error_stat_one_ptr,
    Gamut_input_data_error* error_stat_two_ptr
);

static int calculate_cumulative_error_by_sharp
(
    Error_by_sharp* count_ptr,
    Error_by_sharp* default_to_dull_count_ptr,
    Error_by_sharp* cumulative_error_sum_ptr,
    Error_by_sharp* max_ptr,
    Error_by_sharp* sum_sqr_ptr,
    Error_by_sharp* error_stat_one_ptr,
    Error_by_sharp* error_stat_two_ptr
);

static int calculate_cumulative_error_by_algorithm
(
    Error_by_algorithm* count_ptr,
    Error_by_algorithm* default_to_dull_count_ptr,
    Error_by_algorithm* cumulative_error_sum_ptr,
    Error_by_algorithm* max_ptr,
    Error_by_algorithm* sum_sqr_ptr,
    Error_by_algorithm* error_stat_one_ptr,
    Error_by_algorithm* error_stat_two_ptr
);

static int calculate_best_fit_cumulative_error
(
    Best_fit_error* count_ptr,
    Best_fit_error* default_to_dull_count_ptr,
    Best_fit_error* cumulative_error_sum_ptr,
    Best_fit_error* max_ptr,
    Best_fit_error* sum_sqr_ptr,
    Best_fit_error* error_stat_one_ptr,
    Best_fit_error* error_stat_two_ptr
);

static int calculate_standard_cumulative_error
(
    Standard_error* count_ptr,
    Standard_error* default_to_dull_count_ptr,
    Standard_error* cumulative_error_sum_ptr,
    Standard_error* max_ptr,
    Standard_error* sum_sqr_ptr,
    Standard_error* error_stat_one_ptr,
    Standard_error* error_stat_two_ptr
);

static int calculate_gamut_cumulative_error
(
    Gamut_error* count_ptr,
    Gamut_error* default_to_dull_count_ptr,
    Gamut_error* cumulative_error_sum_ptr,
    Gamut_error* max_ptr,
    Gamut_error* sum_sqr_ptr,
    Gamut_error* error_stat_one_ptr,
    Gamut_error* error_stat_two_ptr
);

static int calculate_correlation_cumulative_error
(
    Correlation_error* count_ptr,
    Correlation_error* default_to_dull_count_ptr,
    Correlation_error* cumulative_error_sum_ptr,
    Correlation_error* max_ptr,
    Correlation_error* sum_sqr_ptr,
    Correlation_error* error_stat_one_ptr,
    Correlation_error* error_stat_two_ptr
);

#ifdef SIMPLE_TEMPLATE
static int calculate_simple_template_cumulative_error
(
    Simple_template_error* count_ptr,
    Simple_template_error* default_to_dull_count_ptr,
    Simple_template_error* cumulative_error_sum_ptr,
    Simple_template_error* max_ptr,
    Simple_template_error* sum_sqr_ptr,
    Simple_template_error* error_stat_one_ptr,
    Simple_template_error* error_stat_two_ptr
);

#endif

#ifdef TEMPLATE
static int calculate_template_cumulative_error
(
    Template_error* count_ptr,
    Template_error* default_to_dull_count_ptr,
    Template_error* cumulative_error_sum_ptr,
    Template_error* max_ptr,
    Template_error* sum_sqr_ptr,
    Template_error* error_stat_one_ptr,
    Template_error* error_stat_two_ptr
);

#endif

#ifdef MULTI_TEMPLATE
static int calculate_multi_template_cumulative_error
(
    Multi_template_error* count_ptr,
    Multi_template_error* default_to_dull_count_ptr,
    Multi_template_error* cumulative_error_sum_ptr,
    Multi_template_error* max_ptr,
    Multi_template_error* sum_sqr_ptr,
    Multi_template_error* error_stat_one_ptr,
    Multi_template_error* error_stat_two_ptr
);

#endif

static int calculate_cumulative_error_stat_item
(
    double     count,
    double     default_to_dull_count,
    double     error_sum,
    double     error_max,
    double     error_sum_sqr,
    Error_stat error_stat_request,
    double*    error_stat_ptr
);

/* ======================================================================---- */

EXTERN_C int set_stat_options(const char* option, const char* value)
{
    char       lc_option[ 100 ];
    Error_stat temp_error_stat;
    int        temp_boolean_value;
    int        result           = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "error-stat-one"))
         || (match_pattern(lc_option, "error-one-stat"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(print_error_stat_option("first", fs_error_stat_one_request));
        }
        else
        {
            temp_error_stat = parse_error_stat_request(value);

            if (temp_error_stat == INVALID_ERROR_STAT) return ERROR;

            fs_error_stat_one_request = temp_error_stat;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "error-stat-two"))
         || (match_pattern(lc_option, "error-two-stat"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(print_error_stat_option("second", fs_error_stat_two_request));
        }
        else
        {
            temp_error_stat = parse_error_stat_request(value);

            if (temp_error_stat == INVALID_ERROR_STAT) return ERROR;

            fs_error_stat_two_request = temp_error_stat;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "default-dull"))
         || (match_pattern(lc_option, "default-to-dull"))
         || (match_pattern(lc_option, "default-to-dull-error"))
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("default-to-dull-error = %s\n",
                    fs_default_to_dull_error ? "t" : "f"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Unavaible sharp results %s.\n",
                     fs_default_to_dull_error ? "default to dull results"
                                           : "are not replaced"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_default_to_dull_error = temp_boolean_value;
        }

        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static Error_stat parse_error_stat_request(const char* request)
{

    if (is_no_value_word(request))
    {
        return NO_ERROR_STAT;
    }
    else if (IC_STRCMP_EQ(request, "count"))
    {
        return COUNT_ERROR_STAT;
    }
    else if (    (IC_STRCMP_EQ(request, "default-to-dull"))
              || (IC_STRCMP_EQ(request, "default-to-dull-count"))
            )
    {
        return DEFAULT_TO_DULL_COUNT_ERROR_STAT;
    }
    else if (IC_STRCMP_EQ(request, "rms"))
    {
        return RMS_ERROR_STAT;
    }
    else if (    (IC_STRCMP_EQ(request, "mean"))
              || (IC_STRCMP_EQ(request, "ave"))
              || (IC_STRCMP_EQ(request, "average"))
            )
    {
        return MEAN_ERROR_STAT;
    }
    else if (IC_STRCMP_EQ(request, "max"))
    {
        return MAX_ERROR_STAT;
    }
    else if (IC_STRCMP_EQ(request, "stdev"))
    {
        return STDEV_ERROR_STAT;
    }
    else
    {
        set_error("%q is an invalid error statistic designation.", request);
        return INVALID_ERROR_STAT;
    }

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int print_error_stat_option
(
    const char* id_str,
    Error_stat  error_stat_option_value
)
{
    const char* error_stat_option_str;

    if (error_stat_option_value == NO_ERROR_STAT)
    {
        error_stat_option_str = "not used";
    }
    else if (error_stat_option_value == COUNT_ERROR_STAT)
    {
        error_stat_option_str = "data count";
    }
    else if (error_stat_option_value == DEFAULT_TO_DULL_COUNT_ERROR_STAT)
    {
        error_stat_option_str = "default to dull count";
    }
    else if (error_stat_option_value == RMS_ERROR_STAT)
    {
        error_stat_option_str = "rms";
    }
    else if (error_stat_option_value == MEAN_ERROR_STAT)
    {
        error_stat_option_str = "the mean";
    }
    else if (error_stat_option_value == MAX_ERROR_STAT)
    {
        error_stat_option_str = "the max";
    }
    else if (error_stat_option_value == STDEV_ERROR_STAT)
    {
        error_stat_option_str = "the stdev";
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    ERE(pso("The %s error statistic request is %s.\n",
            id_str, error_stat_option_str));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_error_stat_heading_str
(
    Error_stat error_stat_option_value,
    char*      buff,
    size_t     buff_size
)
{
    const char* error_stat_option_str;

    if (error_stat_option_value == NO_ERROR_STAT)
    {
        error_stat_option_str = "";
    }
    else if (error_stat_option_value == COUNT_ERROR_STAT)
    {
        error_stat_option_str = "Count";
    }
    else if (error_stat_option_value == DEFAULT_TO_DULL_COUNT_ERROR_STAT)
    {
        error_stat_option_str = "D-to-D";
    }
    else if (error_stat_option_value == RMS_ERROR_STAT)
    {
        error_stat_option_str = "RMS";
    }
    else if (error_stat_option_value == MEAN_ERROR_STAT)
    {
        error_stat_option_str = "Mean";
    }
    else if (error_stat_option_value == MAX_ERROR_STAT)
    {
        error_stat_option_str = "Max";
    }
    else if (error_stat_option_value == STDEV_ERROR_STAT)
    {
        error_stat_option_str = "Stdev";
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    kjb_strncpy(buff, error_stat_option_str, buff_size);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_grand_total_error(void)
{

    fs_grand_input_count = 0;
    initialize_cumulative_error();
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void initialize_cumulative_error(void)
{

    fs_cumulative_input_count = 0;

    initialize_error(&fs_cumulative_error_count, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&fs_cumulative_default_to_dull_count, 0.0);
    initialize_error(&fs_cumulative_error_sum, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&fs_cumulative_error_sum_sqr, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&fs_cumulative_error_max, DBL_HALF_MOST_NEGATIVE);

    initialize_input_data_error(&fs_cumulative_input_data_error_count,
                                DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&fs_cumulative_input_data_default_to_dull_count,
                                0.0);
    initialize_input_data_error(&fs_cumulative_input_data_error_sum,
                                DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&fs_cumulative_input_data_error_sum_sqr,
                                DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&fs_cumulative_input_data_error_max,
                                DBL_HALF_MOST_NEGATIVE);

    if (fs_grand_input_count == 0)
    {
        initialize_error(&fs_grand_error_count, DBL_HALF_MOST_NEGATIVE);
        initialize_error(&fs_grand_default_to_dull_count, 0.0);
        initialize_error(&fs_grand_error_sum, DBL_HALF_MOST_NEGATIVE);
        initialize_error(&fs_grand_error_sum_sqr, DBL_HALF_MOST_NEGATIVE);
        initialize_error(&fs_grand_error_max, DBL_HALF_MOST_NEGATIVE);

        initialize_input_data_error(&fs_grand_input_data_error_count,
                                    DBL_HALF_MOST_NEGATIVE);
        initialize_input_data_error(&fs_grand_input_data_default_to_dull_count,
                                    0.0);
        initialize_input_data_error(&fs_grand_input_data_error_sum,
                                    DBL_HALF_MOST_NEGATIVE);
        initialize_input_data_error(&fs_grand_input_data_error_sum_sqr,
                                    DBL_HALF_MOST_NEGATIVE);
        initialize_input_data_error(&fs_grand_input_data_error_max,
                                    DBL_HALF_MOST_NEGATIVE);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int update_cumulative_input_data_error
(
    int*              sharp_methods,
    Input_data_error* error_ptr
)
{
    int sharp_method;

    sharp_method = DONT_SHARPEN;

    while (sharp_method < NUM_SHARP_METHODS)
    {
        if (sharp_methods[ sharp_method ])
        {
            ERE(update_cumulative_input_data_error_by_sharp((Sharp_method)sharp_method,
                                                            error_ptr,
                                   &fs_cumulative_input_data_error_count,
                                   &fs_cumulative_input_data_default_to_dull_count,
                                   &fs_cumulative_input_data_error_sum,
                                   &fs_cumulative_input_data_error_max,
                                   &fs_cumulative_input_data_error_sum_sqr));

            ERE(update_cumulative_input_data_error_by_sharp((Sharp_method)sharp_method,
                                                            error_ptr,
                                      &fs_grand_input_data_error_count,
                                      &fs_grand_input_data_default_to_dull_count,
                                      &fs_grand_input_data_error_sum,
                                      &fs_grand_input_data_error_max,
                                      &fs_grand_input_data_error_sum_sqr));
        }
        sharp_method++;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_input_data_error_by_sharp
(
    Sharp_method      sharp_method,
    Input_data_error* error_ptr,
    Input_data_error* cum_input_data_error_count_ptr,
    Input_data_error* cum_input_data_default_to_dull_count_ptr,
    Input_data_error* cum_input_data_error_sum_ptr,
    Input_data_error* cum_input_data_error_max_ptr,
    Input_data_error* cum_input_data_error_sum_sqr_ptr
)
{
    int j = sharp_method;
    int i;

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(update_cumulative_gamut_input_data_error(
              &(error_ptr->sharp[ j ].gamut_2D[ i ]),
              &(error_ptr->sharp[ DONT_SHARPEN ].gamut_2D[ i ]),
              &(cum_input_data_error_count_ptr->sharp[ j ].gamut_2D[ i ]),
              &(cum_input_data_default_to_dull_count_ptr->sharp[ j ].gamut_2D[ i ]),
              &(cum_input_data_error_sum_ptr->sharp[ j ].gamut_2D[ i ]),
              &(cum_input_data_error_max_ptr->sharp[ j ].gamut_2D[ i ]),
              &(cum_input_data_error_sum_sqr_ptr->sharp[ j ].gamut_2D[ i ])));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(update_cumulative_gamut_input_data_error(
              &(error_ptr->sharp[ j ].gamut_3D[ i ]),
              &(error_ptr->sharp[ DONT_SHARPEN ].gamut_3D[ i ]),
              &(cum_input_data_error_count_ptr->sharp[ j ].gamut_3D[ i ]),
              &(cum_input_data_default_to_dull_count_ptr->sharp[ j ].gamut_3D[ i ]),
              &(cum_input_data_error_sum_ptr->sharp[ j ].gamut_3D[ i ]),
              &(cum_input_data_error_max_ptr->sharp[ j ].gamut_3D[ i ]),
              &(cum_input_data_error_sum_sqr_ptr->sharp[ j ].gamut_3D[ i ])));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_gamut_input_data_error
(
    Gamut_input_data_error* error_ptr,
    Gamut_input_data_error* dull_error_ptr,
    Gamut_input_data_error* count_ptr,
    Gamut_input_data_error* default_to_dull_count_ptr,
    Gamut_input_data_error* cumulative_error_sum_ptr,
    Gamut_input_data_error* max_ptr,
    Gamut_input_data_error* sum_sqr_ptr
)
{

    ERE(update_cumulative_error_item(
                error_ptr->absolute_pixel_error,
                dull_error_ptr->absolute_pixel_error,
                &(count_ptr->absolute_pixel_error),
                &(default_to_dull_count_ptr->absolute_pixel_error),
                &(cumulative_error_sum_ptr->absolute_pixel_error),
                &(max_ptr->absolute_pixel_error),
                &(sum_sqr_ptr->absolute_pixel_error)));

    ERE(update_cumulative_error_item(
                error_ptr->relative_pixel_error,
                dull_error_ptr->relative_pixel_error,
                &(count_ptr->relative_pixel_error),
                &(default_to_dull_count_ptr->relative_pixel_error),
                &(cumulative_error_sum_ptr->relative_pixel_error),
                &(max_ptr->relative_pixel_error),
                &(sum_sqr_ptr->relative_pixel_error)));

    ERE(update_cumulative_error_item(
                        error_ptr->diag_error,
                        dull_error_ptr->diag_error,
                        &(count_ptr->diag_error),
                        &(default_to_dull_count_ptr->diag_error),
                        &(cumulative_error_sum_ptr->diag_error),
                        &(max_ptr->diag_error),
                        &(sum_sqr_ptr->diag_error)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int update_cumulative_error
(
    int*                    sharp_methods,
    Colour_constancy_error* error_ptr
)
{
    int sharp_method;

    sharp_method = DONT_SHARPEN;

    while (sharp_method < NUM_SHARP_METHODS)
    {
        if (sharp_methods[ sharp_method ])
        {
            ERE(update_cumulative_error_by_sharp((Sharp_method)sharp_method,
                                                 error_ptr,
                                                 &fs_cumulative_error_count,
                                             &fs_cumulative_default_to_dull_count,
                                                 &fs_cumulative_error_sum,
                                                 &fs_cumulative_error_max,
                                                 &fs_cumulative_error_sum_sqr));

            ERE(update_cumulative_error_by_sharp((Sharp_method)sharp_method,
                                                 error_ptr,
                                                 &fs_grand_error_count,
                                                 &fs_grand_default_to_dull_count,
                                                 &fs_grand_error_sum,
                                                 &fs_grand_error_max,
                                                 &fs_grand_error_sum_sqr));
        }
        sharp_method++;
    }

    fs_grand_input_count++;
    fs_cumulative_input_count++;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_error_by_sharp
(
    Sharp_method            sharp_method,
    Colour_constancy_error* error_ptr,
    Colour_constancy_error* cum_error_count_ptr,
    Colour_constancy_error* cum_default_to_dull_count_ptr,
    Colour_constancy_error* cum_error_sum_ptr,
    Colour_constancy_error* cum_error_max_ptr,
    Colour_constancy_error* cum_error_sum_sqr_ptr
)
{
    int i = sharp_method;

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_angular.sharp[ i ]),
            &(error_ptr->illum_angular.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_angular.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_angular.sharp[ i ]),
            &(cum_error_sum_ptr->illum_angular.sharp[ i ]),
            &(cum_error_max_ptr->illum_angular.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_angular.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_sum_chrom.sharp[ i ]),
            &(error_ptr->illum_sum_chrom.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_sum_chrom.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_sum_chrom.sharp[ i ]),
            &(cum_error_sum_ptr->illum_sum_chrom.sharp[ i ]),
            &(cum_error_max_ptr->illum_sum_chrom.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_sum_chrom.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_rgb.sharp[ i ]),
            &(error_ptr->illum_rgb.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_rgb.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_rgb.sharp[ i ]),
            &(cum_error_sum_ptr->illum_rgb.sharp[ i ]),
            &(cum_error_max_ptr->illum_rgb.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_rgb.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_r_div_sum.sharp[ i ]),
            &(error_ptr->illum_r_div_sum.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_r_div_sum.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_r_div_sum.sharp[ i ]),
            &(cum_error_sum_ptr->illum_r_div_sum.sharp[ i ]),
            &(cum_error_max_ptr->illum_r_div_sum.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_r_div_sum.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_g_div_sum.sharp[ i ]),
            &(error_ptr->illum_g_div_sum.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_g_div_sum.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_g_div_sum.sharp[ i ]),
            &(cum_error_sum_ptr->illum_g_div_sum.sharp[ i ]),
            &(cum_error_max_ptr->illum_g_div_sum.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_g_div_sum.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_sum_lum.sharp[ i ]),
            &(error_ptr->illum_sum_lum.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_sum_lum.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_sum_lum.sharp[ i ]),
            &(cum_error_sum_ptr->illum_sum_lum.sharp[ i ]),
            &(cum_error_max_ptr->illum_sum_lum.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_sum_lum.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->illum_lab.sharp[ i ]),
            &(error_ptr->illum_lab.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->illum_lab.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->illum_lab.sharp[ i ]),
            &(cum_error_sum_ptr->illum_lab.sharp[ i ]),
            &(cum_error_max_ptr->illum_lab.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->illum_lab.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->world_map_blue_chrom_rms.sharp[ i ]),
            &(error_ptr->world_map_blue_chrom_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->world_map_blue_chrom_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->world_map_blue_chrom_rms.sharp[i]),
            &(cum_error_sum_ptr->world_map_blue_chrom_rms.sharp[ i ]),
            &(cum_error_max_ptr->world_map_blue_chrom_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->world_map_blue_chrom_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->world_map_sum_chrom_rms.sharp[ i ]),
            &(error_ptr->world_map_sum_chrom_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->world_map_sum_chrom_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->world_map_sum_chrom_rms.sharp[i]),
            &(cum_error_sum_ptr->world_map_sum_chrom_rms.sharp[ i ]),
            &(cum_error_max_ptr->world_map_sum_chrom_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->world_map_sum_chrom_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->world_map_3D_rms.sharp[ i ]),
            &(error_ptr->world_map_3D_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->world_map_3D_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->world_map_3D_rms.sharp[ i ]),
            &(cum_error_sum_ptr->world_map_3D_rms.sharp[ i ]),
            &(cum_error_max_ptr->world_map_3D_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->world_map_3D_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->world_map_3D_lab.sharp[ i ]),
            &(error_ptr->world_map_3D_lab.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->world_map_3D_lab.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->world_map_3D_lab.sharp[ i ]),
            &(cum_error_sum_ptr->world_map_3D_lab.sharp[ i ]),
            &(cum_error_max_ptr->world_map_3D_lab.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->world_map_3D_lab.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->scene_map_blue_chrom_rms.sharp[ i ]),
            &(error_ptr->scene_map_blue_chrom_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->scene_map_blue_chrom_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->scene_map_blue_chrom_rms.sharp[i]),
            &(cum_error_sum_ptr->scene_map_blue_chrom_rms.sharp[ i ]),
            &(cum_error_max_ptr->scene_map_blue_chrom_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->scene_map_blue_chrom_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->scene_map_sum_chrom_rms.sharp[ i ]),
            &(error_ptr->scene_map_sum_chrom_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->scene_map_sum_chrom_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->scene_map_sum_chrom_rms.sharp[i]),
            &(cum_error_sum_ptr->scene_map_sum_chrom_rms.sharp[ i ]),
            &(cum_error_max_ptr->scene_map_sum_chrom_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->scene_map_sum_chrom_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->scene_map_3D_rms.sharp[ i ]),
            &(error_ptr->scene_map_3D_rms.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->scene_map_3D_rms.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->scene_map_3D_rms.sharp[ i ]),
            &(cum_error_sum_ptr->scene_map_3D_rms.sharp[ i ]),
            &(cum_error_max_ptr->scene_map_3D_rms.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->scene_map_3D_rms.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->scene_map_3D_lab.sharp[ i ]),
            &(error_ptr->scene_map_3D_lab.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->scene_map_3D_lab.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->scene_map_3D_lab.sharp[ i ]),
            &(cum_error_sum_ptr->scene_map_3D_lab.sharp[ i ]),
            &(cum_error_max_ptr->scene_map_3D_lab.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->scene_map_3D_lab.sharp[ i ])));

    ERE(update_cumulative_error_by_algorithm(
            &(error_ptr->cpu_time.sharp[ i ]),
            &(error_ptr->cpu_time.sharp[ DONT_SHARPEN ]),
            &(cum_error_count_ptr->cpu_time.sharp[ i ]),
            &(cum_default_to_dull_count_ptr->cpu_time.sharp[ i ]),
            &(cum_error_sum_ptr->cpu_time.sharp[ i ]),
            &(cum_error_max_ptr->cpu_time.sharp[ i ]),
            &(cum_error_sum_sqr_ptr->cpu_time.sharp[ i ])));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_error_by_algorithm
(
    Error_by_algorithm* error_ptr,
    Error_by_algorithm* dull_error_ptr,
    Error_by_algorithm* count_ptr,
    Error_by_algorithm* default_to_dull_count_ptr,
    Error_by_algorithm* cumulative_error_sum_ptr,
    Error_by_algorithm* max_ptr,
    Error_by_algorithm* sum_sqr_ptr
)
{
    int i;

    ERE(update_cumulative_best_fit_error(&(error_ptr->best_fit),
                                         &(dull_error_ptr->best_fit),
                             &(count_ptr->best_fit),
                             &(default_to_dull_count_ptr->best_fit),
                             &(cumulative_error_sum_ptr->best_fit),
                             &(max_ptr->best_fit),
                             &(sum_sqr_ptr->best_fit)));

    ERE(update_cumulative_standard_error(&(error_ptr->standard),
                                         &(dull_error_ptr->standard),
                                 &(count_ptr->standard),
                                 &(default_to_dull_count_ptr->standard),
                                 &(cumulative_error_sum_ptr->standard),
                                 &(max_ptr->standard),
                                 &(sum_sqr_ptr->standard)));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(update_cumulative_gamut_error(&(error_ptr->gamut_2D[ i ]),
                                          &(dull_error_ptr->gamut_2D[ i ]),
                                          &(count_ptr->gamut_2D[ i ]),
                                 &(default_to_dull_count_ptr->gamut_2D[ i ]),
                                 &(cumulative_error_sum_ptr->gamut_2D[ i ]),
                                 &(max_ptr->gamut_2D[ i ]),
                                 &(sum_sqr_ptr->gamut_2D[ i ])));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(update_cumulative_gamut_error(&(error_ptr->gamut_3D[ i ]),
                                  &(dull_error_ptr->gamut_3D[ i ]),
                                  &(count_ptr->gamut_3D[ i ]),
                                  &(default_to_dull_count_ptr->gamut_3D[ i ]),
                                  &(cumulative_error_sum_ptr->gamut_3D[ i ]),
                                  &(max_ptr->gamut_3D[ i ]),
                                  &(sum_sqr_ptr->gamut_3D[ i ])));
    }

    /* Lindsay: Feb 17/98 */
    ERE(update_cumulative_correlation_error(&(error_ptr->correlation),
                                            &(dull_error_ptr->correlation),
                           &(count_ptr->correlation),
                           &(default_to_dull_count_ptr->correlation),
                           &(cumulative_error_sum_ptr->correlation),
                           &(max_ptr->correlation),
                           &(sum_sqr_ptr->correlation)));
    /* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
    ERE(update_cumulative_simple_template_error(&(error_ptr->simple_template),
                                         &(dull_error_ptr->simple_template),
                             &(count_ptr->simple_template),
                             &(default_to_dull_count_ptr->simple_template),
                             &(cumulative_error_sum_ptr->simple_template),
                             &(max_ptr->simple_template),
                             &(sum_sqr_ptr->simple_template)));
#endif

#ifdef TEMPLATE
    ERE(update_cumulative_template_error(&(error_ptr->template),
                                         &(dull_error_ptr->template),
                             &(count_ptr->template),
                             &(default_to_dull_count_ptr->template),
                             &(cumulative_error_sum_ptr->template),
                             &(max_ptr->template),
                             &(sum_sqr_ptr->template)));
#endif

#ifdef MULTI_TEMPLATE
    ERE(update_cumulative_multi_template_error(&(error_ptr->multi_template),
                                         &(dull_error_ptr->multi_template),
                             &(count_ptr->multi_template),
                             &(default_to_dull_count_ptr->multi_template),
                             &(cumulative_error_sum_ptr->multi_template),
                             &(max_ptr->multi_template),
                             &(sum_sqr_ptr->multi_template)));
#endif

    return NO_ERROR;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_best_fit_error
(
    Best_fit_error* error_ptr,
    Best_fit_error* dull_error_ptr,
    Best_fit_error* count_ptr,
    Best_fit_error* default_to_dull_count_ptr,
    Best_fit_error* cumulative_error_sum_ptr,
    Best_fit_error* max_ptr,
    Best_fit_error* sum_sqr_ptr
)
{

    ERE(update_cumulative_error_item(error_ptr->world_best_3D_fit,
                                     dull_error_ptr->world_best_3D_fit,
                             &(count_ptr->world_best_3D_fit),
                             &(default_to_dull_count_ptr->world_best_3D_fit),
                             &(cumulative_error_sum_ptr->world_best_3D_fit),
                             &(max_ptr->world_best_3D_fit),
                             &(sum_sqr_ptr->world_best_3D_fit)));

    ERE(update_cumulative_error_item(error_ptr->world_best_2D_fit,
                                     dull_error_ptr->world_best_2D_fit,
                             &(count_ptr->world_best_2D_fit),
                             &(default_to_dull_count_ptr->world_best_2D_fit),
                             &(cumulative_error_sum_ptr->world_best_2D_fit),
                             &(max_ptr->world_best_2D_fit),
                             &(sum_sqr_ptr->world_best_2D_fit)));

    ERE(update_cumulative_error_item(error_ptr->world_best_3D_diag_fit,
                                     dull_error_ptr->world_best_3D_diag_fit,
                &(count_ptr->world_best_3D_diag_fit),
                &(default_to_dull_count_ptr->world_best_3D_diag_fit),
                &(cumulative_error_sum_ptr->world_best_3D_diag_fit),
                &(max_ptr->world_best_3D_diag_fit),
                &(sum_sqr_ptr->world_best_3D_diag_fit)));

    ERE(update_cumulative_error_item(error_ptr->world_best_2D_diag_fit,
                                     dull_error_ptr->world_best_2D_diag_fit,
                &(count_ptr->world_best_2D_diag_fit),
                &(default_to_dull_count_ptr->world_best_2D_diag_fit),
                &(cumulative_error_sum_ptr->world_best_2D_diag_fit),
                &(max_ptr->world_best_2D_diag_fit),
                &(sum_sqr_ptr->world_best_2D_diag_fit)));

    ERE(update_cumulative_error_item(error_ptr->scene_best_3D_fit,
                                     dull_error_ptr->scene_best_3D_fit,
                &(count_ptr->scene_best_3D_fit),
                &(default_to_dull_count_ptr->scene_best_3D_fit),
                &(cumulative_error_sum_ptr->scene_best_3D_fit),
                &(max_ptr->scene_best_3D_fit),
                &(sum_sqr_ptr->scene_best_3D_fit)));

    ERE(update_cumulative_error_item(error_ptr->scene_best_2D_fit,
                                     dull_error_ptr->scene_best_2D_fit,
                &(count_ptr->scene_best_2D_fit),
                &(default_to_dull_count_ptr->scene_best_2D_fit),
                &(cumulative_error_sum_ptr->scene_best_2D_fit),
                &(max_ptr->scene_best_2D_fit),
                &(sum_sqr_ptr->scene_best_2D_fit)));

    ERE(update_cumulative_error_item(error_ptr->scene_best_3D_diag_fit,
                                     dull_error_ptr->scene_best_3D_diag_fit,
                &(count_ptr->scene_best_3D_diag_fit),
                &(default_to_dull_count_ptr->scene_best_3D_diag_fit),
                &(cumulative_error_sum_ptr->scene_best_3D_diag_fit),
                &(max_ptr->scene_best_3D_diag_fit),
                &(sum_sqr_ptr->scene_best_3D_diag_fit)));

    ERE(update_cumulative_error_item(error_ptr->scene_best_2D_diag_fit,
                                     dull_error_ptr->scene_best_2D_diag_fit,
                &(count_ptr->scene_best_2D_diag_fit),
                &(default_to_dull_count_ptr->scene_best_2D_diag_fit),
                &(cumulative_error_sum_ptr->scene_best_2D_diag_fit),
                &(max_ptr->scene_best_2D_diag_fit),
                &(sum_sqr_ptr->scene_best_2D_diag_fit)));

    ERE(update_cumulative_error_item(error_ptr->actual,
                                     dull_error_ptr->actual,
                             &(count_ptr->actual),
                             &(default_to_dull_count_ptr->actual),
                             &(cumulative_error_sum_ptr->actual),
                             &(max_ptr->actual),
                             &(sum_sqr_ptr->actual)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_standard_error
(
    Standard_error* error_ptr,
    Standard_error* dull_error_ptr,
    Standard_error* count_ptr,
    Standard_error* default_to_dull_count_ptr,
    Standard_error* cumulative_error_sum_ptr,
    Standard_error* max_ptr,
    Standard_error* sum_sqr_ptr
)
{

    ERE(update_cumulative_error_item(error_ptr->nothing,
                                     dull_error_ptr->nothing,
                             &(count_ptr->nothing),
                             &(default_to_dull_count_ptr->nothing),
                             &(cumulative_error_sum_ptr->nothing),
                             &(max_ptr->nothing),
                             &(sum_sqr_ptr->nothing)));

    ERE(update_cumulative_error_item(error_ptr->ave_illum,
                                     dull_error_ptr->ave_illum,
                             &(count_ptr->ave_illum),
                             &(default_to_dull_count_ptr->ave_illum),
                             &(cumulative_error_sum_ptr->ave_illum),
                             &(max_ptr->ave_illum),
                             &(sum_sqr_ptr->ave_illum)));

    ERE(update_cumulative_error_item(error_ptr->grey_world,
                                     dull_error_ptr->grey_world,
                             &(count_ptr->grey_world),
                            &(default_to_dull_count_ptr->grey_world),
                             &(cumulative_error_sum_ptr->grey_world),
                             &(max_ptr->grey_world),
                             &(sum_sqr_ptr->grey_world)));

    ERE(update_cumulative_error_item(error_ptr->db_grey_world,
                                     dull_error_ptr->db_grey_world,
                        &(count_ptr->db_grey_world),
                        &(default_to_dull_count_ptr->db_grey_world),
                        &(cumulative_error_sum_ptr->db_grey_world),
                        &(max_ptr->db_grey_world),
                        &(sum_sqr_ptr->db_grey_world)));

    ERE(update_cumulative_error_item(error_ptr->scale_by_max,
                                     dull_error_ptr->scale_by_max,
                             &(count_ptr->scale_by_max),
                             &(default_to_dull_count_ptr->scale_by_max),
                             &(cumulative_error_sum_ptr->scale_by_max),
                             &(max_ptr->scale_by_max),
                             &(sum_sqr_ptr->scale_by_max)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_gamut_error
(
    Gamut_error* error_ptr,
    Gamut_error* dull_error_ptr,
    Gamut_error* count_ptr,
    Gamut_error* default_to_dull_count_ptr,
    Gamut_error* cumulative_error_sum_ptr,
    Gamut_error* max_ptr,
    Gamut_error* sum_sqr_ptr
)
{
    int i;

    ERE(update_cumulative_error_item(error_ptr->mv_surface,
                                     dull_error_ptr->mv_surface,
                         &(count_ptr->mv_surface),
                         &(default_to_dull_count_ptr->mv_surface),
                         &(cumulative_error_sum_ptr->mv_surface),
                         &(max_ptr->mv_surface),
                         &(sum_sqr_ptr->mv_surface)));

    ERE(update_cumulative_error_item(error_ptr->mv_surface_illum,
                                     dull_error_ptr->mv_surface_illum,
                   &(count_ptr->mv_surface_illum),
                   &(default_to_dull_count_ptr->mv_surface_illum),
                   &(cumulative_error_sum_ptr->mv_surface_illum),
                   &(max_ptr->mv_surface_illum),
                   &(sum_sqr_ptr->mv_surface_illum)));

    ERE(update_cumulative_error_item(error_ptr->ha_surface,
                                     dull_error_ptr->ha_surface,
                         &(count_ptr->ha_surface),
                         &(default_to_dull_count_ptr->ha_surface),
                         &(cumulative_error_sum_ptr->ha_surface),
                         &(max_ptr->ha_surface),
                         &(sum_sqr_ptr->ha_surface)));

    ERE(update_cumulative_error_item(error_ptr->ha_surface_illum,
                                     dull_error_ptr->ha_surface_illum,
                     &(count_ptr->ha_surface_illum),
                     &(default_to_dull_count_ptr->ha_surface_illum),
                     &(cumulative_error_sum_ptr->ha_surface_illum),
                     &(max_ptr->ha_surface_illum),
                     &(sum_sqr_ptr->ha_surface_illum)));

    ERE(update_cumulative_error_item(error_ptr->icsa_surface_illum,
                                     dull_error_ptr->icsa_surface_illum,
                 &(count_ptr->icsa_surface_illum),
                 &(default_to_dull_count_ptr->icsa_surface_illum),
                 &(cumulative_error_sum_ptr->icsa_surface_illum),
                 &(max_ptr->icsa_surface_illum),
                 &(sum_sqr_ptr->icsa_surface_illum)));

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        ERE(update_cumulative_error_item(error_ptr->scia_surface_illum[ i ],
                                         dull_error_ptr->scia_surface_illum[ i ],
                     &(count_ptr->scia_surface_illum[ i ]),
                     &(default_to_dull_count_ptr->scia_surface_illum[ i ]),
                     &(cumulative_error_sum_ptr->scia_surface_illum[ i ]),
                     &(max_ptr->scia_surface_illum[ i ]),
                     &(sum_sqr_ptr->scia_surface_illum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->scca_surface_illum[ i ],
                                         dull_error_ptr->scca_surface_illum[ i ],
                     &(count_ptr->scca_surface_illum[ i ]),
                     &(default_to_dull_count_ptr->scca_surface_illum[ i ]),
                     &(cumulative_error_sum_ptr->scca_surface_illum[ i ]),
                     &(max_ptr->scca_surface_illum[ i ]),
                     &(sum_sqr_ptr->scca_surface_illum[ i ])));
    }

    ERE(update_cumulative_error_item(error_ptr->wa_surface,
                                     dull_error_ptr->wa_surface,
                         &(count_ptr->wa_surface),
                         &(default_to_dull_count_ptr->wa_surface),
                         &(cumulative_error_sum_ptr->wa_surface),
                         &(max_ptr->wa_surface),
                         &(sum_sqr_ptr->wa_surface)));

    ERE(update_cumulative_error_item(error_ptr->wa_surface_illum,
                                     dull_error_ptr->wa_surface_illum,
                   &(count_ptr->wa_surface_illum),
                   &(default_to_dull_count_ptr->wa_surface_illum),
                   &(cumulative_error_sum_ptr->wa_surface_illum),
                   &(max_ptr->wa_surface_illum),
                   &(sum_sqr_ptr->wa_surface_illum)));

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        ERE(update_cumulative_error_item(error_ptr->ha_surface_lum[ i ],
                                         dull_error_ptr->ha_surface_lum[ i ],
                             &(count_ptr->ha_surface_lum[ i ]),
                             &(default_to_dull_count_ptr->ha_surface_lum[ i ]),
                             &(cumulative_error_sum_ptr->ha_surface_lum[ i ]),
                             &(max_ptr->ha_surface_lum[ i ]),
                             &(sum_sqr_ptr->ha_surface_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->mv_surface_lum[ i ],
                                         dull_error_ptr->mv_surface_lum[ i ],
                             &(count_ptr->mv_surface_lum[ i ]),
                             &(default_to_dull_count_ptr->mv_surface_lum[ i ]),
                             &(cumulative_error_sum_ptr->mv_surface_lum[ i ]),
                             &(max_ptr->mv_surface_lum[ i ]),
                             &(sum_sqr_ptr->mv_surface_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->ha_surface_illum_lum[ i ],
                        dull_error_ptr->ha_surface_illum_lum[ i ],
                        &(count_ptr->ha_surface_illum_lum[ i ]),
                        &(default_to_dull_count_ptr->ha_surface_illum_lum[ i ]),
                        &(cumulative_error_sum_ptr->ha_surface_illum_lum[ i ]),
                        &(max_ptr->ha_surface_illum_lum[ i ]),
                        &(sum_sqr_ptr->ha_surface_illum_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->mv_surface_illum_lum[ i ],
                        dull_error_ptr->mv_surface_illum_lum[ i ],
                        &(count_ptr->mv_surface_illum_lum[ i ]),
                        &(default_to_dull_count_ptr->mv_surface_illum_lum[ i ]),
                        &(cumulative_error_sum_ptr->mv_surface_illum_lum[ i ]),
                        &(max_ptr->mv_surface_illum_lum[ i ]),
                        &(sum_sqr_ptr->mv_surface_illum_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->scla_surface_illum_lum[ i ],
                 dull_error_ptr->scla_surface_illum_lum[ i ],
                 &(count_ptr->scla_surface_illum_lum[ i ]),
             &(default_to_dull_count_ptr->scla_surface_illum_lum[ i ]),
                 &(cumulative_error_sum_ptr->scla_surface_illum_lum[ i ]),
                 &(max_ptr->scla_surface_illum_lum[ i ]),
                 &(sum_sqr_ptr->scla_surface_illum_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->lcsa_surface_illum_lum[ i ],
                     dull_error_ptr->lcsa_surface_illum_lum[ i ],
                     &(count_ptr->lcsa_surface_illum_lum[ i ]),
                     &(default_to_dull_count_ptr->lcsa_surface_illum_lum[ i ]),
                     &(cumulative_error_sum_ptr->lcsa_surface_illum_lum[ i ]),
                     &(max_ptr->lcsa_surface_illum_lum[ i ]),
                     &(sum_sqr_ptr->lcsa_surface_illum_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->scia_surface_illum_lum[ i ],
                                 dull_error_ptr->scia_surface_illum_lum[ i ],
                     &(count_ptr->scia_surface_illum_lum[ i ]),
                     &(default_to_dull_count_ptr->scia_surface_illum_lum[ i ]),
                     &(cumulative_error_sum_ptr->scia_surface_illum_lum[ i ]),
                     &(max_ptr->scia_surface_illum_lum[ i ]),
                     &(sum_sqr_ptr->scia_surface_illum_lum[ i ])));

        ERE(update_cumulative_error_item(error_ptr->scca_surface_illum_lum[ i ],
                                 dull_error_ptr->scca_surface_illum_lum[ i ],
                     &(count_ptr->scca_surface_illum_lum[ i ]),
                     &(default_to_dull_count_ptr->scca_surface_illum_lum[ i ]),
                     &(cumulative_error_sum_ptr->scca_surface_illum_lum[ i ]),
                     &(max_ptr->scca_surface_illum_lum[ i ]),
                     &(sum_sqr_ptr->scca_surface_illum_lum[ i ])));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay - June 7/99 */
static int update_cumulative_correlation_error
(
    Correlation_error* error_ptr,
    Correlation_error* dull_error_ptr,
    Correlation_error* count_ptr,
    Correlation_error* default_to_dull_count_ptr,
    Correlation_error* cumulative_error_sum_ptr,
    Correlation_error* max_ptr,
    Correlation_error* sum_sqr_ptr
)
{

    int i; /* Modified April 14/99 */

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                                error_ptr->correlation_01[i],
                                dull_error_ptr->correlation_01[i],
                                &(count_ptr->correlation_01[i]),
                                &(default_to_dull_count_ptr->correlation_01[i]),
                                &(cumulative_error_sum_ptr->correlation_01[i]),
                                &(max_ptr->correlation_01[i]),
                                &(sum_sqr_ptr->correlation_01[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                                error_ptr->correlation_01_alt[i],
                                dull_error_ptr->correlation_01_alt[i],
                                &(count_ptr->correlation_01_alt[i]),
                                &(default_to_dull_count_ptr->correlation_01_alt[i]),
                                &(cumulative_error_sum_ptr->correlation_01_alt[i]),
                                &(max_ptr->correlation_01_alt[i]),
                                &(sum_sqr_ptr->correlation_01_alt[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                        error_ptr->correlation_gaussian_MAP[i],
                        dull_error_ptr->correlation_gaussian_MAP[i],
                        &(count_ptr->correlation_gaussian_MAP[i]),
                      &(default_to_dull_count_ptr->correlation_gaussian_MAP[i]),
                      &(cumulative_error_sum_ptr->correlation_gaussian_MAP[i]),
                        &(max_ptr->correlation_gaussian_MAP[i]),
                        &(sum_sqr_ptr->correlation_gaussian_MAP[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                    error_ptr->correlation_uniform_MAP_inv_dist_fill[i],
                    dull_error_ptr->correlation_uniform_MAP_inv_dist_fill[i],
                    &(count_ptr->correlation_uniform_MAP_inv_dist_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MAP_inv_dist_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MAP_inv_dist_fill[i]),
            &(max_ptr->correlation_uniform_MAP_inv_dist_fill[i]),
            &(sum_sqr_ptr->correlation_uniform_MAP_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                    error_ptr->correlation_uniform_MAP_gaussian_fill[i],
                    dull_error_ptr->correlation_uniform_MAP_gaussian_fill[i],
                    &(count_ptr->correlation_uniform_MAP_gaussian_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MAP_gaussian_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MAP_gaussian_fill[i]),
                    &(max_ptr->correlation_uniform_MAP_gaussian_fill[i]),
                    &(sum_sqr_ptr->correlation_uniform_MAP_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                            error_ptr->correlation_uniform_MAP_fitted[i],
                            dull_error_ptr->correlation_uniform_MAP_fitted[i],
                            &(count_ptr->correlation_uniform_MAP_fitted[i]),
                &(default_to_dull_count_ptr->correlation_uniform_MAP_fitted[i]),
                &(cumulative_error_sum_ptr->correlation_uniform_MAP_fitted[i]),
                &(max_ptr->correlation_uniform_MAP_fitted[i]),
                &(sum_sqr_ptr->correlation_uniform_MAP_fitted[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                                error_ptr->correlation_gaussian_MLM[i],
                                dull_error_ptr->correlation_gaussian_MLM[i],
                                &(count_ptr->correlation_gaussian_MLM[i]),
                    &(default_to_dull_count_ptr->correlation_gaussian_MLM[i]),
                    &(cumulative_error_sum_ptr->correlation_gaussian_MLM[i]),
                    &(max_ptr->correlation_gaussian_MLM[i]),
                    &(sum_sqr_ptr->correlation_gaussian_MLM[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                    error_ptr->correlation_uniform_MLM_inv_dist_fill[i],
                    dull_error_ptr->correlation_uniform_MLM_inv_dist_fill[i],
                    &(count_ptr->correlation_uniform_MLM_inv_dist_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MLM_inv_dist_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MLM_inv_dist_fill[i]),
                &(max_ptr->correlation_uniform_MLM_inv_dist_fill[i]),
                &(sum_sqr_ptr->correlation_uniform_MLM_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                error_ptr->correlation_uniform_MLM_gaussian_fill[i],
                dull_error_ptr->correlation_uniform_MLM_gaussian_fill[i],
                &(count_ptr->correlation_uniform_MLM_gaussian_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MLM_gaussian_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MLM_gaussian_fill[i]),
                &(max_ptr->correlation_uniform_MLM_gaussian_fill[i]),
                &(sum_sqr_ptr->correlation_uniform_MLM_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                error_ptr->correlation_uniform_MLM_fitted[i],
                dull_error_ptr->correlation_uniform_MLM_fitted[i],
                &(count_ptr->correlation_uniform_MLM_fitted[i]),
                &(default_to_dull_count_ptr->correlation_uniform_MLM_fitted[i]),
                &(cumulative_error_sum_ptr->correlation_uniform_MLM_fitted[i]),
                &(max_ptr->correlation_uniform_MLM_fitted[i]),
                &(sum_sqr_ptr->correlation_uniform_MLM_fitted[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                                error_ptr->correlation_gaussian_MMSE[i],
                                dull_error_ptr->correlation_gaussian_MMSE[i],
                                &(count_ptr->correlation_gaussian_MMSE[i]),
                    &(default_to_dull_count_ptr->correlation_gaussian_MMSE[i]),
                    &(cumulative_error_sum_ptr->correlation_gaussian_MMSE[i]),
                    &(max_ptr->correlation_gaussian_MMSE[i]),
                    &(sum_sqr_ptr->correlation_gaussian_MMSE[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                    error_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
                    dull_error_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
                    &(count_ptr->correlation_uniform_MMSE_inv_dist_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MMSE_inv_dist_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MMSE_inv_dist_fill[i]),
                &(max_ptr->correlation_uniform_MMSE_inv_dist_fill[i]),
                &(sum_sqr_ptr->correlation_uniform_MMSE_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                error_ptr->correlation_uniform_MMSE_gaussian_fill[i],
                dull_error_ptr->correlation_uniform_MMSE_gaussian_fill[i],
                &(count_ptr->correlation_uniform_MMSE_gaussian_fill[i]),
        &(default_to_dull_count_ptr->correlation_uniform_MMSE_gaussian_fill[i]),
        &(cumulative_error_sum_ptr->correlation_uniform_MMSE_gaussian_fill[i]),
                &(max_ptr->correlation_uniform_MMSE_gaussian_fill[i]),
                &(sum_sqr_ptr->correlation_uniform_MMSE_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(update_cumulative_error_item(
                error_ptr->correlation_uniform_MMSE_fitted[i],
                dull_error_ptr->correlation_uniform_MMSE_fitted[i],
                &(count_ptr->correlation_uniform_MMSE_fitted[i]),
               &(default_to_dull_count_ptr->correlation_uniform_MMSE_fitted[i]),
               &(cumulative_error_sum_ptr->correlation_uniform_MMSE_fitted[i]),
                &(max_ptr->correlation_uniform_MMSE_fitted[i]),
                &(sum_sqr_ptr->correlation_uniform_MMSE_fitted[i])));
    }

    return NO_ERROR;

}
/* End Lindsay: June 07/99 */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static int update_cumulative_simple_template_error
(
    Simple_template_error* error_ptr,
    Simple_template_error* dull_error_ptr,
    Simple_template_error* count_ptr,
    Simple_template_error* default_to_dull_count_ptr,
    Simple_template_error* cumulative_error_sum_ptr,
    Simple_template_error* max_ptr,
    Simple_template_error* sum_sqr_ptr
)
{

    ERE(update_cumulative_error_item(error_ptr->alg,
                                     dull_error_ptr->alg,
                                     &(count_ptr->alg),
                                     &(default_to_dull_count_ptr->alg),
                                     &(cumulative_error_sum_ptr->alg),
                                     &(max_ptr->alg),
                                     &(sum_sqr_ptr->alg)));

    return NO_ERROR;

}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static int update_cumulative_template_error
(
    Template_error* error_ptr,
    Template_error* dull_error_ptr,
    Template_error* count_ptr,
    Template_error* default_to_dull_count_ptr,
    Template_error* cumulative_error_sum_ptr,
    Template_error* max_ptr,
    Template_error* sum_sqr_ptr
)
{

    ERE(update_cumulative_error_item(error_ptr->first_alg,
                                     dull_error_ptr->first_alg,
                                     &(count_ptr->first_alg),
                                     &(default_to_dull_count_ptr->first_alg),
                                     &(cumulative_error_sum_ptr->first_alg),
                                     &(max_ptr->first_alg),
                                     &(sum_sqr_ptr->first_alg)));

    ERE(update_cumulative_error_item(error_ptr->second_alg,
                                     dull_error_ptr->second_alg,
                                     &(count_ptr->second_alg),
                                     &(default_to_dull_count_ptr->second_alg),
                                     &(cumulative_error_sum_ptr->second_alg),
                                     &(max_ptr->second_alg),
                                     &(sum_sqr_ptr->second_alg)));

    return NO_ERROR;

}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE

static int update_cumulative_multi_template_error
(
    Multi_template_error* error_ptr,
    Multi_template_error* dull_error_ptr,
    Multi_template_error* count_ptr,
    Multi_template_error* default_to_dull_count_ptr,
    Multi_template_error* cumulative_error_sum_ptr,
    Multi_template_error* max_ptr,
    Multi_template_error* sum_sqr_ptr
)
{
    int i;

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        ERE(update_cumulative_error_item(error_ptr->alg[ i ],
                                         dull_error_ptr->alg[ i ],
                                         &(count_ptr->alg[ i ]),
                                         &(default_to_dull_count_ptr->alg[ i ]),
                                         &(cumulative_error_sum_ptr->alg[ i ]),
                                         &(max_ptr->alg[ i ]),
                                         &(sum_sqr_ptr->alg[ i ])));
    }

    return NO_ERROR;

}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_cumulative_error_item
(
    double  error,
    double  dull_error,
    double* count_ptr,
    double* default_to_dull_count_ptr,
    double* cumulative_error_sum_ptr,
    double* max_ptr,
    double* sum_sqr_ptr
)
{

    /* Check for lingering use of the old method of specifying missing value. */
    ASSERT( ! IS_EQUAL_DBL(error, DBL_NOT_SET));

    if (error < DBL_HALF_MOST_NEGATIVE/2.0)     /* Error is not set/used */
    {
        if (fs_default_to_dull_error && (dull_error > DBL_HALF_MOST_NEGATIVE/2.0))
        {
            error = dull_error;
            *default_to_dull_count_ptr += 1.0;
        }
        else
        {
            return NO_ERROR;
        }
    }

    if (*count_ptr < DBL_HALF_MOST_NEGATIVE/2.0)
    {
        *count_ptr = 0.0;
        *cumulative_error_sum_ptr = 0.0;
        *max_ptr = 0.0;
        *sum_sqr_ptr = 0.0;
    }

    *count_ptr += 1.0;

    *cumulative_error_sum_ptr += error;
    *sum_sqr_ptr += error*error;

    if (ABS_OF(error) > *max_ptr)
    {
        *max_ptr = ABS_OF(error);
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_cumulative_error_report(char* label)
{
    static const char*            twenty_equals  = "====================";
    Colour_constancy_error        error_stat_one;
    Colour_constancy_error        error_stat_two;
    Input_data_error              input_data_error_stat_one;
    Input_data_error              input_data_error_stat_two;
    char                          heading[ 1000 ];
    int                           i, j;
    char                          error_stat_two_str[ 100 ];
    char                          error_stat_one_str[ 100 ];

    if (fs_cumulative_input_count <= 0)
    {
        set_error("No data yet collected for cumulative report.");
        return ERROR;
    }

    initialize_error(&error_stat_one, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&error_stat_two, DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&input_data_error_stat_one,
                                DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&input_data_error_stat_two,
                                DBL_HALF_MOST_NEGATIVE);

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++)
        {
            ERE(calculate_cumulative_gamut_input_data_error(
                   &(fs_cumulative_input_data_error_count.sharp[ j ].gamut_2D[ i ]),
                   &(fs_cumulative_input_data_default_to_dull_count.sharp[ j ].gamut_2D[ i ]),
                   &(fs_cumulative_input_data_error_sum.sharp[ j ].gamut_2D[ i ]),
                   &(fs_cumulative_input_data_error_max.sharp[ j ].gamut_2D[ i ]),
                   &(fs_cumulative_input_data_error_sum_sqr.sharp[ j ].gamut_2D[ i ]),
                   &(input_data_error_stat_one.sharp[ j ].gamut_2D[ i ]),
                   &(input_data_error_stat_two.sharp[ j ].gamut_2D[ i ])));

            ERE(calculate_cumulative_gamut_input_data_error(
                   &(fs_cumulative_input_data_error_count.sharp[ j ].gamut_3D[ i ]),
                   &(fs_cumulative_input_data_default_to_dull_count.sharp[ j ].gamut_3D[ i ]),
                   &(fs_cumulative_input_data_error_sum.sharp[ j ].gamut_3D[ i ]),
                   &(fs_cumulative_input_data_error_max.sharp[ j ].gamut_3D[ i ]),
                   &(fs_cumulative_input_data_error_sum_sqr.sharp[ j ].gamut_3D[ i ]),
                   &(input_data_error_stat_one.sharp[ j ].gamut_3D[ i ]),
                   &(input_data_error_stat_two.sharp[ j ].gamut_3D[ i ])));
        }
    }

    ERE(calculate_cumulative_error_by_sharp(
               &(fs_cumulative_error_count.illum_angular),
               &(fs_cumulative_default_to_dull_count.illum_angular),
               &(fs_cumulative_error_sum.illum_angular),
               &(fs_cumulative_error_max.illum_angular),
               &(fs_cumulative_error_sum_sqr.illum_angular),
               &(error_stat_one.illum_angular),
               &(error_stat_two.illum_angular)));

    ERE(calculate_cumulative_error_by_sharp(
               &(fs_cumulative_error_count.illum_sum_chrom),
               &(fs_cumulative_default_to_dull_count.illum_sum_chrom),
               &(fs_cumulative_error_sum.illum_sum_chrom),
               &(fs_cumulative_error_max.illum_sum_chrom),
               &(fs_cumulative_error_sum_sqr.illum_sum_chrom),
               &(error_stat_one.illum_sum_chrom),
               &(error_stat_two.illum_sum_chrom)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_cumulative_error_count.illum_rgb),
                               &(fs_cumulative_default_to_dull_count.illum_rgb),
                               &(fs_cumulative_error_sum.illum_rgb),
                               &(fs_cumulative_error_max.illum_rgb),
                               &(fs_cumulative_error_sum_sqr.illum_rgb),
                               &(error_stat_one.illum_rgb),
                               &(error_stat_two.illum_rgb)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.illum_r_div_sum),
                           &(fs_cumulative_default_to_dull_count.illum_r_div_sum),
                           &(fs_cumulative_error_sum.illum_r_div_sum),
                           &(fs_cumulative_error_max.illum_r_div_sum),
                           &(fs_cumulative_error_sum_sqr.illum_r_div_sum),
                           &(error_stat_one.illum_r_div_sum),
                           &(error_stat_two.illum_r_div_sum)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.illum_g_div_sum),
                           &(fs_cumulative_default_to_dull_count.illum_g_div_sum),
                           &(fs_cumulative_error_sum.illum_g_div_sum),
                           &(fs_cumulative_error_max.illum_g_div_sum),
                           &(fs_cumulative_error_sum_sqr.illum_g_div_sum),
                           &(error_stat_one.illum_g_div_sum),
                           &(error_stat_two.illum_g_div_sum)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.illum_sum_lum),
                           &(fs_cumulative_default_to_dull_count.illum_sum_lum),
                           &(fs_cumulative_error_sum.illum_sum_lum),
                           &(fs_cumulative_error_max.illum_sum_lum),
                           &(fs_cumulative_error_sum_sqr.illum_sum_lum),
                           &(error_stat_one.illum_sum_lum),
                           &(error_stat_two.illum_sum_lum)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.illum_lab),
                           &(fs_cumulative_default_to_dull_count.illum_lab),
                           &(fs_cumulative_error_sum.illum_lab),
                           &(fs_cumulative_error_max.illum_lab),
                           &(fs_cumulative_error_sum_sqr.illum_lab),
                           &(error_stat_one.illum_lab),
                           &(error_stat_two.illum_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                   &(fs_cumulative_error_count.world_map_sum_chrom_rms),
                   &(fs_cumulative_default_to_dull_count.world_map_sum_chrom_rms),
                   &(fs_cumulative_error_sum.world_map_sum_chrom_rms),
                   &(fs_cumulative_error_max.world_map_sum_chrom_rms),
                   &(fs_cumulative_error_sum_sqr.world_map_sum_chrom_rms),
                   &(error_stat_one.world_map_sum_chrom_rms),
                   &(error_stat_two.world_map_sum_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                   &(fs_cumulative_error_count.world_map_blue_chrom_rms),
                   &(fs_cumulative_default_to_dull_count.world_map_blue_chrom_rms),
                   &(fs_cumulative_error_sum.world_map_blue_chrom_rms),
                   &(fs_cumulative_error_max.world_map_blue_chrom_rms),
                   &(fs_cumulative_error_sum_sqr.world_map_blue_chrom_rms),
                   &(error_stat_one.world_map_blue_chrom_rms),
                   &(error_stat_two.world_map_blue_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.world_map_3D_rms),
                           &(fs_cumulative_default_to_dull_count.world_map_3D_rms),
                           &(fs_cumulative_error_sum.world_map_3D_rms),
                           &(fs_cumulative_error_max.world_map_3D_rms),
                           &(fs_cumulative_error_sum_sqr.world_map_3D_rms),
                           &(error_stat_one.world_map_3D_rms),
                           &(error_stat_two.world_map_3D_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                           &(fs_cumulative_error_count.world_map_3D_lab),
                           &(fs_cumulative_default_to_dull_count.world_map_3D_lab),
                           &(fs_cumulative_error_sum.world_map_3D_lab),
                           &(fs_cumulative_error_max.world_map_3D_lab),
                           &(fs_cumulative_error_sum_sqr.world_map_3D_lab),
                           &(error_stat_one.world_map_3D_lab),
                           &(error_stat_two.world_map_3D_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                   &(fs_cumulative_error_count.scene_map_sum_chrom_rms),
                   &(fs_cumulative_default_to_dull_count.scene_map_sum_chrom_rms),
                   &(fs_cumulative_error_sum.scene_map_sum_chrom_rms),
                   &(fs_cumulative_error_max.scene_map_sum_chrom_rms),
                   &(fs_cumulative_error_sum_sqr.scene_map_sum_chrom_rms),
                   &(error_stat_one.scene_map_sum_chrom_rms),
                   &(error_stat_two.scene_map_sum_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                   &(fs_cumulative_error_count.scene_map_blue_chrom_rms),
                   &(fs_cumulative_default_to_dull_count.scene_map_blue_chrom_rms),
                   &(fs_cumulative_error_sum.scene_map_blue_chrom_rms),
                   &(fs_cumulative_error_max.scene_map_blue_chrom_rms),
                   &(fs_cumulative_error_sum_sqr.scene_map_blue_chrom_rms),
                   &(error_stat_one.scene_map_blue_chrom_rms),
                   &(error_stat_two.scene_map_blue_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_cumulative_error_count.scene_map_3D_rms),
                       &(fs_cumulative_default_to_dull_count.scene_map_3D_rms),
                       &(fs_cumulative_error_sum.scene_map_3D_rms),
                       &(fs_cumulative_error_max.scene_map_3D_rms),
                       &(fs_cumulative_error_sum_sqr.scene_map_3D_rms),
                       &(error_stat_one.scene_map_3D_rms),
                       &(error_stat_two.scene_map_3D_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_cumulative_error_count.scene_map_3D_lab),
                       &(fs_cumulative_default_to_dull_count.scene_map_3D_lab),
                       &(fs_cumulative_error_sum.scene_map_3D_lab),
                       &(fs_cumulative_error_max.scene_map_3D_lab),
                       &(fs_cumulative_error_sum_sqr.scene_map_3D_lab),
                       &(error_stat_one.scene_map_3D_lab),
                       &(error_stat_two.scene_map_3D_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                                   &(fs_cumulative_error_count.cpu_time),
                                   &(fs_cumulative_default_to_dull_count.cpu_time),
                                   &(fs_cumulative_error_sum.cpu_time),
                                   &(fs_cumulative_error_max.cpu_time),
                                   &(fs_cumulative_error_sum_sqr.cpu_time),
                                   &(error_stat_one.cpu_time),
                                   &(error_stat_two.cpu_time)));

    ERE(kjb_sprintf(heading, sizeof(heading),
                    "%s%s%s%s\n\n%s (%d input sets)\n%s%s%s\n",
                    twenty_equals, twenty_equals, twenty_equals, twenty_equals,
                    "Cumulative Error Report Since Last Reset",
                    fs_cumulative_input_count,
                    (label && *label) ? "[ " : "",
                    (label && *label) ? label : "" ,
                    (label && *label) ? " ]\n" : ""));

    ERE(get_error_stat_heading_str(fs_error_stat_one_request, error_stat_one_str,
                              sizeof(error_stat_one_str)));

    ERE(get_error_stat_heading_str(fs_error_stat_two_request, error_stat_two_str,
                              sizeof(error_stat_two_str)));

    ERE(do_error_report(&error_stat_one, &error_stat_two,
                        &input_data_error_stat_one, &input_data_error_stat_two,
                        heading, error_stat_one_str, error_stat_two_str));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int do_grand_total_error_report(char* label)
{
    static const char*            twenty_equals = "====================";
    Colour_constancy_error        error_stat_one;
    Colour_constancy_error        error_stat_two;
    Input_data_error              input_data_error_stat_two;
    Input_data_error              input_data_error_stat_one;
    char                          heading[ 1000 ];
    int                           i, j;
    char                          error_stat_two_str[ 100 ];
    char                          error_stat_one_str[ 100 ];

    if (fs_grand_input_count <= 0)
    {
        set_error("No data yet collected for grand total error report.");
        return ERROR;
    }

    initialize_error(&error_stat_one, DBL_HALF_MOST_NEGATIVE);
    initialize_error(&error_stat_two, DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&input_data_error_stat_two,
                                DBL_HALF_MOST_NEGATIVE);
    initialize_input_data_error(&input_data_error_stat_one,
                                DBL_HALF_MOST_NEGATIVE);

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        for (j=0; j<NUM_SHARP_METHODS; j++)
        {
            ERE(calculate_cumulative_gamut_input_data_error(
                     &(fs_grand_input_data_error_count.sharp[ j ].gamut_2D[ i ]),
                     &(fs_grand_input_data_default_to_dull_count.sharp[ j ].gamut_2D[ i ]),
                     &(fs_grand_input_data_error_sum.sharp[ j ].gamut_2D[ i ]),
                     &(fs_grand_input_data_error_max.sharp[ j ].gamut_2D[ i ]),
                     &(fs_grand_input_data_error_sum_sqr.sharp[ j ].gamut_2D[ i ]),
                     &(input_data_error_stat_one.sharp[ j ].gamut_2D[ i ]),
                     &(input_data_error_stat_two.sharp[ j ].gamut_2D[ i ])));

            ERE(calculate_cumulative_gamut_input_data_error(
                     &(fs_grand_input_data_error_count.sharp[ j ].gamut_3D[ i ]),
                     &(fs_grand_input_data_default_to_dull_count.sharp[ j ].gamut_3D[ i ]),
                     &(fs_grand_input_data_error_sum.sharp[ j ].gamut_3D[ i ]),
                     &(fs_grand_input_data_error_max.sharp[ j ].gamut_3D[ i ]),
                     &(fs_grand_input_data_error_sum_sqr.sharp[ j ].gamut_3D[ i ]),
                     &(input_data_error_stat_one.sharp[ j ].gamut_3D[ i ]),
                     &(input_data_error_stat_two.sharp[ j ].gamut_3D[ i ])));
        }
    }

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_angular),
                               &(fs_grand_default_to_dull_count.illum_angular),
                               &(fs_grand_error_sum.illum_angular),
                               &(fs_grand_error_max.illum_angular),
                               &(fs_grand_error_sum_sqr.illum_angular),
                               &(error_stat_one.illum_angular),
                               &(error_stat_two.illum_angular)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_sum_chrom),
                               &(fs_grand_default_to_dull_count.illum_sum_chrom),
                               &(fs_grand_error_sum.illum_sum_chrom),
                               &(fs_grand_error_max.illum_sum_chrom),
                               &(fs_grand_error_sum_sqr.illum_sum_chrom),
                               &(error_stat_one.illum_sum_chrom),
                               &(error_stat_two.illum_sum_chrom)));

    ERE(calculate_cumulative_error_by_sharp(
                                   &(fs_grand_error_count.illum_rgb),
                                   &(fs_grand_default_to_dull_count.illum_rgb),
                                   &(fs_grand_error_sum.illum_rgb),
                                   &(fs_grand_error_max.illum_rgb),
                                   &(fs_grand_error_sum_sqr.illum_rgb),
                                   &(error_stat_one.illum_rgb),
                                   &(error_stat_two.illum_rgb)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_r_div_sum),
                               &(fs_grand_default_to_dull_count.illum_r_div_sum),
                               &(fs_grand_error_sum.illum_r_div_sum),
                               &(fs_grand_error_max.illum_r_div_sum),
                               &(fs_grand_error_sum_sqr.illum_r_div_sum),
                               &(error_stat_one.illum_r_div_sum),
                               &(error_stat_two.illum_r_div_sum)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_g_div_sum),
                               &(fs_grand_default_to_dull_count.illum_g_div_sum),
                               &(fs_grand_error_sum.illum_g_div_sum),
                               &(fs_grand_error_max.illum_g_div_sum),
                               &(fs_grand_error_sum_sqr.illum_g_div_sum),
                               &(error_stat_one.illum_g_div_sum),
                               &(error_stat_two.illum_g_div_sum)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_sum_lum),
                               &(fs_grand_default_to_dull_count.illum_sum_lum),
                               &(fs_grand_error_sum.illum_sum_lum),
                               &(fs_grand_error_max.illum_sum_lum),
                               &(fs_grand_error_sum_sqr.illum_sum_lum),
                               &(error_stat_one.illum_sum_lum),
                               &(error_stat_two.illum_sum_lum)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.illum_lab),
                               &(fs_grand_default_to_dull_count.illum_lab),
                               &(fs_grand_error_sum.illum_lab),
                               &(fs_grand_error_max.illum_lab),
                               &(fs_grand_error_sum_sqr.illum_lab),
                               &(error_stat_one.illum_lab),
                               &(error_stat_two.illum_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_grand_error_count.world_map_sum_chrom_rms),
                       &(fs_grand_default_to_dull_count.world_map_sum_chrom_rms),
                       &(fs_grand_error_sum.world_map_sum_chrom_rms),
                       &(fs_grand_error_max.world_map_sum_chrom_rms),
                       &(fs_grand_error_sum_sqr.world_map_sum_chrom_rms),
                       &(error_stat_one.world_map_sum_chrom_rms),
                       &(error_stat_two.world_map_sum_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_grand_error_count.world_map_blue_chrom_rms),
                       &(fs_grand_default_to_dull_count.world_map_blue_chrom_rms),
                       &(fs_grand_error_sum.world_map_blue_chrom_rms),
                       &(fs_grand_error_max.world_map_blue_chrom_rms),
                       &(fs_grand_error_sum_sqr.world_map_blue_chrom_rms),
                       &(error_stat_one.world_map_blue_chrom_rms),
                       &(error_stat_two.world_map_blue_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.world_map_3D_rms),
                               &(fs_grand_default_to_dull_count.world_map_3D_rms),
                               &(fs_grand_error_sum.world_map_3D_rms),
                               &(fs_grand_error_max.world_map_3D_rms),
                               &(fs_grand_error_sum_sqr.world_map_3D_rms),
                               &(error_stat_one.world_map_3D_rms),
                               &(error_stat_two.world_map_3D_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.world_map_3D_lab),
                               &(fs_grand_default_to_dull_count.world_map_3D_lab),
                               &(fs_grand_error_sum.world_map_3D_lab),
                               &(fs_grand_error_max.world_map_3D_lab),
                               &(fs_grand_error_sum_sqr.world_map_3D_lab),
                               &(error_stat_one.world_map_3D_lab),
                               &(error_stat_two.world_map_3D_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_grand_error_count.scene_map_sum_chrom_rms),
                       &(fs_grand_default_to_dull_count.scene_map_sum_chrom_rms),
                       &(fs_grand_error_sum.scene_map_sum_chrom_rms),
                       &(fs_grand_error_max.scene_map_sum_chrom_rms),
                       &(fs_grand_error_sum_sqr.scene_map_sum_chrom_rms),
                       &(error_stat_one.scene_map_sum_chrom_rms),
                       &(error_stat_two.scene_map_sum_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                       &(fs_grand_error_count.scene_map_blue_chrom_rms),
                       &(fs_grand_default_to_dull_count.scene_map_blue_chrom_rms),
                       &(fs_grand_error_sum.scene_map_blue_chrom_rms),
                       &(fs_grand_error_max.scene_map_blue_chrom_rms),
                       &(fs_grand_error_sum_sqr.scene_map_blue_chrom_rms),
                       &(error_stat_one.scene_map_blue_chrom_rms),
                       &(error_stat_two.scene_map_blue_chrom_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.scene_map_3D_rms),
                               &(fs_grand_default_to_dull_count.scene_map_3D_rms),
                               &(fs_grand_error_sum.scene_map_3D_rms),
                               &(fs_grand_error_max.scene_map_3D_rms),
                               &(fs_grand_error_sum_sqr.scene_map_3D_rms),
                               &(error_stat_one.scene_map_3D_rms),
                               &(error_stat_two.scene_map_3D_rms)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.scene_map_3D_lab),
                               &(fs_grand_default_to_dull_count.scene_map_3D_lab),
                               &(fs_grand_error_sum.scene_map_3D_lab),
                               &(fs_grand_error_max.scene_map_3D_lab),
                               &(fs_grand_error_sum_sqr.scene_map_3D_lab),
                               &(error_stat_one.scene_map_3D_lab),
                               &(error_stat_two.scene_map_3D_lab)));

    ERE(calculate_cumulative_error_by_sharp(
                               &(fs_grand_error_count.cpu_time),
                               &(fs_grand_default_to_dull_count.cpu_time),
                               &(fs_grand_error_sum.cpu_time),
                               &(fs_grand_error_max.cpu_time),
                               &(fs_grand_error_sum_sqr.cpu_time),
                               &(error_stat_one.cpu_time),
                               &(error_stat_two.cpu_time)));

    ERE(kjb_sprintf(heading, sizeof(heading),
                    "%s%s%s%s\n\n%s (%d input sets)\n%s%s%s\n",
                    twenty_equals, twenty_equals, twenty_equals, twenty_equals,
                    "Grand Total Error Report",
                    fs_grand_input_count,
                    (label && *label) ? "[ " : "",
                    (label && *label) ? label : "",
                    (label && *label) ? " ]\n" : ""));

    ERE(get_error_stat_heading_str(fs_error_stat_one_request, error_stat_one_str,
                              sizeof(error_stat_one_str)));

    ERE(get_error_stat_heading_str(fs_error_stat_two_request, error_stat_two_str,
                              sizeof(error_stat_two_str)));

    ERE(do_error_report(&error_stat_one, &error_stat_two,
                        &input_data_error_stat_one, &input_data_error_stat_two,
                        heading, error_stat_one_str, error_stat_two_str));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_cumulative_gamut_input_data_error
(
    Gamut_input_data_error* count_ptr,
    Gamut_input_data_error* default_to_dull_count_ptr,
    Gamut_input_data_error* cumulative_error_sum_ptr,
    Gamut_input_data_error* max_ptr,
    Gamut_input_data_error* sum_sqr_ptr,
    Gamut_input_data_error* error_stat_one_ptr,
    Gamut_input_data_error* error_stat_two_ptr
)
{

    ERE(calculate_cumulative_error_stats(
                   (count_ptr->relative_pixel_error),
                   (default_to_dull_count_ptr->relative_pixel_error),
                   (cumulative_error_sum_ptr->relative_pixel_error),
                   (max_ptr->relative_pixel_error),
                   (sum_sqr_ptr->relative_pixel_error),
                   &(error_stat_one_ptr->relative_pixel_error),
                   &(error_stat_two_ptr->relative_pixel_error)));

    ERE(calculate_cumulative_error_stats(
                   (count_ptr->absolute_pixel_error),
                   (default_to_dull_count_ptr->absolute_pixel_error),
                   (cumulative_error_sum_ptr->absolute_pixel_error),
                   (max_ptr->absolute_pixel_error),
                   (sum_sqr_ptr->absolute_pixel_error),
                   &(error_stat_one_ptr->absolute_pixel_error),
                   &(error_stat_two_ptr->absolute_pixel_error)));

    ERE(calculate_cumulative_error_stats(
                        (count_ptr->diag_error),
                        (default_to_dull_count_ptr->diag_error),
                        (cumulative_error_sum_ptr->diag_error),
                        (max_ptr->diag_error),
                        (sum_sqr_ptr->diag_error),
                        &(error_stat_one_ptr->diag_error),
                        &(error_stat_two_ptr->diag_error)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_cumulative_error_by_sharp
(
    Error_by_sharp* count_ptr,
    Error_by_sharp* default_to_dull_count_ptr,
    Error_by_sharp* cumulative_error_sum_ptr,
    Error_by_sharp* max_ptr,
    Error_by_sharp* sum_sqr_ptr,
    Error_by_sharp* error_stat_one_ptr,
    Error_by_sharp* error_stat_two_ptr
)
{
    int i;

    for (i=0; i<NUM_SHARP_METHODS; i++)
    {
        ERE(calculate_cumulative_error_by_algorithm(
                             &(count_ptr->sharp[ i ]),
                             &(default_to_dull_count_ptr->sharp[ i ]),
                             &(cumulative_error_sum_ptr->sharp[ i ]),
                             &(max_ptr->sharp[ i ]),
                             &(sum_sqr_ptr->sharp[ i ]),
                             &(error_stat_one_ptr->sharp[ i ]),
                             &(error_stat_two_ptr->sharp[ i ])));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_cumulative_error_by_algorithm
(
    Error_by_algorithm* count_ptr,
    Error_by_algorithm* default_to_dull_count_ptr,
    Error_by_algorithm* cumulative_error_sum_ptr,
    Error_by_algorithm* max_ptr,
    Error_by_algorithm* sum_sqr_ptr,
    Error_by_algorithm* error_stat_one_ptr,
    Error_by_algorithm* error_stat_two_ptr
)
{
    int i;

    ERE(calculate_best_fit_cumulative_error(
                                    &(count_ptr->best_fit),
                                    &(default_to_dull_count_ptr->best_fit),
                                    &(cumulative_error_sum_ptr->best_fit),
                                    &(max_ptr->best_fit),
                                    &(sum_sqr_ptr->best_fit),
                                    &(error_stat_one_ptr->best_fit),
                                    &(error_stat_two_ptr->best_fit)));

    ERE(calculate_standard_cumulative_error(
                                    &(count_ptr->standard),
                                    &(default_to_dull_count_ptr->standard),
                                    &(cumulative_error_sum_ptr->standard),
                                    &(max_ptr->standard),
                                    &(sum_sqr_ptr->standard),
                                    &(error_stat_one_ptr->standard),
                                    &(error_stat_two_ptr->standard)));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(calculate_gamut_cumulative_error(
                                    &(count_ptr->gamut_2D[ i ]),
                                    &(default_to_dull_count_ptr->gamut_2D[ i ]),
                                    &(cumulative_error_sum_ptr->gamut_2D[ i ]),
                                    &(max_ptr->gamut_2D[ i ]),
                                    &(sum_sqr_ptr->gamut_2D[ i ]),
                                    &(error_stat_one_ptr->gamut_2D[ i ]),
                                    &(error_stat_two_ptr->gamut_2D[ i ])));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(calculate_gamut_cumulative_error(
                                    &(count_ptr->gamut_3D[ i ]),
                                    &(default_to_dull_count_ptr->gamut_3D[ i ]),
                                    &(cumulative_error_sum_ptr->gamut_3D[ i ]),
                                    &(max_ptr->gamut_3D[ i ]),
                                    &(sum_sqr_ptr->gamut_3D[ i ]),
                                    &(error_stat_one_ptr->gamut_3D[ i ]),
                                    &(error_stat_two_ptr->gamut_3D[ i ])));
    }

    /* Lindsay: Feb 17/98 */
    ERE(calculate_correlation_cumulative_error(
                                 &(count_ptr->correlation),
                                 &(default_to_dull_count_ptr->correlation),
                                 &(cumulative_error_sum_ptr->correlation),
                                 &(max_ptr->correlation),
                                 &(sum_sqr_ptr->correlation),
                                 &(error_stat_one_ptr->correlation),
                                 &(error_stat_two_ptr->correlation)));
    /* End Lindsay: Feb 17/98 */

#ifdef SIMPLE_TEMPLATE
    ERE(calculate_simple_template_cumulative_error(
                                    &(count_ptr->simple_template),
                                    &(default_to_dull_count_ptr->simple_template),
                                    &(cumulative_error_sum_ptr->simple_template),
                                    &(max_ptr->simple_template),
                                    &(sum_sqr_ptr->simple_template),
                                    &(error_stat_one_ptr->simple_template),
                                    &(error_stat_two_ptr->simple_template)));
#endif

#ifdef TEMPLATE
    ERE(calculate_template_cumulative_error(
                                    &(count_ptr->template),
                                    &(default_to_dull_count_ptr->template),
                                    &(cumulative_error_sum_ptr->template),
                                    &(max_ptr->template),
                                    &(sum_sqr_ptr->template),
                                    &(error_stat_one_ptr->template),
                                    &(error_stat_two_ptr->template)));
#endif

#ifdef MULTI_TEMPLATE
    ERE(calculate_multi_template_cumulative_error(
                                    &(count_ptr->multi_template),
                                    &(default_to_dull_count_ptr->multi_template),
                                    &(cumulative_error_sum_ptr->multi_template),
                                    &(max_ptr->multi_template),
                                    &(sum_sqr_ptr->multi_template),
                                    &(error_stat_one_ptr->multi_template),
                                    &(error_stat_two_ptr->multi_template)));
#endif

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_best_fit_cumulative_error
(
    Best_fit_error* count_ptr,
    Best_fit_error* default_to_dull_count_ptr,
    Best_fit_error* cumulative_error_sum_ptr,
    Best_fit_error* max_ptr,
    Best_fit_error* sum_sqr_ptr,
    Best_fit_error* error_stat_one_ptr,
    Best_fit_error* error_stat_two_ptr
)
{

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->world_best_3D_fit),
                            (default_to_dull_count_ptr->world_best_3D_fit),
                            (cumulative_error_sum_ptr->world_best_3D_fit),
                            (max_ptr->world_best_3D_fit),
                            (sum_sqr_ptr->world_best_3D_fit),
                            &(error_stat_one_ptr->world_best_3D_fit),
                            &(error_stat_two_ptr->world_best_3D_fit)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->world_best_2D_fit),
                            (default_to_dull_count_ptr->world_best_2D_fit),
                            (cumulative_error_sum_ptr->world_best_2D_fit),
                            (max_ptr->world_best_2D_fit),
                            (sum_sqr_ptr->world_best_2D_fit),
                            &(error_stat_one_ptr->world_best_2D_fit),
                            &(error_stat_two_ptr->world_best_2D_fit)));

    ERE(calculate_cumulative_error_stats(
                    (count_ptr->world_best_3D_diag_fit),
                    (default_to_dull_count_ptr->world_best_3D_diag_fit),
                    (cumulative_error_sum_ptr->world_best_3D_diag_fit),
                    (max_ptr->world_best_3D_diag_fit),
                    (sum_sqr_ptr->world_best_3D_diag_fit),
                    &(error_stat_one_ptr->world_best_3D_diag_fit),
                    &(error_stat_two_ptr->world_best_3D_diag_fit)));

    ERE(calculate_cumulative_error_stats(
                    (count_ptr->world_best_2D_diag_fit),
                    (default_to_dull_count_ptr->world_best_2D_diag_fit),
                    (cumulative_error_sum_ptr->world_best_2D_diag_fit),
                    (max_ptr->world_best_2D_diag_fit),
                    (sum_sqr_ptr->world_best_2D_diag_fit),
                    &(error_stat_one_ptr->world_best_2D_diag_fit),
                    &(error_stat_two_ptr->world_best_2D_diag_fit)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->scene_best_3D_fit),
                            (default_to_dull_count_ptr->scene_best_3D_fit),
                            (cumulative_error_sum_ptr->scene_best_3D_fit),
                            (max_ptr->scene_best_3D_fit),
                            (sum_sqr_ptr->scene_best_3D_fit),
                            &(error_stat_one_ptr->scene_best_3D_fit),
                            &(error_stat_two_ptr->scene_best_3D_fit)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->scene_best_2D_fit),
                            (default_to_dull_count_ptr->scene_best_2D_fit),
                            (cumulative_error_sum_ptr->scene_best_2D_fit),
                            (max_ptr->scene_best_2D_fit),
                            (sum_sqr_ptr->scene_best_2D_fit),
                            &(error_stat_one_ptr->scene_best_2D_fit),
                            &(error_stat_two_ptr->scene_best_2D_fit)));

    ERE(calculate_cumulative_error_stats(
                    (count_ptr->scene_best_3D_diag_fit),
                    (default_to_dull_count_ptr->scene_best_3D_diag_fit),
                    (cumulative_error_sum_ptr->scene_best_3D_diag_fit),
                    (max_ptr->scene_best_3D_diag_fit),
                    (sum_sqr_ptr->scene_best_3D_diag_fit),
                    &(error_stat_one_ptr->scene_best_3D_diag_fit),
                    &(error_stat_two_ptr->scene_best_3D_diag_fit)));

    ERE(calculate_cumulative_error_stats(
                    (count_ptr->scene_best_2D_diag_fit),
                    (default_to_dull_count_ptr->scene_best_2D_diag_fit),
                    (cumulative_error_sum_ptr->scene_best_2D_diag_fit),
                    (max_ptr->scene_best_2D_diag_fit),
                    (sum_sqr_ptr->scene_best_2D_diag_fit),
                    &(error_stat_one_ptr->scene_best_2D_diag_fit),
                    &(error_stat_two_ptr->scene_best_2D_diag_fit)));

    ERE(calculate_cumulative_error_stats(
                                (count_ptr->actual),
                                (default_to_dull_count_ptr->actual),
                                (cumulative_error_sum_ptr->actual),
                                (max_ptr->actual),
                                (sum_sqr_ptr->actual),
                                &(error_stat_one_ptr->actual),
                                &(error_stat_two_ptr->actual)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_standard_cumulative_error
(
    Standard_error* count_ptr,
    Standard_error* default_to_dull_count_ptr,
    Standard_error* cumulative_error_sum_ptr,
    Standard_error* max_ptr,
    Standard_error* sum_sqr_ptr,
    Standard_error* error_stat_one_ptr,
    Standard_error* error_stat_two_ptr
)
{

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->nothing),
                            (default_to_dull_count_ptr->nothing),
                            (cumulative_error_sum_ptr->nothing),
                            (max_ptr->nothing),
                            (sum_sqr_ptr->nothing),
                            &(error_stat_one_ptr->nothing),
                            &(error_stat_two_ptr->nothing)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->ave_illum),
                            (default_to_dull_count_ptr->ave_illum),
                            (cumulative_error_sum_ptr->ave_illum),
                            (max_ptr->ave_illum),
                            (sum_sqr_ptr->ave_illum),
                            &(error_stat_one_ptr->ave_illum),
                            &(error_stat_two_ptr->ave_illum)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->grey_world),
                            (default_to_dull_count_ptr->grey_world),
                            (cumulative_error_sum_ptr->grey_world),
                            (max_ptr->grey_world),
                            (sum_sqr_ptr->grey_world),
                            &(error_stat_one_ptr->grey_world),
                            &(error_stat_two_ptr->grey_world)));

    ERE(calculate_cumulative_error_stats(
                (count_ptr->db_grey_world),
                (default_to_dull_count_ptr->db_grey_world),
                (cumulative_error_sum_ptr->db_grey_world),
                (max_ptr->db_grey_world),
                (sum_sqr_ptr->db_grey_world),
                &(error_stat_one_ptr->db_grey_world),
                &(error_stat_two_ptr->db_grey_world)));

    ERE(calculate_cumulative_error_stats(
                            (count_ptr->scale_by_max),
                            (default_to_dull_count_ptr->scale_by_max),
                            (cumulative_error_sum_ptr->scale_by_max),
                            (max_ptr->scale_by_max),
                            (sum_sqr_ptr->scale_by_max),
                            &(error_stat_one_ptr->scale_by_max),
                            &(error_stat_two_ptr->scale_by_max)));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_gamut_cumulative_error
(
    Gamut_error* count_ptr,
    Gamut_error* default_to_dull_count_ptr,
    Gamut_error* cumulative_error_sum_ptr,
    Gamut_error* max_ptr,
    Gamut_error* sum_sqr_ptr,
    Gamut_error* error_stat_one_ptr,
    Gamut_error* error_stat_two_ptr
)
{
    int i;

    ERE(calculate_cumulative_error_stats(
                                count_ptr->mv_surface,
                                default_to_dull_count_ptr->mv_surface,
                                cumulative_error_sum_ptr->mv_surface,
                                max_ptr->mv_surface,
                                sum_sqr_ptr->mv_surface,
                                &(error_stat_one_ptr->mv_surface),
                                &(error_stat_two_ptr->mv_surface)));

    ERE(calculate_cumulative_error_stats(
                                count_ptr->mv_surface_illum,
                                default_to_dull_count_ptr->mv_surface_illum,
                                cumulative_error_sum_ptr->mv_surface_illum,
                                max_ptr->mv_surface_illum,
                                sum_sqr_ptr->mv_surface_illum,
                                &(error_stat_one_ptr->mv_surface_illum),
                                &(error_stat_two_ptr->mv_surface_illum)));

    ERE(calculate_cumulative_error_stats(
                                count_ptr->ha_surface,
                                default_to_dull_count_ptr->ha_surface,
                                cumulative_error_sum_ptr->ha_surface,
                                max_ptr->ha_surface,
                                sum_sqr_ptr->ha_surface,
                                &(error_stat_one_ptr->ha_surface),
                                &(error_stat_two_ptr->ha_surface)));

    ERE(calculate_cumulative_error_stats(
                                count_ptr->ha_surface_illum,
                                default_to_dull_count_ptr->ha_surface_illum,
                                cumulative_error_sum_ptr->ha_surface_illum,
                                max_ptr->ha_surface_illum,
                                sum_sqr_ptr->ha_surface_illum,
                                &(error_stat_one_ptr->ha_surface_illum),
                                &(error_stat_two_ptr->ha_surface_illum)));

    ERE(calculate_cumulative_error_stats(
                              count_ptr->icsa_surface_illum,
                              default_to_dull_count_ptr->icsa_surface_illum,
                              cumulative_error_sum_ptr->icsa_surface_illum,
                              max_ptr->icsa_surface_illum,
                              sum_sqr_ptr->icsa_surface_illum,
                              &(error_stat_one_ptr->icsa_surface_illum),
                              &(error_stat_two_ptr->icsa_surface_illum)));

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        ERE(calculate_cumulative_error_stats(
                              count_ptr->scia_surface_illum[ i ],
                             default_to_dull_count_ptr->scia_surface_illum[ i ],
                              cumulative_error_sum_ptr->scia_surface_illum[ i ],
                              max_ptr->scia_surface_illum[ i ],
                              sum_sqr_ptr->scia_surface_illum[ i ],
                              &(error_stat_one_ptr->scia_surface_illum[ i ]),
                              &(error_stat_two_ptr->scia_surface_illum[ i ])));

        ERE(calculate_cumulative_error_stats(
                              count_ptr->scca_surface_illum[ i ],
                             default_to_dull_count_ptr->scca_surface_illum[ i ],
                              cumulative_error_sum_ptr->scca_surface_illum[ i ],
                              max_ptr->scca_surface_illum[ i ],
                              sum_sqr_ptr->scca_surface_illum[ i ],
                              &(error_stat_one_ptr->scca_surface_illum[ i ]),
                              &(error_stat_two_ptr->scca_surface_illum[ i ])));
    }

    ERE(calculate_cumulative_error_stats(
                                       count_ptr->wa_surface,
                                       default_to_dull_count_ptr->wa_surface,
                                       cumulative_error_sum_ptr->wa_surface,
                                       max_ptr->wa_surface,
                                       sum_sqr_ptr->wa_surface,
                                       &(error_stat_one_ptr->wa_surface),
                                       &(error_stat_two_ptr->wa_surface)));

    ERE(calculate_cumulative_error_stats(
                                count_ptr->wa_surface_illum,
                                default_to_dull_count_ptr->wa_surface_illum,
                                cumulative_error_sum_ptr->wa_surface_illum,
                                max_ptr->wa_surface_illum,
                                sum_sqr_ptr->wa_surface_illum,
                                &(error_stat_one_ptr->wa_surface_illum),
                                &(error_stat_two_ptr->wa_surface_illum)));

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                count_ptr->ha_surface_illum_lum[ i ],
                default_to_dull_count_ptr->ha_surface_illum_lum[ i ],
                cumulative_error_sum_ptr->ha_surface_illum_lum[ i ],
                max_ptr->ha_surface_illum_lum[ i ],
                sum_sqr_ptr->ha_surface_illum_lum[ i ],
                &(error_stat_one_ptr->ha_surface_illum_lum[ i ]),
                &(error_stat_two_ptr->ha_surface_illum_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                                count_ptr->mv_surface_illum_lum[ i ],
                default_to_dull_count_ptr->mv_surface_illum_lum[ i ],
                cumulative_error_sum_ptr->mv_surface_illum_lum[ i ],
                max_ptr->mv_surface_illum_lum[ i ],
                sum_sqr_ptr->mv_surface_illum_lum[ i ],
                &(error_stat_one_ptr->mv_surface_illum_lum[ i ]),
                &(error_stat_two_ptr->mv_surface_illum_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                count_ptr->ha_surface_lum[ i ],
                default_to_dull_count_ptr->ha_surface_lum[ i ],
                cumulative_error_sum_ptr->ha_surface_lum[ i ],
                max_ptr->ha_surface_lum[ i ],
                sum_sqr_ptr->ha_surface_lum[ i ],
                &(error_stat_one_ptr->ha_surface_lum[ i ]),
                &(error_stat_two_ptr->ha_surface_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                count_ptr->mv_surface_lum[ i ],
                default_to_dull_count_ptr->mv_surface_lum[ i ],
                cumulative_error_sum_ptr->mv_surface_lum[ i ],
                max_ptr->mv_surface_lum[ i ],
                sum_sqr_ptr->mv_surface_lum[ i ],
                &(error_stat_one_ptr->mv_surface_lum[ i ]),
                &(error_stat_two_ptr->mv_surface_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                count_ptr->scla_surface_illum_lum[ i ],
                default_to_dull_count_ptr->scla_surface_illum_lum[ i ],
                cumulative_error_sum_ptr->scla_surface_illum_lum[ i ],
                max_ptr->scla_surface_illum_lum[ i ],
                sum_sqr_ptr->scla_surface_illum_lum[ i ],
                &(error_stat_one_ptr->scla_surface_illum_lum[ i ]),
                &(error_stat_two_ptr->scla_surface_illum_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                count_ptr->lcsa_surface_illum_lum[ i ],
                default_to_dull_count_ptr->lcsa_surface_illum_lum[ i ],
                cumulative_error_sum_ptr->lcsa_surface_illum_lum[ i ],
                max_ptr->lcsa_surface_illum_lum[ i ],
                sum_sqr_ptr->lcsa_surface_illum_lum[ i ],
                &(error_stat_one_ptr->lcsa_surface_illum_lum[ i ]),
                &(error_stat_two_ptr->lcsa_surface_illum_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                      count_ptr->scia_surface_illum_lum[ i ],
                      default_to_dull_count_ptr->scia_surface_illum_lum[ i ],
                      cumulative_error_sum_ptr->scia_surface_illum_lum[ i ],
                      max_ptr->scia_surface_illum_lum[ i ],
                      sum_sqr_ptr->scia_surface_illum_lum[ i ],
                      &(error_stat_one_ptr->scia_surface_illum_lum[ i ]),
                      &(error_stat_two_ptr->scia_surface_illum_lum[ i ])));

        ERE(calculate_cumulative_error_stats(
                      count_ptr->scca_surface_illum_lum[ i ],
                      default_to_dull_count_ptr->scca_surface_illum_lum[ i ],
                      cumulative_error_sum_ptr->scca_surface_illum_lum[ i ],
                      max_ptr->scca_surface_illum_lum[ i ],
                      sum_sqr_ptr->scca_surface_illum_lum[ i ],
                      &(error_stat_one_ptr->scca_surface_illum_lum[ i ]),
                      &(error_stat_two_ptr->scca_surface_illum_lum[ i ])));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/* Lindsay - June 07/99 */
static int calculate_correlation_cumulative_error
(
    Correlation_error* count_ptr,
    Correlation_error* default_to_dull_count_ptr,
    Correlation_error* cumulative_error_sum_ptr,
    Correlation_error* max_ptr,
    Correlation_error* sum_sqr_ptr,
    Correlation_error* error_stat_one_ptr,
    Correlation_error* error_stat_two_ptr
)
{
    int i; /* Modified April 14/99 */

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                count_ptr->correlation_01[i],
                                default_to_dull_count_ptr->correlation_01[i],
                                cumulative_error_sum_ptr->correlation_01[i],
                                max_ptr->correlation_01[i],
                                sum_sqr_ptr->correlation_01[i],
                                &(error_stat_one_ptr->correlation_01[i]),
                                &(error_stat_two_ptr->correlation_01[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                count_ptr->correlation_01_alt[i],
                                default_to_dull_count_ptr->correlation_01_alt[i],
                                cumulative_error_sum_ptr->correlation_01_alt[i],
                                max_ptr->correlation_01_alt[i],
                                sum_sqr_ptr->correlation_01_alt[i],
                                &(error_stat_one_ptr->correlation_01_alt[i]),
                                &(error_stat_two_ptr->correlation_01_alt[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                    count_ptr->correlation_uniform_MAP_inv_dist_fill[i],
            default_to_dull_count_ptr->correlation_uniform_MAP_inv_dist_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MAP_inv_dist_fill[i],
                    max_ptr->correlation_uniform_MAP_inv_dist_fill[i],
                    sum_sqr_ptr->correlation_uniform_MAP_inv_dist_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MAP_inv_dist_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MAP_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                        count_ptr->correlation_uniform_MAP_gaussian_fill[i],
            default_to_dull_count_ptr->correlation_uniform_MAP_gaussian_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MAP_gaussian_fill[i],
                        max_ptr->correlation_uniform_MAP_gaussian_fill[i],
                        sum_sqr_ptr->correlation_uniform_MAP_gaussian_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MAP_gaussian_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MAP_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                            count_ptr->correlation_uniform_MAP_fitted[i],
                default_to_dull_count_ptr->correlation_uniform_MAP_fitted[i],
                cumulative_error_sum_ptr->correlation_uniform_MAP_fitted[i],
                                max_ptr->correlation_uniform_MAP_fitted[i],
                                sum_sqr_ptr->correlation_uniform_MAP_fitted[i],
                &(error_stat_one_ptr->correlation_uniform_MAP_fitted[i]),
                &(error_stat_two_ptr->correlation_uniform_MAP_fitted[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                        count_ptr->correlation_gaussian_MAP[i],
                        default_to_dull_count_ptr->correlation_gaussian_MAP[i],
                        cumulative_error_sum_ptr->correlation_gaussian_MAP[i],
                                    max_ptr->correlation_gaussian_MAP[i],
                                    sum_sqr_ptr->correlation_gaussian_MAP[i],
                        &(error_stat_one_ptr->correlation_gaussian_MAP[i]),
                        &(error_stat_two_ptr->correlation_gaussian_MAP[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                    count_ptr->correlation_uniform_MLM_inv_dist_fill[i],
            default_to_dull_count_ptr->correlation_uniform_MLM_inv_dist_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MLM_inv_dist_fill[i],
                    max_ptr->correlation_uniform_MLM_inv_dist_fill[i],
                    sum_sqr_ptr->correlation_uniform_MLM_inv_dist_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MLM_inv_dist_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MLM_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                        count_ptr->correlation_uniform_MLM_gaussian_fill[i],
            default_to_dull_count_ptr->correlation_uniform_MLM_gaussian_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MLM_gaussian_fill[i],
                        max_ptr->correlation_uniform_MLM_gaussian_fill[i],
                        sum_sqr_ptr->correlation_uniform_MLM_gaussian_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MLM_gaussian_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MLM_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                            count_ptr->correlation_uniform_MLM_fitted[i],
                default_to_dull_count_ptr->correlation_uniform_MLM_fitted[i],
                cumulative_error_sum_ptr->correlation_uniform_MLM_fitted[i],
                                max_ptr->correlation_uniform_MLM_fitted[i],
                                sum_sqr_ptr->correlation_uniform_MLM_fitted[i],
                &(error_stat_one_ptr->correlation_uniform_MLM_fitted[i]),
                &(error_stat_two_ptr->correlation_uniform_MLM_fitted[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                        count_ptr->correlation_gaussian_MLM[i],
                        default_to_dull_count_ptr->correlation_gaussian_MLM[i],
                        cumulative_error_sum_ptr->correlation_gaussian_MLM[i],
                                    max_ptr->correlation_gaussian_MLM[i],
                                    sum_sqr_ptr->correlation_gaussian_MLM[i],
                        &(error_stat_one_ptr->correlation_gaussian_MLM[i]),
                        &(error_stat_two_ptr->correlation_gaussian_MLM[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                    count_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
           default_to_dull_count_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
                    max_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
                    sum_sqr_ptr->correlation_uniform_MMSE_inv_dist_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MMSE_inv_dist_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MMSE_inv_dist_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                        count_ptr->correlation_uniform_MMSE_gaussian_fill[i],
           default_to_dull_count_ptr->correlation_uniform_MMSE_gaussian_fill[i],
            cumulative_error_sum_ptr->correlation_uniform_MMSE_gaussian_fill[i],
                        max_ptr->correlation_uniform_MMSE_gaussian_fill[i],
                        sum_sqr_ptr->correlation_uniform_MMSE_gaussian_fill[i],
            &(error_stat_one_ptr->correlation_uniform_MMSE_gaussian_fill[i]),
            &(error_stat_two_ptr->correlation_uniform_MMSE_gaussian_fill[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                            count_ptr->correlation_uniform_MMSE_fitted[i],
                default_to_dull_count_ptr->correlation_uniform_MMSE_fitted[i],
                cumulative_error_sum_ptr->correlation_uniform_MMSE_fitted[i],
                                max_ptr->correlation_uniform_MMSE_fitted[i],
                                sum_sqr_ptr->correlation_uniform_MMSE_fitted[i],
                &(error_stat_one_ptr->correlation_uniform_MMSE_fitted[i]),
                &(error_stat_two_ptr->correlation_uniform_MMSE_fitted[i])));
    }

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(calculate_cumulative_error_stats(
                                        count_ptr->correlation_gaussian_MMSE[i],
                        default_to_dull_count_ptr->correlation_gaussian_MMSE[i],
                        cumulative_error_sum_ptr->correlation_gaussian_MMSE[i],
                                    max_ptr->correlation_gaussian_MMSE[i],
                                    sum_sqr_ptr->correlation_gaussian_MMSE[i],
                        &(error_stat_one_ptr->correlation_gaussian_MMSE[i]),
                        &(error_stat_two_ptr->correlation_gaussian_MMSE[i])));
    }

    return NO_ERROR;
}
/* End Lindsay: June 07/99 */

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static int calculate_simple_template_cumulative_error
(
    Simple_template_error* count_ptr,
    Simple_template_error* default_to_dull_count_ptr,
    Simple_template_error* cumulative_error_sum_ptr,
    Simple_template_error* max_ptr,
    Simple_template_error* sum_sqr_ptr,
    Simple_template_error* error_stat_one_ptr,
    Simple_template_error* error_stat_two_ptr
)
{

    ERE(calculate_cumulative_error_stats(count_ptr->alg,
                                         default_to_dull_count_ptr->alg,
                                         cumulative_error_sum_ptr->alg,
                                         max_ptr->alg,
                                         sum_sqr_ptr->alg,
                                         &(error_stat_one_ptr->alg),
                                         &(error_stat_two_ptr->alg)));

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static int calculate_template_cumulative_error
(
    Template_error* count_ptr,
    Template_error* default_to_dull_count_ptr,
    Template_error* cumulative_error_sum_ptr,
    Template_error* max_ptr,
    Template_error* sum_sqr_ptr,
    Template_error* error_stat_one_ptr,
    Template_error* error_stat_two_ptr
)
{

    ERE(calculate_cumulative_error_stats(count_ptr->first_alg,
                                         default_to_dull_count_ptr->first_alg,
                                         cumulative_error_sum_ptr->first_alg,
                                         max_ptr->first_alg,
                                         sum_sqr_ptr->first_alg,
                                         &(error_stat_one_ptr->first_alg),
                                         &(error_stat_two_ptr->first_alg)));

    ERE(calculate_cumulative_error_stats(count_ptr->second_alg,
                                         default_to_dull_count_ptr->second_alg,
                                         cumulative_error_sum_ptr->second_alg,
                                         max_ptr->second_alg,
                                         sum_sqr_ptr->second_alg,
                                         &(error_stat_one_ptr->second_alg),
                                         &(error_stat_two_ptr->second_alg)));

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE

static int calculate_multi_template_cumulative_error
(
    Multi_template_error* count_ptr,
    Multi_template_error* default_to_dull_count_ptr,
    Multi_template_error* cumulative_error_sum_ptr,
    Multi_template_error* max_ptr,
    Multi_template_error* sum_sqr_ptr,
    Multi_template_error* error_stat_one_ptr,
    Multi_template_error* error_stat_two_ptr
)
{
    int i;

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        ERE(calculate_cumulative_error_stats(count_ptr->alg[ i ],
                                            default_to_dull_count_ptr->alg[ i ],
                                             cumulative_error_sum_ptr->alg[ i ],
                                             max_ptr->alg[ i ],
                                             sum_sqr_ptr->alg[ i ],
                                             &(error_stat_one_ptr->alg[ i ]),
                                             &(error_stat_two_ptr->alg[ i ])));
    }

    return NO_ERROR;
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int calculate_cumulative_error_stats
(
    double  count,
    double  default_to_dull_count,
    double  error_sum,
    double  error_max,
    double  error_sum_sqr,
    double* error_stat_one_ptr,
    double* error_stat_two_ptr
)
{

    if (error_stat_one_ptr != NULL)
    {
        ERE(calculate_cumulative_error_stat_item(count, default_to_dull_count,
                                                 error_sum, error_max,
                                                 error_sum_sqr,
                                                 fs_error_stat_one_request,
                                                 error_stat_one_ptr));
    }

    if (error_stat_two_ptr != NULL)
    {
        ERE(calculate_cumulative_error_stat_item(count, default_to_dull_count,
                                                 error_sum, error_max,
                                                 error_sum_sqr,
                                                 fs_error_stat_two_request,
                                                 error_stat_two_ptr));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int calculate_cumulative_error_stat_item
(
    double     count,
    double     default_to_dull_count,
    double     error_sum,
    double     error_max,
    double     error_sum_sqr,
    Error_stat error_stat_request,
    double*    error_stat_ptr
)
{

    /* Check for lingering use of the old method of specifying missing value. */
    ASSERT( ! IS_EQUAL_DBL(count, DBL_NOT_SET));
    ASSERT( ! IS_EQUAL_DBL(default_to_dull_count, DBL_NOT_SET));
    ASSERT( ! IS_EQUAL_DBL(error_sum, DBL_NOT_SET));
    ASSERT( ! IS_EQUAL_DBL(error_sum_sqr, DBL_NOT_SET));
    ASSERT( ! IS_EQUAL_DBL(error_max, DBL_NOT_SET));

    if (error_stat_request == COUNT_ERROR_STAT)
    {
        if (count < DBL_HALF_MOST_NEGATIVE/2.0)
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            *error_stat_ptr = count;
        }
    }
    else if (error_stat_request == DEFAULT_TO_DULL_COUNT_ERROR_STAT)
    {
        if (default_to_dull_count < DBL_HALF_MOST_NEGATIVE/2.0)
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            *error_stat_ptr = default_to_dull_count;
        }
    }
    else if (error_stat_request == RMS_ERROR_STAT)
    {
        if (error_sum_sqr < DBL_HALF_MOST_NEGATIVE/2.0)
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            *error_stat_ptr = sqrt((error_sum_sqr/(double)count));
        }
    }
    else if (error_stat_request == MEAN_ERROR_STAT)
    {
        if (error_sum < DBL_HALF_MOST_NEGATIVE/2.0)
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            *error_stat_ptr = error_sum / count;
        }
    }
    else if (error_stat_request == MAX_ERROR_STAT)
    {
        if (error_max < DBL_HALF_MOST_NEGATIVE/2.0)
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            *error_stat_ptr = error_max;
        }
    }
    else if (error_stat_request == STDEV_ERROR_STAT)
    {
        if (    (error_sum < DBL_HALF_MOST_NEGATIVE/2.0)
             || (error_sum_sqr < DBL_HALF_MOST_NEGATIVE/2.0)
           )
        {
            *error_stat_ptr = DBL_HALF_MOST_NEGATIVE;
        }
        else
        {
            double mean = error_sum / count;
            double temp = error_sum_sqr / count - mean * mean;

            if (temp < 0.0)
            {
#ifdef TEST
                dbe(temp);
#endif
                temp = 0.0;
            }

            *error_stat_ptr = sqrt(temp);
        }
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

