
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "sample/sample_gauss.h" 

#include "spectra.h"
#include "synthetic_data.h"

/* -------------------------------------------------------------------------- */

static Matrix* fs_cached_data_mp                = NULL;
static Matrix* fs_cached_canon_data_mp          = NULL;
static int     fs_data_version                  = 0;
static int     fs_canon_data_version            = 0;

static Matrix* fs_cached_matte_data_mp          = NULL;
static Matrix* fs_cached_matte_canon_data_mp    = NULL;
static int     fs_matte_data_version            = 0;
static int     fs_matte_canon_data_version      = 0;

static double    fs_gaussian_relative_noise_sigma = DBL_NOT_SET;
static double    fs_gaussian_absolute_noise_sigma = DBL_NOT_SET;
static double    fs_uniform_relative_noise_amount = DBL_NOT_SET;
static double    fs_uniform_absolute_noise_amount = DBL_NOT_SET;

/* -------------------------------------------------------------------------- */

static int make_matte_data(int dry_run);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif

/* -------------------------------------------------------------------------- */

EXTERN_C int set_synthetic_data_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "gaussian-relative-noise-sigma"))
       )
    {
        double        temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("gaussian-relative-noise-sigma = %.2f\n",
                    fs_gaussian_relative_noise_sigma));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_gaussian_relative_noise_sigma <= 0.0)
            {
                ERE(pso("Relative Gaussian noise is not added.\n"));
            }
            else
            {
                ERE(pso("Relative Gaussian noise is added with sigma %.2f.\n",
                        fs_gaussian_relative_noise_sigma));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_gaussian_relative_noise_sigma = DBL_NOT_SET;
            }
            else
            {
                ERE(ss1d(value, &temp_real));

                if (temp_real < 0.0) temp_real = NOT_SET;
                fs_gaussian_relative_noise_sigma = temp_real;
            }
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "gaussian-absolute-noise-sigma"))
       )
    {
        double        temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("gaussian-absolute-noise-sigma = %.2f\n",
                    fs_gaussian_absolute_noise_sigma));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_gaussian_absolute_noise_sigma <= 0.0)
            {
                ERE(pso("Absolute gaussian noise is not added.\n"));
            }
            else
            {
                ERE(pso("Absolute gaussian noise is added with sigma %.2f.\n",
                        fs_gaussian_absolute_noise_sigma));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_gaussian_absolute_noise_sigma = DBL_NOT_SET;
            }
            else
            {
                ERE(ss1d(value, &temp_real));

                if (temp_real < 0.0) temp_real = NOT_SET;
                fs_gaussian_absolute_noise_sigma = temp_real;
            }
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "uniform-relative-noise-amount"))
       )
    {
        double        temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("uniform-relative-noise-amount = %.2f\n",
                    fs_uniform_relative_noise_amount));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_uniform_relative_noise_amount <= 0.0)
            {
                ERE(pso("Relative uniform noise is not added.\n"));
            }
            else
            {
                ERE(pso("Relative uniform noise is added with amount %.2f.\n",
                        fs_uniform_relative_noise_amount));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_uniform_relative_noise_amount = DBL_NOT_SET;
            }
            else
            {
                ERE(ss1d(value, &temp_real));

                if (temp_real < 0.0) temp_real = NOT_SET;
                fs_uniform_relative_noise_amount = temp_real;
            }
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "uniform-absolute-noise-amount"))
       )
    {
        double        temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("uniform-absolute-noise-amount = %.2f\n",
                    fs_uniform_absolute_noise_amount));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_uniform_absolute_noise_amount <= 0.0)
            {
                ERE(pso("Absolute uniform noise is not added.\n"));
            }
            else
            {
                ERE(pso("Absolute uniform noise is added with amount %.2f.\n",
                        fs_uniform_absolute_noise_amount));
            }
        }
        else
        {
            if (is_no_value_word(value))
            {
                fs_uniform_absolute_noise_amount = DBL_NOT_SET;
            }
            else
            {
                ERE(ss1d(value, &temp_real));

                if (temp_real < 0.0) temp_real = NOT_SET;
                fs_uniform_absolute_noise_amount = temp_real;
            }
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_synthetic_data_points(void)
{

    ERE(make_data((const char*)NULL, (const char*)NULL));

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0));

    return fs_cached_data_mp->num_rows;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_synthetic_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums
)
{

    ERE(make_data((const char*)NULL, (const char*)NULL));

    ASSERT((fs_cached_data_mp != NULL) && (fs_cached_data_mp->num_rows > 0));

    ERE(select_data(target_mpp, fs_cached_data_mp, num_data_points,
                    data_point_nums, (int*)NULL));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_registered_synthetic_data
(
    Matrix** target_mpp,
    int      num_data_points,
    int*     data_point_nums
)
{

    ERE(make_data((const char*)NULL, (const char*)NULL));

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0));

    ERE(select_data(target_mpp, fs_cached_data_mp, num_data_points,
                    data_point_nums, (int*)NULL));

    return fs_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_registered_synthetic_data(Matrix** target_mpp)
{

    ERE(make_data((const char*)NULL, (const char*)NULL));

    ASSERT((fs_cached_data_mp != NULL)&&(fs_cached_data_mp->num_rows > 0));

    ERE(copy_matrix(target_mpp, fs_cached_data_mp));

    return fs_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_synthetic_matte_data(Matrix** target_mpp)
{

    ERE(make_matte_data(FALSE /* dry_run */ ));

    ASSERT(    (fs_cached_matte_data_mp != NULL)
            && (fs_cached_matte_data_mp->num_rows > 0));

    ERE(copy_matrix(target_mpp, fs_cached_matte_data_mp));

    return fs_matte_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int synthetic_matte_data_is_up_to_date(int proposed_version)
{

    if (proposed_version != fs_matte_data_version)
    {
        return FALSE;
    }
    else
    {
        return make_matte_data( TRUE  /* dry_run */ );
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_selected_synthetic_canon_data
(
    Matrix** target_mpp,
    int      num_canon_data_points,
    int*     canon_data_point_nums
)
{

    ERE(make_canon_data((char*)NULL, (char*)NULL));

    ASSERT(    (fs_cached_canon_data_mp != NULL)
            && (fs_cached_canon_data_mp->num_rows > 0));

    ERE(select_data(target_mpp, fs_cached_canon_data_mp, num_canon_data_points,
                    canon_data_point_nums, (int*)NULL));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_synthetic_canon_data(Matrix** target_mpp)
{

    ERE(make_canon_data((char*)NULL, (char*)NULL));

    ASSERT(    (fs_cached_canon_data_mp != NULL)
            && (fs_cached_canon_data_mp->num_rows > 0));

    ERE(copy_matrix(target_mpp, fs_cached_canon_data_mp));

    return fs_canon_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_synthetic_matte_canon_data(Matrix** target_mpp)
{

    ERE(make_matte_canon_data( FALSE /* dry_run */ ));

    ASSERT(    (fs_cached_matte_canon_data_mp != NULL)
            && (fs_cached_matte_canon_data_mp->num_rows > 0));

    ERE(copy_matrix(target_mpp, fs_cached_matte_canon_data_mp));

    return fs_matte_canon_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int synthetic_matte_canon_data_is_up_to_date(int proposed_version)
{

    if (proposed_version != fs_matte_canon_data_version)
    {
        return FALSE;
    }
    else
    {
        return make_matte_canon_data( TRUE  /* dry_run */ );
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_synthetic_var_data_points(void)
{
    set_error("Synthetic var data code not yet ready.");
    return ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_synthetic_var_data
(
    Matrix** __attribute__((unused)) dummy_target_mpp,
    int      __attribute__((unused)) dummy_num_data_points,
    int*     __attribute__((unused)) dummy_data_point_nums,
    Matrix** __attribute__((unused)) dummy_var_data_mpp
)
{
#ifdef NOT_YET
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version;
    Matrix*        dull_data_mp = NULL;
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR;

    ERE(initialize_var_image_data());

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0));

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else
    {
        result = select_data(data_mpp, var_cached_input_data_mp,
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp);

    return cached_var_input_data_version;
#else
    set_error("Synthetic var data code not yet ready.");
    return ERROR;
#endif

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_selected_registered_synthetic_var_data
(
    Matrix** __attribute__((unused)) dummy_target_mpp,
    int      __attribute__((unused)) dummy_num_data_points,
    int*     __attribute__((unused)) dummy_data_point_nums
)
{
#ifdef NOT_YET
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version;
    Matrix*        dull_data_mp = NULL;
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR;

    ERE(initialize_var_image_data());

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0));

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else
    {
        result = select_data(data_mpp, var_cached_input_data_mp,
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp);

    return cached_var_input_data_version;
#else
    set_error("Synthetic var data code not yet ready.");
    return ERROR;
#endif

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int get_registered_synthetic_var_data(Matrix** __attribute__((unused)) dummy_target_mpp)
{
#ifdef NOT_YET
    extern Matrix* cached_var_input_data_mp;
    extern int     cached_var_input_data_version;
    Matrix*        dull_data_mp = NULL;
    Matrix**       data_mpp  = (sharp_flag ? &dull_data_mp : target_mpp);
    int            result = NO_ERROR;

    ERE(initialize_var_image_data());

    ASSERT(    (cached_input_data_mp != NULL)
            && (cached_input_data_mp->num_rows > 0));

    if ((num_data_points == 0) || (data_point_nums == NULL))
    {
        result = copy_matrix(data_mpp, cached_var_input_data_mp);
    }
    else
    {
        result = select_data(data_mpp, var_cached_input_data_mp,
                             num_data_points, data_point_nums, (int*)NULL);
    }

    if (sharp_flag && (result != ERROR))
    {
        result = sharpen_matrix_data(target_mpp, *data_mpp);
    }

    free_matrix(dull_data_mp);

    return cached_var_input_data_version;
#else
    set_error("Synthetic var data code not yet ready.");
    return ERROR;
#endif

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_data(const char* make_directive, const char* file_name)
{
    static int  previous_sensor_version          = NOT_SET;
    static int  previous_data_version            = NOT_SET;
    static int  previous_actual_spectra_version  = NOT_SET;
    static int  previous_reflect_spectra_version = NOT_SET;
    static int  previous_fluorescent_db_version = NOT_SET;
    static double previous_abs_sigma               = DBL_NOT_SET;
    static double previous_rel_sigma               = DBL_NOT_SET;
    static double previous_abs_amount              = DBL_NOT_SET;
    static double previous_rel_amount              = DBL_NOT_SET;
    int         result                           = NO_ERROR;
    char        lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all");
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    (*lc_make_directive != '\0')
         && ! STRCMP_EQ(lc_make_directive, "all")
         && ! STRCMP_EQ(lc_make_directive, "data")
       )
    {
        return NOT_FOUND;
    }

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    ERE(make_reflect_spectra_with_specularities());

    if (    (fs_data_version == previous_data_version)
         && (actual_spectra_is_up_to_date(previous_actual_spectra_version))
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (reflect_spectra_with_spec_is_up_to_date(previous_reflect_spectra_version))
         && (fluorescent_db_is_up_to_date(previous_fluorescent_db_version))
         && (IS_EQUAL_DBL(fs_gaussian_absolute_noise_sigma, previous_abs_sigma))
         && (IS_EQUAL_DBL(fs_gaussian_relative_noise_sigma, previous_rel_sigma))
         && (IS_EQUAL_DBL(fs_uniform_absolute_noise_amount, previous_abs_amount))
         && (IS_EQUAL_DBL(fs_uniform_relative_noise_amount, previous_rel_amount))
       )
    {
        verbose_pso(10, "Data is up to date. Skipping make.\n");
    }
    else
    {
        Fluorescent_database* fl_db_ptr               = NULL;
        Spectra*              actual_sp               = NULL;
        Spectra*              reflect_sp              = NULL;
        Matrix*               data_mp                 = NULL;
        int                   actual_spectra_version;
        int                   reflect_spectra_version = NOT_SET;
        int                   fluorescent_db_version = NOT_SET;

        verbose_pso(5, "Data is out of date. Remaking it.\n");

        actual_spectra_version = get_actual_spectra(&actual_sp);

        if (actual_spectra_version == ERROR) result = ERROR;

        if (result != ERROR)
        {
            reflect_spectra_version =
                           get_reflect_spectra_with_specularities(&reflect_sp);
            if (reflect_spectra_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            fluorescent_db_version =
                                 get_optional_fluorescent_database(&fl_db_ptr);
            if (fluorescent_db_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            previous_sensor_version = generate_synthetic_RGB_data(&data_mp,
                                                                  actual_sp,
                                                                  0, reflect_sp,
                                                                  fl_db_ptr);
            if (previous_sensor_version == ERROR) result = ERROR;
        }

        if ((result != ERROR) && (fs_gaussian_relative_noise_sigma > 0.0))
        {
            int num_rows = data_mp->num_rows;
            int num_cols = data_mp->num_cols;
            int i, j;

            for (i=0; i<num_rows; i++)
            {
                for (j=0; j<num_cols; j++)
                {
                    double temp = data_mp->elements[ i ][ j ];
                    double noise = fs_gaussian_relative_noise_sigma * gauss_rand();

                    temp *= (1.0 + noise);
                    data_mp->elements[ i ][ j ] = MAX_OF(temp, 0.0);
                }
            }
        }

        if ((result != ERROR) && (fs_uniform_relative_noise_amount > 0.0))
        {
            int num_rows = data_mp->num_rows;
            int num_cols = data_mp->num_cols;
            int i, j;

            for (i=0; i<num_rows; i++)
            {
                for (j=0; j<num_cols; j++)
                {
                    double temp = data_mp->elements[ i ][ j ];
                    double noise = fs_uniform_relative_noise_amount * 2.0 * (kjb_rand()-0.5);

                    temp *= (1.0 + noise);
                    data_mp->elements[ i ][ j ] = MAX_OF(temp, 0.0);
                }
            }
        }

        if ((result != ERROR) && (fs_gaussian_absolute_noise_sigma > 0.0))
        {
            int num_rows = data_mp->num_rows;
            int num_cols = data_mp->num_cols;
            int i, j;

            for (i=0; i<num_rows; i++)
            {
                for (j=0; j<num_cols; j++)
                {
                    double temp = data_mp->elements[ i ][ j ];

                    temp += fs_gaussian_absolute_noise_sigma * gauss_rand();
                    data_mp->elements[ i ][ j ] = MAX_OF(temp, 0.0);
                }
            }
        }

        if ((result != ERROR) && (fs_uniform_absolute_noise_amount > 0.0))
        {
            int num_rows = data_mp->num_rows;
            int num_cols = data_mp->num_cols;
            int i, j;

            for (i=0; i<num_rows; i++)
            {
                for (j=0; j<num_cols; j++)
                {
                    double temp = data_mp->elements[ i ][ j ];

                    temp += fs_uniform_absolute_noise_amount * 2.0 * (kjb_rand() - 0.5);
                    data_mp->elements[ i ][ j ] = MAX_OF(temp, 0.0);
                }
            }
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_data_mp, data_mp);
        }

        if (result != ERROR)
        {
            fs_data_version++;

            previous_actual_spectra_version  = actual_spectra_version;

            ASSERT(IS_SET(reflect_spectra_version));
            previous_reflect_spectra_version = reflect_spectra_version;

            ASSERT(IS_SET(fluorescent_db_version));
            previous_fluorescent_db_version  = fluorescent_db_version;

            previous_data_version            = fs_data_version;
            previous_rel_sigma               = fs_gaussian_relative_noise_sigma;
            previous_abs_sigma               = fs_gaussian_absolute_noise_sigma;
            previous_rel_amount              = fs_uniform_relative_noise_amount;
            previous_abs_amount              = fs_uniform_absolute_noise_amount;

            verbose_pso(10, "Data remade.\n");
        }

        free_matrix(data_mp);
        free_spectra(actual_sp);
        free_spectra(reflect_sp);
        free_fluorescent_database(fl_db_ptr);
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_matrix(fs_cached_data_mp, file_name) == ERROR) result = ERROR;
    }

    return (result == ERROR) ? ERROR : fs_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_canon_data(const char* make_directive, const char* file_name)
{
    static int previous_sensor_version          = NOT_SET;
    static int previous_canon_data_version      = NOT_SET;
    static int previous_canon_spectra_version   = NOT_SET;
    static int previous_reflect_spectra_version = NOT_SET;
    static int previous_fluorescent_db_version  = NOT_SET;
    int        result                           = NO_ERROR;
    char       lc_make_directive[ 100 ];

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all");
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    (*lc_make_directive != '\0')
         && ! STRCMP_EQ(lc_make_directive, "all")
         && ! STRCMP_EQ(lc_make_directive, "canon-data")
       )
    {
        return NOT_FOUND;
    }

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    ERE(make_reflect_spectra_with_specularities());

    if (    (fs_canon_data_version == previous_canon_data_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (canon_spectra_is_up_to_date(previous_canon_spectra_version))
         && (reflect_spectra_with_spec_is_up_to_date(previous_reflect_spectra_version))
         && (fluorescent_db_is_up_to_date(previous_fluorescent_db_version))
       )
    {
        verbose_pso(10, "Canon data is up to date. Skipping make.\n");
    }
    else
    {
        Fluorescent_database* fl_db_ptr               = NULL;
        Spectra*              canon_illum_sp          = NULL;
        Spectra*              reflect_sp              = NULL;
        Matrix*               canon_data_mp           = NULL;
        int                   canon_spectra_version;
        int                   reflect_spectra_version = NOT_SET;
        int                   fluorescent_db_version = NOT_SET;

        verbose_pso(5, "Canon data is out of date. Remaking it.\n");

        canon_spectra_version = get_canon_spectra(&canon_illum_sp);

        if (canon_spectra_version == ERROR) result = ERROR;

        if (result != ERROR)
        {
            reflect_spectra_version =
                           get_reflect_spectra_with_specularities(&reflect_sp);
            if (reflect_spectra_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            fluorescent_db_version =
                                 get_optional_fluorescent_database(&fl_db_ptr);
            if (fluorescent_db_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            previous_sensor_version = generate_synthetic_RGB_data(
                                                            &canon_data_mp,
                                                            canon_illum_sp, 0,
                                                            reflect_sp,
                                                            fl_db_ptr);

            if (previous_sensor_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_canon_data_mp, canon_data_mp);
        }

        if (result != ERROR)
        {
            fs_canon_data_version++;

            previous_canon_spectra_version   = canon_spectra_version;

            ASSERT(IS_SET(reflect_spectra_version));
            previous_reflect_spectra_version = reflect_spectra_version;

            ASSERT(IS_SET(fluorescent_db_version));
            previous_fluorescent_db_version = fluorescent_db_version;

            previous_canon_data_version      = fs_canon_data_version;

            verbose_pso(10, "Canon data remade.\n");
        }

        free_matrix(canon_data_mp);
        free_spectra(canon_illum_sp);
        free_spectra(reflect_sp);
        free_fluorescent_database(fl_db_ptr);
    }

    if ((result != ERROR) && (file_name != NULL) && (file_name[ 0 ] != '\0'))
    {
        if (write_matrix(fs_cached_canon_data_mp, file_name) == ERROR)
        {
            result = ERROR;
        }
    }

    return (result == ERROR) ? ERROR : fs_canon_data_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int make_matte_data(int dry_run)
{
    static int previous_sensor_version          = NOT_SET;
    static int previous_matte_data_version      = NOT_SET;
    static int previous_actual_spectra_version  = NOT_SET;
    static int previous_reflect_spectra_version = NOT_SET;
    static int previous_fluorescent_db_version  = NOT_SET;
    int        result                           = NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    if (    (fs_matte_data_version == previous_matte_data_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (actual_spectra_is_up_to_date(previous_actual_spectra_version))
         && (reflect_spectra_is_up_to_date(previous_reflect_spectra_version))
         && (fluorescent_db_is_up_to_date(previous_fluorescent_db_version))
       )
    {
        if (dry_run) return TRUE;

        verbose_pso(10, "Matte data is up to date. Skipping make.\n");
    }
    else
    {
        Fluorescent_database* fl_db_ptr               = NULL;
        Spectra*              actual_sp               = NULL;
        Spectra*              reflect_sp              = NULL;
        Matrix*               data_mp                 = NULL;
        int                   actual_spectra_version;
        int                   reflect_spectra_version = NOT_SET;
        int                   fluorescent_db_version = NOT_SET;

        if (dry_run) return FALSE;

        verbose_pso(5, "Matte data is out of date. Remaking it.\n");

        if ((actual_spectra_version = get_actual_spectra(&actual_sp)) == ERROR)
        {
            result = ERROR;
        }
        else if ((reflect_spectra_version = get_reflect_spectra(&reflect_sp))
                                                                       == ERROR)
        {
            result = ERROR;
        }
        else if ((fluorescent_db_version = get_optional_fluorescent_database(&fl_db_ptr))
                                                                       == ERROR)
        {
            result = ERROR;
        }
        else
        {
            previous_sensor_version = generate_RGB_data(&data_mp, actual_sp,
                                                        0, reflect_sp,
                                                        (Spectra*)NULL);
            if (previous_sensor_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_matte_data_mp, data_mp);
        }

        if (result != ERROR)
        {
            fs_matte_data_version++;

            previous_actual_spectra_version  = actual_spectra_version;
            previous_reflect_spectra_version = reflect_spectra_version;
            previous_fluorescent_db_version  = fluorescent_db_version;
            previous_matte_data_version      = fs_matte_data_version;

            verbose_pso(10, "Matte data remade.\n");
        }

        free_matrix(data_mp);
        free_spectra(actual_sp);
        free_spectra(reflect_sp);
        free_fluorescent_database(fl_db_ptr);

    }

    if (result != ERROR)
    {
        result = fs_matte_data_version;
    }

    return result;  /* Either ERROR or version. */
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_matte_canon_data(int dry_run)
{
    static int previous_sensor_version           = NOT_SET;
    static int previous_matte_canon_data_version = NOT_SET;
    static int previous_canon_spectra_version   = NOT_SET;
    static int previous_reflect_spectra_version  = NOT_SET;
    static int previous_fluorescent_db_version  = NOT_SET;
    int        result                            = NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    if (    (fs_matte_canon_data_version == previous_matte_canon_data_version)
         && (sensor_data_is_up_to_date(previous_sensor_version))
         && (canon_spectra_is_up_to_date(previous_canon_spectra_version))
         && (reflect_spectra_is_up_to_date(previous_reflect_spectra_version))
         && (fluorescent_db_is_up_to_date(previous_fluorescent_db_version))
       )
    {
        if (dry_run) return TRUE;

        verbose_pso(10, "Matte canon data is up to date. Skipping make.\n");
    }
    else
    {
        Fluorescent_database* fl_db_ptr                  = NULL;
        Spectra*              canon_sp                   = NULL;
        Spectra*              reflect_sp                 = NULL;
        Matrix*               data_mp                    = NULL;
        int                   canon_spectra_version;
        int                   reflect_spectra_version = NOT_SET;
        int                   fluorescent_db_version = NOT_SET;

        if (dry_run) return FALSE;

        verbose_pso(5, "Matte canon data is out of date. Remaking it.\n");

        if ((canon_spectra_version = get_canon_spectra(&canon_sp)) == ERROR)
        {
            result = ERROR;
        }
        else if ((reflect_spectra_version = get_reflect_spectra(&reflect_sp))
                                                                      == ERROR)
        {
            result = ERROR;
        }
        else if ((fluorescent_db_version = get_optional_fluorescent_database(&fl_db_ptr))
                                                                      == ERROR)
        {
            result = ERROR;
        }
        else
        {
            previous_sensor_version = generate_RGB_data(&data_mp, canon_sp,
                                                        0, reflect_sp,
                                                        (Spectra*)NULL);
            if (previous_sensor_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_matte_canon_data_mp, data_mp);
        }

        if (result != ERROR)
        {
            fs_matte_canon_data_version++;

            previous_canon_spectra_version    = canon_spectra_version;
            previous_reflect_spectra_version  = reflect_spectra_version;
            previous_fluorescent_db_version   = fluorescent_db_version;
            previous_matte_canon_data_version = fs_matte_canon_data_version;

            verbose_pso(10, "Matte canon data remade.\n");
        }

        free_matrix(data_mp);
        free_spectra(canon_sp);
        free_spectra(reflect_sp);
        free_fluorescent_database(fl_db_ptr);

    }

    if (result != ERROR)
    {
        result = fs_matte_canon_data_version;
    }

    return result;  /* Either ERROR or version. */
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_matrix(fs_cached_data_mp);
    free_matrix(fs_cached_canon_data_mp);
    free_matrix(fs_cached_matte_data_mp);
    free_matrix(fs_cached_matte_canon_data_mp);
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

