
/*
// KOBUS : Modified this code (written by Mike Brockington) to
// be more consistent with the rest of the gamut code.
*/

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "black_body_spectrum.h"

#define C1     3.741832e-16    /*  W m^2  */
#define C2     1.4388e-2       /*  m K    */

/* -------------------------------------------------------------------------- */

double* black_body_spectrum
(
    double   temp,
    double   minwave,
    double   maxwave,
    double   wavestep,
    double** spectrum,
    int*     out_nelm,
    int      verbose
)
{
    double wave;
    double wave2;
    double power;
    int    nelm;
    double *outspect;
    int ind = 0;
    int do_alloc = 0;

    if (temp < 0.0)
    {
        set_error("black_body_spectrum error: temp < 0.0." );
        return( NULL );
    }

    if (wavestep == 0.0)
    {
        set_error("black_body_spectrum error: wavestep=0, infinte loop." );
        return( NULL );
    }

    if (minwave > maxwave)
    {
        set_error("black_body_spectrum error: minwave > maxwave." );
        return( NULL );
    }
    else
    {
        if (wavestep < 0.0)
        {
            if (verbose > 1)
            {
                p_stderr("black_body_spectrum - Warning: wavestep going wrong direction.\n" );
                p_stderr("  (will reverse sign.)\n" );
            }
            wavestep = -wavestep;
        }
    }

    nelm = (int)(fabs( (maxwave - minwave) / wavestep ) + 1);

    if ((spectrum == NULL) || (*spectrum == NULL))
        ++do_alloc;
    else if  (*out_nelm < nelm)
    {
        ++do_alloc;
        kjb_free( *spectrum );
    }

    if (do_alloc)
    {
        if ((outspect = (double *)kjb_calloc( (unsigned int)nelm, sizeof(double) )) == NULL)
        {
            set_error( "black_body_spectrum: couldn't allocate spectrum." );
            return( NULL );
        }
        if (spectrum != NULL)
            *spectrum = outspect;
        *out_nelm = nelm;
        if (verbose > 1)
            p_stderr("black_body_spectrum: Allocating new spectrum vector.\n" );
    }
    else
    {
        outspect = *spectrum;
        *out_nelm = nelm;
        if (verbose > 1)
            p_stderr("black_body_spectrum: Reusing user's spectrum vector.\n" );
    }

    /*
     * Calculate blackbody radiant exitance for each wavelength
     * Main loop is inclusive of the maxwave value
    */

    for (wave=minwave; wave<=maxwave; wave+=wavestep)
    {
        /*
         * Convert wavelength from nm to m to match units of constants.
         */

        wave2 = wave / 1.0e9;

        power = (C1 / pow( wave2, 5.0)) /
            ( exp( C2 / (wave2 * temp)) - 1.0);

        outspect[ind++] = power;
    }

    return( outspect );
}

