
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "illum_data.h"
#include "image_input.h"
#include "image_input.h"
#include "compute_output.h"

/* -------------------------------------------------------------------------- */

typedef enum Output_brightness_method
{
    DONT_ADJUST_BRIGHTNESS,
    ADJUST_MAXIMUM_BRIGHTNESS,
    ADJUST_AVERGE_BRIGHTNESS,
    ADJUST_PIXEL_BRIGHTNESS
}
Output_brightness_method;

typedef enum Output_brightness_source
{
    USE_INPUT_IMAGE,
    USE_TARGET_IMAGE,
    USE_PRESCRIBED_VALUE
}
Output_brightness_source;

/* -------------------------------------------------------------------------- */

static double                   fs_output_brightness_value = 100.0;
static Output_brightness_method fs_output_brightness_method =
                                                 ADJUST_MAXIMUM_BRIGHTNESS;
static Output_brightness_source fs_output_brightness_source = USE_TARGET_IMAGE;

static int                      fs_normalize_3D_output = FALSE;

/* -------------------------------------------------------------------------- */

static int get_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          diag_vp
);

static int get_2D_sharp_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          diag_vp
);

static int get_3D_sharp_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          diag_vp
);

static int get_2D_map_image(KJB_image**, const KJB_image*, Matrix*);

static int get_3D_map_image(KJB_image**, const KJB_image*, Matrix*);

static int compute_illumination_image
(
    KJB_image**     target_ipp,
    Image_map_data* image_map_data_ptr
);

static int get_illumination
(
    Vector**     illum_estimate_vpp,
    Vector*      diag_vp,
    Sharp_method sharp_method
);

static int ow_normalize_image
(
    KJB_image*       ip,
    const KJB_image* input_ip,
    const KJB_image* canon_ip
);

static int ow_match_image_brightness
(
    KJB_image*       target_ip,
    const KJB_image* image_to_match_ip
);

static int ow_normalize_data
(
    Matrix*       data_mp,
    const Matrix* input_data_mp,
    const Matrix* canon_data_mp
);

static int ow_match_data_brightness
(
    Matrix*       target_mp,
    const Matrix* image_to_match_mp
);

/* -------------------------------------------------------------------------- */

EXTERN_C int set_output_image_options(const char* option, const char* value)
{
    char   lc_option[ 100 ];
    double temp_double;
    int    temp_boolean_value;
    int    result             = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (    (lc_option[ 0 ] == '\0') || (match_pattern(lc_option, "obv"))
         || (match_pattern(lc_option, "output-brightness-value"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Output brightness value (obv) is %5.1f.\n",
                    fs_output_brightness_value));
        }
        else
        {
            ERE(ss1d(value, &temp_double));

            if (temp_double < 0.0)
            {
                set_error(
                         "Output-brightness-value (obv) must be non-negative.");
                return ERROR;
            }

            fs_output_brightness_value = temp_double;
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "obm"))
         || (match_pattern(lc_option, "output-brightness-method"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_output_brightness_method == DONT_ADJUST_BRIGHTNESS)
            {
                ERE(pso("Output image brightness is not adjusted. "));
                ERE(pso("(obm=none)\n"));
            }
            else if (fs_output_brightness_method == ADJUST_AVERGE_BRIGHTNESS)
            {
                ERE(pso("Output image brightness is adjusted by image "));
                ERE(pso("averages. (obm=ave)\n"));
            }
            else if (fs_output_brightness_method == ADJUST_MAXIMUM_BRIGHTNESS)
            {
                ERE(pso("Output image brightness is adjusted by image "));
                ERE(pso("maximums. (obm=max)\n"));
            }
            else if (fs_output_brightness_method == ADJUST_PIXEL_BRIGHTNESS)
            {
                ERE(pso("Output image brightness is adjusted pixel by "));
                ERE(pso("pixel. (obm=pixel)\n"));
            }
            else
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }
        }
        else
        {
            if (IC_STRCMP_EQ(value, "ave"))
            {
                fs_output_brightness_method = ADJUST_AVERGE_BRIGHTNESS;
            }
            else if (IC_STRCMP_EQ(value, "max"))
            {
                fs_output_brightness_method = ADJUST_MAXIMUM_BRIGHTNESS;
            }
            else if (IC_STRCMP_EQ(value, "pixel"))
            {
                fs_output_brightness_method = ADJUST_PIXEL_BRIGHTNESS;
            }
            else if (    (IC_STRCMP_EQ(value, "off"))
                      || (IC_STRCMP_EQ(value, "none"))
                    )
            {
                fs_output_brightness_method = DONT_ADJUST_BRIGHTNESS;
            }
            else
            {
                set_error("%q is an invalid output brightness method.", value);
                return ERROR;
            }
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "obs"))
         || (match_pattern(lc_option, "output-brightness-source"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_output_brightness_source == USE_INPUT_IMAGE)
            {
                ERE(pso("Output image brightness is adjusted relative to "));
                ERE(pso("input image. (obs=input)\n"));
            }
            else if (fs_output_brightness_source == USE_TARGET_IMAGE)
            {
                ERE(pso("Output image brightness is adjusted relative to "));
                ERE(pso("target image. (obs=target)\n"));
            }
            else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
            {
                ERE(pso("Output image brightness is adjusted to "));
                ERE(pso("prescribed value. (obs=value)\n"));
            }
            else
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }
        }
        else
        {
            if (IC_STRCMP_EQ(value, "input"))
            {
                fs_output_brightness_source = USE_INPUT_IMAGE;
            }
            else if (IC_STRCMP_EQ(value, "target"))
            {
                fs_output_brightness_source = USE_TARGET_IMAGE;
            }
            else if (IC_STRCMP_EQ(value, "value"))
            {
                fs_output_brightness_source = USE_PRESCRIBED_VALUE;
            }
            else
            {
                set_error("%q is an invalid output brightness source.", value);
                return ERROR;
            }
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "n3do"))
         || (match_pattern(lc_option, "n-3d-o"))
         || (match_pattern(lc_option, "norm-3d-output"))
         || (match_pattern(lc_option, "normalize-3d-output"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("3D output %s normalized.\n",
                    fs_normalize_3D_output ? "is" : "is not"));
        }
        else
        {
            ERE(temp_boolean_value = get_boolean_value(value));
            fs_normalize_3D_output = temp_boolean_value;
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int print_output_dependence_message(void)
{
    ERE(pso("Note :\n"));
    ERE(pso("   2D results depend on output brightness settings.\n"));
    ERE(pso("   "));

    if (fs_normalize_3D_output)
    {
        ERE(pso("3D results depend on output brightness settings.\n"));
    }
    else
    {
        ERE(pso("Some 3D results depend on cif/ciw brightnesses.\n"));
    }

    ERE(pso("\n"));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// This routine takes the user input in "request" and determines the image and
// map needed to calculate the results. This information is put into
// *image_map_data_ptr, which is used by compute_result_image() to compute the
// result. Although it may seem more elegent to simply compute the result, I
// broke the work into two parts for several reasons. First, this makes it
// possible to completely error check a list of requests without doing time
// consuming image processing. Second, it is more flexible, as the information
// stored is not limited to driving compute_result_image().
//
// I have decided not to use the following, and so it is neither currently used,
// nor tested. However, it may be useful in the future, so I will leave it in
// for a version or two. The unused code segments are preceeded by the
//     This routine has become even more ugly now : It also computes a
//     corresponding error pointer based on the user input request. So, in
//     short, this routine can be seen to be a major hack job which takes user
//     requests and calculates various internal things based on it, in order to
//     have the parsing all in one place.
*/
int parse_image_request
(
    char*                    request,
    Colour_constancy_result* result_ptr,
    int                      reduction_flag,
    int                      illumination_image_flag,
    int                      chromaticity_image_flag,
    Image_map_data*          image_map_data_ptr,
    Error_by_sharp*          error_ptr,
    double**                 error_item_ptr_ptr
)
{
    char lc_request[ 100 ];

    BUFF_CPY(image_map_data_ptr->request, request);
    image_map_data_ptr->illumination_image_flag = illumination_image_flag;
    image_map_data_ptr->chromaticity_image_flag = chromaticity_image_flag;

    image_map_data_ptr->input_ip = (reduction_flag ?
                                            get_reduced_input_image()
                                          : get_input_image((int*)NULL));

    image_map_data_ptr->canon_ip = (reduction_flag ?
                                            get_reduced_canon_image()
                                          : get_canon_image((int*)NULL));

    EXTENDED_LC_BUFF_CPY(lc_request, request);

    if (STRCMP_EQ(lc_request, "input"))
    {
        if ((error_ptr != NULL) && (error_item_ptr_ptr != NULL))
        {
            *error_item_ptr_ptr =
                &(error_ptr->sharp[ DONT_SHARPEN ].standard.nothing);
        }

        NRE(image_map_data_ptr->image_ip =
                (reduction_flag ?
                    get_reduced_input_image()
                  : get_input_image((int*)NULL)));
    }
    else if (    (STRCMP_EQ(lc_request, "target"))
              || (STRCMP_EQ(lc_request, "canon"))
              || (STRCMP_EQ(lc_request, "canonical"))
            )
    {
        if ((error_ptr != NULL) && (error_item_ptr_ptr != NULL))
        {
            *error_item_ptr_ptr = NULL;
        }

        NRE(image_map_data_ptr->image_ip =
                (reduction_flag ?
                    get_reduced_canon_image()
                 :  get_canon_image((int*)NULL)));
    }
    else
    {
            NRE(image_map_data_ptr->image_ip =
                (reduction_flag ?
                    get_reduced_input_image()
                 :  get_input_image((int*)NULL)));

    }

    if (    (STRCMP_EQ(lc_request, "input"))
         || (STRCMP_EQ(lc_request, "target"))
         || (STRCMP_EQ(lc_request, "no-var"))
         || (STRCMP_EQ(lc_request, "canon"))
         || (STRCMP_EQ(lc_request, "canonical"))
       )
    {
        free_vector(image_map_data_ptr->map_vp);
        image_map_data_ptr->map_vp = NULL;
        free_matrix(image_map_data_ptr->map_mp);
        image_map_data_ptr->map_mp = NULL;
    }
    else
    {
        ERE(parse_result_request(request, result_ptr,
                                 &(image_map_data_ptr->map_vp),
                                 &(image_map_data_ptr->map_mp),
                                 &(image_map_data_ptr->sharp_method),
                                 error_ptr, error_item_ptr_ptr));
    }

    if ((illumination_image_flag) && (image_map_data_ptr->map_mp != NULL))
    {
        set_error("Illumination estimate is not available for non-digaonal ");
        cat_error("estimates.");
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Changes to this routine must be synched with changes to
// compute_result_data(). See comments there for futher information.
*/

int compute_result_image
(
    KJB_image**     target_ipp,
    Image_map_data* image_map_data_ptr
)
{
    Vector*            map_vp              = image_map_data_ptr->map_vp;
    Matrix*            map_mp              = image_map_data_ptr->map_mp;
    const KJB_image* image_ip            = image_map_data_ptr->image_ip;
    Sharp_method       sharp_method        = image_map_data_ptr->sharp_method;
    int                result;
    int                dimension = NOT_SET;

    if (image_map_data_ptr->illumination_image_flag)
    {
        result = compute_illumination_image(target_ipp, image_map_data_ptr);
    }
    else if (map_vp != NULL)
    {
        dimension = map_vp->length;

        if (sharp_method != DONT_SHARPEN)
        {
            if (map_vp->length == 2)
            {
                result = get_2D_sharp_diag_map_image(target_ipp, image_ip,
                                                     map_vp);
            }
            else
            {
                result = get_3D_sharp_diag_map_image(target_ipp, image_ip,
                                                     map_vp);
            }
        }
        else
        {
            result = get_diag_map_image(target_ipp, image_ip, map_vp);
        }
    }
    else if (map_mp != NULL)
    {
        dimension = map_mp->num_rows;

        ASSERT(image_map_data_ptr->sharp_method == DONT_SHARPEN);

        if (map_mp->num_cols == 2)
        {
            result = get_2D_map_image(target_ipp, image_ip, map_mp);
        }
        else
        {
            result = get_3D_map_image(target_ipp, image_ip, map_mp);
        }
    }
    else
    {
        result = kjb_copy_image(target_ipp, image_ip);
    }

    if (result != ERROR)
    {
        if (image_map_data_ptr->chromaticity_image_flag)
        {
            result = ow_make_chromaticity_image(*target_ipp,
                                                fs_output_brightness_value);
        }
        else if ((dimension == 2) || fs_normalize_3D_output)
        {
            result = ow_normalize_image(*target_ipp,
                                        image_map_data_ptr->input_ip,
                                        image_map_data_ptr->canon_ip);
            verbose_pso(5, "Normalizing image for %s.\n",
                        image_map_data_ptr->request);
        }
        else
        {
            verbose_pso(5, "No normalization is being done for %s image.\n",
                        image_map_data_ptr->request);
        }
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          map_vp
)
{
    Vector* back_projected_map_vp = NULL;
    int     result                = NO_ERROR;

    if (map_vp == NULL)
    {
        set_error("Result has not been computed.");
        return ERROR;
    }

    if (map_vp->length == 2)
    {
        ERE(back_project_vector(&back_projected_map_vp, map_vp, DIVIDE_BY_BLUE));
        map_vp = back_projected_map_vp;
    }

    if (result != ERROR)
        result = scale_image_by_channel(target_ipp, in_ip, map_vp);

    free_vector(back_projected_map_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_sharp_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          map_vp
)
{
    Matrix* diag_mp;
    int     result;

    if (map_vp == NULL)
    {
        set_error("Result has not been computed.");
        return ERROR;
    }

    NRE(diag_mp = create_diagonal_matrix(map_vp));

    result = map_projected_sharp_image(target_ipp, in_ip, diag_mp);

    free_matrix(diag_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_3D_sharp_diag_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Vector*          map_vp
)
{
    Matrix* diag_mp;
    int     result;

    if (map_vp == NULL)
    {
        set_error("Result has not been computed.");
        return ERROR;
    }

    NRE(diag_mp = create_diagonal_matrix(map_vp));

    result = map_sharp_image(target_ipp, in_ip, diag_mp);

    free_matrix(diag_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_2D_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp
)
{

    if (map_mp == NULL)
    {
        set_error("Result has not been computed.");
        return ERROR;
    }

    return post_map_projected_image(target_ipp, in_ip, map_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_3D_map_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp
)
{

    if (map_mp == NULL)
    {
        set_error("Result has not been computed.");
        return ERROR;
    }

    return post_map_image(target_ipp, in_ip, map_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Image_map_data* create_image_map_data(void)
{
    Image_map_data* image_map_data_ptr;

    NRN(image_map_data_ptr = TYPE_MALLOC(Image_map_data));
    image_map_data_ptr->map_vp = NULL;
    image_map_data_ptr->map_mp = NULL;

    return image_map_data_ptr;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void queue_free_image_map_data(void* image_map_data_ptr)
{

    free_image_map_data((Image_map_data*)image_map_data_ptr);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_image_map_data(Image_map_data *image_map_data_ptr)
{

    free_vector(image_map_data_ptr->map_vp);
    free_matrix(image_map_data_ptr->map_mp);
    kjb_free(image_map_data_ptr);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int compute_illumination_image
(
    KJB_image**     target_ipp,
    Image_map_data* image_map_data_ptr
)
{
    Vector*            map_vp     = image_map_data_ptr->map_vp;
    Matrix*            map_mp     = image_map_data_ptr->map_mp;
    const KJB_image* image_ip   = image_map_data_ptr->image_ip;
    Sharp_method       sharp_method = image_map_data_ptr->sharp_method;
    int                result;

    if (map_mp != NULL)
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR;
    }
    else if (map_vp != NULL)
    {
        Vector* illum_estimate_vp = NULL;

        result = get_illumination(&illum_estimate_vp, map_vp, sharp_method);

        if (result != ERROR)
        {
            result = scale_image_by_channel(target_ipp, image_ip,
                                            illum_estimate_vp);
        }

        free_vector(illum_estimate_vp);
    }
    else
    {
        result = kjb_copy_image(target_ipp, image_ip);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_illumination
(
    Vector**     illum_estimate_vpp,
    Vector*      diag_vp,
    Sharp_method sharp_method
)
{
    Vector* diag_3D_vp             = NULL;
    Vector* illum_estimate_vp      = NULL;
    Vector* canon_illum_vp         = NULL;
    Vector* derived_canon_illum_vp;
    Vector* sharp_canon_illum_vp   = NULL;
    int     result                 = NO_ERROR;

    ERE(get_canon_illum(&canon_illum_vp));

    if (sharp_method != DONT_SHARPEN)
    {
        result = sharpen_vector_data(&sharp_canon_illum_vp, canon_illum_vp);
        derived_canon_illum_vp = sharp_canon_illum_vp;
    }
    else
    {
        derived_canon_illum_vp = canon_illum_vp;
    }

    if (result != ERROR)
    {
        if (diag_vp->length == 2)
        {
            result = back_project_vector(&diag_3D_vp, diag_vp, DIVIDE_BY_BLUE);
            diag_vp = diag_3D_vp;
        }
    }

    if (result != ERROR)
    {
        result = divide_vectors(&illum_estimate_vp, derived_canon_illum_vp,
                                diag_vp);
    }

    if (result != ERROR)
    {
        if (sharp_method != DONT_SHARPEN)
        {
            result = unsharpen_vector_data(illum_estimate_vpp,
                                           illum_estimate_vp);
        }
        else
        {
            result = copy_vector(illum_estimate_vpp, illum_estimate_vp);
        }
    }

    free_vector(diag_3D_vp);
    free_vector(illum_estimate_vp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Changes to this routine must be synched with changes to
// ow_normalize_data(). See comments there for futher information.
*/

static int ow_normalize_image
(
    KJB_image*       ip,
    const KJB_image* input_ip,
    const KJB_image* canon_ip
)
{
    if (fs_output_brightness_method == DONT_ADJUST_BRIGHTNESS)
    {
        return NO_ERROR;
    }
    else if (fs_output_brightness_method == ADJUST_AVERGE_BRIGHTNESS)
    {
        double  current_brightness;
        double  desired_brightness;
        double  factor;
        Vector* temp_rgb_vp        = NULL;

        ERE(get_ave_rgb(&temp_rgb_vp, ip));
        current_brightness = sum_vector_elements(temp_rgb_vp);

        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_ip == NULL)
            {
                set_error("Unable to match output image brightness to input.");
                add_error("No input image is available.");
                return ERROR;
            }

            ERE(get_ave_rgb(&temp_rgb_vp, input_ip));
            desired_brightness = sum_vector_elements(temp_rgb_vp);

        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_ip == NULL)
            {
                set_error("Unable to match output image brightness to target.");
                add_error("No target (canonical) image is available.");
                return ERROR;
            }

            ERE(get_ave_rgb(&temp_rgb_vp, canon_ip));
            desired_brightness = sum_vector_elements(temp_rgb_vp);

        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            desired_brightness = fs_output_brightness_value;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }

        free_vector(temp_rgb_vp);

        factor = desired_brightness / current_brightness;

        return ow_scale_image(ip, factor);
    }
    else if (fs_output_brightness_method == ADJUST_MAXIMUM_BRIGHTNESS)
    {
        double  current_brightness;
        double  desired_brightness;
        double  factor;
        Vector* temp_rgb_vp        = NULL;

        ERE(get_max_rgb(&temp_rgb_vp, ip));

        current_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                          temp_rgb_vp->elements[ 1 ]) ,
                                   temp_rgb_vp->elements[ 2 ]);

        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_ip == NULL)
            {
                set_error("Unable to match output image brightness to input.");
                add_error("No input image is available.");
                return ERROR;
            }

            ERE(get_max_rgb(&temp_rgb_vp, input_ip));

            desired_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                              temp_rgb_vp->elements[ 1 ]) ,
                                       temp_rgb_vp->elements[ 2 ]);
        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_ip == NULL)
            {
                set_error("Unable to match output image brightness to target.");
                add_error("No target (canonical) image is available.");
                return ERROR;
            }

            ERE(get_max_rgb(&temp_rgb_vp, canon_ip));

            desired_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                              temp_rgb_vp->elements[ 1 ]) ,
                                       temp_rgb_vp->elements[ 2 ]);
        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            desired_brightness = fs_output_brightness_value;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }

        free_vector(temp_rgb_vp);

        factor = desired_brightness / current_brightness;

        return ow_scale_image(ip, factor);
    }
    else if (fs_output_brightness_method == ADJUST_PIXEL_BRIGHTNESS)
    {
        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_ip == NULL)
            {
                set_error("Unable to match output image brightness to input.");
                add_error("No input image is available.");
                return ERROR;
            }

            return ow_match_image_brightness(ip, input_ip);
        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_ip == NULL)
            {
                set_error("Unable to match output image brightness to target.");
                add_error("No target (canonical) image is available.");
                return ERROR;
            }

            return ow_match_image_brightness(ip, canon_ip);
        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            set_error("Output brightness adjustment method \"pixel\" is ");
            cat_error("invalid with source \"value\".");
            return ERROR;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int ow_match_image_brightness
(
    KJB_image*       target_ip,
    const KJB_image* image_to_match_ip
)
{
    int   i;
    int   j;
    float current_sum;
    float desired_sum;
    float factor;

    for (i=0; i<target_ip->num_rows; i++)
    {
        for (j=0; j<target_ip->num_cols; j++)
        {
            current_sum = target_ip->pixels[ i ][ j ].r
                                + target_ip->pixels[ i ][ j ].g
                                + target_ip->pixels[ i ][ j ].b;

            desired_sum = image_to_match_ip->pixels[ i ][ j ].r
                                + image_to_match_ip->pixels[ i ][ j ].g
                                + image_to_match_ip->pixels[ i ][ j ].b;

            /*
            // Leave black pixels alone.
            */
            if (current_sum > 0.001)
            {
                factor = desired_sum / current_sum;

                target_ip->pixels[ i ][ j ].r *= factor;
                target_ip->pixels[ i ][ j ].g *= factor;
                target_ip->pixels[ i ][ j ].b *= factor;
            }
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Hack time! This routine is meant to do the same thing that
// compute_result_image() does for images, but I don't want to break out the
// logic and put it into a sub-routine, because doing so will be too
// computationally expensive. Note that we have to have the routine in this
// file, because it depends on the settings that this file controls. Basically
// this is the result of unanticipated requirements (not much of an excuse).
//
// This routine is currently used inefficiently. We end up calculating the same
// normalized canonical data over and over again. Ideally, the result should be
// cached, but then we need to deal with the dependencies on the output
// brightness options (as well as the dependencies an the canonical data).
*/

int compute_result_data
(
    Matrix**      result_data_mpp,
    const Matrix* data_mp,
    const Matrix* sharp_data_mp,
    const Matrix* canon_data_mp,
    const Matrix* map_mp
)
{
    Matrix* projected_sharp_data_mp = NULL;
    const Matrix* sharp_data_to_map_mp;

    if (map_mp->num_rows == 2)
    {
        ERE(get_projection_matrix(&projected_sharp_data_mp, sharp_data_mp,
                                  DIVIDE_BY_BLUE, TRUNCATE_INVALID_CHROMATICITY,
                                  MAX_X_DIV_B, DBL_DUMMY_ARG, DBL_DUMMY_ARG,
                                  (Matrix**)NULL, (Matrix**)NULL));
        sharp_data_to_map_mp = projected_sharp_data_mp;
    }
    else
    {
        sharp_data_to_map_mp = sharp_data_mp;
    }

    if (map_mp->num_rows == 2)
    {
        Matrix* mapped_data_mp = NULL;

        ERE(multiply_matrices(&mapped_data_mp, sharp_data_to_map_mp, map_mp));

        ERE(back_project_matrix(result_data_mpp, mapped_data_mp,
                                DIVIDE_BY_BLUE));
        free_matrix(mapped_data_mp);
    }
    else
    {
        ERE(multiply_matrices(result_data_mpp, sharp_data_to_map_mp,
                              map_mp));
    }

    ERE(ow_unsharpen_matrix_data(*result_data_mpp));

    if ((map_mp->num_rows == 2) || (fs_normalize_3D_output))
    {
        ERE(ow_normalize_data(*result_data_mpp, data_mp, canon_data_mp));
    }

    free_matrix(projected_sharp_data_mp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_normalize_canon_data
(
    Matrix*       canon_data_mp,
    const Matrix* input_data_mp
)
{
    if (fs_normalize_3D_output)
    {
        ERE(ow_normalize_data(canon_data_mp, input_data_mp, canon_data_mp));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Changes to this routine must be synched with changes to ow_normalize_image().
// See comments for compute_result_data() for more info.
*/

static int ow_normalize_data
(
    Matrix*       data_mp,
    const Matrix* input_data_mp,
    const Matrix* canon_data_mp
)
{
    if (fs_output_brightness_method == DONT_ADJUST_BRIGHTNESS)
    {
        return NO_ERROR;
    }
    else if (fs_output_brightness_method == ADJUST_AVERGE_BRIGHTNESS)
    {
        double  current_brightness;
        double  desired_brightness;
        double  factor;
        Vector* temp_rgb_vp        = NULL;

        ERE(average_matrix_rows(&temp_rgb_vp, data_mp));
        current_brightness = sum_vector_elements(temp_rgb_vp);

        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_data_mp == NULL)
            {
                set_error("Unable to match output brightness to input.");
                add_error("No input image data is available.");
                return ERROR;
            }

            ERE(average_matrix_rows(&temp_rgb_vp, input_data_mp));
            desired_brightness = sum_vector_elements(temp_rgb_vp);

        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_data_mp == NULL)
            {
                set_error("Unable to match output brightness to target.");
                add_error("No target (canonical) image data is available.");
                return ERROR;
            }

            ERE(average_matrix_rows(&temp_rgb_vp, canon_data_mp));
            desired_brightness = sum_vector_elements(temp_rgb_vp);

        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            desired_brightness = fs_output_brightness_value;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }

        free_vector(temp_rgb_vp);

        factor = desired_brightness / current_brightness;

        return ow_multiply_matrix_by_scalar(data_mp, factor);
    }
    else if (fs_output_brightness_method == ADJUST_MAXIMUM_BRIGHTNESS)
    {
        double  current_brightness;
        double  desired_brightness;
        double  factor;
        Vector* temp_rgb_vp        = NULL;

        ERE(get_max_matrix_col_elements(&temp_rgb_vp, data_mp));

        current_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                          temp_rgb_vp->elements[ 1 ]) ,
                                   temp_rgb_vp->elements[ 2 ]);

        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_data_mp == NULL)
            {
                set_error("Unable to match output brightness to input.");
                add_error("No input image data is available.");
                return ERROR;
            }

            ERE(get_max_matrix_col_elements(&temp_rgb_vp, input_data_mp));

            desired_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                              temp_rgb_vp->elements[ 1 ]) ,
                                       temp_rgb_vp->elements[ 2 ]);
        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_data_mp == NULL)
            {
                set_error("Unable to match output brightness to target.");
                add_error("No target (canonical) image data is available.");
                return ERROR;
            }

            ERE(get_max_matrix_col_elements(&temp_rgb_vp, canon_data_mp));

            desired_brightness = MAX_OF(MAX_OF(temp_rgb_vp->elements[ 0 ],
                                              temp_rgb_vp->elements[ 1 ]) ,
                                       temp_rgb_vp->elements[ 2 ]);
        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            desired_brightness = fs_output_brightness_value;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }

        free_vector(temp_rgb_vp);

        factor = desired_brightness / current_brightness;

        return ow_multiply_matrix_by_scalar(data_mp, factor);
    }
    else if (fs_output_brightness_method == ADJUST_PIXEL_BRIGHTNESS)
    {
        if (fs_output_brightness_source == USE_INPUT_IMAGE)
        {
            if (input_data_mp == NULL)
            {
                set_error("Unable to match output brightness to input.");
                add_error("No input image data is available.");
                return ERROR;
            }

            return ow_match_data_brightness(data_mp, input_data_mp);
        }
        else if (fs_output_brightness_source == USE_TARGET_IMAGE)
        {
            if (canon_data_mp == NULL)
            {
                set_error("Unable to match output brightness to target.");
                add_error("No target (canonical) image data is available.");
                return ERROR;
            }

            return ow_match_data_brightness(data_mp, canon_data_mp);
        }
        else if (fs_output_brightness_source == USE_PRESCRIBED_VALUE)
        {
            set_error("Output brightness adjustment method \"pixel\" is ");
            cat_error("invalid with source \"value\".");
            return ERROR;
        }
        else
        {
            SET_CANT_HAPPEN_BUG();
            return ERROR;
        }
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*
// Changes to this routine must be synched with changes to ow_normalize_image().
// See comments for compute_result_data() for more info.
*/

static int ow_match_data_brightness
(
    Matrix*       target_mp,
    const Matrix* image_to_match_mp
)
{
    int  i;
    int  j;
    double current_sum;
    double desired_sum;
    double factor;

    for (i=0; i<target_mp->num_rows; i++)
    {

        current_sum = 0.0;
        desired_sum = 0.0;

        for (j=0; j<target_mp->num_cols; j++)
        {
            current_sum += target_mp->elements[ i ][ j ];
            desired_sum += image_to_match_mp->elements[ i ][ j ];
        }

        /*
        // Leave black elements alone.
        */
        if (current_sum > 0.001)
        {
            factor = desired_sum / current_sum;

            target_mp->elements[ i ][ 0 ] *= factor;
            target_mp->elements[ i ][ 1 ] *= factor;
            target_mp->elements[ i ][ 2 ] *= factor;
        }
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

