
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Lindsay Martin and Kobus Barnard (authors).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"

#include "corr_lib.h"

static int fp_read_correlation_matrix_header
(
    FILE*              fp,               /* Pointer to file to search for header. */
    Corr_algorithm*    alg_ptr,          /* Ptr to correlation algorithm type.    */
    Corr_matrix_src*   src_ptr,          /* Ptr to matrix data source type.       */
    Corr_matrix_dist*  dist_ptr,         /* Ptr to matrix distribution type.      */
    int*               num_illum_ptr,    /* Number of illuminants.                */
    int*               num_surf_ptr,     /* Number of surfaces.                   */
    int*               num_bins_ptr,
    int*               num_bins_3_ptr,
    Corr_colour_space* colour_space_ptr, /* Number of colour axes in histogram.*/
    double*            sigma_ptr,        /* Sigma value if gaussian distribution. */
    double*            max_RGB_ptr       /* Maximum RGB value for 3D.             */
);

static int fp_write_correlation_matrix_header
(
    FILE*              fp,
    const Corr_matrix* source_cp
);

/* -------------------------------------------------------------------------- */

/* ===========================================================================
 *                       get_target_correlation_matrix
 *
 * Creates a correlation matrix with default values.
 *
 * This routine creates a correlation matrix structure, and fills in default
 * initial values for contained fields.
 *
 * A correlation matrix (type Corr_matrix) is a data structure that contains
 * the following fields:
 *
 *   TYPE             FIELD NAME    FIELD DESCRIPTION
 *   Corr_algorithm   algorithm     Algorithm used to generate matrix.
 *   Corr_matrix_src  data_source   Type of source data matrix is built from.
 *   Corr_matrix_dist distribution  Type of distribution applied to reflectances.
 *   int              num_illum     Number of illuminants in source database.
 *   int              num_surf      Number of surfaces in source database.
 *   int              num_bins      Number of bins in first two axes.
 *   int              num_bins_3    Number of bins in third axis.
 *   int              corr_colour_space
 *   double           sigma         Option sigma value for gaussian fill.
 *   double           make_max_RGB  Maximum RGB value for 3D indexing.
 *   Matrix*          cormat_mp     Pointer to matrix containing the matrix.
 *   Spectra*         illum_sp      Pointer to spectra used to build the matrix.
 *
 * "algorithm" can be one of CMA_ZERO_ONE or CMA_PROBABILITY
 * "data_source" can be one of CMS_UNDEFINED, CMS_RGB, or CMS_SPECTRA.
 * "distribution" can be one of CMD_UNIFORM or CMD_GAUSSIAN.
 *
 * If this routine is successful, the fields will have the following values:
 *
 *   FIELD NAME       DEFAULT VALUE
 *   cormat_mp        num_illums X (num_bins ^ num_dims) matrix, with
 *                     elements set to zero.
 *   illum_sp         NULL
 *
 * NOTE: If the correlation matrix is derived from spectra, the source illuminant
 *       spectra must be manually bound to the "illum_sp" field using a call to
 *       "copy_spectra". For example:
 *
 *       Corr_matrix* corr_cp;
 *       Spectra*     illum_sp = get_illuminant_spectra_from_somewhere();
 *
 *       get_target_correlation_matrix( &corr_cp, ... );
 *       copy_spectra( &(corr_cp->illum_sp), illum_sp );
 *
 * Returns:
 *    NO_ERROR on success, ERROR on failure.
 *
 * See Also:
 *    copy_correlation_matrix, free_correlation_matrix.
 *
 * Authors:
 *    Lindsay Martin
 *
 * Documentors:
 *    Lindsay Martin
 *
 * Index: gamut, colour by correlation, colour by correlation matrices
 *
 * -----------------------------------------------------------------------------*/

int get_target_correlation_matrix
(
    Corr_matrix**     output_correl_cp_ptr,
    Corr_algorithm    algorithm,
    Corr_matrix_src   src,
    Corr_matrix_dist  dist,
    int               num_illum,
    int               num_surf,
    int               num_bins,
    int               num_bins_3,
    Corr_colour_space colour_space,
    double            sigma,
    double            make_max_RGB
)
{

    Corr_matrix* out_cp  = *output_correl_cp_ptr;
    Matrix*      data_mp = NULL;
    int          row_size;

    if ((num_illum < 1) || (num_bins < 1))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (out_cp == NULL)
    {
        NRE(out_cp = TYPE_MALLOC(Corr_matrix ));
    }
    else
    {
        free_matrix(out_cp->cormat_mp);
        free_spectra(out_cp->illum_sp);
        free_matrix(out_cp->illum_chrom_mp);
    }

    out_cp->algorithm    = algorithm;
    out_cp->data_source  = src;
    out_cp->distribution = dist;
    out_cp->num_illum    = num_illum;
    out_cp->num_surf     = num_surf;
    out_cp->num_bins     = num_bins;
    out_cp->num_bins_3   = num_bins_3;
    out_cp->colour_space = colour_space;
    out_cp->make_max_RGB = make_max_RGB;
    out_cp->sigma        = sigma;
    out_cp->illum_sp     = NULL;
    out_cp->illum_chrom_mp = NULL;

    if (colour_space == CORR_RG_SPACE)
    {
        row_size = num_bins * num_bins;
    }
    else if (colour_space == CORR_RGB_SPACE)
    {
        if (num_bins_3 < 1)
        {
            SET_ARGUMENT_BUG();
            return ERROR;
        }
        row_size = num_bins * num_bins * num_bins_3;
    }
    else if (colour_space == CORR_RG_LUM_SPACE)
    {
        if (num_bins_3 < 1)
        {
            SET_ARGUMENT_BUG();
            return ERROR;
        }
        row_size = num_bins_3 * num_bins * (num_bins + 1) / 2;
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(get_zero_matrix(&data_mp, num_illum, row_size));
    out_cp->cormat_mp    = data_mp;

    *output_correl_cp_ptr = out_cp;
    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void free_correlation_matrix(Corr_matrix* corr_cp)
{
    if (corr_cp != NULL)
    {
        free_matrix(corr_cp->cormat_mp);
        free_spectra(corr_cp->illum_sp);
        free_matrix(corr_cp->illum_chrom_mp);
        kjb_free(corr_cp);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int copy_correlation_matrix
(
    Corr_matrix**      target_cp_ptr,
    const Corr_matrix* source_cp
)
{
    if (source_cp == NULL)
    {
        set_error("copy_correlation_matrix() - source correlation matrix is NULL");
        return ERROR;
    }
    ASSERT(source_cp != NULL);

    /* Do a deep copy of the source matrix. */
    ERE(get_target_correlation_matrix(target_cp_ptr,
                                      source_cp->algorithm,
                                      source_cp->data_source,
                                      source_cp->distribution,
                                      source_cp->num_illum,
                                      source_cp->num_surf,
                                      source_cp->num_bins,
                                      source_cp->num_bins_3,
                                      source_cp->colour_space,
                                      source_cp->sigma,
                                      source_cp->make_max_RGB));

    if ( (*target_cp_ptr)->cormat_mp != NULL )
    {
        free_matrix((*target_cp_ptr)->cormat_mp);
        (*target_cp_ptr)->cormat_mp = NULL;
    }
    ERE(copy_matrix( &( (*target_cp_ptr)->cormat_mp),
                     source_cp->cormat_mp ));

    if ( (*target_cp_ptr)->illum_sp != NULL )
    {
        free_spectra((*target_cp_ptr)->illum_sp);
        (*target_cp_ptr)->illum_sp = NULL;
    }
    ERE(copy_matrix(&((*target_cp_ptr)->illum_chrom_mp),
                    source_cp->illum_chrom_mp));

    ERE(copy_spectra( &( (*target_cp_ptr)->illum_sp),
                      source_cp->illum_sp  ));

    return NO_ERROR;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

int read_correlation_matrix
(
    Corr_matrix** target_cp_ptr,
    const char*   file_name,
    const char*   error_msg
)
{
    /*
    // Lindsay - May 07, 1999
    // Changed the read behaviour for log probability correlation matrices.
    //
    // If a log probability matrix is indicated by the header, then all matrix
    // values read from the file that are equal to DBL_NOT_SET are converted
    // to CORRELATION_MINUS_INFINITY. This is intended to avoid the read/write
    // precision dichotomy that has caused problems.
    //
    // "write_correlation_matrix() has had a similar change.
    */

#ifdef NEW_CORRELATION_IO_VERSION
    Matrix*  temp_mp;
    int      i, j;
    double     value;
#endif
    Spectra* illum_from_file_sp = NULL;
    FILE*    fp;
    int      result;

    /* Data from correlation matrix file header. */
    Corr_algorithm   algorithm;
    Corr_matrix_src  source;
    Corr_matrix_dist distribution;
    int              num_illum = NOT_SET;
    int              num_surf  = NOT_SET;
    int              num_bins  = NOT_SET;
    int              num_bins_3 = NOT_SET;
    Corr_colour_space colour_space = CORR_RG_SPACE; /* Backwards compatability */
    double             sigma     = DBL_NOT_SET;
    double             max_RGB   = DBL_NOT_SET;

    /*
    // Untested since adjustments for new conventions on reading matrices.
    */

    if ((file_name == NULL) || (file_name[ 0 ] == '\0'))
    {
        SET_CANT_HAPPEN_BUG();
        return ERROR;
    }

    fp = open_gamut_config_file(NULL, file_name, error_msg);

    if (fp == NULL)
    {
        return ERROR;
    }

    result = fp_read_correlation_matrix_header(fp, &algorithm, &source,
                                               &distribution,
                                               &num_illum, &num_surf,
                                               &num_bins, &num_bins_3,
                                               &colour_space,
                                               &sigma, &max_RGB);

    if ( result == ERROR )
    {
        add_error("Failed to read correlation matrix header from \"%s\".",
                  file_name);
        return ERROR;
    }

    /*
    // Check the important header options for validity. The other options are just
    // descriptive metadata.
    */
    if (   (num_illum == NOT_SET)
        || (num_illum < 1)        )
    {
        set_error("Invalid number of illuminants (cm_noi=%d) in correlation matrix header", num_illum);
        return ERROR;
    }

    if (   (num_bins == NOT_SET)
        || (num_bins < 1)       )
    {
        set_error("Invalid number of histogram bins (cm_nob=%d) in correlation matrix header", num_bins);
        return ERROR;
    }

    result = get_target_correlation_matrix(target_cp_ptr,
                                           algorithm, source, distribution,
                                           num_illum, num_surf,
                                           num_bins, num_bins_3, colour_space,
                                           sigma, max_RGB);

    if ( result == ERROR )
    {
        add_error("Unable to obtain correlation matrix.");
        return ERROR;
    }

#ifdef NEW_CORRELATION_IO_VERSION

    TEST_PSE((">>>>>> New Correlation I/O Model Used <<<<<<\n"));

    if ( (algorithm == CMA_ZERO_ONE) || (algorithm == CMA_PROBABILITY) )
    {
        /* Read correlation data verbatim. Identical to old version */

        TEST_PSE((">>>>>>   Non-log values detected. Data read verbatim.\n"));
        if (fp_ow_read_matrix_by_rows((*target_cp_ptr)->cormat_mp, fp) == ERROR )
        {
            add_error("Failed to read correlation matrix data from \"%s\".",
                      file_name);
            return ERROR;
        }
    }
    else if (algorithm == CMA_LOG_PROBABILITY)
    {
        TEST_PSE((">>>>>>   Log values detected. Minus Inifinity values corrected.\n"));

        temp_mp = create_matrix(num_illum,
                                (*target_cp_ptr)->cormat_mp->num_cols);

        if (temp_mp == NULL)
        {
            set_error("Failed to allocate matrix for correlation data");
            return ERROR;
        }

        ASSERT(temp_mp->num_rows == (*target_cp_ptr)->cormat_mp->num_rows);

        /* Read the correlation matrix into a temporary matrix. */
        if (fp_ow_read_matrix_by_rows(temp_mp, fp) == ERROR)
        {
            add_error("Failed to read correlation matrix data from \"%s\".",
                      file_name);
            return ERROR;
        }

        /*
        // Copy values from the temporary matrix to the correlation matrix,
        // translating any of the dummy -infinity values to the true one.
        */
        for (i = 0; i < temp_mp->num_rows; i++)
        {
            for (j = 0; j < temp_mp->num_cols; j++)
            {
                value = temp_mp->elements[ i ][ j ];

                if ( IS_EQUAL_DBL(value, FILE_MINUS_INFINITY ) )
                {
                    (*target_cp_ptr)->cormat_mp->elements[ i ][ j ]
                                                 = CORRELATION_MINUS_INFINITY;
                }
                else
                {
                    (*target_cp_ptr)->cormat_mp->elements[ i ][ j ] = value;
                }
            }
        }

        free_matrix(temp_mp);
    }
    else
    {
        set_error("Invalid alorithm code in correlation matrix header");
        add_error(" (algorithm=%d)", algorithm);
        add_error(" (file_name=%s)", file_name);
        return ERROR;
    }

#else /* Old I/O Version - read correlation data verbatim */

    TEST_PSE((">>>>>> Old Correlation I/O Model Used <<<<<<\n"));

    if ( fp_ow_read_matrix_by_rows(fp,(*target_cp_ptr)->cormat_mp) == ERROR )
    {
        add_error("Failed to read correlation matrix data from \"%s\".",
                  file_name);
        return ERROR;
    }
#endif

    /* Read in the illuminant chromaticities. */
    if (fp_read_matrix_with_header(&((*target_cp_ptr)->illum_chrom_mp), fp) == ERROR)
    {
        add_error("Failed to read illuminant spectra from \"%s\".", file_name);
        return ERROR;
    }

    /*
    // Read in the illuminant spectra. It may not be there, but it is not very
    // important.
    */
    (void)fp_read_spectra(&illum_from_file_sp, fp);

    if (copy_spectra( &((*target_cp_ptr)->illum_sp), illum_from_file_sp) == ERROR )
    {
        add_error("Unable to copy illum spectra into correlation matrix.");
        result = ERROR;
    }

    if (kjb_fclose(fp) == ERROR )
    {
        result = ERROR;
    }

    free_spectra(illum_from_file_sp);

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

int write_correlation_matrix
(
    const char*        file_name,
    const Corr_matrix* source_cp
)
{
    /*
    // Lindsay - May 07, 1999
    // Changed the write behaviour for log probability correlation matrices.
    //
    // If a log probability matrix is indicated by the header, then all
    // matrix values that are equal to CORRELATION_MINUS_INFINITY are written
    // as DBL_NOT_SET. This is intended to avoid the read/write precision
    // dichotomy that has caused problems.
    //
    // "read_correlation_matrix() has had a similar change.
    */

    FILE*        fp;

#ifdef NEW_CORRELATION_IO_VERSION
    Matrix*  temp_mp = NULL;
    int      i, j;
#endif

    fp = kjb_fopen(file_name, "w");

    if (fp == NULL)
    {
        add_error("Could not open file \"%s\" for correlation matrix output.",
                  file_name);
        return ERROR;
    }

    if (source_cp == NULL)
    {
        set_error("Cannot write empty correlation matrix to \"%s\".",
                  file_name);
        return ERROR;
    }

    ERE(fp_write_correlation_matrix_header(fp, source_cp));

#ifdef NEW_CORRELATION_IO_VERSION
    TEST_PSE((">>>>>> New Correlation I/O Model Used <<<<<<\n"));

    if (source_cp->algorithm == CMA_LOG_PROBABILITY)
    {
        TEST_PSE((">>>>>>   Log values detected. Minus Inifinity values corrected.\n"));

        ERE(copy_matrix(&temp_mp, source_cp->cormat_mp));

        /*
        // Copy values from the temporary matrix to the correlation matrix,
        // translating any of the true -infinity values to the dummy file
        // placeholder.
        */
        for (i = 0; i < temp_mp->num_rows; i++)
        {
            for (j = 0; j < temp_mp->num_cols; j++)
            {
                if ( IS_EQUAL_DBL( temp_mp->elements[ i ][ j ],
                                    CORRELATION_MINUS_INFINITY )
                   )
                {
                    temp_mp->elements[ i ][ j ] = FILE_MINUS_INFINITY;
                }
            }
        }

        ERE(fp_write_matrix_rows_full_precision(temp_mp, fp));

        free_matrix(temp_mp);
    }
    else
        /* Write the data verbatim - same as Old I/O Version */
    {
        TEST_PSE((">>>>>>   Non-log values detected. Data written verbatim.\n"));
        ERE(fp_write_matrix_rows_full_precision(source_cp->cormat_mp, fp));
    }

#else /* Old I/O Version - write correlation data verbatim */

    TEST_PSE((">>>>>> Old Correlation I/O Model Used <<<<<<\n"));

    ERE(fp_write_matrix_rows_full_precision(fp, source_cp->cormat_mp));

#endif

    ERE(fp_write_matrix_with_header(source_cp->illum_chrom_mp, fp));

    if (source_cp->illum_sp != NULL)
    {
        ERE(fp_write_spectra_full_precision(source_cp->illum_sp, fp));
    }

    return kjb_fclose(fp);
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int fp_read_correlation_matrix_header
(
    FILE*              fp,
    Corr_algorithm*    alg_ptr,
    Corr_matrix_src*   src_ptr,
    Corr_matrix_dist*  dist_ptr,
    int*               num_illum_ptr,
    int*               num_surf_ptr,
    int*               num_bins_ptr,
    int*               num_bins_3_ptr,
    Corr_colour_space* colour_space_ptr,
    double*            sigma_ptr,
    double*            max_RGB_ptr
)
{
    IMPORT int kjb_comment_char;
    IMPORT int kjb_header_char;
    int        i;
    int        num_options;
    char**     option_list;
    char**     option_list_pos;
    char**     value_list;
    char**     value_list_pos;

    char       line[ 5000 ];
    char*      line_pos;
    long       save_file_pos;
    int        algorithm       = NOT_SET;
    int        source          = NOT_SET;
    int        distribution    = NOT_SET;
    int        num_illum       = NOT_SET;
    int        num_surf        = NOT_SET;
    int        num_bins        = NOT_SET;
    int        num_bins_3      = NOT_SET;
    Corr_colour_space colour_space    = CORR_INVALID_COLOUR_SPACE;
    double            sigma           = DBL_NOT_SET;
    double            max_RGB         = DBL_NOT_SET;
    int               result          = NO_ERROR;

    if (fp == NULL)
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(save_file_pos = kjb_ftell(fp));

    /*CONSTCOND*/
    while ( TRUE )
    {
        int read_result;

        read_result  = BUFF_FGET_LINE(fp, line);

        if (read_result == ERROR)
        {
            add_error("Error reading data from correlation matrix data file \"%F\"", fp);
            result = ERROR;
        }

        if (read_result < 0)
        {
            add_error("No data in correlation matrix data file \"%F\"", fp);
            break;
        }

        line_pos = line;

        trim_beg(&line_pos);

        if (*line_pos == '\0')
        {
            /*EMPTY*/
            ;  /* Do nothing. */
        }
        else if (*line_pos == kjb_comment_char)
        {
            line_pos++;

            trim_beg(&line_pos);

            if (*line_pos == kjb_header_char)
            {
                line_pos++;

                num_options = parse_options(line_pos, &option_list, &value_list);

                option_list_pos = option_list;
                value_list_pos  = value_list;

                if (num_options == ERROR)
                {
                    add_error("Unable to parse options in correlation matrix header.");
                    result = ERROR;
                    break;
                }

                for (i=0; i<num_options; i++)
                {
                    /* algorithm. */
                    if (IC_STRCMP_EQ(*option_list_pos,"cm_alg"))
                    {
                        if (ss1pi(*value_list_pos, &algorithm) == ERROR)
                        {
                            add_error("Unable to read \"cm_alg\" (algorithm) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* data source. */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_src"))
                    {
                        if (ss1pi(*value_list_pos, &source) == ERROR)
                        {
                            add_error("Unable to read \"cm_src\" (data source) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* distribution of surface reflectances. */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_dst"))
                    {
                        if (ss1pi(*value_list_pos, &distribution) == ERROR)
                        {
                            add_error("Unable to read \"cm_dst\" (distribution) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* number of illuminants */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_noi"))
                    {
                        if (ss1pi(*value_list_pos, &num_illum) == ERROR)
                        {
                            add_error("Unable to read \"cm_noi\" (num illum) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* number of surfaces */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_nos"))
                    {
                        if (ss1pi(*value_list_pos, &num_surf) == ERROR)
                        {
                            add_error("Unable to read \"cm_nos\" (num surfaces) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* number of histogram bins */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_nob"))
                    {
                        if (ss1pi(*value_list_pos, &num_bins) == ERROR)
                        {
                            add_error("Unable to read \"cm_nob\" (num bins) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* number of histogram bins */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_nob3"))
                    {
                        if (ss1pi(*value_list_pos, &num_bins_3) == ERROR)
                        {
                            add_error("Unable to read \"cm_nob3\" (num bins 3)");
                            cat_error(" option from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    /* number of histogram axes--now colour space! */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_dim"))
                    {
                        int temp_int;

                        if (ss1pi(*value_list_pos, &temp_int) == ERROR)
                        {
                            add_error("Unable to read \"cm_dim\" (num dimensions) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                        colour_space = (Corr_colour_space)temp_int;
                    }

                    /* colour space */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_cs"))
                    {
                        int temp_int;

                        if (ss1pi(*value_list_pos, &temp_int) == ERROR)
                        {
                            add_error("Unable to read \"cm_cs\" (colour space) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                        colour_space = (Corr_colour_space)temp_int;
                    }

                    /* 3D average RGB luminance. */
                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_3Dl"))
                    {
                        warn_pso("Obsolete option cm_3Dl in %F is ignored.\n",
                                 fp);
                    }

                    else if (IC_STRCMP_EQ(*option_list_pos,"cm_3Dm"))
                    {
                        if (ss1d(*value_list_pos, &max_RGB) == ERROR)
                        {
                            add_error("Unable to read \"cm_3Dm\" (max RGB) option");
                            cat_error(" from correlation matrix header.");
                            result = ERROR;
                            break;
                        }
                    }

                    value_list_pos++;
                    option_list_pos++;
                }

                free_options(num_options, option_list, value_list);

                if (result == ERROR)
                {
                    break;
                }
            }
        }
        else
        {
            result =  NO_ERROR;
            break;
        }
    }

    if (result != ERROR)
    {
        if (IS_SET(algorithm))
        {
            *alg_ptr = (Corr_algorithm)algorithm;
        }

        if (IS_SET(source))
        {
            *src_ptr = (Corr_matrix_src)source;
        }

        if (IS_SET(distribution))
        {
            *dist_ptr = (Corr_matrix_dist)distribution;
        }

        if (IS_SET(num_illum))
        {
            *num_illum_ptr = num_illum;
        }

        if (IS_SET(num_surf))
        {
            *num_surf_ptr = num_surf;
        }

        if (IS_SET(num_bins))
        {
            *num_bins_ptr = num_bins;
        }

        if (IS_SET(num_bins_3))
        {
            *num_bins_3_ptr = num_bins_3;
        }

        if (colour_space != CORR_INVALID_COLOUR_SPACE)
        {
            *colour_space_ptr = colour_space;
        }

        /*CHECK Logic here with respect to error checking (what if sigma is
        //zero, etc).
        */
        if (sigma > 0.0)
        {
            *sigma_ptr = sigma;
        }

        if (max_RGB != DBL_NOT_SET)
        {
            *max_RGB_ptr = max_RGB;
        }
    }

    ERE(kjb_fseek(fp, save_file_pos, SEEK_SET));

    return result;
}

/* /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\  */

static int fp_write_correlation_matrix_header
(
    FILE*              fp,
    const Corr_matrix* source_cp
)
{
    char   comment_str[100];

    if (fp == NULL)
    {
        if (is_interactive())
        {
            ERE(pso("Invalid file pointer when writing "));
            ERE(pso("correlation matrix header.\n"));
        }

        return ERROR;
    }

    /* Write out the correlation matrix header. */
    ERE(kjb_fprintf(fp, "#!"));
    ERE(kjb_fprintf(fp, " cm_alg=%d",     source_cp->algorithm));
    ERE(kjb_fprintf(fp, " cm_src=%d",     source_cp->data_source));
    ERE(kjb_fprintf(fp, " cm_dst=%d",     source_cp->distribution));
    ERE(kjb_fprintf(fp, " cm_noi=%d",     source_cp->num_illum));
    ERE(kjb_fprintf(fp, " cm_nos=%d",     source_cp->num_surf));
    ERE(kjb_fprintf(fp, " cm_nob=%d",     source_cp->num_bins));

    if (source_cp->num_bins_3 > 0)
    {
        ERE(kjb_fprintf(fp, " cm_nob3=%d",    source_cp->num_bins_3));
    }
    ERE(kjb_fprintf(fp, " cm_cs=%d",      source_cp->colour_space));
    ERE(kjb_fprintf(fp, " cm_sig=%f",     source_cp->sigma));

    if (    (source_cp->colour_space == CORR_RGB_SPACE)
         || (source_cp->colour_space == CORR_RG_LUM_SPACE)
       )
    {
        ERE(kjb_fprintf(fp, " cm_3Dm=%f", source_cp->make_max_RGB));
    }

    ERE(kjb_fprintf(fp, "\n\n"));

    /* Write out an English version of the header as a comment. */
    if (source_cp->algorithm == CMA_ZERO_ONE)
    {
        BUFF_CPY(comment_str, "0/1 Correlation Matrix");
    }
    else if (source_cp->algorithm == CMA_PROBABILITY)
    {
        BUFF_CPY(comment_str, "Probability Correlation Matrix");
    }
    else if (source_cp->algorithm == CMA_LOG_PROBABILITY)
    {
        BUFF_CPY(comment_str, "log(Probability) Correlation Matrix");
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(kjb_fprintf(fp, "# %s\n", comment_str));
    ERE(kjb_fprintf(fp, "#   Num illuminants          : %d\n",
                    source_cp->num_illum));
    ERE(kjb_fprintf(fp, "#   Num surfaces             : %d\n",
                    source_cp->num_surf));
    ERE(kjb_fprintf(fp, "#   Colour space is          : %d\n",
                    source_cp->colour_space));

    ERE(kjb_fprintf(fp, "#   Num histogram bins 1 & 2 : %d / axis\n",
                    source_cp->num_bins));

    if (    (source_cp->colour_space == CORR_RG_LUM_SPACE)
         || (source_cp->colour_space == CORR_RGB_SPACE)
       )
    {
        ERE(kjb_fprintf(fp, "#   Num histogram bins 3     : %d / axis\n",
                        source_cp->num_bins_3));
    }
    if (source_cp->colour_space == CORR_RG_SPACE)
    {
        ERE(kjb_fprintf(fp, "#   (chromaticity matrix)\n"));
    }
    else if (source_cp->colour_space == CORR_RG_LUM_SPACE)
    {
        ERE(kjb_fprintf(fp, "#   (rg-lum matrix)\n"));
        ERE(kjb_fprintf(fp, "#   3D maximum RGB value     : %f\n",
                        source_cp->make_max_RGB));
    }
    else if (source_cp->colour_space == CORR_RGB_SPACE)
    {
        ERE(kjb_fprintf(fp, "#   (RGB matrix)\n"));
        ERE(kjb_fprintf(fp, "#   3D maximum RGB value     : %f\n",
                        source_cp->make_max_RGB));
    }
    else
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    if (source_cp->data_source == CMS_UNDEFINED)
    {
        ERE(kjb_fprintf(fp, "#   Data Source              : Undefined\n"));
    }
    else if (source_cp->data_source == CMS_RGB)
    {
        ERE(kjb_fprintf(fp, "#   Data Source              : RGB\n"));
    }
    else if (source_cp->data_source == CMS_SPECTRA)
    {
        ERE(kjb_fprintf(fp, "#   Data Source              : Spectra\n"));
    }
    else
    {
        ERE(kjb_fprintf(fp, "#   Data Source              : WARNING: "));
        ERE(kjb_fprintf(fp, "Source not specified\n"));
    }

    if (source_cp->distribution == CMD_UNIFORM)
    {
        ERE(kjb_fprintf(fp,
                        "#   Reflectance Distribution : Uniform\n\n"));
    }
    else if (source_cp->distribution == CMD_GAUSSIAN)
    {
        ERE(kjb_fprintf(fp, "#   Reflectance Distribution : Gaussian\n\n"));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ */

