
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "canon_gamut.h"
#include "data.h"
#include "do_cc.h"
#include "grey.h"
#include "lib.h"
#include "synthetic_data.h"
#include "illum_data.h"
#include "spectra.h"

#include "generate.h"

/* -------------------------------------------------------------------------- */

static int process_generate_option
(
    char* option,
    char* value,
    int*  count_ptr,
    int*  num_surfaces_ptr,
    int*  num_illuminants_ptr,
    int*  num_canon_illuminants_ptr,
    long* seed_ptr,
    long* surface_seed_ptr
);

static int generate_canon_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surfaces,
    int                       num_illuminants,
    int                       num_canon_illuminants,
    long                      surface_seed
);

static int generate_illum_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surfaces,
    int                       num_illuminants,
    long                      surface_seed
);

static int generate_scene_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surface,
    long                      surface_seed
);

/* -------------------------------------------------------------------------- */

int process_generate_command(char* arg)
{
    int    num_options;
    char** option_list;
    char** option_list_pos;
    char** value_list;
    char** value_list_pos;
    int    count           = 1;
    int    num_surfaces    = 0;
    int    num_illuminants = 0;        /* Use current one as default. */
    int    num_canon_illuminants = 0;  /* Use current one as default. */
    int    i;
    int    result = NO_ERROR;
    Colour_constancy_methods cc_methods;
    long   seed = NOT_SET;
    long   surface_seed = NOT_SET;

    ERE(num_options=parse_options(arg, &option_list, &value_list));

    option_list_pos = option_list;
    value_list_pos = value_list;

    for (i = 0; i<num_options; i++)
    {
        result = process_generate_option(*option_list_pos,
                                         *value_list_pos,
                                         &count,
                                         &num_surfaces,
                                         &num_illuminants,
                                         &num_canon_illuminants,
                                         &seed, &surface_seed);

        if (result == ERROR) break;

        value_list_pos++;
        option_list_pos++;
    }

    free_options(num_options, option_list, value_list);

    if (IS_SET(seed))
    {
        kjb_seed_rand(seed, (long)0);
    }

    ERE(get_cc_methods(arg, &cc_methods));

    if (*arg != '\0')
    {
        set_error("Invalid generate specifications: %q", arg);
        return ERROR;
    }

    if (cc_methods.plot_enable == NOT_SET)   cc_methods.plot_enable = FALSE;
    if (cc_methods.report_result == NOT_SET) cc_methods.report_result = FALSE;

    if (result != ERROR)
    {
        /* kjb_disable_paging(); */

        result = request_synthetic_data();
    }

    if (result != ERROR)
    {
        result = generate_canon_and_do_cc(&cc_methods, count, num_surfaces,
                                          num_illuminants,
                                          num_canon_illuminants,
                                          surface_seed);
        if (restore_data_source() == ERROR)
        {
            result = ERROR;
        }

        /* kjb_restore_paging(); */
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int process_generate_option
(
    char* option,
    char* value,
    int*  count_ptr,
    int*  num_surfaces_ptr,
    int*  num_illuminants_ptr,
    int*  num_canon_illuminants_ptr,
    long  int *seed_ptr,
    long  int *surface_seed_ptr
)
{
    char  lc_option[ 100 ];

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if (match_pattern(lc_option, "seed"))
    {
        ERE(ss1l(value, seed_ptr));
    }
    else if (match_pattern(lc_option, "surface-seed"))
    {
        ERE(ss1l(value, surface_seed_ptr));
    }
    else if (match_pattern(lc_option, "c"))
    {
        ERE(ss1pi(value, count_ptr));
    }
    else if (match_pattern(lc_option, "n"))
    {
        ERE(ss1pi(value, num_surfaces_ptr));
    }
    else if (match_pattern(lc_option, "ni"))
    {
        ERE(ss1pi(value, num_illuminants_ptr));
    }
    else if (match_pattern(lc_option, "illum-num"))
    {
        ERE(ss1pi(value, num_illuminants_ptr));
        *num_illuminants_ptr *= -1;  /* Overload num illum with illum num. */
    }
    else if (match_pattern(lc_option, "nc"))
    {
        ERE(ss1pi(value, num_canon_illuminants_ptr));
    }
    else if (match_pattern(lc_option, "canon-num"))
    {
        ERE(ss1pi(value, num_canon_illuminants_ptr));
        /* Overload num canon illum with negative of canon illum num. */
        *num_canon_illuminants_ptr *= -1;
    }
    else
    {
        set_error("%q is an invalid generate option.", option);
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int generate_canon_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surfaces,
    int                       num_illuminants,
    int                       num_canon_illuminants,
    long                      int surface_seed
)
{
    int result = NO_ERROR;
    int i;
    int canon_illuminant_num = NOT_SET;

    if (num_canon_illuminants <= 0)
    {
        canon_illuminant_num = -num_canon_illuminants;
        num_canon_illuminants = 1;
    }

    i = 0;

    while ((result != ERROR) && (i < num_canon_illuminants))
    {
        if (canon_illuminant_num > 0)
        {
            result = set_canon_spectra_by_number( canon_illuminant_num );
        }
        else if (canon_illuminant_num != 0)
        {
            result = set_random_canon_spectra();

            if (result != ERROR)
                verbose_pso(2, "Canon illum number is %d.\n", result);
        }

        if (    (result != ERROR)
             && (canon_illuminant_num != 0)
           )
        {
            result = make_canon_illum("", "");

            if (result != ERROR)
            {
                result = make_canon_data("", "");
            }

            if (result != ERROR)
            {
                result = make_canon_gamut("", "");
            }

            if (result != ERROR)
            {
                result = make_grey("", "");
            }
        }

        if (result != ERROR)
        {
           result = generate_illum_and_do_cc(cc_methods_ptr,
                                             count, num_surfaces,
                                             num_illuminants, surface_seed);
        }

        i++;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int generate_illum_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surfaces,
    int                       num_illuminants,
    long                      int surface_seed
)
{
    int result = NO_ERROR;
    int i;
    int illuminant_num = NOT_SET;
    int max_num_illuminants;

    ERE(get_num_illum_spectra(&max_num_illuminants));

    if (num_illuminants < 0)
    {
        illuminant_num = -num_illuminants;
        num_illuminants = 1;
    }
    else if (num_illuminants == 0)
    {
        num_illuminants = max_num_illuminants;
        illuminant_num = 0;
    }

    i = 0;

    dbi(num_illuminants);
    dbi(max_num_illuminants);
    dbi(illuminant_num);

    while ((result != ERROR) && (i < num_illuminants))
    {
        dbi(i);

        if (illuminant_num > 0)
        {
            result = set_actual_spectra_by_number(illuminant_num);
        }
        else if (illuminant_num == 0)
        {
            result = set_actual_spectra_by_number(i + 1);

            if (result != ERROR)
            {
                dbw();
                verbose_pso(3, "Illum number is %d (%d / %d).\n", i + 1,
                            i + 1, num_illuminants);
            }
        }
        else
        {
            static int first_time = TRUE;
            static unsigned short save_seed_buff[ 3 ];
            static unsigned short illum_seed_buff[ 3 ];

            get_rand_seed(save_seed_buff);

            if (first_time)
            {
                first_time = FALSE;
            }
            else
            {
                kjb_seed_rand_with_3_short(illum_seed_buff);
            }

            result = set_random_actual_spectra();

            if (result != ERROR)
            {
                dbw();
                verbose_pso(3, "Illum number is %d (%d / %d).\n", result,
                            i + 1, num_illuminants);
                get_rand_seed(illum_seed_buff);
            }

            kjb_seed_rand_with_3_short(save_seed_buff);
        }

        if (result != ERROR)
        {
            result = make_actual_illum("", "");

            if (result != ERROR)
            {
                result = make_data("", "");
            }
        }

        if (result != ERROR)
        {
            result = generate_scene_and_do_cc(cc_methods_ptr, count,
                                              num_surfaces, surface_seed);
        }

        i++;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int generate_scene_and_do_cc
(
    Colour_constancy_methods* cc_methods_ptr,
    int                       count,
    int                       num_surfaces,
    long                      int surface_seed
)
{
    int                     num_data_points;
    int*                    data_point_nums;
    int                     result                        = NO_ERROR;
    int                     prompt_to_increase_error_flag = FALSE;
    int                     i;
    Colour_constancy_result cc_result;
    int                     total_num_data_points;

    ERE(total_num_data_points = get_num_data_points());

    initialize_cc_result(&cc_result);

    if (num_surfaces == 0)
    {
        num_data_points = total_num_data_points;
    }
    else
    {
        num_data_points = num_surfaces;
    }

    NRE(data_point_nums = INT_MALLOC(num_data_points));

    i = 0;

    if (is_interactive())
    {
        set_io_atn_trap();
    }

    while ((result != ERROR) && (i<count))
    {
        if (num_surfaces == 0)
        {
            int ii;

            for (ii = 0; ii < num_data_points; ii++)
            {
                data_point_nums[ ii ] = ii + 1;
            }
        }
        else
        {
            if (IS_SET(surface_seed))
            {
                kjb_seed_rand(surface_seed, (long)0);
            }

            result = get_random_integer_list(num_surfaces, 1,
                                             total_num_data_points,
                                             data_point_nums);
        }

        if (result != ERROR)
        {
            IMPORT volatile int io_atn_flag;
            IMPORT volatile int halt_all_output;

            result = do_cc(cc_methods_ptr, prompt_to_increase_error_flag,
                           num_data_points, data_point_nums, &cc_result);

            clear_result(&cc_result);   /* Also frees all allocated memory. */

            /*
            // Likely this error message has already been set, and
            // halt_all_output has already been set to FALSE. Note that
            // io_atn_flag does not get set to FALSE until unset_io_atn_trap is
            // called.
            */
            if (io_atn_flag)
            {
                set_error("Processing interrupted.");
                halt_all_output = FALSE;
                result =  ERROR;
            }
            else if (result == ERROR)
            {
                kjb_print_error();
                result = NO_ERROR;
            }
        }

        i++;
    }

    if (is_interactive())
    {
        unset_io_atn_trap();
    }

    kjb_free(data_point_nums);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

