
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author). (Lindsay Martin
|  contributed the interface with the correlation code).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "c2/c2_sharp.h"
#include "i4/i4_map.h"

#include "data.h"
#include "synthetic_data.h"
#include "spectra.h"
#include "illum_data.h"
#include "sharp.h"

/* -------------------------------------------------------------------------- */

typedef enum Sharp_init_method
{
    IDENTITY_SHARP_INIT,
    AVE_ILLUM_SHARP_INIT,
    FILE_SHARP_INIT
}
Sharp_init_method;

/* -------------------------------------------------------------------------- */

static Matrix* fs_initial_mip_db_sharpen_mp = NULL;
static Sharp_init_method fs_mip_db_sharp_init_method = IDENTITY_SHARP_INIT;

static Matrix* fs_sharpen_mp       = NULL;
static Matrix* fs_inv_sharpen_mp   = NULL;

static Matrix* fs_fixed_sharpen_mp = NULL;
static int     fs_sharp_version    = 0;

static Matrix* fs_cached_db_sharpen_mp = NULL;
static int     fs_cached_db_sharp_version = 0;

static Matrix* fs_cached_mip_db_sharpen_mp = NULL;
static int     fs_cached_mip_db_sharp_version = 0;

static Sharp_method fs_current_sharp_method = DONT_SHARPEN;

/* -------------------------------------------------------------------------- */

static int set_fixed_sharpening(void);
static int set_synthetic_data_base_sharpening(void);
static int make_synthetic_db_sharp_transform(void);
static int set_synthetic_mip_db_sharpening(void);
static int make_synthetic_mip_db_sharp_transform(const char* file_name);
static int set_world_data_base_sharpening(void);

static int set_scene_data_base_sharpening
(
    int  num_data_points,
    int* data_point_nums
);

static int set_iterative_sharpening(Vector* illum_rgb_vp);

static int get_average_illum_db_sharp(Matrix** sharp_mpp);

static int set_sharp_matrix(Matrix* mp);

static int map_rgb
(
    double* map_R_ptr,
    double* map_G_ptr,
    double* map_B_ptr,
    Matrix* map_mp,
    double  R,
    double  G,
    double  B
);

static int get_best_fit_fixed_sharp_maps
(
    Best_fit_result* sharp_result_ptr,
    Best_fit_result* dull_result_ptr,
    Vector*          canon_illum_vp
);

static int get_standard_fixed_sharp_maps
(
    Standard_result* sharp_result_ptr,
    Standard_result* dull_result_ptr,
    Vector*          canon_illum_vp
);

static int get_best_fit_fixed_sharp_maps
(
    Best_fit_result* sharp_result_ptr,
    Best_fit_result* dull_result_ptr,
    Vector*          canon_illum_vp
);

static int get_standard_fixed_sharp_maps
(
    Standard_result* sharp_result_ptr,
    Standard_result* dull_result_ptr,
    Vector*          canon_illum_vp
);

static int get_gamut_fixed_sharp_maps
(
    Gamut_result* sharp_result_ptr,
    Gamut_result* dull_result_ptr,
    Vector*       canon_illum_vp
);

static int get_correlation_fixed_sharp_maps
(
    Correlation_result* sharp_result_ptr,
    Correlation_result* dull_result_ptr,
    Vector*             canon_illum_vp
);

#ifdef SIMPLE_TEMPLATE
    static int get_simple_template_fixed_sharp_maps
    (
        Simple_template_result* sharp_result_ptr,
        Simple_template_result* dull_result_ptr,
        Vector*                canon_illum_vp
    );
#endif

#ifdef TEMPLATE
    static int get_template_fixed_sharp_maps
    (
        Template_result* sharp_result_ptr,
        Template_result* dull_result_ptr,
        Vector*          canon_illum_vp
    );
#endif

#ifdef MULTI_TEMPLATE
    static int get_multi_template_fixed_sharp_maps
    (
        Multi_template_result* sharp_result_ptr,
        Multi_template_result* dull_result_ptr,
        Vector*                canon_illum_vp
    );
#endif

static int get_fixed_sharp_diagonal_map
(
    Vector** sharp_diag_vpp,
    Vector*  diag_vp,
    Vector*  canon_illum_vp
);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif

/* -------------------------------------------------------------------------- */

EXTERN_C int set_sharp_file(const char* option, const char* value)
{
    char lc_option[ 100 ];
    Matrix* temp_mp = NULL;
    int     result  = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "smf"))
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("Under construction\n"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_sharpen_mp == NULL)
            {
                ERE(pso("No sharpening matrix is active.\n"));
            }
            else
            {
                ERE(pso("The current sharpening matrix is :\n"));
                ERE(write_matrix(fs_sharpen_mp, (char*)NULL));
            }

            result = NO_ERROR;
        }
        else
        {
            if ((IC_STRCMP_EQ(value, "ave")) || (IC_STRCMP_EQ(value,"average")))
            {
                ERE(get_average_illum_db_sharp(&temp_mp));
            }
            else
            {
                ERE(read_matrix(&temp_mp, value));
            }

#ifdef TRACK_MEMORY_ALLOCATION
            prepare_memory_cleanup();
#endif

            result = copy_matrix(&fs_fixed_sharpen_mp, temp_mp);

            free_matrix(temp_mp);
        }
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option, "mip-db-sharp-init"))
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_mip_db_sharp_init_method == IDENTITY_SHARP_INIT)
            {
                ERE(pso("mip-db-sharp-init = identity\n"));
            }
            else if (fs_mip_db_sharp_init_method == AVE_ILLUM_SHARP_INIT)
            {
                ERE(pso("mip-db-sharp-init = ave\n"));
            }
            else if (fs_mip_db_sharp_init_method == FILE_SHARP_INIT)
            {
                ERE(pso("mip-db-sharp-init = <file>\n"));
            }
            else
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }

            result = NO_ERROR;
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_mip_db_sharp_init_method == IDENTITY_SHARP_INIT)
            {
                ERE(pso("MIP DB sharp inititialization is identity matrix.\n"));
            }
            else if (fs_mip_db_sharp_init_method == AVE_ILLUM_SHARP_INIT)
            {
                ERE(pso("MIP DB sharp inititialization is the ave illum sharp matrix.\n"));
            }
            else if (fs_mip_db_sharp_init_method == FILE_SHARP_INIT)
            {
                ERE(pso("MIP DB sharp inititialization matrix is:\n"));
                ERE(write_matrix(fs_initial_mip_db_sharpen_mp, (char*)NULL));
            }
            else
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }

            result = NO_ERROR;
        }
        else
        {
            if ((IC_STRCMP_EQ(value, "ave")) || (IC_STRCMP_EQ(value,"average")))
            {
                fs_mip_db_sharp_init_method = AVE_ILLUM_SHARP_INIT;
                result = NO_ERROR;
            }
            else if (    (IC_STRCMP_EQ(value, "i"))
                      || (IC_STRCMP_EQ(value, "id"))
                      || (IC_STRCMP_EQ(value, "identity"))
                    )
            {
                fs_mip_db_sharp_init_method = IDENTITY_SHARP_INIT;
                result = NO_ERROR;
            }
            else
            {
                ERE(read_matrix(&temp_mp, value));

#ifdef TRACK_MEMORY_ALLOCATION
                prepare_memory_cleanup();
#endif

                result = copy_matrix(&fs_initial_mip_db_sharpen_mp, temp_mp);

                free_matrix(temp_mp);

                if (result != ERROR)
                {
                    fs_mip_db_sharp_init_method = FILE_SHARP_INIT;
                }
            }
        }
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

/*ARGSUSED*/
int set_sharpening
(
    Sharp_method sharp_method,
    int          num_data_points,
    int*         data_point_nums,
    Vector*      illum_vp
)
{
    int result;

    if (sharp_method == DONT_SHARPEN)
    {
        result = set_sharp_matrix((Matrix*)NULL);
    }
    else if (    (sharp_method == FIXED_SHARP_CORRECTION)
              || (sharp_method == FIXED_SHARP)
            )
    {
        result = set_fixed_sharpening();
    }
    else if (    (sharp_method == SYNTHETIC_DB_SHARP_CORRECTION)
              || (sharp_method == SYNTHETIC_DB_SHARP)
            )
    {
        result = set_synthetic_data_base_sharpening();
    }
    else if (    (sharp_method == SYNTHETIC_MIP_DB_SHARP_CORRECTION)
              || (sharp_method == SYNTHETIC_MIP_DB_SHARP)
            )
    {
        result = set_synthetic_mip_db_sharpening();
    }
    else if (    (sharp_method == WORLD_DB_SHARP_CORRECTION)
              || (sharp_method == WORLD_DB_SHARP)
            )
    {
        result = set_world_data_base_sharpening();
    }
    else if (    (sharp_method == SCENE_DB_SHARP_CORRECTION)
              || (sharp_method == SCENE_DB_SHARP)
            )
    {
        result = set_scene_data_base_sharpening(num_data_points,
                                                data_point_nums);
    }
    else if (    (sharp_method == ITERATIVE_SHARP_CORRECTION)
              || (sharp_method == ITERATIVE_SHARP)
            )
    {
        result = set_iterative_sharpening(illum_vp);
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        result = ERROR;
    }

    if (result != ERROR)
    {
        fs_current_sharp_method = sharp_method;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int write_sharpening_transform
(
    char* request,
    char* output_file,
    int   append_flag
)
{
    Matrix* data_mp;
    char    lc_request[ 100 ];

    EXTENDED_LC_BUFF_CPY(lc_request, request);

    if (    (STRCMP_EQ(lc_request, "fixed-sharp"))
         || (STRCMP_EQ(lc_request, "fixed"))
       )
    {
        data_mp = fs_fixed_sharpen_mp;
    }
    else if (    (STRCMP_EQ(lc_request, "synthetic-db-sharp"))
              || (STRCMP_EQ(lc_request, "synthetic-db"))
            )
    {
        data_mp = fs_cached_db_sharpen_mp;
    }

    else if (    (STRCMP_EQ(lc_request, "synthetic-mip-db-sharp"))
              || (STRCMP_EQ(lc_request, "synthetic-mip-db"))
            )
    {
        data_mp = fs_cached_mip_db_sharpen_mp;
    }
    else
    {
        set_error("%q is an invalid sharpening transform for output.",
                  request);
        return ERROR;
    }

    if (data_mp == NULL)
    {
        set_error("No %q sharpening transform is available.", request);
        return ERROR;
    }
    else
    {
        return write_data_matrix(data_mp, output_file, append_flag);
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int write_sharp_sensors
(
    char* request,
    char* output_file,
    int   append_flag
)
{
    Matrix* sharp_mp;
    char    lc_request[ 100 ];

    EXTENDED_LC_BUFF_CPY(lc_request, request);

    if (    (STRCMP_EQ(lc_request, "fixed-sharp"))
         || (STRCMP_EQ(lc_request, "fixed"))
       )
    {
        sharp_mp = fs_fixed_sharpen_mp;
    }
    else if (    (STRCMP_EQ(lc_request, "synthetic-db-sharp"))
              || (STRCMP_EQ(lc_request, "synthetic-db"))
            )
    {
        sharp_mp = fs_cached_db_sharpen_mp;
    }

    else if (    (STRCMP_EQ(lc_request, "synthetic-mip-db-sharp"))
              || (STRCMP_EQ(lc_request, "synthetic-mip-db"))
            )
    {
        sharp_mp = fs_cached_mip_db_sharpen_mp;
    }
    else
    {
        set_error("%q is an invalid sharpening transform for output.",
                  request);
        return ERROR;
    }

    if (sharp_mp == NULL)
    {
        set_error("No %q sharpening transform is available.", request);
        return ERROR;
    }
    else
    {
        Spectra* sp = NULL;
        Matrix* sensor_mp = NULL;

        ERE(get_RGB_sensors(&sp, 101, 380.0, 4.0));
        ERE(multiply_with_transpose(&sensor_mp, sp->spectra_mp, sharp_mp));

        ERE(write_data_matrix(sensor_mp, output_file, append_flag));

        free_spectra(sp);
        free_matrix(sensor_mp);

        return NO_ERROR;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

Sharp_method get_current_sharp_method(void)
{
    return fs_current_sharp_method;

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void get_current_sharp_method_string(char* buff, size_t max_len)
{

    get_sharp_method_string(fs_current_sharp_method, buff, max_len);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void get_sharp_method_string
(
    Sharp_method sharp_method,
    char*        buff,
    size_t       max_len
)
{
    const char* str;

    if (sharp_method == DONT_SHARPEN)
    {
        str = "dull";
    }
    else if (sharp_method == FIXED_SHARP_CORRECTION)
    {
        str = "fixed-sc";
    }
    else if (sharp_method == SYNTHETIC_DB_SHARP_CORRECTION)
    {
        str = "syn-db-sc";
    }
    else if (sharp_method == SYNTHETIC_MIP_DB_SHARP_CORRECTION)
    {
        str = "syn-mip-db-sc";
    }
    else if (sharp_method == WORLD_DB_SHARP_CORRECTION)
    {
        str = "world-db-sc";
    }
    else if (sharp_method == SCENE_DB_SHARP_CORRECTION)
    {
        str = "scene-db-sc";
    }
    else if (sharp_method == ITERATIVE_SHARP_CORRECTION)
    {
        str = "iterative-sc";
    }
    else if (sharp_method == FIXED_SHARP)
    {
        str = "fixed-sharp";
    }
    else if (sharp_method == SYNTHETIC_DB_SHARP)
    {
        str = "syn-db-sharp";
    }
    else if (sharp_method == SYNTHETIC_MIP_DB_SHARP)
    {
        str = "syn-mip-db-sh";
    }
    else if (sharp_method == WORLD_DB_SHARP)
    {
        str = "world-db-sharp";
    }
    else if (sharp_method == SCENE_DB_SHARP)
    {
        str = "scene-db-sharp";
    }
    else if (sharp_method == ITERATIVE_SHARP)
    {
        str = "iterative-sharp";
    }
    else
    {
        SET_CANT_HAPPEN_BUG();
        return;
    }

    kjb_strncpy(buff, str, max_len);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_fixed_sharpening(void)
{

    if (fs_fixed_sharpen_mp == NULL)
    {
        set_error("No fixed sharpening matrix is available.\n");
        return ERROR;
    }

    return set_sharp_matrix(fs_fixed_sharpen_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_synthetic_data_base_sharpening(void)
{
    static int previous_sharp_version    = NOT_SET;
    static int previous_db_sharp_version = NOT_SET;
    int        db_sharp_version;

    ERE(db_sharp_version = make_synthetic_db_sharp_transform());

    if (    (fs_sharp_version == previous_sharp_version)
         && (db_sharp_version == previous_db_sharp_version)
       )
    {
        verbose_pso(10, "Sharpening transform is up to date. ");
        verbose_pso(10, "Skipping re-copy of DB sharpening transform.\n");
    }
    else
    {
        verbose_pso(5, "Updating sharpening transform with DB sharp.\n");

        ERE(set_sharp_matrix(fs_cached_db_sharpen_mp));
    }

    previous_db_sharp_version = db_sharp_version;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int make_synthetic_db_sharp_transform(void)
{
    static int previous_data_version       = NOT_SET;
    static int previous_canon_data_version = NOT_SET;
    int        canon_data_version;
    int        data_version;
    Matrix*    data_mp                     = NULL;
    Matrix*    canon_data_mp               = NULL;
    Matrix*    T_mp                        = NULL;
    int        result                      = NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    if (    (synthetic_matte_data_is_up_to_date(previous_data_version))
         && (synthetic_matte_canon_data_is_up_to_date(previous_canon_data_version))
       )
    {
        verbose_pso(10, "Database sharpening is up to date. ");
        verbose_pso(10, "Skipping recalculation.\n");

        result = fs_cached_db_sharp_version;
    }
    else
    {
        verbose_pso(5, "Database sharpening is out of date. Re-calculating.\n");

        ERE(data_version = get_synthetic_matte_data(&data_mp));

        canon_data_version= get_synthetic_matte_canon_data(&canon_data_mp);

        if (canon_data_version == ERROR) result = ERROR;

        if (result != ERROR)
        {
            result = get_db_sharp_transform(&T_mp, data_mp, canon_data_mp);
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_db_sharpen_mp, T_mp);
        }

        if (result != ERROR)
        {
            previous_data_version = data_version;
            previous_canon_data_version = canon_data_version;

            fs_cached_db_sharp_version++;

            result = fs_cached_db_sharp_version;
        }
    }

    free_matrix(T_mp);
    free_matrix(data_mp);
    free_matrix(canon_data_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_synthetic_mip_db_sharpening(void)
{

    if (fs_cached_mip_db_sharpen_mp == NULL)
    {
        set_error("No MIP DB sharpening matrix is available.\n");
        return ERROR;
    }

    ERE(set_sharp_matrix(fs_cached_mip_db_sharpen_mp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_sharp(const char* make_directive, const char* file_name)
{
    char lc_make_directive[ 100 ];
    int  result = NOT_FOUND;

    if (make_directive == NULL)
    {
        BUFF_CPY(lc_make_directive, "all");
    }
    else
    {
        EXTENDED_LC_BUFF_CPY(lc_make_directive, make_directive);
    }

    if (    STRCMP_EQ(lc_make_directive, "all")
         || STRCMP_EQ(lc_make_directive, "sharp")
         || STRCMP_EQ(lc_make_directive, "sharp-all")
         || STRCMP_EQ(lc_make_directive, "sharp-mip-db-transform")
         || STRCMP_EQ(lc_make_directive, "sharp-synthetic-mip-db-transform")
       )
    {
        ERE(make_synthetic_mip_db_sharp_transform(file_name));
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int make_synthetic_mip_db_sharp_transform(const char* file_name)
{
    static int previous_illum_spectra_version   = NOT_SET;
    static int previous_reflect_spectra_version = NOT_SET;
    static int previous_canon_spectra_version   = NOT_SET;
    static int previous_sensor_version          = NOT_SET;
    static int previous_env_version             = NOT_SET;
    int        illum_spectra_version            = NOT_SET;
    int        reflect_spectra_version          = NOT_SET;
    int        canon_spectra_version            = NOT_SET;
    int        sensor_version                   = NOT_SET;
    int        env_version                      = NOT_SET;
    Spectra*   reflect_spectra_sp               = NULL;
    Spectra*   illum_spectra_sp                 = NULL;
    Spectra*   canon_spectra_sp                 = NULL;
    Matrix*    T_mp                             = NULL;
    Matrix*    initial_T_mp                     = NULL;
    int        result                           = NO_ERROR;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    if (    (sensor_data_is_up_to_date(previous_sensor_version))
         && (illum_spectra_is_up_to_date(previous_illum_spectra_version))
         && (reflect_spectra_is_up_to_date(previous_reflect_spectra_version))
         && (canon_spectra_is_up_to_date(previous_canon_spectra_version))
         && (sharp_db_mip_env_is_up_to_date(previous_env_version))
       )
    {
        verbose_pso(10, "MIP DB sharpening transform is up to date. ");
        verbose_pso(10, "Skipping re-calculation.\n");

        result = fs_cached_mip_db_sharp_version;
    }
    else
    {
        verbose_pso(5, "Re-making MIP DB sharpening transform.\n");

        if (result != ERROR)
        {
            illum_spectra_version = get_illum_spectra(&illum_spectra_sp);
            if (illum_spectra_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            reflect_spectra_version = get_reflect_spectra(&reflect_spectra_sp);
            if (reflect_spectra_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            canon_spectra_version = get_canon_spectra(&canon_spectra_sp);
            if (canon_spectra_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            if (fs_mip_db_sharp_init_method == IDENTITY_SHARP_INIT)
            {
                result = get_identity_matrix(&initial_T_mp, 3);
            }
            else if (fs_mip_db_sharp_init_method == AVE_ILLUM_SHARP_INIT)
            {
                result = get_average_illum_db_sharp(&initial_T_mp);
            }
            else if (fs_initial_mip_db_sharpen_mp != NULL)
            {
                result = copy_matrix(&initial_T_mp, fs_initial_mip_db_sharpen_mp);
            }
            else
            {
                SET_CANT_HAPPEN_BUG();
                return ERROR;
            }
        }

        if (result != ERROR)
        {
            env_version = get_mip_db_sharp_transform(&T_mp, reflect_spectra_sp,
                                                     illum_spectra_sp,
                                                     canon_spectra_sp,
                                                     initial_T_mp);
            if (env_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            result = copy_matrix(&fs_cached_mip_db_sharpen_mp, T_mp);
        }

        if (result != ERROR)
        {
            sensor_version = get_sensor_version();

            if (sensor_version == ERROR) result = ERROR;
        }

        if (result != ERROR)
        {
            ASSERT(IS_SET(sensor_version));
            previous_sensor_version = sensor_version;

            ASSERT(IS_SET(illum_spectra_version));
            previous_illum_spectra_version = illum_spectra_version;

            ASSERT(IS_SET(reflect_spectra_version));
            previous_reflect_spectra_version = reflect_spectra_version;

            ASSERT(IS_SET(canon_spectra_version));
            previous_canon_spectra_version = canon_spectra_version;

            ASSERT(IS_SET(env_version));
            previous_env_version = env_version;

            fs_cached_mip_db_sharp_version++;

            result = fs_cached_mip_db_sharp_version;
        }
    }

    if (result != ERROR)
    {
        result = write_matrix(fs_cached_mip_db_sharpen_mp, file_name);
    }

    free_matrix(T_mp);
    free_matrix(initial_T_mp);
    free_spectra(illum_spectra_sp);
    free_spectra(reflect_spectra_sp);
    free_spectra(canon_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_world_data_base_sharpening(void)
{
    Matrix* data_mp          = NULL;
    Matrix* canon_data_mp    = NULL;
    Matrix* sharp_mp         = NULL;
    int     result;

    /*
    // Can't cache the result here, unless data is cached in data.c (with a
    // version count), and this is not worth the trouble. Thus we have to
    // recompute the result everytime.  This is justified because we rarely use
    // this sharpening method.
    */
    result = get_registered_data(&data_mp);

    if (result != ERROR)
    {
        result = get_canon_data(&canon_data_mp);
    }

    if (result != ERROR)
    {
        result = get_db_sharp_transform(&sharp_mp, data_mp, canon_data_mp);
    }

    if (result != ERROR)
    {
        result = set_sharp_matrix(sharp_mp);
    }

    free_matrix(data_mp);
    free_matrix(canon_data_mp);
    free_matrix(sharp_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_scene_data_base_sharpening
(
    int  num_data_points,
    int* data_point_nums
)
{
    Matrix* selected_data_mp       = NULL;
    Matrix* selected_canon_data_mp = NULL;
    Matrix* sharp_mp               = NULL;
    int     result;

    /*
    // Can't cache the result here, unless data is cached in data.c (with a
    // version count), and this is not worth the trouble. Thus we have to
    // recompute the result everytime.  This is justified because we rarely use
    // this sharpening method.
    */
    result = get_selected_registered_data(&selected_data_mp, num_data_points,
                                          data_point_nums);

    if (result != ERROR)
    {
        result = get_selected_canon_data(&selected_canon_data_mp,
                                         num_data_points, data_point_nums);
    }

    if (result != ERROR)
    {
        result = get_db_sharp_transform(&sharp_mp, selected_data_mp,
                                        selected_canon_data_mp);
    }

    if (result != ERROR)
    {
        result = set_sharp_matrix(sharp_mp);
    }

    free_matrix(selected_data_mp);
    free_matrix(selected_canon_data_mp);
    free_matrix(sharp_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_iterative_sharpening(Vector* illum_rgb_vp)
{
    Spectra* illum_sp               = NULL;
    Spectra* canon_illum_sp         = NULL;
    Spectra* reflect_sp             = NULL;
    Matrix*  illum_rgb_mp           = NULL;
    Matrix*  projected_illum_rgb_mp = NULL;
    Matrix*  data_mp                = NULL;
    Matrix*  canon_data_mp          = NULL;
    Matrix*  sharp_mp               = NULL;
    int      result;
    double     illum_r;
    double     illum_g;

    if (illum_rgb_vp == NULL)
    {
        set_error("No illuminant estimate available for iterative sharpening.");
        return ERROR;
    }

    ERE(make_illum_gamut((char*)NULL, (char*)NULL));

    ERE(get_illum_spectra(&illum_sp));

    result = get_canon_spectra(&canon_illum_sp);

    if (result != ERROR)
        result = get_reflect_spectra(&reflect_sp);

    if (result != ERROR)
        result = get_RGB_data_from_spectra(&illum_rgb_mp, illum_sp,
                                           (Spectra*)NULL);

    if (result != ERROR)
    {
        result = get_projection_matrix(&projected_illum_rgb_mp, illum_rgb_mp,
                                       DIVIDE_BY_SUM, SKIP_INVALID_CHROMATICITY,
                                       MAX_X_DIV_SUM, DBL_NOT_SET, DBL_NOT_SET,
                                       (Matrix**)NULL, (Matrix**)NULL);
    }

    if (result != ERROR)
    {
        result = get_rg_from_RGB(illum_rgb_vp->elements[ 0 ],
                                 illum_rgb_vp->elements[ 1 ],
                                 illum_rgb_vp->elements[ 2 ],
                                 &illum_r, &illum_g);
    }

    if (result != ERROR)
    {
        int  min_index = 0;
        double r_diff = projected_illum_rgb_mp->elements[ 0 ][ 0 ] - illum_r;
        double g_diff = projected_illum_rgb_mp->elements[ 0 ][ 1 ] - illum_g;
        double min_diff = r_diff * r_diff + g_diff * g_diff;
        int  i;
        double diff;

        for (i=1; i<projected_illum_rgb_mp->num_rows; i++)
        {
            r_diff = projected_illum_rgb_mp->elements[ i ][ 0 ] - illum_r;
            g_diff = projected_illum_rgb_mp->elements[ i ][ 1 ] - illum_g;
            diff = r_diff * r_diff + g_diff * g_diff;

            if (diff < min_diff)
            {
                min_diff = diff;
                min_index = i;
            }
        }

        result = generate_RGB_data(&data_mp, illum_sp, min_index, reflect_sp,
                                   (Spectra*)NULL);
    }

    if (result != ERROR)
    {
        result = generate_RGB_data(&canon_data_mp, canon_illum_sp, 0,
                                   reflect_sp,
                                   (Spectra*)NULL);
    }

    if (result != ERROR)
    {
        result = get_db_sharp_transform(&sharp_mp, data_mp, canon_data_mp);
    }

    if (result != ERROR)
    {
        result = set_sharp_matrix(sharp_mp);
    }

    free_matrix(data_mp);
    free_matrix(canon_data_mp);

    free_matrix(sharp_mp);

    free_spectra(canon_illum_sp);
    free_spectra(illum_sp);
    free_spectra(reflect_sp);

    free_matrix(projected_illum_rgb_mp);
    free_matrix(illum_rgb_mp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_average_illum_db_sharp(Matrix** sharp_mpp)
{
    Spectra* ave_illum_sp               = NULL;
    Spectra* illum_sp               = NULL;
    Spectra* canon_illum_sp         = NULL;
    Spectra* reflect_sp             = NULL;
    Vector*  ave_illum_vp           = NULL;
    Matrix*  data_mp                = NULL;
    Matrix*  canon_data_mp          = NULL;
    int      result;

    ERE(get_illum_spectra(&illum_sp));

    /*
    // This should no longer be needed: Illuminant spectra are now always
    // normalized.
    //
    // ERE(ow_normalize_matrix_rows(illum_sp->spectra_mp));
    */

    ERE(average_matrix_rows(&ave_illum_vp, illum_sp->spectra_mp));

    ERE(get_target_spectra(&ave_illum_sp, 1, illum_sp->spectra_mp->num_cols,
                           illum_sp->offset, illum_sp->step, illum_sp->type));

    ERE(put_matrix_row(ave_illum_sp->spectra_mp, ave_illum_vp, 0));

    result = get_canon_spectra(&canon_illum_sp);

    if (result != ERROR)
    {
        result = get_reflect_spectra(&reflect_sp);
    }

    if (result != ERROR)
    {
        result = generate_RGB_data(&data_mp, ave_illum_sp, 0, reflect_sp,
                                   (Spectra*)NULL);
    }

    if (result != ERROR)
    {
        result = generate_RGB_data(&canon_data_mp, canon_illum_sp, 0,
                                   reflect_sp, (Spectra*)NULL);
    }

    if (result != ERROR)
    {
        result = get_db_sharp_transform(sharp_mpp, data_mp, canon_data_mp);
    }

    free_matrix(data_mp);
    free_matrix(canon_data_mp);

    free_vector(ave_illum_vp);
    free_spectra(canon_illum_sp);
    free_spectra(illum_sp);
    free_spectra(reflect_sp);
    free_spectra(ave_illum_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_sharp_matrix(Matrix* mp)
{
    Matrix* temp_inv_mp = NULL;
    int     result;

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_sharp_version++;

    if (mp == NULL)
    {
        free_matrix(fs_sharpen_mp);
        fs_sharpen_mp = NULL;

        free_matrix(fs_inv_sharpen_mp);
        fs_inv_sharpen_mp = NULL;

        return NO_ERROR;
    }

    ERE(get_matrix_inverse(&temp_inv_mp, mp));

    result = copy_matrix(&fs_sharpen_mp, mp);

    if (result != ERROR)
    {
        result = copy_matrix(&fs_inv_sharpen_mp, temp_inv_mp);
    }

    free_matrix(temp_inv_mp);

    if (result == ERROR)
    {
        free_matrix(fs_sharpen_mp);
        fs_sharpen_mp = NULL;

        free_matrix(fs_inv_sharpen_mp);
        fs_inv_sharpen_mp = NULL;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_sharp_version(void)
{
    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int sharpen_matrix_data(Matrix** sharp_data_mpp, Matrix* data_mp)
{

    if (fs_sharpen_mp == NULL)
    {
        ERE(copy_matrix(sharp_data_mpp, data_mp));
    }
    else if (data_mp == NULL)
    {
        free_matrix(*sharp_data_mpp);
        *sharp_data_mpp = NULL;
    }
    else
    {
        ERE(multiply_matrices(sharp_data_mpp, data_mp, fs_sharpen_mp));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_sharpen_matrix_data(Matrix* data_mp)
{
    int result = fs_sharp_version;

    if ((fs_sharpen_mp != NULL) && (data_mp != NULL))
    {
        Matrix* sharp_data_mp = NULL;

        ERE(multiply_matrices(&sharp_data_mp, data_mp, fs_sharpen_mp));

        if (copy_matrix(&data_mp, sharp_data_mp) == ERROR)
        {
            result = ERROR;
        }

        free_matrix(sharp_data_mp);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int unsharpen_matrix_data(Matrix** data_mpp, Matrix* sharp_data_mp)
{

    if (fs_inv_sharpen_mp == NULL)
    {
        ERE(copy_matrix(data_mpp, sharp_data_mp));
    }
    else if (sharp_data_mp == NULL)
    {
        free_matrix(*data_mpp);
        *data_mpp = NULL;
    }
    else
    {
        ERE(multiply_matrices(data_mpp, sharp_data_mp, fs_inv_sharpen_mp));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_unsharpen_matrix_data(Matrix* data_mp)
{
    int result = fs_sharp_version;

    if ((fs_inv_sharpen_mp != NULL) && (data_mp != NULL))
    {
        Matrix* dull_data_mp = NULL;

        ERE(multiply_matrices(&dull_data_mp, data_mp, fs_inv_sharpen_mp));

        if (copy_matrix(&data_mp, dull_data_mp) == ERROR)
        {
            result = ERROR;
        }

        free_matrix(dull_data_mp);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_sharpen_vector_data(Vector* data_vp)
{
    int result = fs_sharp_version;

    if (fs_sharpen_mp != NULL)
    {
        Vector* sharp_data_vp = NULL;

        ERE(sharpen_vector_data(&sharp_data_vp, data_vp));

        if (copy_vector(&data_vp, sharp_data_vp) == ERROR)
        {
            result = ERROR;
        }

        free_vector(sharp_data_vp);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int sharpen_vector_data(Vector** sharp_data_vpp, Vector* data_vp)
{

    if (fs_sharpen_mp == NULL)
    {
        ERE(copy_vector(sharp_data_vpp, data_vp));
    }
    else if (data_vp == NULL)
    {
        free_vector(*sharp_data_vpp);
        *sharp_data_vpp = NULL;
        return NO_ERROR;
    }
    else
    {
        ERE(multiply_vector_and_matrix(sharp_data_vpp, data_vp, fs_sharpen_mp));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int ow_unsharpen_vector_data(Vector* data_vp)
{
    int result = fs_sharp_version;

    if (fs_inv_sharpen_mp != NULL)
    {
        Vector* dull_data_vp = NULL;

        ERE(unsharpen_vector_data(&dull_data_vp, data_vp));

        if (copy_vector(&data_vp, dull_data_vp) == ERROR)
        {
            result = ERROR;
        }

        free_vector(dull_data_vp);
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int unsharpen_vector_data(Vector** data_vpp, Vector* sharp_data_vp)
{

    if (fs_inv_sharpen_mp == NULL)
    {
        ERE(copy_vector(data_vpp, sharp_data_vp));
    }
    else if (sharp_data_vp == NULL)
    {
        free_vector(*data_vpp);
        *data_vpp = NULL;
    }
    else
    {
        ERE(multiply_vector_and_matrix(data_vpp, sharp_data_vp,
                                       fs_inv_sharpen_mp));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int sharpen_rgb
(
    double* sharp_R_ptr,
    double* sharp_G_ptr,
    double* sharp_B_ptr,
    double  R,
    double  G,
    double  B
)
{

    if (fs_sharpen_mp == NULL)
    {
        *sharp_R_ptr = R;
        *sharp_G_ptr = G;
        *sharp_B_ptr = B;
    }
    else
    {
        ERE(map_rgb(sharp_R_ptr, sharp_G_ptr, sharp_B_ptr, fs_sharpen_mp, R, G, B));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int unsharpen_rgb
(
    double* R_ptr,
    double* G_ptr,
    double* B_ptr,
    double  sharp_R,
    double  sharp_G,
    double  sharp_B
)
{

    if (fs_inv_sharpen_mp == NULL)
    {
        *R_ptr = sharp_R;
        *G_ptr = sharp_G;
        *B_ptr = sharp_B;
    }
    else
    {
        ERE(map_rgb(R_ptr, G_ptr, B_ptr, fs_inv_sharpen_mp, sharp_R,
                    sharp_G,sharp_B));
    }

    return fs_sharp_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int map_rgb
(
    double* map_R_ptr,
    double* map_G_ptr,
    double* map_B_ptr,
    Matrix* map_mp,
    double  R,
    double  G,
    double  B
)
{
    static double    data[ 3 ];
    static Vector  data_vector = { 3, 3, data };
    static double    map_data[ 3 ];
    static Vector  map_data_vector = { 3, 3, map_data };
    Vector*        map_data_vp;

    data[ 0 ] = R;
    data[ 1 ] = G;
    data[ 2 ] = B;

    map_data_vp = &map_data_vector;

    ERE(multiply_vector_and_matrix(&map_data_vp, &data_vector, map_mp));

    *map_R_ptr = map_data[ 0 ];
    *map_G_ptr = map_data[ 1 ];
    *map_B_ptr = map_data[ 2 ];

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int map_projected_sharp_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp
)
{

    if (fs_sharpen_mp == NULL) return ERROR;

    return change_basis_post_map_projected_image(target_ipp, in_ip,
                                                 fs_sharpen_mp, map_mp);

}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int map_sharp_image
(
    KJB_image**      target_ipp,
    const KJB_image* in_ip,
    Matrix*          map_mp
)
{

    if (fs_sharpen_mp == NULL) return ERROR;

    return change_basis_post_map_image(target_ipp, in_ip, fs_sharpen_mp,
                                       map_mp);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_fixed_sharp_maps_by_algorithm
(
    Result_by_algorithm* fixed_sharp_result_ptr,
    Result_by_algorithm* dull_result_ptr
)
{
    Vector* canon_illum_vp = NULL;
    int i;

    (void)get_canon_illum(&canon_illum_vp);

    ERE(get_best_fit_fixed_sharp_maps(&(fixed_sharp_result_ptr->best_fit),
                                      &(dull_result_ptr->best_fit),
                                      canon_illum_vp));

    ERE(get_standard_fixed_sharp_maps(&(fixed_sharp_result_ptr->standard),
                                      &(dull_result_ptr->standard),
                                      canon_illum_vp));

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(get_gamut_fixed_sharp_maps(&(fixed_sharp_result_ptr->gamut_2D[ i ]),
                                       &(dull_result_ptr->gamut_2D[ i ]),
                                       canon_illum_vp));
    }

    for (i=0; i<MAX_NUM_CANON_GAMUTS; i++)
    {
        ERE(get_gamut_fixed_sharp_maps(&(fixed_sharp_result_ptr->gamut_3D[ i ]),
                                       &(dull_result_ptr->gamut_3D[ i ]),
                                       canon_illum_vp));
    }

    ERE(get_correlation_fixed_sharp_maps(&(fixed_sharp_result_ptr->correlation),
                                         &(dull_result_ptr->correlation),
                                         canon_illum_vp));

#ifdef SIMPLE_TEMPLATE
    ERE(get_simple_template_fixed_sharp_maps(
                                      &(fixed_sharp_result_ptr->simple_template),
                                      &(dull_result_ptr->simple_template),
                                      canon_illum_vp));
#endif

#ifdef TEMPLATE
    ERE(get_template_fixed_sharp_maps(&(fixed_sharp_result_ptr->template),
                                      &(dull_result_ptr->template),
                                      canon_illum_vp));
#endif

#ifdef MULTI_TEMPLATE
    ERE(get_multi_template_fixed_sharp_maps(
                                      &(fixed_sharp_result_ptr->multi_template),
                                      &(dull_result_ptr->multi_template),
                                      canon_illum_vp));
#endif

    free_vector(canon_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_best_fit_fixed_sharp_maps
(
    Best_fit_result* sharp_result_ptr,
    Best_fit_result* dull_result_ptr,
    Vector*          canon_illum_vp
)
{

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->world_best_3D_diag_fit_vp),
                                     dull_result_ptr->world_best_3D_diag_fit_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->world_best_2D_diag_fit_vp),
                                     dull_result_ptr->world_best_2D_diag_fit_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->scene_best_3D_diag_fit_vp),
                                     dull_result_ptr->scene_best_3D_diag_fit_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->scene_best_2D_diag_fit_vp),
                                     dull_result_ptr->scene_best_2D_diag_fit_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->actual_vp),
                                     dull_result_ptr->actual_vp,
                                     canon_illum_vp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_standard_fixed_sharp_maps
(
    Standard_result* sharp_result_ptr,
    Standard_result* dull_result_ptr,
    Vector*          canon_illum_vp
)
{

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->ave_illum_vp),
                                     dull_result_ptr->ave_illum_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->grey_world_vp),
                                     dull_result_ptr->grey_world_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->db_grey_vp),
                                     dull_result_ptr->db_grey_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->scale_by_max_vp),
                                     dull_result_ptr->scale_by_max_vp,
                                     canon_illum_vp));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_gamut_fixed_sharp_maps
(
    Gamut_result* sharp_result_ptr,
    Gamut_result* dull_result_ptr,
    Vector*       canon_illum_vp
)
{
    int i;

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->mv_surface_vp),
                                     dull_result_ptr->mv_surface_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->mv_surface_illum_vp),
                                     dull_result_ptr->mv_surface_illum_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->ha_surface_vp),
                                     dull_result_ptr->ha_surface_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->ha_surface_illum_vp),
                                     dull_result_ptr->ha_surface_illum_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->icsa_surface_illum_vp),
                                     dull_result_ptr->icsa_surface_illum_vp,
                                     canon_illum_vp));

    for (i=0; i<MAX_NUM_SCA_WEIGHT_FUNCTIONS; i++)
    {
        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->scia_surface_illum_vp[ i ]),
                                         dull_result_ptr->scia_surface_illum_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->scca_surface_illum_vp[ i ]),
                                         dull_result_ptr->scca_surface_illum_vp[ i ],
                                         canon_illum_vp));
    }

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->wa_surface_vp),
                                     dull_result_ptr->wa_surface_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->wa_surface_illum_vp),
                                     dull_result_ptr->wa_surface_illum_vp,
                                     canon_illum_vp));

    for (i=0; i<MAX_NUM_LUMINANCE_METHODS; i++)
    {
        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->ha_surface_lum_vp[ i ]),
                    dull_result_ptr->ha_surface_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->mv_surface_lum_vp[ i ]),
                    dull_result_ptr->mv_surface_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->ha_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->ha_surface_illum_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->mv_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->mv_surface_illum_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->scla_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->scla_surface_illum_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->lcsa_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->scla_surface_illum_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->scia_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->scla_surface_illum_lum_vp[ i ],
                    canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(
                    &(sharp_result_ptr->scca_surface_illum_lum_vp[ i ]),
                    dull_result_ptr->scla_surface_illum_lum_vp[ i ],
                    canon_illum_vp));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_correlation_fixed_sharp_maps
(
    Correlation_result* sharp_result_ptr,
    Correlation_result* dull_result_ptr,
    Vector*             canon_illum_vp
)
{
    int i; /* Modified April 14/99 */

    for (i=0; i<NUM_CORMATS_PER_ALG; i++)
    {
        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_01_vp[ i ]),
                                         dull_result_ptr->correlation_01_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_01_alt_vp[ i ]),
                                         dull_result_ptr->correlation_01_alt_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_gaussian_MAP_vp[ i ]),
                                         dull_result_ptr->correlation_gaussian_MAP_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MAP_inv_dist_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MAP_gaussian_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MAP_fitted_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MAP_fitted_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_gaussian_MLM_vp[ i ]),
                                         dull_result_ptr->correlation_gaussian_MLM_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MLM_inv_dist_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MLM_gaussian_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MLM_fitted_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MLM_fitted_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_gaussian_MMSE_vp[ i ]),
                                         dull_result_ptr->correlation_gaussian_MMSE_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MMSE_inv_dist_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MMSE_gaussian_fill_vp[ i ],
                                         canon_illum_vp));

        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->correlation_uniform_MMSE_fitted_vp[ i ]),
                                         dull_result_ptr->correlation_uniform_MMSE_fitted_vp[ i ],
                                         canon_illum_vp));
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef SIMPLE_TEMPLATE

static int get_simple_template_fixed_sharp_maps
(
    Simple_template_result* sharp_result_ptr,
    Simple_template_result* dull_result_ptr,
    Vector*          canon_illum_vp
)
{

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->alg_vp),
                                     dull_result_ptr->alg_vp,
                                     canon_illum_vp));

    return NO_ERROR;
}
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TEMPLATE

static int get_template_fixed_sharp_maps
(
    Template_result* sharp_result_ptr,
    Template_result* dull_result_ptr,
    Vector*          canon_illum_vp
)
{

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->first_alg_vp),
                                     dull_result_ptr->first_alg_vp,
                                     canon_illum_vp));

    ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->second_alg_vp),
                                     dull_result_ptr->second_alg_vp,
                                     canon_illum_vp));

    return NO_ERROR;
}
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef MULTI_TEMPLATE
static int get_multi_template_fixed_sharp_maps
(
    Multi_template_result* sharp_result_ptr,
    Multi_template_result* dull_result_ptr,
    Vector*                canon_illum_vp
)
{
    int i;

    for (i=0; i<MAX_NUM_MULTI_TEMPLATE_ALGS; i++)
    {
        ERE(get_fixed_sharp_diagonal_map(&(sharp_result_ptr->alg_vp[ i ]),
                                         dull_result_ptr->alg_vp[ i ],
                                         canon_illum_vp));
    }

    return NO_ERROR;
}
#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int get_fixed_sharp_diagonal_map
(
    Vector** sharp_diag_vpp,
    Vector*  diag_vp,
    Vector*  canon_illum_vp
)
{
    Vector* scratch_rgb_diag_vp = NULL;
    Vector* rgb_diag_vp;
    Vector* white_rgb_vp = NULL;
    Vector* sharp_white_rgb_vp = NULL;
    Vector* sharp_canon_illum_vp = NULL;
    Vector* sharp_3D_diag_vp = NULL;

    if (diag_vp == NULL)
    {
        free_vector(*sharp_diag_vpp);
        *sharp_diag_vpp = NULL;
	
        return NO_ERROR;
    }
    else if (diag_vp->length == 2)
    {
        ERE(back_project_vector(&scratch_rgb_diag_vp, diag_vp, DIVIDE_BY_BLUE));
        rgb_diag_vp = scratch_rgb_diag_vp;
    }
    else
    {
        rgb_diag_vp = diag_vp;
    }

    ERE(divide_vectors(&white_rgb_vp, canon_illum_vp, rgb_diag_vp));

    ERE(sharpen_vector_data(&sharp_white_rgb_vp, white_rgb_vp));
    ERE(sharpen_vector_data(&sharp_canon_illum_vp, canon_illum_vp));

    ERE(divide_vectors(&sharp_3D_diag_vp, sharp_canon_illum_vp,
                       sharp_white_rgb_vp));

    if (diag_vp->length == 2)
    {
        ERE(get_projection_vector(sharp_diag_vpp, sharp_3D_diag_vp,
                                  DIVIDE_BY_BLUE, MAX_X_DIV_B));
    }
    else
    {
        ERE(copy_vector(sharp_diag_vpp, sharp_3D_diag_vp));
    }

    free_vector(sharp_3D_diag_vp);
    free_vector(scratch_rgb_diag_vp);
    free_vector(white_rgb_vp);
    free_vector(sharp_white_rgb_vp);
    free_vector(sharp_canon_illum_vp);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{
    free_matrix(fs_initial_mip_db_sharpen_mp);
    free_matrix(fs_fixed_sharpen_mp);
    free_matrix(fs_sharpen_mp);
    free_matrix(fs_inv_sharpen_mp);
    free_matrix(fs_cached_mip_db_sharpen_mp);
    free_matrix(fs_cached_db_sharpen_mp);
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

