
/* =========================================================================== *
|
|  Copyright (c) 1994-2002 by Kobus Barnard (author).
|
|  Personal and educational use of this code is granted, provided that this
|  header is kept intact, and that the authorship is not misrepresented, that
|  its use is acknowleged in publications, and relevent papers are cited.
|
|  For other use contact the author (kobus AT cs DOT arizona DOT edu).
|
|  Please note that the code in this file has not necessarily been adequately
|  tested. Naturally, there is no guarentee of performance, support, or fitness
|  for any particular task. Nonetheless, I am interested in hearing about
|  problems that you encounter.
|
* =========================================================================== */

#include "gen.h"     /* Only safe as first include in a ".c" file. */
                       /* Otherwise use gen.h.                       */

#include "c2/c2_fluorescence.h"

#include "sample/sample_gauss.h" 

#include "spectra.h"

/* -------------------------------------------------------------------------- */

#define ILLUM_SPECTRA_FILE             "illum_db.illum"
#define REFLECT_SPECTRA_FILE           "reflect_db.reflect"
#define CANON_ILLUM_SPECTRA_FILE       "canon_illum.illum"
#define ACTUAL_ILLUM_SPECTRA_FILE      "actual_illum.illum"

#ifdef PROVIDE_DEFAULT_SPECULAR_REFLECT_SPECTRA_FILE
#    define SPECULAR_REFLECT_SPECTRA_FILE  "specular_reflect_db.reflect"
#endif

#ifdef PROVIDE_DEFAULT_FLUORESCENT_DATABASE_FILE
#    define FLUORESCENT_DATABASE_FILE      "fluorescent_db.fl"
#endif

#define USE_RANDOM (-1)

/* -------------------------------------------------------------------------- */

static Spectra* fs_cached_illum_spectra_sp             = NULL;
static Spectra* fs_cached_reflect_spectra_sp           = NULL;
static Spectra* fs_cached_specular_spectra_sp          = NULL;
static Spectra* fs_cached_canon_spectra_sp             = NULL;
static Spectra* fs_cached_actual_spectra_sp            = NULL;
static Spectra* fs_cached_normalized_actual_spectra_sp = NULL;
static Spectra* fs_cached_reflect_spectra_with_spec_sp = NULL;
static Fluorescent_database* fs_cached_fluorescent_database_ptr = NULL;

static char fs_illum_spectra_config_file[ MAX_FILE_NAME_SIZE ]   = { '\0' };
static char fs_reflect_spectra_config_file[ MAX_FILE_NAME_SIZE ] = { '\0' };
static char fs_canon_spectra_config_file[ MAX_FILE_NAME_SIZE ]   = { '\0' };
static char fs_actual_spectra_config_file[ MAX_FILE_NAME_SIZE ]  = { '\0' };
static char fs_specular_spectra_config_file[ MAX_FILE_NAME_SIZE ] = {'\0'};
static char fs_fluorescent_database_config_file[ MAX_FILE_NAME_SIZE ] = {'\0'};

static int fs_world_version                     = 0;
static int fs_illum_spectra_version             = 0;
static int fs_canon_spectra_version             = 0;
static int fs_actual_spectra_version            = 0;
static int fs_reflect_spectra_version           = 0;
static int fs_specular_spectra_version  = 0;
static int fs_reflect_spectra_with_spec_version = 0;
static int fs_fluorescent_database_version      = 0;

/*
// This option is relatively useless, except for debugging. All algorithms
// should be invariant with respect to this option (except the average
// illuminant option, as of 04/98).
*/
static double fs_actual_spectra_scale = DBL_NOT_SET;

static double fs_specular_fraction                  = 0.0;
static double fs_specular_amount                    = 0.0;
static double fs_highlight_fraction                 = 0.0;
static double fs_highlight_amount                   = 0.0;
static double fs_white_specular_reflection_fraction = 1.0;
static int  fs_fluorescent_weight_factor          = 1;

static double fs_actual_illum_normalization_level = 255.0;

/* -------------------------------------------------------------------------- */

static int read_illum_spectra_from_file(const char* file_name);

static int read_reflect_spectra_from_file(const char* file_name);

static int read_canon_spectra_from_file(const char* file_name);

static int read_actual_spectra_from_file(const char* file_name);

static int read_specular_spectra_from_file(const char* file_name);

static int read_fluorescent_database_from_file(const char* file_name);

static int set_illum_spectra
(
    Spectra* new_illum_spectra_sp,
    char*    config_file
);

static int set_reflect_spectra
(
    Spectra* new_reflect_spectra_sp,
    char*    config_file
);

static int set_canon_spectra
(
    Spectra* new_canon_spectra_sp,
    char*    config_file
);

static int set_actual_spectra
(
    Spectra* new_actual_spectra_sp,
    char*    config_file
);

static int set_specular_spectra
(
    Spectra* new_specular_spectra_sp,
    char*    config_file
);

static int set_fluorescent_database
(
    Fluorescent_database* new_fl_db_ptr,
    char*                 config_file
);

static int initialize_illum_spectra(void);
static int initialize_reflect_spectra(void);
static int initialize_fluorescent_database(void);
static int update_world_version(void);

#ifdef TRACK_MEMORY_ALLOCATION
    static void free_allocated_static_data(void);
    static void prepare_memory_cleanup(void);
#endif

/* -------------------------------------------------------------------------- */

EXTERN_C int set_spectra_options(const char* option, const char* value)
{
    char lc_option[ 100 ];
    int  result = NOT_FOUND;

    EXTENDED_LC_BUFF_CPY(lc_option, option);

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "isf"))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("isf = %s\n",
                    fs_illum_spectra_config_file[ 0 ] ? fs_illum_spectra_config_file
                                                   : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_illum_spectra_sp == NULL)
            {
                ERE(pso("No illuminant spectra database has been set.\n"));
            }
            else
            {
                ERE(pso("Illumiant spectra database source is %s.\n",
                        fs_illum_spectra_config_file));
            }
        }
        else
        {
            ERE(read_illum_spectra_from_file(value));
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "rsf"))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("rsf = %s\n",
                    fs_reflect_spectra_config_file[ 0 ] ? fs_reflect_spectra_config_file
                                                   : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_reflect_spectra_sp == NULL)
            {
                ERE(pso("No reflectance spectra database has been set.\n"));
            }
            else
            {
                ERE(pso("Reflectance spectra database source is %s.\n",
                        fs_reflect_spectra_config_file));
            }
        }
        else
        {
            ERE(read_reflect_spectra_from_file(value));
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "aisf"))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("aisf = %s\n",
                    fs_actual_spectra_config_file[ 0 ] ? fs_actual_spectra_config_file
                                                   : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_actual_spectra_sp == NULL)
            {
                ERE(pso("No actual illuminant spectra has been set.\n"));
            }
            else
            {
                ERE(pso("Actual illuminant spectra source is %s.\n",
                        fs_actual_spectra_config_file));
            }
        }
        else if (all_digits(value))
        {
            int temp_int;

            ERE(ss1pi(value, &temp_int));
            ERE(set_actual_spectra_by_number(temp_int));
        }
        else if (STRCMP_EQ(value, "#"))
        {
            ERE(set_random_actual_spectra());
        }
        else
        {
            ERE(read_actual_spectra_from_file(value));
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || (match_pattern(lc_option,
                           "actual-illumination-spectra-normalization-level"))
         || (match_pattern(lc_option,
                           "actual-illumination-normalization-level"))
         || (match_pattern(lc_option, "actual-illum-normalization-level"))
       )
    {
        double temp;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_actual_illum_normalization_level >= 0.0)
            {
                ERE(pso("actual-illum-normalization-level=%6.1f\n",
                        fs_actual_illum_normalization_level));
            }
            else
            {
                ERE(pso("actual-illum-normalization-level=off\n"));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_actual_illum_normalization_level >= 0.0)
            {
                ERE(pso("Actual illum spectra normalization level is %6.1f.\n",
                        fs_actual_illum_normalization_level));
            }
            else
            {
                ERE(pso("Actual illumination spectra is not normalized.\n"));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_actual_illum_normalization_level = DBL_NOT_SET;
            fs_actual_spectra_version++;
        }
        else
        {
            ERE(ss1d(value, &temp));
            fs_actual_illum_normalization_level = temp;
            fs_actual_spectra_version++;
        }
        result = NO_ERROR;
    }

    if ((lc_option[ 0 ] == '\0') || match_pattern(lc_option, "cisf"))
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("cisf = %s\n",
                    fs_canon_spectra_config_file[ 0 ] ? fs_canon_spectra_config_file
                                                   : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_canon_spectra_sp == NULL)
            {
                ERE(pso("No canonical illuminant spectra has been set.\n"));
            }
            else
            {
                ERE(pso("Canonical illuminant spectra source is %s.\n",
                        fs_canon_spectra_config_file));
            }
        }
        else if (all_digits(value))
        {
            int temp_int;

            ERE(ss1pi(value, &temp_int));
            ERE(set_canon_spectra_by_number(temp_int));
        }
        else if (STRCMP_EQ(value, "#"))
        {
            ERE(set_random_canon_spectra());
        }
        else
        {
            ERE(read_canon_spectra_from_file(value));
        }
        result = NO_ERROR;
    }

    /*
    // This option is relatively useless, except for debugging. All algorithms
    // should be invariant with respect to this option (except the average
    // illuminant option, as of 04/98).
    */
    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "actual-spectra-scale"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            if (fs_actual_spectra_scale <= DBL_EPSILON)
            {
                ERE(pso("actual-spectra-scale = off\n"));
            }
            else
            {
                ERE(pso("actual-spectra-scale = %.2f\n",
                        fs_actual_spectra_scale));
            }
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_actual_spectra_scale <= DBL_EPSILON)
            {
                ERE(pso("Actual spectra scale factor is not being used.\n"));
            }
            else
            {
                ERE(pso("Actual spectra scale factor is %.2f.\n",
                        fs_actual_spectra_scale));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_actual_spectra_scale = DBL_NOT_SET;
            fs_actual_spectra_version++;
        }
        else
        {
            ERE(ss1d(value, &temp_real));
            fs_actual_spectra_scale = temp_real;
            fs_actual_spectra_version++;

        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "specular-reflect-spectra-file")
         || match_pattern(lc_option, "specular-rsf")
       )
    {

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-reflect-spectra-file = %s\n",
                    fs_specular_spectra_config_file[ 0 ]
                                ? fs_specular_spectra_config_file
                                : "off"));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_cached_specular_spectra_sp == NULL)
            {
                ERE(pso("No specular reflectance database has been set.\n"));
            }
            else
            {
                ERE(pso("Specular reflectance spectra database source is %s.\n",
                        fs_specular_spectra_config_file));
            }
        }
        else
        {
            ERE(read_specular_spectra_from_file(value));
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "specular-fraction"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-fraction = %.2f%%\n",
                    100.0 * fs_specular_fraction));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_specular_fraction <= 0.0)
            {
                ERE(pso("No generated pixels are made specular.\n"));
            }
            else
            {
                ERE(pso("The fraction of pixels made specular is %3.0f%%.\n",
                        100.0 * fs_specular_fraction));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_specular_fraction = 0.0;
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < 0.0) temp_real = 0.0;
            fs_specular_fraction = temp_real;

            invalidate_reflectance_with_spec();
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "specular-amount"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-amount = %.2f\n", fs_specular_amount));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The maximum added specularity is %.1f times actual.\n",
                    fs_specular_amount));
        }
        else if (is_no_value_word(value))
        {
            fs_specular_amount = 0.0;
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < 0.0) temp_real = 0.0;
            fs_specular_amount = temp_real;

            invalidate_reflectance_with_spec();
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "highlight-fraction"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("highlight-fraction = %.2f%%\n",
                    100.0 * fs_highlight_fraction));
        }
        else if (value[ 0 ] == '\0')
        {
            if (fs_highlight_fraction <= 0.0)
            {
                ERE(pso("No generated pixels have added highlights.\n"));
            }
            else
            {
                ERE(pso("The fraction of pixels with added highlights is %3.0f%%.\n",
                        100.0 * fs_highlight_fraction));
            }
        }
        else if (is_no_value_word(value))
        {
            fs_highlight_fraction = 0.0;
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < 0.0) temp_real = 0.0;
            fs_highlight_fraction = temp_real;

            invalidate_reflectance_with_spec();
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "highlight-amount"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("highlight-amount = %.2f\n", fs_specular_amount));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The maximum added highlight is %.1f times actual.\n",
                    fs_highlight_amount));
        }
        else if (is_no_value_word(value))
        {
            fs_highlight_amount = 0.0;
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < 0.0) temp_real = 0.0;
            fs_highlight_amount = temp_real;

            invalidate_reflectance_with_spec();
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "white-specular-reflection-fraction"))
        || (match_pattern(lc_option, "specular-white-reflection-fraction"))
       )
    {
        double temp_real;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("specular-white-reflection-fraction = %.1f%%\n",
                    100.0 * fs_white_specular_reflection_fraction));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("The fraction of specular reflection that is white is %3.0f%%.\n",
                    100.0 * fs_white_specular_reflection_fraction));
        }
        else if (is_no_value_word(value))
        {
            fs_white_specular_reflection_fraction = 0.0;
        }
        else
        {
            ERE(ss1d(value, &temp_real));

            if (temp_real < 0.0) temp_real = 0.0;
            fs_white_specular_reflection_fraction = temp_real;

            invalidate_reflectance_with_spec();
        }
        result = NO_ERROR;
    }

    if (    (lc_option[ 0 ] == '\0')
         || match_pattern(lc_option, "fluorescent-database-file")
         || match_pattern(lc_option, "fl-db-file")
         || match_pattern(lc_option, "fluorescent-database")
         || match_pattern(lc_option, "fl-db")
       )
    {
        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {

            ERE(pso("fluorescent-database-file = %s\n",
                    fs_fluorescent_database_config_file[ 0 ] ?
                                        fs_fluorescent_database_config_file
                                        : "off"));
        }
        else if (value[ 0 ] == '\0')
        {

            if (fs_fluorescent_database_config_file[ 0 ] == '\0')
            {
                ERE(pso("No fluorescent database has been set.\n"));
            }
            else
            {
                ERE(pso("Fluorescent database source is %s.\n",
                        fs_fluorescent_database_config_file));
            }
        }
        else
        {
            ERE(read_fluorescent_database_from_file(value));
        }
        result = NO_ERROR;
    }

    if (   (lc_option[ 0 ] == '\0')
        || (match_pattern(lc_option, "fluorescent-weight-factor"))
        || (match_pattern(lc_option, "fl-weight-factor"))
       )
    {
        int temp_int;

        if (value == NULL)
        {
            return NO_ERROR;
        }
        else if (value[ 0 ] == '?')
        {
            ERE(pso("fluorescent-weight-factor = %d\n",
                    fs_fluorescent_weight_factor));
        }
        else if (value[ 0 ] == '\0')
        {
            ERE(pso("Fluorescent surfaces are weighted by a factor of %d.\n",
                    fs_fluorescent_weight_factor));
        }
        else
        {
            ERE(ss1pi(value, &temp_int));

            if (fs_fluorescent_weight_factor != temp_int)
            {

                fs_fluorescent_weight_factor = temp_int;
                fs_world_version++;
                fs_fluorescent_database_version++;
            }
        }
        result = NO_ERROR;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_synthetic_RGB(Matrix** rgb_data_mpp, int illum)
{

    ERE(initialize_illum_spectra());
    ERE(initialize_reflect_spectra());
    (void)initialize_fluorescent_database();

    ASSERT(fs_cached_illum_spectra_sp != NULL);
    ASSERT(fs_cached_reflect_spectra_sp != NULL);

    if ((illum < 0) || (illum >= fs_cached_illum_spectra_sp->spectra_mp->num_rows))
    {
        SET_ARGUMENT_BUG();
        return ERROR;
    }

    ERE(generate_synthetic_RGB_data(rgb_data_mpp, fs_cached_illum_spectra_sp,
                                    illum, fs_cached_reflect_spectra_sp,
                                    fs_cached_fluorescent_database_ptr));

    return update_world_version();
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int generate_synthetic_RGB_data
(
    Matrix**              rgb_data_mpp,
    Spectra*              illum_spectra_sp,
    int                   illum,
    Spectra*              reflect_spectra_sp,
    Fluorescent_database* fl_db_ptr
)
{
    Matrix* rgb_matrix_list[ 2 ] = {NULL, NULL};
    int     result                 = NO_ERROR;
    int     sensor_version         = NOT_SET;

    /* Since minor tweaks revolving to ensure library consistency. */

    if (    (reflect_spectra_sp == NULL)
         && ((fl_db_ptr == NULL) || (fs_fluorescent_weight_factor < 1))
       )
    {
        set_error("No reflectance data is available.");
        return ERROR;
    }

    if (reflect_spectra_sp != NULL)
    {
        sensor_version = generate_RGB_data(&(rgb_matrix_list[ 0 ]),
                                           illum_spectra_sp,
                                           illum, reflect_spectra_sp,
                                           (Spectra*)NULL);

        if (sensor_version == ERROR)
        {
            result = ERROR;
        }
    }

    verbose_pso(30, "%d matte RGB generated.\n",
                rgb_matrix_list[ 0 ] ? rgb_matrix_list[ 0 ]->num_rows : 0);

    if (    (result != ERROR)
         && (fl_db_ptr != NULL)
         && (fs_fluorescent_weight_factor > 0)
       )
    {
        int i;

        verbose_pso(35, "Illum %d sent to generate_fluorescent_RGB_data.\n",
                    illum);
        verbose_pso(35, "Componant 50 of Illum %d is %e.\n",
                    illum, illum_spectra_sp->spectra_mp->elements[illum][50]);

        sensor_version = generate_fluorescent_RGB_data(&(rgb_matrix_list[ 1 ]),
                                                       illum_spectra_sp, illum,
                                                       fl_db_ptr,
                                                       (Spectra*)NULL);
        if (sensor_version == ERROR) result = ERROR;

        verbose_pso(40, "Fluourescent RGB output:\n");

        for (i=0; i<(rgb_matrix_list[ 1 ])->num_rows; i++)
        {
            verbose_pso(40, "     %.1f  %.1f  %.1f\n",
                        (rgb_matrix_list[ 1 ])->elements[ i ][ 0 ],
                        (rgb_matrix_list[ 1 ])->elements[ i ][ 1 ],
                        (rgb_matrix_list[ 1 ])->elements[ i ][ 2 ]);
        }

        if ((result != ERROR) && (fs_fluorescent_weight_factor > 1))
        {
            Matrix** fl_rgb_matrix_list = N_TYPE_MALLOC(Matrix*,
                                                  fs_fluorescent_weight_factor);
            Matrix* temp_mp = NULL;

            if (fl_rgb_matrix_list == NULL)
            {
                result = ERROR;
            }
            else
            {
                for (i=0; i<fs_fluorescent_weight_factor; i++)
                {
                    /*
                    // We living a bit on the edge here, as we are just copying
                    // pointers (this is why we need temp_mp).
                    */
                    fl_rgb_matrix_list[ i ] = rgb_matrix_list[ 1 ];
                }
                result = concat_matrices_vertically(&temp_mp,
                                                    fs_fluorescent_weight_factor,
                                                    fl_rgb_matrix_list);

                if (result != ERROR)
                {
                    copy_matrix(&(rgb_matrix_list[ 1 ]), temp_mp);
                }
            }
            free_matrix(temp_mp);
            kjb_free(fl_rgb_matrix_list);
        }

        verbose_pso(30, "%d fluorescent RGB generated.\n",
                    rgb_matrix_list[ 1 ] ? rgb_matrix_list[ 1 ]->num_rows
                                          : 0);
    }

    if (result != ERROR)
    {
        int     concat_res;
        concat_res = concat_matrices_vertically(rgb_data_mpp, 2,
                                                rgb_matrix_list);

        if (concat_res == ERROR) result = ERROR;
    }

    if (result != ERROR)
    {
        verbose_pso(30, "%d RGB generated.\n", (*rgb_data_mpp)->num_rows);
    }

    free_matrix(rgb_matrix_list[ 0 ]);
    free_matrix(rgb_matrix_list[ 1 ]);

    if (result == ERROR)
    {
        return result;
    }
    else
    {
        ASSERT(IS_SET(sensor_version));
        return sensor_version;
    }
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int world_is_up_to_date(int proposed_world_version)
{

    ERE(update_world_version());

    return (proposed_world_version == fs_world_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int update_world_version(void)
{
    static int cached_sensor_version = NOT_SET;
    int        sensor_version;

    ERE(sensor_version = get_sensor_version());

    if (sensor_version != cached_sensor_version)
    {
        cached_sensor_version = sensor_version;
        fs_world_version++;
    }

    return fs_world_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_illum_spectra_from_file(const char* file_name)
{
    Spectra* illum_spectra_sp = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_spectra_from_gamut_config_file(&illum_spectra_sp,
                                            (const char*)NULL, file_name,
                                            "illum spectra",
                                            config_file, sizeof(config_file)));

    result = set_illum_spectra(illum_spectra_sp, config_file);

    free_spectra(illum_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_reflect_spectra_from_file(const char* file_name)
{
    Spectra* reflect_spectra_sp = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_spectra_from_gamut_config_file(&reflect_spectra_sp,
                                            (const char*)NULL,
                                            file_name, "reflect spectra",
                                            config_file, sizeof(config_file)));

    result = set_reflect_spectra(reflect_spectra_sp, config_file);

    free_spectra(reflect_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_specular_spectra_from_file(const char* file_name)
{
    Spectra* specular_spectra_sp = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_spectra_from_gamut_config_file(&specular_spectra_sp,
                                            (const char*)NULL,
                                            file_name,
                                            "specular reflect spectra",
                                            config_file, sizeof(config_file)));

    result = set_specular_spectra(specular_spectra_sp, config_file);

    free_spectra(specular_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_canon_spectra_from_file(const char* file_name)
{
    Spectra* canon_spectra_sp = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_spectra_from_gamut_config_file(&canon_spectra_sp,
                                            (const char*)NULL,
                                            file_name, "canon illum spectra",
                                            config_file, sizeof(config_file)));

    if (    (canon_spectra_sp != NULL)
         && (canon_spectra_sp->spectra_mp->num_rows != 1))
    {
        set_error("Canon illuminant spectra file %q has more than one spectra.",
                  file_name);
        result = ERROR;
    }
    else
    {
        result = set_canon_spectra(canon_spectra_sp, config_file);
    }

    free_spectra(canon_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_actual_spectra_from_file(const char* file_name)
{
    Spectra* actual_spectra_sp = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_spectra_from_gamut_config_file(&actual_spectra_sp,
                                            (const char*)NULL,
                                            file_name, "actual illum spectra",
                                            config_file, sizeof(config_file)));

    if (    (actual_spectra_sp != NULL)
         && (actual_spectra_sp->spectra_mp->num_rows != 1))
    {
        set_error(
                 "Actual illuminant spectra file %q has more than one spectra.",
                 file_name);
        result = ERROR;
    }
    else
    {
        result = set_actual_spectra(actual_spectra_sp, config_file);
    }

    free_spectra(actual_spectra_sp);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int read_fluorescent_database_from_file(const char* file_name)
{
    Fluorescent_database* fl_db_ptr = NULL;
    char     config_file[ MAX_FILE_NAME_SIZE ];
    int      result;

    ERE(read_fl_db_from_gamut_config_file(&fl_db_ptr, (const char*)NULL,
                                           file_name, "fluorescent database",
                                           config_file, sizeof(config_file)));

    result = set_fluorescent_database(fl_db_ptr, config_file);

    free_fluorescent_database(fl_db_ptr);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_illum_spectra
(
    Spectra* new_illum_spectra_sp,
    char*    config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_illum_spectra_version++;

    ERE(copy_spectra(&fs_cached_illum_spectra_sp, new_illum_spectra_sp));
    BUFF_CPY(fs_illum_spectra_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_reflect_spectra
(
    Spectra* new_reflect_spectra_sp,
    char*    config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_reflect_spectra_version++;
    fs_world_version++;

    ERE(copy_spectra(&fs_cached_reflect_spectra_sp, new_reflect_spectra_sp));
    BUFF_CPY(fs_reflect_spectra_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_specular_spectra
(
    Spectra* new_specular_spectra_sp,
    char*    config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_specular_spectra_version++;

    ERE(copy_spectra(&fs_cached_specular_spectra_sp, new_specular_spectra_sp));
    BUFF_CPY(fs_specular_spectra_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_random_canon_spectra(void)
{

    return set_canon_spectra_by_number(USE_RANDOM);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_canon_spectra_by_number(int number)
{
    Spectra* new_canon_spectra_sp = NULL;
    Vector*  row_vp               = NULL;
    int      result               = NO_ERROR;

    ERE(initialize_illum_spectra());

    if (number > fs_cached_illum_spectra_sp->spectra_mp->num_rows)
    {
        set_error("Can't select illum %d. There are only %d illum.",
                  number, fs_cached_illum_spectra_sp->spectra_mp->num_rows);
        result = ERROR;
    }
    else if (number == USE_RANDOM)
    {
        number = (int)(kjb_rand() * fs_cached_illum_spectra_sp->spectra_mp->num_rows);

        if (number == fs_cached_illum_spectra_sp->spectra_mp->num_rows)
        {
            number--;
        }

        number++;
    }
    else if (number < 1)
    {
        set_error("Illuminant selection must be greater than zero.");
        result = ERROR;
    }

    if (result != ERROR)
    {
        result = get_target_spectra(&new_canon_spectra_sp, 1,
                                 fs_cached_illum_spectra_sp->spectra_mp->num_cols,
                                 fs_cached_illum_spectra_sp->offset,
                                 fs_cached_illum_spectra_sp->step,
                                 fs_cached_illum_spectra_sp->type);
    }

    if (result != ERROR)
    {
        result = get_matrix_row(&row_vp, fs_cached_illum_spectra_sp->spectra_mp,
                                number - 1);
    }

    if (result != ERROR)
    {
        result = put_matrix_row(new_canon_spectra_sp->spectra_mp, row_vp, 0);
    }

    if (result != ERROR)
    {
        char file_name_analog[ 100 ];

        result = kjb_sprintf(file_name_analog, sizeof(file_name_analog),
                             "<illum spectra %d>", number);

        if (result != ERROR)
        {
            result = set_canon_spectra(new_canon_spectra_sp, file_name_analog);
        }
    }

    free_spectra(new_canon_spectra_sp);
    free_vector(row_vp);

    if (result != ERROR)
    {
        result = number;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_canon_spectra
(
    Spectra* new_canon_spectra_sp,
    char*    config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    dbs(config_file);

    fs_canon_spectra_version++;

    ERE(copy_spectra(&fs_cached_canon_spectra_sp, new_canon_spectra_sp));

    BUFF_CPY(fs_canon_spectra_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_random_actual_spectra(void)
{

    return set_actual_spectra_by_number(USE_RANDOM);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int set_actual_spectra_by_number(int number)
{
    Spectra* new_actual_spectra_sp = NULL;
    Vector*  row_vp                = NULL;
    int      result                = NO_ERROR;

    ERE(initialize_illum_spectra());

    if (number > fs_cached_illum_spectra_sp->spectra_mp->num_rows)
    {
        set_error("Can't select illum %d. There are only %d illum.",
                  number, fs_cached_illum_spectra_sp->spectra_mp->num_rows);
        result = ERROR;
    }
    else if (number == USE_RANDOM)
    {
        number = (int)(kjb_rand() * fs_cached_illum_spectra_sp->spectra_mp->num_rows);

        if (number == fs_cached_illum_spectra_sp->spectra_mp->num_rows)
        {
            number--;
        }

        number++;
    }
    else if (number < 1)
    {
        set_error("Illuminant selection must be greater than zero.");
        result = ERROR;
    }

    if (result != ERROR)
    {
        result = get_target_spectra(&new_actual_spectra_sp, 1,
                                 fs_cached_illum_spectra_sp->spectra_mp->num_cols,
                                 fs_cached_illum_spectra_sp->offset,
                                 fs_cached_illum_spectra_sp->step,
                                 fs_cached_illum_spectra_sp->type);
    }

    if (result != ERROR)
    {
        result = get_matrix_row(&row_vp, fs_cached_illum_spectra_sp->spectra_mp,
                                number - 1);
    }

    if (result != ERROR)
    {
        result = put_matrix_row(new_actual_spectra_sp->spectra_mp, row_vp, 0);
    }

    if (result != ERROR)
    {
        char file_name_analog[ 100 ];

        result = kjb_sprintf(file_name_analog, sizeof(file_name_analog),
                             "<illum spectra %d>", number);

        if (result != ERROR)
        {
            result = set_actual_spectra(new_actual_spectra_sp,
                                        file_name_analog);
        }
    }

    free_spectra(new_actual_spectra_sp);
    free_vector(row_vp);

    if (result != ERROR)
    {
        result = number;
    }

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_actual_spectra
(
    Spectra* new_actual_spectra_sp,
    char*    config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_actual_spectra_version++;

    ERE(copy_spectra(&fs_cached_actual_spectra_sp, new_actual_spectra_sp));
    BUFF_CPY(fs_actual_spectra_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int set_fluorescent_database
(
    Fluorescent_database* new_fl_db_ptr,
    char*                 config_file
)
{

#ifdef TRACK_MEMORY_ALLOCATION
    prepare_memory_cleanup();
#endif

    fs_fluorescent_database_version++;
    fs_world_version++;

    ERE(copy_fluorescent_database(&fs_cached_fluorescent_database_ptr,
                                  new_fl_db_ptr));
    BUFF_CPY(fs_fluorescent_database_config_file, config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_num_illum_spectra(int* num_illum_spectra_ptr)
{

    ERE(initialize_illum_spectra());
    *num_illum_spectra_ptr = fs_cached_illum_spectra_sp->spectra_mp->num_rows;

    return fs_illum_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_illum_spectra(Spectra** target_sp_ptr)
{

    ERE(initialize_illum_spectra());
    ERE(copy_spectra(target_sp_ptr, fs_cached_illum_spectra_sp));

    return fs_illum_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_illum_spectra(void)
{
    static int      first_time = TRUE;
    static int      previous_illum_spectra_version = NOT_SET;
    static double     previous_sensor_version = NOT_SET;
    int             sensor_version;

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

    if (fs_cached_illum_spectra_sp == NULL)
    {
        if (first_time)
        {
            first_time = FALSE;
            (void)read_illum_spectra_from_file(ILLUM_SPECTRA_FILE);
        }
        else
        {
            set_error("No illuminant spectra database is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }

    if (fs_cached_illum_spectra_sp == NULL)
    {
        insert_error("No illuminant spectra database is available.");
        return ERROR;
    }

    sensor_version = get_sensor_version();

    if (    (previous_illum_spectra_version != fs_illum_spectra_version)
         || (previous_sensor_version != sensor_version)
       )
    {
        verbose_pso(5, "Re-normalizing illum spectra.\n");

        ERE(ow_normalize_illum_spectra(fs_cached_illum_spectra_sp,
                                       (Spectra*)NULL, 255.0));
    }

    previous_illum_spectra_version = fs_illum_spectra_version;
    previous_sensor_version = sensor_version;

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_reflect_spectra_with_specularities(Spectra** target_sp_ptr)
{

    ERE(make_reflect_spectra_with_specularities());
    ERE(copy_spectra(target_sp_ptr, fs_cached_reflect_spectra_with_spec_sp));

    return fs_reflect_spectra_with_spec_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int make_reflect_spectra_with_specularities(void)
{
    static int previous_reflect_spectra_version           = NOT_SET;
    static int previous_specular_spectra_version          = NOT_SET;
    static int previous_reflect_spectra_with_spec_version = NOT_SET;

    if (    (fs_cached_reflect_spectra_with_spec_sp != NULL)
         && (previous_reflect_spectra_with_spec_version == fs_reflect_spectra_with_spec_version)
         && (previous_reflect_spectra_version == fs_reflect_spectra_version)
         && (previous_specular_spectra_version == fs_specular_spectra_version)
       )
    {
        verbose_pso(10, "Specular reflectance is up to date. Skipping make.\n");
    }
    else
    {
        Spectra* specular_spectra_sp = NULL;

        verbose_pso(5, "Specular reflectance is out of date. Re-making it.\n");

        ERE(get_reflect_spectra(&fs_cached_reflect_spectra_with_spec_sp));

        if (fs_white_specular_reflection_fraction < 1.0)
        {
            ERE(get_specular_spectra(&specular_spectra_sp));
            ERE(check_spectra_are_comparable(specular_spectra_sp,
                                    fs_cached_reflect_spectra_with_spec_sp));
            ERE(ow_scale_matrix_rows_by_max(specular_spectra_sp->spectra_mp));
        }

        if (    (fs_specular_fraction > 0.0)
             || (fs_highlight_fraction > 0.0)
           )
        {
            Matrix* spectra_mp = fs_cached_reflect_spectra_with_spec_sp->spectra_mp;
            int     num_rows   = spectra_mp->num_rows;
            int     num_cols   = spectra_mp->num_cols;
            int     i, j;
            Vector* specular_reflect_vp = NULL;

            verbose_pso(5, "Specular or highlight fraction is non-zero. ");
            verbose_pso(5, "Adding specularities.\n");

            for (i=0; i<num_rows; i++)
            {
                if (kjb_rand() <= fs_white_specular_reflection_fraction)
                {
                    ERE(get_initialized_vector(&specular_reflect_vp,
                                               spectra_mp->num_cols, 1.0));
                }
                else
                {
                    ERE(get_random_matrix_row(&specular_reflect_vp,
                                              specular_spectra_sp->spectra_mp));

                }

                if (kjb_rand() < fs_highlight_fraction)
                {
                    double random_factor = gauss_rand();
                    double extra = fs_highlight_amount * ABS_OF(random_factor);

                    for (j=0; j<num_cols; j++)
                    {
                        ASSERT(extra >= 0.0);
                        ASSERT(spectra_mp->elements[ i ][ j ] >= 0.0);

                        spectra_mp->elements[ i ][ j ] +=
                                    extra * specular_reflect_vp->elements[ j ];

                        ASSERT(spectra_mp->elements[ i ][ j ] >= 0.0);
                    }
                }
                if (kjb_rand() < fs_specular_fraction)
                {
                    double random_factor = kjb_rand();
                    double extra = fs_specular_amount * ABS_OF(random_factor);

                    for (j=0; j<num_cols; j++)
                    {
                        ASSERT(extra >= 0.0);
                        ASSERT(spectra_mp->elements[ i ][ j ] >= 0.0);

                        spectra_mp->elements[ i ][ j ] +=
                                    extra * specular_reflect_vp->elements[ j ];

                        ASSERT(spectra_mp->elements[ i ][ j ] >= 0.0);
                    }
                }
            }

            free_vector(specular_reflect_vp);
        }

        fs_reflect_spectra_with_spec_version++;
        fs_world_version++;

        previous_reflect_spectra_with_spec_version =
                                           fs_reflect_spectra_with_spec_version;
        previous_reflect_spectra_version = fs_reflect_spectra_version;
        previous_specular_spectra_version = fs_specular_spectra_version;

        verbose_pso(10, "Specular reflectance remade.\n");

        free_spectra(specular_spectra_sp);
    }

    return fs_reflect_spectra_with_spec_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

void invalidate_reflectance_with_spec(void)
{

    free_spectra(fs_cached_reflect_spectra_with_spec_sp);
    fs_cached_reflect_spectra_with_spec_sp = NULL;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_reflect_spectra(Spectra** target_sp_ptr)
{

    ERE(initialize_reflect_spectra());
    ERE(copy_spectra(target_sp_ptr, fs_cached_reflect_spectra_sp));

    return fs_reflect_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_reflect_spectra(void)
{

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

    if (fs_cached_reflect_spectra_sp == NULL)
    {
        if (fs_reflect_spectra_version == 0)
        {
            (void)read_reflect_spectra_from_file(REFLECT_SPECTRA_FILE);
        }
        else
        {
            set_error("No reflectance spectra is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }

    if (fs_cached_reflect_spectra_sp == NULL)
    {
        insert_error("No reflectance spectra is available.");
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_specular_spectra(Spectra** target_sp_ptr)
{

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

#ifdef PROVIDE_DEFAULT_SPECULAR_REFLECT_SPECTRA_FILE
    if (fs_cached_specular_spectra_sp == NULL)
    {
        if (fs_specular_spectra_version == 0)
        {
            read_specular_spectra_from_file(SPECULAR_REFLECT_SPECTRA_FILE);
        }
        else
        {
            set_error("No specular reflectance spectra is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }
#endif

    if (fs_cached_specular_spectra_sp == NULL)
    {
        insert_error("No specular reflectance spectra is available.");
        return ERROR;
    }

    ERE(copy_spectra(target_sp_ptr, fs_cached_specular_spectra_sp));

    return fs_specular_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_canon_spectra(Spectra** target_sp_ptr)
{
    static int previous_canon_spectra_version = NOT_SET;
    static int previous_sensor_version        = NOT_SET;
    int        sensor_version;

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

    if (fs_cached_canon_spectra_sp == NULL)
    {
        if (fs_canon_spectra_version == 0)
        {
            (void)read_canon_spectra_from_file(CANON_ILLUM_SPECTRA_FILE);
        }
        else
        {
            set_error("No canoncical illuminant spectra is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }

    if (fs_cached_canon_spectra_sp == NULL)
    {
        insert_error("No canonical illuminant spectra is available.");
        return ERROR;
    }

    sensor_version = get_sensor_version();

    if (    (previous_canon_spectra_version != fs_canon_spectra_version)
         || (previous_sensor_version != sensor_version)
       )
    {
        verbose_pso(5, "Re-normalizing canon spectra.\n");

        ERE(ow_normalize_illum_spectra(fs_cached_canon_spectra_sp,
                                       (Spectra*)NULL, 255.0));
    }

    ERE(copy_spectra(target_sp_ptr, fs_cached_canon_spectra_sp));

    previous_canon_spectra_version = fs_canon_spectra_version;
    previous_sensor_version = sensor_version;

    return fs_canon_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_actual_spectra(Spectra** target_sp_ptr)
{
    static int    previous_actual_spectra_version = NOT_SET;
    static double previous_actual_spectra_scale   = DBL_NOT_SET;
    static int    previous_sensor_version         = NOT_SET;
    /*
    // Technically a value of DBL_NOT_SET would work, but DBL_NOT_SET is a valid
    // value for fs_actual_illum_normalization_level, so to keep things
    // absolutely clear, initialize previous_actual_normalization_level to some
    // value which fs_actual_illum_normalization_level can never have in order
    // to force the normalization the first time.
    */
    static double   previous_actual_normalization_level = 2.0 * DBL_NOT_SET;
    int             sensor_version;

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

    if (fs_cached_actual_spectra_sp == NULL)
    {
        if (fs_actual_spectra_version == 0)
        {
            (void)read_actual_spectra_from_file(ACTUAL_ILLUM_SPECTRA_FILE);
        }
        else
        {
            set_error("No actual illuminant spectra is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }

    if (fs_cached_actual_spectra_sp == NULL)
    {
        insert_error("No actual (test) illuminant spectra is available.");
        return ERROR;
    }

    sensor_version = get_sensor_version();

    if (    (previous_actual_spectra_version != fs_actual_spectra_version)
         || (previous_sensor_version != sensor_version)
         || ( ! IS_EQUAL_DBL(previous_actual_spectra_scale,
                              fs_actual_spectra_scale))
         || ( ! IS_EQUAL_DBL(previous_actual_normalization_level,
                             fs_actual_illum_normalization_level))
         /*
         // Likely not necessary, if the previous_actual_normalization_level
         // initialization is correct.
         */
         || (fs_cached_normalized_actual_spectra_sp == NULL)
       )
    {
        verbose_pso(5, "Re-normalizing actual spectra.\n");

        ERE(normalize_illum_spectra(&fs_cached_normalized_actual_spectra_sp,
                                    fs_cached_actual_spectra_sp,
                                    (Spectra*)NULL,
                                    fs_actual_illum_normalization_level));
        /*
        // This option is relatively useless, except for debugging. All
        // algorithms should be invariant with respect to this option (except
        // the average illuminant option, as of 04/98).
        */
        if (fs_actual_spectra_scale > DBL_EPSILON)
        {
            double factor = fs_actual_spectra_scale * kjb_rand();

            ERE(ow_multiply_matrix_by_scalar(
                                fs_cached_normalized_actual_spectra_sp->spectra_mp,
                                factor));
        }
    }

    ERE(copy_spectra(target_sp_ptr, fs_cached_normalized_actual_spectra_sp));

    previous_actual_normalization_level = fs_actual_illum_normalization_level;
    previous_actual_spectra_scale = fs_actual_spectra_scale;
    previous_actual_spectra_version = fs_actual_spectra_version;
    previous_sensor_version = sensor_version;

    return fs_actual_spectra_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_optional_fluorescent_database
(
    Fluorescent_database** fl_db_ptr_ptr
)
{

    (void)initialize_fluorescent_database();

    ERE(copy_fluorescent_database(fl_db_ptr_ptr,
                                  fs_cached_fluorescent_database_ptr));

    return fs_fluorescent_database_version;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

static int initialize_fluorescent_database(void)
{

    /*
    // Clear the error since there is an execution path which uses
    // "insert_error" without a previous "set_error". Although it should be the
    // case that if this path is taken, then the error should be set, we are
    // better off not taking any chances.
    */
    kjb_clear_error();

#ifdef PROVIDE_DEFAULT_FLUORESCENT_DATABASE_FILE
    if (fs_cached_fluorescent_database_ptr == NULL)
    {
        if (fs_fluorescent_database_version == 0)
        {
            read_fluorescent_database_from_file(FLUORESCENT_DATABASE_FILE);
        }
        else
        {
            set_error("No fluorescent database is currently cached.");
            add_error("No attempt to read default, as this is only done the ");
            cat_error("first time it is sought.");
        }
    }
#endif

    if (fs_cached_fluorescent_database_ptr == NULL)
    {
        insert_error("No fluorescent database is available.");
        return ERROR;
    }

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int illum_spectra_is_up_to_date(int proposed_version)
{

    return (proposed_version == fs_illum_spectra_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int reflect_spectra_is_up_to_date(int proposed_version)
{

    if (fs_cached_reflect_spectra_sp == NULL) return FALSE;
    return (proposed_version == fs_reflect_spectra_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int specular_spectra_is_up_to_date(int proposed_version)
{

    if (fs_cached_specular_spectra_sp == NULL) return FALSE;
    return (proposed_version == fs_specular_spectra_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int reflect_spectra_with_spec_is_up_to_date(int proposed_version)
{

    if (fs_cached_reflect_spectra_with_spec_sp == NULL) return FALSE;
    return (proposed_version == fs_reflect_spectra_with_spec_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int actual_spectra_is_up_to_date(int proposed_version)
{

    return (proposed_version == fs_actual_spectra_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int canon_spectra_is_up_to_date(int proposed_version)
{

    return (proposed_version == fs_canon_spectra_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int fluorescent_db_is_up_to_date(int proposed_version)
{

    return (proposed_version == fs_fluorescent_database_version);
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int get_current_world_info(World_info* world_info_ptr)
{

    BUFF_CPY(world_info_ptr->illum_spectra_config_file,
             fs_illum_spectra_config_file);
    BUFF_CPY(world_info_ptr->reflect_spectra_config_file,
             fs_reflect_spectra_config_file);
    BUFF_CPY(world_info_ptr->canon_spectra_config_file,
             fs_canon_spectra_config_file);
    BUFF_CPY(world_info_ptr->specular_spectra_config_file,
             fs_specular_spectra_config_file);
    BUFF_CPY(world_info_ptr->fluorescent_database_config_file,
             fs_fluorescent_database_config_file);

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int echo_world_info(World_info* world_info_ptr)
{

    ERE(pso("     illum_spectra_config_file = %s\n",
            world_info_ptr->illum_spectra_config_file[ 0 ]
            ?    world_info_ptr->illum_spectra_config_file
            :    "<none>"));

    ERE(pso("     reflect_spectra_config_file = %s\n",
            world_info_ptr->reflect_spectra_config_file[ 0 ]
            ?    world_info_ptr->reflect_spectra_config_file
            :    "<none>"));

    ERE(pso("     canon_spectra_config_file = %s\n",
            world_info_ptr->canon_spectra_config_file[ 0 ]
            ?    world_info_ptr->canon_spectra_config_file
            :    "<none>"));

    ERE(pso("     specular_spectra_config_file = %s\n",
            world_info_ptr->specular_spectra_config_file[ 0 ]
            ?    world_info_ptr->specular_spectra_config_file
            :    "<none>"));

    ERE(pso("     fluorescent_database_config_file = %s\n",
            world_info_ptr->fluorescent_database_config_file[ 0 ]
            ?    world_info_ptr->fluorescent_database_config_file
            :    "<none>"));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int write_world_info(FILE* fp, World_info* world_info_ptr)
{

    ERE(kjb_fprintf(fp, "#! illum_spectra_config_file = %s\n",
                    world_info_ptr->illum_spectra_config_file[ 0 ]
                    ?   world_info_ptr->illum_spectra_config_file
                    :   "<none>"));

    ERE(kjb_fprintf(fp, "#! reflect_spectra_config_file = %s\n",
                    world_info_ptr->reflect_spectra_config_file[ 0 ]
                    ?    world_info_ptr->reflect_spectra_config_file
                    :    "<none>"));

    ERE(kjb_fprintf(fp, "#! canon_spectra_config_file = %s\n",
                    world_info_ptr->canon_spectra_config_file[ 0 ]
                    ?    world_info_ptr->canon_spectra_config_file
                    :    "<none>"));

    ERE(kjb_fprintf(fp, "#! specular_spectra_config_file = %s\n",
                    world_info_ptr->specular_spectra_config_file[ 0 ]
                    ?    world_info_ptr->specular_spectra_config_file
                    :    "<none>"));

    ERE(kjb_fprintf(fp, "#! fluorescent_database_config_file = %s\n",
                    world_info_ptr->fluorescent_database_config_file[ 0 ]
                    ?    world_info_ptr->fluorescent_database_config_file
                    :    "<none>"));

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int read_world_info(FILE* fp, World_info* world_info_ptr)
{
    char** option_list;
    char** value_list;
    char*  value;
    char*  option;
    int    num_options;
    int    i;
    long save_file_pos;
    int result = NO_ERROR;
    Error_action save_error_action;

    ERE(initialize_world_info(world_info_ptr));

    ERE(save_file_pos = kjb_ftell(fp));

    while ((num_options = read_header_options(fp, &option_list, &value_list))
           != EOF)
    {
        if (num_options == ERROR)
        {
            result = ERROR;
            break;
        }

        for (i=0; i<num_options; i++)
        {
            option = option_list[ i ];
            value  = value_list[ i ];

            if (STRCMP_EQ(option, "illum_spectra_config_file"))
            {
                BUFF_CPY(world_info_ptr->illum_spectra_config_file, value);
            }
            else if (STRCMP_EQ(option, "reflect_spectra_config_file"))
            {
                BUFF_CPY(world_info_ptr->reflect_spectra_config_file, value);
            }
            else if (STRCMP_EQ(option, "canon_spectra_config_file"))
            {
                BUFF_CPY(world_info_ptr->canon_spectra_config_file, value);
            }
            else if (STRCMP_EQ(option, "luorescent_database_config_file"))
            {
                BUFF_CPY(world_info_ptr->fluorescent_database_config_file,
                         value);
            }
        }

        free_options(num_options, option_list, value_list);
    }

    save_error_action = get_error_action();
    set_error_action(FORCE_ADD_ERROR_ON_ERROR);
    if (kjb_fseek(fp, save_file_pos, SEEK_SET) == ERROR) result = ERROR;
    set_error_action(save_error_action);

    return result;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

int initialize_world_info(World_info* world_info_ptr)
{

    world_info_ptr->illum_spectra_config_file[ 0 ] = '\0';
    world_info_ptr->reflect_spectra_config_file[ 0 ] = '\0';
    world_info_ptr->canon_spectra_config_file[ 0 ] = '\0';
    world_info_ptr->specular_spectra_config_file[ 0 ] = '\0';
    world_info_ptr->fluorescent_database_config_file[ 0 ] = '\0';

    return NO_ERROR;
}

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION
static void prepare_memory_cleanup(void)
{
    static int first_time = TRUE;

    if (first_time)
    {
        add_cleanup_function(free_allocated_static_data);
        first_time = FALSE;
    }
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

#ifdef TRACK_MEMORY_ALLOCATION

static void free_allocated_static_data(void)
{

    free_spectra(fs_cached_reflect_spectra_with_spec_sp);
    free_spectra(fs_cached_specular_spectra_sp);
    free_spectra(fs_cached_reflect_spectra_sp);
    free_spectra(fs_cached_illum_spectra_sp);
    free_spectra(fs_cached_canon_spectra_sp);
    free_spectra(fs_cached_actual_spectra_sp);
    free_spectra(fs_cached_normalized_actual_spectra_sp);
    free_fluorescent_database(fs_cached_fluorescent_database_ptr);
}

#endif

/*  /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\ /\   */

